theory Environ
imports Main Infinite_Set

begin

types name = "nat"

lemma fresh_exists:
  fixes L :: "name set"
  assumes "finite L"
  shows "\<exists>x. x \<notin> L"
using prems by (auto simp add: ex_new_if_finite)

constdefs dom :: "(name * 'a) set \<Rightarrow> name set"
  "dom \<delta> \<equiv> { fst p | p. p \<in> \<delta> }"

lemma dom_insert[simp]:
  shows "dom (insert p \<delta>) = insert (fst p) (dom \<delta>)"
by (cases p, auto simp add: dom_def)

lemma dom_finite:
  assumes "finite \<delta>"
  shows "finite (dom \<delta>)"
using prems proof induct
  case empty thus ?case by (simp add: dom_def)
qed auto

syntax
  in_env :: "(name * 'a) set \<Rightarrow> name \<Rightarrow> bool"    ("_ defines _" [80,80] 81)

translations
  "\<delta> defines x" \<rightleftharpoons> "x \<in> dom \<delta>"

lemma defines_sanity:
  assumes "\<delta> defines x"
  shows "\<exists>y. (x, y) \<in> \<delta>"
using prems by (auto simp add: dom_def)

lemma env_permute:
  shows "insert (x, bx) (insert (y, by) \<delta>) = insert (y, by) (insert (x, bx) \<delta>)"
by auto

constdefs without :: "(name * 'a) set \<Rightarrow> name \<Rightarrow> (name * 'a) set"    (infixl "\<bar>" 80)
  "\<delta>\<bar>x \<equiv> { p | p. p \<in> \<delta> \<and> fst p \<noteq> x }"

end
