theory MemTyping
imports Typing Semantics

begin 
inductive memTyping :: "env \<Rightarrow> ty_env \<Rightarrow> store \<Rightarrow> env \<Rightarrow> bool"      ("_;_ \<Turnstile>' _:_" [50,50,50,50] 50)
  where
    mt_nil[intro!]   : "\<lbrakk> \<Delta> \<Turnstile>' TY \<rbrakk> \<Longrightarrow> \<Gamma>;\<Delta> \<Turnstile>' {} : {}"
  | mt_cons[intro!]  : "\<And>\<Gamma> \<mu> M x \<tau> l.
                        \<lbrakk> \<Gamma>;\<Delta> \<Turnstile>' \<mu> : M; \<Gamma> \<sim> {x\<rhd>\<tau>}; \<Gamma>,(x:\<tau>);\<Delta> \<Turnstile> OK; \<not> \<mu> defines l; 
                          M;\<Delta> \<Turnstile> OK; \<mu> \<Turnstile> Mem \<rbrakk> \<Longrightarrow>
                          \<Gamma>,(x:\<tau>);\<Delta> \<Turnstile>' \<mu>,(l\<mapsto>TmFreeVar x) : M,(l:\<tau>)"

inductive_cases mt_nil_elim [elim!]:  "\<Gamma>;\<Delta> \<Turnstile>' {} : {}"
inductive_cases mt_cons_elim [elim!]: "\<Gamma>;\<Delta> \<Turnstile>' \<mu>,(l\<mapsto>TmFreeVar x) : M,(x:\<tau>)"

lemma memTyping_dom_equal:
  assumes "\<Gamma>;\<Delta> \<Turnstile>' M : \<mu>"
  shows "dom M = dom \<mu>"
using prems by (induct, auto simp add: dom_def)

lemma mem_dom_in:
  assumes "\<Gamma>;\<Delta> \<Turnstile>' \<mu> : M"
  and "M defines l"
  shows "\<mu> defines l"
using prems proof induct
  case mt_cons
  thus ?case by (cases "z=l", auto simp add: dom_def)
qed (auto simp add: dom_def)

lemma mem_dom_not_in:
  assumes "\<Gamma>;\<Delta> \<Turnstile>' \<mu> : M"
  and "\<not> \<mu> defines l"
  shows "\<not> M defines l"
using prems by (auto simp add: mem_dom_in)

lemma memTyping_wf_mem:
  assumes "\<Gamma>;\<Delta> \<Turnstile>' \<mu> : M"
  shows "\<mu> \<Turnstile> Mem"
using prems proof induct
  case mt_nil thus ?case by (auto simp add: wf_mem_def)
next
  case (mt_cons M \<Gamma> \<mu> \<tau> l x)
  thus ?case using wf_mem_cons by auto
qed

lemma memTyping_wf_loc:
  assumes "\<Gamma>;\<Delta> \<Turnstile>' \<mu> : M"
  shows "M;\<Delta> \<Turnstile> OK"
using prems proof induct
  case (mt_cons \<Delta> \<Gamma> \<mu> M x \<tau> l)
  have "\<not> M defines l" using mt_cons mem_dom_not_in by auto
  moreover
  have "\<Delta> \<Turnstile>' \<tau>" using mt_cons poly_env_wf_ty by auto
  ultimately
  show ?case using mt_cons by (auto simp add: wf_env_cons)
qed (auto simp add: wf_env_def)

lemma memTyping_weakening:
  assumes "\<Gamma>;\<Delta> \<Turnstile>' \<mu> : M"
  and "\<Gamma>,(x:\<tau>);\<Delta> \<Turnstile> OK"
  and "\<Gamma> \<sim> {x\<rhd>\<tau>}"
  shows "\<Gamma>,(x:\<tau>);\<Delta> \<Turnstile>' \<mu> : M"
using prems proof induct
  case (mt_cons \<Delta> \<Gamma> \<mu> M z \<tau>z lz)
  thus ?case proof (cases "z = x")
    case True
    have "\<Gamma>,(z:\<tau>z),(x:\<tau>);\<Delta> \<Turnstile> OK" by assumption
    hence "\<tau> = \<tau>z" using True wf_env_def[of "\<Gamma>,(z:\<tau>z),(x:\<tau>)"] by blast
    thus ?thesis using prems by auto
  next
    case False
    have "\<Gamma>,(x:\<tau>),(z:\<tau>z);\<Delta> \<Turnstile>' \<mu>,(lz\<mapsto>TmFreeVar z):M,(lz:\<tau>z)" 
      using prems proof auto
      have "\<Gamma>,(x:\<tau>);\<Delta> \<Turnstile> OK" 
        using prems wf_env_subset[of "\<Gamma>,(z:\<tau>z),(x:\<tau>)" \<Delta> "\<Gamma>,(x:\<tau>)"] by auto
      moreover
      have "\<Gamma> \<sim> {(x\<rhd>\<tau>)}" using prems by (auto simp add: env_compat_def)
      ultimately
      show "\<Gamma>,(x:\<tau>);\<Delta> \<Turnstile>' \<mu>:M" using prems by auto
    next
      show "\<Gamma>,(x:\<tau>) \<sim> {(z, Term \<tau>z)}"
        using prems by (auto simp add: env_compat_def)
    next
      show "\<Gamma>,(x:\<tau>),(z:\<tau>z);\<Delta> \<Turnstile> OK" using prems by (auto simp add: env_permute)
    qed
    thus ?thesis by (auto simp add: env_permute)
  qed
qed auto

lemma memTyping_elim_qU:
  assumes "\<Gamma>;\<Delta> \<Turnstile>' \<mu> : M"
  and "(l\<mapsto>TmFreeVar x) \<in> \<mu>"
  and "(l\<rhd>\<tau>) \<in> M"
  and "qualT \<tau> = qU"
  shows "\<Gamma>;\<Delta> \<Turnstile>' \<mu>\<bar>l : M\<bar>l \<and> (x\<rhd>\<tau>) \<in> \<Gamma>"
using prems proof induct
  case (mt_cons \<Delta> \<Gamma> \<mu> M xz \<tau>z lz)
  have H\<tau>z: "\<Delta> \<Turnstile>' \<tau>z" using mt_cons poly_env_wf_ty by auto
  have Hnotinm: "\<not> M defines lz" using mt_cons mem_dom_not_in by auto
  hence Hmok: "M,(lz:\<tau>z);\<Delta> \<Turnstile> OK" using mt_cons wf_env_cons H\<tau>z by auto
  have Hmt: "\<mu>,(lz\<mapsto>TmFreeVar xz) \<Turnstile> Mem" using mt_cons wf_mem_cons by auto
  show ?case proof (cases "l = lz")
    case True
    hence Hteq: "\<tau> = \<tau>z" using Hmok mt_cons wf_env_def[of "M,(lz:\<tau>z)"] by fast
    have Hxeq: "x = xz" 
      using Hmt True mt_cons wf_mem_def[of "\<mu>,(lz\<mapsto>TmFreeVar xz)"] by fast
    have Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>,(xz:\<tau>z)" using Hteq Hxeq by auto
    have "\<Gamma>;\<Delta> \<Turnstile>' \<mu>:M" by assumption
    also have "\<mu> = \<mu>,(lz\<mapsto>TmFreeVar xz)\<bar>l" using mt_cons True
      by (auto simp add: without_def dom_def)
    also have "M = M,(lz:\<tau>z)\<bar>l" using Hnotinm True 
      by (auto simp add: without_def dom_def)
    finally have "\<Gamma>,(xz:\<tau>z);\<Delta> \<Turnstile>' \<mu>,(lz\<mapsto>TmFreeVar xz)\<bar>l : M,(lz:\<tau>z)\<bar>l"
      using memTyping_weakening mt_cons by auto
    thus ?thesis using Hin by auto
  next
    case False
    hence "(l\<mapsto>TmFreeVar x) \<in> \<mu>" and "(l\<rhd>\<tau>) \<in> M" and "qualT \<tau> = qU" 
      using prems by auto
    hence Hmt': "\<Gamma>;\<Delta> \<Turnstile>' \<mu>\<bar>l : M\<bar>l" and Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>" using mt_cons.hyps 
      by auto
    have "\<Gamma>,(xz:\<tau>z);\<Delta> \<Turnstile>' (\<mu>\<bar>l),(lz\<mapsto>TmFreeVar xz) : (M\<bar>l),(lz:\<tau>z)" proof
      show "\<Gamma>;\<Delta> \<Turnstile>' \<mu>\<bar>l : M\<bar>l" using Hmt' by auto
    next
      show "\<not> \<mu>\<bar>l defines lz" using mt_cons False 
        by (auto simp add: dom_def without_def)
    next
      show "M\<bar>l;\<Delta> \<Turnstile> OK" using mt_cons wf_env_subset[of M  \<Delta> "M\<bar>l"] 
        by (auto simp add: without_def)
    next
      show "\<mu>\<bar>l \<Turnstile> Mem" using wf_mem_subset[of \<mu> "\<mu>\<bar>l"] mt_cons 
        by (auto simp add: without_def)
    qed
    also have "(\<mu>\<bar>l),(lz\<mapsto>TmFreeVar xz) = \<mu>,(lz\<mapsto>TmFreeVar xz)\<bar>l"
      using False by (auto simp add: without_def)
    also have "(M\<bar>l),(lz:\<tau>z) = M,(lz:\<tau>z)\<bar>l" using False 
      by (auto simp add: without_def)
    finally show ?thesis using Hin by auto
  qed
qed auto

lemma memTyping_elim_qL:
  assumes "\<Gamma>;\<Delta> \<Turnstile>' \<mu> : M"
  and "(l\<mapsto>TmFreeVar x) \<in> \<mu>"
  and "(l\<rhd>\<tau>) \<in> M"
  and "qualT \<tau> = qL"
  shows "\<Gamma>\<bar>x;\<Delta> \<Turnstile>' \<mu>\<bar>l : M\<bar>l \<and> (x\<rhd>\<tau>) \<in> \<Gamma>"
using prems proof induct
  case (mt_cons \<Delta> \<Gamma> \<mu> M xz \<tau>z lz)
  have H\<tau>z: "\<Delta> \<Turnstile>' \<tau>z" using mt_cons poly_env_wf_ty by auto
  have Hnotinm: "\<not> M defines lz" using mt_cons mem_dom_not_in by auto
  hence Hmok: "M,(lz:\<tau>z);\<Delta> \<Turnstile> OK" using mt_cons wf_env_cons H\<tau>z by auto
  have Hmt: "\<mu>,(lz\<mapsto>TmFreeVar xz) \<Turnstile> Mem" using mt_cons wf_mem_cons by auto
  show ?case proof (cases "l = lz")
    case True
    hence Hteq: "\<tau> = \<tau>z" using Hmok mt_cons wf_env_def[of "M,(lz:\<tau>z)"] by fast
    have Hxeq: "x = xz" 
      using Hmt True mt_cons wf_mem_def[of "\<mu>,(lz\<mapsto>TmFreeVar xz)"] by fast
    have Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>,(xz:\<tau>z)" using Hteq Hxeq by auto
    have "\<Gamma>;\<Delta> \<Turnstile>' \<mu>:M" by assumption
    also have "\<mu> = \<mu>,(lz\<mapsto>TmFreeVar xz)\<bar>l" using mt_cons True 
      by (auto simp add: without_def dom_def)
    also have "M = M,(lz:\<tau>z)\<bar>l" using Hnotinm True 
      by (auto simp add: without_def dom_def)
    also have "\<not> \<Gamma> defines xz"
      using Hteq compat_ok_sanity_right[of \<Gamma> "{xz\<rhd>\<tau>z}" _ xz \<tau>z] mt_cons
      by auto
    hence "\<Gamma> = \<Gamma>,(xz:\<tau>z)\<bar>x" using Hxeq by (auto simp add: without_def dom_def)
    finally have "\<Gamma>,(xz:\<tau>z)\<bar>x;\<Delta> \<Turnstile>' \<mu>,(lz\<mapsto>TmFreeVar xz)\<bar>l : M,(lz:\<tau>z)\<bar>l"
      using memTyping_weakening mt_cons by auto
    thus ?thesis using Hin by auto
  next
    case False
    hence "(l\<mapsto>TmFreeVar x) \<in> \<mu>" and "(l\<rhd>\<tau>) \<in> M" using prems by auto
    hence Hmt': "\<Gamma>\<bar>x;\<Delta> \<Turnstile>' \<mu>\<bar>l : M\<bar>l" and Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>" using mt_cons by auto
    hence Hxneq: "xz \<noteq> x"
      using mt_cons compat_ok_sanity[of \<Gamma> "{xz\<rhd>\<tau>z}" _ x \<tau>] 
      by (auto simp add: dom_def)
    have "(\<Gamma>\<bar>x),(xz:\<tau>z);\<Delta> \<Turnstile>' (\<mu>\<bar>l),(lz\<mapsto>TmFreeVar xz) : (M\<bar>l),(lz:\<tau>z)" proof
      show "\<Gamma>\<bar>x;\<Delta> \<Turnstile>' \<mu>\<bar>l : M\<bar>l" using Hmt' by auto
    next
      show "\<not> \<mu>\<bar>l defines lz" using mt_cons False
        by (auto simp add: dom_def without_def)
    next
      show "M\<bar>l;\<Delta> \<Turnstile> OK" using mt_cons wf_env_subset[of M _ "M\<bar>l"] 
        by (auto simp add: without_def)
    next
      show "\<mu>\<bar>l \<Turnstile> Mem" using wf_mem_subset[of \<mu> "\<mu>\<bar>l"] mt_cons 
        by (auto simp add: without_def)
    next
      show "\<Gamma>\<bar>x \<sim> {(xz\<rhd>\<tau>z)}" using mt_cons 
        by (auto simp add: without_def env_compat_def)
    next
      have "\<Gamma>,(xz:\<tau>z)\<bar>x;\<Delta> \<Turnstile> OK"
        using mt_cons wf_env_subset[of "\<Gamma>,(xz:\<tau>z)" _ "\<Gamma>,(xz:\<tau>z)\<bar>x"]
        by (auto simp add: without_def)
      also have "\<Gamma>,(xz:\<tau>z)\<bar>x = (\<Gamma>\<bar>x),(xz:\<tau>z)" using Hxneq 
        by (auto simp add: without_def dom_def)
      finally show "(\<Gamma>\<bar>x),(xz:\<tau>z);\<Delta> \<Turnstile> OK" by auto
    qed
    also have "(\<Gamma>\<bar>x),(xz:\<tau>z) = \<Gamma>,(xz:\<tau>z)\<bar>x" using Hxneq
      by (auto simp add: without_def dom_def)
    also have "(\<mu>\<bar>l),(lz\<mapsto>TmFreeVar xz) = \<mu>,(lz\<mapsto>TmFreeVar xz)\<bar>l"
      using False by (auto simp add: without_def)
    also have "(M\<bar>l),(lz:\<tau>z) = M,(lz:\<tau>z)\<bar>l" using False 
      by (auto simp add: without_def)
    finally show ?thesis using Hin by auto
  qed
qed auto

lemma memTyping_update:
  assumes "\<Gamma>m;\<Delta> \<Turnstile>' \<mu> : M"
  and "\<mu> defines l"
  and "l\<rhd>\<tau> \<in> M"
  and "\<Gamma>m,(z:\<tau>);\<Delta> \<Turnstile> OK"
  and "\<Gamma>m \<sim> {z\<rhd>\<tau>}"
  shows "\<Gamma>m,(z:\<tau>);\<Delta> \<Turnstile>' update \<mu> l (TmFreeVar z) : M"
using prems proof induct
  case mt_nil thus ?case by (auto simp add: update_def dom_def)
next
  case (mt_cons \<Delta> \<Gamma> \<mu> M xz \<tau>z lz)
  {
    assume Heq: "l = lz"
    have H\<tau>z: "\<Delta> \<Turnstile>' \<tau>z" using mt_cons poly_env_wf_ty by auto
    have "lz\<rhd>\<tau> \<in> M,(lz:\<tau>z)" using mt_cons Heq by auto
    moreover
    have "\<not> M defines lz" using mt_cons mem_dom_not_in by auto
    hence "M,(lz:\<tau>z);\<Delta> \<Turnstile> OK" using mt_cons wf_env_cons H\<tau>z by auto
    ultimately
    have Ht: "\<tau> = \<tau>z" by (auto simp add: wf_env_def)
    hence "\<Gamma>,(xz:\<tau>);\<Delta> \<Turnstile> OK" using prems by auto
    moreover
    have "\<Gamma> \<sim> {(xz\<rhd>\<tau>)}" using prems Ht by auto
    ultimately have "\<Gamma>,(xz:\<tau>);\<Delta> \<Turnstile>' \<mu>:M"
      using prems memTyping_weakening[of \<Gamma> _ \<mu> M xz \<tau>] by auto
    moreover
    have "\<Gamma>,(xz:\<tau>),(z:\<tau>);\<Delta> \<Turnstile> OK" using Ht prems by auto
    ultimately
    have "(\<Gamma>,(xz:\<tau>)),(z:\<tau>);\<Delta> \<Turnstile>' \<mu>,(lz\<mapsto>TmFreeVar z) : M,(lz:\<tau>)"
      using prems Ht by (auto simp add: env_compat_def)
    also
    have "\<not> \<mu> defines l" using prems by auto
    hence "\<mu>,(lz\<mapsto>TmFreeVar z) = update (\<mu>,(lz\<mapsto>TmFreeVar xz)) l (TmFreeVar z)"
      using Heq by (auto simp add: update_def dom_def)
    finally have ?case using Ht by auto
  }
  moreover
  {
    assume Hneq: "l \<noteq> lz"
    hence "\<mu> defines l" using prems by (auto simp add: update_def dom_def)
    moreover
    have "(l\<rhd>\<tau>) \<in> M" using prems by auto
    moreover
    have "\<Gamma>,(z:\<tau>);\<Delta> \<Turnstile> OK"
      using prems wf_env_subset[of "\<Gamma>,(xz:\<tau>z),(z:\<tau>)" _ "\<Gamma>,(z:\<tau>)"] by auto
    moreover
    have "\<Gamma> \<sim> {z\<rhd>\<tau>}" using prems by (auto simp add: env_compat_def)
    ultimately
    have Hmt: "\<Gamma>,(z:\<tau>);\<Delta> \<Turnstile>' update \<mu> l (TmFreeVar z):M" using prems by auto
    moreover
    have "\<Gamma>,(z:\<tau>) \<sim> {xz\<rhd>\<tau>z}" using prems by (auto simp add: env_compat_def)
    moreover
    have Heq: "\<Gamma>,(z:\<tau>),(xz:\<tau>z) = \<Gamma>,(xz:\<tau>z),(z:\<tau>)" by auto
    hence "\<Gamma>,(z:\<tau>),(xz:\<tau>z);\<Delta> \<Turnstile> OK" using prems by auto
    moreover
    have "\<not>  (update \<mu> l (TmFreeVar z)) defines lz"
      using prems Hneq by (auto simp add: update_def dom_def)
    moreover
    have "update \<mu> l (TmFreeVar z) \<Turnstile> Mem" using Hmt memTyping_wf_mem by auto
    ultimately
    have "\<Gamma>,(z:\<tau>),(xz:\<tau>z);\<Delta> \<Turnstile>'
      (update \<mu> l (TmFreeVar z)),(lz\<mapsto>TmFreeVar xz) : M,(lz:\<tau>z)"
      using prems by auto
    also note Heq
    also
    have "(update \<mu> l (TmFreeVar z)),(lz\<mapsto>TmFreeVar xz) =
          update (\<mu>,(lz\<mapsto>TmFreeVar xz)) l (TmFreeVar z)"
      using Hneq by (auto simp add: update_def )
    finally have ?case by auto
  }
  ultimately
  show ?case by auto
qed

lemma memTyping_update_qL_without:
  assumes "\<Gamma>;\<Delta> \<Turnstile>' \<mu> : M"
  and "(l\<mapsto>TmFreeVar x) \<in> \<mu>"
  and "l\<rhd>\<tau> \<in> M"
  and "x\<rhd>\<tau> \<in> \<Gamma>"
  and "qL \<preceq> qualT \<tau>"
  and "\<not> \<Gamma> defines z"
  and "z \<noteq> x"
  shows "\<Gamma>,(z:\<tau>)\<bar>x;\<Delta> \<Turnstile>' update \<mu> l (TmFreeVar z) : M"
using prems proof induct
  case mt_nil thus ?case by auto
next
  case (mt_cons \<Delta> \<Gamma> \<mu> M xa \<tau>a la)
  {
    assume Heq: "la = l"
    have H\<tau>a: "\<Delta> \<Turnstile>' \<tau>a" using mt_cons poly_env_wf_ty by auto
    have H\<tau>: "\<Delta> \<Turnstile>' \<tau>" using mt_cons poly_env_wf_ty by auto
    have "\<not> M defines la" using mt_cons mem_dom_not_in by auto
    hence Hmok: "M,(la:\<tau>a);\<Delta> \<Turnstile> OK" using wf_env_cons mt_cons H\<tau>a by auto
    hence Hteq: "\<tau> = \<tau>a" using Heq mt_cons wf_env_def[of "M,(la:\<tau>a)"] by blast
    have "\<mu>,(l\<mapsto>TmFreeVar xa) \<Turnstile> Mem" using mt_cons wf_mem_cons Heq by auto
    hence Hxeq: "x = xa"
      using mt_cons Heq wf_mem_def[of "\<mu>,(l\<mapsto>TmFreeVar xa)"] by fast
    have "qualT \<tau> = qL" using mt_cons by (cases "qualT \<tau>", auto)
    hence "\<not> \<Gamma> defines x" 
      using mt_cons Hxeq Hteq compat_ok_sanity_right[of \<Gamma> "{xa\<rhd>\<tau>a}" _ x \<tau>]
      by auto
    hence Hgeq: "\<Gamma> = \<Gamma>,(xa:\<tau>)\<bar>x" using Hxeq 
      by (auto simp add: without_def dom_def)
    have "\<Gamma> \<sim> {z\<rhd>\<tau>}" using mt_cons by (auto simp add: env_compat_def dom_def)
    moreover
    have "\<Gamma>;\<Delta> \<Turnstile> OK" using mt_cons wf_env_subset[of "\<Gamma>,(xa:\<tau>a)"] by blast
    hence "\<Gamma>,(z:\<tau>);\<Delta> \<Turnstile> OK" using Hteq mt_cons wf_env_cons H\<tau> by (auto simp add: dom_def)
    ultimately
    have "\<Gamma>,(z:\<tau>);\<Delta> \<Turnstile>' \<mu>,(l\<mapsto>TmFreeVar z) : M,(l:\<tau>)" using mt_cons Heq by auto
    also note Hgeq
    also have "(\<Gamma>,(xa:\<tau>)\<bar>x),(z:\<tau>) = \<Gamma>,(xa:\<tau>),(z:\<tau>)\<bar>x"
      using mt_cons by (auto simp add: without_def)
    finally have "\<Gamma>,(xa:\<tau>a),(z:\<tau>)\<bar>x;\<Delta> \<Turnstile>' \<mu>,(l\<mapsto>TmFreeVar z) : M,(l:\<tau>a)"
      using Hteq by auto
    also have "\<mu>,(l\<mapsto>TmFreeVar z) =
      update (\<mu>,(l\<mapsto>TmFreeVar xa)) l (TmFreeVar z)"
      using Heq mt_cons by (auto simp add: update_def dom_def)
    finally have ?case using Heq by auto
  }
  moreover
  {
    assume Hneq: "la \<noteq> l"
    have H\<tau>: "\<Delta> \<Turnstile>' \<tau>" using mt_cons poly_env_wf_ty by auto
    have Hzxa: "z \<noteq> xa" using mt_cons by (auto simp add: dom_def)
    have Hznotin: "\<not> (\<Gamma>\<bar>x \<union> {xa\<rhd>\<tau>a}) defines z"
      using mt_cons by (auto simp add: dom_def without_def)
    have Hxm: "l\<mapsto>TmFreeVar x \<in> \<mu>" using Hneq mt_cons by auto
    moreover
    have Hlm: "l\<rhd>\<tau> \<in> M" using Hneq mt_cons by auto
    moreover
    have Htql: "qualT \<tau> = qL" using mt_cons by (cases "qualT \<tau>", auto)
    have Hgx: "x\<rhd>\<tau> \<in> \<Gamma>" using mt_cons Hlm Hxm memTyping_elim_qL[of \<Gamma> _ \<mu> M l x \<tau>] by auto
    ultimately
    have "\<Gamma>,(z:\<tau>)\<bar>x;\<Delta> \<Turnstile>' update \<mu> l (TmFreeVar z):M" using mt_cons
      by (auto simp add: dom_def)
    moreover have Hgeq: "(\<Gamma>\<bar>x),(z:\<tau>) = \<Gamma>,(z:\<tau>)\<bar>x" using mt_cons
      by (auto simp add: without_def)
    moreover have "\<not> (update \<mu> l (TmFreeVar z)) defines la"
      using Hneq mt_cons by (auto simp add: update_def dom_def)
    moreover have "update \<mu> l (TmFreeVar z) \<Turnstile> Mem" 
      using mt_cons mem_update_ok by auto
    moreover
    have "(\<Gamma>\<bar>x) \<sim> {xa\<rhd>\<tau>a}"
      using mt_cons env_compat_subset[of \<Gamma> _ "\<Gamma>\<bar>x"]
      by (auto simp add: without_def)
    hence "(\<Gamma>\<bar>x),(z:\<tau>) \<sim> {xa\<rhd>\<tau>a}" using Hzxa by (auto simp add: env_compat_def)
    moreover
    {
      have "(\<Gamma>\<bar>x) \<union> {xa\<rhd>\<tau>a};\<Delta> \<Turnstile> OK"
        using mt_cons wf_env_subset[of "\<Gamma>,(xa:\<tau>a)" _ "(\<Gamma>\<bar>x) \<union> {xa\<rhd>\<tau>a}"]
        by (auto simp add: without_def)
      hence "((\<Gamma>\<bar>x) \<union> {xa\<rhd>\<tau>a}),(z:\<tau>);\<Delta> \<Turnstile> OK"
        using mt_cons Hzxa wf_env_cons[of "(\<Gamma>\<bar>x) \<union> {xa\<rhd>\<tau>a}" _ z \<tau>] Hznotin H\<tau>
        by auto
      also have "((\<Gamma>\<bar>x) \<union> {xa\<rhd>\<tau>a}),(z:\<tau>) = (\<Gamma>\<bar>x),(z:\<tau>) \<union> {xa\<rhd>\<tau>a}" by auto
      finally have "(\<Gamma>\<bar>x),(z:\<tau>) \<union> {xa\<rhd>\<tau>a};\<Delta> \<Turnstile> OK" by auto
    }
    ultimately
    have "(\<Gamma>\<bar>x),(z:\<tau>),(xa:\<tau>a);\<Delta> \<Turnstile>'
      (update \<mu> l (TmFreeVar z)),(la\<mapsto>TmFreeVar xa) : M,(la:\<tau>a)"
      using mt_cons by auto
    also note Hgeq
    also
    have "x \<noteq> xa"
      using mt_cons Htql Hgx compat_ok_sanity[of \<Gamma> "{xa\<rhd>\<tau>a}" _ x \<tau>] 
      by (auto simp add: dom_def)
    hence "(\<Gamma>,(z:\<tau>)\<bar>x),(xa:\<tau>a) = \<Gamma>,(xa:\<tau>a),(z:\<tau>)\<bar>x" 
      by (auto simp add: without_def)
    also have "(update \<mu> l (TmFreeVar z)),(la\<mapsto>TmFreeVar xa) =
               update (\<mu>,(la\<mapsto>TmFreeVar xa)) l (TmFreeVar z)"
      using Hneq by (auto simp add: update_def)
    finally have ?case by auto
  }
  ultimately show ?case by auto
qed

end
