theory Metatheory
imports Typing Semantics

begin

consts
  INV2 :: "store \<Rightarrow> bool"
  INV3 :: "store \<Rightarrow> bool"
  INV4 :: "store \<Rightarrow> bool"
  INV5 :: "store \<Rightarrow> bool"
  INV6 :: "tm \<Rightarrow> scopeset \<Rightarrow> bool"
  INV7 :: "tm \<Rightarrow> tm \<Rightarrow> scopeset \<Rightarrow> bool" 
  INV8 :: "scopeset \<Rightarrow> scopeset \<Rightarrow> tm \<Rightarrow> bool" 
  INV10 :: "env \<Rightarrow> env \<Rightarrow> store \<Rightarrow> scopeset \<Rightarrow> bool" 
defs
  INV2[simp]: "INV2 S \<equiv> \<forall>x l \<pi>. (x\<mapsto>TmLoc qL \<pi> l) \<in> S \<longrightarrow> \<not> l free loc in st (S\<bar>x)"
  INV3[simp]: "INV3 S \<equiv> \<forall>x v. (x\<mapsto>v) \<in> S \<longrightarrow> (\<exists>q l \<pi>. v = TmLoc q \<pi> l) \<or> (\<forall>l. \<not> l free loc in v)"
  INV4: "INV4 S \<equiv> \<forall>x l \<pi>. (x\<mapsto>TmLoc qU \<pi> l) \<in> S \<longrightarrow> 
               (\<forall>x v. (x\<mapsto>v) \<in> S \<and> l free loc in v \<longrightarrow> scopeV v = \<pi>)"
  INV5: "INV5 S \<equiv> \<forall>x v. (x\<mapsto>v) \<in> S \<longrightarrow> \<not> hasDollars v"
  INV6[simp]: "INV6 e Z \<equiv> \<forall>x. x is free dollar sc in e \<longrightarrow> \<not> Z defines x"
  INV7[simp]: "INV7 e e' Zs \<equiv> 
      \<forall>\<rho>. \<rho> \<in> set (free_let_dollar_scope e') \<and> \<rho> \<notin> set (free_let_dollar_scope e) 
          \<longrightarrow> \<not> Zs defines \<rho>"
  INV8[simp]: "INV8 Z' Z e' \<equiv> dom Z' \<subseteq> (dom Z \<union> set (free_let_dollar_scope e'))"
  INV10: "INV10 M G S Zl \<equiv>
  \<forall>z l q \<pi> \<tau>. (((l\<rhd>\<tau>) \<in> (M\<union>G) \<and> ((z\<mapsto>TmLoc q \<pi> l) \<in> (filterStore S Zl))) \<longrightarrow> (l\<rhd>\<tau>) \<in> M)"

inductive INV9 :: "tm \<Rightarrow> bool" ("INV9 _")
  where
    inv_unit[intro!]:   "INV9 (TmUnit q \<pi>)"
  | inv_var[intro!]:    "INV9 (TmFreeVar n)"
  | inv_bvar[intro!]:    "INV9 (TmVar n)"
  | inv_abs[intro!]:    "\<lbrakk> INV9 e; \<not> hasDollars e \<rbrakk> \<Longrightarrow> INV9 (\<lambda>^q^\<pi>[\<tau>].e)"
  | inv_app[intro!]:    "\<lbrakk> INV9 e1; INV9 e2 \<rbrakk> \<Longrightarrow> INV9 e1\<cdot>e2"
  | inv_ptyabs[intro!]: "\<lbrakk> INV9 e; \<not> hasDollars e \<rbrakk> \<Longrightarrow> INV9 (\<Lambda>^q^\<pi>[*].e)"
  | inv_ptyapp[intro!]: "\<lbrakk> INV9 e \<rbrakk> \<Longrightarrow> INV9 e\<langle>\<phi>\<rangle>"
  | inv_new[intro!]:    "\<lbrakk> INV9 e \<rbrakk> \<Longrightarrow> INV9 (TmNew e)"
  | inv_free[intro!]:   "\<lbrakk> INV9 e \<rbrakk> \<Longrightarrow> INV9 (TmFree e)"
  | inv_assign[intro!]: "\<lbrakk> INV9 e1; INV9 e2 \<rbrakk> \<Longrightarrow> INV9 e1:=e2"
  | inv_swap[intro!]:   "\<lbrakk> INV9 e1; INV9 e2 \<rbrakk> \<Longrightarrow> INV9 e1:=:e2"
  | inv_free[intro!]:   "\<lbrakk> INV9 e \<rbrakk> \<Longrightarrow> INV9 (TmDeref e)"
  | inv_let[intro!]:    "\<lbrakk> INV9 e; INV9 e1; INV9 e2; \<not> hasDollars e2 \<rbrakk> \<Longrightarrow> INV9 (let! (e) e1 in e2)"
  | inv_let'[intro!]:   "\<lbrakk> INV9 e; INV9 e1; INV9 e2; \<not> hasDollars e2 \<rbrakk> \<Longrightarrow> INV9 (let$ \<rho> (e) e1 in e2)"
  | inv_loc[intro!]:    "INV9 (TmLoc q \<pi> l)"

inductive_cases inv_elim[elim]: "INV9 e"

lemma inv2_subset:
  assumes "INV2 S"
  and "S' \<subseteq> S"
  shows "INV2 S'"
using prems proof auto
  fix x l \<pi>
  assume "(x\<mapsto>TmLoc qL \<pi> l) \<in> S'" 
    and "l free loc in st (S'\<bar>x)"
  have "(S'\<bar>x) \<subseteq> (S\<bar>x)" using prems by (auto simp add: without_def)
  hence "l free loc in st (S\<bar>x)" using prems by (auto simp add: free_loc_store_def)
  thus False using prems by auto
qed

lemma inv3_subset:
  assumes "INV3 S"
  and "S' \<subseteq> S"
  shows "INV3 S'"
using prems proof auto
  fix x v
  assume "(x\<mapsto>v) \<in> S'" and "free_loc v \<noteq> []"
  have "(x\<mapsto>v) \<in> S" using prems by auto
  thus "\<exists>q l \<pi>. v = TmLoc q \<pi> l" using prems by auto
qed

lemma inv2_cons:
  assumes "INV2 S"
  and "\<not> S defines z"
  and "\<forall>l. \<not> l free loc in v"
  shows "INV2 (S,(z\<mapsto>v))"
using prems proof auto
  fix x l \<pi>
  assume "(x\<mapsto>TmLoc qL \<pi> l) \<in> S" 
    and "l free loc in st (S,(z\<mapsto>v) \<bar> x)"
  hence "\<not> l free loc in st (S\<bar>x)" using prems by auto
  moreover
  have "\<not> l free loc in v" using prems by auto
  ultimately
  have "\<not> l free loc in st ((S\<bar>x),(z\<mapsto>v))" using free_loc_store_def by auto
  also
  have "z \<noteq> x" using prems by (auto simp add: dom_def)
  hence "(S\<bar>x),(z\<mapsto>v) = (S,(z\<mapsto>v))\<bar>x" by (auto simp add: without_def)
  finally show False using prems by auto
qed

lemma inv2_cons_loc:
  assumes "INV2 S"
  and "\<not> S defines z"
  and "\<not> l free loc in st S"
  shows "INV2 (S,(z\<mapsto>TmLoc q \<pi> l))"
using prems proof auto
  assume "l free loc in st (S,(z\<mapsto>TmLoc qL \<pi> l)\<bar>z)"
  have "S,(z\<mapsto>TmLoc qL \<pi> l)\<bar>z = S" 
    using prems by (auto simp add: without_def dom_def)
  thus False using prems by auto
next
  fix x lx \<pi>l
  assume "(x\<mapsto>TmLoc qL \<pi>l lx) \<in> S" 
    and "lx free loc in st (S,(z\<mapsto>TmLoc q \<pi> l)\<bar>x)"
  hence "S,(z\<mapsto>TmLoc q \<pi> l)\<bar>x = (S\<bar>x),(z\<mapsto>TmLoc q \<pi> l)"
    using prems by (auto simp add: without_def dom_def)
  hence "lx free loc in st ((S\<bar>x),(z\<mapsto>TmLoc q \<pi> l))" using prems by auto
  moreover
  hence "\<not> lx free loc in st (S\<bar>x)" using prems by auto
  ultimately
  have Heq: "l = lx" by (auto simp add: free_loc_store_def)
  have "lx free loc in st S" proof (auto simp add: free_loc_store_def)
    show "\<exists>x. (\<exists>z v. x = set (free_loc v) \<and> (z\<mapsto>v) \<in> S) \<and> lx \<in> x" proof (rule exI | rule conjI)+
      show "(x\<mapsto>TmLoc qL \<pi>l lx) \<in> S" using prems by auto
    qed auto
  qed
  thus False using Heq prems by auto
qed
  
lemma inv2_without_conv:
  assumes "INV2 S"
  and "INV3 S"
  and "S \<Turnstile> Store"
  and "x\<mapsto>v \<in> S"
  shows "INV2 ((S\<bar>x),(x\<mapsto>convV \<rho> v))"
proof -
  have Hsx: "INV2 (S\<bar>x)" using inv2_subset[of S "S\<bar>x"] assms
    by (auto simp add: without_def dom_def)
  have Hval: "value v"  using wf_store_def assms by auto
  show ?thesis proof (cases v)
    case (TmUnit q \<pi>)
    have "INV2 ((S \<bar> x),(x\<mapsto>TmUnit qU (ScFree \<rho>)))"
      using inv2_cons[of "S\<bar>x" x "TmUnit qU (ScFree \<rho>)"] Hsx 
      by (auto simp add: dom_def without_def)
    thus ?thesis using TmUnit by auto
  next
    case (TmAbs q \<pi>l \<tau> e)
    have "INV3 S" and "x\<mapsto>v \<in> S" .
    hence "(\<exists>q l \<pi>. v = TmLoc q \<pi> l) \<or> (\<forall>l. l \<notin> set (free_loc v))" by auto
    hence "\<forall>l. \<not> l free loc in v" using TmAbs by auto
    hence "\<forall>l. \<not> l free loc in (\<lambda>^q^\<pi>l[\<tau>].e)" using TmAbs by auto
    hence "\<forall>l. \<not> l free loc in (\<lambda>^qU^(ScFree \<rho>)[\<tau>].e)" by auto
    hence "INV2 ((S\<bar>x),(x\<mapsto>\<lambda>^qU^(ScFree \<rho>)[\<tau>].e))"
      using inv2_cons[of "S\<bar>x" x "\<lambda>^qU^(ScFree \<rho>)[\<tau>].e"] Hsx 
      by (auto simp add: dom_def without_def)
    thus ?thesis using TmAbs by auto
  next
    case (TmLoc q \<pi> l)
    {
      assume "q = qL"
      hence "l \<notin> free_loc_store (S\<bar>x)" using assms TmLoc INV2[of S] by auto
      hence "INV2 ((S\<bar>x),(x\<mapsto>TmLoc qU (ScFree \<rho>) l))"
        using inv2_cons_loc[of "S\<bar>x" x l qU] Hsx prems
        by (auto simp add: dom_def without_def)
      hence ?thesis using TmLoc by auto
    }
    moreover
    {
      assume Hu: "q = qU"
      have Hinv2: "INV2 S" .
      hence "INV2 (S\<bar>x)" using inv2_subset[of S "S\<bar>x"] by (auto simp add: without_def)
      moreover
      have "x\<mapsto>TmLoc q \<pi> l \<in> S" using prems by auto
      ultimately
      have "INV2 ((S\<bar>x),(x\<mapsto> TmLoc qU (ScFree \<rho>) l))" proof auto
        fix xa la \<pi>'
        assume "\<forall>xa l. (\<exists>\<pi>. xa\<mapsto>TmLoc qL \<pi> l \<in> S\<bar>x) \<longrightarrow> 
                             l \<notin> free_loc_store (S \<bar> x \<bar> xa)"
          and Hinx: "x\<mapsto>TmLoc q \<pi> l \<in> S"
          and Hina: "xa\<mapsto>TmLoc qL \<pi>' la \<in> S\<bar>x"
          and H: "la free loc in st (S\<bar>x),(x\<mapsto>TmLoc qU (ScFree \<rho>) l) \<bar>xa"
        {
          assume "x=xa"
          hence False using prems by (auto simp add: without_def)
        }
        moreover
        {
          assume Hneq: "x\<noteq>xa"
          hence "(S\<bar>x),(x\<mapsto>TmLoc qU (ScFree \<rho>) l)\<bar>xa  = 
                (S\<bar>x\<bar>xa),(x\<mapsto>TmLoc qU (ScFree \<rho>) l)"
            by (auto simp add: without_def)
          hence "la free loc in st (S\<bar>x\<bar>xa),(x\<mapsto>TmLoc qU (ScFree \<rho>) l)"
            using H by auto
          hence "la free loc in st ((S\<bar>x\<bar>xa) \<union> {x\<mapsto>TmLoc qU (ScFree \<rho>) l})"
            by auto
          moreover
          have "\<not> la free loc in st (S\<bar>x\<bar>xa)" using prems by blast
          ultimately 
          have Htrue: "la free loc in (TmLoc qU (ScFree \<rho>) l)" 
            using free_loc_store_def by auto
          have "xa\<mapsto> TmLoc qL \<pi>' la \<in> S" using Hneq Hina
            by (auto simp add: without_def)
          hence "\<not> la free loc in st (S\<bar>xa)" using Hinv2 by auto
          moreover
          have "x\<mapsto>TmLoc q \<pi> l \<in> S\<bar>xa" using Hinx Hneq by (auto simp add: without_def)
          ultimately
          have "\<not> la free loc in (TmLoc q \<pi> l)"
            using free_loc_store_def by blast
          hence False using Htrue by auto
        }
        ultimately show False by auto
      qed
      hence ?thesis using prems by auto
    }
    ultimately show ?thesis by (cases q, auto)
  next
    case (TmProd ql \<pi> e)
    have "INV3 S" and "x\<mapsto>v \<in> S" .
    hence "(\<exists>q l \<pi>. v = TmLoc q \<pi> l) \<or> (\<forall>l. l \<notin> set (free_loc v))" by auto
    hence "\<forall>l. \<not> l free loc in v" using TmProd by auto
    hence "\<forall>l. \<not> l free loc in (\<Lambda>^ql^\<pi>[*].e)" using TmProd by auto
    hence "\<forall>l. \<not> l free loc in (\<Lambda>^qU^(ScFree \<rho>)[*].e)" by auto
    hence "INV2 ((S\<bar>x),(x\<mapsto>\<Lambda>^qU^(ScFree \<rho>)[*].e))"
      using inv2_cons[of "S\<bar>x" x "\<Lambda>^qU^(ScFree \<rho>)[*].e"] Hsx 
      by (auto simp add: dom_def without_def)
    thus ?thesis using TmProd by auto
  qed (insert Hsx Hval, auto)
qed

lemma inv3_without_conv:
  assumes "INV3 S"
  and "S \<Turnstile> Store"
  and "x\<mapsto>v \<in> S"
  shows "INV3 ((S\<bar>x),(x\<mapsto>convV \<rho> v))"
proof -
  have "(\<exists>q l \<pi>. v = TmLoc q \<pi> l) \<or> (\<forall>l. l \<notin> set (free_loc v))"
    using assms by auto
  moreover
  have "value v" using wf_store_def assms by auto
  hence "free_loc v = free_loc (convV \<rho> v)" by auto
  ultimately
  have "(\<exists>q l \<pi>. convV \<rho> v = TmLoc qU (ScFree \<rho>) l) \<or> 
                 (\<forall>l. l \<notin> set (free_loc (convV \<rho> v)))"
    by auto
  moreover
  have "INV3 (S\<bar>x)" using inv3_subset[of S "S\<bar>x"] assms
    by (auto simp add: without_def dom_def)
  ultimately show ?thesis by auto
qed

lemma inv4_elim:
  assumes "INV4 S"
  and "x\<mapsto>TmLoc qU (ScFree \<rho>) l \<in> S"
  and H: "z\<mapsto>TmLoc q \<pi> l \<in> S"
  shows "\<pi> = ScFree \<rho>"
proof -
  have "\<forall>x v. ((x\<mapsto>v) \<in> S \<and> l free loc in v) \<longrightarrow> scopeV v = ScFree \<rho>" using prems INV4 by auto
  moreover
  have "z\<mapsto>TmLoc q \<pi> l \<in> S" and "l free loc in TmLoc q \<pi> l" using prems by auto
  ultimately
  have "scopeV (TmLoc q \<pi> l) = ScFree \<rho>" by blast
  thus ?thesis by auto
qed

lemma inv4_subset:
  assumes "INV4 S"
  and "S' \<subseteq> S"
  shows "INV4 S'"
using prems INV4 proof auto
  fix x l \<pi> xa v
  assume H1: "\<forall>x l \<pi>. x\<mapsto>TmLoc qU \<pi> l \<in> S \<longrightarrow> (\<forall>x v. x\<mapsto>v \<in> S \<and> l free loc in v \<longrightarrow> scopeV v = \<pi>)"
    and H2: "x\<mapsto>TmLoc qU \<pi> l \<in> S'"
    and H3: "xa\<mapsto>v \<in> S'"
    and H4: "l free loc in v"
  hence "x\<mapsto>TmLoc qU \<pi> l \<in> S" using assms by auto
  hence "\<forall>x v. x\<mapsto>v \<in> S \<and> l free loc in v \<longrightarrow> scopeV v = \<pi>" using H1 by auto
  moreover
  have "xa\<mapsto>v \<in> S" using assms H3 by auto
  ultimately
  show "scopeV v = \<pi>" using H4 by auto
qed

lemma inv4_cons_loc:
  assumes "INV4 S"
  and "\<not> l free loc in st S"
  shows "INV4 (S,(z\<mapsto>TmLoc q \<pi> l))"
using prems proof -
  have H: "\<not> l free loc in st S" using assms by auto
  thus "INV4 (S,(z\<mapsto>TmLoc q \<pi> l))" using assms INV4 proof (auto simp add: free_loc_store_def)
    fix x \<pi>'
    assume Has: "x\<mapsto>TmLoc qU \<pi>' l \<in> S"
    have "\<forall>x v. x\<mapsto>v \<in> S \<longrightarrow> \<not> l free loc in v" using H by (auto simp add: free_loc_store_def)
    hence "\<not> l free loc in (TmLoc qU \<pi>' l)" using Has by blast
    thus "\<pi> = \<pi>'" by auto
  qed
qed

lemma inv4_cons:
  assumes "INV4 S"
  and "\<not> S defines z"
  and "\<forall>l. \<not> l free loc in v"
  shows "INV4 (S,(z\<mapsto>v))"
using prems by (auto simp add: INV4)

lemma inv5_subset:
  assumes "INV5 S"
  and "S' \<subseteq> S"
  shows "INV5 S'"
using prems INV5 by auto

lemma INV9_elim:
  assumes "value e"
  and "INV9 e"
  shows "\<not> hasDollars e"
using prems by (cases e, auto)

lemma INV9_intro:
  assumes "\<not> hasDollars e"
  shows "INV9 e"
using prems by (induct e, auto)

lemma hasDollars_exists: 
  assumes "hasDollars e"
  shows "\<exists>\<rho>. \<rho> is free dollar sc in e"
using prems by (induct e, auto)

lemma not_hasDollars_exists:
  assumes "\<not> hasDollars e"
  shows "\<not> (\<exists>\<rho>. \<rho> is free dollar sc in e)"
using prems by (induct e, auto)

lemma inv9_vsubst:
  shows "INV9 e = INV9 (vsubst_tm (TmFreeVar x) i (vsubst_sc (ScFree \<rho>) j e ))"
proof (induct e arbitrary:i j)
  case (TmAbs q \<pi> \<tau> e) 
  hence "INV9 e = INV9 vsubst_tm (TmFreeVar x) (Suc i) (vsubst_sc (ScFree \<rho>) j e)" by auto
  moreover
  have "(\<not> hasDollars e) =
    (\<not> hasDollars (vsubst_tm (TmFreeVar x) (Suc i) (vsubst_sc (ScFree \<rho>) j e)))"
  proof auto
    assume "hasDollars e"
    then obtain \<rho>' where "\<rho>' is free dollar sc in e" using hasDollars_exists by auto
    hence "\<rho>' is free dollar sc in (vsubst_tm (TmFreeVar x) (Suc i) (vsubst_sc (ScFree \<rho>) j e))"
      using free_dollars_vsubst[of e x "Suc i" \<rho> j] by auto
    thus "hasDollars (vsubst_tm (TmFreeVar x) (Suc i) (vsubst_sc (ScFree \<rho>) j e))"
      using not_hasDollars_exists by blast
  next
    assume "hasDollars (vsubst_tm (TmFreeVar x) (Suc i) (vsubst_sc (ScFree \<rho>) j e))"
    then obtain \<rho>' where "\<rho>' is free dollar sc in
      (vsubst_tm (TmFreeVar x) (Suc i) (vsubst_sc (ScFree \<rho>) j e))" using hasDollars_exists by auto
    hence "\<rho>' is free dollar sc in e"
      using free_dollars_vsubst[of e x "Suc i" \<rho> j] by auto
    thus "hasDollars e"
      using not_hasDollars_exists by blast
  qed
  ultimately
  show ?case by auto
next
  case (TmProd q \<pi> e) 
  hence "INV9 e = INV9 vsubst_tm (TmFreeVar x) i (vsubst_sc (ScFree \<rho>) j e)" by auto
  moreover
  have "(\<not> hasDollars e) =
    (\<not> hasDollars (vsubst_tm (TmFreeVar x) i (vsubst_sc (ScFree \<rho>) j e)))"
  proof auto
    assume "hasDollars e"
    then obtain \<rho>' where "\<rho>' is free dollar sc in e" using hasDollars_exists by auto
    hence "\<rho>' is free dollar sc in (vsubst_tm (TmFreeVar x) i (vsubst_sc (ScFree \<rho>) j e))"
      using free_dollars_vsubst[of e x i \<rho> j] by auto
    thus "hasDollars (vsubst_tm (TmFreeVar x) i (vsubst_sc (ScFree \<rho>) j e))"
      using not_hasDollars_exists by blast
  next
    assume "hasDollars (vsubst_tm (TmFreeVar x) i (vsubst_sc (ScFree \<rho>) j e))"
    then obtain \<rho>' where "\<rho>' is free dollar sc in
      (vsubst_tm (TmFreeVar x) i (vsubst_sc (ScFree \<rho>) j e))" using hasDollars_exists by auto
    hence "\<rho>' is free dollar sc in e"
      using free_dollars_vsubst[of e x i \<rho> j] by auto
    thus "hasDollars e"
      using not_hasDollars_exists by blast
  qed
  ultimately
  show ?case by auto
next
  case (TmLetbangSc \<rho>x e e1 e2 i j) thus ?case proof auto
    assume "INV9 let$ \<rho>x e e1 in e2"
     and H: "hasDollars (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_sc (ScFree \<rho>) j e2))"
    hence "\<not> hasDollars e2" by auto
    hence "free_let_dollar_scope e2 = []" using not_hasDollars_exists by auto
    hence "free_let_dollar_scope 
      (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_sc (ScFree \<rho>) j e2)) = []"
      using free_dollars_vsubst by auto
    hence "\<not> (\<exists>\<rho>'. \<rho>' is free dollar sc in
      (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_sc (ScFree \<rho>) j e2)))"
      by auto
    hence "\<not> hasDollars (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_sc (ScFree \<rho>) j e2))"
      using hasDollars_exists by blast
    thus False using H by auto
  next
     assume "INV9 let$ \<rho>x vsubst_tm (TmFreeVar x) i (vsubst_sc (ScFree \<rho>) j e)
                          vsubst_tm (TmFreeVar x) i (vsubst_sc (ScFree \<rho>) j e1) in
                          vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_sc (ScFree \<rho>) j e2)"
      and H: "hasDollars e2"
    hence "\<not> hasDollars (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_sc (ScFree \<rho>) j e2))"
      by auto
    hence "free_let_dollar_scope (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_sc (ScFree \<rho>) j e2))
            = []"
      using not_hasDollars_exists by auto
    hence "free_let_dollar_scope e2 = []" using free_dollars_vsubst by auto
    hence "\<not> (\<exists>\<rho>'. \<rho>' is free dollar sc in e2)" by auto
    hence "\<not> hasDollars e2" using hasDollars_exists by blast
    thus False using H by auto
  qed
next
  case (TmLetbang e e1 e2 i j) thus ?case proof auto
    assume "INV9 let! e e1 in e2"
     and H: "hasDollars (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_sc (ScFree \<rho>) j e2))"
    hence "\<not> hasDollars e2" by auto
    hence "free_let_dollar_scope e2 = []" using not_hasDollars_exists by auto
    hence "free_let_dollar_scope 
      (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_sc (ScFree \<rho>) j e2)) = []"
      using free_dollars_vsubst by auto
    hence "\<not> (\<exists>\<rho>'. \<rho>' is free dollar sc in
      (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_sc (ScFree \<rho>) j e2)))"
      by auto
    hence "\<not> hasDollars (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_sc (ScFree \<rho>) j e2))"
      using hasDollars_exists by blast
    thus False using H by auto
  next
     assume "INV9 let! vsubst_tm (TmFreeVar x) i (vsubst_sc (ScFree \<rho>) j e)
                          vsubst_tm (TmFreeVar x) (Suc i) (vsubst_sc (ScFree \<rho>) (Suc j) e1) in
                          vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_sc (ScFree \<rho>) j e2)"
      and H: "hasDollars e2"
    hence "\<not> hasDollars (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_sc (ScFree \<rho>) j e2))"
      by auto
    hence "free_let_dollar_scope (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_sc (ScFree \<rho>) j e2))
            = []"
      using not_hasDollars_exists by auto
    hence "free_let_dollar_scope e2 = []" using free_dollars_vsubst by auto
    hence "\<not> (\<exists>\<rho>'. \<rho>' is free dollar sc in e2)" by auto
    hence "\<not> hasDollars e2" using hasDollars_exists by blast
    thus False using H by auto
  qed
qed auto

lemma inv9_vsubst_tm:
  shows "INV9 e = INV9 (vsubst_tm (TmFreeVar x) i (vsubst_tm (TmFreeVar y) j e ))"
proof (induct e arbitrary:i j)
  case (TmAbs q \<pi> \<tau> e) 
  hence "INV9 e = INV9 vsubst_tm (TmFreeVar x) (Suc i) (vsubst_tm (TmFreeVar y) (Suc j) e)" by auto
  moreover
  have "(\<not> hasDollars e) =
    (\<not> hasDollars (vsubst_tm (TmFreeVar x) (Suc i) (vsubst_tm (TmFreeVar y) (Suc j) e)))"
  proof auto
    assume "hasDollars e"
    then obtain \<rho>' where "\<rho>' is free dollar sc in e" using hasDollars_exists by auto
    hence "\<rho>' is free dollar sc in 
      (vsubst_tm (TmFreeVar x) (Suc i) (vsubst_tm (TmFreeVar y) (Suc j) e))"
      using free_dollars_vsubst_tm[of e x "Suc i" y "Suc j"] by auto
    thus "hasDollars (vsubst_tm (TmFreeVar x) (Suc i) (vsubst_tm (TmFreeVar y) (Suc j) e))"
      using not_hasDollars_exists by blast
  next
    assume "hasDollars (vsubst_tm (TmFreeVar x) (Suc i) (vsubst_tm (TmFreeVar y) (Suc j) e))"
    then obtain \<rho>' where "\<rho>' is free dollar sc in
      (vsubst_tm (TmFreeVar x) (Suc i) (vsubst_tm (TmFreeVar y) (Suc j) e))"
      using hasDollars_exists by auto
    hence "\<rho>' is free dollar sc in e"
      using free_dollars_vsubst_tm[of e x "Suc i" y "Suc j"] by auto
    thus "hasDollars e"
      using not_hasDollars_exists by blast
  qed
  ultimately
  show ?case by auto
next
  case (TmProd q \<pi> e)
  hence "INV9 e = INV9 vsubst_tm (TmFreeVar x) i (vsubst_tm (TmFreeVar y) j e)" by auto
  moreover
  have "(\<not> hasDollars e) =
    (\<not> hasDollars (vsubst_tm (TmFreeVar x) i (vsubst_tm (TmFreeVar y) j e)))"
  proof auto
    assume "hasDollars e"
    then obtain \<rho>' where "\<rho>' is free dollar sc in e" using hasDollars_exists by auto
    hence "\<rho>' is free dollar sc in 
      (vsubst_tm (TmFreeVar x) i (vsubst_tm (TmFreeVar y) j e))"
      using free_dollars_vsubst_tm[of e x i y j] by auto
    thus "hasDollars (vsubst_tm (TmFreeVar x) i (vsubst_tm (TmFreeVar y) j e))"
      using not_hasDollars_exists by blast
  next
    assume "hasDollars (vsubst_tm (TmFreeVar x) i (vsubst_tm (TmFreeVar y) j e))"
    then obtain \<rho>' where "\<rho>' is free dollar sc in
      (vsubst_tm (TmFreeVar x) i (vsubst_tm (TmFreeVar y) j e))"
      using hasDollars_exists by auto
    hence "\<rho>' is free dollar sc in e"
      using free_dollars_vsubst_tm[of e x i y j] by auto
    thus "hasDollars e"
      using not_hasDollars_exists by blast
  qed
  ultimately
  show ?case by auto
next
  case (TmLetbangSc \<rho>x e e1 e2 i j) thus ?case proof auto
    assume "INV9 let$ \<rho>x e e1 in e2"
     and H: "hasDollars (vsubst_tm (TmFreeVar x) (Suc (Suc i))
                          (vsubst_tm (TmFreeVar y) (Suc (Suc j)) e2))"
    hence "\<not> hasDollars e2" by auto
    hence "free_let_dollar_scope e2 = []" using not_hasDollars_exists by auto
    hence "free_let_dollar_scope 
      (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_tm (TmFreeVar y) (Suc (Suc j)) e2)) = []"
      using free_dollars_vsubst_tm by auto
    hence "\<not> (\<exists>\<rho>'. \<rho>' is free dollar sc in
      (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_tm (TmFreeVar y) (Suc (Suc j)) e2)))"
      by auto
    hence "\<not> hasDollars (vsubst_tm (TmFreeVar x) (Suc (Suc i)) 
                         (vsubst_tm (TmFreeVar y) (Suc (Suc j)) e2))"
      using hasDollars_exists by blast
    thus False using H by auto
  next
     assume "INV9 let$ \<rho>x vsubst_tm (TmFreeVar x) i (vsubst_tm (TmFreeVar y) j e)
                          vsubst_tm (TmFreeVar x) i (vsubst_tm (TmFreeVar y) j e1) in
                          vsubst_tm (TmFreeVar x) (Suc (Suc i)) 
                           (vsubst_tm (TmFreeVar y) (Suc (Suc j)) e2)"
      and H: "hasDollars e2"
    hence "\<not> hasDollars (vsubst_tm (TmFreeVar x) (Suc (Suc i)) 
                         (vsubst_tm (TmFreeVar y) (Suc (Suc j)) e2))"
      by auto
    hence "free_let_dollar_scope (vsubst_tm (TmFreeVar x) (Suc (Suc i)) 
                                   (vsubst_tm (TmFreeVar y) (Suc (Suc j)) e2)) = []"
      using not_hasDollars_exists by auto
    hence "free_let_dollar_scope e2 = []" using free_dollars_vsubst_tm by auto
    hence "\<not> (\<exists>\<rho>'. \<rho>' is free dollar sc in e2)" by auto
    hence "\<not> hasDollars e2" using hasDollars_exists by blast
    thus False using H by auto
  qed
next
  case (TmLetbang e e1 e2 i j) thus ?case proof auto
    assume "INV9 let! e e1 in e2"
     and H: "hasDollars (vsubst_tm (TmFreeVar x) (Suc (Suc i))
                          (vsubst_tm (TmFreeVar y) (Suc (Suc j)) e2))"
    hence "\<not> hasDollars e2" by auto
    hence "free_let_dollar_scope e2 = []" using not_hasDollars_exists by auto
    hence "free_let_dollar_scope 
      (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_tm (TmFreeVar y) (Suc (Suc j)) e2)) = []"
      using free_dollars_vsubst_tm by auto
    hence "\<not> (\<exists>\<rho>'. \<rho>' is free dollar sc in
      (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_tm (TmFreeVar y) (Suc (Suc j)) e2)))"
      by auto
    hence "\<not> hasDollars (vsubst_tm (TmFreeVar x) (Suc (Suc i)) 
                         (vsubst_tm (TmFreeVar y) (Suc (Suc j)) e2))"
      using hasDollars_exists by blast
    thus False using H by auto
  next
     assume "INV9 let! vsubst_tm (TmFreeVar x) i (vsubst_tm (TmFreeVar y) j e)
                          vsubst_tm (TmFreeVar x) (Suc i) (vsubst_tm (TmFreeVar y) (Suc j) e1) in
                          vsubst_tm (TmFreeVar x) (Suc (Suc i)) 
                           (vsubst_tm (TmFreeVar y) (Suc (Suc j)) e2)"
      and H: "hasDollars e2"
    hence "\<not> hasDollars (vsubst_tm (TmFreeVar x) (Suc (Suc i)) 
                         (vsubst_tm (TmFreeVar y) (Suc (Suc j)) e2))"
      by auto
    hence "free_let_dollar_scope (vsubst_tm (TmFreeVar x) (Suc (Suc i)) 
                                   (vsubst_tm (TmFreeVar y) (Suc (Suc j)) e2)) = []"
      using not_hasDollars_exists by auto
    hence "free_let_dollar_scope e2 = []" using free_dollars_vsubst_tm by auto
    hence "\<not> (\<exists>\<rho>'. \<rho>' is free dollar sc in e2)" by auto
    hence "\<not> hasDollars e2" using hasDollars_exists by blast
    thus False using H by auto
  qed
qed auto

lemma weakening:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>"
  and "qualG \<Gamma>' \<preceq> qU"
  and "\<Gamma> \<union> \<Gamma>';\<Delta> \<Turnstile> OK"
  shows "\<Gamma>\<union>\<Gamma>';\<Delta>;M;Z \<turnstile> e : \<tau>"
using prems proof induct
  case (t_var \<Delta> Z M \<Gamma> \<tau> x) 
  hence "qualG ((\<Gamma> \<union> \<Gamma>') \<bar> x) \<preceq> qU" 
    by (auto simp add: qualGrelL_def without_def)
  thus ?case using t_var by auto
next
  case (t_abs \<Delta> X Z M \<Gamma> \<tau>1 \<tau>2 e q \<pi> Lx) 
  show ?case apply (rule Typing.t_abs[where L=Lx]) using t_abs proof auto
    fix x
    assume "(\<Gamma>\<union>\<Gamma>'),(x:\<tau>1);\<Delta> \<Turnstile> OK" and "(\<Gamma>\<union>\<Gamma>') \<sim> {x\<rhd>\<tau>1}" and "x \<notin> Lx"
    hence "\<Gamma>,(x:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma> \<sim> {x\<rhd>\<tau>1}" and "x \<notin> Lx"
      using wf_env_subset[of "(\<Gamma>\<union>\<Gamma>'),(x:\<tau>1)" _ "\<Gamma>,(x:\<tau>1)"]
    env_compat_subset[of "\<Gamma>\<union>\<Gamma>'" "{x\<rhd>\<tau>1}" \<Gamma>] by auto 
    moreover
    hence "\<Gamma>,(x:\<tau>1) \<union> \<Gamma>';\<Delta> \<Turnstile> OK" using prems wf_env_cons by auto
    ultimately
    have "(\<Gamma>,(x:\<tau>1)\<union>\<Gamma>');\<Delta>;M;(filterScFreeX X) \<turnstile> freshen_tm x e : \<tau>2" 
      using t_abs by blast
    hence "(\<Gamma>\<union>\<Gamma>'),(x:\<tau>1);\<Delta>;M;(filterScFreeX X) \<turnstile> vsubst_tm (TmFreeVar x) 0 e : \<tau>2" 
      by auto
    thus "(\<Gamma>\<union>\<Gamma>'),(x:\<tau>1);\<Delta>;M;{(\<rho>, EmptyScope) |\<rho>. ScFree \<rho> \<in> X} \<turnstile> vsubst_tm (TmFreeVar x) 0 e : \<tau>2"
      using t_abs by auto
  next
    assume "qualG \<Gamma> \<preceq> q"
    thus "qualG (\<Gamma> \<union> \<Gamma>') \<preceq> q" using t_abs.prems 
      by (auto simp add: qualGrelL_def qualT_unrestricted)
  qed
next
  case (t_let \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z e e1 e2 \<tau> \<tau>1 \<tau>2)
  hence "\<Gamma> \<union> \<Gamma>';\<Delta> \<Turnstile> OK" using wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>'" \<Delta> "\<Gamma>\<union>\<Gamma>'"]
    t_let by auto
  hence Ht: "\<Gamma> \<union> \<Gamma>';\<Delta>;M;Z \<turnstile> e : \<tau>" using t_let by auto
  have "\<Gamma> \<sim> \<Gamma>1" and "\<Gamma> \<sim> \<Gamma>2" 
    using t_let by (auto simp add: env_compat_def qualGrelL_def)
  hence Hcompat2: "\<Gamma> \<union> \<Gamma>' \<sim> \<Gamma>2" and Hcompat1: "\<Gamma> \<union> \<Gamma>' \<sim> \<Gamma>1" 
    using t_let env_compat_weakening_left by auto
  have "(\<Gamma>\<union> \<Gamma>') \<union> \<Gamma>1 \<union> \<Gamma>2 ;\<Delta>;M;Z2 \<union> Z1 \<union> Z \<turnstile> let! e e1 in e2 : \<tau>2"
    apply (rule Typing.t_let[of \<Gamma>1 \<Gamma>2 "\<Gamma>\<union>\<Gamma>'"]) 
    using t_let Ht Hcompat1 Hcompat2 by (auto simp add: Un_ac)
  thus ?case by (auto simp only: Un_ac)
next
  case (t_let' \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z \<rho> x e1 e2 \<tau> \<tau>1 \<tau>2 Zf)
  hence "\<Gamma> \<union> \<Gamma>';\<Delta> \<Turnstile> OK" using wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>'" \<Delta> "\<Gamma>\<union>\<Gamma>'"] t_let' 
    by auto
  hence Ht: "\<Gamma> \<union> \<Gamma>';\<Delta>;M;Z \<turnstile> TmFreeVar x : convT \<rho> \<tau>" using t_let' by auto
  have "\<Gamma> \<sim> \<Gamma>1" and "\<Gamma> \<sim> \<Gamma>2" 
    using t_let' by (auto simp add: env_compat_def qualGrelL_def)
  hence Hcompat2: "\<Gamma> \<union> \<Gamma>' \<sim> \<Gamma>2" and Hcompat1: "\<Gamma> \<union> \<Gamma>' \<sim> \<Gamma>1" 
    using t_let' env_compat_weakening_left by auto
  have "(\<Gamma>\<union> \<Gamma>') \<union> \<Gamma>1 \<union> \<Gamma>2 ;\<Delta>;M;Z2 \<union> Z1 \<union> Z \<turnstile> let$ \<rho> (TmFreeVar x) e1 in e2 : \<tau>2"
    apply (rule Typing.t_let'[of \<Gamma>1 \<Gamma>2 "\<Gamma>\<union>\<Gamma>'"]) 
    using t_let' Ht Hcompat1 Hcompat2 by (auto simp add: Un_ac)
  thus ?case by (auto simp only: Un_ac)
next
  case (t_assign \<Delta> Z1 Z2 M \<Gamma>1 \<Gamma>2 \<tau> e1 e2 \<pi>) 
  have "\<Gamma>' \<sim> \<Gamma>2" using t_assign by (auto simp add: env_compat_def qualGrelL_def)
  hence Hcompat: "\<Gamma>1 \<union> \<Gamma>' \<sim> \<Gamma>2" using t_assign env_compat_weakening_left by auto
  have "\<Gamma>1 \<union> \<Gamma>';\<Delta> \<Turnstile> OK" using wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>'" \<Delta> "\<Gamma>1\<union>\<Gamma>'"] 
    using t_assign by (auto simp only: Un_ac)
  hence Ht: "\<Gamma>1 \<union> \<Gamma>';\<Delta>;M;Z1 \<turnstile> e1 : Qual qU \<pi> (TyRef \<tau>)" using t_assign by auto
  have "\<Gamma>1 \<union> \<Gamma>' \<union> \<Gamma>2;\<Delta>;M;Z1 \<union> Z2 \<turnstile> e1 := e2 : Qual qU ScBot TyUnit"
    apply (rule Typing.t_assign[of "\<Gamma>1\<union>\<Gamma>'" \<Gamma>2])
    using t_assign Ht Hcompat by (auto simp add: Un_ac)
  thus ?case by (auto simp only: Un_ac)
next
  case (t_swap \<Delta> Z1 Z2 M \<Gamma>1 \<Gamma>2 \<tau> e1 e2 \<pi>) 
  have "\<Gamma>' \<sim> \<Gamma>2" using t_swap by (auto simp add: env_compat_def qualGrelL_def)
  hence Hcompat: "\<Gamma>1 \<union> \<Gamma>' \<sim> \<Gamma>2" using t_swap env_compat_weakening_left by auto
  have "\<Gamma>1 \<union> \<Gamma>';\<Delta> \<Turnstile> OK" using wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>'" \<Delta> "\<Gamma>1\<union>\<Gamma>'"] 
    using t_swap by (auto simp only: Un_ac)
  hence Ht: "\<Gamma>1 \<union> \<Gamma>';\<Delta>;M;Z1 \<turnstile> e1 : Qual qU \<pi> (TyRef \<tau>)" using t_swap by auto
  have "\<Gamma>1 \<union> \<Gamma>' \<union> \<Gamma>2;\<Delta>;M;Z2 \<union> Z1 \<turnstile> e1 :=: e2 : \<tau>"
    apply (rule Typing.t_swap[of "\<Gamma>1\<union>\<Gamma>'" \<Gamma>2])
    using t_swap Ht Hcompat by (auto simp add: Un_ac)
  thus ?case by (auto simp only: Un_ac)
next
  case (t_app \<Delta> X Z1 Z2 Z M \<Gamma>1 \<Gamma>2 \<tau>1 \<tau>2 e1 e2 q)
  have "\<Gamma>' \<sim> \<Gamma>2" using t_app by (auto simp add: env_compat_def qualGrelL_def)
  hence Hcompat: "\<Gamma>1 \<union> \<Gamma>' \<sim> \<Gamma>2" using t_app env_compat_weakening_left by auto
  have "\<Gamma>1 \<union> \<Gamma>';\<Delta> \<Turnstile> OK" using wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>'" \<Delta> "\<Gamma>1\<union>\<Gamma>'"] 
    using t_app by (auto simp only: Un_ac)
  hence Ht: "\<Gamma>1 \<union> \<Gamma>';\<Delta>;M;Z1 \<turnstile> e1 : Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)" using t_app by auto
  have "\<Gamma>1 \<union> \<Gamma>' \<union> \<Gamma>2;\<Delta>;M;Z1\<union>Z2\<union>(filterScFreeX X)\<union>Z \<turnstile> e1 \<cdot> e2 : \<tau>2"
    apply (rule Typing.t_app[of "\<Gamma>1\<union>\<Gamma>'" \<Gamma>2])
    using t_app Ht Hcompat by (auto simp add: Un_ac)
  thus ?case by (auto simp only: Un_ac)
next
  case (t_poly \<Delta> Z X M \<Gamma> \<tau> v q \<pi> La) 
  {
    fix a
    assume Hla: "a \<notin> La"
    have "\<Gamma> \<union> \<Gamma>';\<Delta>,(a:\<star>) \<Turnstile> OK" using t_poly poly_env_ty_cons by auto
    hence "\<Gamma> \<union> \<Gamma>';\<Delta>,(a:\<star>);M;(filterScFreeX X) \<turnstile> freshen_tmty a v : freshen_ty a \<tau>" 
      using t_poly Hla by auto
  }
  moreover
  have "qualG (\<Gamma> \<union> \<Gamma>') \<preceq> q" using t_poly by (auto simp add: qualGrelL_def qualT_unrestricted)
  ultimately
  show ?case using t_poly by auto
next
  case t_ptyapp show ?case apply rule using t_ptyapp by auto
qed (auto simp add: env_compat_weakening_left qualGrelL_def)

lemma eval_preserves_let_dollars:
  assumes "\<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> \<not> Z defines \<rho>"
  and "Z \<subseteq> Zs"
  and Hinv: "INV7 e e' Zs"
  shows "\<forall>\<rho>. \<rho> is free dollar sc in e' \<longrightarrow> \<not> Z defines \<rho>"
proof auto
  fix \<rho>
  assume Hassm1: "\<rho> is free dollar sc in e'"
    and Hassm2: "\<rho> \<in> Environ.dom Z"
  thus False using assms by (cases "\<rho> is free dollar sc in e", auto simp add: dom_def INV7)
qed

lemma fresh_dollars_are_free:
  assumes "\<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> \<rho> \<notin> Environ.dom Z" 
  and "INV7 e e' Zs"
  and "Z \<subseteq> Zs"
  shows "\<forall>\<rho>. \<rho> is free dollar sc in e' \<longrightarrow> \<rho> \<notin> Environ.dom Z"
using prems proof auto
  fix \<rho>
  assume Hr: "\<rho> is free dollar sc in e'"
    and Hfalse: "\<rho> \<in> dom Z"
  {
    assume "\<rho> is free dollar sc in e"
    hence "\<rho> \<notin> Environ.dom Z" using assms by auto
  }
  moreover
  {
    assume "\<not> \<rho> is free dollar sc in e"
    hence "\<not> Zs defines \<rho>" using assms Hr by auto
    hence "\<not> Z defines \<rho>" using assms by (auto simp only: dom_def)
  }
  ultimately
  have "\<not> Z defines \<rho>" by auto
  thus False using Hfalse by auto
qed

lemma free_dollars_in_scopeset:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>"
  and "\<rho> is free dollar sc in e"
  and "INV9 e"
  shows "Z defines \<rho>"
proof -
have H: "Z \<Turnstile> ScopeSet" using ty_wf_scope prems by auto
show ?thesis using prems H proof induct
  case (t_let' \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z \<rho>n x e1 e2 \<tau> \<tau>1 \<tau>2 Zf Lx Ly) 
  have "\<Gamma>;\<Delta>;M;Z \<turnstile> TmFreeVar x : convT \<rho>n \<tau>" by assumption
  hence "validScope (scopeT (convT \<rho>n \<tau>)) Z" using ty_valid_scope by auto
  hence Hrn: "Z defines \<rho>n" by (cases \<tau>, auto)
  have "\<rho> = \<rho>n \<or> \<rho> is free dollar sc in e1 \<or> \<rho> is free dollar sc in e2" using t_let' by auto
  thus ?case proof auto
    show "\<rho>n \<in> dom (Z2 \<union> Z1 \<union> Z)" using Hrn by (auto simp add: dom_def)
  next
    assume Hassm: "\<rho> is free dollar sc in e1"
    hence Hin: "Zf defines \<rho>" using t_let' free_dollars_vsubst by (auto simp add: wf_scopeset_def)
    have "\<rho> \<notin> Environ.dom (Z\<union>Z2)" using Hassm t_let' by auto
    hence Hneq: "\<rho> \<noteq> \<rho>n" using Hrn by (auto simp add: dom_def)
    have "Zf = Z1 \<or> Zf = Z1,(\<rho>n:\<section>)" by assumption
    hence "Z1 defines \<rho>" using Hin Hneq by auto
    thus "\<rho> \<in> Environ.dom (Z2 \<union> Z1 \<union> Z)" by (auto simp add: dom_def)
  next
    assume Hass: "\<rho> is free dollar sc in e2"
    have Hlx: "finite Lx" and Hly: "finite Ly" by assumption
    have "\<Delta> \<Turnstile>' convT \<rho>n \<tau>" using t_let' poly_ty_wf_type by auto
    hence H\<tau>: "\<Delta> \<Turnstile>' \<tau>" by (cases \<tau>, auto)
    have H\<tau>1: "\<Delta> \<Turnstile>' \<tau>1" using t_let' poly_ty_wf_type by auto
    have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
    then obtain x y where "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
      and "x \<notin> Lx" and "y \<notin> Ly" and "x \<noteq> y"
      using new_variable_let[where Lx=Lx and Ly=Ly] H\<tau> H\<tau>1 Hlx Hly by blast
    moreover
    have "INV9 e2" using t_let' by auto
    hence "INV9 vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2)" using inv9_vsubst_tm by auto 
    ultimately
    have "Z2 defines \<rho>" using t_let' free_dollars_vsubst_tm Hass by (auto simp add: wf_scopeset_def)
    thus "\<rho> \<in> Environ.dom (Z2 \<union> Z1 \<union> Z)" by (auto simp add: dom_def)
  qed
next
  case (t_let \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z e e1 e2 \<tau> \<tau>1 \<tau>2 Lx Ly)
  hence "\<rho> is free dollar sc in e \<or> \<rho> is free dollar sc in e1 \<or> \<rho> is free dollar sc in e2" 
    by auto
  have Hlx: "finite Lx" and Hly: "finite Ly" by assumption
  have "\<Gamma>1;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>1] by auto
  moreover
  have "Z1\<union>Z2 \<Turnstile> ScopeSet" using t_let by (auto simp add: wf_scopeset_def)
  moreover
  have H\<tau>: "\<Delta> \<Turnstile>' \<tau>" using poly_ty_wf_type t_let by auto
  ultimately
  obtain \<rho>out xf
    where Hc1: "\<Gamma>1,(xf:convT \<rho>out \<tau>);\<Delta> \<Turnstile> OK" and Hc2: "\<Gamma>1 \<sim> {xf\<rhd>convT \<rho>out \<tau>}" 
    and Hc3: "\<rho>out \<notin> set (free_let_dollar_scope e1)" and Hc4: "\<rho>out \<notin> Environ.dom (Z2 \<union> Z1)"
    and Hc5: "xf \<notin> Lx"
    using new_scope_let[where Lx=Lx] Hlx by blast
  hence Htyp: "\<Gamma>1,(xf:convT \<rho>out \<tau>);\<Delta>;M;Z1 \<turnstile> freshen_tm xf (freshen_sc \<rho>out e1) : \<tau>1 \<or>
    \<Gamma>1,(xf:convT \<rho>out \<tau>);\<Delta>;M;Z1,(\<rho>out:\<section>) \<turnstile> freshen_tm xf (freshen_sc \<rho>out e1) : \<tau>1" 
    using t_let by (auto simp add: dom_def wf_scopeset_def)
  hence H\<tau>1: "\<Delta> \<Turnstile>' \<tau>1" using poly_ty_wf_type by auto
  show ?case using t_let proof auto
    assume "\<rho> is free dollar sc in e"
    hence "\<rho> \<in> Environ.dom Z" using t_let by (auto simp add: wf_scopeset_def)
    thus "\<rho> \<in> Environ.dom (Z2 \<union> Z1 \<union> Z)" by (auto simp add: dom_def)
  next
    assume Hassm: "\<rho> is free dollar sc in e1"
    moreover
    have "INV9 e1" using t_let by auto
    hence "INV9 freshen_tm xf (freshen_sc \<rho>out e1)" using inv9_vsubst by auto
    ultimately
    have Hin: "\<rho> \<in> Environ.dom Z1" using t_let.hyps free_dollars_vsubst Hassm Hc1 Hc2 Hc3 Hc4 Hc5
      by (auto simp add: wf_scopeset_def)
    have "\<rho> \<noteq> \<rho>out" using Hc3 Hassm by auto
    hence "\<rho> \<in> Environ.dom Z1" using Hin by (auto simp add: dom_def)
    thus "\<rho> \<in> Environ.dom (Z2 \<union> Z1 \<union> Z)" by (auto simp add: dom_def)
  next
    assume Hass: "\<rho> is free dollar sc in e2"
    have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
    then obtain x y where "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
      and "x \<notin> Lx" and "y \<notin> Ly" and "x \<noteq> y"
      using new_variable_let[where Lx=Lx and Ly=Ly] H\<tau> H\<tau>1 Hlx Hly by blast
    moreover
    have "INV9 e2" using t_let by auto
    hence "INV9 vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2)" using inv9_vsubst_tm by auto
    ultimately
    have "Z2 defines \<rho>" using t_let free_dollars_vsubst_tm Hass by (auto simp add: wf_scopeset_def)
    thus "\<rho> \<in> Environ.dom (Z2 \<union> Z1 \<union> Z)" by (auto simp add: dom_def)
  qed
next
  case (t_abs \<Delta> X Z M \<Gamma> \<tau>1 \<tau>2 e q \<pi> Lx)
  hence "value(\<lambda>^q^\<pi>[\<tau>1].e)" and "INV9 \<lambda>^q^\<pi>[\<tau>1].e" by auto
  hence "\<not> hasDollars (\<lambda>^q^\<pi>[\<tau>1].e)" using INV9_elim[of "\<lambda>^q^\<pi>[\<tau>1].e"] by auto
  hence "\<not> (\<exists>\<rho>. \<rho> is free dollar sc in (\<lambda>^q^\<pi>[\<tau>1].e))" using not_hasDollars_exists by auto
  moreover
  have "\<rho> is free dollar sc in (\<lambda>^q^\<pi>[\<tau>1].e)" by assumption
  ultimately
  have False by auto
  thus ?case by auto
next
  case (t_poly \<Delta> Z X M \<Gamma> \<tau> v q \<pi>) 
  hence "value(\<Lambda>^q^\<pi>[*].v)" and "INV9 \<Lambda>^q^\<pi>[*].v" by auto
  hence "\<not> hasDollars (\<Lambda>^q^\<pi>[*].v)" using INV9_elim[of "\<Lambda>^q^\<pi>[*].v"] by auto
  hence "\<not> (\<exists>\<rho>. \<rho> is free dollar sc in (\<Lambda>^q^\<pi>[*].v))" using not_hasDollars_exists by auto
  moreover
  have "\<rho> is free dollar sc in (\<Lambda>^q^\<pi>[*].v)" by assumption
  ultimately
  have False by auto
  thus ?case by auto
next
  case (t_deref \<pi> \<Delta> Z Z\<tau> M \<Gamma> \<tau> e)
  hence "\<rho> is free dollar sc in e" by auto
  moreover
  have "Z \<Turnstile> ScopeSet" using ty_wf_scope t_deref by auto
  ultimately
  have "\<rho> \<in> dom Z" using t_deref by auto
  thus ?case using t_deref by (auto simp add: dom_def) 
next
  case (t_free \<Delta> Z Z\<tau> M \<Gamma> \<tau> e)
  hence "\<rho> is free dollar sc in e" by auto
  moreover
  have "Z \<Turnstile> ScopeSet" using ty_wf_scope t_free by auto
  ultimately
  have "\<rho> \<in> dom Z" using t_free by auto
  thus ?case using t_free by (auto simp add: dom_def) 
next
  case (t_app \<Delta> X Z1 Z2 Z M \<Gamma>1 \<Gamma>2 \<tau>1 \<tau>2 e1 e2 q) 
  hence "\<rho> is free dollar sc in e1 \<or> \<rho> is free dollar sc in e2" by auto
  moreover
  have "Z1 \<Turnstile> ScopeSet" and "Z2 \<Turnstile> ScopeSet" using t_app ty_wf_scope by auto
  ultimately
  have "\<rho> \<in> Environ.dom Z1 \<or> \<rho> \<in> Environ.dom Z2" using t_app by auto
  thus ?case by (cases, auto simp add: dom_def)
next 
  case (t_ptyapp \<Delta> Z Z\<tau> M \<Gamma> \<tau> \<phi> e q X) 
  hence "\<rho> \<in> Environ.dom Z" using t_ptyapp.hyps ty_wf_scope by auto
  thus ?case by (auto simp add: dom_def wf_scopeset_def)
qed (auto simp add: dom_def wf_scopeset_def)
qed

lemma eval_preserves_let_dollars_scope:
  assumes "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom Z1"
  and "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom Z2"
  and "INV7 e1 e1' Zs"
  and "INV8 Ze' Z1 e1'"
  and "INV9 e2"
  and "Z1\<union>Z2 \<subseteq> Zs"
  and "\<Gamma>;\<Delta>;M;Z2 \<turnstile> e2 : \<tau>"
  shows "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom Ze'"
proof auto
  fix \<rho>
  assume Hassm: "\<rho> is free dollar sc in e2"
    and Hfalse: "\<rho> \<in> Environ.dom Ze'"
  have H1: "Z2 defines \<rho>" using free_dollars_in_scopeset Hassm assms by auto
  {
    assume "\<rho> is free dollar sc in e1"
    hence "\<not> Z2 defines \<rho>" using assms by auto
    hence False using H1 by auto
  }
  moreover
  {
    assume Hass: "\<not> \<rho> is free dollar sc in e1"
    { 
      assume "\<rho> is free dollar sc in e1'"
      hence "\<not> Zs defines \<rho>" using INV7 Hass assms by auto
      moreover
      have "Z1\<union>Z2 \<subseteq> Zs" by assumption
      ultimately
      have "\<not> Z2 defines \<rho>" by (auto simp add: dom_def)
      hence False using H1 by auto
    }
    moreover
    {
      assume Hass2: "\<not> \<rho> is free dollar sc in e1'"
      have "\<rho> \<notin> Environ.dom Z1" using Hassm assms by auto
      hence "\<not> Ze' defines \<rho>" using INV8 Hass2 assms by auto
      hence False using Hfalse by auto
    }
    ultimately have False by auto
  }
  ultimately show False by auto
qed

end
