theory Semantics
imports Syntax

begin

datatype StTerm = StTerm "tm"
types store = "(name * StTerm) set"

syntax
  var_stterm :: "name \<Rightarrow> StTerm \<Rightarrow> (name * StTerm)"  ("(_\<mapsto>_)" [90,90] 90)
translations
  "(x\<mapsto>\<tau>)" \<rightleftharpoons> "(x, StTerm \<tau>)"

syntax
  addmapping :: "store \<Rightarrow> name \<Rightarrow> tm \<Rightarrow> store"      ("_','(_\<mapsto>_')" [90,0,0] 91)
translations
  "S,(x\<mapsto>v)" \<rightleftharpoons> "insert (x\<mapsto>v) S"

consts qualV :: "tm \<Rightarrow> qual"
primrec
  "qualV (TmUnit q \<pi>) = q"
  "qualV (\<lambda>^q^\<pi>[\<tau>].e) = q"
  "qualV (TmLoc q \<pi> l) = q"
  "qualV (\<Lambda>^q^\<pi>[*].v) = q"

consts scopeV :: "tm \<Rightarrow> scope"
primrec
  "scopeV (TmUnit q \<pi>) = \<pi>"
  "scopeV (\<lambda>^q^\<pi>[\<tau>].e) = \<pi>"
  "scopeV (TmLoc q \<pi> l) = \<pi>"
  "scopeV (\<Lambda>^q^\<pi>[*].v) = \<pi>"

consts convV :: "name \<Rightarrow> tm \<Rightarrow> tm"
primrec
  "convV \<rho> (TmUnit q \<pi>) = TmUnit qU (ScFree \<rho>)"
  "convV \<rho> (\<lambda>^q^\<pi>[\<tau>].e) = \<lambda>^qU^(ScFree \<rho>)[\<tau>].e"
  "convV \<rho> (TmLoc q \<pi> l) = TmLoc qU (ScFree \<rho>) l"
  "convV \<rho> (\<Lambda>^q^\<pi>[*].v) = \<Lambda>^qU^(ScFree \<rho>)[*].v"

consts convRevV :: "tm \<Rightarrow> tm"
primrec
  "convRevV (TmUnit q \<pi>) = TmUnit qL ScBot"
  "convRevV (\<lambda>^q^\<pi>[\<tau>].e) = \<lambda>^qL^ScBot[\<tau>].e"
  "convRevV (TmLoc q \<pi> l) = TmLoc qL ScBot l"
  "convRevV (\<Lambda>^q^\<pi>[*].v) = \<Lambda>^qL^ScBot[*].v"

inductive Var :: "tm \<Rightarrow> bool" ("isVar _")
  where
  v_var[intro]  : "isVar (TmFreeVar x)"

inductive_cases isVar_var_elim [elim!]: "isVar e"

constdefs update :: "store \<Rightarrow> name \<Rightarrow> tm \<Rightarrow> store"
  "update S x v \<equiv> { p | p. p \<in> S \<and> fst p \<noteq> x},(x\<mapsto>v)"

constdefs
  wf_store :: "store \<Rightarrow> bool" ("_ \<Turnstile> Store" [50] 50)
  "wf_store S \<equiv> finite S \<and> (\<forall>x. \<forall>v1. \<forall>v2. (x\<mapsto>v1) \<in> S \<and> (x\<mapsto>v2) \<in> S \<longrightarrow> v1 = v2 \<and> value v1)"

constdefs filterStore :: "store \<Rightarrow> scopeset \<Rightarrow> store"
  "filterStore S Z \<equiv> { (x\<mapsto>v) | x v. (x\<mapsto>v) \<in> S \<and> validScope (scopeV v) Z }"

lemma wf_store_cons:
  assumes "S \<Turnstile> Store"
  and "\<not> S defines z"
  and "value v"
  shows "S,(z\<mapsto>v) \<Turnstile> Store"
using prems by (auto simp add: wf_store_def dom_def)

lemma store_ok_is_value:
  assumes "S \<Turnstile> Store"
  and "(x\<mapsto>v) \<in> S"
  shows "value v"
using prems by (auto simp add: wf_store_def)

lemma store_ok_exists_value:
  assumes "S \<Turnstile> Store"
  and H: "S defines x"
  shows "\<exists>v. (x\<mapsto>v) \<in> S \<and> value v"
using prems proof (auto simp add: dom_def store_ok_is_value)
  fix b
  assume "(x, b) \<in> S"
  then obtain v where "(x\<mapsto>v) \<in> S" by (cases b, auto)
  moreover have "value v" using prems by (auto simp add: store_ok_is_value)
  ultimately show "\<exists>v. x\<mapsto>v \<in> S \<and> value v" by auto
qed

lemma wf_store_defines_without:
  assumes "S \<Turnstile> Store"
  and "(x\<mapsto>v) \<in> S"
  shows "S = (S\<bar>x),(x\<mapsto>v)"
using prems proof (auto simp add: without_def)
  fix y b
  assume "(y\<mapsto>v) \<in> S" and "(y, b) \<in> S" and "b \<noteq> StTerm v"
  then obtain v' where "b = StTerm v'" by (cases b)
  hence "v = v'" using prems by (auto simp add: wf_store_def)
  thus False using prems by auto
qed

lemma wf_store_subset:
  assumes "S \<Turnstile> Store"
  and "S' \<subseteq> S"
  shows "S' \<Turnstile> Store"
using prems finite_subset[of S' S] by (unfold wf_store_def, fast)

lemma wf_store_sanity:
  assumes "S \<Turnstile> Store"
  and "x\<mapsto>v1 \<in> S"
  and "x\<mapsto>v2 \<in> S"
  shows "v1 = v2"
using prems by (auto simp add: wf_store_def)

constdefs
  wf_mem :: "store \<Rightarrow> bool" ("_ \<Turnstile> Mem" [50] 50)
  "wf_mem \<mu> \<equiv> finite \<mu> \<and> (\<forall>x. \<forall>v1. \<forall>v2. (x\<mapsto>v1) \<in> \<mu> \<and> (x\<mapsto>v2) \<in> \<mu> \<longrightarrow> v1 = v2 \<and> isVar v1)"

lemma mem_ok_is_var:
  assumes "\<mu> \<Turnstile> Mem"
  and "(l\<mapsto>v) \<in> \<mu>"
  shows "isVar v"
using prems by (auto simp add: wf_mem_def)

lemma mem_ok_exists_var:
  assumes "\<mu> \<Turnstile> Mem"
  and "\<mu> defines l"
  shows "\<exists>x. (l\<mapsto>TmFreeVar x) \<in> \<mu>"
using prems proof (auto simp add: wf_mem_def dom_def)
  fix b
  assume "(l, b) \<in> \<mu>"
  then obtain v where "(l\<mapsto>v) \<in> \<mu>" by (cases b, auto)
  moreover have "isVar v" using prems by (auto simp add: mem_ok_is_var)
  ultimately show "\<exists>x. (l\<mapsto>TmFreeVar x) \<in> \<mu>" by auto
qed

lemma wf_mem_cons:
  assumes "\<mu> \<Turnstile> Mem"
  and "\<not> \<mu> defines z"
  and "isVar e"
  shows "\<mu>,(z\<mapsto>e) \<Turnstile> Mem"
using prems by (auto simp add: wf_mem_def dom_def)

lemma mem_update_ok:
  assumes "\<mu> \<Turnstile> Mem"
  shows "update \<mu> l (TmFreeVar x) \<Turnstile> Mem"
using prems by (auto simp add: wf_mem_def update_def)

lemma wf_mem_subset:
  assumes "\<mu> \<Turnstile> Mem"
  and "\<mu>' \<subseteq> \<mu>"
  shows "\<mu>' \<Turnstile> Mem"
using prems finite_subset[of \<mu>' \<mu>] by (unfold wf_mem_def, fast)

consts ValComp :: "store \<Rightarrow> name \<Rightarrow> store \<Rightarrow> tm \<Rightarrow> bool"
  ("_;_\<Down>_;_" [50,80,50,80] 50)
defs ValCom[simp]:
  "S;z\<Down>S';v \<equiv> (z\<mapsto>v) \<in> S \<and> S' = (if qualV v = qU then S else S\<bar>z)"

constdefs free_loc_store :: "store \<Rightarrow> name set"
  "free_loc_store S \<equiv> \<Union> { set (free_loc v) | z v. (z\<mapsto>v) \<in> S }"

syntax
  is_free_in_store :: "name \<Rightarrow> store \<Rightarrow> bool" ("_ free loc in st _" [900,50] 50)

translations
  "a free loc in st S" \<rightleftharpoons> "a \<in> free_loc_store S"

constdefs free_var_store :: "store \<Rightarrow> name set"
  "free_var_store S \<equiv> \<Union> { set (free_tm v) | z v. (z\<mapsto>v) \<in> S }"

syntax
  is_free_var_in_store :: "name \<Rightarrow> store \<Rightarrow> bool" ("_ free var in st _" [900,50] 50)

translations
  "a free var in st S" \<rightleftharpoons> "a \<in> free_var_store S"

constdefs free_scope_store :: "store \<Rightarrow> name set"
  "free_scope_store S \<equiv> \<Union> { set (free_scope_scope (scopeV v)) | z v. (z\<mapsto>v) \<in> S }"

syntax
  is_free_scope_in_store :: "name \<Rightarrow> store \<Rightarrow> bool" ("_ free sc in st _" [900,50] 50)

translations
  "a free sc in st S" \<rightleftharpoons> "a \<in> free_scope_store S"

lemma free_scope_store_cons_eq:
  assumes Hv: "value v"
  shows "free_scope_store (S,(x\<mapsto> convV \<rho> v)) =
  free_scope_store S \<union> {\<rho>}"
proof -
  have Heq: "free_scope_store (S,(x\<mapsto> convV \<rho> v)) =
  free_scope_store S \<union> (set (free_scope_scope (scopeV (convV \<rho> v))))"
    by (auto simp add: free_scope_store_def)
  have "free_scope_store (S,(x\<mapsto> convV \<rho> v)) = free_scope_store S \<union> (set [\<rho>])"
    using prems Heq by (cases v, auto)
  thus ?thesis by auto
qed

inductive Eval :: "store \<Rightarrow> store \<Rightarrow> tm \<Rightarrow> store \<Rightarrow> store \<Rightarrow> tm \<Rightarrow> bool"
("_';_';_\<hookrightarrow>_';_';_" [50,50,80,50,50,80] 50)
  where
    e_val[intro!]       : "\<And>S \<mu> v z.
                           \<lbrakk> \<not> S defines z; S \<Turnstile> Store; \<mu> \<Turnstile> Mem; value v \<rbrakk> \<Longrightarrow>
                             S;\<mu>;v \<hookrightarrow> S,(z\<mapsto>v);\<mu>;TmFreeVar z"
  | e_app_p1[intro!]    : "\<And>S S' \<mu> \<mu>' e1 e1' e2.
                           \<lbrakk> S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1' \<rbrakk> \<Longrightarrow>
                             S;\<mu>;e1\<cdot>e2 \<hookrightarrow> S';\<mu>';e1'\<cdot>e2"
  | e_app_p2[intro!]    : "\<And>S S' \<mu> \<mu>' x e2' e2.
                           \<lbrakk> S;\<mu>;e2\<hookrightarrow>S';\<mu>';e2' \<rbrakk> \<Longrightarrow>
                             S;\<mu>;(TmFreeVar x)\<cdot>e2 \<hookrightarrow> S';\<mu>';(TmFreeVar x)\<cdot>e2'"
  | e_beta[intro!]      : "\<And>S S' \<mu> \<tau> e q y z \<pi>.
                           \<lbrakk> S \<Turnstile> Store; \<mu> \<Turnstile> Mem; S;z\<Down>S';(\<lambda>^q^\<pi>[\<tau>].e) \<rbrakk> \<Longrightarrow>
                             S;\<mu>;(TmFreeVar z)\<cdot>(TmFreeVar y) \<hookrightarrow> S';\<mu>;freshen_tm y e"
  | e_tyapp_p[intro!]   : "\<And>S S' \<mu> \<mu>' e e' \<phi>.
                           \<lbrakk> S;\<mu>;e\<hookrightarrow>S';\<mu>';e' \<rbrakk> \<Longrightarrow> S;\<mu>;e\<langle>\<phi>\<rangle> \<hookrightarrow> S';\<mu>';e'\<langle>\<phi>\<rangle>"
  | e_tybeta[intro!]    : "\<And>S S' \<mu> z e q \<pi> \<phi>.
                           \<lbrakk> S \<Turnstile> Store; \<mu> \<Turnstile> Mem; S;z\<Down>S';\<Lambda>^q^\<pi>[*].e \<rbrakk> \<Longrightarrow>
                           S;\<mu>;(TmFreeVar z)\<langle>\<phi>\<rangle> \<hookrightarrow> S';\<mu>;vsubst_tmty \<phi> 0 e"
  | e_new_p[intro!]     : "\<And>S S' \<mu> \<mu>' e e'.
                           \<lbrakk> S;\<mu>;e\<hookrightarrow>S';\<mu>';e' \<rbrakk> \<Longrightarrow>
                             S;\<mu>;TmNew e \<hookrightarrow> S';\<mu>';TmNew e'"
  | e_new_c[intro!]     : "\<And>S \<mu> l x z.
                           \<lbrakk> S \<Turnstile> Store; \<mu> \<Turnstile> Mem; \<not> \<mu> defines l; \<not> S defines z \<rbrakk> \<Longrightarrow>
                             S;\<mu>;TmNew (TmFreeVar x) \<hookrightarrow>
                               S,(z\<mapsto>TmLoc qL ScBot l);\<mu>,(l\<mapsto>TmFreeVar x);TmFreeVar z"
  | e_free_p[intro!]    : "\<And>S S' \<mu> \<mu>' e e'.
                           \<lbrakk> S;\<mu>;e\<hookrightarrow>S';\<mu>';e' \<rbrakk> \<Longrightarrow>
                             S;\<mu>;TmFree e \<hookrightarrow> S';\<mu>';TmFree e'"
  | e_free_c[intro!]    : "\<And>S S' \<mu> l q x y \<pi>.
                           \<lbrakk> S \<Turnstile> Store; \<mu> \<Turnstile> Mem; S;x\<Down>S';TmLoc q \<pi> l;
                             (l\<mapsto>TmFreeVar y) \<in> \<mu> \<rbrakk> \<Longrightarrow>
                             S;\<mu>;TmFree (TmFreeVar x) \<hookrightarrow> S';\<mu>\<bar>l;TmFreeVar y"
  | e_assign_p1[intro!] : "\<And>S S' \<mu> \<mu>' e1 e1' e2.
                           \<lbrakk> S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1' \<rbrakk> \<Longrightarrow>
                             S;\<mu>;e1 := e2 \<hookrightarrow> S';\<mu>';e1' := e2"
  | e_assign_p2[intro!] : "\<And>S S' \<mu> \<mu>' x e2' e2.
                           \<lbrakk> S;\<mu>;e2\<hookrightarrow>S';\<mu>';e2' \<rbrakk> \<Longrightarrow>
                             S;\<mu>;(TmFreeVar x) := e2 \<hookrightarrow> S';\<mu>';(TmFreeVar x) := e2'"
  | e_assign_c[intro!]  : "\<And>S S' \<mu> l q x y \<pi>.
                           \<lbrakk> S \<Turnstile> Store; \<mu> \<Turnstile> Mem; S;x\<Down>S';(TmLoc q \<pi> l) \<rbrakk> \<Longrightarrow>
                             S;\<mu>;(TmFreeVar x) := (TmFreeVar y) \<hookrightarrow>
                               S';update \<mu> l (TmFreeVar y); TmUnit qU ScBot"
  | e_swap_p1[intro!]   : "\<And>S S' \<mu> \<mu>' e1 e1' e2.
                           \<lbrakk> S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1' \<rbrakk> \<Longrightarrow>
                             S;\<mu>;e1 :=: e2 \<hookrightarrow> S';\<mu>';e1' :=: e2"
  | e_swap_p2[intro!]   : "\<And>S S' \<mu> \<mu>' x e2' e2.
                           \<lbrakk> S;\<mu>;e2\<hookrightarrow>S';\<mu>';e2' \<rbrakk> \<Longrightarrow>
                             S;\<mu>;(TmFreeVar x) :=: e2 \<hookrightarrow> S';\<mu>';(TmFreeVar x) :=: e2'"
  | e_swap_c[intro!]    : "\<And>S S' \<mu> l q x z y \<pi>.
                           \<lbrakk> S \<Turnstile> Store; \<mu> \<Turnstile> Mem; S;x\<Down>S';(TmLoc q \<pi> l); (l\<mapsto>TmFreeVar z) \<in> \<mu> \<rbrakk> \<Longrightarrow>
                             S;\<mu>;(TmFreeVar x) :=: (TmFreeVar y) \<hookrightarrow>
                               S';update \<mu> l (TmFreeVar y); TmFreeVar z"
  | e_deref_p[intro!]   : "\<And>S S' \<mu> \<mu>' e e'.
                           \<lbrakk> S;\<mu>;e\<hookrightarrow>S';\<mu>';e' \<rbrakk> \<Longrightarrow>
                             S;\<mu>;TmDeref e \<hookrightarrow> S';\<mu>';TmDeref e'"
  | e_deref_c[intro!]   : "\<And>S S' \<mu> l q x z \<pi>.
                           \<lbrakk> S \<Turnstile> Store; \<mu> \<Turnstile> Mem; S;x\<Down>S';(TmLoc q \<pi> l);
                             (l\<mapsto>TmFreeVar z) \<in> \<mu> \<rbrakk> \<Longrightarrow>
                             S;\<mu>;TmDeref (TmFreeVar x) \<hookrightarrow> S';\<mu>;TmFreeVar z"
  | e_let_p[intro!]     : "\<And>S S' \<mu> \<mu>' e e' e1 e2.
                           \<lbrakk> S \<Turnstile> Store; \<mu> \<Turnstile> Mem; S;\<mu>; e \<hookrightarrow> S';\<mu>';e' \<rbrakk> \<Longrightarrow>
                             S;\<mu>; let! (e) e1 in e2 \<hookrightarrow>
                             S';\<mu>'; let! (e') e1 in e2"
  | e_let_p'[intro!]    : "\<And>S \<mu> x e1 e2 \<rho> v.
                           \<lbrakk> S \<Turnstile> Store; \<mu> \<Turnstile> Mem; \<not> \<rho> free sc in st S;
                             S;x\<Down>S';v  \<rbrakk> \<Longrightarrow>
                             S;\<mu>; let! (TmFreeVar x) e1 in e2 \<hookrightarrow>
                             S',(x\<mapsto>convV \<rho> v);\<mu>; 
                               let$ \<rho> (TmFreeVar x) (freshen_tm x (freshen_sc \<rho> e1)) in e2"
  | e_let_p1[intro!]    : "\<And>S S' \<mu> \<mu>' x e1 e1' e2 \<rho>.
                           \<lbrakk> S \<Turnstile> Store; \<mu> \<Turnstile> Mem;
                             S;\<mu>; e1 \<hookrightarrow> S';\<mu>'; e1' \<rbrakk> \<Longrightarrow>
                             S;\<mu>; let$ \<rho> (TmFreeVar x) e1 in e2 \<hookrightarrow>
                              S';\<mu>'; let$ \<rho> (TmFreeVar x) e1' in e2"
  | e_let_c[intro!]    : "\<And>S S' \<mu> x z \<rho> v y e2 .
                           \<lbrakk> S \<Turnstile> Store; \<mu> \<Turnstile> Mem; S;x\<Down>S';v; \<not> S defines z \<rbrakk> \<Longrightarrow>
                             S;\<mu>; let$ \<rho> (TmFreeVar x) (TmFreeVar y) in e2 \<hookrightarrow>
                               S,(z\<mapsto>convRevV v);\<mu>; vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm z e2)"

inductive_cases e_eval_unit[elim!]:      "S;\<mu>;TmUnit q \<pi> \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_var[elim!]:       "S;\<mu>;TmVar n \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_freevar[elim!]:   "S;\<mu>;TmFreeVar n \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_app_p1[elim!]:    "S;\<mu>;e1 \<cdot> e2 \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_app_p2[elim!]:    "S;\<mu>;(TmFreeVar x) \<cdot> e2 \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_app_beta[elim!]:  "S;\<mu>;(TmFreeVar x) \<cdot> (TmFreeVar y) \<hookrightarrow>
                                           S';\<mu>';e'"
inductive_cases e_eval_poly[elim!]:      "S;\<mu>;\<Lambda>^q^\<pi>[*].e \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_tapp_beta[elim!]: "S;\<mu>;(TmFreeVar x) \<langle>\<phi>\<rangle> \<hookrightarrow> S';\<mu>';e'" 
inductive_cases e_eval_new_p[elim!]:     "S;\<mu>;TmNew e \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_new_c[elim!]:     "S;\<mu>;TmNew (TmFreeVar x) \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_free_p[elim!]:    "S;\<mu>;TmFree e \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_free_c[elim!]:    "S;\<mu>;TmFree (TmFreeVar x) \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_assign_p1[elim!]: "S;\<mu>;e1 := e2 \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_assign_p2[elim!]: "S;\<mu>;(TmFreeVar x):=e2 \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_assign_c[elim!]:  "S;\<mu>;(TmFreeVar x):=(TmFreeVar y) \<hookrightarrow>
                                           S';\<mu>';e'"
inductive_cases e_eval_swap_p1[elim!]:   "S;\<mu>;e1 :=: e2 \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_swap_p2[elim!]:   "S;\<mu>;(TmFreeVar x):=:e2 \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_swap_c[elim!]:    "S;\<mu>;(TmFreeVar x):=:(TmFreeVar y) \<hookrightarrow>
                                           S';\<mu>';e'"
inductive_cases e_eval_deref_p[elim!]:   "S;\<mu>;TmDeref e \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_deref_c[elim!]:   "S;\<mu>;TmDeref (TmFreeVar x) \<hookrightarrow>
                                           S';\<mu>';e'"
inductive_cases e_eval_loc[elim!]:       "S;\<mu>;TmLoc q \<pi> n \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_let_p[elim!]:     "S;\<mu>;let! (e) e1 in e2 \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_let_p'[elim!]:     "S;\<mu>;let$ \<rho> (e) e1 in e2 \<hookrightarrow> S';\<mu>';e'"
inductive_cases e_eval_let_p1[elim!]:    "S;\<mu>;let$ \<rho> (TmFreeVar x) e1 in e2 \<hookrightarrow>
                                           S';\<mu>';e'"
inductive_cases e_eval_let_c[elim!]:     "S;\<mu>;let$ \<rho> (TmFreeVar x) (TmFreeVar y)
                                          in e2 \<hookrightarrow> S';\<mu>';e'"

consts makes_progress :: "tm \<Rightarrow> store \<Rightarrow> store \<Rightarrow> bool"
defs makes_progress[simp]: "makes_progress e S \<mu> \<equiv> \<exists>e'. \<exists>S'. \<exists>\<mu>'. S;\<mu>;e \<hookrightarrow>
                                                                   S';\<mu>';e'"
consts not_stuck :: "tm \<Rightarrow> store \<Rightarrow> store \<Rightarrow> bool"
defs not_stuck[simp]: "not_stuck e S \<mu> \<equiv> isVar e \<or> makes_progress e S \<mu>"

consts stuck :: "tm \<Rightarrow> store \<Rightarrow> store \<Rightarrow> bool"
defs stuck[simp]: "stuck e S \<mu> \<equiv> \<not> not_stuck e S \<mu>"

inductive EvalMany :: "store \<Rightarrow> store \<Rightarrow> tm \<Rightarrow> store \<Rightarrow> store \<Rightarrow> tm \<Rightarrow> bool"
("_';_';_\<hookrightarrow>*_';_';_" [50,50,80,50,50,80] 50)
  where
    e_refl[intro!]  : "S;\<mu>;e \<hookrightarrow>* S;\<mu>;e"
  | e_trans[intro!] : "\<lbrakk> S;\<mu>;e \<hookrightarrow> S';\<mu>';e'; S';\<mu>';e' \<hookrightarrow>* S'';\<mu>'';e'' \<rbrakk> \<Longrightarrow>
                         S;\<mu>;e \<hookrightarrow>* S'';\<mu>'';e''"

end
