theory StoreTyping
imports Typing Semantics TypingLemmas

begin

inductive storeTyping :: "env \<Rightarrow> ty_env \<Rightarrow> scopeset \<Rightarrow> scopeset \<Rightarrow> store \<Rightarrow> env \<Rightarrow> bool" ("_;_;_;_ \<Turnstile> _:_" [50,50,50,50,50] 50)
  where
    st_nil[intro!]   : "\<lbrakk> M;\<Delta> \<Turnstile> OK \<rbrakk> \<Longrightarrow> M;\<Delta>;Z;Z' \<Turnstile> {} : {}"
  | st_cons[intro!]  : "\<And> \<Delta> Z Zs M S \<Gamma>s \<Gamma>v \<tau> v x Zl.
                        \<lbrakk> \<Gamma>v;\<Delta>;M;Z \<turnstile> v : \<tau>; M;\<Delta>;Zs;Zl \<Turnstile> S : \<Gamma>s \<union> \<Gamma>v; Z \<subseteq> Zs;
                          \<Gamma>s \<sim> \<Gamma>v; S \<Turnstile> Store; validScope (scopeV v) Zl;
                          \<Gamma>s \<union> \<Gamma>v;\<Delta> \<Turnstile> OK; value v; \<not> S defines x; un \<Gamma>v \<subseteq> un \<Gamma>s \<rbrakk> \<Longrightarrow>
                          M;\<Delta>;Zs;Zl \<Turnstile> S,(x\<mapsto>v) : \<Gamma>s,(x:\<tau>)"
  | st_dummy[intro!] : "\<And> \<Delta> Z Zl M S \<Gamma>s \<tau> v x.
                        \<lbrakk> M;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>s; \<not> validScope (scopeV v) Zl; \<Delta> \<Turnstile>' \<tau>;
                          \<not> S defines x; value v; scopeV v = scopeT \<tau> \<rbrakk> \<Longrightarrow>
                          M;\<Delta>;Z;Zl \<Turnstile> S,(x\<mapsto>v) : \<Gamma>s,(x:\<tau>)"

inductive_cases st_nil_elim [elim!]:  "M;\<Delta>;{};Zl \<Turnstile> {} : {}"
inductive_cases st_cons_elim [elim!]: "M;\<Delta>;Z;Zl \<Turnstile> S,(x\<mapsto>v) : \<Gamma>,(x:\<tau>)"

lemma store_dom_in:
  assumes "M;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>"
  and "\<Gamma> defines z"
  shows "S defines z"
using prems proof induct
  case st_cons
  thus ?case by (cases "x=z", auto simp add: dom_def)
qed (auto simp add: dom_def)

lemma store_dom_not_in:
  assumes "M;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>"
  and "\<not> S defines z"
  shows "\<not> \<Gamma> defines z"
using prems by (auto simp only: store_dom_in)

lemma storeTyping_wf_store:
  assumes "M;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>"
  shows "S \<Turnstile> Store"
using prems proof induct
  case st_nil thus ?case by (auto simp add: wf_store_def)
next
  case st_cons thus ?case by (auto simp add: wf_store_cons)
next
  case st_dummy thus ?case by (auto simp add: wf_store_cons)
qed 

lemma storeTyping_wf_env:
  assumes "M;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>"
  shows "\<Gamma>;\<Delta> \<Turnstile> OK"
using prems proof induct
  case st_nil thus ?case by (auto simp add: wf_env_def)
next
  case (st_cons \<Delta> Z Zs M S \<Gamma>s \<Gamma>v \<tau> v x Zl)
  have "\<Gamma>s;\<Delta> \<Turnstile> OK" using st_cons wf_env_subset[of "\<Gamma>s\<union>\<Gamma>v" \<Delta> \<Gamma>s] by auto
  moreover have "\<not> \<Gamma>s defines x"
    using st_cons store_dom_not_in[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>v" x] by (auto simp add: dom_def)
  ultimately show ?case using wf_env_cons st_cons poly_ty_wf_type by auto
next
  case (st_dummy \<Delta> Z Zl M S \<Gamma>s \<tau> v x)
  have "\<Gamma>s;\<Delta> \<Turnstile> OK" using st_dummy by auto
  moreover have "\<not> \<Gamma>s defines x"
    using st_dummy store_dom_not_in[of M \<Delta> Z Zl S _ x] by (auto simp add: dom_def)
  ultimately show ?case using wf_env_cons st_dummy by auto
qed

lemma storeTyping_wf_mem:
  assumes "M;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>"
  shows "M;\<Delta> \<Turnstile> OK"
using prems by induct

lemma storeTyping_weakening:
  assumes "M;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>s"
  and "M,(l:\<tau>);\<Delta> \<Turnstile> OK"
  shows "M,(l:\<tau>);\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>s"
using prems proof induct
  case (st_cons \<Delta> Z1 Z2 M S \<Gamma>s \<Gamma>v \<tau>z vz z Zl)
  hence "\<Gamma>v;\<Delta>;M,(l:\<tau>);Z1 \<turnstile> vz : \<tau>z" using prems weakening_m[where M'="{l\<rhd>\<tau>}"] 
    by auto
  thus ?case using st_cons storeTyping.st_cons by auto
qed auto

lemma storeTyping_no_loc:
  assumes "M;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>"
  and "\<not> M defines l"
  shows "\<not> l free loc in st (filterStore S Zl)"
using prems by (induct, auto simp add: ty_no_loc free_loc_store_def filterStore_def)

lemma storeTyping_sanity:
  assumes "M;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>s"
  and "(x\<mapsto>v) \<in> S"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>s"
  and "validScope (scopeV v) Zl"
  shows "qualV v = qualT \<tau> \<and> scopeV v = scopeT \<tau> \<and>
  (\<exists>\<Gamma>v Z'. \<Gamma>v;\<Delta>;M;Z' \<turnstile> v : \<tau> \<and> un \<Gamma>v \<subseteq> un \<Gamma>s \<and> Z' \<subseteq> Z)"
using prems proof induct
  case (st_cons \<Delta> Z Zs M S \<Gamma>s \<Gamma>c \<sigma> w z Zl)
  show ?case proof (cases "x=z")
    case True
    have "\<not> (\<Gamma>s\<union>\<Gamma>c) defines z" using prems store_dom_not_in by auto
    hence "\<not> \<Gamma>s defines z" by (auto simp add: dom_def)
    moreover
    have "\<Gamma>s;\<Delta> \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s\<union>\<Gamma>c" \<Delta> \<Gamma>s] by auto
    moreover
    have "\<Delta> \<Turnstile>' \<sigma>" using st_cons poly_ty_wf_type by auto
    ultimately
    have "\<Gamma>s,(z:\<sigma>);\<Delta> \<Turnstile> OK" using wf_env_cons by auto
    hence Hteq: "\<tau> = \<sigma>" using True wf_env_def[of "\<Gamma>s,(z:\<sigma>)"] st_cons.prems 
      by fast
    have "S,(z\<mapsto>w) \<Turnstile> Store" using st_cons wf_store_cons by auto
    hence Hweq: "v = w" using True wf_store_def[of "S,(z\<mapsto>w)"] st_cons.prems
      by fast
    have "un \<Gamma>s \<subseteq> un (\<Gamma>s,(x:\<tau>))" by (auto simp add: un_def)
    hence "un \<Gamma>c \<subseteq> un (\<Gamma>s,(x:\<tau>))" using st_cons by auto
    thus ?thesis using prems exI[of _ \<Gamma>c] Hteq Hweq 
      by (auto simp add: ty_qual_sanity ty_scope_sanity)
  next
    case False
    hence Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>s" using st_cons by auto
    have "(x\<mapsto>v) \<in> S" and "(x\<rhd>\<tau>) \<in> (\<Gamma>s \<union> \<Gamma>c)"
      using False st_cons by (auto simp add: without_def dom_def)
    then obtain \<Gamma>v Z'
      where Htyp: "\<Gamma>v;\<Delta>;M;Z' \<turnstile> v : \<tau>"
      and Hun: "un \<Gamma>v \<subseteq> un (\<Gamma>s \<union> \<Gamma>c)"
      and Hzsub: "Z' \<subseteq> Zs" using st_cons by auto
    have "un (\<Gamma>s \<union> \<Gamma>c) = un \<Gamma>s \<union> un \<Gamma>c" by (auto simp add: un_def)
    hence "un \<Gamma>v \<subseteq> un \<Gamma>s \<union> un \<Gamma>c" using Hun by auto
    hence "un \<Gamma>v \<subseteq> un \<Gamma>s" using prems by auto
    moreover have "un \<Gamma>s \<subseteq> un (\<Gamma>s,(z:\<sigma>))" by (auto simp add: un_def)
    ultimately have "un \<Gamma>v \<subseteq> un (\<Gamma>s,(z:\<sigma>))" by auto
    moreover have "value v" using prems store_ok_is_value by auto
    ultimately show ?thesis 
      using exI[of _ \<Gamma>v] exI[of _ Z'] Htyp Hzsub ty_qual_sanity ty_scope_sanity by auto
  qed
next 
  case (st_dummy \<Delta> Z Zl M S \<Gamma>s \<sigma> w z)
  show ?case proof (cases "x=z")
    case True
    have "S,(z\<mapsto>w) \<Turnstile> Store" apply (rule wf_store_cons[of S]) using st_dummy storeTyping_wf_store 
      by auto
    hence Hweq: "v = w" using True wf_store_def[of "S,(z\<mapsto>w)"] st_dummy.prems
      by fast
    thus ?thesis using Hweq st_dummy by auto
  next
    case False
    hence Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>s" using st_dummy by auto
    have "(x\<mapsto>v) \<in> S" using False st_dummy by (auto simp add: without_def dom_def)
    hence Hg1: "qualV v = qualT \<tau>" and Hg2: "scopeV v = scopeT \<tau>"
      and H: "\<exists>\<Gamma>v Z'. \<Gamma>v;\<Delta>;M;Z' \<turnstile> v : \<tau> \<and> un \<Gamma>v \<subseteq> un \<Gamma>s \<and> Z' \<subseteq> Z"
      using st_dummy Hin by auto
    then obtain \<Gamma>v Z'
      where Htyp: "\<Gamma>v;\<Delta>;M;Z' \<turnstile> v : \<tau>"
      and Hun: "un \<Gamma>v \<subseteq> un \<Gamma>s"
      and Hz: "Z' \<subseteq> Z"
      by auto
    have Hg3: "\<exists>\<Gamma>v Z'. \<Gamma>v;\<Delta>;M;Z' \<turnstile> v : \<tau> \<and> un \<Gamma>v \<subseteq> un (\<Gamma>s,(z:\<sigma>)) \<and> Z' \<subseteq> Z"
      apply (rule exI[of _ \<Gamma>v], rule exI[of _ Z'], rule, assumption, rule, insert Hun) proof -
      assume H: "un \<Gamma>v \<subseteq> un \<Gamma>s"
      have "un \<Gamma>s \<subseteq> un (\<Gamma>s,(z:\<sigma>))" by (auto simp add: un_def)
      thus "un \<Gamma>v \<subseteq> un (\<Gamma>s,(z:\<sigma>))" using H by auto
    qed
    thus ?thesis using Hg1 Hg2 Hg3 by auto
  qed
qed auto

lemma storeTyping_elim:
  assumes "M;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>s"
  and "(x\<mapsto>v) \<in> S"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>s"
  and "qualT \<tau> = qL"
  and "validScope (scopeV v) Zl"
  shows "\<exists>\<Gamma>v Z'. \<Gamma>v;\<Delta>;M;Z' \<turnstile> v : \<tau> \<and> M;\<Delta>;Z;Zl \<Turnstile> S\<bar>x : \<Gamma>s\<bar>x \<union> \<Gamma>v
  \<and> un \<Gamma>v \<subseteq> un \<Gamma>s \<and> \<Gamma>s\<bar>x \<sim> \<Gamma>v \<and> \<Gamma>s\<bar>x \<union> \<Gamma>v;\<Delta> \<Turnstile> OK \<and> Z' \<subseteq> Z"
using prems proof induct
  case (st_cons \<Delta> Z Zs M S \<Gamma>s \<Gamma>c \<sigma> w z Zl)
  thus ?case proof (cases "x=z")
    case True
    have "\<not> (\<Gamma>s \<union> \<Gamma>c) defines z" using prems store_dom_not_in by auto
    hence Hznoting: "\<not> \<Gamma>s defines z" by (auto simp add: dom_def)
    moreover
    have "\<Gamma>s;\<Delta> \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s \<union> \<Gamma>c" \<Delta> \<Gamma>s] by auto
    moreover
    have "\<Delta> \<Turnstile>' \<sigma>" using st_cons poly_ty_wf_type by auto
    ultimately
    have "\<Gamma>s,(z:\<sigma>);\<Delta> \<Turnstile> OK" using wf_env_cons by auto
    hence Hteq: "\<tau> = \<sigma>" using True wf_env_def[of "\<Gamma>s,(z:\<sigma>)"] st_cons.prems 
      by fast
    have "S,(z\<mapsto>w) \<Turnstile> Store" using st_cons wf_store_cons by auto
    hence Hweq: "v = w" using True wf_store_def[of "S,(z\<mapsto>w)"] st_cons.prems 
      by fast
    have "un \<Gamma>s \<subseteq> un (\<Gamma>s,(x:\<tau>))" by (auto simp add: un_def)
    hence H: "un \<Gamma>c \<subseteq> un (\<Gamma>s,(x:\<tau>))" using st_cons by auto
    hence "\<Gamma>c;\<Delta>;M;Z \<turnstile> v: \<tau> \<and> M;\<Delta>;Zs;Zl \<Turnstile> S : \<Gamma>s \<union> \<Gamma>c \<and> \<Gamma>s \<sim> \<Gamma>c \<and> \<Gamma>s \<union> \<Gamma>c;\<Delta> \<Turnstile> OK \<and> Z \<subseteq> Zs"
      using st_cons Hteq Hweq True storeTyping.st_cons by auto
    also have "\<Gamma>s = \<Gamma>s,(z:\<sigma>)\<bar>x" using Hznoting True 
      by (auto simp add: without_def dom_def)
    also have "S = S,(z\<mapsto>w)\<bar>x" using st_cons True 
      by (auto simp add: without_def dom_def)
    finally show ?thesis using H exI[of _ \<Gamma>c] Hteq Hweq True by auto
  next
    case False
    hence Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>s" using st_cons by auto
    have Hnotin: "\<not> \<Gamma>c defines x"
      using st_cons Hin compat_ok_sanity[of \<Gamma>s \<Gamma>c \<Delta> x \<tau>] by auto
    hence Hceq: "\<Gamma>c = \<Gamma>c\<bar>x" by (auto simp add: without_def dom_def)
    have "(x\<mapsto>v) \<in> S" and "(x\<rhd>\<tau>) \<in> (\<Gamma>s \<union> \<Gamma>c)"
      using False st_cons by (auto simp add: without_def dom_def)
    then obtain \<Gamma>v Z' where Htyp: "\<Gamma>v;\<Delta>;M;Z' \<turnstile> v : \<tau>"
      and Hst: "M;\<Delta>;Zs;Zl \<Turnstile> S\<bar>x : (\<Gamma>s \<union> \<Gamma>c)\<bar>x \<union> \<Gamma>v"
      and Hcompat: "(\<Gamma>s \<union> \<Gamma>c)\<bar>x \<sim> \<Gamma>v"
      and Hsubset: "un \<Gamma>v \<subseteq> un (\<Gamma>s\<union>\<Gamma>c)"
      and Hok: "(\<Gamma>s \<union> \<Gamma>c)\<bar>x \<union> \<Gamma>v;\<Delta> \<Turnstile> OK"
      and Hz: "Z' \<subseteq> Zs" using st_cons by auto
    have "(\<Gamma>s \<union> \<Gamma>c)\<bar>x = \<Gamma>s\<bar>x \<union> \<Gamma>c" using Hnotin 
      by (auto simp add: without_def dom_def)
    hence "(\<Gamma>s \<union> \<Gamma>c)\<bar>x \<union> \<Gamma>v = (\<Gamma>s\<bar>x \<union> \<Gamma>v) \<union> \<Gamma>c" by auto
    hence Hgst': "M;\<Delta>;Zs;Zl \<Turnstile> S\<bar>x : (\<Gamma>s\<bar>x \<union> \<Gamma>v) \<union> \<Gamma>c" using Hst by auto
    have Hnsxz: "\<not> S\<bar>x defines z" using st_cons
      by (auto simp add: without_def dom_def)
    have "M;\<Delta>;Zs;Zl \<Turnstile> (S\<bar>x),(z\<mapsto>w) : (\<Gamma>s\<bar>x  \<union> \<Gamma>v),(z:\<sigma>)" proof (rule StoreTyping.st_cons)
      have "\<Gamma>s\<bar>x \<sim> \<Gamma>c" using st_cons 
        by (auto simp add: env_compat_def without_def)
      moreover
      have "\<Gamma>v \<sim> \<Gamma>c\<bar>x" using Hcompat
        by (auto simp add: env_compat_def without_def)
      hence "\<Gamma>v \<sim> \<Gamma>c" using Hceq by auto
      ultimately
      show "\<Gamma>s\<bar>x \<union> \<Gamma>v \<sim> \<Gamma>c" by (auto simp add: env_compat_union_right)
    next
      show "S\<bar>x \<Turnstile> Store" using wf_store_subset[of S "S\<bar>x"] prems 
        by (auto simp add: without_def)
    next
      have "(\<Gamma>s \<union> \<Gamma>c) \<bar> x \<union> \<Gamma>v = \<Gamma>s\<bar>x \<union> \<Gamma>v \<union> \<Gamma>c" using Hceq 
        by (auto simp add: without_def)
      thus "\<Gamma>s\<bar>x \<union> \<Gamma>v \<union> \<Gamma>c;\<Delta> \<Turnstile> OK" using Hok by auto
    next
      have "\<Gamma>s;\<Delta> \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s \<union> \<Gamma>c" \<Delta> \<Gamma>s] by auto
      hence "un \<Gamma>s = un (\<Gamma>s\<bar>x)" using prems un_without_l_sanity by auto
      hence "un \<Gamma>c \<subseteq> un (\<Gamma>s\<bar>x) \<union> un \<Gamma>v" using st_cons by auto
      also have "un (\<Gamma>s\<bar>x) \<union> un \<Gamma>v = un (\<Gamma>s\<bar>x \<union> \<Gamma>v)" 
        by (auto simp add: un_def)
      finally show "un \<Gamma>c \<subseteq> un (\<Gamma>s\<bar>x \<union> \<Gamma>v)" by auto
    next
      show "Z \<subseteq> Zs" by assumption
    qed (auto simp add: Hgst' Hnsxz Hz)
    also have "(S\<bar>x),(z\<mapsto>w) =  S,(z\<mapsto>w)\<bar>x" using False 
      by (auto simp add: without_def)
    also have "(\<Gamma>s\<bar>x \<union> \<Gamma>v),(z:\<sigma>) = \<Gamma>s,(z:\<sigma>)\<bar>x \<union> \<Gamma>v" using False 
      by (auto simp add: without_def)
    finally
    have Hgst: "M;\<Delta>;Zs;Zl \<Turnstile> S,(z\<mapsto>w)\<bar>x : ((\<Gamma>s,(z:\<sigma>)\<bar>x) \<union> \<Gamma>v)" by auto
    hence "\<not> (\<Gamma>s\<bar>x \<union> \<Gamma>v \<union> \<Gamma>c) defines z" 
      using Hgst' Hnsxz store_dom_not_in[where S="S\<bar>x"] by auto
    hence Htemp: "\<not> (\<Gamma>s\<bar>x \<union> \<Gamma>v) defines z" by (auto simp add: dom_def)
    hence Hznotin: "\<not> \<Gamma>v defines z" by (auto simp add: dom_def)
    have "\<Gamma>s\<bar>x \<sim> \<Gamma>v" using Hcompat 
      by (auto simp add: without_def env_compat_def)
    hence "(\<Gamma>s\<bar>x),(z:\<sigma>) \<sim> \<Gamma>v" using Hznotin
      by (auto simp add: env_compat_def dom_def)
    also have "(\<Gamma>s\<bar>x),(z:\<sigma>) = \<Gamma>s,(z:\<sigma>)\<bar>x" using False
      by (auto simp add: without_def dom_def)
    finally have Hgcomp: "\<Gamma>s,(z:\<sigma>)\<bar>x \<sim> \<Gamma>v" using False by auto
    have H\<sigma>: "\<Delta> \<Turnstile>' \<sigma>" using st_cons poly_ty_wf_type by auto
    have "\<Gamma>s\<bar>x \<union> \<Gamma>v;\<Delta> \<Turnstile> OK" using Hok wf_env_subset[of _ _ "\<Gamma>s\<bar>x \<union> \<Gamma>v"]
      by (auto simp add: without_def)
    hence "(\<Gamma>s\<bar>x \<union> \<Gamma>v),(z:\<sigma>);\<Delta> \<Turnstile> OK" using Htemp H\<sigma>
      by (auto simp add: wf_env_def dom_def)
    also have "(\<Gamma>s\<bar>x \<union> \<Gamma>v),(z:\<sigma>) = \<Gamma>s,(z:\<sigma>)\<bar>x \<union> \<Gamma>v"   
      using False by (auto simp add: without_def dom_def)
    finally have Hgok: "\<Gamma>s,(z:\<sigma>)\<bar>x \<union> \<Gamma>v;\<Delta> \<Turnstile> OK" by auto
    have "un (\<Gamma>s \<union> \<Gamma>c) = un \<Gamma>s \<union> un \<Gamma>c" by (auto simp add: un_def)
    hence "un \<Gamma>v \<subseteq> un \<Gamma>s \<union> un \<Gamma>c" using Hsubset by auto
    hence "un \<Gamma>v \<subseteq> un \<Gamma>s" using prems by auto
    moreover have "un \<Gamma>s \<subseteq> un (\<Gamma>s,(z:\<sigma>))" by (auto simp add: un_def)
    ultimately have "un \<Gamma>v \<subseteq> un (\<Gamma>s,(z:\<sigma>))" by auto
    thus ?thesis using Htyp Hgst Hgok Hgcomp Hz by auto
  qed
next 
  case (st_dummy \<Delta> Z Zl M S \<Gamma>s \<sigma> w z)
  show ?case proof (cases "x=z")
    case True
    have "S,(z\<mapsto>w) \<Turnstile> Store" apply (rule wf_store_cons[of S]) using st_dummy storeTyping_wf_store 
      by auto
    hence Hweq: "v = w" using True wf_store_def[of "S,(z\<mapsto>w)"] st_dummy.prems
      by fast
    thus ?thesis using Hweq st_dummy by auto
  next
    case False
    hence Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>s" using st_dummy by auto
    have "(x\<mapsto>v) \<in> S" using False st_dummy by (auto simp add: without_def dom_def)
    from this st_dummy
    obtain \<Gamma>v Z'
      where Ht: "\<Gamma>v;\<Delta>;M;Z' \<turnstile> v : \<tau>" and Hst: "M;\<Delta>;Z;Zl \<Turnstile> S\<bar>x :\<Gamma>s\<bar>x \<union> \<Gamma>v" and 
      Hun: "un \<Gamma>v \<subseteq> un \<Gamma>s" and Hcompat: "\<Gamma>s\<bar>x \<sim> \<Gamma>v" and Hok: "\<Gamma>s\<bar>x \<union> \<Gamma>v;\<Delta> \<Turnstile> OK" 
      and Heq: "Z' \<subseteq> Z"
      using Hin by blast
    have Hnsxz: "\<not> S\<bar>x defines z" using st_dummy by (auto simp add: without_def dom_def)
    have "M;\<Delta>;Z;Zl \<Turnstile> (S\<bar>x),(z\<mapsto>w) : (\<Gamma>s\<bar>x \<union> \<Gamma>v),(z:\<sigma>)" apply (rule StoreTyping.st_dummy)
      using st_dummy Hst Hnsxz by auto
    also have "(S\<bar>x),(z\<mapsto>w) =  S,(z\<mapsto>w)\<bar>x" using False 
      by (auto simp add: without_def)
    also have "(\<Gamma>s\<bar>x \<union> \<Gamma>v),(z:\<sigma>) = \<Gamma>s,(z:\<sigma>)\<bar>x \<union> \<Gamma>v" using False 
      by (auto simp add: without_def)
    finally
    have Hgst: "M;\<Delta>;Z;Zl \<Turnstile> S,(z\<mapsto>w)\<bar>x : ((\<Gamma>s,(z:\<sigma>)\<bar>x) \<union> \<Gamma>v)" by auto
    hence "\<not> (\<Gamma>s\<bar>x \<union> \<Gamma>v) defines z" 
      using Hst Hnsxz store_dom_not_in[where S="S\<bar>x"] by auto
    hence Htemp: "\<not> (\<Gamma>s\<bar>x \<union> \<Gamma>v) defines z" by (auto simp add: dom_def)
    hence Hznotin: "\<not> \<Gamma>v defines z" by (auto simp add: dom_def)
    have "\<Gamma>s\<bar>x \<sim> \<Gamma>v" using Hcompat 
      by (auto simp add: without_def env_compat_def)
    hence "(\<Gamma>s\<bar>x),(z:\<sigma>) \<sim> \<Gamma>v" using Hznotin
      by (auto simp add: env_compat_def dom_def)
    also have "(\<Gamma>s\<bar>x),(z:\<sigma>) = \<Gamma>s,(z:\<sigma>)\<bar>x" using False
      by (auto simp add: without_def dom_def)
    finally have Hgcomp: "\<Gamma>s,(z:\<sigma>)\<bar>x \<sim> \<Gamma>v" using False by auto
    have "\<Gamma>s\<bar>x \<union> \<Gamma>v;\<Delta> \<Turnstile> OK" using Hok wf_env_subset[of _ _ "\<Gamma>s\<bar>x \<union> \<Gamma>v"]
      by (auto simp add: without_def)
    hence "(\<Gamma>s\<bar>x \<union> \<Gamma>v),(z:\<sigma>);\<Delta> \<Turnstile> OK" using Htemp 
      by (auto simp add: wf_env_def dom_def)
    also have "(\<Gamma>s\<bar>x \<union> \<Gamma>v),(z:\<sigma>) = \<Gamma>s,(z:\<sigma>)\<bar>x \<union> \<Gamma>v"   
      using False by (auto simp add: without_def dom_def)
    finally have Hgok: "\<Gamma>s,(z:\<sigma>)\<bar>x \<union> \<Gamma>v;\<Delta> \<Turnstile> OK" by auto
    have "un \<Gamma>s \<subseteq> un (\<Gamma>s,(z:\<sigma>))" by (auto simp add: un_def)
    hence H: "un \<Gamma>v \<subseteq> un (\<Gamma>s,(z:\<sigma>))" using Hun by auto
    show ?thesis apply (rule exI[of _ \<Gamma>v], rule exI[of _ Z'])
      using Ht Hgst Hgok Hgcomp H Heq by auto
  qed
qed auto

lemma storeTyping_strengthen:
  assumes "M;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>"
  and "\<not> l free loc in st (filterStore S Zl)"
  shows "M\<bar>l;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>"
using prems proof induct
  case (st_cons \<Delta> Z Zs M S \<Gamma>s \<Gamma>v \<tau> v x Zl)
  have "validScope (scopeV v) Zl" by assumption
  hence "x\<mapsto>v \<in> (filterStore (S,(x\<mapsto>v)) Zl)" by (auto simp add: filterStore_def)
  moreover
  have "\<not> l free loc in st (filterStore (S,(x\<mapsto>v)) Zl)" by assumption
  ultimately
  have "\<not> l free loc in v" using st_cons by (auto simp add: free_loc_store_def)
  hence H: "\<Gamma>v;\<Delta>; M\<bar>l; Z \<turnstile> v : \<tau>" using st_cons ty_mem_strengthen by auto 
  have "\<not> l free loc in st (filterStore (S,(x\<mapsto>v)) Zl)" by assumption
  moreover
  have "filterStore S Zl \<subseteq>  filterStore (S,(x\<mapsto>v)) Zl"
    by (auto simp add: filterStore_def dom_def)
  ultimately
  have "\<not> l free loc in st (filterStore S Zl)" by (auto simp add: free_loc_store_def)
  hence "M\<bar>l;\<Delta>;Zs;Zl \<Turnstile> S : (\<Gamma>s \<union> \<Gamma>v)" using st_cons by auto
  thus ?case using st_cons H StoreTyping.st_cons by auto
next
  case (st_nil M \<Delta> Z)
  moreover
  have "M\<bar>l \<subseteq> M" by (auto simp add: without_def)
  hence "M\<bar>l;\<Delta> \<Turnstile> OK" using wf_env_subset st_nil by auto
  thus ?case by auto
next
  case (st_dummy \<Delta> Z Zl M S \<Gamma>s \<tau> v x)
  moreover
  have "filterStore S Zl \<subseteq>  filterStore (S,(x\<mapsto>v)) Zl"
    by (auto simp add: filterStore_def dom_def)
  ultimately
  have "\<not> l free loc in st (filterStore S Zl)" by (auto simp add: free_loc_store_def)
  hence "M\<bar>l;\<Delta>;Z;Zl \<Turnstile> S:\<Gamma>s" using st_dummy by auto
  thus ?case using st_dummy by auto
qed

lemma storeTyping_weaken_scopes:
  assumes "M;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>s"
  and "\<not> \<rho> free sc in st S"
  shows "M;\<Delta>;Z;Zl,(\<rho>:\<section>) \<Turnstile> S : \<Gamma>s"
using prems proof induct
  case (st_dummy \<Delta> Z Zl M S \<Gamma>s \<tau> v x)
  hence "\<not> \<rho> free sc in st S" by (auto simp add: free_scope_store_def)
  hence H1: "M;\<Delta>;Z;Zl,(\<rho>:\<section>) \<Turnstile> S:\<Gamma>s" using st_dummy by auto
  have "\<rho> \<notin> free_scope_store (S,(x\<mapsto>v))" by assumption
  hence "\<rho> \<notin> set (free_scope_scope (scopeV v))" by (auto simp add: free_scope_store_def)
  moreover
  have "\<not>  validScope (scopeV v) Zl" by assumption
  ultimately
  have "\<not> validScope (scopeV v) (Zl,(\<rho>:\<section>))" by (cases "scopeV v", auto simp add: dom_def)
  thus ?case using H1 st_dummy by auto
next
  case (st_cons \<Delta> Z Zs M S \<Gamma>s \<Gamma>v \<tau> v x Zl)
  have "\<not> \<rho> free sc in st S,(x\<mapsto>v)" by assumption
  hence "\<not> \<rho> free sc in st S" by (auto simp only: free_scope_store_def)
  hence H1: "M;\<Delta>;Zs;Zl,(\<rho>:\<section>) \<Turnstile> S:\<Gamma>s\<union>\<Gamma>v" using st_cons by auto
  have "validScope (scopeV v) (Zl,(\<rho>:\<section>))" using st_cons by (cases "scopeV v", auto simp add: dom_def)
  thus ?case using H1 StoreTyping.st_cons st_cons by auto
qed auto

lemma storeTyping_Zs_weaken:
  assumes "M;\<Delta>;Zs;Zl \<Turnstile> S : \<Gamma>s"
  shows "M;\<Delta>;Zs,(\<rho>:\<section>);Zl \<Turnstile> S : \<Gamma>s"
using prems proof induct
  case st_nil thus ?case by auto
next
  case st_dummy thus ?case by auto
next
  case (st_cons Z Zs M S \<Gamma>s \<Gamma>v \<tau> v x Zl) 
  show ?case apply (rule StoreTyping.st_cons) using st_cons by auto
qed

lemma storeTyping_without_conv_qL:
  assumes "M;\<Delta>;Z;Zl,(\<rho>:\<section>) \<Turnstile> S : \<Gamma>s"
  and "x\<mapsto>v \<in> S"
  and "x\<rhd>\<tau> \<in> \<Gamma>s"
  and "validScope (scopeV v) Zl"
  and "isBangable \<tau>"
  shows "M;\<Delta>;Z,(\<rho>:\<section>);Zl,(\<rho>:\<section>) \<Turnstile> (S\<bar>x),(x\<mapsto>convV \<rho> v) : (\<Gamma>s\<bar>x),(x:convT \<rho> \<tau>)"
proof -
  have Hsc: "validScope (scopeV v) (Zl,(\<rho>:\<section>))" 
    using assms by (cases "scopeV v", auto simp add: dom_def)
  have "qualT \<tau> = qL" using assms by auto
  then obtain \<Gamma>v Z'
    where Ht: "\<Gamma>v;\<Delta>;M;Z' \<turnstile> v : \<tau>"
    and Hg2: "M;\<Delta>;Z;Zl,(\<rho>:\<section>) \<Turnstile> (S\<bar>x) : (\<Gamma>s\<bar>x) \<union> \<Gamma>v"
    and Hun: "un \<Gamma>v \<subseteq> un \<Gamma>s"
    and Hg3: "\<Gamma>s\<bar>x \<sim> \<Gamma>v"
    and Hg4: "\<Gamma>s\<bar>x \<union> \<Gamma>v;\<Delta> \<Turnstile> OK"
    and Hzeq: "Z' \<subseteq> Z"
    using storeTyping_elim[of M \<Delta> Z "Zl,(\<rho>:\<section>)" S \<Gamma>s x v \<tau>] Hsc assms by auto 
  have Hg2: "M;\<Delta>;Z,(\<rho>:\<section>);Zl,(\<rho>:\<section>) \<Turnstile> (S\<bar>x) : (\<Gamma>s\<bar>x) \<union> \<Gamma>v" 
    using storeTyping_Zs_weaken Hg2 by auto
  have Hst: "S \<Turnstile> Store" using storeTyping_wf_store assms by auto
  hence Hv: "value v" using wf_store_def assms by auto
  hence Hg1: "\<Gamma>v;\<Delta>;M;{},(\<rho>:\<section>) \<turnstile> convV \<rho> v : convT \<rho> \<tau>" using Ht ty_conv_ty assms
    by auto
  have Hg5: "value convV \<rho> v" using Hv by (cases v, auto)
  have Hg6: "\<not> (S\<bar>x) defines x" by (auto simp add: dom_def without_def)
  have Hg7: "S\<bar>x \<Turnstile> Store"
    using Hst wf_store_subset[of S "S\<bar>x"] by (auto simp add: without_def)
  have "\<Gamma>s;\<Delta> \<Turnstile> OK" using storeTyping_wf_env assms by auto
  hence "un (\<Gamma>s\<bar>x) = un (\<Gamma>s)" using prems un_without_l_sanity by auto
  hence Hg8: "un \<Gamma>v \<subseteq> un (\<Gamma>s \<bar> x)" using Hun by auto
  have "S \<Turnstile> Store" using storeTyping_wf_store assms by auto
  hence "value v" using assms by (auto simp add: wf_store_def)
  hence Hg9: "validScope (scopeV (convV \<rho> v)) (Zl,(\<rho>:\<section>))" by (cases v, auto)
  show "M;\<Delta>;Z,(\<rho>:\<section>);Zl,(\<rho>:\<section>) \<Turnstile> (S\<bar>x),(x\<mapsto>convV \<rho> v):(\<Gamma>s\<bar>x),(x:convT \<rho> \<tau>)" 
    apply (rule st_cons[of \<Gamma>v \<Delta> M "{},(\<rho>:\<section>)" "convV \<rho> v" "convT \<rho> \<tau>" "Z,(\<rho>:\<section>)" "Zl,(\<rho>:\<section>)" "S\<bar>x" "\<Gamma>s\<bar>x" x])
    using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg8 Hg9 by auto
qed

lemma storeTyping_without_conv_qU:
  assumes "M;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>s"
  and "x\<mapsto>v \<in> S"
  and "x\<rhd>convT \<rho> \<tau> \<in> \<Gamma>s"
  and "isBangable \<tau>"
  and "\<not> S defines y"
  and "value v"
  and "scopeV v = ScFree \<rho>"
  and "qualV v = qU"
  and "validScope (ScFree \<rho>) Zl"
  shows "M;\<Delta>;Z;Zl \<Turnstile> S,(y\<mapsto>convRevV v) : \<Gamma>s,(y:\<tau>)"
using prems proof induct
  case (st_cons \<Delta> Z Zs M S \<Gamma>s \<Gamma>c \<sigma> w z Zl)
  show ?case proof (cases "x=z")
    case True
    have "\<not> (\<Gamma>s\<union>\<Gamma>c) defines z" using prems store_dom_not_in by auto
    hence "\<not> \<Gamma>s defines z" by (auto simp add: dom_def)
    moreover
    have "\<Gamma>s;\<Delta> \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s\<union>\<Gamma>c" \<Delta> \<Gamma>s] by auto
    moreover
    have H\<sigma>: "\<Delta> \<Turnstile>' \<sigma>" using st_cons poly_ty_wf_type by auto
    ultimately
    have "\<Gamma>s,(z:\<sigma>);\<Delta> \<Turnstile> OK" using wf_env_cons by auto
    hence Hteq: "convT \<rho> \<tau> = \<sigma>"
      using True wf_env_def[of "\<Gamma>s,(z:\<sigma>)"] st_cons.prems 
      by fast
    have "S,(z\<mapsto>w) \<Turnstile> Store" using st_cons wf_store_cons by auto
    hence Hweq: "v = w" 
      using True wf_store_def[of "S,(z\<mapsto>w)"] st_cons.prems
      by fast
    have Htyp_c: "\<Gamma>c;\<Delta>;M;Z \<turnstile> v : convT \<rho> \<tau>"
      using st_cons Hteq Hweq by auto
    then obtain Zn where Htyp: "\<Gamma>c;\<Delta>;M;Zn \<turnstile> convRevV v : \<tau>" and Hsub: "Zn \<subseteq> Z"
      using ty_conv_ty_rev[of \<Gamma>c \<Delta> M Z v \<rho> \<tau>] prems by auto
    have HqualG: "qualG \<Gamma>c \<preceq> qU" 
      using Htyp_c ty_bang_unrestricted_env st_cons  by auto
    show "M;\<Delta>;Zs;Zl \<Turnstile> S,(z\<mapsto>w),(y\<mapsto>convRevV v):\<Gamma>s,(z:\<sigma>),(y:\<tau>)" 
      apply (rule StoreTyping.st_cons) 
    proof -
      show "\<Gamma>c;\<Delta>;M;Zn \<turnstile> convRevV v : \<tau>" using Htyp by auto
    next
      have Hst: "M;\<Delta>;Zs;Zl \<Turnstile> S,(z\<mapsto>w):\<Gamma>s,(z:\<sigma>)" using StoreTyping.st_cons st_cons by auto
      also
      have "\<Gamma>c = un \<Gamma>c" using equal_unrestricted HqualG by auto
      hence "\<Gamma>c \<subseteq> un \<Gamma>s" using st_cons by auto
      hence "\<Gamma>c \<subseteq> \<Gamma>s" by (auto simp add: un_def)
      hence "\<Gamma>s,(z:\<sigma>) = \<Gamma>s,(z:\<sigma>) \<union> \<Gamma>c" by auto
      finally
      show "M;\<Delta>;Zs;Zl \<Turnstile> S,(z\<mapsto>w):\<Gamma>s,(z:\<sigma>) \<union> \<Gamma>c" by auto
    next
      show "\<Gamma>s,(z:\<sigma>) \<sim> \<Gamma>c" using HqualG by (auto simp add: qualGrelL_def env_compat_def)
    next
      show "S,(z\<mapsto>w) \<Turnstile> Store" using wf_store_cons st_cons by auto
    next
      have "\<not> (\<Gamma>s\<union>\<Gamma>c) defines z"
        using store_dom_not_in[of _ _ _ _ S "\<Gamma>s\<union>\<Gamma>c" z] st_cons by auto
      thus "\<Gamma>s,(z:\<sigma>) \<union> \<Gamma>c;\<Delta> \<Turnstile> OK" using wf_env_cons[of "\<Gamma>s\<union>\<Gamma>c" _ z \<sigma>] st_cons H\<sigma> by auto
    next
      have "value v" by assumption
      thus "value (convRevV v)" by (cases v, auto)
    next
      show "y \<notin> Environ.dom (S,(z\<mapsto>w))" using True st_cons by (auto simp add: dom_def)
    next
      have "un \<Gamma>c \<subseteq> un \<Gamma>s" using st_cons by auto
      also 
      have "un \<Gamma>s \<subseteq> un (\<Gamma>s,(z:\<sigma>))" by (auto simp add: un_def)
      finally
      show "un \<Gamma>c \<subseteq> un (\<Gamma>s,(z:\<sigma>))" by auto
    next
      have "scopeT \<tau> = ScBot" using st_cons.prems by auto
      thus "validScope (scopeV (convRevV v)) Zl" using st_cons.prems by auto
    next
      show "Zn \<subseteq> Zs" using Hsub st_cons by auto
    qed
  next
    case False
    hence Hin: "(x\<rhd>convT \<rho> \<tau>) \<in> \<Gamma>s" using st_cons by auto
    have Hynin: "\<not> S defines y" using st_cons False by (auto simp add: dom_def) 
    have "\<not> S defines y" using st_cons by (auto simp add: dom_def)
    hence Hysc: "\<not> (\<Gamma>s\<union>\<Gamma>c) defines y" using store_dom_not_in st_cons by auto
    hence Hyc: "\<not> \<Gamma>c defines y" by (auto simp add: dom_def)
    have "(x\<mapsto>v) \<in> S" and "(x\<rhd>convT \<rho> \<tau>) \<in> (\<Gamma>s \<union> \<Gamma>c)"
      using False st_cons by (auto simp add: without_def dom_def)
    hence Hst: "M;\<Delta>;Zs;Zl \<Turnstile> S,(y\<mapsto>convRevV v):(\<Gamma>s \<union> \<Gamma>c),(y:\<tau>)" using st_cons Hin Hynin by auto
    have "M;\<Delta>;Zs;Zl \<Turnstile> S,(y\<mapsto>convRevV v),(z\<mapsto>w):\<Gamma>s,(y:\<tau>),(z:\<sigma>)"
      apply (rule StoreTyping.st_cons[of \<Gamma>c _ M Z w \<sigma> Zs Zl "S,(y\<mapsto>convRevV v)" "\<Gamma>s,(y:\<tau>)" z])
    proof -
      show "\<Gamma>c;\<Delta>;M;Z \<turnstile> w : \<sigma>" by assumption
    next
      show "M;\<Delta>;Zs;Zl \<Turnstile> S,(y\<mapsto>convRevV v):\<Gamma>s,(y:\<tau>) \<union> \<Gamma>c" using Hst by auto
    next
      have "\<Gamma>s \<sim> \<Gamma>c" by assumption
      thus "\<Gamma>s,(y:\<tau>) \<sim> \<Gamma>c" using Hyc by (auto simp add: env_compat_def dom_def)
    next
      show "S,(y\<mapsto>convRevV v) \<Turnstile> Store" using Hst storeTyping_wf_store by auto
    next
      have "(x\<rhd>convT \<rho> \<tau>) \<in> \<Gamma>s" using st_cons False by auto
      hence "\<Delta> \<Turnstile>' convT \<rho> \<tau>" 
        using st_cons poly_env_wf_ty[of \<Gamma>s \<Delta> x "convT \<rho> \<tau>"]  wf_env_subset[of "\<Gamma>s\<union>\<Gamma>c" \<Delta> \<Gamma>s] by blast
      hence "\<Delta> \<Turnstile>' \<tau>" by (cases \<tau>, auto)
      thus "\<Gamma>s,(y:\<tau>) \<union> \<Gamma>c;\<Delta> \<Turnstile> OK" using wf_env_cons[of "\<Gamma>s\<union>\<Gamma>c" _ y \<tau>] st_cons Hysc by auto
    next
      show "value w" by assumption
    next
      show "z \<notin> dom (S,(y\<mapsto>convRevV v))" using st_cons False by (auto simp add: dom_def)
    next
      have "un \<Gamma>s \<subseteq> un (\<Gamma>s,(y:\<tau>))" by (auto simp add: un_def)
      thus "un \<Gamma>c \<subseteq> un (\<Gamma>s,(y:\<tau>))" using st_cons by auto
    qed
    also have "S,(y\<mapsto>convRevV v),(z\<mapsto>w) = S,(z\<mapsto>w),(y\<mapsto>convRevV v)" by auto
    also have "\<Gamma>s,(y:\<tau>),(z:\<sigma>) = \<Gamma>s,(z:\<sigma>),(y:\<tau>)" by auto
    finally show ?thesis by auto
  qed 
next 
  case (st_dummy \<Delta> Z Zl M S \<Gamma>s \<sigma> w z)
  show ?case proof (cases "x=z")
    case True
    have "S,(z\<mapsto>w) \<Turnstile> Store" apply (rule wf_store_cons[of S]) using st_dummy storeTyping_wf_store 
      by auto
    hence Hweq: "v = w" using True wf_store_def[of "S,(z\<mapsto>w)"] st_dummy.prems
      by fast
    moreover
    have "value v" and "scopeV v = ScFree \<rho>" by assumption
    hence "scopeV v = scopeT (convT \<rho> \<tau>)" by (cases \<tau>, cases v, auto)
    ultimately
    show ?thesis using Hweq st_dummy by auto
  next
    case False
    hence Hin: "(x\<rhd>convT \<rho> \<tau>) \<in> \<Gamma>s" using st_dummy by auto
    have "(x\<mapsto>v) \<in> S" using False st_dummy by (auto simp add: without_def dom_def)
    hence "M;\<Delta>;Z;Zl \<Turnstile> S,(y\<mapsto>convRevV v):\<Gamma>s,(y:\<tau>)" using Hin st_dummy by auto
    hence "M;\<Delta>;Z;Zl \<Turnstile> S,(y\<mapsto>convRevV v),(z\<mapsto>w):\<Gamma>s,(y:\<tau>),(z:\<sigma>)"
      using st_dummy StoreTyping.st_dummy by auto
    thus ?thesis by (auto simp add: insert_commute)
  qed
qed auto

lemma storeTyping_scope_weakening:
  assumes "M;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>s"
  shows "M;\<Delta>;Z \<union> Z';Zl \<Turnstile> S : \<Gamma>s"
using prems proof induct
  case (st_cons \<Delta> Z Zs M S \<Gamma>s \<Gamma>v \<tau> v x Zl)
  show "M;\<Delta>;Zs \<union>Z';Zl \<Turnstile> S,(x\<mapsto>v):\<Gamma>s,(x:\<tau>)" apply (rule StoreTyping.st_cons) using st_cons by auto
qed auto

lemma storeTyping_strenghten_scopes:
  assumes "M;\<Delta>;Zs;Zl \<Turnstile> S : \<Gamma>s"
  shows "\<exists>\<Gamma>s'. M;\<Delta>;Zs;(Zl\<bar>\<rho>) \<Turnstile> S : \<Gamma>s' \<and> \<Gamma>s \<subseteq> \<Gamma>s'"
using prems proof induct
  case (st_cons \<Delta> Z Zs M S \<Gamma>s \<Gamma>c \<tau> v x Zl) 
  then  obtain \<Gamma>s'
    where H: "M;\<Delta>;Zs;(Zl\<bar>\<rho>) \<Turnstile> S : \<Gamma>s'" 
    and Hsub: "\<Gamma>s\<union>\<Gamma>c \<subseteq> \<Gamma>s'"
    by auto
  show ?case proof (cases "validScope (scopeV v) (Zl\<bar>\<rho>)")
    case False
    have Hscope: "scopeV v = scopeT \<tau>" using ty_scope_sanity st_cons by auto
    have H\<tau>: "\<Delta> \<Turnstile>' \<tau>" using st_cons poly_ty_wf_type by auto
    have H: "M;\<Delta>;Zs;Zl\<bar>\<rho> \<Turnstile> S,(x\<mapsto>v):\<Gamma>s',(x:\<tau>)" 
      apply (rule StoreTyping.st_dummy) using Hscope H\<tau> st_cons.hyps False by auto
    show ?thesis 
      apply (rule exI[of _ "\<Gamma>s',(x:\<tau>)"]) using H Hsub by auto
  next
    case True
    show ?thesis apply (rule exI[of _ "(\<Gamma>s'-lin(\<Gamma>c)),(x:\<tau>)"]) apply rule
    proof (rule StoreTyping.st_cons) 
      show "\<Gamma>c;\<Delta>;M;Z \<turnstile> v : \<tau>" by assumption
    next
      note H
      also have "\<Gamma>s' = \<Gamma>s'-lin(\<Gamma>c) \<union> \<Gamma>c" using Hsub by (auto simp add: lin_def)
      finally show "M;\<Delta>;Zs;Zl\<bar>\<rho> \<Turnstile> S:\<Gamma>s'-lin(\<Gamma>c) \<union> \<Gamma>c" by auto
    next
      show "un \<Gamma>c \<subseteq> un (\<Gamma>s' - lin \<Gamma>c)" using Hsub by (auto simp add: un_def lin_def)
    next
      show "\<Gamma>s' - lin \<Gamma>c \<sim> \<Gamma>c" proof (auto simp add: lin_def env_compat_def)
        fix x \<tau>
        assume "qualT \<tau> \<noteq> qL"
        thus "qualT \<tau> \<preceq> qU" by (cases "qualT \<tau>", auto)
      qed
    next
      have "\<Gamma>s';\<Delta> \<Turnstile> OK" using storeTyping_wf_env H by auto
      moreover
      have "\<Gamma>s' - lin \<Gamma>c \<union> \<Gamma>c \<subseteq> \<Gamma>s'" using Hsub by auto
      ultimately
      show "\<Gamma>s' - lin \<Gamma>c \<union> \<Gamma>c;\<Delta> \<Turnstile> OK" using wf_env_subset by auto
    next
      have "\<Gamma>s \<sim> \<Gamma>c" by assumption
      hence "\<Gamma>s = \<Gamma>s - lin \<Gamma>c" proof (auto simp add: lin_def env_compat_def)
        fix x \<tau>
        assume "\<forall>x \<tau>. x\<rhd>\<tau> \<in> \<Gamma>s \<and> x\<rhd>\<tau> \<in> \<Gamma>c \<longrightarrow> qualT \<tau> \<preceq> qU"
          and "x\<rhd>\<tau> \<in> \<Gamma>s" and "x\<rhd>\<tau> \<in> \<Gamma>c" and Hqual: "qualT \<tau> = qL"
        hence "qualT \<tau> \<preceq> qU" by fast
        thus False using Hqual by auto
      qed
      thus "\<Gamma>s,(x:\<tau>) \<subseteq> (\<Gamma>s' - lin \<Gamma>c),(x:\<tau>)" using Hsub by (auto simp add: env_compat_def)
    qed
  qed
next
  case (st_dummy \<Delta> Zs Zl M S \<Gamma>s \<tau> v x) 
  then obtain \<Gamma>s'
    where H: "M;\<Delta>;Zs;(Zl\<bar>\<rho>) \<Turnstile> S : \<Gamma>s'" 
    and Hsub: "\<Gamma>s \<subseteq> \<Gamma>s'" 
    by auto
  have "\<not> validScope (scopeV v) Zl" by assumption
  hence Hnv: "\<not> validScope (scopeV v) (Zl\<bar>\<rho>)" by (cases "scopeV v", auto simp add: dom_def without_def)
  have H: "M;\<Delta>;Zs;Zl\<bar>\<rho> \<Turnstile> S,(x\<mapsto>v):\<Gamma>s',(x:\<tau>)"
    apply (rule StoreTyping.st_dummy) using st_dummy Hnv H by auto
  show ?case apply (rule exI[of _ "\<Gamma>s',(x:\<tau>)"]) using H Hsub st_dummy by auto
qed auto

end
