theory Syntax
imports Environ

begin

datatype qual = qL | qU

datatype EmptyScope = EmptyScope

datatype scope = ScBot | ScFree name | ScVar nat

types scopeset = "(name * EmptyScope) set"
types scopearrow = "scope set"


datatype pty =
    TyUnit
  | TyBase name
  | TyFreeVar name
  | TyVar nat
  | TyArrow ty scopearrow ty    ("_ \<rightarrow>^_ _" [900, 700, 900] 900)
  | TyForall ty scopearrow      ("\<forall>._^_" [900, 700] 900)
  | TyRef ty

and ty = Qual qual scope pty

datatype tm =
    TmUnit qual scope
  | TmFreeVar name
  | TmVar nat
  | TmAbs qual scope ty tm  ("\<lambda>^_^_[_]._" [900,900,800] 800)
  | TmApp tm tm       (infixl "\<cdot>" 900)
  | TmNew tm
  | TmProd qual scope tm  ("\<Lambda>^_^_[*]._" [900] 900)
  | TmTapp tm pty      ("_\<langle>_\<rangle>" [900,0] 900)
  | TmFree tm
  | TmAssign tm tm    (infixr ":=" 900)
  | TmSwap tm tm      (infixr ":=:" 900)
  | TmDeref tm
  | TmLetbang tm tm tm ("let! (_) _ in _" [700, 900,800] 800)
  | TmLetbangSc name tm tm tm ("let$ _ (_) _ in _" [700, 700, 900,800] 800)
  | TmLoc qual scope name

syntax
  addscope :: "scopeset \<Rightarrow> name \<Rightarrow> scopeset"
    ("_','(_':'\<section>')" [90,0] 91)
translations

  "Z,(\<rho>:\<section>)" \<rightleftharpoons> "insert (\<rho>, EmptyScope) Z"

(* vsubst: freshening substitution *)
consts vsubst_tm:: "tm \<Rightarrow> nat \<Rightarrow> tm \<Rightarrow> tm"
primrec
  "vsubst_tm s i (TmUnit q \<pi>) = TmUnit q \<pi>"
  "vsubst_tm s i (TmFreeVar b) = TmFreeVar b"
  "vsubst_tm s i (TmVar n) = (if i = n then s else (TmVar n))"
  "vsubst_tm s i (\<lambda>^q^\<pi>[\<tau>].e) = \<lambda>^q^\<pi>[\<tau>].(vsubst_tm s (Suc i) e)"
  "vsubst_tm s i (e1\<cdot>e2) = (vsubst_tm s i e1)\<cdot>(vsubst_tm s i e2)"
  "vsubst_tm s i (\<Lambda>^q^\<pi>[*].e) = \<Lambda>^q^\<pi>[*]. (vsubst_tm s i e)"
  "vsubst_tm s i (e\<langle>\<tau>\<rangle>) = (vsubst_tm s i e)\<langle>\<tau>\<rangle>"
  "vsubst_tm s i (TmNew e) = TmNew (vsubst_tm s i e)"
  "vsubst_tm s i (TmFree e) = TmFree (vsubst_tm s i e)"
  "vsubst_tm s i (TmAssign e1 e2) = 
     TmAssign (vsubst_tm s i e1) (vsubst_tm s i e2)"
  "vsubst_tm s i (TmSwap e1 e2) = TmSwap (vsubst_tm s i e1) (vsubst_tm s i e2)"
  "vsubst_tm s i (TmDeref e) = TmDeref (vsubst_tm s i e)"
  "vsubst_tm s i (let! (e) e1 in e2) = 
     let! (vsubst_tm s i e) (vsubst_tm s (Suc i) e1) in (vsubst_tm s (Suc (Suc i)) e2)"
  "vsubst_tm s i (let$ \<rho> (e) e1 in e2) = 
     let$ \<rho> (vsubst_tm s i e) (vsubst_tm s i e1) in (vsubst_tm s (Suc (Suc i)) e2)"
  "vsubst_tm s i (TmLoc q \<pi> l) = TmLoc q \<pi> l"

consts freshen_tm :: "name \<Rightarrow> tm \<Rightarrow> tm"
defs freshen_tm[simp]: "freshen_tm x e \<equiv> vsubst_tm (TmFreeVar x) 0 e"

consts vsubst_sc_sc:: "scope \<Rightarrow> nat \<Rightarrow> scope \<Rightarrow> scope"
primrec
  "vsubst_sc_sc s i ScBot = ScBot"
  "vsubst_sc_sc s i (ScFree \<rho>) = ScFree \<rho>"
  "vsubst_sc_sc s i (ScVar \<eta>) = (if i = \<eta> then s else (ScVar \<eta>))"

consts vsubst_sc_set:: "scope \<Rightarrow> nat \<Rightarrow> scopearrow \<Rightarrow> scopearrow"
defs vsubst_sc_set[simp]: "vsubst_sc_set s i X \<equiv> { vsubst_sc_sc s i \<pi> | \<pi>. \<pi> \<in> X }"

consts vsubst_sc_ty:: "scope \<Rightarrow> nat \<Rightarrow> ty \<Rightarrow> ty"
consts vsubst_sc_pty:: "scope \<Rightarrow> nat \<Rightarrow> pty \<Rightarrow> pty"
primrec
  "vsubst_sc_ty s i (Qual q \<pi> \<phi>)  = Qual q (vsubst_sc_sc s i \<pi>) (vsubst_sc_pty s i \<phi>)"
  "vsubst_sc_pty s i TyUnit = TyUnit"
  "vsubst_sc_pty s i (TyBase n) = TyBase n"
  "vsubst_sc_pty s i (TyFreeVar n) = TyFreeVar n"
  "vsubst_sc_pty s i (TyVar n) = TyVar n"
  "vsubst_sc_pty s i (\<forall>.\<tau>^X) = \<forall>.(vsubst_sc_ty s i \<tau>)^X"
  "vsubst_sc_pty s i (TyRef \<tau>) = TyRef (vsubst_sc_ty s i \<tau>)"
  "vsubst_sc_pty s i (\<tau>1 \<rightarrow>^X \<tau>2) = (vsubst_sc_ty s i \<tau>1) \<rightarrow>^(vsubst_sc_set s i X) (vsubst_sc_ty s i \<tau>2)"

consts vsubst_sc:: "scope \<Rightarrow> nat \<Rightarrow> tm \<Rightarrow> tm"
primrec
  "vsubst_sc s i (TmUnit q \<pi>) = TmUnit q (vsubst_sc_sc s i \<pi>)"
  "vsubst_sc s i (TmFreeVar b) = TmFreeVar b"
  "vsubst_sc s i (TmVar n) = TmVar n"
  "vsubst_sc s i (\<lambda>^q^\<pi>[\<tau>].e) = \<lambda>^q^(vsubst_sc_sc s i \<pi>)[vsubst_sc_ty s i \<tau>]. vsubst_sc s i e"
  "vsubst_sc s i (e1\<cdot>e2) = (vsubst_sc s i e1)\<cdot>(vsubst_sc s i e2)"
  "vsubst_sc s i (\<Lambda>^q^\<pi>[*]. e) = \<Lambda>^q^\<pi>[*].(vsubst_sc s i e)"
  "vsubst_sc s i (e\<langle>\<phi>\<rangle>) = (vsubst_sc s i e)\<langle>vsubst_sc_pty s i \<phi>\<rangle>"
  "vsubst_sc s i (TmNew e) = TmNew (vsubst_sc s i e)"
  "vsubst_sc s i (TmFree e) = TmFree (vsubst_sc s i e)"
  "vsubst_sc s i (TmAssign e1 e2) = 
     TmAssign (vsubst_sc s i e1) (vsubst_sc s i e2)"
  "vsubst_sc s i (TmSwap e1 e2) = TmSwap (vsubst_sc s i e1) (vsubst_sc s i e2)"
  "vsubst_sc s i (TmDeref e) = TmDeref (vsubst_sc s i e)"
  "vsubst_sc s i (let! (e) e1 in e2) = 
     let! (vsubst_sc s i e) (vsubst_sc s (Suc i) e1) in (vsubst_sc s i e2)"
  "vsubst_sc s i (let$ \<rho> (e) e1 in e2) = 
     let$ \<rho> (vsubst_sc s i e) (vsubst_sc s i e1) in (vsubst_sc s i e2)"
  "vsubst_sc s i (TmLoc q \<pi> l) = TmLoc q (vsubst_sc_sc s i \<pi>) l"

consts freshen_sc :: "name \<Rightarrow> tm \<Rightarrow> tm"
defs freshen_sc[simp]: "freshen_sc \<rho> e \<equiv> vsubst_sc (ScFree \<rho>) 0 e"

consts free_tm :: "tm \<Rightarrow> name list"
primrec
  "free_tm (TmUnit q \<pi>)          = []"
  "free_tm (TmFreeVar b)         = [b]"
  "free_tm (TmVar n)             = []"
  "free_tm (\<lambda>^q^\<pi>[\<tau>].e)          = free_tm e"
  "free_tm (e1\<cdot>e2)               = (free_tm e1)@(free_tm e2)"
  "free_tm (\<Lambda>^q^\<pi>[*].e)          = free_tm e"
  "free_tm (e\<langle>\<tau>\<rangle>)                 = free_tm e"
  "free_tm (TmNew e)             = free_tm e"
  "free_tm (TmFree e)            = free_tm e"
  "free_tm (TmAssign e1 e2)      = (free_tm e1)@(free_tm e2)"
  "free_tm (TmSwap e1 e2)        = (free_tm e1)@(free_tm e2)"
  "free_tm (TmDeref e)           = free_tm e"
  "free_tm (let! (e) e1 in e2)   = (free_tm e)@(free_tm e1)@(free_tm e2)"
  "free_tm (let$ \<rho> (e) e1 in e2)= (free_tm e)@(free_tm e1)@(free_tm e2)"
  "free_tm (TmLoc q \<pi> l)         = []"

syntax
  is_free_in_tm :: "name \<Rightarrow> tm \<Rightarrow> bool"  ("_ free in _" [900,900] 91)
translations
  "x free in e" \<rightleftharpoons> "x \<in> set (free_tm e)"

consts free_loc :: "tm \<Rightarrow> name list"
primrec
  "free_loc (TmUnit q \<pi>)          = []"
  "free_loc (TmFreeVar b)         = []"
  "free_loc (TmVar n)             = []"
  "free_loc (\<lambda>^q^\<pi>[\<tau>].e)          = free_loc e"
  "free_loc (e1\<cdot>e2)               = (free_loc e1)@(free_loc e2)"
  "free_loc (\<Lambda>^q^\<pi>[*].e)          = free_loc e"
  "free_loc (e\<langle>\<tau>\<rangle>)                 = free_loc e"
  "free_loc (TmNew e)             = free_loc e"
  "free_loc (TmFree e)            = free_loc e"
  "free_loc (TmAssign e1 e2)      = (free_loc e1)@(free_loc e2)"
  "free_loc (TmSwap e1 e2)        = (free_loc e1)@(free_loc e2)"
  "free_loc (TmDeref e)           = free_loc e"
  "free_loc (let! (e) e1 in e2)   = (free_loc e)@(free_loc e1)@(free_loc e2)"
  "free_loc (let$ \<rho> (e) e1 in e2)= (free_loc e)@(free_loc e1)@(free_loc e2)"
  "free_loc (TmLoc q \<pi> l)         = [l]"

syntax
  is_free_in_loc :: "name \<Rightarrow> tm \<Rightarrow> bool"  ("_ free loc in _" [900,900] 91)
translations
  "x free loc in e" \<rightleftharpoons> "x \<in> set (free_loc e)"

consts free_scope_scope:: "scope \<Rightarrow> name list"
primrec
  "free_scope_scope ScBot       = []"
  "free_scope_scope (ScVar n)   = []"
  "free_scope_scope (ScFree n)  = [n]"

consts shift_many_db_list :: "nat \<Rightarrow> nat list \<Rightarrow> nat list"
primrec
  "shift_many_db_list k [] = []"
  "shift_many_db_list k (n#l) = (if n < k then shift_many_db_list k l
                                 else (n - k)#(shift_many_db_list k l))"

consts shift_db_list :: "nat list \<Rightarrow> nat list"
defs shift_db_list[simp]: "shift_db_list l \<equiv> shift_many_db_list 1 l"
consts free_db_tm :: "tm \<Rightarrow> nat list"
primrec
  "free_db_tm (TmUnit q \<pi>)        = []"
  "free_db_tm (TmFreeVar b)       = []"
  "free_db_tm (TmVar n)           = [n]"
  "free_db_tm (\<lambda>^q^\<pi>[\<tau>]. e)       = shift_db_list (free_db_tm e)"
  "free_db_tm (e1\<cdot>e2)             = (free_db_tm e1)@(free_db_tm e2)"
  "free_db_tm (\<Lambda>^q^\<pi>[*].e)        = free_db_tm e"
  "free_db_tm (e\<langle>\<tau>\<rangle>)               = free_db_tm e"
  "free_db_tm (TmNew e)           = free_db_tm e"
  "free_db_tm (TmFree e)          = free_db_tm e"
  "free_db_tm (TmAssign e1 e2)    = (free_db_tm e1)@(free_db_tm e2)"
  "free_db_tm (TmSwap e1 e2)      = (free_db_tm e1)@(free_db_tm e2)"
  "free_db_tm (TmDeref e)         = free_db_tm e"
  "free_db_tm (let! (e) e1 in e2) =
     (free_db_tm e)@(shift_db_list (free_db_tm e1))@(shift_db_list (shift_db_list (free_db_tm e2)))"
  "free_db_tm (let$ \<rho> (e) e1 in e2) =
     (free_db_tm e)@(shift_db_list (free_db_tm e1))@(shift_db_list (shift_db_list (free_db_tm e2)))"
  "free_db_tm (TmLoc q \<pi> l)       = []"

lemma shift_many_sanity:
  shows "set (shift_many_db_list k l) = { x-k | x. x \<ge> k \<and> x \<in> set l }"
by (induct l, auto)

lemma not_free_subst:
  assumes "\<not> x free in e"
  and "\<not> x free in e'"
  shows "\<not> x free in (vsubst_tm e' k e)"
using prems by (induct e arbitrary: k, auto)

lemma free_subst:
  assumes "x free in e"
  shows "x free in (vsubst_tm e' k e)"
using prems by (induct e arbitrary: k, auto)

lemma free_subst_sc:
  assumes "x free in e"
  shows "x free in (vsubst_sc \<pi> k e)"
using prems by (induct e arbitrary: k, auto)

lemma vsubst_var_same_size[simp]:
  shows "size (vsubst_tm (TmFreeVar x) k e) = size e"
by (induct e arbitrary: k, auto)

lemma subst_subst_swap:
  assumes "i \<noteq> j"
  and "\<forall>k. vsubst_tm (TmFreeVar x) k e_s = e_s"
  shows "vsubst_tm (TmFreeVar x) i (vsubst_tm e_s j e) =
         vsubst_tm e_s j (vsubst_tm (TmFreeVar x) i e)"
using prems by (induct e arbitrary: i j, auto)

lemma vsubst_tm_same_loc:
  shows "free_loc e = free_loc (vsubst_tm (TmFreeVar x) i e)"
by (induct e arbitrary: i, auto)

lemma vsubst_sc_same_loc:
  shows "free_loc e = free_loc (vsubst_sc \<pi> i e)"
by (induct e arbitrary: i, auto)

lemma not_free_subst_same_subst:
  assumes "\<not> x free in vsubst_tm (TmFreeVar x) i e"
  shows "vsubst_tm e_s i e = e"
using prems by (induct e arbitrary: i, auto)

consts hasDollars:: "tm \<Rightarrow> bool"
primrec
  "hasDollars (TmUnit q \<pi>) = False"
  "hasDollars (TmFreeVar b) = False"
  "hasDollars (TmVar n) = False"
  "hasDollars (\<lambda>^q^\<pi>[\<tau>].e) = hasDollars e"
  "hasDollars (e1\<cdot>e2) = (hasDollars e1 \<or> hasDollars e2)"
  "hasDollars (\<Lambda>^q^\<pi>[*].e) = hasDollars e"
  "hasDollars (e\<langle>\<tau>\<rangle>) = hasDollars e"
  "hasDollars (TmNew e) = hasDollars e"
  "hasDollars (TmFree e) = hasDollars e"
  "hasDollars (e1 := e2) = (hasDollars e1 \<or> hasDollars e2)"
  "hasDollars (e1 :=: e2) = (hasDollars e1 \<or> hasDollars e2)"
  "hasDollars (TmDeref e) = hasDollars e"
  "hasDollars (let! (e) e1 in e2) = (hasDollars e \<or> hasDollars e1 \<or> hasDollars e2)"
  "hasDollars (let$ \<rho> (e) e1 in e2) = True"
  "hasDollars (TmLoc q \<pi> l) = False"

consts free_let_dollar_scope:: "tm \<Rightarrow> name list"
primrec
  "free_let_dollar_scope (TmUnit q \<pi>) = []"
  "free_let_dollar_scope (TmFreeVar b) = []"
  "free_let_dollar_scope (TmVar n) = []"
  "free_let_dollar_scope (\<lambda>^q^\<pi>[\<tau>].e) = free_let_dollar_scope e"
  "free_let_dollar_scope (e1\<cdot>e2) = free_let_dollar_scope e1 @ free_let_dollar_scope e2"
  "free_let_dollar_scope (\<Lambda>^q^\<pi>[*].e) = free_let_dollar_scope e"
  "free_let_dollar_scope (e\<langle>\<tau>\<rangle>) = free_let_dollar_scope e"
  "free_let_dollar_scope (TmNew e) = free_let_dollar_scope e"
  "free_let_dollar_scope (TmFree e) = free_let_dollar_scope e"
  "free_let_dollar_scope (e1 := e2) = free_let_dollar_scope e1 @ free_let_dollar_scope e2"
  "free_let_dollar_scope (e1 :=: e2) = free_let_dollar_scope e1 @ free_let_dollar_scope e2"
  "free_let_dollar_scope (TmDeref e) = free_let_dollar_scope e"
  "free_let_dollar_scope (let! (e) e1 in e2) = 
      free_let_dollar_scope e @ free_let_dollar_scope e1 @ free_let_dollar_scope e2"
  "free_let_dollar_scope (let$ \<rho> (e) e1 in e2) = 
      [\<rho>] @ free_let_dollar_scope e @ free_let_dollar_scope e1 @ free_let_dollar_scope e2"
  "free_let_dollar_scope (TmLoc q \<pi> l) = []"

syntax
  free_in_dollars :: "scope \<Rightarrow> tm \<Rightarrow> bool"  ("_ is free dollar sc in _" [900,900] 91)
translations
  "\<rho> is free dollar sc in e" \<rightleftharpoons> "\<rho> \<in> set (free_let_dollar_scope e)"

lemma free_dollars_vsubst:
  shows "free_let_dollar_scope e = 
  free_let_dollar_scope (vsubst_tm (TmFreeVar x) i (vsubst_sc (ScFree \<rho>) j e))" 
by (induct e arbitrary: i j, auto)

lemma free_dollars_vsubst_tm:
  shows "free_let_dollar_scope e = 
  free_let_dollar_scope (vsubst_tm (TmFreeVar x) i (vsubst_tm (TmFreeVar y) j e))" 
by (induct e arbitrary: i j, auto)

lemma free_dollars_vsubst_simpl:
  shows "free_let_dollar_scope e = 
  free_let_dollar_scope (vsubst_tm (TmFreeVar x) i e)" 
by (induct e arbitrary: i, auto)

consts validScope :: "scope \<Rightarrow> scopeset \<Rightarrow> bool"
primrec
  "validScope ScBot Z = True"
  "validScope (ScFree \<rho>) Z = Z defines \<rho>"
  "validScope (ScVar \<eta>) Z = False"


(* type-polymorphism related definitions and lemmata *)

consts vsubst_pty:: "pty \<Rightarrow> nat \<Rightarrow> pty \<Rightarrow> pty"
consts vsubst_ty:: "pty \<Rightarrow> nat \<Rightarrow> ty \<Rightarrow> ty"
primrec
  "vsubst_ty s i (Qual q \<pi> \<phi>) = Qual q \<pi> (vsubst_pty s i \<phi>)"
  "vsubst_pty s i TyUnit = TyUnit"
  "vsubst_pty s i (TyBase n) = (TyBase n)"
  "vsubst_pty s i (TyFreeVar b) = TyFreeVar b"
  "vsubst_pty s i (TyVar n) = (if n=i then s else (TyVar n))"
  "vsubst_pty s i (\<tau>1\<rightarrow>^Z \<tau>2) = (vsubst_ty s i \<tau>1) \<rightarrow>^Z (vsubst_ty s i \<tau>2)"
  "vsubst_pty s i (\<forall>.\<tau>^X) = \<forall>.(vsubst_ty s (Suc i) \<tau>)^X"
  "vsubst_pty s i (TyRef \<tau>) = TyRef (vsubst_ty s i \<tau>)"

consts convert_ty:: "nat \<Rightarrow> nat \<Rightarrow> ty \<Rightarrow> ty"
consts convert_pty:: "nat \<Rightarrow> nat \<Rightarrow> pty \<Rightarrow> pty"
primrec
  "convert_ty j i (Qual q \<pi> \<phi>) = Qual q \<pi> (convert_pty j i \<phi>)"
  "convert_pty j i TyUnit = TyUnit"
  "convert_pty j i (TyBase n) = (TyBase n)"
  "convert_pty j i (TyFreeVar b) = TyFreeVar b"
  "convert_pty j i (TyVar n) = (if n=i then (TyVar j) else (TyVar n))"
  "convert_pty j i (\<tau>1\<rightarrow>^Z \<tau>2) = (convert_ty j i \<tau>1) \<rightarrow>^Z (convert_ty j i \<tau>2)"
  "convert_pty j i (\<forall>.\<tau>^X) = \<forall>.(convert_ty (Suc j) (Suc i) \<tau>)^X"
  "convert_pty j i (TyRef \<tau>) = TyRef (convert_ty j i \<tau>)"

consts vsubst_tmty:: "pty \<Rightarrow> nat \<Rightarrow> tm \<Rightarrow> tm"
primrec
  "vsubst_tmty s i (TmUnit q \<pi>) = TmUnit q \<pi>"
  "vsubst_tmty s i (TmFreeVar b) = TmFreeVar b"
  "vsubst_tmty s i (TmVar n) = (TmVar n)"
  "vsubst_tmty s i (\<lambda>^q^\<pi>[\<tau>].e) = \<lambda>^q^\<pi>[vsubst_ty s i \<tau>].(vsubst_tmty s i e)"
  "vsubst_tmty s i (e1\<cdot>e2) = (vsubst_tmty s i e1)\<cdot>(vsubst_tmty s i e2)"
  "vsubst_tmty s i (\<Lambda>^q^\<pi>[*].e) = (\<Lambda>^q^\<pi>[*].(vsubst_tmty s (Suc i) e))"
  "vsubst_tmty s i (e\<langle>\<phi>\<rangle>) = (vsubst_tmty s i e)\<langle>vsubst_pty s i \<phi>\<rangle>"
  "vsubst_tmty s i (TmNew e) = TmNew (vsubst_tmty s i e)"
  "vsubst_tmty s i (e1 := e2) = (vsubst_tmty s i e1) := (vsubst_tmty s i e2)"
  "vsubst_tmty s i (e1 :=: e2) = (vsubst_tmty s i e1) :=: (vsubst_tmty s i e2)"
  "vsubst_tmty s i (TmFree e) = TmFree (vsubst_tmty s i e)"
  "vsubst_tmty s i (TmDeref e) = TmDeref (vsubst_tmty s i e)"
  "vsubst_tmty s i (let! (e) e1 in e2) = 
      let! (vsubst_tmty s i e) (vsubst_tmty s i e1) in (vsubst_tmty s i e2)"
  "vsubst_tmty s i (let$ \<rho> (e) e1 in e2) = 
      let$ \<rho> (vsubst_tmty s i e) (vsubst_tmty s i e1) in (vsubst_tmty s i e2)"
  "vsubst_tmty s i (TmLoc q \<pi> l) = TmLoc q \<pi> l"

consts
  freshen_ty :: "name \<Rightarrow> ty \<Rightarrow> ty"
  freshen_tmty :: "name \<Rightarrow> tm \<Rightarrow> tm"

defs freshen_ty[simp]: "freshen_ty x \<tau> \<equiv> vsubst_ty (TyFreeVar x) 0 \<tau>"
defs freshen_tmty[simp]: "freshen_tmty x e \<equiv> vsubst_tmty (TyFreeVar x) 0 e"

lemma poly_free_loc_tmty_equal:
  shows "free_loc e = free_loc (vsubst_tmty \<phi> i e)"
by (induct e arbitrary: i, auto)

lemma poly_free_var_tmty_equal:
  shows "free_tm e = free_tm (vsubst_tmty \<phi> i e)"
by (induct e arbitrary: i, auto)

lemma poly_free_dollars_tmty_equal:
  shows "free_let_dollar_scope e = free_let_dollar_scope (vsubst_tmty \<phi> i e)"
by (induct e arbitrary: i, auto)

consts free_ty :: "ty \<Rightarrow> name list"
consts free_pty :: "pty \<Rightarrow> name list"
primrec
  "free_ty (Qual q \<pi> \<phi>) = free_pty \<phi>"
  "free_pty (TyUnit) = []"
  "free_pty (TyBase n) = []"
  "free_pty (TyFreeVar a) = [a]"
  "free_pty (TyVar n) = []"
  "free_pty (\<tau>1 \<rightarrow>^Z \<tau>2) = free_ty \<tau>1 @ free_ty \<tau>2"
  "free_pty (\<forall>.\<tau>^X) = free_ty \<tau>"
  "free_pty (TyRef \<tau>) = free_ty \<tau>"

consts free_tmty :: "tm \<Rightarrow> name list"
primrec
  "free_tmty (TmUnit q \<pi>)     = []"
  "free_tmty (TmFreeVar b)    = []"
  "free_tmty (TmVar n)        = []"
  "free_tmty (\<lambda>^q^\<pi>[\<tau>].e)     = free_ty \<tau> @ free_tmty e"
  "free_tmty (e1\<cdot>e2)          = free_tmty e1 @ free_tmty e2"
  "free_tmty (\<Lambda>^q^\<pi>[*].e)     = free_tmty e"
  "free_tmty (e\<langle>\<phi>\<rangle>)            = free_tmty e @ free_pty \<phi>"
  "free_tmty (TmNew e)        = free_tmty e"
  "free_tmty (e1 := e2)       = free_tmty e1 @ free_tmty e2"
  "free_tmty (e1 :=: e2)      = free_tmty e1 @ free_tmty e2"
  "free_tmty (TmDeref e)      = free_tmty e"
  "free_tmty (TmFree e)       = free_tmty e"
  "free_tmty (let! (e) e1 in e2) = free_tmty e @ free_tmty e1 @ free_tmty e2"
  "free_tmty (let$ \<rho> (e) e1 in e2) = free_tmty e @ free_tmty e1 @ free_tmty e2"
  "free_tmty (TmLoc q \<pi> l)    = []"

syntax
  is_free_in_ty :: "name \<Rightarrow> ty \<Rightarrow> bool"  ("_ free in type _" [900,900] 91)
  is_free_in_pty :: "name \<Rightarrow> pty \<Rightarrow> bool"  ("_ free in pretype _" [900,900] 91)
  is_free_in_tmty :: "name \<Rightarrow> tm \<Rightarrow> bool"  ("_ free in term _" [900,900] 91)
translations
  "x free in type \<tau>" \<rightleftharpoons> "x \<in> set (free_ty \<tau>)"
  "x free in pretype \<phi>" \<rightleftharpoons> "x \<in> set (free_pty \<phi>)"
  "x free in term e" \<rightleftharpoons> "x \<in> set (free_tmty e)"

consts free_db_ty :: "ty \<Rightarrow> nat list"
consts free_db_pty :: "pty \<Rightarrow> nat list"
primrec
  "free_db_ty (Qual q \<pi> \<phi>) = free_db_pty \<phi>"
  "free_db_pty (TyUnit) = []"
  "free_db_pty (TyBase n) = []"
  "free_db_pty (TyFreeVar a) = []"
  "free_db_pty (TyVar n) = [n]"
  "free_db_pty (\<tau>1 \<rightarrow>^Z \<tau>2) = free_db_ty \<tau>1 @ free_db_ty \<tau>2"
  "free_db_pty (\<forall>.\<tau>^X) = shift_db_list (free_db_ty \<tau>)"
  "free_db_pty (TyRef \<tau>) = free_db_ty \<tau>"

consts free_db_tmty :: "tm \<Rightarrow> nat list"
primrec
  "free_db_tmty (TmUnit q \<pi>)     = []"
  "free_db_tmty (TmFreeVar b)    = []"
  "free_db_tmty (TmVar n)        = []"
  "free_db_tmty (\<lambda>^q^\<pi>[\<tau>].e)     = free_db_ty \<tau> @ free_db_tmty e"
  "free_db_tmty (e1\<cdot>e2)          = free_db_tmty e1 @ free_db_tmty e2"
  "free_db_tmty (\<Lambda>^q^\<pi>[*].e)     = shift_db_list (free_db_tmty e)"
  "free_db_tmty (e\<langle>\<phi>\<rangle>)            = free_db_tmty e @ free_db_pty \<phi>"
  "free_db_tmty (TmNew e)        = free_db_tmty e"
  "free_db_tmty (e1 := e2)       = free_db_tmty e1 @ free_db_tmty e2"
  "free_db_tmty (e1 :=: e2)      = free_db_tmty e1 @ free_db_tmty e2"
  "free_db_tmty (TmDeref e)      = free_db_tmty e"
  "free_db_tmty (TmFree e)       = free_db_tmty e"
  "free_db_tmty (let! (e) e1 in e2) = free_db_tmty e @ free_db_tmty e1 @ free_db_tmty e2"
  "free_db_tmty (let$ \<rho> (e) e1 in e2) = free_db_tmty e @ free_db_tmty e1 @ free_db_tmty e2"
  "free_db_tmty (TmLoc q \<pi> l)    = []"

lemma not_free_subst_ty:
  assumes "\<not> a free in type \<tau>"
  and "\<not> a free in pretype \<phi>"
  shows "\<not> a free in type (vsubst_ty \<phi> k \<tau>)"
using prems by (induct \<tau> arbitrary: k taking: 
"%\<phi>'. \<not> a free in pretype \<phi>' \<longrightarrow> \<not> a free in pretype \<phi> \<longrightarrow> 
  (\<forall>k. \<not> a free in pretype (vsubst_pty \<phi> k \<phi>'))", auto)

lemma not_free_subst_pty:
  assumes "\<not> a free in pretype \<phi>"
  and "\<not> a free in pretype \<phi>s"
  shows "\<not> a free in pretype (vsubst_pty \<phi>s i \<phi>)"
using prems by (induct \<phi> arbitrary: i, auto simp add: not_free_subst_ty); 

lemma free_subst_ty:
  assumes "a free in type \<tau>"
  shows "a free in type (vsubst_ty \<phi> k \<tau>)"
using prems by (induct \<tau> arbitrary: k taking: 
  "%\<phi>'. a free in pretype \<phi>' \<longrightarrow> 
  (\<forall>k. a free in pretype (vsubst_pty \<phi> k \<phi>'))", auto)

lemma inv_not_free_subst_ty:
  assumes "\<not> a free in type (vsubst_ty \<tau>' k \<tau>)"
  shows "\<not> a free in type \<tau>"
using prems free_subst_ty by auto

lemma not_free_subst_tmty:
  assumes "\<not> a free in term e"
  and "\<not> a free in pretype \<phi>"
  shows "\<not> a free in term (vsubst_tmty \<phi> i e)"
using prems by (induct e arbitrary: i, auto simp add: not_free_subst_ty not_free_subst_pty)

lemma not_free_convert_pty:
  assumes "\<not> a free in pretype \<phi>"
  shows "\<not> a free in pretype convert_pty j i \<phi>"
using prems by (induct \<phi> arbitrary: i j taking:
  "%\<tau>. \<forall> i j.  \<not> a free in type \<tau> \<longrightarrow> \<not> a free in type convert_ty j i \<tau>", auto)

lemma not_free_convert_ty:
  assumes "\<not> a free in type \<tau>"
  shows "\<not> a free in type convert_ty j i \<tau>"
using prems by (induct \<tau> arbitrary: i j, auto simp add: not_free_convert_pty)

lemma vsubst_ty_var_same_size[simp]:
  shows "size (vsubst_ty (TyFreeVar x) i \<tau>) = size \<tau>"
by (induct \<tau> arbitrary: i taking: "%\<phi>. \<forall>i. size (vsubst_pty (TyFreeVar x) i \<phi>) = size \<phi>", auto)

lemma vsubst_tmty_var_same_size[simp]:
  shows "size (vsubst_tmty (TyFreeVar x) k e) = size e"
by (induct e arbitrary: k, auto)

lemma subst_subst_ty_swap:
  assumes "i \<noteq> j"
  and "\<forall>k. vsubst_pty (TyFreeVar a) k \<phi>_s = \<phi>_s"
  shows "vsubst_ty (TyFreeVar a) i (vsubst_ty \<phi>_s j \<tau>) =
         vsubst_ty \<phi>_s j (vsubst_ty (TyFreeVar a) i \<tau>)"
using prems by (induct \<tau> arbitrary: i j taking:
  "%\<phi>. \<forall>i j. i \<noteq> j \<longrightarrow> (\<forall>k. vsubst_pty (TyFreeVar a) k \<phi>_s = \<phi>_s) \<longrightarrow>
  vsubst_pty (TyFreeVar a) i (vsubst_pty \<phi>_s j \<phi>) = vsubst_pty \<phi>_s j (vsubst_pty (TyFreeVar a) i \<phi>)",
  auto)

lemma subst_subst_pty_swap:
  assumes "i \<noteq> j"
  and "\<forall>k. vsubst_pty (TyFreeVar a) k \<phi>_s = \<phi>_s"
  shows "vsubst_pty (TyFreeVar a) i (vsubst_pty \<phi>_s j \<phi>) = 
  vsubst_pty \<phi>_s j (vsubst_pty (TyFreeVar a) i \<phi>)"
using prems by (induct \<phi>, auto simp add: subst_subst_ty_swap)

lemma subst_subst_tmty_swap:
  assumes "\<forall>k. vsubst_tmty (TyFreeVar a) k e = e"
  shows "vsubst_tm e i (vsubst_tmty (TyFreeVar a) j eb) =
         vsubst_tmty (TyFreeVar a) j (vsubst_tm e i eb)"
using prems by (induct eb arbitrary: i j, auto)

lemma subst_subst_tytm_swap:
  shows "vsubst_tm (TmFreeVar x) i (vsubst_tmty \<tau> j eb) =
         vsubst_tmty \<tau> j (vsubst_tm (TmFreeVar x) i eb)"
by (induct eb arbitrary: i j, auto)

lemma subst_subst_tyty_swap:
  assumes "i \<noteq> j"
  and "\<forall>k. vsubst_pty (TyFreeVar a) k \<tau> = \<tau>"
  shows "vsubst_tmty \<tau> i (vsubst_tmty (TyFreeVar a) j eb) =
         vsubst_tmty (TyFreeVar a) j (vsubst_tmty \<tau> i eb)"
using prems by (induct eb arbitrary: i j, auto simp add: subst_subst_pty_swap subst_subst_ty_swap)

lemma subst_convert_subst_swap:
  assumes "i \<noteq> k"
  and "j \<noteq> k"
  and "\<forall>k. vsubst_pty (TyFreeVar a) k \<phi>s = \<phi>s"
  shows "vsubst_ty \<phi>s i (convert_ty i j (vsubst_ty (TyFreeVar a) k \<tau>)) =
         vsubst_ty (TyFreeVar a) k (vsubst_ty \<phi>s i (convert_ty i j \<tau>))"
using prems by (induct \<tau> arbitrary: i j k taking:
  "%\<phi>. \<forall>i j k. i \<noteq> k \<longrightarrow> j \<noteq> k \<longrightarrow> (\<forall>k. vsubst_pty (TyFreeVar a) k \<phi>s = \<phi>s) \<longrightarrow>
       vsubst_pty \<phi>s i (convert_pty i j (vsubst_pty (TyFreeVar a) k \<phi>)) =
       vsubst_pty (TyFreeVar a) k (vsubst_pty \<phi>s i (convert_pty i j \<phi>))", auto)

lemma vsubst_ty_free_db_ty:
  assumes "free_db_pty \<phi>s = []"
  shows "set (free_db_ty (vsubst_ty \<phi>s k \<tau>)) = set (free_db_ty \<tau>) - {k}"
using prems by (induct \<tau> arbitrary: k taking: 
 "%\<phi>. \<forall>k. free_db_pty \<phi>s = [] \<longrightarrow> set (free_db_pty (vsubst_pty \<phi>s k \<phi>)) = set (free_db_pty \<phi>) - {k}",
  auto simp add: shift_many_sanity)

(* not used *)
(*
lemma vsubst_tm_free_db_tm:
  assumes "free_db_tm e_s = []"
  shows "set (free_db_tm (vsubst_tm e_s k e)) = set (free_db_tm e) - {k}"
using prems by (induct e arbitrary: k, auto simp add: shift_many_sanity)
*)

lemma vsubst_tmty_free_db:
  shows "free_db_tm e = free_db_tm (vsubst_tmty \<tau> i e)"
by (induct e arbitrary: i, auto)

lemma subst_subst_tmsc_swap:
  shows "vsubst_tm (TmFreeVar x) i (vsubst_sc (ScFree \<rho>) j e) =
         vsubst_sc (ScFree \<rho>) j (vsubst_tm (TmFreeVar x) i e)"
by (induct e arbitrary: i j, auto)

lemma subst_subst_ptysc_swap:
  assumes "\<forall>i. vsubst_sc_pty (ScFree \<rho>) i \<phi>s = \<phi>s"
  shows "vsubst_pty \<phi>s i (vsubst_sc_pty (ScFree \<rho>) j \<phi>) =
         vsubst_sc_pty (ScFree \<rho>) j (vsubst_pty \<phi>s i \<phi>)"
using prems by (induct \<phi>  arbitrary: i j taking:
  "%\<tau>. \<forall>i j. (\<forall>i. vsubst_sc_pty (ScFree \<rho>) i \<phi>s = \<phi>s) \<longrightarrow>
  vsubst_ty \<phi>s i (vsubst_sc_ty (ScFree \<rho>) j \<tau>) =
         vsubst_sc_ty (ScFree \<rho>) j (vsubst_ty \<phi>s i \<tau>)"
 , auto)

lemma subst_subst_tysc_swap:
  assumes "\<forall>i. vsubst_sc_pty (ScFree \<rho>) i \<phi>s = \<phi>s"
  shows "vsubst_ty \<phi>s i (vsubst_sc_ty (ScFree \<rho>) j \<tau>) =
         vsubst_sc_ty (ScFree \<rho>) j (vsubst_ty \<phi>s i \<tau>)"
using prems subst_subst_ptysc_swap by (cases \<tau>, auto)

lemma subst_subst_tmtysc_swap:
  assumes "\<forall>i. vsubst_sc_pty (ScFree \<rho>) i \<phi>s = \<phi>s"
  shows "vsubst_tmty \<phi>s i (vsubst_sc (ScFree \<rho>) j e) =
         vsubst_sc (ScFree \<rho>) j (vsubst_tmty \<phi>s i e)"
using prems subst_subst_tysc_swap subst_subst_ptysc_swap by (induct e arbitrary: i j, auto)

lemma not_free_subst_tm:
  assumes "\<not> a free in term e"
  shows "\<not> a free in term (vsubst_tm (TmFreeVar x) i e)" 
using prems by (induct e arbitrary: i, auto)

lemma not_free_subst_ptysc:
  assumes "\<not> a free in pretype \<phi>"
  shows "\<not> a free in pretype (vsubst_sc_pty (ScFree \<rho>) i \<phi>)" 
using prems by (induct \<phi> arbitrary: i taking: 
  "%\<tau>. \<forall>i. \<not> a free in type \<tau> \<longrightarrow> \<not> a free in type (vsubst_sc_ty (ScFree \<rho>) i \<tau>)", auto)

lemma not_free_subst_tysc:
  assumes "\<not> a free in type \<tau>"
  shows "\<not> a free in type (vsubst_sc_ty (ScFree \<rho>) i \<tau>)"
using prems not_free_subst_ptysc by (cases \<tau>, auto)

lemma not_free_subst_sc:
  assumes "\<not> a free in term e"
  shows "\<not> a free in term (vsubst_sc (ScFree \<rho>) i e)" 
using prems not_free_subst_tysc not_free_subst_ptysc by (induct e arbitrary: i, auto)

lemma vsubst_sc_var_same_size[simp]:
  shows "size (vsubst_sc (ScFree \<rho>) k e) = size e"
by (induct e arbitrary: k, auto)

lemma not_free_subst_sc_term:
  assumes "\<not> a free in  e"
  shows "\<not> a free in (vsubst_sc (ScFree \<rho>) i e)" 
using prems by (induct e arbitrary: i, auto)


(* not used *)
(*
lemma vsubst_tmty_free_db_tmty:
  assumes "free_db_pty \<tau>_s = []"
  shows "set (free_db_tmty (vsubst_tmty \<tau>_s k e)) = set (free_db_tmty e) - {k}"
using prems by (induct e arbitrary: k, auto simp add: shift_many_sanity vsubst_ty_free_db_ty)
*)

(*
lemma vsubst_tmty_free_db_tm:
  assumes "free_db_tmty e_s = []"
  shows "set (free_db_tmty (vsubst_tm e_s k e)) = set (free_db_tmty e)"
using prems by (induct e arbitrary: k, auto simp add: shift_many_sanity)

lemma not_in_free_db_same_subst:
  assumes "k \<notin> set (free_db_tm e)"
  shows "vsubst_tm e' k e = e"
using prems proof (induct e arbitrary: k)
  case (TmAbs \<tau> eb k')
  thus ?case using shift_many_sanity[of "1" "free_db_tm eb"] by auto
qed auto

lemma not_in_free_db_same_subst_tmty:
  assumes "k \<notin> set (free_db_tmty e)"
  shows "vsubst_tmty \<tau>' k e = e"
using prems proof (induct e arbitrary: k)
  case (TmProd eb k')
  thus ?case using shift_many_sanity[of "1" "free_db_tmty eb"] 
    by (auto simp add: not_in_free_db_same_subst_ty)
qed (auto simp add: not_in_free_db_same_subst_ty)

lemma no_free_db_vars_same_subst:
  assumes "free_db_tm e = []"
  shows "vsubst_tm e' k e = e"
using prems by (auto simp add: not_in_free_db_same_subst)

lemma no_free_db_vars_same_subst_tmty:
  assumes "free_db_tmty e = []"
  shows "vsubst_tmty \<tau>' k e = e"
using prems by (auto simp add: not_in_free_db_same_subst_tmty)
*)

lemma poly_not_in_free_db_same_subst_pty:
  assumes "k \<notin> set (free_db_pty \<phi>)"
  shows "vsubst_pty \<phi>s k \<phi> = \<phi>"
using prems proof (induct \<phi> arbitrary: k taking:
  "%\<tau>. \<forall>k. k \<notin> set (free_db_ty \<tau>) \<longrightarrow> vsubst_ty \<phi>s k \<tau> = \<tau>")
  case (TyForall \<tau> k') thus ?case using shift_many_sanity[of "1" "free_db_ty \<tau>"] by auto
qed auto

lemma poly_not_in_free_db_same_subst_ty:
  assumes "k \<notin> set (free_db_ty \<tau>)"
  shows "vsubst_ty \<phi>s k \<tau> = \<tau>"
using prems poly_not_in_free_db_same_subst_pty by (induct \<tau>, auto)

lemma poly_no_free_db_vars_same_subst_pty:
  assumes "free_db_pty \<phi>  = []"
  shows "vsubst_pty \<phi>s k \<phi> = \<phi>"
using prems by (auto simp add: poly_not_in_free_db_same_subst_pty)

lemma poly_no_free_db_vars_same_subst_ty:
  assumes "free_db_ty \<tau> = []"
  shows "vsubst_ty \<phi>s k \<tau> = \<tau>"
using prems by (auto simp add: poly_not_in_free_db_same_subst_ty)

inductive Val :: "tm \<Rightarrow> bool" ("value _")
  where
    v_unit[intro]  : "value (TmUnit q \<pi>)"
  | v_abs[intro]   : "value (\<lambda>^q^\<pi>[\<tau>].e)"
  | v_poly[intro!] : "\<lbrakk> value v \<rbrakk> \<Longrightarrow> value (\<Lambda>^q^\<pi>[*].v)"
  | v_loc[intro]   : "value (TmLoc q \<pi> l)"

inductive_cases value_elim[elim]: "value e"

end
