theory Typing
imports Syntax

begin

(**************************************************************************)
consts qualRel :: "qual \<Rightarrow> qual \<Rightarrow> bool" (infix "\<preceq>" 50)
primrec
  "qU \<preceq> q = True"
  "qL \<preceq> q = (q = qL)"

consts qualT :: "ty \<Rightarrow> qual"
primrec
  "qualT (Qual q \<pi> \<phi>) = q"

consts scopeT :: "ty \<Rightarrow> scope"
primrec
  "scopeT (Qual q \<pi> \<phi>) = \<pi>"

datatype Term = Term "ty"
datatype Type = Type
types env = "(name * Term) set"
types ty_env = "(name * Type) set"

syntax
  var_term :: "name \<Rightarrow> Term \<Rightarrow> (name * Term)"  ("(_\<rhd>_)" [90,90] 90)
  var_type :: "name \<Rightarrow> (name * Type)"  ("(_\<rhd>'\<star>)" [90] 90)
translations
  "(x\<rhd>\<tau>)" \<rightleftharpoons> "(x, Term \<tau>)"
  "(x\<rhd>\<star>)" \<rightleftharpoons> "(x, Type)"

syntax
  addbinding :: "(name * Term) set \<Rightarrow> name \<Rightarrow> ty \<Rightarrow> (name * Term) set"
    ("_','(_':_')" [90,0,0] 91)
  addbinding_ty :: "(name * Type) set \<Rightarrow> name \<Rightarrow> (name * Type) set"
    ("_','(_':'\<star>')" [90,0] 91)
translations
  "\<Gamma>,(x:\<tau>)" \<rightleftharpoons> "insert (x\<rhd>\<tau>) \<Gamma>"
  "\<Delta>,(x:\<star>)" \<rightleftharpoons> "insert (x\<rhd>\<star>) \<Delta>"

constdefs un :: "env \<Rightarrow> env"
 "un \<Gamma> \<equiv> { (x\<rhd>\<tau>) | x \<tau>. (x\<rhd>\<tau>) \<in> \<Gamma> \<and> qualT \<tau> \<preceq> qU }"

constdefs lin :: "env \<Rightarrow> env"
 "lin \<Gamma> \<equiv> { (x\<rhd>\<tau>) | x \<tau>. (x\<rhd>\<tau>) \<in> \<Gamma> \<and> qL \<preceq> qualT \<tau> }"

constdefs free_ty_env :: "env \<Rightarrow> name set"
  "free_ty_env \<Gamma> \<equiv> \<Union> { set (free_ty \<tau>) | x \<tau>. (x \<rhd> \<tau>) \<in> \<Gamma> }"

constdefs free_db_ty_env :: "env \<Rightarrow> name set"
  "free_db_ty_env \<Gamma> \<equiv> \<Union> { set (free_db_ty \<tau>) | x \<tau>. (x \<rhd> \<tau>) \<in> \<Gamma> }"

constdefs vsubst_ty_env:: "pty \<Rightarrow> nat \<Rightarrow> env \<Rightarrow> env"
  "vsubst_ty_env s i \<Gamma> \<equiv> { (x\<rhd>vsubst_ty s i \<tau>) | x \<tau>. (x \<rhd> \<tau>) \<in> \<Gamma> }"

constdefs convert_ty_env:: "nat \<Rightarrow> nat \<Rightarrow> env \<Rightarrow> env"
  "convert_ty_env j i \<Gamma> \<equiv> { (x\<rhd>convert_ty j i \<tau>) | x \<tau>. (x \<rhd> \<tau>) \<in> \<Gamma> }"

constdefs vsubst_ty_func :: "pty \<Rightarrow> nat \<Rightarrow> (name * Term) \<Rightarrow> (name * Term)"
  "vsubst_ty_func s i m \<equiv> case m of (Pair x b) \<Rightarrow>
                            case b of (Term \<tau>) \<Rightarrow>
                              (x\<rhd>vsubst_ty s i \<tau>)"

syntax
  is_free_in_env :: "name \<Rightarrow> env \<Rightarrow> bool"  ("_ free in environ _" [900,900] 91)
translations
  "a free in environ \<Gamma>" \<rightleftharpoons> "a \<in> free_ty_env \<Gamma>"

lemma free_env_finite:
  assumes "finite \<Gamma>"
  shows "finite (free_ty_env \<Gamma>)"
proof -
  let ?f = "%m. case m of (Pair x b) \<Rightarrow>
                  case b of (Term \<tau>) \<Rightarrow>
                    set (free_ty \<tau>)"
  let ?A = "{F. \<exists>x \<tau>. F = set (free_ty \<tau>) \<and> (x\<rhd>\<tau>) \<in> \<Gamma>}"
  have "finite (\<Union>{F. \<exists>x \<tau>. F = set (free_ty \<tau>) \<and> (x\<rhd>\<tau>) \<in> \<Gamma>})"
    using prems finite_surj[of \<Gamma> ?A ?f] by force
  thus ?thesis using prems finite_Union by (auto simp add: free_ty_env_def)
qed

lemma free_ty_env_subset:
  assumes "\<not> x free in environ \<Gamma>"
  and "\<Gamma>' \<subseteq> \<Gamma>"
  shows "\<not> x free in environ \<Gamma>'"
using prems by (auto simp add: free_ty_env_def)

lemma only_term:
  assumes "\<Gamma> defines x"
  shows "\<exists>\<tau>. (x\<rhd>\<tau>) \<in> \<Gamma>"
proof -
  obtain b where "(x, b) \<in> \<Gamma>" using prems defines_sanity by blast
  thus ?thesis by (cases b, auto)
qed

lemma vsubst_ty_env_dom_sanity:
  shows "dom \<Gamma> = dom (vsubst_ty_env s i \<Gamma>)"
proof (auto simp add: dom_def vsubst_ty_env_def)
  fix x b
  assume "(x, b) \<in> \<Gamma>"
  hence "\<Gamma> defines x" by (auto simp add: dom_def)
  thus "\<exists>\<tau>. (x\<rhd>\<tau>) \<in> \<Gamma>" by (rule only_term)
qed

lemma vsubst_ty_env_dom_sanity:
  shows "dom \<Gamma> = dom (vsubst_ty_env s i \<Gamma>)"
proof (auto simp add: dom_def vsubst_ty_env_def)
  fix x b
  assume "(x, b) \<in> \<Gamma>"
  hence "\<Gamma> defines x" by (auto simp add: dom_def)
  thus "\<exists>\<tau>. (x\<rhd>\<tau>) \<in> \<Gamma>" by (rule only_term)
qed

constdefs
  wf_env_ty :: "ty_env \<Rightarrow> bool" ("_ \<Turnstile>' TY")
  "wf_env_ty \<Delta> \<equiv> finite \<Delta>"

consts isBangableP :: "pty \<Rightarrow> bool"
primrec
  "isBangableP TyUnit = False"
  "isBangableP (TyFreeVar n) = False"
  "isBangableP (TyVar n) = False"
  "isBangableP (TyBase n) = False"
  "isBangableP (\<tau>1 \<rightarrow>^\<delta> \<tau>2) = False"
  "isBangableP (\<forall>.\<tau>^X) = False"
  "isBangableP (TyRef \<tau>) = True"

inductive isBangable :: "ty \<Rightarrow> bool"
  where
  banger[intro!]: "\<lbrakk> isBangableP \<phi> \<rbrakk> \<Longrightarrow> isBangable (Qual qL ScBot \<phi>)"

inductive_cases banger_elim[elim]:   "isBangable \<tau>"


consts convT :: "name \<Rightarrow> ty \<Rightarrow> ty"
primrec
  "convT \<rho> (Qual q \<pi> \<phi>) = Qual qU (ScFree \<rho>) \<phi>"

constdefs qualGrelL :: "env \<Rightarrow> qual \<Rightarrow> bool" ("qualG _ \<preceq> _" [90,90] 50)
  "qualGrelL \<Gamma> q \<equiv> (\<forall>x \<tau>. (x\<rhd>\<tau>) \<in> \<Gamma> \<longrightarrow> qualT \<tau> \<preceq> q)"

constdefs qualGrelR :: "qual \<Rightarrow> env \<Rightarrow> bool" ("_ \<preceq> qualG _" [90,90] 50)
  "qualGrelR q \<Gamma> \<equiv> (\<forall>x \<tau>. (x\<rhd>\<tau>) \<in> \<Gamma> \<longrightarrow> q \<preceq> qualT \<tau>)"

lemma qualT_unrestricted:
  assumes "qualT \<tau> \<preceq> qU"
  shows "qualT \<tau> = qU"
using prems by (cases "qualT \<tau>", auto)

lemma qualG_relL_subset:
  assumes "qualG \<Gamma> \<preceq> q"
  and "\<Gamma>' \<subseteq> \<Gamma>"
  shows "qualG \<Gamma>' \<preceq> q"
using prems by (auto simp add: qualGrelL_def)

lemma equal_unrestricted:
  assumes "qualG \<Gamma> \<preceq> qU"
  shows "un \<Gamma> = \<Gamma>"
using prems proof auto
  fix x b
  assume "(x, b) \<in> \<Gamma>"
  then obtain \<tau> where "b = Term \<tau>" by (cases b)
  hence "(x\<rhd>\<tau>) \<in> \<Gamma>" and "qualT \<tau> \<preceq> qU" using prems by (auto simp add: qualGrelL_def)
  thus "(x, b) \<in> un \<Gamma>" using prems by (auto simp add: un_def)
qed (auto simp add: un_def)

constdefs wf_scopeset :: "scopeset \<Rightarrow> bool" ("_ \<Turnstile> ScopeSet" [50] 50)
  "wf_scopeset Z \<equiv> finite Z"

constdefs minimalValidScope :: "scope \<Rightarrow> scopeset \<Rightarrow> bool"
   "minimalValidScope \<pi> Z \<equiv>
    (\<pi> = ScBot \<and> Z = {}) \<or> (\<exists>\<rho>. \<pi> = ScFree \<rho> \<and> Z = {},(\<rho>:\<section>))"

constdefs minimalValidScopeT :: "ty \<Rightarrow> scopeset \<Rightarrow> bool"
   "minimalValidScopeT \<tau> Z \<equiv>
    (scopeT \<tau> = ScBot \<and> Z = {}) \<or> (\<exists>\<rho>. scopeT \<tau> = ScFree \<rho> \<and> Z = {},(\<rho>:\<section>))"

consts filterScFreeX :: "scopearrow \<Rightarrow> scopeset"
defs filterScFreeX[simp]: "filterScFreeX X \<equiv> { (\<rho>, EmptyScope) | \<rho>. (ScFree \<rho>) \<in> X }"

consts filterScVarX :: "scopearrow \<Rightarrow> scopeset"
defs filterScVarX[simp]: "filterScVarX X \<equiv> { (n, EmptyScope) | n. (ScVar n) \<in> X }"

inductive 
  wf_ptype :: "ty_env \<Rightarrow> pty \<Rightarrow> bool" ("_ \<Turnstile>' _" [50,900])
 and
  wf_type  :: "ty_env \<Rightarrow> ty \<Rightarrow> bool" ("_ \<Turnstile>' _" [50,900])
  where
  wf_unit[intro!]   : "\<lbrakk> \<Delta> \<Turnstile>' TY \<rbrakk> \<Longrightarrow> \<Delta> \<Turnstile>' TyUnit"
  | wf_base[intro!]   : "\<lbrakk> \<Delta> \<Turnstile>' TY \<rbrakk> \<Longrightarrow> \<Delta> \<Turnstile>' TyBase n"
  | wf_tvar[intro!]   : "\<lbrakk> (x\<rhd>\<star>) \<in> \<Delta>; \<Delta> \<Turnstile>' TY \<rbrakk> \<Longrightarrow> \<Delta> \<Turnstile>' TyFreeVar x"
  | wf_arrow[intro!]  : "\<lbrakk> \<Delta> \<Turnstile>' \<tau>1; \<Delta> \<Turnstile>' \<tau>2 \<rbrakk> \<Longrightarrow> \<Delta> \<Turnstile>' \<tau>1 \<rightarrow>^Z \<tau>2"
  | wf_forall[intro!] : "\<lbrakk> \<Delta> \<Turnstile>' TY; finite La; \<forall>a. a \<notin> La \<longrightarrow> \<Delta>,(a:\<star>) \<Turnstile>' freshen_ty a \<tau> \<rbrakk> \<Longrightarrow> \<Delta> \<Turnstile>' \<forall>.\<tau>^X"
  | wf_ref[intro!]    : "\<lbrakk> \<Delta> \<Turnstile>' \<tau> \<rbrakk> \<Longrightarrow> \<Delta> \<Turnstile>' TyRef \<tau>"
  | wf_ty[intro!]     : "\<lbrakk> \<Delta> \<Turnstile>' \<phi> \<rbrakk> \<Longrightarrow> \<Delta> \<Turnstile>' Qual q \<pi> \<phi>"

inductive_cases wf_ty_unit [elim!]:   "\<Delta> \<Turnstile>' TyUnit"
inductive_cases wf_ty_base [elim!]:   "\<Delta> \<Turnstile>' TyBase n"
inductive_cases wf_ty_tvar [elim!]:   "\<Delta> \<Turnstile>' TyFreeVar x"
inductive_cases wf_ty_arrow [elim!]:  "\<Delta> \<Turnstile>' \<tau>1 \<rightarrow>^Z \<tau>2"
inductive_cases wf_ty_forall [elim!]: "\<Delta> \<Turnstile>' \<forall>.\<tau>^X"
inductive_cases wf_ty_ref [elim!]:    "\<Delta> \<Turnstile>' TyRef \<tau>"
inductive_cases wf_ty [elim!]:        "\<Delta> \<Turnstile>' Qual q \<pi> \<phi>"

constdefs wf_env :: "env \<Rightarrow> ty_env \<Rightarrow> bool" ("_;_ \<Turnstile> OK" [50] 50)
  "wf_env \<Gamma> \<Delta> \<equiv> finite \<Gamma> \<and> \<Delta> \<Turnstile>' TY \<and> (\<forall>x \<tau>1 \<tau>2. (x\<rhd>\<tau>1) \<in> \<Gamma> \<and> (x\<rhd>\<tau>2) \<in> \<Gamma> \<longrightarrow> \<tau>1 = \<tau>2 \<and> 
  \<Delta> \<Turnstile>' \<tau>1)"

lemma wf_env_subset:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "\<Gamma>' \<subseteq> \<Gamma>"
  shows "\<Gamma>';\<Delta> \<Turnstile> OK"
using prems finite_subset[of \<Gamma>' \<Gamma>] by (unfold wf_env_def, fast)

lemma env_cons_without_eq_left:
  assumes "\<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK"
  and "(y\<rhd>\<tau>) \<in> \<Gamma>2"
  shows "\<Gamma>1,(y:\<tau>) \<union> \<Gamma>2\<bar>y = \<Gamma>1 \<union> \<Gamma>2"
using prems proof (auto simp add: wf_env_def without_def)
  fix x b
  assume "(x\<rhd>\<tau>) \<in> \<Gamma>2" and "(x, b) \<in> \<Gamma>2" and "b \<noteq> Term \<tau>"
  have "\<Gamma>2;\<Delta> \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>1 \<union> \<Gamma>2" \<Delta> \<Gamma>2] by auto
  hence "b = Term \<tau>" using prems by (cases b, auto simp add: wf_env_def)
  thus False by contradiction
qed

lemma env_cons_without_eq_right:
  assumes "\<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK"
  and "(y\<rhd>\<tau>) \<in> \<Gamma>1"
  shows "\<Gamma>1\<bar>y \<union> \<Gamma>2,(y:\<tau>) = \<Gamma>1 \<union> \<Gamma>2"
proof -
  have "\<Gamma>2 \<union> \<Gamma>1;\<Delta> \<Turnstile> OK" using prems by (auto simp add: Un_commute)
  hence "\<Gamma>2,(y:\<tau>) \<union> \<Gamma>1\<bar>y = \<Gamma>2 \<union> \<Gamma>1" using prems env_cons_without_eq_left by auto
  thus ?thesis by (auto simp add: Un_commute)
qed

lemma wf_env_cons:
  fixes \<tau> :: ty
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "\<not> \<Gamma> defines x"
  and "\<Delta> \<Turnstile>' \<tau>"
  shows "\<Gamma>,(x:\<tau>);\<Delta> \<Turnstile> OK"
using prems by (auto simp add: dom_def wf_env_def)

lemma wf_env_union_sanity:
  assumes "\<Gamma>1\<union>\<Gamma>2;\<Delta> \<Turnstile> OK"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>2"
  and "\<Gamma>1 defines x"
  shows "(x\<rhd>\<tau>) \<in> \<Gamma>1"
proof -
  obtain b where "(x, b) \<in> \<Gamma>1" using prems by (auto simp add: dom_def)
  then obtain \<tau>' where "(x\<rhd>\<tau>') \<in> \<Gamma>1" by (cases b, auto)
  also hence "\<tau>' = \<tau>" using prems by (auto simp add: wf_env_def)
  finally show ?thesis .
qed

lemma wf_env_union_sanity_cons:
  assumes "\<Gamma>1\<union>\<Gamma>2;\<Delta> \<Turnstile> OK"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>1"
  shows "\<Gamma>2,(x:\<tau>);\<Delta> \<Turnstile> OK"
using prems proof (cases "\<Gamma>2 defines x")
  case False
  have "\<Gamma>2;\<Delta> \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
  moreover
  have "\<Delta> \<Turnstile>' \<tau>" using prems by (auto simp add: wf_env_def)
  ultimately
  show ?thesis using False wf_env_cons by auto 
next
  case True
  then obtain b where "(x, b) \<in> \<Gamma>2" using prems by (auto simp add: dom_def)
  then obtain \<tau>' where "(x\<rhd>\<tau>') \<in> \<Gamma>2" by (cases b, auto)
  also hence "\<tau>' = \<tau>" using prems by (auto simp add: wf_env_def)
  finally have "\<Gamma>2,(x:\<tau>) = \<Gamma>2" by auto
  thus ?thesis using prems wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
qed

lemma wf_env_subset_defines:
  assumes Hok: "\<Gamma>s;\<Delta> \<Turnstile> OK"
  and Hsub: "\<Gamma>r \<subseteq> \<Gamma>s"
  and Hx: "(x\<rhd>\<tau>) \<in> \<Gamma>s"
  and Hd: "\<Gamma>r defines x"
  shows "(x\<rhd>\<tau>) \<in> \<Gamma>r"
proof -
  obtain b where "(x, b) \<in> \<Gamma>r" using Hd defines_sanity by auto
  then obtain \<tau>' where "(x\<rhd>\<tau>') \<in> \<Gamma>r" by (cases b, auto)
  hence "(x\<rhd>\<tau>') \<in> \<Gamma>s" using Hsub by auto
  also hence "\<tau>' = \<tau>" using Hx Hok by (auto simp add: wf_env_def)
  thus ?thesis using prems by auto
qed

lemma un_without_l_sanity:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>"
  and "qualT \<tau> = qL"
  shows "un \<Gamma> = un (\<Gamma>\<bar>x)"
proof (auto simp add: un_def without_def)
  fix \<tau>'
  assume "(x\<rhd>\<tau>') \<in> \<Gamma>" and "qualT \<tau>' \<preceq> qU"
  hence "\<tau> = \<tau>'" using prems by (auto simp add: wf_env_def)
  hence "qualT \<tau> = qualT \<tau>'" by auto
  hence "qL \<preceq> qU" using prems by auto
  thus False by auto
qed

constdefs env_compat :: "env \<Rightarrow> env \<Rightarrow> bool"  (infix "\<sim>" 50)
   "\<Gamma>1 \<sim> \<Gamma>2 \<equiv> \<forall>x \<tau>. (x\<rhd>\<tau>) \<in> (\<Gamma>1 \<inter> \<Gamma>2) \<longrightarrow> qualT \<tau> \<preceq> qU"

lemma env_compat_weakening_left:
  assumes "\<Gamma>1 \<sim> \<Gamma>2"
  and "qualG \<Gamma> \<preceq> qU"
  shows "\<Gamma>1 \<union> \<Gamma> \<sim> \<Gamma>2"
using prems by (auto simp add: env_compat_def qualGrelL_def)

lemma env_compat_symm:
  assumes "\<Gamma>1 \<sim> \<Gamma>2"
  shows "\<Gamma>2 \<sim> \<Gamma>1"
using prems by (auto simp add: env_compat_def)

lemma env_compat_union_left:
  assumes "\<Gamma> \<sim> \<Gamma>1"
  and "\<Gamma> \<sim> \<Gamma>2"
  shows "\<Gamma> \<sim> \<Gamma>1 \<union> \<Gamma>2"
using prems by (auto simp add: env_compat_def)

lemma env_compat_union_right:
  assumes "\<Gamma>1 \<sim> \<Gamma>"
  and "\<Gamma>2 \<sim> \<Gamma>"
  shows "\<Gamma>1 \<union> \<Gamma>2 \<sim> \<Gamma>"
using prems by (auto simp add: env_compat_def)

lemma env_compat_subset:
  assumes "\<Gamma>1 \<sim> \<Gamma>2"
  and "\<Gamma>1' \<subseteq> \<Gamma>1"
  shows "\<Gamma>1' \<sim> \<Gamma>2"
using prems by (auto simp add: env_compat_def)

lemma env_compat_subset_right:
  assumes "\<Gamma>1 \<sim> \<Gamma>2"
  and "\<Gamma>2' \<subseteq> \<Gamma>2"
  shows "\<Gamma>1 \<sim> \<Gamma>2'"
using prems by (auto simp add: env_compat_def)

lemma compat_ok_sanity:
  assumes "\<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK"
  and "\<Gamma>1 \<sim> \<Gamma>2"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>1"
  and "qualT \<tau> = qL"
  shows "\<not> \<Gamma>2 defines x"
proof
  assume "\<Gamma>2 defines x"
  then obtain b where "(x, b) \<in> \<Gamma>2" by (auto simp add: dom_def)
  then obtain \<tau>' where "(x\<rhd>\<tau>') \<in> \<Gamma>2" by (cases b, auto)
  also hence "\<tau>' = \<tau>" using prems by (auto simp add: wf_env_def)
  finally have "(x\<rhd>\<tau>) \<in> \<Gamma>1 \<inter> \<Gamma>2" by auto
  hence "qualT \<tau> \<preceq> qU" using prems env_compat_def[of \<Gamma>1 \<Gamma>2] by fast
  thus False using prems by auto
qed

lemma compat_ok_sanity_right:
  assumes "\<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK"
  and "\<Gamma>1 \<sim> \<Gamma>2"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>2"
  and "qualT \<tau> = qL"
  shows "\<not> \<Gamma>1 defines x"
using prems by (auto simp add: env_compat_symm[of \<Gamma>1 \<Gamma>2] Un_commute[of \<Gamma>1 \<Gamma>2]
                compat_ok_sanity[of \<Gamma>2 \<Gamma>1])

inductive
  ty_pretype :: "scopeset \<Rightarrow> pty \<Rightarrow> bool" ("_ \<Turnstile> _" [50,900])
  and
  ty_type :: "scopeset \<Rightarrow> ty \<Rightarrow> bool" ("_ \<Turnstile> _" [50,900])
  where
    ty_unit[intro!] :   "{} \<Turnstile> TyUnit"
  | ty_base[intro!] :   "{} \<Turnstile> TyBase n"
  | ty_tvar[intro!] :   "{} \<Turnstile> TyFreeVar x"
  | ty_tbvar[intro!] :  "{} \<Turnstile> TyVar x"
  | ty_arrow[intro!] :  " \<lbrakk> Z1 \<Turnstile> \<tau>1 ; Z2 \<Turnstile> \<tau>2; Z \<Turnstile> ScopeSet;
                          filterScVarX X = {}; Z = filterScFreeX X \<rbrakk> 
                          \<Longrightarrow> Z1\<union>Z2\<union>Z \<Turnstile>  \<tau>1 \<rightarrow>^X \<tau>2"
  | ty_forall[intro!] : "\<lbrakk> Z \<Turnstile> \<tau>; filterScVarX X = {}; 
                           Zx = filterScFreeX X; Zx \<Turnstile> ScopeSet \<rbrakk> \<Longrightarrow> Z\<union>Zx \<Turnstile> \<forall>.\<tau>^X"
  | ty_ref[intro!] :    "\<lbrakk> Z \<Turnstile> \<tau> \<rbrakk> \<Longrightarrow> Z \<Turnstile> TyRef \<tau>"
  | ty_qual[intro!] :   "\<lbrakk> Z\<phi> \<Turnstile> \<phi>; minimalValidScope \<pi> Z \<rbrakk> \<Longrightarrow> Z\<union>Z\<phi> \<Turnstile>  Qual q \<pi> \<phi>"

inductive_cases ty_type_unit_elim [elim]:   "Z \<Turnstile> TyUnit"
inductive_cases ty_type_base_elim [elim]:   "Z \<Turnstile> TyBase n"
inductive_cases ty_type_var_elim [elim]:    "Z \<Turnstile> TyFreeVar x"
inductive_cases ty_type_bvar_elim [elim]:   "Z \<Turnstile> TyVar x"
inductive_cases ty_type_arrow_elim [elim]:  "Z \<Turnstile> \<tau>1 \<rightarrow>^Z' \<tau>2"
inductive_cases ty_type_forall_elim [elim]: "Z \<Turnstile> \<forall>.\<tau>^X"
inductive_cases ty_type_ref_elim [elim]:    "Z \<Turnstile> TyRef \<tau>"
inductive_cases ty_type_qual [elim]:        "Z \<Turnstile> Qual q \<pi> \<phi>"

inductive
  Typing :: "env \<Rightarrow> ty_env \<Rightarrow> env \<Rightarrow> scopeset \<Rightarrow> tm \<Rightarrow> ty \<Rightarrow> bool"
    ("_;_;_;_ \<turnstile> _ : _" [50,50,50,50,80,80] 50)
  where
    t_unit[intro!]  : "\<And>\<Delta> Z M \<Gamma> q \<pi>.
                       \<lbrakk> \<Gamma>;\<Delta> \<Turnstile> OK; qualG \<Gamma> \<preceq> qU; M;\<Delta> \<Turnstile> OK;
                         minimalValidScope \<pi> Z \<rbrakk> \<Longrightarrow>
                         \<Gamma>;\<Delta>;M;Z \<turnstile> TmUnit q \<pi>: Qual q \<pi> TyUnit"
  | t_var[intro!]   : "\<And>\<Delta> Z M \<Gamma> \<tau> x.
                       \<lbrakk> \<Gamma>;\<Delta> \<Turnstile> OK; (x\<rhd>\<tau>) \<in> \<Gamma>; qualG (\<Gamma>\<bar>x) \<preceq> qU; M;\<Delta> \<Turnstile> OK;
                         minimalValidScope (scopeT \<tau>) Z \<rbrakk> \<Longrightarrow>
                         \<Gamma>;\<Delta>;M;Z \<turnstile> TmFreeVar x : \<tau>"
  | t_abs[intro!]   : "\<And>\<Delta> X Z M \<Gamma> \<tau>1 \<tau>2 e q \<pi> L.
                       \<lbrakk> \<forall>x. \<Gamma>,(x:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma> \<sim> {x\<rhd>\<tau>1} \<and> x \<notin> L \<longrightarrow>
                           \<Gamma>,(x:\<tau>1);\<Delta>;M;(filterScFreeX X) \<turnstile> freshen_tm x e : \<tau>2;
                         qualG \<Gamma> \<preceq> q; \<Gamma>;\<Delta> \<Turnstile> OK; M;\<Delta> \<Turnstile> OK; finite L;
                         filterScVarX X = {}; \<Delta> \<Turnstile>' \<tau>1; minimalValidScope \<pi> Z \<rbrakk> \<Longrightarrow>
                         \<Gamma>;\<Delta>;M;Z \<turnstile> \<lambda>^q^\<pi>[\<tau>1]. e : Qual q \<pi> (\<tau>1 \<rightarrow>^X \<tau>2)"
  | t_app[intro!]   : "\<And>\<Delta> X Z1 Z2 Z M \<Gamma>1 \<Gamma>2 \<tau>1 \<tau>2 e1 e2 q.
                       \<lbrakk> \<Gamma>1 \<sim> \<Gamma>2; \<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK;
                         \<Gamma>1;\<Delta>;M;Z1 \<turnstile> e1 : Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2);
                         \<Gamma>2;\<Delta>;M;Z2 \<turnstile> e2 : \<tau>1;
                         filterScFreeX X \<Turnstile> ScopeSet;                         
                         minimalValidScope (scopeT \<tau>2) Z;
                         \<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<not> (Z2 \<union> filterScFreeX X \<union> Z) defines \<rho>;
                         \<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<not> (Z1 \<union> filterScFreeX X \<union> Z) defines \<rho> 
                       \<rbrakk>\<Longrightarrow>
                         \<Gamma>1\<union>\<Gamma>2;\<Delta>;M;Z1\<union>Z2\<union>(filterScFreeX X)\<union>Z \<turnstile> e1\<cdot>e2 : \<tau>2"
  | t_poly[intro!]  : "\<And>\<Delta> Z X M \<Gamma> \<tau> v q \<pi> La.
                       \<lbrakk> \<Gamma>;\<Delta> \<Turnstile> OK; M;\<Delta> \<Turnstile> OK; value v; qualG \<Gamma> \<preceq> q; finite La;
                        \<forall>a. a \<notin> La \<longrightarrow> \<Gamma>;\<Delta>,(a:\<star>);M;(filterScFreeX X) \<turnstile> freshen_tmty a v : freshen_ty a \<tau>;
                        minimalValidScope \<pi> Z \<rbrakk> \<Longrightarrow>
                        \<Gamma>;\<Delta>;M;Z \<turnstile> \<Lambda>^q^\<pi>[*].v : Qual q \<pi> (\<forall>.\<tau>^X)"
  | t_ptyapp[intro!]: "\<And>\<Delta> Z Z\<tau> M \<Gamma> \<tau> \<phi> e q X.
                       \<lbrakk> \<Gamma>;\<Delta>;M;Z \<turnstile> e : Qual q ScBot (\<forall>.\<tau>^X); \<Delta> \<Turnstile>' \<phi>; (*Z\<phi> \<Turnstile> \<phi>;*)
                         \<forall>i \<rho>. vsubst_sc_pty (ScFree \<rho>) i \<phi> = \<phi>;
                         minimalValidScope (scopeT \<tau>) Z\<tau>; filterScFreeX X \<Turnstile> ScopeSet;
                         \<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> \<not> (filterScFreeX X\<union>Z\<tau>) defines \<rho> \<rbrakk> \<Longrightarrow> 
                        \<Gamma>;\<Delta>;M;Z\<union>(filterScFreeX X)\<union>Z\<tau> \<turnstile> e\<langle>\<phi>\<rangle> : vsubst_ty \<phi> 0 \<tau>"
  | t_new[intro!]   : "\<And>\<Delta> Z M \<Gamma> \<tau> e.
                       \<lbrakk> \<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau> \<rbrakk> \<Longrightarrow>
                         \<Gamma>;\<Delta>;M;Z \<turnstile> TmNew e : Qual qL ScBot (TyRef \<tau>)"
  | t_free[intro!]  : "\<And>\<Delta> Z Z\<tau> M \<Gamma> \<tau> e.
                       \<lbrakk> \<Gamma>;\<Delta>;M;Z \<turnstile> e : Qual qL ScBot (TyRef \<tau>);
                         minimalValidScope (scopeT \<tau>) Z\<tau>;
                         \<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> \<not> Z\<tau> defines \<rho> \<rbrakk>  \<Longrightarrow>
                         \<Gamma>;\<Delta>;M;Z\<union>Z\<tau> \<turnstile> TmFree e : \<tau>"
  | t_assign[intro!]: "\<And>\<Delta> Z1 Z2 M \<Gamma>1 \<Gamma>2 \<tau> e1 e2 \<pi>.
                       \<lbrakk> \<Gamma>1 \<sim> \<Gamma>2; \<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK; qualT \<tau> \<preceq> qU;
                         \<Gamma>1;\<Delta>;M;Z1 \<turnstile> e1 : Qual qU \<pi> (TyRef \<tau>);
                         \<Gamma>2;\<Delta>;M;Z2 \<turnstile> e2 : \<tau>;
                         \<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<not> Z2 defines \<rho>;
                         \<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<not> Z1 defines \<rho> \<rbrakk> \<Longrightarrow>
                         \<Gamma>1\<union>\<Gamma>2;\<Delta>;M;Z1\<union>Z2 \<turnstile> e1 := e2 : Qual qU ScBot TyUnit"
  | t_swap[intro!]:   "\<And>\<Delta> Z1 Z2 M \<Gamma>1 \<Gamma>2 \<tau> e1 e2 \<pi>.
                       \<lbrakk> \<Gamma>1 \<sim> \<Gamma>2; \<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK; qL \<preceq> qualT \<tau>;
                         \<Gamma>1;\<Delta>;M;Z1 \<turnstile> e1 : Qual qU \<pi> (TyRef \<tau>);
                         \<Gamma>2;\<Delta>;M;Z2 \<turnstile> e2 : \<tau>;
                         \<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<not> Z2 defines \<rho>;
                         \<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<not> Z1 defines \<rho> \<rbrakk> \<Longrightarrow>
                         \<Gamma>1\<union>\<Gamma>2;\<Delta>;M;Z2\<union>Z1 \<turnstile> e1 :=: e2 : \<tau>"
  | t_deref[intro!] : "\<And>\<Delta> Z Z\<tau> M \<Gamma> \<tau> e.
                       \<lbrakk> \<Gamma>;\<Delta>;M;Z \<turnstile> e : Qual qU \<pi> (TyRef \<tau>); qualT \<tau> \<preceq> qU;
                         minimalValidScope (scopeT \<tau>) Z\<tau>;
                         \<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> \<not> Z\<tau> defines \<rho> \<rbrakk> \<Longrightarrow>
                         \<Gamma>;\<Delta>;M;Z\<union>Z\<tau> \<turnstile> TmDeref e : \<tau>"
  | t_let[intro!]   : "\<And>\<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z e e1 e2 \<tau> \<tau>1 \<tau>2 Lx Ly.
                       \<lbrakk> \<Gamma>1 \<sim> \<Gamma>2; \<Gamma> \<sim> \<Gamma>2; \<Gamma> \<sim> \<Gamma>1; \<Gamma> \<union> \<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK;
                         Z1 \<Turnstile> ScopeSet; Z2 \<Turnstile> ScopeSet; validScope (scopeT \<tau>1) Z1;
                         isBangable \<tau>; \<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>; finite Lx; finite Ly;
                         (\<forall>x \<rho>. \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK \<and> \<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>} \<and> x \<notin> Lx
                             \<and> \<not> \<rho> is free dollar sc in e1 \<and> \<not> (Z2\<union>Z1) defines \<rho> \<longrightarrow> (
                             \<Gamma>1,(x:convT \<rho> \<tau>); \<Delta>; M; Z1 \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1
                             \<or>
                             \<Gamma>1,(x:convT \<rho> \<tau>); \<Delta>; M; Z1,(\<rho>:\<section>) \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1
                         ));
                         (\<forall>x y. \<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1) \<and>
                             {x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1} \<and> x \<notin> Lx \<and> y \<notin> Ly \<and> x \<noteq> y \<longrightarrow>
                             \<Gamma>2,(x:\<tau>),(y:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2) : \<tau>2);
                         \<forall>\<rho>. \<rho> is free dollar sc in e  \<longrightarrow> \<not> (Z1\<union>Z2) defines \<rho>;
                         \<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<not> (Z2\<union>Z) defines \<rho>;
                         \<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<not> (Z1\<union>Z) defines \<rho>
                        \<rbrakk> \<Longrightarrow>
                        \<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2;\<Delta>;M;Z2\<union>Z1\<union>Z \<turnstile> let! (e) e1 in e2 : \<tau>2"
  | t_let'[intro!]  : "\<And>\<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z \<rho> x e1 e2 \<tau> \<tau>1 \<tau>2 Zf Lx Ly.
                       \<lbrakk> \<Gamma>1 \<sim> \<Gamma>2; \<Gamma> \<sim> \<Gamma>2; \<Gamma> \<sim> \<Gamma>1; \<Gamma> \<union> \<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK;
                         \<not> (Z2\<union>Z1) defines \<rho>;
                         Z1 \<Turnstile> ScopeSet; Z2 \<Turnstile> ScopeSet; validScope (scopeT \<tau>1) Z1;
                         isBangable \<tau>; \<Gamma>;\<Delta>;M;Z \<turnstile> TmFreeVar x : convT \<rho> \<tau>;
                         \<not> (\<Gamma>1\<union>\<Gamma>2) defines x;
                         finite Lx; finite Ly;
                         (Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>));
                         \<Gamma>1,(x:convT \<rho> \<tau>); \<Delta>; M; Zf \<turnstile> e1 : \<tau>1;
                         (\<forall>x y. \<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1) \<and>
                             {x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1} \<and> x \<notin> Lx \<and> y \<notin> Ly \<and> x \<noteq> y \<longrightarrow>
                           \<Gamma>2,(x:\<tau>),(y:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2) : \<tau>2);
                         \<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<not> (Z\<union>Z2) defines \<rho>;
                         \<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<not> (Z\<union>Z1) defines \<rho>
                       \<rbrakk> \<Longrightarrow>
                         \<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2;\<Delta>;M;Z2\<union>Z1\<union>Z \<turnstile> let$ \<rho> (TmFreeVar x) e1 in e2 : \<tau>2"
  | t_loc[intro!]   : "\<And>M \<Gamma> \<tau> l \<pi> q.
                       \<lbrakk> \<Gamma>;\<Delta> \<Turnstile> OK; qualG \<Gamma> \<preceq> qU; M;\<Delta> \<Turnstile> OK; (l\<rhd>\<tau>) \<in> M;
                         minimalValidScope \<pi> Z \<rbrakk> \<Longrightarrow>
                         \<Gamma>;\<Delta>;M;Z \<turnstile> TmLoc q \<pi> l : Qual q \<pi> (TyRef \<tau>)"

inductive_cases typing_unit_elim [elim]:   "\<Gamma>;\<Delta>;M;Z \<turnstile> TmUnit q \<pi> : \<tau>"
inductive_cases typing_var_elim [elim]:    "\<Gamma>;\<Delta>;M;Z \<turnstile> TmFreeVar x : \<tau>"
inductive_cases typing_abs_elim [elim]:    "\<Gamma>;\<Delta>;M;Z \<turnstile> \<lambda>^q^\<pi>[\<tau>1]. e : \<tau>2"
inductive_cases typing_app_elim [elim]:    "\<Gamma>;\<Delta>;M;Z \<turnstile> e1\<cdot>e2 : \<tau>"
inductive_cases typing_poly_elim [elim]:   "\<Gamma>;\<Delta>;M;Z \<turnstile> \<Lambda>^q^\<pi>[*]. e : \<tau>"
inductive_cases typing_tyapp_elim [elim]:  "\<Gamma>;\<Delta>;M;Z \<turnstile> e\<langle>\<phi>\<rangle> : \<tau>"
inductive_cases typing_new_elim [elim]:    "\<Gamma>;\<Delta>;M;Z \<turnstile> TmNew e : \<tau>"
inductive_cases typing_free_elim [elim]:   "\<Gamma>;\<Delta>;M;Z \<turnstile> TmFree e : \<tau>"
inductive_cases typing_assign_elim [elim]: "\<Gamma>;\<Delta>;M;Z \<turnstile> e1 := e2 : \<tau>"
inductive_cases typing_swap_elim [elim]:   "\<Gamma>;\<Delta>;M;Z \<turnstile> e1 :=: e2 : \<tau>"
inductive_cases typing_deref_elim [elim]:  "\<Gamma>;\<Delta>;M;Z \<turnstile> TmDeref e : \<tau>"
inductive_cases typing_let_elim [elim]:    "\<Gamma>;\<Delta>;M;Z \<turnstile> let! (e) e1 in e2 : \<tau>"
inductive_cases typing_loc_elim [elim]:    "\<Gamma>;\<Delta>;M;Z \<turnstile> TmLoc q \<pi> l : \<tau>"
inductive_cases typing_letsc_elim [elim]:  "\<Gamma>;\<Delta>;M;Z \<turnstile> let$ \<rho> (e) e1 in e2 : \<tau>" 

lemma ty_wf_env:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>"
  shows "\<Gamma>;\<Delta> \<Turnstile> OK"
using prems by (induct, auto)

lemma ty_wf_env_loc:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>"
  shows "M;\<Delta> \<Turnstile> OK"
using prems by (induct, auto)

lemma poly_env_wf_env_ty:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  shows "\<Delta> \<Turnstile>' TY"
using prems wf_env_def by auto

lemma poly_wf_env_ty:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>"
  shows "\<Delta> \<Turnstile>' TY"
proof -
  have "\<Gamma>;\<Delta> \<Turnstile> OK" using ty_wf_env prems by auto
  thus ?thesis using poly_env_wf_env_ty by auto
qed

lemma poly_env_wf_ty:
  fixes \<tau> :: ty
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>"
  shows "\<Delta> \<Turnstile>' \<tau>"
using prems wf_env_def[of \<Gamma> \<Delta>] by auto

lemma ty_finite_dom:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>"
  shows "finite (dom \<Gamma>)"
proof -
  have "\<Gamma>;\<Delta> \<Turnstile> OK" using prems ty_wf_env by auto
  hence "finite \<Gamma>" by (auto simp add: wf_env_def)
  thus ?thesis using dom_finite by auto
qed

lemma poly_wf_env_ty_fin:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  shows "finite (dom \<Delta>)"
using wf_env_ty_def[of \<Delta>] wf_env_def[of \<Gamma> \<Delta>] dom_finite prems by auto

lemma exists_x:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "finite L"
  shows "\<exists>x. x \<notin> set (free_tm e) \<and> x \<notin> L \<and> \<not> \<Gamma> defines x"
proof -
  have "finite (dom \<Gamma>)" using prems wf_env_def dom_finite by auto
  thus ?thesis using fresh_exists[of "(set (free_tm e))\<union>L\<union>(dom \<Gamma>)"] prems
    by blast
qed

lemma wf_type_scopeset:
  fixes \<tau> :: ty 
  assumes "Z \<Turnstile> \<tau>"
  shows "Z \<Turnstile> ScopeSet"
using prems by (induct taking: "%Z. %\<phi>::pty. Z \<Turnstile> \<phi> \<longrightarrow> Z \<Turnstile> ScopeSet" , 
  auto simp add: wf_scopeset_def minimalValidScope_def)

lemma wf_pretype_scopeset:
  fixes \<phi> :: pty 
  assumes "Z \<Turnstile> \<phi>"
  shows "Z \<Turnstile> ScopeSet"
using prems wf_type_scopeset by (induct, auto simp add: wf_scopeset_def)

lemma minimal_is_scope:
  assumes "minimalValidScope \<pi> Z"
  shows "Z \<Turnstile> ScopeSet"
using prems by (cases \<pi>, auto simp add: minimalValidScope_def wf_scopeset_def)

lemma new_variable_abs:
  fixes \<tau> :: ty
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK" 
  and "finite Lx"
  and "\<Delta> \<Turnstile>' \<tau>"
  shows "\<exists>x. (\<Gamma>,(x:\<tau>);\<Delta> \<Turnstile> OK \<and> \<Gamma> \<sim> {x\<rhd>\<tau>} \<and> x \<notin> Lx)"
proof -
  have "finite (dom \<Gamma>)" using prems dom_finite by (auto simp add: wf_env_def)
  then obtain x where "x \<notin> dom \<Gamma> \<and> x \<notin> Lx" using fresh_exists[of "dom \<Gamma> \<union> Lx"] using assms by auto
  moreover
  hence H1: "\<Gamma>,(x:\<tau>);\<Delta> \<Turnstile> OK" and H2: "\<Gamma> \<sim> {x\<rhd>\<tau>}" using prems
    by (auto simp add: dom_def wf_env_def env_compat_def)
  ultimately 
  show ?thesis by auto
qed

lemma new_variable_let:
  fixes \<tau> \<tau>' :: ty 
  fixes Lx Ly :: "nat set"
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK" 
  and "\<Delta> \<Turnstile>' \<tau>" and "\<Delta> \<Turnstile>' \<tau>'"
  and "finite Lx"
  and "finite Ly"
  shows "\<exists>x y. (\<Gamma>,(x:\<tau>),(y:\<tau>');\<Delta> \<Turnstile> OK \<and> \<Gamma> \<sim> {x\<rhd>\<tau>},(y:\<tau>') \<and> {x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>'} \<and> x \<notin> Lx \<and> y \<notin> Ly \<and> x \<noteq> y)"
proof -
  have Hfin: "finite (dom \<Gamma>)" using prems dom_finite by (auto simp add: wf_env_def)
  then obtain x where Hx: "x \<notin> dom \<Gamma>" and Hlx: "x \<notin> Lx" using fresh_exists[of "dom \<Gamma> \<union> Lx"]
    using assms by auto
 obtain y where Hy: "y \<notin> dom \<Gamma>" and Hyx: "y \<noteq> x" and Hly: "y \<notin> Ly"
    using Hfin assms fresh_exists[of "(dom \<Gamma>) \<union> {x} \<union> Ly"] by auto 
  have "\<Gamma> \<sim> {x\<rhd>\<tau>}" and "\<Gamma> \<sim> {y\<rhd>\<tau>'}"  using Hx Hy by (auto simp add: env_compat_def dom_def)
  hence Hg2: "\<Gamma> \<sim> {x\<rhd>\<tau>},(y:\<tau>')" using env_compat_union_left[of \<Gamma> "{x\<rhd>\<tau>}" "{y\<rhd>\<tau>'}"] by auto
  have Hg3: "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>'}" using Hyx by (auto simp add: env_compat_def)
  have "\<Gamma>,(x:\<tau>);\<Delta> \<Turnstile> OK" using assms Hx by (auto simp add: wf_env_def dom_def)
  hence "\<Gamma>,(x:\<tau>),(y:\<tau>');\<Delta> \<Turnstile> OK" using Hy Hyx assms by (auto simp add: wf_env_cons)
  thus ?thesis using Hg2 Hg3 Hlx Hly Hyx by auto
qed

lemma new_scope_let:
  fixes \<tau> :: ty
  fixes Lx :: "nat set"
  assumes "(Z\<union>Z') \<Turnstile> ScopeSet" 
  and "finite Lx"
  and "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "\<Delta> \<Turnstile>' \<tau>"
  shows "\<exists>\<rho> Zf x. \<Gamma>,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK \<and> \<Gamma> \<sim> {x\<rhd>convT \<rho> \<tau>} \<and> (Zf = Z \<or> Zf = Z,(\<rho>:\<section>)) 
                   \<and> \<not> \<rho> is free dollar sc in e \<and> \<not> (Z'\<union>Z) defines \<rho> \<and> x \<notin> Lx" 
proof -
  have Hfin: "finite (dom \<Gamma>)" using prems dom_finite by (auto simp add: wf_env_def)
  then obtain x where Hx: "x \<notin> dom \<Gamma>" and Hlx: "x \<notin> Lx" 
    using fresh_exists[of "dom \<Gamma> \<union> Lx"] using assms by auto
  have "finite (Z'\<union>Z)" using assms by (auto simp add: wf_scopeset_def)
  hence "finite ((dom (Z'\<union>Z)) \<union> set (free_let_dollar_scope e))" using dom_finite by auto
  then obtain \<rho> Zf where Hg1: "Zf = Z \<or> Zf = Z,(\<rho>:\<section>)" 
    and Hg2: "\<not> (Z'\<union>Z) defines \<rho>" and Hg3: "\<not> \<rho> is free dollar sc in e"
    using fresh_exists[of "((dom (Z'\<union>Z)) \<union> set (free_let_dollar_scope e))"] by auto
  have Hg4: "\<Gamma> \<sim> {x\<rhd>convT \<rho> \<tau>}" using Hx by (auto simp add: env_compat_def dom_def)
  have "\<Delta> \<Turnstile>' convT \<rho> \<tau>" using assms by (cases \<tau>, auto)
  hence Hg5: "\<Gamma>,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK" using assms Hx by (auto simp add: wf_env_cons)
  show ?thesis using Hg1 Hg2 Hg3 Hg4 Hg5 Hlx by auto
qed

lemma poly_tenv_cons:
  fixes \<tau> :: ty
  assumes "\<Delta> \<Turnstile>' \<tau>"
  shows "\<Delta>,(z:\<star>) \<Turnstile>' \<tau>"
proof -
  have H: "\<And>\<Delta>'. \<Delta>' \<Turnstile>' TY \<Longrightarrow> \<Delta>',(z:\<star>) \<Turnstile>' TY" using wf_env_ty_def by auto
  show ?thesis using prems H by (induct taking: 
      "%\<Delta>. %\<phi>::pty. \<Delta> \<Turnstile>' \<phi> \<longrightarrow> \<Delta>,(z:\<star>) \<Turnstile>' \<phi>", auto simp add: env_permute)
qed

lemma poly_env_ty_cons:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  shows "\<Gamma>;\<Delta>,(a:\<star>) \<Turnstile> OK"
proof -
  have "\<Delta>,(a:\<star>) \<Turnstile>' TY" using poly_env_wf_env_ty wf_env_ty_def prems by auto
  moreover
  have "(\<forall>x \<tau>1 \<tau>2. x\<rhd>\<tau>1 \<in> \<Gamma> \<and> x\<rhd>\<tau>2 \<in> \<Gamma> \<longrightarrow> \<tau>1 = \<tau>2 \<and> \<Delta> \<Turnstile>' \<tau>1)" 
    using wf_env_def prems by auto
  hence "(\<forall>x \<tau>1 \<tau>2. x\<rhd>\<tau>1 \<in> \<Gamma> \<and> x\<rhd>\<tau>2 \<in> \<Gamma> \<longrightarrow> \<tau>1 = \<tau>2 \<and> \<Delta>,(a:\<star>) \<Turnstile>' \<tau>1)"
    using poly_tenv_cons by auto
  ultimately
  show ?thesis using prems wf_env_def by auto
qed

lemma poly_type_wf_env_ty:
  fixes \<tau> :: ty
  assumes "\<Delta> \<Turnstile>' \<tau>"
  shows "\<Delta> \<Turnstile>' TY"
using prems by (induct, auto)

lemma poly_wf_ty_strength:
  fixes \<tau> :: ty
  assumes "\<Delta>,(z:\<star>) \<Turnstile>' \<tau>"
  and "\<not> z free in type \<tau>"
  shows "\<Delta> \<Turnstile>' \<tau>"
proof -
  { fix \<Delta>o
    assume H: "\<Delta>o = \<Delta>,(z:\<star>)"
    hence "\<Delta>o \<Turnstile>' \<tau>" using assms by auto
    hence ?thesis using H assms proof (induct arbitrary: \<Delta>
        taking: "%\<Delta>o. %\<phi> :: pty. \<forall>\<Delta>. \<Delta>o \<Turnstile>' \<phi> \<longrightarrow> \<Delta>o = \<Delta>,(z:\<star>) \<longrightarrow> \<not> z free in pretype \<phi> \<longrightarrow> \<Delta> \<Turnstile>' \<phi>")
      case (wf_forall \<Delta>n La \<tau>n X)
      show ?case proof rule+
        fix \<Delta>
        assume "\<Delta>n \<Turnstile>' \<forall>.\<tau>n^X" and H\<Delta>_eq: "\<Delta>n = \<Delta>,(z:\<star>)" and Hz: "z \<notin> set (free_pty (\<forall>.\<tau>n^X))"
        show "\<forall>a. a \<notin> La \<longrightarrow> \<Delta>,(a:\<star>) \<Turnstile>' freshen_ty a \<tau>n" proof rule+
          fix a
          assume Ha_la: "a \<notin> La"
          have IH: "\<forall>a. a \<notin> La \<longrightarrow>
            \<Delta>n,(a:\<star>) \<Turnstile>' freshen_ty a \<tau>n \<and>
            (\<forall>\<Delta>. \<Delta>n,(a:\<star>) = \<Delta>,(z:\<star>) \<longrightarrow>
            \<Delta>,(z:\<star>) \<Turnstile>' freshen_ty a \<tau>n \<longrightarrow> z \<notin> set (free_ty (freshen_ty a \<tau>n)) \<longrightarrow> \<Delta> \<Turnstile>' freshen_ty a \<tau>n)"
            by assumption
          hence "\<forall>\<Delta>. \<Delta>n,(a:\<star>) = \<Delta>,(z:\<star>) \<longrightarrow>
            \<Delta>,(z:\<star>) \<Turnstile>' freshen_ty a \<tau>n \<longrightarrow> z \<notin> set (free_ty (freshen_ty a \<tau>n)) \<longrightarrow> \<Delta> \<Turnstile>' freshen_ty a \<tau>n"
            using Ha_la by auto
          hence "\<Delta>,(a:\<star>),(z:\<star>) \<Turnstile>' freshen_ty a \<tau>n \<longrightarrow> z \<notin> set (free_ty (freshen_ty a \<tau>n)) \<longrightarrow> 
            \<Delta>,(a:\<star>) \<Turnstile>' freshen_ty a \<tau>n" using H\<Delta>_eq by auto
          moreover
          have "\<Delta>n \<Turnstile>' \<forall>.\<tau>n^X" by assumption
          hence "\<Delta>n,(a:\<star>) \<Turnstile>' freshen_ty a \<tau>n" using wf_forall Ha_la by auto
          hence Hwf: "\<Delta>,(z:\<star>),(a:\<star>) \<Turnstile>' freshen_ty a \<tau>n" using H\<Delta>_eq by auto
          ultimately
          have Hg: "z \<notin> set (free_ty (freshen_ty a \<tau>n)) \<longrightarrow> \<Delta>,(a:\<star>) \<Turnstile>' freshen_ty a \<tau>n"
            by (auto simp add: env_permute)
          show "\<Delta>,(a:\<star>) \<Turnstile>' freshen_ty a \<tau>n" proof (cases "z = a")
            assume "z = a"
            thus ?thesis using Hwf by auto
          next
            assume Hdiff: "z \<noteq> a"
            have "\<not> z free in type \<tau>n" using Hz by auto
            hence "\<forall>i. \<not> z free in type (vsubst_ty (TyFreeVar a) i \<tau>n)" using Hdiff
              by (induct \<tau>n arbitrary: i taking: 
                "%\<phi>. \<forall>i. \<not> z free in pretype \<phi> \<longrightarrow> \<not> z free in pretype (vsubst_pty (TyFreeVar a) i \<phi>)", 
                auto)
            thus ?thesis using Hg by auto
          qed
        qed
      next
        fix \<Delta>
        assume "\<Delta>n = \<Delta>,(z:\<star>)"
        thus "\<Delta> \<Turnstile>' TY" using prems by (auto simp add: wf_env_ty_def)
      qed auto
    next      
      case wf_unit thus ?case by (auto simp add: wf_env_ty_def)
    next      
      case wf_base thus ?case by (auto simp add: wf_env_ty_def)
    next      
      case wf_ref thus ?case by (auto simp add: wf_env_ty_def)
    next      
      case wf_arrow thus ?case by (auto simp add: wf_env_ty_def)
    next      
      case wf_tvar thus ?case by (auto simp add: wf_env_ty_def)
    next      
      case (wf_ty \<Delta> \<phi> q \<pi> \<Delta>') 
      hence "\<Delta> \<Turnstile>' \<phi>" and "z \<notin> set (free_pty \<phi>)" and "\<Delta> = \<Delta>',(z:\<star>)" by auto
      thus ?case using wf_ty.hyps by auto
    qed
  }
  thus ?thesis by auto
qed

lemma poly_wf_no_free_db_pty:
  fixes \<phi> :: pty
  assumes "\<Delta> \<Turnstile>' \<phi>"
  shows "free_db_pty \<phi> = []"
using prems proof (induct taking: "%\<Delta> \<tau>. \<Delta> \<Turnstile>' \<tau> \<longrightarrow> free_db_ty \<tau> = []")
  case (wf_forall \<Delta> La \<tau>ins) 
  moreover
  then obtain z where "z \<notin> La" using fresh_exists by auto
  ultimately
  have "set (free_db_ty (freshen_ty z \<tau>ins)) = {}" by auto
  hence "set (free_db_ty \<tau>ins) - {0} = {}" by (auto simp add: vsubst_ty_free_db_ty)
  hence "set (free_db_ty \<tau>ins) \<subseteq> {0}" by auto
  hence "set (shift_db_list (free_db_ty \<tau>ins)) = {}" by (auto simp add: shift_many_sanity)
  hence "shift_db_list (free_db_ty \<tau>ins) = []" by auto
  thus ?case by auto
qed auto

lemma poly_wf_no_free_db_ty:
  fixes \<tau> :: ty
  assumes "\<Delta> \<Turnstile>' \<tau>"
  shows "free_db_ty \<tau> = []"
using prems poly_wf_no_free_db_pty by (induct, auto)

lemma poly_wf_ty_subst_equal:
  fixes \<tau> :: ty
  assumes "\<Delta> \<Turnstile>' \<tau>"
  shows "\<tau> = vsubst_ty \<phi>s i \<tau>"
using prems by (auto simp add: poly_wf_no_free_db_ty poly_no_free_db_vars_same_subst_ty)

lemma wf_env_subst_eq:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  shows "\<Gamma> = vsubst_ty_env \<phi> i \<Gamma>"
proof (auto simp add: vsubst_ty_env_def)
  fix x b
  assume "(x, b) \<in> \<Gamma>"
  then obtain \<tau> where "b = Term \<tau>" by (cases b)
  hence "\<Delta> \<Turnstile>' \<tau>" using prems by (auto simp add: poly_env_wf_ty)
  hence "\<tau> = vsubst_ty \<phi> i \<tau>" by (rule poly_wf_ty_subst_equal)
  thus "\<exists>\<tau>. b = Term (vsubst_ty \<phi> i \<tau>) \<and> (x\<rhd>\<tau>) \<in> \<Gamma>" using prems by auto
next
  fix x \<tau>
  assume "(x\<rhd>\<tau>) \<in> \<Gamma>"
  hence "\<Delta> \<Turnstile>' \<tau>" using prems using poly_env_wf_ty by auto
  hence "\<tau> = vsubst_ty \<phi> i \<tau>" by (rule poly_wf_ty_subst_equal)
  thus "(x\<rhd>(vsubst_ty \<phi> i \<tau>)) \<in> \<Gamma>" by auto
qed

lemma poly_subst_var_type_ok:
  fixes \<phi>s :: pty
  assumes "\<Delta> \<Turnstile>' vsubst_ty (TyFreeVar z) i \<tau>"
  and "\<Delta> \<Turnstile>' \<phi>s"
  shows "\<Delta> \<Turnstile>' vsubst_ty \<phi>s i \<tau>"
proof -
  { fix \<Delta>o \<tau>o
    assume Ho: "\<tau>o = vsubst_ty (TyFreeVar z) i \<tau>"
    hence "\<Delta> \<Turnstile>' \<tau>o" using assms by auto
    hence ?thesis using assms Ho
    proof (induct arbitrary: i \<tau> taking: 
      "%\<Delta>. %\<phi>o. \<forall>i \<phi>. \<Delta> \<Turnstile>' \<phi>s \<longrightarrow> \<phi>o = vsubst_pty (TyFreeVar z) i \<phi> \<longrightarrow> \<Delta> \<Turnstile>' \<phi>o \<longrightarrow> 
        \<Delta> \<Turnstile>' vsubst_pty \<phi>s i \<phi>")
      case (wf_ty \<Delta> \<phi>o q \<pi> i) 
      then obtain \<phi> where Heq: "\<tau> = Qual q \<pi> \<phi>" by (cases \<tau>, auto)
      hence "\<phi>o = vsubst_pty (TyFreeVar z) i \<phi>" using wf_ty by auto
      hence "\<Delta> \<Turnstile>' vsubst_pty \<phi>s i \<phi>" using wf_ty by auto
      thus ?case using Heq by auto
    next
      case (wf_unit \<Delta>) thus ?case proof auto
        fix i \<phi>
        assume "TyUnit = vsubst_pty (TyFreeVar z) i \<phi>"
          and "\<Delta> \<Turnstile>' TY"
        hence "\<phi> = TyUnit" by (cases \<phi>, auto simp add: split_if)
        thus "\<Delta> \<Turnstile>' vsubst_pty \<phi>s i \<phi>" by auto
      qed
    next
      case (wf_base \<Delta> n) thus ?case proof auto
        fix i \<phi>
        assume "TyBase n = vsubst_pty (TyFreeVar z) i \<phi>"
          and "\<Delta> \<Turnstile>' TY"
        hence "\<phi> = TyBase n" by (cases \<phi>, auto simp add: split_if)
        thus "\<Delta> \<Turnstile>' vsubst_pty \<phi>s i \<phi>" by auto
      qed
    next
      case (wf_tvar a \<Delta>) thus ?case proof auto
        fix i \<phi>
        assume H: "TyFreeVar a = vsubst_pty (TyFreeVar z) i \<phi>"
          and Hs: "\<Delta> \<Turnstile>' \<phi>s"
          and "\<Delta> \<Turnstile>' TY"
        thus "\<Delta> \<Turnstile>' vsubst_pty \<phi>s i \<phi>" proof (cases "\<phi> = TyVar i", auto)
          case False 
          hence "\<phi> = TyFreeVar a" using H by (cases \<phi>, auto)
          thus ?thesis by auto
        qed
      qed
    next
      case (wf_ref \<Delta> \<tau>n) thus ?case proof auto
        fix i \<phi>
        assume H: "\<Delta> \<Turnstile>' \<phi>s"
          and Heq: "TyRef \<tau>n = vsubst_pty (TyFreeVar z) i \<phi>"
          and Hwf: "\<Delta> \<Turnstile>' vsubst_pty (TyFreeVar z) i \<phi>"
          and IH: "\<And>i \<tau>. \<lbrakk>\<Delta> \<Turnstile>' vsubst_ty (TyFreeVar z) i \<tau>; \<tau>n = vsubst_ty (TyFreeVar z) i \<tau>\<rbrakk>
                 \<Longrightarrow> \<Delta> \<Turnstile>' vsubst_ty \<phi>s i \<tau>"
        then obtain \<tau> where H\<tau>: "\<phi> = TyRef \<tau>" by (cases \<phi>, auto simp add: split_if)
        hence Hih: "\<Delta> \<Turnstile>' vsubst_ty (TyFreeVar z) i \<tau>" using Hwf by auto
        hence "\<Delta> \<Turnstile>' vsubst_ty \<phi>s i \<tau>" using IH Heq Hih Hwf H\<tau> by auto
        hence "\<Delta> \<Turnstile>' vsubst_pty \<phi>s i (TyRef \<tau>)" by auto
        thus "\<Delta> \<Turnstile>' vsubst_pty \<phi>s i \<phi>" using H\<tau> by auto
      qed
    next
      case (wf_arrow \<Delta> \<tau>1 \<tau>2 Z) thus ?case proof auto
        fix i \<phi> 
        assume "\<Delta> \<Turnstile>' \<tau>1"
          and IH1: "\<And>i \<tau>. \<lbrakk>\<Delta> \<Turnstile>' vsubst_ty (TyFreeVar z) i \<tau>;
                  \<tau>1 = vsubst_ty (TyFreeVar z) i \<tau>\<rbrakk>
                 \<Longrightarrow> \<Delta> \<Turnstile>' vsubst_ty \<phi>s i \<tau>"
          and "\<Delta> \<Turnstile>' \<tau>2"
          and IH2: "\<And>i \<tau>. \<lbrakk>\<Delta> \<Turnstile>' vsubst_ty (TyFreeVar z) i \<tau>;
                  \<tau>2 = vsubst_ty (TyFreeVar z) i \<tau>\<rbrakk>
                 \<Longrightarrow> \<Delta> \<Turnstile>' vsubst_ty \<phi>s i \<tau>"
          and "\<Delta> \<Turnstile>' \<phi>s"
          and Heq: "\<tau>1 \<rightarrow>^Z \<tau>2 = vsubst_pty (TyFreeVar z) i \<phi>"
          and Hwf:  "\<Delta> \<Turnstile>' vsubst_pty (TyFreeVar z) i \<phi>"
        obtain \<tau>n1 \<tau>n2 where H\<phi>: "\<phi> = \<tau>n1 \<rightarrow>^Z \<tau>n2" 
          using Heq by (cases \<phi>, auto simp add: split_if)
        hence Hih1: "\<Delta> \<Turnstile>' vsubst_ty (TyFreeVar z) i \<tau>n1" using Hwf by auto
        hence "\<Delta> \<Turnstile>' vsubst_ty \<phi>s i \<tau>n1" using Hwf IH1 H\<phi> Heq by auto
        moreover
        have Hih2: "\<Delta> \<Turnstile>' vsubst_ty (TyFreeVar z) i \<tau>n2" using Hwf H\<phi> by auto
        hence "\<Delta> \<Turnstile>' vsubst_ty \<phi>s i \<tau>n2" using Hwf IH2 H\<phi> Heq by auto
        ultimately
        show "\<Delta> \<Turnstile>' vsubst_pty \<phi>s i \<phi>" using H\<phi> by auto
      qed
    next
      case (wf_forall \<Delta> La \<tau> X) thus ?case proof auto
        fix i \<phi> 
        assume "\<Delta> \<Turnstile>' TY"
          and "finite La"
          and IH: "\<forall>a. a \<notin> La \<longrightarrow> \<Delta>,(a:\<star>) \<Turnstile>' (vsubst_ty (TyFreeVar a) 0 \<tau>) \<and>
               (\<forall>x xa. \<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty (TyFreeVar z) x xa \<longrightarrow>
                       \<Delta>,(a:\<star>) \<Turnstile>' \<phi>s \<longrightarrow>
                       vsubst_ty (TyFreeVar a) 0 \<tau> = vsubst_ty (TyFreeVar z) x xa \<longrightarrow>
                       \<Delta>,(a:\<star>) \<Turnstile>' (vsubst_ty \<phi>s x xa))"
          and "\<Delta> \<Turnstile>' \<phi>s"
          and Heq: "\<forall>.\<tau>^X = vsubst_pty (TyFreeVar z) i \<phi>"
          and Hwf: "\<Delta> \<Turnstile>' vsubst_pty (TyFreeVar z) i \<phi>"
        obtain \<tau>n where H\<phi>: "\<phi> = \<forall>.\<tau>n^X" using Heq by (cases \<phi>, auto simp add: split_if)
        have H: "\<forall>a. a \<notin> La \<longrightarrow> \<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty (TyFreeVar a) 0 (vsubst_ty \<phi>s (Suc i) \<tau>n)" proof auto
          fix a
          assume Hla: "a \<notin> La"
          have Hwf\<tau>: "\<Delta>,(a:\<star>) \<Turnstile>' (vsubst_ty (TyFreeVar a) 0 \<tau>)" and 
               IH': "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty (TyFreeVar z) (Suc i) (vsubst_ty (TyFreeVar a) 0 \<tau>n) \<longrightarrow>
                       \<Delta>,(a:\<star>) \<Turnstile>' \<phi>s \<longrightarrow>
                       vsubst_ty (TyFreeVar a) 0 \<tau> = 
                       vsubst_ty (TyFreeVar z) (Suc i) (vsubst_ty (TyFreeVar a) 0 \<tau>n) \<longrightarrow>
                       \<Delta>,(a:\<star>) \<Turnstile>' (vsubst_ty \<phi>s (Suc i) (vsubst_ty (TyFreeVar a) 0 \<tau>n))"
            using IH Hla by auto
          hence "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty (TyFreeVar z) (Suc i) (vsubst_ty (TyFreeVar a) 0 \<tau>n)" 
            using Hwf\<tau> Heq H\<phi> subst_subst_ty_swap[of 0 "Suc i"] by auto
          moreover
          have "\<Delta>,(a:\<star>) \<Turnstile>' \<phi>s" using poly_tenv_cons prems by auto
          moreover
          have "vsubst_ty (TyFreeVar a) 0 \<tau> = 
            vsubst_ty (TyFreeVar z) (Suc i) (vsubst_ty (TyFreeVar a) 0 \<tau>n)" 
            using Hwf\<tau> Heq H\<phi> subst_subst_ty_swap[of 0 "Suc i"] by auto
          ultimately
          have "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty \<phi>s (Suc i) (vsubst_ty (TyFreeVar a) 0 \<tau>n)" using IH' by auto
          moreover
          have "free_db_pty \<phi>s = []" using poly_wf_no_free_db_pty prems by auto
          hence "\<forall>k. vsubst_pty (TyFreeVar a) k \<phi>s = \<phi>s"
            using poly_no_free_db_vars_same_subst_pty by auto
          ultimately
          show "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty (TyFreeVar a) 0 (vsubst_ty \<phi>s (Suc i) \<tau>n)" 
            using subst_subst_ty_swap[of 0 "Suc i" a \<phi>s \<tau>n] by auto
        qed
        have "\<Delta> \<Turnstile>' \<forall>.(vsubst_ty \<phi>s (Suc i) \<tau>n)^X" apply rule using H prems by auto
        thus "\<Delta> \<Turnstile>' vsubst_pty \<phi>s i \<phi>" using H\<phi> by auto
      qed
    qed
  }
  thus ?thesis by auto
qed

lemma poly_ty_wf_type:
  fixes \<tau> :: ty
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>"
  shows "\<Delta> \<Turnstile>' \<tau>"
proof -
  have A: "\<Gamma>;\<Delta> \<Turnstile> OK" using ty_wf_env prems by auto
  have B: "\<Delta> \<Turnstile>' TY" using poly_wf_env_ty prems by auto
  from prems A B show ?thesis proof induct
    case t_unit thus ?case by auto
  next
    case t_var thus ?case using poly_env_wf_ty by auto
  next
    case (t_abs \<Delta> X Z M \<Gamma> \<tau>1 \<tau>2 e q \<pi> Lx)
    have Hgok: "\<Gamma>;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<tau>1" and "finite Lx" by assumption
    then obtain x where  "\<Gamma>,(x:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma> \<sim> {x\<rhd>\<tau>1}" and "x \<notin> Lx"
      using new_variable_abs by blast
    hence "\<Delta> \<Turnstile>' \<tau>2" using t_abs by auto
    thus ?case using t_abs by auto
  next
    case (t_app \<Delta> X Z1 Z2 Z M \<Gamma>1 \<Gamma>2 \<tau>1 \<tau>2 e1 e2 q)
    hence "\<Delta> \<Turnstile>' Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)" using wf_env_subset by blast
    thus ?case by auto
  next
    case (t_poly \<Delta> Z Z' M \<Gamma> \<tau> v q \<pi> La)
    have "\<Delta> \<Turnstile>' \<forall>.\<tau>^Z'" apply (rule wf_forall) proof -
      {
        fix z
        assume Hla: "z \<notin> La"
        have "\<Delta>,(z:\<star>) \<Turnstile>' TY" using t_poly by (auto simp add: wf_env_ty_def)
        moreover
        have "\<Gamma>;\<Delta>,(z:\<star>) \<Turnstile> OK" using poly_env_ty_cons t_poly by auto
        ultimately
        have "\<Delta>,(z:\<star>) \<Turnstile>' freshen_ty z \<tau>" using t_poly Hla by auto
      }
      thus "\<forall>a. a \<notin> La \<longrightarrow> \<Delta>,(a:\<star>) \<Turnstile>' freshen_ty a \<tau>" by auto
    qed 
  next
    case (t_ptyapp \<Delta> Z Z\<phi> M \<Gamma> \<tau> \<phi>s e1 q X)
    have "finite (dom \<Delta>)" using t_ptyapp poly_wf_env_ty_fin by auto
    have "\<Delta> \<Turnstile>' Qual q ScBot (\<forall>.\<tau>^X)" using t_ptyapp by auto
    then obtain La where "\<forall>a. a \<notin> La \<longrightarrow> \<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty (TyFreeVar a) 0 \<tau>" 
      and "finite La" by auto
    moreover
    then obtain a where "a \<notin> La" and  Hfresh: "\<not> a free in type (vsubst_ty \<phi>s 0 \<tau>)"
      using fresh_exists by blast
    ultimately
    have "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty (TyFreeVar a) 0 \<tau>" by auto
    moreover
    have "\<Delta>,(a:\<star>) \<Turnstile>' \<phi>s" using poly_tenv_cons t_ptyapp by auto
    ultimately
    have "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty \<phi>s 0 \<tau>" using poly_subst_var_type_ok[of "\<Delta>,(a:\<star>)" a 0 \<tau> \<phi>s] by auto
    thus ?case using poly_wf_ty_strength Hfresh by blast
  next
    case t_loc thus ?case using poly_env_wf_ty by auto
  next
    case (t_swap \<Delta> Z1 Z2 M \<Gamma>1 \<Gamma>2 \<tau> e1 e2 \<pi>)
    hence "\<Gamma>2;\<Delta> \<Turnstile> OK" using wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2" _ \<Gamma>2] by auto 
    thus ?case using t_swap by auto
  next
    case (t_let \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z e e1 e2 \<tau> \<tau>1 \<tau>2 Lx Ly) 
    have "\<Gamma>;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>] by auto
    hence HG1: "\<Delta> \<Turnstile>' \<tau>" using t_let by auto
    have "\<Gamma>1;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>1] by auto
    moreover
    have "Z1\<union>Z2 \<Turnstile> ScopeSet" using t_let by (auto simp add: wf_scopeset_def)
    moreover
    have "finite Lx" by assumption
    ultimately
    obtain \<rho> x
      where "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK" and "\<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>}" 
      and "\<rho> \<notin> set (free_let_dollar_scope e1)" and "\<rho> \<notin> Environ.dom (Z2 \<union> Z1)"
      and "x \<notin> Lx"
      using new_scope_let HG1 by blast
    hence HG2: "\<Delta> \<Turnstile>' \<tau>1" using t_let by auto
    have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
    moreover
    have "finite Lx" and "finite Ly" by assumption
    ultimately
    obtain x y where "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
      and "x \<notin> Lx" and "y \<notin> Ly" and "x \<noteq> y"
      using new_variable_let[of \<Gamma>2 \<Delta> \<tau> \<tau>1 Lx Ly] HG1 HG2 by auto
    thus ?case using t_let by auto
  next
    case (t_let' \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z \<rho> x e1 e2 \<tau> \<tau>1 \<tau>2 Zf Lx Ly) 
    have "\<Gamma>;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>] by auto
    hence HG1: "\<Delta> \<Turnstile>' \<tau>" using t_let' by auto
    have HG2: "\<Delta> \<Turnstile>' \<tau>1" using t_let' ty_wf_env by auto
    have Hlx: "finite Lx" and Hly: "finite Ly" by assumption
    have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
    then obtain x y where "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
      and "x \<notin> Lx" and "y \<notin> Ly" and "x \<noteq> y"
      using new_variable_let[of \<Gamma>2 \<Delta> \<tau> \<tau>1 Lx Ly] HG1 HG2 Hlx Hly by auto
    thus ?case using t_let' by auto
  next
    case (t_poly \<Delta> Z X M \<Gamma> \<tau> v q \<pi> La) show ?case apply rule proof (rule wf_forall[where La=La])
      {
        fix a 
        assume Hla: "a \<notin> La"
        hence "\<Gamma>;\<Delta>,(a:\<star>) \<Turnstile> OK \<longrightarrow> \<Delta>,(a:\<star>) \<Turnstile>' TY \<longrightarrow> \<Delta>,(a:\<star>) \<Turnstile>' freshen_ty a \<tau>" using t_poly by auto
        moreover
        have "\<Gamma>;\<Delta>,(a:\<star>) \<Turnstile> OK" using t_poly poly_env_ty_cons by auto 
        moreover have "\<Delta>,(a:\<star>) \<Turnstile>' TY" using t_poly by (auto simp add: wf_env_ty_def)
        ultimately
        have "\<Delta>,(a:\<star>) \<Turnstile>' freshen_ty a \<tau>" by auto
      }
      thus "\<forall>a. a \<notin> La \<longrightarrow> \<Delta>,(a:\<star>) \<Turnstile>' freshen_ty a \<tau>" by auto 
    qed
  qed auto
qed

lemma ty_wf_scope:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>"
  shows "Z \<Turnstile> ScopeSet"
using prems proof induct
  case (t_abs \<Delta> X Z M \<Gamma> \<tau>1 \<tau>2 e q \<pi> Lx)
  have "Z \<Turnstile> ScopeSet" using t_abs minimal_is_scope by auto
  moreover
  have Hgok: "\<Gamma>;\<Delta> \<Turnstile> OK" and "finite Lx" and "\<Delta> \<Turnstile>' \<tau>1" by assumption
  then obtain x where  "\<Gamma>,(x:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma> \<sim> {x\<rhd>\<tau>1}" and "x \<notin> Lx" 
    using new_variable_abs by blast
  hence "filterScFreeX X \<Turnstile> ScopeSet" using t_abs by auto
  ultimately
  show ?case using t_abs by (auto simp add: wf_scopeset_def)
next
  case (t_let \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z e e1 e2 \<tau> \<tau>1 \<tau>2 Lx Ly)
  have Hlx: "finite Lx" and Hly: "finite Ly" by assumption
  have "Z \<Turnstile> ScopeSet" using t_let by auto
  moreover
  have H\<tau>: "\<Delta> \<Turnstile>' \<tau>" using t_let poly_ty_wf_type by auto 
  have H1: "\<Gamma>1;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>1] by auto
  have "Z1\<union>Z2 \<Turnstile> ScopeSet" using t_let by (auto simp add: wf_scopeset_def)
  then obtain \<rho> x
    where "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK" and "\<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>}" 
    and "\<rho> \<notin> set (free_let_dollar_scope e1)" and "\<rho> \<notin> Environ.dom (Z2 \<union> Z1)"
    and "x \<notin> Lx"
    using new_scope_let H1 H\<tau> Hlx by blast
  hence "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M;Z1 \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1
    \<or> \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1" 
    using t_let.hyps by auto
  hence H\<tau>1: "\<Delta> \<Turnstile>' \<tau>1" using poly_ty_wf_type by auto
  have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
  then obtain x y where "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
    and "x \<notin> Lx" and "y \<notin> Ly" 
    using new_variable_let[where Lx=Lx and Ly=Ly] H\<tau> H\<tau>1 Hlx Hly by blast
  hence "Z2 \<Turnstile> ScopeSet" using t_let by auto
  ultimately
  show ?case using t_let by (auto simp add: wf_scopeset_def)
next
  case (t_let' \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z \<rho> x e1 e2 \<tau> \<tau>1 \<tau>2 _ Lx Ly)
  have Hlx: "finite Lx" and Hly: "finite Ly" by assumption
  have "Z \<Turnstile> ScopeSet" using t_let' by (auto simp add: wf_scopeset_def)
  moreover
  have "\<Delta> \<Turnstile>' convT \<rho> \<tau>" using t_let' poly_ty_wf_type by auto
  hence H\<tau>: "\<Delta> \<Turnstile>' \<tau>" by (cases \<tau>, auto)
  have H\<tau>1: "\<Delta> \<Turnstile>' \<tau>1" using t_let' ty_wf_env poly_ty_wf_type by auto
  have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
  then obtain x y where "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
    using new_variable_let[where Lx=Lx and Ly=Ly] H\<tau> H\<tau>1 Hlx Hly by blast
  hence "Z2 \<Turnstile> ScopeSet" using t_let' by auto
  ultimately
  show ?case using t_let' by (auto simp add: wf_scopeset_def)
qed (insert wf_type_scopeset wf_pretype_scopeset, 
    auto simp add: wf_scopeset_def exists_x minimalValidScope_def)

lemma valid_scope_weaken:
  assumes "validScope \<pi>  Z"
  shows "validScope \<pi> (Z \<union> Z')"
using prems by (cases \<pi>, auto simp add: dom_def)

lemma minimal_is_valid:
  assumes "minimalValidScope \<pi> Z"
  shows "validScope \<pi> Z"
using prems by (auto simp add: minimalValidScope_def dom_def)

lemma wf_subst_ty_wf_ty:
  fixes \<tau> :: ty
  assumes "Z \<Turnstile> vsubst_ty (TyFreeVar a) i \<tau>"
  shows "Z \<Turnstile> \<tau>"
proof -
  {
    fix \<tau>n
    assume Heq: "\<tau>n = vsubst_ty (TyFreeVar a) i \<tau>"
    hence "Z \<Turnstile> \<tau>n" using assms by auto
    hence ?thesis using Heq
    proof (induct arbitrary: i \<tau> taking:
        "%Z. %\<phi>n::pty. \<forall>i \<phi>. Z \<Turnstile> \<phi>n \<longrightarrow> \<phi>n = vsubst_pty (TyFreeVar a) i \<phi> \<longrightarrow> Z \<Turnstile> \<phi>")
      case ty_unit thus ?case proof auto
        fix i \<phi>
        assume "TyUnit = vsubst_pty (TyFreeVar a) i \<phi>"
        hence "\<phi> = TyUnit" by (cases \<phi>, auto simp add: split_if)
        thus "{} \<Turnstile> \<phi>" by auto
      qed
    next
      case (ty_base n) thus ?case proof auto
        fix i \<phi> 
        assume "TyBase n = vsubst_pty (TyFreeVar a) i \<phi>"
        hence "\<phi> = TyBase n" by (cases \<phi>, auto simp add: split_if)
        thus "{} \<Turnstile> \<phi>" by auto
      qed
    next
      case (ty_tvar x) thus ?case proof auto
        fix i \<phi> 
        assume H: "TyFreeVar x = vsubst_pty (TyFreeVar a) i \<phi>"
        thus "{} \<Turnstile> \<phi>" proof (cases "\<phi> = TyVar i", auto)
          case False 
          hence "\<phi> = TyFreeVar x" using H by (cases \<phi>, auto)
          thus ?thesis by auto
        qed
      qed
    next
      case (ty_tbvar n) thus ?case proof auto
        fix i \<phi> 
        assume H: "TyVar n = vsubst_pty (TyFreeVar a) i \<phi>"
        hence "\<phi> = TyVar n" by (cases \<phi>, auto simp add: split_if)
        thus "{} \<Turnstile> \<phi>" by auto
      qed
    next
      case (ty_arrow Z1 \<tau>1 Z2 \<tau>2 Z X) thus ?case proof (auto simp del: filterScFreeX filterScVarX)
        fix i \<phi> 
        assume "Z1 \<Turnstile> \<tau>1"
          and IH1: "\<And>i \<tau>. \<tau>1 = vsubst_ty (TyFreeVar a) i \<tau> \<Longrightarrow> Z1 \<Turnstile> \<tau>"
          and "Z2 \<Turnstile> \<tau>2"
          and IH2: "\<And>i \<tau>. \<tau>2 = vsubst_ty (TyFreeVar a) i \<tau> \<Longrightarrow> Z2 \<Turnstile> \<tau>"
          and Hg1: "filterScFreeX X \<Turnstile> ScopeSet"
          and Hg2: "filterScVarX X = {}"
          and Hg3: "Z1 \<union> Z2 \<union> filterScFreeX X \<Turnstile> vsubst_pty (TyFreeVar a) i \<phi>"
          and Heq: "\<tau>1 \<rightarrow>^X \<tau>2 = vsubst_pty (TyFreeVar a) i \<phi>"
        obtain \<tau>n1 \<tau>n2 where 
          H1: "\<tau>1 = vsubst_ty (TyFreeVar a) i \<tau>n1" and
          H2: "\<tau>2 = vsubst_ty (TyFreeVar a) i \<tau>n2" and
          Heq\<phi>: "\<phi> = \<tau>n1 \<rightarrow>^X \<tau>n2" using Heq by (cases \<phi>, auto simp add: split_if)
        have "Z1 \<Turnstile> \<tau>n1" using IH1 H1 by auto
        moreover
        have "Z2 \<Turnstile> \<tau>n2" using IH2 H2 by auto
        ultimately
        show "Z1 \<union> Z2 \<union> filterScFreeX X \<Turnstile> \<phi>" 
          using Heq\<phi> Hg1 Hg2 Hg3 by auto 
      qed
    next
      case (ty_forall Zi \<tau>i X) thus ?case proof (auto simp del: filterScFreeX filterScVarX)
        fix i \<phi> 
        assume "Zi \<Turnstile> \<tau>i"
          and IH: "\<And>i \<tau>. \<tau>i = vsubst_ty (TyFreeVar a) i \<tau> \<Longrightarrow> Zi \<Turnstile> \<tau>"
          and "Zi \<union> filterScFreeX X \<Turnstile> vsubst_pty (TyFreeVar a) i \<phi>"
          and "\<forall>.\<tau>i^X = vsubst_pty (TyFreeVar a) i \<phi>"
          and Hg1: "filterScVarX X = {}"
          and Hg2: "filterScFreeX X \<Turnstile> ScopeSet"
        then obtain \<tau>n where Heq\<phi>: "\<phi> = \<forall>.\<tau>n^X" and "\<tau>i = vsubst_ty (TyFreeVar a) (Suc i) \<tau>n"
          by (cases \<phi>, auto simp add: split_if)
        hence "Zi \<Turnstile> \<tau>n" using IH by auto
        thus "Zi \<union> filterScFreeX X \<Turnstile> \<phi>" using Heq\<phi> Hg1 Hg2 by auto
      qed
    next
      case (ty_ref Zi \<tau>i) thus ?case proof auto
        fix i \<phi> 
        assume "Zi \<Turnstile> \<tau>i"
          and IH: "\<And>i \<tau>. \<tau>i = vsubst_ty (TyFreeVar a) i \<tau> \<Longrightarrow> Zi \<Turnstile> \<tau>"
          and "Zi \<Turnstile> vsubst_pty (TyFreeVar a) i \<phi>"
          and "TyRef \<tau>i = vsubst_pty (TyFreeVar a) i \<phi>"
        then obtain \<tau>n where Heq\<phi>: "\<phi> = TyRef \<tau>n" and "\<tau>i = vsubst_ty (TyFreeVar a) i \<tau>n"
          by (cases \<phi>, auto simp add: split_if)
        hence "Zi \<Turnstile> \<tau>n" using IH by auto
        thus "Zi \<Turnstile> \<phi>" using Heq\<phi> by auto
      qed
    next
      case (ty_qual Z\<phi> \<phi> \<pi> Z q i) 
      then obtain \<phi>n where Heq: "\<tau> = Qual q \<pi> \<phi>n" by (cases \<tau>, auto)
      hence "Z\<phi> \<Turnstile> \<phi>" using ty_qual by auto
      show ?case using ty_qual Heq by auto
    qed 
  }
  thus ?thesis by auto
qed

lemma wf_ty_wf_subst_ty:
  fixes \<tau> :: ty
  fixes \<phi>s :: pty
  assumes "Z \<Turnstile> \<tau>"
  and "Zs \<Turnstile> \<phi>s"
  shows "\<exists>Zr. Zr \<Turnstile> vsubst_ty \<phi>s i \<tau> \<and> Zr \<subseteq> Z \<union> Zs"
using prems proof (induct arbitrary: i taking:
  "%Z \<phi>. \<forall>i. Z \<Turnstile> \<phi> \<longrightarrow> Zs \<Turnstile> \<phi>s \<longrightarrow> (\<exists>Zr. Zr \<Turnstile> vsubst_pty \<phi>s i \<phi> \<and> Zr \<subseteq> Z \<union> Zs)")
  case (ty_arrow Z1 \<tau>1 Z2 \<tau>2 Z X) thus ?case proof (auto simp del: filterScFreeX)
    fix i
    assume Hs: "Zs \<Turnstile> \<phi>s"
    obtain Zr1 where "Zr1 \<Turnstile> vsubst_ty \<phi>s i \<tau>1" and Hsub1: "Zr1 \<subseteq> Z1 \<union> Zs" using Hs ty_arrow by blast
    moreover
    obtain Zr2 where "Zr2 \<Turnstile> vsubst_ty \<phi>s i \<tau>2" and Hsub2: "Zr2 \<subseteq> Z2 \<union> Zs" using Hs ty_arrow by blast
    ultimately
    have "Zr1\<union>Zr2\<union>Z \<Turnstile> vsubst_ty \<phi>s i \<tau>1 \<rightarrow>^X vsubst_ty \<phi>s i \<tau>2" using ty_arrow by auto
    moreover
    have "Zr1\<union>Zr2\<union>Z \<subseteq> Z1 \<union> Z2 \<union> filterScFreeX X \<union> Zs" using Hsub1 Hsub2 ty_arrow by auto
    ultimately
    show "\<exists>Zr. Zr \<Turnstile> vsubst_ty \<phi>s i \<tau>1 \<rightarrow>^X vsubst_ty \<phi>s i \<tau>2 \<and> Zr \<subseteq> Z1 \<union> Z2 \<union> filterScFreeX X \<union> Zs"
      by auto
  qed
next
  case (ty_qual Z\<phi> \<phi> \<pi> Z q i) 
  then obtain Zr where "Zr \<Turnstile> vsubst_pty \<phi>s i \<phi>" and "Zr \<subseteq> Z\<phi> \<union> Zs" by auto
  thus ?case using ty_qual by auto
next
  case (ty_forall Z \<tau> X Zx) thus ?case proof (auto simp del: filterScVarX filterScFreeX)
    fix i
    assume Hs: "Zs \<Turnstile> \<phi>s"
    obtain Zr where "Zr \<Turnstile> vsubst_ty \<phi>s (Suc i) \<tau>" and Hsub: "Zr \<subseteq> Z \<union> Zs" using Hs ty_forall by blast
    hence "Zr\<union>(filterScFreeX X) \<Turnstile> vsubst_pty \<phi>s i (\<forall>.\<tau>^X)" using ty_forall by auto
    moreover
    have "Zr\<union>(filterScFreeX X) \<subseteq> Z \<union> filterScFreeX X \<union> Zs" using Hsub ty_forall by auto
    ultimately
    show "\<exists>Zr. Zr \<Turnstile> (\<forall>.(vsubst_ty \<phi>s (Suc i) \<tau>)^X) \<and> Zr \<subseteq> Z \<union> filterScFreeX X \<union> Zs" 
      by auto
  qed
qed auto

(*
lemma ty_wf_type:
  fixes \<tau> :: ty
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>"
  shows "\<exists>Z'. Z' \<Turnstile> \<tau> \<and> Z' \<subseteq> Z"
using prems proof induct
  case (t_abs \<Delta> Z1 Z2 Z' M \<Gamma> \<tau>1 \<tau>2 e q \<pi> X) 
  have Hgok: "\<Gamma>;\<Delta> \<Turnstile> OK" by assumption
  then obtain x where  Hx1: "\<Gamma>,(x:\<tau>1);\<Delta> \<Turnstile> OK" and Hx2: "\<Gamma> \<sim> {x\<rhd>\<tau>1}" 
    using new_variable_abs t_abs by blast
  hence "\<Gamma>,(x:\<tau>1);\<Delta>;M;Z2 \<turnstile> freshen_tm x e : \<tau>2" using t_abs by auto
  hence Hz: "Z2 \<Turnstile> ScopeSet" using ty_wf_scope by auto
  obtain Zn where "Zn \<Turnstile> \<tau>2" and "Zn \<subseteq> Z2" using t_abs Hx1 Hx2 by auto
  hence "Z1\<union>Zn\<union>Z2 \<Turnstile> (\<tau>1 \<rightarrow>^X \<tau>2) \<and> Z1\<union>Zn\<union>Z' \<subseteq> Z1 \<union> Z2 \<union> Z'"
    using t_abs Hz ty_arrow[of Z1 \<tau>1 Zn \<tau>2 Z2 X] by blast
  thus ?case using t_abs by auto
next  
  case (t_app X \<Delta> Z1 Z2 M \<Gamma>1 \<Gamma>2 \<tau>1 \<tau>2 e1 e2 q)
  then obtain Z' where "Z' \<Turnstile> Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)" and H: "Z' \<subseteq> Z1" by auto
  then obtain Z'' where "Z'' \<Turnstile> \<tau>2" and "Z'' \<subseteq> Z'" using ty_type_arrow_elim[of Z' \<tau>1 X \<tau>2]
    by blast
  thus ?case using exI[of _ Z''] H by auto
next
  case (t_new \<Delta> Z M \<Gamma> \<tau> e)
  then obtain Z' where H: "Z' \<Turnstile> \<tau>" and Hsub: "Z' \<subseteq> Z" by auto
  thus ?case using exI[of _ "Z'"] ty_qual[of Z' "TyRef \<tau>" ScBot "{}"] 
    by(auto simp add: minimalValidScope_def)
next
  case (t_free \<Delta> Z M \<Gamma> \<tau> e)
  then obtain Z' where "Z' \<Turnstile> Qual qL ScBot (TyRef \<tau>)" and H: "Z' \<subseteq> Z" by auto
  hence "Z' \<Turnstile> TyRef \<tau>" using ty_type_qual[of Z' qL ScBot "TyRef \<tau>"] 
    by (auto simp add: minimalValidScope_def)
  thus ?case using exI[of _ Z''] H by auto
next
  case (t_deref \<pi> \<Delta> Z M \<Gamma> \<tau> e)
  then obtain Z' where "Z' \<Turnstile> Qual qU \<pi> (TyRef \<tau>)" and H: "Z' \<subseteq> Z" by auto
  then obtain Z'' where "Z'' \<Turnstile> \<tau>" and "Z'' \<subseteq> Z'" using ty_type_qual[of Z' qU \<pi> "TyRef \<tau>"] by blast
  thus ?case using exI[of _ Z''] H by auto
next  
  case t_assign thus ?case using ty_qual[of "{}" TyUnit ScBot "{}"]
    by (auto simp add: minimalValidScope_def)
next
  case (t_let \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z e e1 e2 \<tau> \<tau>1 \<tau>2) 
  have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
  moreover
  have H\<tau>: "\<Delta> \<Turnstile>' \<tau>" using t_let poly_ty_wf_type by auto 
  moreover
  have H1: "\<Gamma>1;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>1] by auto
  have "Z1\<union>Z2 \<Turnstile> ScopeSet" using t_let by (auto simp add: wf_scopeset_def)
  ultimately obtain \<rho> x
    where "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK" and "\<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>}" 
    and "\<rho> \<notin> set (free_let_dollar_scope e1)" and "\<rho> \<notin> Environ.dom (Z2 \<union> Z1)"
    using new_scope_let H1 H\<tau> by blast
  hence Htyp: "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M;Z1 \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1 \<or>
     \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1" 
      using t_let by (auto simp add: dom_def wf_scopeset_def)
  hence H\<tau>1: "\<Delta> \<Turnstile>' \<tau>1" using poly_ty_wf_type by auto
  have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
  then obtain x y where "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
    using new_variable_let H\<tau> H\<tau>1 by blast  
  hence "\<exists>Z'. Z' \<Turnstile> \<tau>2 \<and> Z' \<subseteq> Z2" using t_let by auto
  then obtain Zn where "Zn \<Turnstile> \<tau>2" and "Zn \<subseteq> Z2" by blast
  thus ?case by auto
next
  case (t_let' \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z \<rho> x e1 e2 \<tau> \<tau>1 \<tau>2)
  have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
  moreover
  have "\<Delta> \<Turnstile>' convT \<rho> \<tau>" using t_let' poly_ty_wf_type by auto 
  hence H\<tau>: "\<Delta> \<Turnstile>' \<tau>" by (cases \<tau>, auto)
  moreover
  have "\<Gamma>1;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>1] by auto
  hence H\<tau>1: "\<Delta> \<Turnstile>' \<tau>1" using t_let' poly_ty_wf_type by auto
  have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
  then obtain x y where "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
    using new_variable_let H\<tau> H\<tau>1 by blast
  then obtain Zn where "Zn \<Turnstile> \<tau>2" and "Zn \<subseteq> Z2" using t_let' by blast
  thus ?case by auto
next
  case (t_poly \<Delta> Z Z' M \<Gamma> \<tau> v q \<pi> X)   
  then obtain Zr where "Zr \<Turnstile> freshen_ty a \<tau>" and Hsub: "Zr \<subseteq> Z" by blast
  hence H\<tau>: "Zr \<Turnstile> \<tau>" using wf_subst_ty_wf_ty by auto
  have HXvar: "filterScVarX X = {}" using t_poly by auto
  have "filterScFreeX X \<Turnstile> ScopeSet" using t_poly ty_wf_scope by blast
  hence "Zr\<union>filterScFreeX X \<Turnstile> \<forall>.\<tau>^X" using H\<tau> HXvar by auto
  hence "Z'\<union>(Zr\<union>filterScFreeX X) \<Turnstile> Qual q \<pi> (\<forall>.\<tau>^X)" using t_poly by auto
  thus ?case using Hsub t_poly by auto
next
  case (t_ptyapp \<Delta> Z Zsub M \<Gamma> \<tau> \<phi> e1 q X)
  then obtain Zr where H1: "Zr \<Turnstile> Qual q ScBot (\<forall>.\<tau>^X)" and Hsub: "Zr \<subseteq> Z" by blast
  obtain Z' Z\<phi> where Hsub': "Zr = Z' \<union> Z\<phi>" and "minimalValidScope ScBot Z'" and "Z\<phi> \<Turnstile> \<forall>.\<tau>^X"
    using H1 ty_type_qual[of Zr q ScBot "\<forall>.\<tau>^X"] by rule
  then obtain Z1 where "Z1 \<Turnstile> \<tau>" and Hsub'': "Z1 \<subseteq> Z\<phi>" 
    using t_ptyapp by auto
  then obtain Zs where "Zs \<Turnstile> vsubst_ty \<phi> 0 \<tau>" and Hsub''': "Zs \<subseteq> Z1 \<union> Zsub" 
    using wf_ty_wf_subst_ty[of Z1 \<tau> Zsub \<phi> 0] t_ptyapp by blast
  thus ?case using Hsub''' Hsub'' Hsub' Hsub by blast
qed auto

lemma ty_ex_minZ:
  assumes "validScopeT \<tau> Z"
  shows "\<exists>Z'. minimalValidScopeT \<tau> Z' \<and> Z' \<subseteq> Z"
using prems proof (unfold minimalValidScopeT_def, cases \<tau>, simp)
  fix q scope \<phi>
  assume "\<tau> = Qual q scope \<phi>"
    and "validScope scope Z"
  thus "\<exists>Z'. (scope = ScBot \<and> Z' = {} \<or>
    (\<exists>\<rho>. scope = ScFree \<rho> \<and> Z' = {(\<rho>, EmptyScope)})) \<and> Z' \<subseteq> Z"
    proof (cases scope, auto simp add: dom_def)
      fix n b
      assume "(n, b) \<in> Z"
      thus "(n, EmptyScope) \<in> Z" by (cases b, auto)
    qed
qed

lemma wf_type_valid_scope:
  fixes \<tau> :: ty
  assumes "Z \<Turnstile> \<tau>"
  shows "validScopeT \<tau> Z"
using prems proof (cases \<tau>)
case Qual
  fix q \<pi> \<phi>
  assume "Z \<Turnstile> \<tau>"
    and Heq: "\<tau> = Qual q \<pi> \<phi>"
  then obtain Z\<phi> Z\<pi> where Hsub: "Z = Z\<pi> \<union> Z\<phi>" and "minimalValidScope \<pi> Z\<pi>" by blast
  hence "validScope \<pi> Z\<pi>" by (auto simp add: minimal_is_valid_scope minimalValidScope_def)
  hence "validScopeT \<tau> Z\<pi>" using Heq by auto
  thus "validScopeT \<tau> Z" using Hsub valid_scope_weaken[of \<tau> Z\<pi> Z\<phi>] by blast
qed
*)

lemma ty_mem_strengthen:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>"
  and "\<not> l free loc in e"
  shows "\<Gamma>;\<Delta>;M\<bar>l;Z \<turnstile> e : \<tau>"
proof -
  have "M;\<Delta> \<Turnstile> OK" using prems ty_wf_env_loc by auto
  hence "M\<bar>l;\<Delta> \<Turnstile> OK" using wf_env_subset[of M _ "M\<bar>l"]
    by (auto simp add: without_def)
  from prems this show ?thesis proof induct
    case (t_abs \<Delta> X Z M \<Gamma> \<tau>1 \<tau>2 e q \<pi> Lx)
    show ?case apply (rule Typing.t_abs[where L=Lx]) using t_abs proof auto
      fix x
      assume H1: "\<Gamma>,(x:\<tau>1);\<Delta> \<Turnstile> OK" and H2: "\<Gamma> \<sim> {x\<rhd>\<tau>1}" and H3: "x \<notin> Lx"
      hence "\<not> l free loc in (vsubst_tm (TmFreeVar x) 0 e)"
        using t_abs vsubst_tm_same_loc by auto
      thus "\<Gamma>,(x:\<tau>1);\<Delta>;M\<bar>l;{(\<rho>, EmptyScope) |\<rho>. ScFree \<rho> \<in> X} \<turnstile> vsubst_tm (TmFreeVar x) 0 e : \<tau>2" 
        using t_abs H1 H2 H3 by auto
    qed
  next
    case (t_let \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z e e1 e2 \<tau> \<tau>1 \<tau>2 Lx Ly)
    show ?case proof
      show "\<forall>x \<rho>. \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK \<and> \<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>} \<and> x \<notin> Lx \<and>
               \<not> \<rho> is free dollar sc in e1 \<and> \<rho> \<notin> Environ.dom (Z2 \<union> Z1) \<longrightarrow>
               (\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M\<bar>l;Z1 \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1 \<or>
               \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M\<bar>l;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1)" proof -
        {
          fix x \<rho>
          assume Hx1: "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK" and Hx2: "\<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>}" 
            and Hr1: "\<rho> \<notin> Environ.dom (Z2 \<union> Z1)"
            and Hr2: "\<not> \<rho> is free dollar sc in e1" 
            and Hlx: "x \<notin> Lx"
          have "l \<notin> set (free_loc e1)" using t_let by auto
          hence "l \<notin> set (free_loc (freshen_tm x (freshen_sc \<rho> e1)))"
            using vsubst_sc_same_loc vsubst_tm_same_loc by auto
          hence "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M\<bar>l;Z1 \<turnstile> 
            vsubst_tm (TmFreeVar x) 0 (vsubst_sc (ScFree \<rho>) 0 e1) : \<tau>1
          \<or> \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M\<bar>l;Z1,(\<rho>:\<section>) \<turnstile> 
            vsubst_tm (TmFreeVar x) 0 (vsubst_sc (ScFree \<rho>) 0 e1) : \<tau>1"
            using t_let Hx1 Hr1 Hx2 Hr2 Hlx by auto
          }
          thus ?thesis by auto
        qed
      next
        { 
          fix x y
          assume Hxy1: "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" 
            and Hxy2: "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)"
            and Hxy3: "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
            and Hlx: "x \<notin> Lx"
            and Hly: "y \<notin> Ly"
            and Hxy: "x \<noteq> y"
          have "l \<notin> set (free_loc e2)" using t_let by auto
          hence "l \<notin> set (free_loc (vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2)))" 
            using vsubst_sc_same_loc vsubst_tm_same_loc by auto
          hence Hy: "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta>;M\<bar>l;Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2) : \<tau>2" 
            using t_let Hxy1 Hxy2 Hxy3 Hlx Hly Hxy by auto
        }
        thus "\<forall>x y. \<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1) \<and> {x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1} \<and> x \<notin> Lx \<and> y \<notin> Ly \<and> x \<noteq> y
          \<longrightarrow> \<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta>;M\<bar>l;Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2) : \<tau>2"
        by auto
    qed (insert t_let, auto)
  next
    case (t_let' \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z \<rho> x e1 e2 \<tau> \<tau>1 \<tau>2 Zf Lx Ly)
    show ?case proof
      have "l \<notin> set (free_loc e1)" using t_let' by auto
      hence "l \<notin> set (free_loc (freshen_tm x (freshen_sc \<rho> e1)))" 
        using vsubst_sc_same_loc vsubst_tm_same_loc by auto
      thus "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M\<bar>l;Zf \<turnstile> e1 : \<tau>1"
        using t_let' by auto
    next
      {
        fix x y
        assume Hxy1: "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" 
          and Hxy2: "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)"
          and Hxy3: "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
          and Hlx: "x \<notin> Lx"
          and Hly: "y \<notin> Ly"
          and Hxy: "x \<noteq> y"
        have "l \<notin> set (free_loc e2)" using t_let' by auto
        hence "l \<notin> set (free_loc (vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2)))" 
          using vsubst_sc_same_loc vsubst_tm_same_loc by auto
        hence Hy: "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta>;M\<bar>l;Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2) : \<tau>2"
          using t_let' Hxy1 Hxy2 Hxy3 Hlx Hly Hxy by auto
      }
      thus "\<forall>x y. \<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1) \<and> {x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}  \<and> x \<notin> Lx \<and> y \<notin> Ly \<and> x \<noteq> y
        \<longrightarrow> \<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta>;M\<bar>l;Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2) : \<tau>2" by auto
    qed (insert t_let', auto)
  next
    case (t_poly \<Delta> Z X M \<Gamma> \<tau> v q \<pi> La)  
    {
      fix a
      assume Hla: "a \<notin> La"
      have "\<not> l free loc in v" using t_poly by auto
      hence "\<not> l free loc in (freshen_tmty a v)" using poly_free_loc_tmty_equal by auto
      moreover
      have "M\<bar>l;\<Delta>,(a:\<star>) \<Turnstile> OK" using t_poly poly_env_ty_cons by auto
      ultimately
      have "\<Gamma>;\<Delta>,(a:\<star>);M\<bar>l;(filterScFreeX X) \<turnstile> freshen_tmty a v : freshen_ty a \<tau>" 
        using t_poly.hyps Hla by blast
    }
    thus ?case using t_poly by auto
  next
    case (t_app \<Delta> X Z1 Z2 Z M \<Gamma>1 \<Gamma>2 \<tau>1 \<tau>2 e1 e2 q) 
    show ?case apply rule using t_app by auto
  next
    case t_ptyapp
    show ?case apply rule using t_ptyapp by auto
  qed (auto simp add: without_def)
qed

lemma ty_mem_strengthen_all:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>"
  and "\<forall>l. \<not> l free loc in e"
  shows "\<Gamma>;\<Delta>;{};Z \<turnstile> e : \<tau>"
proof -
  have Hmok: "M;\<Delta> \<Turnstile> OK" using prems ty_wf_env_loc by auto
  hence "finite M" by (auto simp add: wf_env_def)
  thus ?thesis using Hmok prems proof induct
    case (insert p M)
    then obtain l b where "p = (l, b)" by (cases p)
    then obtain \<tau>' where Hp: "p = (l\<rhd>\<tau>')" by (cases b, auto)
    have H: "M;\<Delta> \<Turnstile> OK" using insert.prems wf_env_subset[of "insert p M" _ M]
      by auto
    have "\<Gamma>;\<Delta>;(insert p M)\<bar>l;Z \<turnstile> e : \<tau>" using prems
      by (auto simp add: ty_mem_strengthen)
    also have "(insert p M)\<bar>l = M\<bar>l" using Hp by (auto simp add: without_def)
    also have "\<not> M defines l" proof
      assume "M defines l"
      hence "p \<in> M"
        using Hp insert.prems wf_env_subset_defines[of "insert p M" _ "M" l \<tau>']
        by auto
      thus False by contradiction
    qed
    hence "M\<bar>l = M" by (auto simp add: dom_def without_def)
    finally show ?case using insert H by auto
  qed auto
qed

lemma ty_no_loc:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>"
  and "\<not> M defines l"
  shows "\<not> l free loc in e"
proof -
have H: "Z \<Turnstile> ScopeSet" using prems ty_wf_scope by auto
show ?thesis using prems H proof induct
  case (t_abs \<Delta> X Z M \<Gamma> \<tau>1 \<tau>2 e q \<pi> Lx)
  have Hgok: "\<Gamma>;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<tau>1" and "finite Lx" by assumption
  then obtain x where  "\<Gamma>,(x:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma> \<sim> {x\<rhd>\<tau>1}" and "x \<notin> Lx"
    using new_variable_abs by blast
  hence "\<not> l free loc in (freshen_tm x e)" using t_abs ty_wf_scope by auto
  hence "\<not> l free loc in e" using vsubst_tm_same_loc by auto
  thus ?case by auto
next
  case (t_let \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z e e1 e2 \<tau> \<tau>1 \<tau>2 Lx Ly)
  thus ?case apply simp proof -
    have Hlx: "finite Lx" and Hly: "finite Ly" by assumption
    have "\<Gamma>1;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>1] by auto
    moreover
    have "Z1\<union>Z2 \<Turnstile> ScopeSet" using t_let by (auto simp add: wf_scopeset_def)
    moreover
    have H\<tau>: "\<Delta> \<Turnstile>' \<tau>" using poly_ty_wf_type t_let by auto
    ultimately
    obtain \<rho> x
      where Hc1: "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK" and Hc2: "\<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>}" 
      and Hc3: "\<rho> \<notin> set (free_let_dollar_scope e1)" and Hc4: "\<rho> \<notin> Environ.dom (Z2 \<union> Z1)"
      and "x \<notin> Lx"
      using new_scope_let[where Lx=Lx] Hlx by blast
    hence "l \<notin> set (free_loc (freshen_tm x (freshen_sc \<rho> e1)))" 
      and Htyp: "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M;Z1 \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1 \<or>
     \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1" 
      using t_let by (auto simp add: dom_def wf_scopeset_def)
    hence "l \<notin> set (free_loc e1)" 
      using vsubst_sc_same_loc vsubst_tm_same_loc by auto
    moreover
    have H\<tau>1: "\<Delta> \<Turnstile>' \<tau>1" using Htyp poly_ty_wf_type by auto
    have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
    then obtain x y where "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
      and "x \<notin> Lx" and "y \<notin> Ly" and "x \<noteq> y"
      using new_variable_let[where Lx=Lx and Ly=Ly] Hlx Hly H\<tau> H\<tau>1 by blast
    hence "l \<notin> set (free_loc (vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2)))" using t_let 
      by (auto simp add: wf_scopeset_def dom_def)
    hence "l \<notin> set (free_loc e2)" 
      using vsubst_sc_same_loc vsubst_tm_same_loc by auto
    ultimately
    show "l \<notin> set (free_loc e) \<and> l \<notin> set (free_loc e1) \<and> l \<notin> set (free_loc e2)"
      using t_let by (auto simp add: wf_scopeset_def)
  qed
next
  case (t_let' \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z \<rho> x e1 e2 \<tau> \<tau>1 \<tau>2 _ Lx Ly)
  thus ?case apply simp proof -
    have Hlx: "finite Lx" and Hly: "finite Ly" by assumption
    have "l \<notin> set (free_loc e1)"
      using t_let' by (auto simp add: wf_scopeset_def)
    hence "l \<notin> set (free_loc e1)" 
      using vsubst_sc_same_loc vsubst_tm_same_loc by auto
    moreover
    have "\<Delta> \<Turnstile>' convT \<rho> \<tau>" using t_let' poly_ty_wf_type by (cases \<tau>, auto)
    hence H\<tau>: "\<Delta> \<Turnstile>' \<tau>" by (cases \<tau>, auto)
    have H\<tau>1: "\<Delta> \<Turnstile>' \<tau>1" using t_let' poly_ty_wf_type by auto
    have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
    then obtain x y where "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
      and "x \<notin> Lx" and "y \<notin> Ly" and "x \<noteq> y"
      using new_variable_let[where Lx=Lx and Ly=Ly] Hlx Hly H\<tau> H\<tau>1 by blast
    hence "l \<notin> set (free_loc (vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2)))" 
      using t_let' by (auto simp add: wf_scopeset_def)
    hence "l \<notin> set (free_loc e2)" 
      using vsubst_sc_same_loc vsubst_tm_same_loc by auto
    ultimately
    show "l \<notin> set (free_loc e1) \<and> l \<notin> set (free_loc e2)" by auto
  qed
next
  case (t_poly \<Delta> Z X M \<Gamma> \<tau> v q \<pi> La)
  then obtain a where Hla: "a \<notin> La" using fresh_exists by auto
  hence "filterScFreeX X \<Turnstile> ScopeSet" using t_poly ty_wf_scope by blast
  hence "\<not> l free loc in (freshen_tmty a v)" using Hla t_poly by (auto simp add: wf_scopeset_def)
  hence "\<not> l free loc in v" using poly_free_loc_tmty_equal by auto
  thus ?case by auto 
qed (auto simp add: dom_def wf_scopeset_def)
qed

lemma ty_no_locs:
  assumes "\<Gamma>;\<Delta>;{};Z \<turnstile> e : \<tau>"
  shows "\<forall>l. \<not> l free loc in e"
using prems by (auto simp add: ty_no_loc dom_def)

lemma free_ty_in_env:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>"
  and "x free in e"
  shows "\<Gamma> defines x"
proof -
have H: "Z \<Turnstile> ScopeSet" using prems ty_wf_scope by auto
show ?thesis using prems H proof induct
  case (t_abs \<Delta> X Z M \<Gamma> \<tau>1 \<tau>2 eb q \<pi> Lx)
  have Hgok: "\<Gamma>;\<Delta> \<Turnstile> OK" by assumption
  have Hfin: "finite (Lx\<union>{x})" using t_abs by auto
  have "\<Delta> \<Turnstile>' \<tau>1" by assumption
  then obtain z where H1: "\<Gamma>,(z:\<tau>1);\<Delta> \<Turnstile> OK" and H2: "\<Gamma> \<sim> {z\<rhd>\<tau>1}" and Hneq: "z \<noteq> x" and Hlx: "z \<notin> Lx"
    using new_variable_abs[where Lx="Lx\<union>{x}"] Hfin Hgok by blast
  have "x free in freshen_tm z eb" using t_abs by (auto simp add: free_subst)
  thus ?case using t_abs H H1 H2 Hneq Hlx using ty_wf_scope by auto
next
  case (t_let \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z e e1 e2 \<tau> \<tau>1 \<tau>2 Lx Ly)
  have Hlx: "finite Lx" and Hly: "finite Ly" by assumption
  have "\<Gamma>1;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>1] by auto
  moreover
  have "Z1\<union>Z2 \<Turnstile> ScopeSet" using t_let by (auto simp add: wf_scopeset_def)
  moreover
  have H\<tau>: "\<Delta> \<Turnstile>' \<tau>" using poly_ty_wf_type t_let by auto
  ultimately
  obtain \<rho> xf
    where Hc1: "\<Gamma>1,(xf:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK" and Hc2: "\<Gamma>1 \<sim> {xf\<rhd>convT \<rho> \<tau>}" 
    and Hc3: "\<rho> \<notin> set (free_let_dollar_scope e1)" and Hc4: "\<rho> \<notin> Environ.dom (Z2 \<union> Z1)"
    and "xf \<notin> Lx" and Hneq: "xf \<noteq> x"
    using new_scope_let[where Lx="Lx\<union>{x}"] Hlx by blast
  moreover
  have "\<forall>xa \<rho>.
     \<Gamma>1,(xa:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK \<and>
     \<Gamma>1 \<sim> {xa\<rhd>convT \<rho> \<tau>} \<and> xa \<notin> Lx \<and>
     (\<rho> \<notin> set (free_let_dollar_scope e1)) \<and> \<rho> \<notin> Environ.dom (Z2 \<union> Z1) \<longrightarrow>
     ((\<Gamma>1,(xa:convT \<rho> \<tau>);\<Delta>;M;Z1 \<turnstile> freshen_tm xa (freshen_sc \<rho> e1) : \<tau>1 \<and>
     (x free in freshen_tm xa (freshen_sc \<rho> e1) \<longrightarrow>
      Z1 \<Turnstile> ScopeSet \<longrightarrow> x \<in> Environ.dom (\<Gamma>1,(xa:convT \<rho> \<tau>)))) \<or>
     (\<Gamma>1,(xa:convT \<rho> \<tau>);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm xa (freshen_sc \<rho> e1) : \<tau>1 \<and>
     (x free in freshen_tm xa (freshen_sc \<rho> e1) \<longrightarrow>
      Z1,(\<rho>:\<section>) \<Turnstile> ScopeSet \<longrightarrow> x \<in> Environ.dom (\<Gamma>1,(xa:convT \<rho> \<tau>)))))" 
    by assumption
  ultimately have H2: "x free in freshen_tm xf (freshen_sc \<rho> e1) \<longrightarrow> Z1 \<Turnstile> ScopeSet \<longrightarrow>
             x \<in> Environ.dom (\<Gamma>1,(xf:convT \<rho> \<tau>))
    \<or> x free in freshen_tm xf (freshen_sc \<rho> e1) \<longrightarrow> Z1,(\<rho>:\<section>) \<Turnstile> ScopeSet \<longrightarrow>
             x \<in> Environ.dom (\<Gamma>1,(xf:convT \<rho> \<tau>))" 
    and Htyp: "\<Gamma>1,(xf:convT \<rho> \<tau>);\<Delta>;M;Z1 \<turnstile> freshen_tm xf (freshen_sc \<rho> e1) : \<tau>1 \<or>
    \<Gamma>1,(xf:convT \<rho> \<tau>);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm xf (freshen_sc \<rho> e1) : \<tau>1" 
    by auto
  have H\<tau>1: "\<Delta> \<Turnstile>' \<tau>1" using Htyp poly_ty_wf_type by auto
  have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
  then obtain x' y where "\<Gamma>2,(x':\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x'\<rhd>\<tau>},(y:\<tau>1)" and "{x'\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
      and "x' \<notin> Lx" and "y \<notin> Ly" and Hneqx: "x' \<noteq> x" and Hneqy: "y \<noteq> x" and Hxy: "x' \<noteq> y"
    using new_variable_let[where Lx="Lx\<union>{x}" and Ly="Ly\<union>{x}"] Hlx Hly H\<tau> H\<tau>1 by blast
  hence H1: "x free in (vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x' e2)) \<longrightarrow> 
    Z2 \<Turnstile> ScopeSet \<longrightarrow> (\<Gamma>2,(x':\<tau>),(y:\<tau>1)) defines x"
    using t_let by blast
  show ?case proof (cases "x free in e2")
    assume Htrue: "x free in e2"
    have "Z2 \<Turnstile> ScopeSet" using t_let by (auto simp add: wf_scopeset_def)
    hence Hin: "\<Gamma>2,(x':\<tau>),(y:\<tau>1) defines x" using H1 free_subst Htrue by auto
    have "x \<noteq> x'"  and "x \<noteq> y" using Htrue Hneqx Hneqy by auto
    thus ?case using Hin by (auto simp add: dom_def)
  next
    assume "\<not> x free in e2"
    hence Hfalse: "x free in e1 \<or> x free in e" using t_let by auto
    have Hwff1: "Z1 \<Turnstile> ScopeSet" by assumption
    hence Hwff2: "Z1,(\<rho>:\<section>) \<Turnstile> ScopeSet" by (auto simp add: wf_scopeset_def)
    show ?thesis using Hfalse proof
      assume Hin1: "x free in e1"
      hence "x free in freshen_tm xf (freshen_sc \<rho> e1)" 
        using free_subst free_subst_sc by auto
      hence Hin: "x \<in> Environ.dom (\<Gamma>1,(xf:convT \<rho> \<tau>))" using H2 Hwff1 Hwff2 by auto
      have "x \<noteq> xf" using Hneq by auto
      thus ?case using Hin by (auto simp add: dom_def)
    next
      assume Hin2: "x free in e"
      thus ?case using t_let by (auto simp add: dom_def wf_scopeset_def)
    qed
  qed
next
  case (t_let' \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z \<rho> x' e1 e2 \<tau> \<tau>1 \<tau>2 _ Lx Ly)
  have Hlx: "finite Lx" and Hly: "finite Ly" by assumption
  have "\<Delta> \<Turnstile>' convT \<rho> \<tau>" using t_let' poly_ty_wf_type by (cases \<tau>, auto)
  hence H\<tau>: "\<Delta> \<Turnstile>' \<tau>" by (cases \<tau>, auto)
  have H\<tau>1: "\<Delta> \<Turnstile>' \<tau>1" using t_let' poly_ty_wf_type by auto
  have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
  then obtain x_n y where "\<Gamma>2,(x_n:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x_n\<rhd>\<tau>},(y:\<tau>1)"
    and "{x_n\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}" and "x_n \<notin> Lx" and "y \<notin> Ly" and Hneqx: "x_n \<noteq> x" and Hneqy: "y \<noteq> x"
    and Hxy: "x_n \<noteq> y"
    using new_variable_let[where Lx="Lx\<union>{x}" and Ly="Ly\<union>{x}"] Hlx Hly H\<tau> H\<tau>1 by blast
  hence H1: "x free in (vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x_n e2)) \<longrightarrow> 
             Z2 \<Turnstile> ScopeSet \<longrightarrow>(\<Gamma>2,(x_n:\<tau>),(y:\<tau>1)) defines x"
    and  H2: "x free in e1 \<longrightarrow> x \<in> Environ.dom (\<Gamma>1,(x':convT \<rho> \<tau>))"
    using t_let' by (auto simp add: wf_scopeset_def)
  have H3: "\<Gamma> defines x'" using t_let' by (auto simp add: dom_def)
  show ?case proof (cases "x=x'")
    case True thus ?thesis using H3 by (auto simp add: dom_def)
  next
    case False
    show ?thesis proof (cases "x free in e2")
      assume Htrue: "x free in e2"
      hence "x free in (vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x_n e2))"
        using free_subst by auto 
      hence Hin: "\<Gamma>2,(x_n:\<tau>),(y:\<tau>1) defines x" 
        using H1 t_let' by (auto simp add: wf_scopeset_def)
      have "x \<noteq> y" using Htrue Hneqy by auto
      thus ?thesis using Hin False Hneqx by (auto simp add: dom_def)
    next
      assume "\<not> x free in e2"
      hence Hfalse: "x free in e1 \<or> x free in (TmFreeVar x')" using t_let' by auto
      thus ?thesis proof
        assume Hin1: "x free in e1"
        hence Hin: "x \<in> Environ.dom (\<Gamma>1,(x':convT \<rho> \<tau>))" using H2 by auto
        thus ?thesis using Hin False by (auto simp add: dom_def)
      next
        assume Hin2: "x free in (TmFreeVar x')"
        thus ?thesis using False by (auto simp add: dom_def)
      qed
    qed
  qed
next
  case (t_poly \<Delta> Z Z' M \<Gamma> \<tau> v q \<pi> La) 
  then obtain a where Hla: "a \<notin> La" using fresh_exists by auto
  hence "x free in v" using t_poly by auto 
  hence "x free in freshen_tmty a v" using poly_free_var_tmty_equal by auto
  thus ?case using Hla t_poly using ty_wf_scope by auto
qed (auto simp add: dom_def wf_scopeset_def)
qed

lemma not_in_env_ty_same_subst:
  assumes "\<not> \<Gamma> defines x"
  and "\<Gamma>;\<Delta>;M;Z \<turnstile> vsubst_tm (TmFreeVar x) i e : \<tau>"
  shows "\<Gamma>;\<Delta>;M;Z \<turnstile> vsubst_tm e_s i e : \<tau>"
proof -
  have "\<not> x free in vsubst_tm (TmFreeVar x) i e" using prems
    by (auto simp add: free_ty_in_env)
  hence "\<forall>e_s. vsubst_tm e_s i e = e" using not_free_subst_same_subst by fast
  thus ?thesis using prems by auto
qed

lemma wf_env_defines_without:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>"
  shows "\<Gamma> = (\<Gamma>\<bar>x),(x:\<tau>)"
using prems proof (auto simp add: without_def)
  fix y b
  assume "(y\<rhd>\<tau>) \<in> \<Gamma>" and "(y, b) \<in> \<Gamma>" and "b \<noteq> Term \<tau>"
  then obtain \<tau>' where "b = Term \<tau>'" by (cases b)
  hence "\<tau> = \<tau>'" using prems by (auto simp add: wf_env_def)
  thus False using prems by auto
qed

lemma valid_scope_subset:
  assumes "validScope \<pi> Z'"
  and "Z' \<subseteq> Z"
  shows "validScope \<pi> Z"
proof -
  have "validScope \<pi> (Z'\<union>Z)" using valid_scope_weaken prems by auto
  also have "Z'\<union>Z = Z" using prems by auto
  finally show ?thesis by auto
qed

lemma weakening_m:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>"
  and "M \<union> M';\<Delta> \<Turnstile> OK"
  shows "\<Gamma>;\<Delta>;M \<union> M';Z \<turnstile> e : \<tau>"
using prems proof induct
  case (t_abs \<Delta> X Z M \<Gamma> \<tau>1 \<tau>2 eb q \<pi> Lx)
  show ?case proof
    show "\<forall>x. \<Gamma>,(x:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma> \<sim> {x\<rhd>\<tau>1} \<and> x \<notin> Lx \<longrightarrow>
             \<Gamma>,(x:\<tau>1);\<Delta>;M\<union>M';filterScFreeX X \<turnstile> freshen_tm x eb : \<tau>2"
    proof (auto simp del: filterScFreeX freshen_tm)
      fix x
      assume "\<Gamma>,(x:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma> \<sim> {x\<rhd>\<tau>1}" and "x \<notin> Lx"
      thus "\<Gamma>,(x:\<tau>1);\<Delta>;M\<union>M';filterScFreeX X \<turnstile> freshen_tm x eb : \<tau>2"
        using t_abs by auto
    qed
  qed (insert t_abs, blast)
next
  case (t_let' \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z \<rho> x' e1 e2 \<tau> \<tau>1 \<tau>2 Zf Lx Ly)
  have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
  show ?case proof 
    show "\<forall>x y. \<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1) \<and> {x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1} \<and> x \<notin> Lx \<and> y \<notin> Ly \<and> x \<noteq> y \<longrightarrow>
           \<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta>;M\<union>M';Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2) : \<tau>2"
    proof auto
      fix x y
      assume "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
      and "x \<notin> Lx" and "y \<notin> Ly" and "x \<noteq> y"
      thus "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta>;M\<union>M';Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0)
                                       (vsubst_tm (TmFreeVar x) 0 e2) : \<tau>2" using t_let' by auto
    qed
  qed (insert t_let', blast)
next
  case (t_let \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z e e1 e2 \<tau> \<tau>1 \<tau>2 Lx Ly)
  have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
  show ?case proof 
    show "\<forall>x y. \<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1) \<and> {x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1} \<and> x \<notin> Lx \<and> y \<notin> Ly \<and> x \<noteq> y \<longrightarrow>
           \<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta>;M\<union>M';Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2) : \<tau>2"
    proof auto
      fix x y
      assume "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
        and "x \<notin> Lx" and "y \<notin> Ly" and "x \<noteq> y"
      thus "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta>;M\<union>M';Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0)
        (vsubst_tm (TmFreeVar x) 0 e2) : \<tau>2" using t_let by auto
    qed
  next
    show "\<forall>x \<rho>. \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK \<and>
           \<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>} \<and> x \<notin> Lx \<and>
           \<rho> \<notin> set (free_let_dollar_scope e1) \<and> \<rho> \<notin> Environ.dom (Z2 \<union> Z1) \<longrightarrow>
           \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M \<union> M';Z1 \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1 \<or>
           \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M \<union> M';Z1,(\<rho>:\<section>) \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1"
    proof auto
      fix x \<rho>
      assume "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK" and "\<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>}"
        and "\<rho> \<notin> set (free_let_dollar_scope e1)" and "\<rho> \<notin> Environ.dom (Z2 \<union> Z1)"
        and Hfalse: "\<not> \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M \<union> M';Z1,(\<rho>:\<section>) \<turnstile> 
             vsubst_tm (TmFreeVar x) 0 (vsubst_sc (ScFree \<rho>) 0 e1) : \<tau>1"
        and "x \<notin> Lx"
      hence 
        "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M \<union> M';Z1 \<turnstile> vsubst_tm (TmFreeVar x) 0 (vsubst_sc (ScFree \<rho>) 0 e1) : \<tau>1 \<or>
        \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M \<union> M';Z1,(\<rho>:\<section>) \<turnstile> vsubst_tm (TmFreeVar x) 0 (vsubst_sc (ScFree \<rho>) 0 e1) : \<tau>1"
        using t_let by auto
      thus "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M \<union> M';Z1 \<turnstile> vsubst_tm (TmFreeVar x) 0 (vsubst_sc (ScFree \<rho>) 0 e1) : \<tau>1"
        using Hfalse by auto
    qed
  qed (insert t_let, blast)
next
  case (t_poly \<Delta> Z X M \<Gamma> \<tau> v q \<pi> La) 
  { fix a
    assume Hla: "a \<notin> La"
    have "M\<union>M';\<Delta>,(a:\<star>) \<Turnstile> OK" using poly_env_ty_cons t_poly by auto
    hence "\<Gamma>;\<Delta>,(a:\<star>);M\<union>M';filterScFreeX X \<turnstile> freshen_tmty a v : freshen_ty a \<tau>" using t_poly Hla by auto
  }
  thus ?case using t_poly by auto 
next
  case t_app show ?case apply rule using t_app by auto
next
  case t_ptyapp show ?case apply rule using t_ptyapp by auto
qed auto

(*
lemma wf_type_conv:
  fixes \<tau> :: ty
  assumes "Z \<Turnstile> convT \<rho> \<tau>"
  and "isBangable \<tau>"
  shows "\<exists>Zn. Zn \<Turnstile> \<tau> \<and> Zn \<subseteq> Z"
using prems proof (cases \<tau>)
  fix q \<pi> \<phi>
  assume Heq: "\<tau> = Qual q \<pi> \<phi>"
  hence "isBangableP \<phi>" using assms by auto
  then obtain \<tau>r where Heq2: "\<phi> = TyRef \<tau>r" by (cases \<phi>, auto)
  hence Heq1: "\<tau> = Qual qL ScBot (TyRef \<tau>r)" using Heq prems by auto
  have "Z \<Turnstile> Qual qU (ScFree \<rho>) (TyRef \<tau>r)" using prems Heq1 Heq2 by auto
  then obtain Z1 Z2 where Hz1: "Z1 \<Turnstile> \<tau>r" and "minimalValidScope (ScFree \<rho>) Z2" 
    and Hsub: "Z = Z1 \<union> Z2" by blast
  have "{}\<union>Z1 \<Turnstile> Qual qL ScBot (TyRef \<tau>r)" proof
    show "Z1 \<Turnstile> TyRef \<tau>r" using Hz1 by auto
  next
    show "minimalValidScope ScBot {}" by (auto simp add: minimalValidScope_def)
  qed
  thus ?thesis using Heq1 Heq2 Hsub by blast
qed
*)

lemma ty_valid_scope:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> e : \<tau>"
  shows "validScope (scopeT \<tau>) Z"
using prems proof induct
  case (t_app \<Delta> X Z1 Z2 Z M \<Gamma>1 \<Gamma>2 \<tau>1 \<tau>2 e1 e2 q) 
  hence "validScope (scopeT \<tau>2) Z" by (auto simp add: minimal_is_valid)
  thus ?case using valid_scope_weaken[of _ Z "Z1 \<union> Z2 \<union> filterScFreeX X"] by (auto simp add: Un_ac)
next
  case (t_let \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z e e1 e2 \<tau> \<tau>1 \<tau>2 Lx Ly) 
  have Hlx: "finite Lx" and Hly: "finite Ly" by assumption
  hence HG1: "\<Delta> \<Turnstile>' \<tau>" using t_let poly_ty_wf_type by auto
  have "\<Gamma>1;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>1] by auto
  moreover
  have "Z1\<union>Z2 \<Turnstile> ScopeSet" using t_let by (auto simp add: wf_scopeset_def)
  ultimately
  obtain \<rho> x
    where "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK" and "\<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>}" 
    and "\<rho> \<notin> set (free_let_dollar_scope e1)" and "\<rho> \<notin> Environ.dom (Z2 \<union> Z1)"
    and "x \<notin> Lx"
    using new_scope_let HG1 Hlx by blast
  hence "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M;Z1 \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1 \<or>
         \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1"
    using t_let by auto
  hence HG2: "\<Delta> \<Turnstile>' \<tau>1" using poly_ty_wf_type by blast
  have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
  then obtain x y where "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
    and "x \<notin> Lx" and "y \<notin> Ly" and "x \<noteq> y"
    using new_variable_let[where Lx=Lx and Ly=Ly] HG1 HG2 Hlx Hly by blast
  hence "validScope (scopeT \<tau>2) Z2" using t_let by auto
  thus ?case using valid_scope_weaken by auto
next
  case (t_let' \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z \<rho> x e1 e2 \<tau> \<tau>1 \<tau>2 Zf Lx Ly)
  have Hlx: "finite Lx" and Hly: "finite Ly" by assumption
  have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
  moreover
  have "\<Delta> \<Turnstile>' convT \<rho> \<tau>" using t_let' poly_ty_wf_type by auto 
  hence H\<tau>: "\<Delta> \<Turnstile>' \<tau>" by (cases \<tau>, auto)
  moreover
  have "\<Gamma>1;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>1] by auto
  hence H\<tau>1: "\<Delta> \<Turnstile>' \<tau>1" using t_let' poly_ty_wf_type by auto
  have "\<Gamma>2;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] by auto
  then obtain x y where "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
    and "x \<notin> Lx" and "y \<notin> Ly" and "x \<noteq> y"
    using new_variable_let[where Lx=Lx and Ly=Ly] H\<tau> H\<tau>1 Hlx Hly by blast
  hence "validScope (scopeT \<tau>2) Z2" using t_let' by auto
  thus ?case using valid_scope_weaken by auto
next
  case (t_swap \<Delta> Z1 Z2 M \<Gamma>1 \<Gamma>2 \<tau> e1 e2 \<pi>)
  hence "validScope (scopeT \<tau>) Z2" by auto
  thus ?case using valid_scope_subset[of _ Z2 "Z1\<union>Z2"] by (auto simp add: Un_ac)
next
  case (t_assign \<Delta> Z1 Z2 M \<Gamma>1 \<Gamma>2 \<tau> e1 e2 \<pi>)
  hence "validScope (scopeT \<tau>) Z2" by auto
  thus ?case using valid_scope_subset[of _ Z2 "Z1\<union>Z2"] by (auto simp add: Un_ac)
next
  case t_abs thus ?case by (auto simp add: valid_scope_weaken minimal_is_valid Un_ac)
next
  case t_free thus ?case by (auto simp add: valid_scope_weaken minimal_is_valid Un_ac)
next
  case t_new thus ?case by (auto simp add: valid_scope_weaken minimal_is_valid Un_ac)
next
  case t_deref thus ?case by (auto simp add: valid_scope_weaken minimal_is_valid Un_ac)
next
  case t_unit thus ?case by (auto simp add: valid_scope_weaken minimal_is_valid Un_ac)
next
  case t_var thus ?case by (auto simp add: valid_scope_weaken minimal_is_valid Un_ac)
next
  case t_loc thus ?case by (auto simp add: valid_scope_weaken minimal_is_valid Un_ac)
next
  case t_poly thus ?case by (auto simp add: valid_scope_weaken minimal_is_valid Un_ac)
next
  case (t_ptyapp \<Delta> Z Z\<tau> M \<Gamma> \<tau> \<phi> e q X) 
  have "scopeT (vsubst_ty \<phi> 0 \<tau>) = scopeT \<tau>" by (cases \<tau>, auto)
  hence "validScope (scopeT (vsubst_ty \<phi> 0 \<tau>)) Z\<tau>" 
    using t_ptyapp by (auto simp add: minimal_is_valid)
  thus ?case using valid_scope_weaken[of _ Z\<tau> "Z\<union>filterScFreeX X"] by (auto simp add: Un_ac)
qed

end
