theory TypingLemmas
imports Typing Semantics

begin

lemma ty_qual_sanity:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> v : \<tau>"
  and "value v"
  shows "qualV v = qualT \<tau>"
using prems by (cases, auto)

lemma ty_scope_sanity:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> v : \<tau>"
  and "value v"
  shows "scopeV v = scopeT \<tau>"
using prems by (cases, auto)

lemma ty_conv_ty:
  assumes "\<Gamma>v;\<Delta>;M;Z \<turnstile> v : \<tau>"
  and "value v"
  and "isBangable \<tau>"
  shows "\<Gamma>v;\<Delta>;M;{},(\<rho>:\<section>) \<turnstile> convV \<rho> v : convT \<rho> \<tau>"
using prems proof induct
  case t_loc thus ?case by (auto simp add: minimalValidScope_def)
next
  case t_unit thus ?case by (auto simp add: minimalValidScope_def)
next
  case (t_abs \<Delta> X Z M \<Gamma> \<tau>1 \<tau>2 e q \<pi> Lx) show ?case apply simp apply (rule Typing.t_abs[where L=Lx])
    using t_abs by (auto simp add: minimalValidScope_def)
qed auto

lemma ty_conv_ty_rev:
  assumes "\<Gamma>v;\<Delta>;M;Z \<turnstile> v : convT \<rho> \<tau>"
  and "value v"
  and Hbang: "isBangable \<tau>"
  and "scopeV v = ScFree \<rho>"
  and "qualV v = qU"
  shows "\<exists>Zn. \<Gamma>v;\<Delta>;M;Zn \<turnstile> convRevV v : \<tau> \<and> Zn \<subseteq> Z"
using prems proof (cases v)
  case (TmLoc q \<pi> l)
  have Ht: "\<Gamma>v;\<Delta>;M;Z \<turnstile> TmLoc qU (ScFree \<rho>) l : convT \<rho> \<tau>" using prems by auto
  obtain r where Heq: "\<tau> = Qual qL ScBot r" using Hbang by (cases \<tau>, auto)
  then obtain \<tau>r where Heq: "\<tau> = Qual qL ScBot (TyRef \<tau>r)" using Ht by (cases \<tau>, auto)
  hence "\<Gamma>v;\<Delta>;M;{} \<turnstile> TmLoc qL ScBot l : Qual qL ScBot (TyRef \<tau>r)"
    using Ht by (auto simp add: minimalValidScope_def)
  thus ?thesis using TmLoc Heq exI[of _ "{}"] by auto
next
  case (TmUnit q \<pi>)
  have Ht: "\<Gamma>v;\<Delta>;M;Z \<turnstile> TmUnit qU (ScFree \<rho>) : convT \<rho> \<tau>" using prems by auto
  hence Heq: "\<tau> = Qual qL ScBot TyUnit" using prems by (cases \<tau>, auto)
  hence "\<Gamma>v;\<Delta>;M;{} \<turnstile> TmUnit qL ScBot : Qual qL ScBot TyUnit"
    using Ht by (auto simp add: minimalValidScope_def)
  thus ?thesis using TmUnit Heq by auto
next
  case (TmAbs q \<pi> \<tau>arg e)
  have Ht: "\<Gamma>v;\<Delta>;M;Z \<turnstile> \<lambda>^qU^(ScFree \<rho>)[\<tau>arg].e : convT \<rho> \<tau>" using prems by auto
  then obtain \<tau>f where Heq: "\<tau> = Qual qL ScBot \<tau>f" using prems by (cases \<tau>, auto)
  hence "\<Gamma>v;\<Delta>;M;Z \<turnstile> \<lambda>^qU^(ScFree \<rho>)[\<tau>arg].e : Qual qU (ScFree \<rho>) \<tau>f"
    using Ht by auto
  then obtain \<tau>p Xp where "\<tau>f = \<tau>arg\<rightarrow>^Xp \<tau>p" by (cases \<tau>f, auto)
  hence "\<tau> = Qual qL ScBot (\<tau>arg\<rightarrow>^Xp \<tau>p)" using Heq by auto
  hence False using prems by auto
  thus ?thesis by auto
next
  case (TmProd q \<pi> e)
  obtain \<phi> q \<pi> where "\<tau> = Qual q \<pi> \<phi>" by (cases \<tau>, auto)
  thus ?thesis using TmProd assms by (cases \<phi>, auto)
qed auto

lemma ty_bang_unrestricted_env:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> v : convT \<rho> \<tau>"
  and "value v"
  and "qualV v = qU"
  and "isBangable \<tau>"
  shows "qualG \<Gamma> \<preceq> qU"
using prems proof (cases v)
  case (TmProd q \<pi> e)
  obtain \<phi> q \<pi> where "\<tau> = Qual q \<pi> \<phi>" by (cases \<tau>, auto)
  thus ?thesis using TmProd assms by (cases \<phi>, auto)
qed auto

(*
lemma free_scope_store_in_Z:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> v : \<tau>"
  and "value v"
  shows "free_scope_store {z\<mapsto>v} \<subseteq> dom Z"
using prems proof (cases "scopeT \<tau>")
  case (ScFree n)
  have "ScFree n = scopeT \<tau>" using prems by auto
  hence "scopeV v = ScFree n" using ty_scope_sanity prems by auto
  hence "free_scope_store {z\<mapsto>v} = {n}" by (auto simp add: free_scope_store_def)
  moreover
  obtain Z' where "Z' \<Turnstile> \<tau>" and H: "Z' \<subseteq> Z" using ty_wf_type prems by blast
  hence "validScopeT \<tau> Z'" using wf_type_valid_scope by auto
  hence "validScope (ScFree n) Z'" using prems by (cases \<tau>, auto)
  hence "n \<in> dom Z" using H by (auto simp add: dom_def)
  ultimately
  show ?thesis by auto
next
  case ScBot
  have "ScBot = scopeT \<tau>" using prems by auto
  hence "scopeV v = ScBot" using ty_scope_sanity prems by auto
  hence "free_scope_store {z\<mapsto>v} = {}" by (auto simp add: free_scope_store_def)
  thus ?thesis by auto
next
  case (ScVar n)
  have "ScVar n = scopeT \<tau>" using prems by auto
  hence "scopeV v = ScVar n" using ty_scope_sanity prems by auto
  hence "free_scope_store {z\<mapsto>v} = {}" by (auto simp add: free_scope_store_def)
  thus ?thesis by auto
qed
*)

end