theory PreType
imports Typing TypingLemmas

begin

lemma wf_env_sanity:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "x\<rhd>\<tau>1 \<in> \<Gamma>"
  and "x\<rhd>\<tau>2 \<in> \<Gamma>"
  shows "\<tau>1 = \<tau>2"
using prems by (auto simp add: wf_env_def)

lemma trivial_case:
  assumes "\<tau>1 \<noteq> \<tau>3"
  and     "\<tau>2 \<noteq> \<tau>3"
  shows "\<not> (if n = i then \<tau>1 else \<tau>2) = \<tau>3"
using prems by (cases "n=i", auto)

consts
  INV11 :: "tm \<Rightarrow> name \<Rightarrow> bool" 
defs
  INV11[simp]: "INV11 e y \<equiv> \<forall>\<rho> z e1 e2. (e = let$ \<rho> (TmFreeVar z) e1 in e2) \<longrightarrow> z \<noteq> y"


lemma wf_env_union_three:
  assumes "\<Gamma>1\<union>\<Gamma>2;\<Delta> \<Turnstile> OK"
  and "\<Gamma>1\<union>\<Gamma>3;\<Delta> \<Turnstile> OK"
  and "\<Gamma>2\<union>\<Gamma>3;\<Delta> \<Turnstile> OK"
  shows "\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>3;\<Delta> \<Turnstile> OK"
proof -
  have "finite \<Gamma>3" using assms by (auto simp add: wf_env_def dom_def)
  thus ?thesis using prems proof induct
    fix x \<Gamma>3_in
    assume "finite \<Gamma>3_in"
      and H3_nin: "x \<notin> \<Gamma>3_in"
      and IH: "\<lbrakk>\<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK; \<Gamma>1 \<union> \<Gamma>3_in;\<Delta> \<Turnstile> OK; \<Gamma>2 \<union> \<Gamma>3_in;\<Delta> \<Turnstile> OK\<rbrakk> \<Longrightarrow> \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>3_in;\<Delta> \<Turnstile> OK"
      and "\<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK"
      and Hwf1: "\<Gamma>1 \<union> insert x \<Gamma>3_in;\<Delta> \<Turnstile> OK"
      and Hwf2: "\<Gamma>2 \<union> insert x \<Gamma>3_in;\<Delta> \<Turnstile> OK"
    hence H: "\<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>3_in;\<Delta> \<Turnstile> OK" using wf_env_subset by blast
    obtain a b where Hx: "x=(a, b)" by (cases x, auto)
    obtain \<tau> where Ht: "b = Term \<tau>" by (cases b, auto)
    show "\<Gamma>1 \<union> \<Gamma>2 \<union> insert x \<Gamma>3_in;\<Delta> \<Turnstile> OK" using H proof (cases "\<Gamma>1 defines a")
      case True 
      hence "x \<in> \<Gamma>1" using Hwf1 Hx Ht wf_env_union_sanity by blast
      hence "\<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>3_in = insert x (\<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>3_in)" by auto
      thus ?thesis using H by auto
    next
      case False
      thus "\<Gamma>1 \<union> \<Gamma>2 \<union> insert x \<Gamma>3_in;\<Delta> \<Turnstile> OK" proof (cases "\<Gamma>2 defines a")
        case True 
        hence "x \<in> \<Gamma>2" using Hwf2 Hx Ht wf_env_union_sanity by blast
        hence "\<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>3_in = insert x (\<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>3_in)" by auto
        thus ?thesis using H by auto
      next
        case False
        have Hnin: "\<not> (\<Gamma>1 \<union> \<Gamma>2) defines a" using prems by (auto simp add: dom_def)
        have "\<not> \<Gamma>3_in defines a" proof 
          assume H: "a \<in> Environ.dom \<Gamma>3_in"
          have "\<Gamma>3_in \<union> {x}; \<Delta> \<Turnstile> OK" using Hwf1 wf_env_subset[of "\<Gamma>1 \<union> insert x \<Gamma>3_in" \<Delta> "\<Gamma>3_in \<union> {x}"]
            by auto 
          hence "x \<in> \<Gamma>3_in" using Hwf2 Hx Ht wf_env_union_sanity H by blast
          thus False using H3_nin by auto
        qed
        hence Hnin: "\<not> (\<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>3_in) defines a" using Hnin by (auto simp add: dom_def)
        have Hwf_ty: "\<Delta> \<Turnstile>' \<tau>" using poly_env_wf_ty Hwf1 Ht Hx by auto
        hence "(\<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>3_in),(a:\<tau>);\<Delta> \<Turnstile> OK" using H wf_env_cons Hwf_ty Hnin by auto
        thus ?thesis using Ht Hx by auto
      qed
    qed
  qed auto
qed
   
lemma substitution:
  assumes "size e = n"
  and "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> vsubst_tm (TmFreeVar x) i e : \<tau>'"
  and "\<Gamma>,(x:\<tau>);\<Delta> \<Turnstile> OK"
  and "\<Gamma> \<sim> {x\<rhd>\<tau>}"
  and "\<Gamma>,(y:\<tau>);\<Delta> \<Turnstile> OK"
  and "\<Gamma> \<sim> {y\<rhd>\<tau>}"
  and "\<not> x free in e"
  and "\<not> \<Gamma> defines x"
  and "\<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> scopeT \<tau> \<noteq> ScFree \<rho>"
  shows "\<Gamma>,(y:\<tau>);\<Delta>;M;Z \<turnstile> vsubst_tm (TmFreeVar y) i e : \<tau>'"
using prems proof (induct n arbitrary: e i \<Gamma> \<tau>' Z \<Delta> \<tau> rule: nat_less_induct)
  case (1 n_y e_y i \<Gamma> \<tau>' Z \<Delta> \<tau>)
  hence IH: "\<And>e i \<Gamma> \<tau>' Z \<Delta> \<tau>. \<lbrakk> size e < size e_y;
    \<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> vsubst_tm (TmFreeVar x) i e : \<tau>'; \<Gamma>,(x:\<tau>);\<Delta> \<Turnstile> OK; \<Gamma> \<sim> {x\<rhd>\<tau>};
    \<Gamma>,(y:\<tau>);\<Delta> \<Turnstile> OK; \<Gamma> \<sim> {y\<rhd>\<tau>}; \<not> x free in e; \<not> \<Gamma> defines x; \<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> scopeT \<tau> \<noteq> ScFree \<rho> \<rbrakk>
    \<Longrightarrow> \<Gamma>,(y:\<tau>);\<Delta>;M;Z \<turnstile> vsubst_tm (TmFreeVar y) i e : \<tau>'"
    by auto
  show ?case proof (cases e_y)
    case (TmUnit q \<pi>)
    hence Ht: "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> TmUnit q \<pi> : Qual q \<pi> TyUnit" and Heq: "\<tau>' = Qual q \<pi> TyUnit" using 1 by auto
    moreover
    hence "qualG \<Gamma>,(x:\<tau>) \<preceq> qU" by auto
    hence "qualG \<Gamma>,(y:\<tau>) \<preceq> qU" by (auto simp add: qualGrelL_def)
    ultimately
    have "\<Gamma>,(y:\<tau>);\<Delta>;M;Z \<turnstile> TmUnit q \<pi> : Qual q \<pi> TyUnit" using "1.prems" by auto
    thus ?thesis using TmUnit Heq by auto
  next
    case (TmFreeVar n)
    hence Ht: "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> TmFreeVar n : \<tau>'" using 1 by auto
    moreover
    hence "qualG (\<Gamma>,(x:\<tau>)\<bar>n) \<preceq> qU" and "n \<noteq> x" using TmFreeVar "1.prems" by auto
    hence "qualG (\<Gamma>,(y:\<tau>)\<bar>n) \<preceq> qU" by (auto simp add: without_def qualGrelL_def dom_def)
    ultimately
    have Ht: "\<Gamma>,(y:\<tau>);\<Delta>;M;Z \<turnstile> TmFreeVar n : \<tau>'" using "1.prems" TmFreeVar by auto
    thus ?thesis using TmFreeVar by auto
  next
    case (TmVar j) thus ?thesis proof (cases "i = j")
      case False
      hence Ht: "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> TmVar j : \<tau>'" using 1 TmVar by auto
      hence False by (cases, auto)
      thus ?thesis by auto
    next
      case True
      hence Ht: "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> TmFreeVar x : \<tau>'" using 1 TmVar by auto
      moreover
      {
        hence "qualG (\<Gamma>,(x:\<tau>)\<bar>x) \<preceq> qU" by auto
        also have "\<Gamma>,(x:\<tau>)\<bar>x = \<Gamma>\<bar>x" by (auto simp add: without_def dom_def)
        also have "\<Gamma>\<bar>x = \<Gamma>" using 1 by (auto simp add: without_def dom_def)
        finally have "qualG (\<Gamma>,(y:\<tau>)\<bar>y) \<preceq> qU" by (auto simp add: qualGrelL_def without_def)
      }
      moreover
      have "\<Gamma>,(x:\<tau>);\<Delta> \<Turnstile> OK" using Ht ty_wf_env by auto
      hence "\<tau>' = \<tau>" using Ht wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x \<tau> \<tau>'] by auto
      ultimately
      have "\<Gamma>,(y:\<tau>);\<Delta>;M;Z \<turnstile> TmFreeVar y : \<tau>'" using "1.prems" by auto
      thus ?thesis using True TmVar by auto
    qed
  next
    case (TmAbs q \<pi> \<tau>a e)
    hence Ht: "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> \<lambda>^q^\<pi>[\<tau>a].(vsubst_tm (TmFreeVar x) (Suc i) e) : \<tau>'" using 1 by auto
    then obtain X \<tau>2 Lx
      where Hfin: "finite Lx" 
      and H\<tau>: "\<tau>' = Qual q \<pi> (\<tau>a \<rightarrow>^X \<tau>2)"
      and Ht: "\<forall>z. \<Gamma>,(x:\<tau>),(z:\<tau>a);\<Delta> \<Turnstile> OK \<and> \<Gamma>,(x:\<tau>) \<sim> {z\<rhd>\<tau>a} \<and> z \<notin> Lx \<longrightarrow>
          \<Gamma>,(x:\<tau>),(z:\<tau>a);\<Delta>;M;filterScFreeX X \<turnstile> vsubst_tm (TmFreeVar z) 0 (vsubst_tm (TmFreeVar x) (Suc i) e) : \<tau>2"
      and Hqual: "qualG (\<Gamma>,(x:\<tau>)) \<preceq> q" and Hwf: "\<Gamma>,(x:\<tau>);\<Delta> \<Turnstile> OK" and HwfM: "M;\<Delta> \<Turnstile> OK"
      and HX: "filterScVarX X = {}" 
      and H\<Delta>: "\<Delta> \<Turnstile>' \<tau>a"
      and Hmin: "minimalValidScope \<pi> Z"
      apply (rule typing_abs_elim) by auto
    have Ht: "\<Gamma>,(y:\<tau>);\<Delta>;M;Z \<turnstile> \<lambda>^q^\<pi>[\<tau>a].(vsubst_tm (TmFreeVar y) (Suc i) e) : Qual q \<pi> (\<tau>a \<rightarrow>^X \<tau>2)" 
    proof (rule Typing.t_abs[where L="Lx\<union>dom \<Gamma>\<union>set (free_tm e)\<union>{x}\<union>{y}"])
      {
        fix z
        assume "\<Gamma>,(y:\<tau>),(z:\<tau>a);\<Delta> \<Turnstile> OK"
          and "\<Gamma>,(y:\<tau>) \<sim> {z\<rhd>\<tau>a}"
          and Hzl: "z \<notin> Lx"
          and Hzdom: "z \<notin> dom \<Gamma>"
          and "z \<notin> set (free_tm e)"
          and Hzx: "z \<noteq> x"
          and Hzy: "z \<noteq> y"
        have Hwf: "\<Gamma>,(x:\<tau>),(z:\<tau>a);\<Delta> \<Turnstile> OK" apply (rule wf_env_cons) using H\<Delta> Hzdom Hzx "1.prems" 
          by(auto simp add: dom_def)
        moreover
        have Hcomp: "\<Gamma>,(x:\<tau>) \<sim> {z\<rhd>\<tau>a} \<and> z \<notin> Lx" using Hzdom Hzx Hzl by (auto simp add: env_compat_def dom_def)
        ultimately
        have
          "\<Gamma>,(x:\<tau>),(z:\<tau>a);\<Delta>;M;filterScFreeX X \<turnstile> freshen_tm z (vsubst_tm (TmFreeVar x) (Suc i) e) : \<tau>2"
          using Ht by auto
        hence Ht:
          "\<Gamma>,(x:\<tau>),(z:\<tau>a);\<Delta>;M;filterScFreeX X \<turnstile> vsubst_tm (TmFreeVar x) (Suc i) (freshen_tm z e) : \<tau>2"
          using subst_subst_swap[where i="Suc i" and j="0"] by auto
        have "\<Gamma>,(z:\<tau>a),(y:\<tau>);\<Delta>;M;filterScFreeX X \<turnstile> vsubst_tm (TmFreeVar y) (Suc i) (freshen_tm z e) : \<tau>2"
          proof (rule IH) 
          show "\<Gamma>,(z:\<tau>a),(x:\<tau>);\<Delta>;M;filterScFreeX X \<turnstile> vsubst_tm (TmFreeVar x) (Suc i) (freshen_tm z e) : \<tau>2"
            using Ht by (auto simp add: env_permute)
        next
          show "\<Gamma>,(z:\<tau>a),(x:\<tau>);\<Delta> \<Turnstile> OK" using Hwf by (auto simp add: env_permute)
        next
          have "\<not> \<Gamma> defines x" by assumption
          thus "\<Gamma>,(z:\<tau>a) \<sim> {x\<rhd>\<tau>}" using Hcomp Hzx by (auto simp add: env_compat_def dom_def)
        next
          have "\<Gamma>,(y:\<tau>),(z:\<tau>a);\<Delta> \<Turnstile> OK" apply (rule wf_env_cons) using "1.prems" Hzy Hzdom by auto
          thus "\<Gamma>,(z:\<tau>a),(y:\<tau>);\<Delta> \<Turnstile> OK" by (auto simp add: env_permute) 
        next
          show "\<Gamma>,(z:\<tau>a) \<sim> {y\<rhd>\<tau>}" using env_compat_union_right[of \<Gamma> _ "{z\<rhd>\<tau>a}"] using "1.prems" Hzy 
            by (auto simp add: env_compat_def)
        next
          have "\<not> x free in e" using TmAbs "1.prems" by auto
          thus "x \<notin> set (free_tm (freshen_tm z e))" using not_free_subst Hzx by auto
        next
          show "x \<notin> Environ.dom (\<Gamma>,(z:\<tau>a))" using "1.prems" Hzx by (auto simp add: dom_def)
        next
          have "\<forall>\<rho>. \<rho> is free dollar sc in (\<lambda>^q^\<pi>[\<tau>].e) \<longrightarrow> scopeT \<tau> \<noteq> ScFree \<rho>" using prems by auto
          thus "\<forall>\<rho>. \<rho> is free dollar sc in freshen_tm z e \<longrightarrow> scopeT \<tau> \<noteq> ScFree \<rho>"
            using free_dollars_vsubst_simpl by auto
        qed (insert TmAbs, auto)
        hence "\<Gamma>,(y:\<tau>),(z:\<tau>a);\<Delta>;M;filterScFreeX X \<turnstile> freshen_tm z (vsubst_tm (TmFreeVar y) (Suc i) e) : \<tau>2"
          using subst_subst_swap[where i="Suc i" and j="0"] by (auto simp add: env_permute)
      }
      thus "\<forall>z. \<Gamma>,(y:\<tau>),(z:\<tau>a);\<Delta> \<Turnstile> OK \<and> \<Gamma>,(y:\<tau>) \<sim> {z\<rhd>\<tau>a} \<and> z \<notin> (Lx\<union>dom \<Gamma>\<union>set (free_tm e)\<union>{x}\<union>{y}) \<longrightarrow>
        \<Gamma>,(y:\<tau>),(z:\<tau>a);\<Delta>;M;filterScFreeX X \<turnstile> freshen_tm z (vsubst_tm (TmFreeVar y) (Suc i) e) : \<tau>2" by auto
    next
      show "qualG \<Gamma>,(y:\<tau>) \<preceq> q" using Hqual by (auto simp add: qualGrelL_def)
    next
      have "finite \<Gamma>" using Hwf by (auto simp add: wf_env_def)
      thus "finite (Lx\<union>dom \<Gamma>\<union>set (free_tm e)\<union>{x}\<union>{y})" using dom_finite Hfin by auto 
    qed
    thus ?thesis using TmAbs H\<tau> by auto 
  next
    case (TmApp e1 e2)
    have Ht: "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> vsubst_tm (TmFreeVar x) i e1 \<cdot> vsubst_tm (TmFreeVar x) i e2 : \<tau>'" using 1 TmApp by auto
    then obtain X Z1 Z2 Zmin \<Gamma>1 \<Gamma>2 \<tau>arg q
      where H\<Gamma>x_eq: "\<Gamma>,(x:\<tau>) = \<Gamma>1 \<union> \<Gamma>2"
      and HZ_eq: "Z = Z1 \<union> Z2 \<union> filterScFreeX X \<union> Zmin"
      and Hcompat: "\<Gamma>1 \<sim> \<Gamma>2" and Hok: "\<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK"
      and Ht1: "\<Gamma>1;\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar x) i e1 : Qual q ScBot (\<tau>arg \<rightarrow>^X \<tau>')"
      and Ht2: "\<Gamma>2;\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar x) i e2 : \<tau>arg"
      and Hdollar1: "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tm (TmFreeVar x) i e1) \<longrightarrow> \<rho> \<notin> dom (Z2 \<union> filterScFreeX X \<union> Zmin)"
      and Hdollar2: "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tm (TmFreeVar x) i e2) \<longrightarrow> \<rho> \<notin> dom (Z1 \<union> filterScFreeX X \<union> Zmin)" 
      and Hmin: "minimalValidScope (scopeT \<tau>') Zmin"
      and Hscopeset:"filterScFreeX X \<Turnstile> ScopeSet" 
      using typing_app_elim by auto
    {
      assume Hin: "x\<rhd>\<tau> \<in> \<Gamma>1"
      moreover
      have Hwf: "\<Gamma>1;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta> \<Gamma>1] by auto
      ultimately
      have Hg1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" using wf_env_defines_without by auto 
      have Hnotin: "\<not> (\<Gamma>1\<bar>x) defines x" by (auto simp add: without_def dom_def)
      have Ht1: "(\<Gamma>1\<bar>x),(y:\<tau>);\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar y) i e1 : Qual q ScBot (\<tau>arg \<rightarrow>^X \<tau>')" proof (rule IH)
        show "(\<Gamma>1\<bar>x),(x:\<tau>);\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar x) i e1 : Qual q ScBot (\<tau>arg \<rightarrow>^X \<tau>')"
          using Ht1 Hg1 by auto
      next
        show "(\<Gamma>1\<bar>x),(x:\<tau>);\<Delta> \<Turnstile> OK" using Hg1 Hwf by auto
      next
        show "(\<Gamma>1\<bar>x) \<sim> {x\<rhd>\<tau>}" by (auto simp add: env_compat_def dom_def without_def)
      next
        have "\<Gamma>1\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        thus "(\<Gamma>1\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using "1.prems" wf_env_subset[of "\<Gamma>,(y:\<tau>)" \<Delta> "(\<Gamma>1\<bar>x),(y:\<tau>)"] by auto
      next
        have "\<Gamma>1\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        thus "\<Gamma>1\<bar>x \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset[of \<Gamma> "{y\<rhd>\<tau>}" "(\<Gamma>1\<bar>x)"] by auto
      next
        show "\<not> x free in e1" using "1.prems" TmApp by auto
      next
        show "\<not> (\<Gamma>1\<bar>x) defines x" using Hnotin by auto
      qed (insert TmApp "1.prems", auto)
    }
    hence Ht1i: "x\<rhd>\<tau> \<in> \<Gamma>1 \<longrightarrow> (\<Gamma>1\<bar>x),(y:\<tau>);\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar y) i e1 : Qual q ScBot (\<tau>arg \<rightarrow>^X \<tau>')" 
      by auto
    {
      assume Hin: "x\<rhd>\<tau> \<in> \<Gamma>2"
      moreover
      have Hwf: "\<Gamma>2;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta> \<Gamma>2] by auto
      ultimately
      have Hg1: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" using wf_env_defines_without by auto 
      have Hnotin: "\<not> (\<Gamma>2\<bar>x) defines x" by (auto simp add: without_def dom_def)
      have Ht1: "(\<Gamma>2\<bar>x),(y:\<tau>);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar y) i e2 : \<tau>arg" proof (rule IH)
        show "(\<Gamma>2\<bar>x),(x:\<tau>);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar x) i e2 : \<tau>arg"
          using Ht2 Hg1 by auto
      next
        show "(\<Gamma>2\<bar>x),(x:\<tau>);\<Delta> \<Turnstile> OK" using Hg1 Hwf by auto
      next
        show "(\<Gamma>2\<bar>x) \<sim> {x\<rhd>\<tau>}" by (auto simp add: env_compat_def dom_def without_def)
      next
        have "\<Gamma>2\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        thus "(\<Gamma>2\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using "1.prems" wf_env_subset[of "\<Gamma>,(y:\<tau>)" \<Delta> "(\<Gamma>2\<bar>x),(y:\<tau>)"] by auto
      next
        have "\<Gamma>2\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        thus "\<Gamma>2\<bar>x \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset[of \<Gamma> "{y\<rhd>\<tau>}" "(\<Gamma>2\<bar>x)"] by auto
      next
        show "\<not> x free in e2" using "1.prems" TmApp by auto
      next
        show "\<not> (\<Gamma>2\<bar>x) defines x" using Hnotin by auto
      qed (insert TmApp "1.prems", auto)
    }
    hence Ht2i: "x\<rhd>\<tau> \<in> \<Gamma>2 \<longrightarrow> (\<Gamma>2\<bar>x),(y:\<tau>);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar y) i e2 : \<tau>arg"
      by auto
    have Hdom1: "x\<rhd>\<tau> \<notin> \<Gamma>1 \<longrightarrow> \<not> \<Gamma>1 defines x" proof (auto simp add: dom_def)
      fix b
      assume "(x, b) \<in> \<Gamma>1"
      and Hne: "x\<rhd>\<tau> \<notin> \<Gamma>1"
      then obtain \<tau>' where "b = Term \<tau>'" and Hin: "x\<rhd>\<tau>' \<in> \<Gamma>1" by (cases b, auto)
      hence "x\<rhd>\<tau>' \<in> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
      hence "\<tau>' = \<tau>" using Ht ty_wf_env wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x \<tau> \<tau>'] by auto
      thus False using Hne Hin by auto
    qed
    have Hdom2: "x\<rhd>\<tau> \<notin> \<Gamma>2 \<longrightarrow> \<not> \<Gamma>2 defines x" proof (auto simp add: dom_def)
      fix b
      assume "(x, b) \<in> \<Gamma>2"
      and Hne: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      then obtain \<tau>' where "b = Term \<tau>'" and Hin: "x\<rhd>\<tau>' \<in> \<Gamma>2" by (cases b, auto)
      hence "x\<rhd>\<tau>' \<in> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
      hence "\<tau>' = \<tau>" using Ht ty_wf_env wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x \<tau> \<tau>'] by auto
      thus False using Hne Hin by auto
    qed
    hence Ht1o: "x\<rhd>\<tau> \<notin> \<Gamma>1 \<longrightarrow> \<Gamma>1;\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar y) i e1 : Qual q ScBot (\<tau>arg \<rightarrow>^X \<tau>')" 
      using Ht1 not_in_env_ty_same_subst Hdom1 by auto
    hence Ht2o: "x\<rhd>\<tau> \<notin> \<Gamma>2 \<longrightarrow> \<Gamma>2;\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar y) i e2 : \<tau>arg" 
      using Ht2 not_in_env_ty_same_subst Hdom2 by auto
    have Hd1: "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e1 \<longrightarrow> \<rho> \<notin> Environ.dom (Z2 \<union> filterScFreeX X \<union> Zmin)"
      using Hdollar1 free_dollars_vsubst_simpl by auto
    have Hd2: "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e2 \<longrightarrow> \<rho> \<notin> Environ.dom (Z1 \<union> filterScFreeX X \<union> Zmin)"
      using Hdollar2 free_dollars_vsubst_simpl by auto
    have Hwf_all: "\<Gamma>,(y:\<tau>);\<Delta> \<Turnstile> OK" by assumption
    {
      assume H1: "x\<rhd>\<tau> \<in> \<Gamma>1" and H2: "x\<rhd>\<tau> \<in> \<Gamma>2"
      hence Hq: "qualT \<tau> \<preceq> qU" using Hcompat by (auto simp add: env_compat_def)
      have Hwf: "\<Gamma>2;\<Delta> \<Turnstile> OK" and "\<Gamma>1;\<Delta> \<Turnstile> OK" 
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heq2: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" and Heq1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" using wf_env_defines_without H1 H2 by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x)" by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x" by (auto simp add: insert_ident)
      have "(\<Gamma>1\<bar>x) \<sim> (\<Gamma>2\<bar>x)" using Hcompat by (auto simp add: without_def env_compat_def)
      hence "(\<Gamma>1\<bar>x),(y:\<tau>) \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using Hq by (auto simp add: env_compat_def)
      hence Ht: "(\<Gamma>1\<bar>x),(y:\<tau>)\<union>(\<Gamma>2\<bar>x),(y:\<tau>);\<Delta>;M;Z1\<union>Z2\<union>filterScFreeX X\<union>Zmin \<turnstile> 
        vsubst_tm (TmFreeVar y) i e1 \<cdot> vsubst_tm (TmFreeVar y) i e2 : \<tau>'"
        apply rule using Hscopeset Hmin Hd1 Hd2 H1 H2 Ht1i Ht2i H_eq Hwf_all by auto
      hence ?thesis using H_eq TmApp HZ_eq by auto
    }
    moreover
    {
      assume H1: "x\<rhd>\<tau> \<in> \<Gamma>1" and H2: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      have Hwf: "\<Gamma>1;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heq1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" using wf_env_defines_without H1 H2 by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>1\<bar>x \<union> \<Gamma>2),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>1\<bar>x \<union> \<Gamma>2)" using H2 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>1\<bar>x \<union> \<Gamma>2" by (auto simp add: insert_ident)
      have "(\<Gamma>1\<bar>x) \<sim> \<Gamma>2" using Hcompat by (auto simp add: without_def env_compat_def)
      moreover
      have "{y\<rhd>\<tau>} \<sim> \<Gamma>2" using "1.prems" using env_compat_subset[of \<Gamma> "{y\<rhd>\<tau>}" \<Gamma>2] H_eq H2
        by (auto simp add: env_compat_def)
      ultimately
      have "(\<Gamma>1\<bar>x),(y:\<tau>) \<sim> \<Gamma>2" using env_compat_union_right[of "\<Gamma>1\<bar>x" \<Gamma>2 "{y\<rhd>\<tau>}"] by auto
      hence Ht: "(\<Gamma>1\<bar>x),(y:\<tau>)\<union>\<Gamma>2;\<Delta>;M;Z1\<union>Z2\<union>filterScFreeX X\<union>Zmin \<turnstile> 
        vsubst_tm (TmFreeVar y) i e1 \<cdot> vsubst_tm (TmFreeVar y) i e2 : \<tau>'"
        apply rule using Hscopeset Hmin Hd1 Hd2 H1 H2 Ht1i Ht2o H_eq Hwf_all by auto
      hence ?thesis using H_eq TmApp HZ_eq by auto
    }
    moreover
    {
      assume H1: "x\<rhd>\<tau> \<notin> \<Gamma>1" and H2: "x\<rhd>\<tau> \<in> \<Gamma>2"
      have Hwf: "\<Gamma>2;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heq1: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" using wf_env_defines_without H1 H2 by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>1 \<union> (\<Gamma>2\<bar>x)),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>1 \<union> \<Gamma>2\<bar>x)" using H1 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>1 \<union> \<Gamma>2\<bar>x" by (auto simp add: insert_ident)
      have "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x)" using Hcompat by (auto simp add: without_def env_compat_def)
      moreover
      have "\<Gamma>1 \<sim> {y\<rhd>\<tau>}" using "1.prems" using env_compat_subset[of \<Gamma> "{y\<rhd>\<tau>}" \<Gamma>1] H_eq H1 by auto
      ultimately
      have "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using env_compat_union_right[of "\<Gamma>2\<bar>x" \<Gamma>1 "{y\<rhd>\<tau>}"] by (auto simp add: env_compat_def)
      hence Ht: "\<Gamma>1\<union>(\<Gamma>2\<bar>x),(y:\<tau>);\<Delta>;M;Z1\<union>Z2\<union>filterScFreeX X\<union>Zmin \<turnstile> 
        vsubst_tm (TmFreeVar y) i e1 \<cdot> vsubst_tm (TmFreeVar y) i e2 : \<tau>'"
        apply rule using Hscopeset Hmin Hd1 Hd2 H1 H2 Ht1o Ht2i H_eq Hwf_all by auto
      hence ?thesis using H_eq TmApp HZ_eq by auto
    }
    moreover
    {
      assume H1: "x\<rhd>\<tau> \<notin> \<Gamma>1" and H2: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      hence "\<not> \<Gamma>1 defines x" and "\<not> \<Gamma>2 defines x" using Hdom1 Hdom2 by auto
      hence ?thesis using H\<Gamma>x_eq by (auto simp add: dom_def)
    }
    ultimately
    show ?thesis by auto
  next
    case (TmNew er)
    hence "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> TmNew (vsubst_tm (TmFreeVar x) i er) : \<tau>'" using "1.prems" by auto
    then obtain \<tau>r where H\<tau>: "\<tau>' = Qual qL ScBot (TyRef \<tau>r)"
      and Ht: "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> vsubst_tm (TmFreeVar x) i er : \<tau>r" by auto
    have "\<Gamma>,(y:\<tau>);\<Delta>;M;Z \<turnstile> vsubst_tm (TmFreeVar y) i er : \<tau>r" apply (rule IH) using "1.prems" TmNew Ht by auto
    thus ?thesis using H\<tau> TmNew by auto
  next
    case (TmProd q \<pi> e)
    hence Ht: "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> \<Lambda>^q^\<pi>[*]. vsubst_tm (TmFreeVar x) i e : \<tau>'"
      using "1.prems" by auto
    then obtain X \<tau>b La
      where H\<tau>': "\<tau>' = Qual q \<pi> (\<forall>.\<tau>b^X)"
      and Hvalue: "value (vsubst_tm (TmFreeVar x) i e)"
      and Hqual: "qualG \<Gamma>,(x:\<tau>) \<preceq> q"
      and Htyp: "\<forall>c. c \<notin> La \<longrightarrow> \<Gamma>,(x:\<tau>);\<Delta>,(c:\<star>);M;filterScFreeX X \<turnstile>
                                 vsubst_tmty (TyFreeVar c) 0 (vsubst_tm (TmFreeVar x) i e) : vsubst_ty (TyFreeVar c) 0 \<tau>b"
      and "minimalValidScope \<pi> Z"  
      and Hfin: "finite La"
      using typing_poly_elim[of "\<Gamma>,(x:\<tau>)" \<Delta> M Z q \<pi> "vsubst_tm (TmFreeVar x) i e" \<tau>'] by force
    have "\<Gamma>,(y:\<tau>);\<Delta>;M;Z \<turnstile> \<Lambda>^q^\<pi>[*]. vsubst_tm (TmFreeVar y) i e : Qual q \<pi> (\<forall>.\<tau>b^X)" proof rule
      show "value (vsubst_tm (TmFreeVar y) i e)" using Hvalue by (induct e, auto)
    next
      show "qualG \<Gamma>,(y:\<tau>) \<preceq> q" using Hqual by (auto simp add: qualGrelL_def)
    next
      {
        fix a
        assume "a \<notin> La"
        hence Htyp: "\<Gamma>,(x:\<tau>);\<Delta>,(a:\<star>);M;filterScFreeX X \<turnstile> vsubst_tm (TmFreeVar x) i (freshen_tmty a e) : freshen_ty a \<tau>b"
          using subst_subst_tmty_swap Htyp by auto
        have "\<Gamma>,(y:\<tau>);\<Delta>,(a:\<star>);M;filterScFreeX X \<turnstile> vsubst_tm (TmFreeVar y) i (freshen_tmty a e) : freshen_ty a \<tau>b"
          apply (rule IH)
          using TmProd "1.prems" TmProd poly_env_ty_cons poly_free_var_tmty_equal Htyp poly_free_dollars_tmty_equal by auto 
        hence "\<Gamma>,(y:\<tau>);\<Delta>,(a:\<star>);M;filterScFreeX X \<turnstile> freshen_tmty a (vsubst_tm (TmFreeVar y) i e) : freshen_ty a \<tau>b"
          using subst_subst_tmty_swap by auto
      }
      thus "\<forall>c. c \<notin> La \<longrightarrow> \<Gamma>,(y:\<tau>);\<Delta>,(c:\<star>);M;filterScFreeX X \<turnstile>
                                 freshen_tmty c (vsubst_tm (TmFreeVar y) i e) : freshen_ty c \<tau>b" by auto
    qed (insert prems, auto)
    thus ?thesis using TmProd H\<tau>' by auto
  next
    case (TmTapp e \<phi>)
    hence "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> (vsubst_tm (TmFreeVar x) i e)\<langle>\<phi>\<rangle> : \<tau>'" using TmTapp "1.prems" by auto
    then obtain Ze Z\<tau> \<tau>e q X 
      where HZ_eq: "Z = Ze \<union> filterScFreeX X \<union> Z\<tau>"
      and H\<tau>: "\<tau>' = vsubst_ty \<phi> 0 \<tau>e"
      and Htyp: "\<Gamma>,(x:\<tau>);\<Delta>;M;Ze \<turnstile> vsubst_tm (TmFreeVar x) i e : Qual q ScBot (\<forall>.\<tau>e^X)"
      and Hmin: "minimalValidScope (scopeT \<tau>e) Z\<tau>"
      and Hscopeset: "filterScFreeX X \<Turnstile> ScopeSet"
      and Hdollar: "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tm (TmFreeVar x) i e) \<longrightarrow> \<rho> \<notin> dom (filterScFreeX X \<union> Z\<tau>)"
      using typing_tyapp_elim by auto
    have "\<Gamma>,(y:\<tau>);\<Delta>;M;Ze\<union>filterScFreeX X\<union>Z\<tau>\<turnstile> (vsubst_tm (TmFreeVar y) i e)\<langle>\<phi>\<rangle> : vsubst_ty \<phi> 0 \<tau>e" proof rule
      show "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tm (TmFreeVar y) i e) \<longrightarrow> \<rho> \<notin> dom (filterScFreeX X \<union> Z\<tau>)"
        using Hdollar free_dollars_vsubst_simpl by auto
    next
      show "\<Gamma>,(y:\<tau>);\<Delta>;M;Ze \<turnstile> vsubst_tm (TmFreeVar y) i e : Qual q ScBot (\<forall>.\<tau>e^X)" by (rule IH, insert prems, auto)
    qed (insert prems, auto)
    thus ?thesis using TmTapp H\<tau> HZ_eq by auto
  next
    case (TmFree e)
    hence "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> TmFree (vsubst_tm (TmFreeVar x) i e) : \<tau>'" using "1.prems" by auto
    then obtain Z\<tau> Ze 
      where Ht: "\<Gamma>,(x:\<tau>);\<Delta>;M;Ze \<turnstile> vsubst_tm (TmFreeVar x) i e : Qual qL ScBot (TyRef \<tau>')"
      and Hmin: "minimalValidScope (scopeT \<tau>') Z\<tau>"
      and HZ_eq: "Z = Ze \<union> Z\<tau>"
      and Hdollar: "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tm (TmFreeVar x) i e) \<longrightarrow> \<not> Z\<tau> defines \<rho>"
      using typing_free_elim by auto
    have "\<Gamma>,(y:\<tau>);\<Delta>;M;Ze\<union>Z\<tau> \<turnstile> TmFree (vsubst_tm (TmFreeVar y) i e) : \<tau>'" 
      apply rule using prems free_dollars_vsubst_simpl by auto
    thus ?thesis using HZ_eq TmFree by auto
  next
    case (TmAssign e1 e2)
    have Ht: "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> vsubst_tm (TmFreeVar x) i e1 := vsubst_tm (TmFreeVar x) i e2 : \<tau>'" using 1 TmAssign by auto
    then obtain Z1 Z2 \<Gamma>1 \<Gamma>2 \<tau>arg \<pi> 
      where H\<Gamma>x_eq: "\<Gamma>,(x:\<tau>) = \<Gamma>1 \<union> \<Gamma>2"
      and HZ_eq: "Z = Z1 \<union> Z2"
      and Hcompat: "\<Gamma>1 \<sim> \<Gamma>2" and Hok: "\<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK"
      and Ht1: "\<Gamma>1;\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar x) i e1 : Qual qU \<pi> (TyRef \<tau>arg)"
      and Ht2: "\<Gamma>2;\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar x) i e2 : \<tau>arg"
      and Hdollar1: "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tm (TmFreeVar x) i e1) \<longrightarrow> \<rho> \<notin> Environ.dom Z2"
      and Hdollar2: "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tm (TmFreeVar x) i e2) \<longrightarrow> \<rho> \<notin> Environ.dom Z1" 
      and Hunit: "\<tau>' = Qual qU ScBot TyUnit"
      and Hqual: "qualT \<tau>arg \<preceq> qU"
      using typing_assign_elim by auto
    {
      assume Hin: "x\<rhd>\<tau> \<in> \<Gamma>1"
      moreover
      have Hwf: "\<Gamma>1;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta> \<Gamma>1] by auto
      ultimately
      have Hg1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" using wf_env_defines_without by auto 
      have Hnotin: "\<not> (\<Gamma>1\<bar>x) defines x" by (auto simp add: without_def dom_def)
      have Ht1: "(\<Gamma>1\<bar>x),(y:\<tau>);\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar y) i e1 : Qual qU \<pi> (TyRef \<tau>arg)" proof (rule IH)
        show "(\<Gamma>1\<bar>x),(x:\<tau>);\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar x) i e1 : Qual qU \<pi> (TyRef \<tau>arg)"
          using Ht1 Hg1 by auto
      next
        show "(\<Gamma>1\<bar>x),(x:\<tau>);\<Delta> \<Turnstile> OK" using Hg1 Hwf by auto
      next
        show "(\<Gamma>1\<bar>x) \<sim> {x\<rhd>\<tau>}" by (auto simp add: env_compat_def dom_def without_def)
      next
        have "\<Gamma>1\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        thus "(\<Gamma>1\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using "1.prems" wf_env_subset[of "\<Gamma>,(y:\<tau>)" \<Delta> "(\<Gamma>1\<bar>x),(y:\<tau>)"] by auto
      next
        have "\<Gamma>1\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        thus "\<Gamma>1\<bar>x \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset[of \<Gamma> "{y\<rhd>\<tau>}" "(\<Gamma>1\<bar>x)"] by auto
      next
        show "\<not> x free in e1" using "1.prems" TmAssign by auto
      next
        show "\<not> (\<Gamma>1\<bar>x) defines x" using Hnotin by auto
      qed (insert TmAssign "1.prems", auto)
    }
    hence Ht1i: "x\<rhd>\<tau> \<in> \<Gamma>1 \<longrightarrow> (\<Gamma>1\<bar>x),(y:\<tau>);\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar y) i e1 : Qual qU \<pi> (TyRef \<tau>arg)"
      by auto
    {
      assume Hin: "x\<rhd>\<tau> \<in> \<Gamma>2"
      moreover
      have Hwf: "\<Gamma>2;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta> \<Gamma>2] by auto
      ultimately
      have Hg1: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" using wf_env_defines_without by auto 
      have Hnotin: "\<not> (\<Gamma>2\<bar>x) defines x" by (auto simp add: without_def dom_def)
      have Ht1: "(\<Gamma>2\<bar>x),(y:\<tau>);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar y) i e2 : \<tau>arg" proof (rule IH)
        show "(\<Gamma>2\<bar>x),(x:\<tau>);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar x) i e2 : \<tau>arg"
          using Ht2 Hg1 by auto
      next
        show "(\<Gamma>2\<bar>x),(x:\<tau>);\<Delta> \<Turnstile> OK" using Hg1 Hwf by auto
      next
        show "(\<Gamma>2\<bar>x) \<sim> {x\<rhd>\<tau>}" by (auto simp add: env_compat_def dom_def without_def)
      next
        have "\<Gamma>2\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        thus "(\<Gamma>2\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using "1.prems" wf_env_subset[of "\<Gamma>,(y:\<tau>)" \<Delta> "(\<Gamma>2\<bar>x),(y:\<tau>)"] by auto
      next
        have "\<Gamma>2\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        thus "\<Gamma>2\<bar>x \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset[of \<Gamma> "{y\<rhd>\<tau>}" "(\<Gamma>2\<bar>x)"] by auto
      next
        show "\<not> x free in e2" using "1.prems" TmAssign by auto
      next
        show "\<not> (\<Gamma>2\<bar>x) defines x" using Hnotin by auto
      qed (insert TmAssign "1.prems", auto)
    }
    hence Ht2i: "x\<rhd>\<tau> \<in> \<Gamma>2 \<longrightarrow> (\<Gamma>2\<bar>x),(y:\<tau>);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar y) i e2 : \<tau>arg"
      by auto
    have Hdom1: "x\<rhd>\<tau> \<notin> \<Gamma>1 \<longrightarrow> \<not> \<Gamma>1 defines x" proof (auto simp add: dom_def)
      fix b
      assume "(x, b) \<in> \<Gamma>1"
      and Hne: "x\<rhd>\<tau> \<notin> \<Gamma>1"
      then obtain \<tau>' where "b = Term \<tau>'" and Hin: "x\<rhd>\<tau>' \<in> \<Gamma>1" by (cases b, auto)
      hence "x\<rhd>\<tau>' \<in> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
      hence "\<tau>' = \<tau>" using Ht ty_wf_env wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x \<tau> \<tau>'] by auto
      thus False using Hne Hin by auto
    qed
    have Hdom2: "x\<rhd>\<tau> \<notin> \<Gamma>2 \<longrightarrow> \<not> \<Gamma>2 defines x" proof (auto simp add: dom_def)
      fix b
      assume "(x, b) \<in> \<Gamma>2"
      and Hne: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      then obtain \<tau>' where "b = Term \<tau>'" and Hin: "x\<rhd>\<tau>' \<in> \<Gamma>2" by (cases b, auto)
      hence "x\<rhd>\<tau>' \<in> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
      hence "\<tau>' = \<tau>" using Ht ty_wf_env wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x \<tau> \<tau>'] by auto
      thus False using Hne Hin by auto
    qed
    hence Ht1o: "x\<rhd>\<tau> \<notin> \<Gamma>1 \<longrightarrow> \<Gamma>1;\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar y) i e1 : Qual qU \<pi> (TyRef \<tau>arg)"
      using Ht1 not_in_env_ty_same_subst Hdom1 by auto
    hence Ht2o: "x\<rhd>\<tau> \<notin> \<Gamma>2 \<longrightarrow> \<Gamma>2;\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar y) i e2 : \<tau>arg" 
      using Ht2 not_in_env_ty_same_subst Hdom2 by auto
    have Hd1: "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e1 \<longrightarrow> \<rho> \<notin> Environ.dom Z2"
      using Hdollar1 free_dollars_vsubst_simpl by auto
    have Hd2: "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e2 \<longrightarrow> \<rho> \<notin> Environ.dom Z1"
      using Hdollar2 free_dollars_vsubst_simpl by auto
    have Hwf_all: "\<Gamma>,(y:\<tau>);\<Delta> \<Turnstile> OK" by assumption
    {
      assume H1: "x\<rhd>\<tau> \<in> \<Gamma>1" and H2: "x\<rhd>\<tau> \<in> \<Gamma>2"
      hence Hq: "qualT \<tau> \<preceq> qU" using Hcompat by (auto simp add: env_compat_def)
      have Hwf: "\<Gamma>2;\<Delta> \<Turnstile> OK" and "\<Gamma>1;\<Delta> \<Turnstile> OK" 
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heq2: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" and Heq1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" using wf_env_defines_without H1 H2 by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x)" by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x" by (auto simp add: insert_ident)
      have "(\<Gamma>1\<bar>x) \<sim> (\<Gamma>2\<bar>x)" using Hcompat by (auto simp add: without_def env_compat_def)
      hence "(\<Gamma>1\<bar>x),(y:\<tau>) \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using Hq by (auto simp add: env_compat_def)
      hence Ht: "(\<Gamma>1\<bar>x),(y:\<tau>)\<union>(\<Gamma>2\<bar>x),(y:\<tau>);\<Delta>;M;Z1\<union>Z2 \<turnstile> 
        vsubst_tm (TmFreeVar y) i e1 := vsubst_tm (TmFreeVar y) i e2 : Qual qU ScBot TyUnit"
        apply rule using Hd1 Hd2 H1 H2 Ht1i Ht2i H_eq Hwf_all Hqual by auto
      hence ?thesis using H_eq TmAssign HZ_eq Hunit by auto
    }
    moreover
    {
      assume H1: "x\<rhd>\<tau> \<in> \<Gamma>1" and H2: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      have Hwf: "\<Gamma>1;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heq1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" using wf_env_defines_without H1 H2 by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>1\<bar>x \<union> \<Gamma>2),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>1\<bar>x \<union> \<Gamma>2)" using H2 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>1\<bar>x \<union> \<Gamma>2" by (auto simp add: insert_ident)
      have "(\<Gamma>1\<bar>x) \<sim> \<Gamma>2" using Hcompat by (auto simp add: without_def env_compat_def)
      moreover
      have "{y\<rhd>\<tau>} \<sim> \<Gamma>2" using "1.prems" using env_compat_subset[of \<Gamma> "{y\<rhd>\<tau>}" \<Gamma>2] H_eq H2
        by (auto simp add: env_compat_def)
      ultimately
      have "(\<Gamma>1\<bar>x),(y:\<tau>) \<sim> \<Gamma>2" using env_compat_union_right[of "\<Gamma>1\<bar>x" \<Gamma>2 "{y\<rhd>\<tau>}"] by auto
      hence Ht: "(\<Gamma>1\<bar>x),(y:\<tau>)\<union>\<Gamma>2;\<Delta>;M;Z1\<union>Z2 \<turnstile> 
        vsubst_tm (TmFreeVar y) i e1 := vsubst_tm (TmFreeVar y) i e2 : Qual qU ScBot TyUnit"
        apply rule using Hd1 Hd2 H1 H2 Ht1i Ht2o H_eq Hwf_all Hqual by auto
      hence ?thesis using H_eq TmAssign HZ_eq Hunit by auto
    }
    moreover
    {
      assume H1: "x\<rhd>\<tau> \<notin> \<Gamma>1" and H2: "x\<rhd>\<tau> \<in> \<Gamma>2"
      have Hwf: "\<Gamma>2;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heq1: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" using wf_env_defines_without H1 H2 by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>1 \<union> (\<Gamma>2\<bar>x)),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>1 \<union> \<Gamma>2\<bar>x)" using H1 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>1 \<union> \<Gamma>2\<bar>x" by (auto simp add: insert_ident)
      have "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x)" using Hcompat by (auto simp add: without_def env_compat_def)
      moreover
      have "\<Gamma>1 \<sim> {y\<rhd>\<tau>}" using "1.prems" using env_compat_subset[of \<Gamma> "{y\<rhd>\<tau>}" \<Gamma>1] H_eq H1 by auto
      ultimately
      have "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using env_compat_union_right[of "\<Gamma>2\<bar>x" \<Gamma>1 "{y\<rhd>\<tau>}"] by (auto simp add: env_compat_def)
      hence Ht: "\<Gamma>1\<union>(\<Gamma>2\<bar>x),(y:\<tau>);\<Delta>;M;Z1\<union>Z2 \<turnstile> 
        vsubst_tm (TmFreeVar y) i e1 := vsubst_tm (TmFreeVar y) i e2 : Qual qU ScBot TyUnit"
        apply rule using Hd1 Hd2 H1 H2 Ht1o Ht2i H_eq Hwf_all Hqual by auto
      hence ?thesis using H_eq TmAssign HZ_eq Hunit by auto
    }
    moreover
    {
      assume H1: "x\<rhd>\<tau> \<notin> \<Gamma>1" and H2: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      hence "\<not> \<Gamma>1 defines x" and "\<not> \<Gamma>2 defines x" using Hdom1 Hdom2 by auto
      hence ?thesis using H\<Gamma>x_eq by (auto simp add: dom_def)
    }
    ultimately
    show ?thesis by auto
  next
    case (TmSwap e1 e2)
    hence Ht: "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> vsubst_tm (TmFreeVar x) i e1 :=: vsubst_tm (TmFreeVar x) i e2 : \<tau>'" using 1 by auto
    then obtain Z1 Z2 \<Gamma>1 \<Gamma>2 \<pi> 
      where H\<Gamma>x_eq: "\<Gamma>,(x:\<tau>) = \<Gamma>1 \<union> \<Gamma>2"
      and HZ_eq: "Z = Z2 \<union> Z1"
      and Hcompat: "\<Gamma>1 \<sim> \<Gamma>2" and Hok: "\<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK"
      and Ht1: "\<Gamma>1;\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar x) i e1 : Qual qU \<pi> (TyRef \<tau>')"
      and Ht2: "\<Gamma>2;\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar x) i e2 : \<tau>'"
      and Hdollar1: "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tm (TmFreeVar x) i e1) \<longrightarrow> \<rho> \<notin> Environ.dom Z2"
      and Hdollar2: "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tm (TmFreeVar x) i e2) \<longrightarrow> \<rho> \<notin> Environ.dom Z1" 
      and Hqual: "qualT \<tau>' = qL"
      using typing_swap_elim by auto
    {
      assume Hin: "x\<rhd>\<tau> \<in> \<Gamma>1"
      moreover
      have Hwf: "\<Gamma>1;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta> \<Gamma>1] by auto
      ultimately
      have Hg1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" using wf_env_defines_without by auto 
      have Hnotin: "\<not> (\<Gamma>1\<bar>x) defines x" by (auto simp add: without_def dom_def)
      have Ht1: "(\<Gamma>1\<bar>x),(y:\<tau>);\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar y) i e1 : Qual qU \<pi> (TyRef \<tau>')" proof (rule IH)
        show "(\<Gamma>1\<bar>x),(x:\<tau>);\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar x) i e1 : Qual qU \<pi> (TyRef \<tau>')"
          using Ht1 Hg1 by auto
      next
        show "(\<Gamma>1\<bar>x),(x:\<tau>);\<Delta> \<Turnstile> OK" using Hg1 Hwf by auto
      next
        show "(\<Gamma>1\<bar>x) \<sim> {x\<rhd>\<tau>}" by (auto simp add: env_compat_def dom_def without_def)
      next
        have "\<Gamma>1\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        thus "(\<Gamma>1\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using "1.prems" wf_env_subset[of "\<Gamma>,(y:\<tau>)" \<Delta> "(\<Gamma>1\<bar>x),(y:\<tau>)"] by auto
      next
        have "\<Gamma>1\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        thus "\<Gamma>1\<bar>x \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset[of \<Gamma> "{y\<rhd>\<tau>}" "(\<Gamma>1\<bar>x)"] by auto
      next
        show "\<not> x free in e1" using "1.prems" TmSwap by auto
      next
        show "\<not> (\<Gamma>1\<bar>x) defines x" using Hnotin by auto
      qed (insert TmSwap "1.prems", auto)
    }
    hence Ht1i: "x\<rhd>\<tau> \<in> \<Gamma>1 \<longrightarrow> (\<Gamma>1\<bar>x),(y:\<tau>);\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar y) i e1 : Qual qU \<pi> (TyRef \<tau>')"
      by auto
    {
      assume Hin: "x\<rhd>\<tau> \<in> \<Gamma>2"
      moreover
      have Hwf: "\<Gamma>2;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta> \<Gamma>2] by auto
      ultimately
      have Hg1: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" using wf_env_defines_without by auto 
      have Hnotin: "\<not> (\<Gamma>2\<bar>x) defines x" by (auto simp add: without_def dom_def)
      have Ht1: "(\<Gamma>2\<bar>x),(y:\<tau>);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar y) i e2 : \<tau>'" proof (rule IH)
        show "(\<Gamma>2\<bar>x),(x:\<tau>);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar x) i e2 : \<tau>'"
          using Ht2 Hg1 by auto
      next
        show "(\<Gamma>2\<bar>x),(x:\<tau>);\<Delta> \<Turnstile> OK" using Hg1 Hwf by auto
      next
        show "(\<Gamma>2\<bar>x) \<sim> {x\<rhd>\<tau>}" by (auto simp add: env_compat_def dom_def without_def)
      next
        have "\<Gamma>2\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        thus "(\<Gamma>2\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using "1.prems" wf_env_subset[of "\<Gamma>,(y:\<tau>)" \<Delta> "(\<Gamma>2\<bar>x),(y:\<tau>)"] by auto
      next
        have "\<Gamma>2\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        thus "\<Gamma>2\<bar>x \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset[of \<Gamma> "{y\<rhd>\<tau>}" "(\<Gamma>2\<bar>x)"] by auto
      next
        show "\<not> x free in e2" using "1.prems" TmSwap by auto
      next
        show "\<not> (\<Gamma>2\<bar>x) defines x" using Hnotin by auto
      qed (insert TmSwap "1.prems", auto)
    }
    hence Ht2i: "x\<rhd>\<tau> \<in> \<Gamma>2 \<longrightarrow> (\<Gamma>2\<bar>x),(y:\<tau>);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar y) i e2 : \<tau>'"
      by auto
    have Hdom1: "x\<rhd>\<tau> \<notin> \<Gamma>1 \<longrightarrow> \<not> \<Gamma>1 defines x" proof (auto simp add: dom_def)
      fix b
      assume "(x, b) \<in> \<Gamma>1"
      and Hne: "x\<rhd>\<tau> \<notin> \<Gamma>1"
      then obtain \<tau>' where "b = Term \<tau>'" and Hin: "x\<rhd>\<tau>' \<in> \<Gamma>1" by (cases b, auto)
      hence "x\<rhd>\<tau>' \<in> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
      hence "\<tau>' = \<tau>" using Ht ty_wf_env wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x \<tau> \<tau>'] by auto
      thus False using Hne Hin by auto
    qed
    have Hdom2: "x\<rhd>\<tau> \<notin> \<Gamma>2 \<longrightarrow> \<not> \<Gamma>2 defines x" proof (auto simp add: dom_def)
      fix b
      assume "(x, b) \<in> \<Gamma>2"
      and Hne: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      then obtain \<tau>' where "b = Term \<tau>'" and Hin: "x\<rhd>\<tau>' \<in> \<Gamma>2" by (cases b, auto)
      hence "x\<rhd>\<tau>' \<in> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
      hence "\<tau>' = \<tau>" using Ht ty_wf_env wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x \<tau> \<tau>'] by auto
      thus False using Hne Hin by auto
    qed
    hence Ht1o: "x\<rhd>\<tau> \<notin> \<Gamma>1 \<longrightarrow> \<Gamma>1;\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar y) i e1 : Qual qU \<pi> (TyRef \<tau>')"
      using Ht1 not_in_env_ty_same_subst Hdom1 by auto
    hence Ht2o: "x\<rhd>\<tau> \<notin> \<Gamma>2 \<longrightarrow> \<Gamma>2;\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar y) i e2 : \<tau>'" 
      using Ht2 not_in_env_ty_same_subst Hdom2 by auto
    have Hd1: "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e1 \<longrightarrow> \<rho> \<notin> Environ.dom Z2"
      using Hdollar1 free_dollars_vsubst_simpl by auto
    have Hd2: "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e2 \<longrightarrow> \<rho> \<notin> Environ.dom Z1"
      using Hdollar2 free_dollars_vsubst_simpl by auto
    have Hwf_all: "\<Gamma>,(y:\<tau>);\<Delta> \<Turnstile> OK" by assumption
    {
      assume H1: "x\<rhd>\<tau> \<in> \<Gamma>1" and H2: "x\<rhd>\<tau> \<in> \<Gamma>2"
      hence Hq: "qualT \<tau> \<preceq> qU" using Hcompat by (auto simp add: env_compat_def)
      have Hwf: "\<Gamma>2;\<Delta> \<Turnstile> OK" and "\<Gamma>1;\<Delta> \<Turnstile> OK" 
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heq2: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" and Heq1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" using wf_env_defines_without H1 H2 by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x)" by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x" by (auto simp add: insert_ident)
      have "(\<Gamma>1\<bar>x) \<sim> (\<Gamma>2\<bar>x)" using Hcompat by (auto simp add: without_def env_compat_def)
      hence "(\<Gamma>1\<bar>x),(y:\<tau>) \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using Hq by (auto simp add: env_compat_def)
      hence Ht: "(\<Gamma>1\<bar>x),(y:\<tau>)\<union>(\<Gamma>2\<bar>x),(y:\<tau>);\<Delta>;M;Z2\<union>Z1 \<turnstile> 
        vsubst_tm (TmFreeVar y) i e1 :=: vsubst_tm (TmFreeVar y) i e2 : \<tau>'"
        apply rule using Hd1 Hd2 H1 H2 Ht1i Ht2i H_eq Hwf_all Hqual by auto
      hence ?thesis using H_eq TmSwap HZ_eq by auto
    }
    moreover
    {
      assume H1: "x\<rhd>\<tau> \<in> \<Gamma>1" and H2: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      have Hwf: "\<Gamma>1;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heq1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" using wf_env_defines_without H1 H2 by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>1\<bar>x \<union> \<Gamma>2),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>1\<bar>x \<union> \<Gamma>2)" using H2 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>1\<bar>x \<union> \<Gamma>2" by (auto simp add: insert_ident)
      have "(\<Gamma>1\<bar>x) \<sim> \<Gamma>2" using Hcompat by (auto simp add: without_def env_compat_def)
      moreover
      have "{y\<rhd>\<tau>} \<sim> \<Gamma>2" using "1.prems" using env_compat_subset[of \<Gamma> "{y\<rhd>\<tau>}" \<Gamma>2] H_eq H2
        by (auto simp add: env_compat_def)
      ultimately
      have "(\<Gamma>1\<bar>x),(y:\<tau>) \<sim> \<Gamma>2" using env_compat_union_right[of "\<Gamma>1\<bar>x" \<Gamma>2 "{y\<rhd>\<tau>}"] by auto
      hence Ht: "(\<Gamma>1\<bar>x),(y:\<tau>)\<union>\<Gamma>2;\<Delta>;M;Z2\<union>Z1 \<turnstile> 
        vsubst_tm (TmFreeVar y) i e1 :=: vsubst_tm (TmFreeVar y) i e2 : \<tau>'"
        apply rule using Hd1 Hd2 H1 H2 Ht1i Ht2o H_eq Hwf_all Hqual by auto
      hence ?thesis using H_eq TmSwap HZ_eq by auto
    }
    moreover
    {
      assume H1: "x\<rhd>\<tau> \<notin> \<Gamma>1" and H2: "x\<rhd>\<tau> \<in> \<Gamma>2"
      have Hwf: "\<Gamma>2;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heq1: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" using wf_env_defines_without H1 H2 by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>1 \<union> (\<Gamma>2\<bar>x)),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>1 \<union> \<Gamma>2\<bar>x)" using H1 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>1 \<union> \<Gamma>2\<bar>x" by (auto simp add: insert_ident)
      have "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x)" using Hcompat by (auto simp add: without_def env_compat_def)
      moreover
      have "\<Gamma>1 \<sim> {y\<rhd>\<tau>}" using "1.prems" using env_compat_subset[of \<Gamma> "{y\<rhd>\<tau>}" \<Gamma>1] H_eq H1 by auto
      ultimately
      have "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using env_compat_union_right[of "\<Gamma>2\<bar>x" \<Gamma>1 "{y\<rhd>\<tau>}"] by (auto simp add: env_compat_def)
      hence Ht: "\<Gamma>1\<union>(\<Gamma>2\<bar>x),(y:\<tau>);\<Delta>;M;Z2\<union>Z1 \<turnstile> 
        vsubst_tm (TmFreeVar y) i e1 :=: vsubst_tm (TmFreeVar y) i e2 : \<tau>'"
        apply rule using Hd1 Hd2 H1 H2 Ht1o Ht2i H_eq Hwf_all Hqual by auto
      hence ?thesis using H_eq TmSwap HZ_eq by auto
    }
    moreover
    {
      assume H1: "x\<rhd>\<tau> \<notin> \<Gamma>1" and H2: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      hence "\<not> \<Gamma>1 defines x" and "\<not> \<Gamma>2 defines x" using Hdom1 Hdom2 by auto
      hence ?thesis using H\<Gamma>x_eq by (auto simp add: dom_def)
    }
    ultimately
    show ?thesis by auto
  next
    case (TmDeref e)
    hence "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> TmDeref (vsubst_tm (TmFreeVar x) i e) : \<tau>'" using "1.prems" by auto
    then obtain Z\<tau> Ze \<pi>
      where Ht: "\<Gamma>,(x:\<tau>);\<Delta>;M;Ze \<turnstile> vsubst_tm (TmFreeVar x) i e : Qual qU \<pi> (TyRef \<tau>')"
      and Hmin: "minimalValidScope (scopeT \<tau>') Z\<tau>"
      and HZ_eq: "Z = Ze \<union> Z\<tau>"
      and Hdollar: "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tm (TmFreeVar x) i e) \<longrightarrow> \<not> Z\<tau> defines \<rho>"
      and "qualT \<tau>' \<preceq> qU"
      using typing_deref_elim by auto
    have "\<Gamma>,(y:\<tau>);\<Delta>;M;Ze\<union>Z\<tau> \<turnstile> TmDeref (vsubst_tm (TmFreeVar y) i e) : \<tau>'" proof rule
      show "\<Gamma>,(y:\<tau>);\<Delta>;M;Ze \<turnstile> vsubst_tm (TmFreeVar y) i e : Qual qU \<pi> (TyRef \<tau>')" apply (rule IH)
        using prems by auto
    qed (insert prems free_dollars_vsubst_simpl, auto)
    thus ?thesis using HZ_eq TmDeref by auto
  next
    case (TmLoc q \<pi> l)
    hence Ht: "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> TmLoc q \<pi> l : \<tau>'" using "1.prems" by auto
    then obtain \<tau>r
      where H\<tau>r: "\<tau>' = Qual q \<pi> (TyRef \<tau>r)"
      and "\<Gamma>,(x:\<tau>);\<Delta> \<Turnstile> OK" and Hqual: "qualG \<Gamma>,(x:\<tau>) \<preceq> qU" and "M;\<Delta> \<Turnstile> OK" 
      and Hin: "(l\<rhd>\<tau>r) \<in> M" and Hmin: "minimalValidScope \<pi> Z"
      by auto
    have Ht: "\<Gamma>,(y:\<tau>);\<Delta>;M;Z \<turnstile> TmLoc q \<pi> l : Qual q \<pi> (TyRef \<tau>r)" apply rule 
      using prems by (auto simp add: qualGrelL_def)
    thus ?thesis using TmLoc H\<tau>r by auto
  next
    case (TmLetbangSc \<rho> e e1 e2)
    let ?es = "vsubst_tm (TmFreeVar x) i e"
    let ?e1s = "vsubst_tm (TmFreeVar x) i e1"
    let ?e2s = "vsubst_tm (TmFreeVar x) (Suc (Suc i)) e2"
    have Ht: "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> let$ \<rho> ?es ?e1s in ?e2s : \<tau>'" using 1 TmLetbangSc by auto
    then obtain \<Gamma>1 \<Gamma>2 \<Gamma>x Z1 Z2 Zx z \<tau>x \<tau>1 Zf Lx Ly
      where H\<Gamma>x_eq: "\<Gamma>,(x:\<tau>) = \<Gamma>x \<union> \<Gamma>1 \<union> \<Gamma>2"
      and Hcompat1: "\<Gamma>1 \<sim> \<Gamma>2" and Hcompat2: "\<Gamma>x \<sim> \<Gamma>2" and Hcompat3: "\<Gamma>x \<sim> \<Gamma>1"
      and Hwf: "\<Gamma>x \<union> \<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK"
      and HZ_eq: "Z = Z2 \<union> Z1 \<union> Zx"
      and Ht\<rho>: "\<not> (Z2\<union>Z1) defines \<rho>"
      and HZ1: "Z1 \<Turnstile> ScopeSet"
      and HZ2: "Z2 \<Turnstile> ScopeSet"
      and Hvalid: "validScope (scopeT \<tau>1) Z1"
      and Hbang: "isBangable \<tau>x"
      and Htyp: "\<Gamma>x;\<Delta>;M;Zx \<turnstile> TmFreeVar z : convT \<rho> \<tau>x"
      and Hxnotin: "\<not> (\<Gamma>1\<union>\<Gamma>2) defines z"
      and Htyp1: "\<Gamma>1,(z:convT \<rho> \<tau>x); \<Delta>; M; Zf \<turnstile> ?e1s : \<tau>1"
      and Htyp2: "\<forall>z w. \<Gamma>2,(z:\<tau>x),(w:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {z\<rhd>\<tau>x},(w:\<tau>1) \<and> {z\<rhd>\<tau>x} \<sim> {w\<rhd>\<tau>1} \<and> 
                  z \<notin> Lx \<and> w \<notin> Ly \<and> z \<noteq> w \<longrightarrow>
                  \<Gamma>2,(z:\<tau>x),(w:\<tau>1); \<Delta>;M;Z2 \<turnstile> 
                  vsubst_tm (TmFreeVar w) (Suc 0) (vsubst_tm (TmFreeVar z) 0 ?e2s) : \<tau>'"
      and Hdollar1: "\<forall>\<rho>. \<rho> is free dollar sc in ?e1s \<longrightarrow> \<not> (Zx\<union>Z2) defines \<rho>"
      and Hdollar2: "\<forall>\<rho>. \<rho> is free dollar sc in ?e2s \<longrightarrow> \<not> (Zx\<union>Z1) defines \<rho>"
      and Hx_eq: "?es = TmFreeVar z"
      and HZf: "Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>)"
      and Hlx: "finite Lx"
      and Hly: "finite Ly"
      using typing_letsc_elim[of "\<Gamma>,(x:\<tau>)" \<Delta> M Z \<rho> ?es ?e1s ?e2s \<tau>']
      by auto;
    {
      assume Hin: "x\<rhd>\<tau> \<in> \<Gamma>1"
      moreover
      have Hwf: "\<Gamma>1;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta> \<Gamma>1] by auto
      ultimately
      have Hg1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" using wf_env_defines_without by auto 
      have Hneq: "x \<noteq> z" proof 
        assume "x = z"
        hence "\<Gamma>1 defines z" using Hin by (auto simp add: dom_def)
        thus False using Hxnotin by (auto simp add: dom_def)
      qed
      have Hy_eq: "y \<noteq> z \<or> (y = z \<and> \<tau> = convT \<rho> \<tau>x)" proof (cases "y = z")
        assume Heq: "y = z"
        have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>x" using Htyp by auto
        hence "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>,(y:\<tau>)" using H\<Gamma>x_eq Hneq by auto
        hence "convT \<rho> \<tau>x = \<tau>" using Heq "1.prems" wf_env_sanity[of "\<Gamma>,(y:\<tau>)" \<Delta> y] by auto
        thus ?thesis using Heq by auto
      qed auto
      have Hnotin: "\<not> (\<Gamma>1\<bar>x) defines x" by (auto simp add: without_def dom_def)
      have Ht1: "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(y:\<tau>);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar y) i e1 : \<tau>1" proof (rule IH)
        have "(\<Gamma>1\<bar>x),(x:\<tau>),(z:convT \<rho> \<tau>x);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar x) i e1 : \<tau>1"
          using Htyp1 Hg1 by auto
        also have "(\<Gamma>1\<bar>x),(x:\<tau>),(z:convT \<rho> \<tau>x) = (\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(x:\<tau>)" by auto
        finally show "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(x:\<tau>);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar x) i e1 : \<tau>1" by auto
      next
        have "(\<Gamma>1\<bar>x),(x:\<tau>);\<Delta> \<Turnstile> OK" using Hg1 Hwf by auto
        moreover
        have "\<Delta> \<Turnstile>' convT \<rho> \<tau>x" using Htyp poly_ty_wf_type by auto
        moreover
        have "\<not> (\<Gamma>1\<bar>x),(x:\<tau>) defines z" using Hxnotin Hneq by (auto simp add: dom_def without_def)
        ultimately
        have "(\<Gamma>1\<bar>x),(x:\<tau>),(z:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK" using wf_env_cons[where x=z] env_permute by auto
        thus "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(x:\<tau>);\<Delta> \<Turnstile> OK" by (auto simp add: env_permute)
      next
        show "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x) \<sim> {x\<rhd>\<tau>}" using Hneq by (auto simp add: env_compat_def dom_def without_def)
      next
        have "\<Gamma>1\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        hence Hwf: "(\<Gamma>1\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using "1.prems" wf_env_subset[of "\<Gamma>,(y:\<tau>)" \<Delta> "(\<Gamma>1\<bar>x),(y:\<tau>)"] by auto
        show "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using Hy_eq proof rule
          assume Hneq: "y \<noteq> z"
          have "\<Delta> \<Turnstile>' convT \<rho> \<tau>x" using Htyp poly_ty_wf_type by auto
          moreover
          have "\<not> (\<Gamma>1\<bar>x),(y:\<tau>) defines z" using Hxnotin Hneq by (auto simp add: dom_def without_def)
          ultimately
          have "(\<Gamma>1\<bar>x),(y:\<tau>),(z:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK" using wf_env_cons Hwf by auto
          thus ?thesis by (auto simp add: env_permute)
        next
          assume "y = z \<and> \<tau> = convT \<rho> \<tau>x"
          hence "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(y:\<tau>) = (\<Gamma>1\<bar>x),(y:\<tau>)" by auto
          thus ?thesis using Hwf by auto
        qed
      next
        have "\<Gamma>1\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        hence "\<Gamma>1\<bar>x \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset[of \<Gamma> "{y\<rhd>\<tau>}" "(\<Gamma>1\<bar>x)"] by auto
        moreover
        have "{z\<rhd>convT \<rho> \<tau>x} \<sim> {y\<rhd>\<tau>}" by (cases \<tau>x, auto simp add: env_compat_def) 
        ultimately
        show "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x) \<sim> {y\<rhd>\<tau>}" using env_compat_union_right[of "\<Gamma>1\<bar>x" "{y\<rhd>\<tau>}" "{z\<rhd>convT \<rho> \<tau>x}"] by auto
      next
        show "\<not> x free in e1" using "1.prems" TmLetbangSc by auto
      next
        show "\<not> (\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x) defines x" using Hnotin Hneq by auto
      qed (insert TmLetbangSc "1.prems", auto)
    }
    hence Ht1i: "x\<rhd>\<tau> \<in> \<Gamma>1 \<longrightarrow> (\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(y:\<tau>);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar y) i e1 : \<tau>1"
      by auto
    have Hdom1: "x\<rhd>\<tau> \<notin> \<Gamma>1 \<longrightarrow> \<not> \<Gamma>1 defines x" proof (auto simp add: dom_def)
      fix b
      assume "(x, b) \<in> \<Gamma>1"
      and Hne: "x\<rhd>\<tau> \<notin> \<Gamma>1"
      then obtain \<tau>' where "b = Term \<tau>'" and Hin: "x\<rhd>\<tau>' \<in> \<Gamma>1" by (cases b, auto)
      hence "x\<rhd>\<tau>' \<in> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
      hence "\<tau>' = \<tau>" using Ht ty_wf_env wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x \<tau> \<tau>'] by auto
      thus False using Hne Hin by auto
    qed
    have Hdom2: "x\<rhd>\<tau> \<notin> \<Gamma>2 \<longrightarrow> \<not> \<Gamma>2 defines x" proof (auto simp add: dom_def)
      fix b
      assume "(x, b) \<in> \<Gamma>2"
      and Hne: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      then obtain \<tau>' where "b = Term \<tau>'" and Hin: "x\<rhd>\<tau>' \<in> \<Gamma>2" by (cases b, auto)
      hence "x\<rhd>\<tau>' \<in> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
      hence "\<tau>' = \<tau>" using Ht ty_wf_env wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x \<tau> \<tau>'] by auto
      thus False using Hne Hin by auto
    qed
    have Ht1o_eq: "x\<rhd>\<tau> \<notin> \<Gamma>1 \<longrightarrow> x=z \<longrightarrow> \<Gamma>1,(y:convT \<rho> \<tau>x);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar y) i e1 : \<tau>1 \<and> e = TmVar i"
    proof (rule, rule)
      assume Heq: "x=z"
        and Hnin: "x\<rhd>\<tau> \<notin> \<Gamma>1" 
      have "\<not> x free in e" using "1.prems" TmLetbangSc by auto
      hence Hg1: "e = TmVar i" using Hx_eq Heq proof (cases e)
        case (TmVar n)
        hence "vsubst_tm (TmFreeVar z) i (TmVar n) = TmFreeVar z" using Heq Hx_eq by auto
        hence "i = n" by (cases "i=n", auto)
        thus ?thesis by auto
      qed auto
      have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>x" using Htyp by auto
      hence "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
      hence Hteq: "convT \<rho> \<tau>x = \<tau>" using Heq "1.prems" wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x] by auto
      have Ht: "\<Gamma>1,(x:convT \<rho> \<tau>x);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar x) i e1 : \<tau>1" using Htyp1 Heq by auto
      have "\<Gamma>1,(y:convT \<rho> \<tau>x);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar y) i e1 : \<tau>1" proof (rule IH)
        show "\<Gamma>1,(x:convT \<rho> \<tau>x);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar x) i e1 : \<tau>1" using Ht by auto
      next
        show "\<Gamma>1,(x:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK" using Htyp1 ty_wf_env Heq by auto
      next
        show "\<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>x}" by (cases \<tau>x, auto simp add: env_compat_def)
      next
        show "\<Gamma>1 \<sim> {y\<rhd>convT \<rho> \<tau>x}" by (cases \<tau>x, auto simp add: env_compat_def)
      next
        show "\<not> \<Gamma>1 defines x" using Hdom1 Hnin by auto
      next
        have "\<Gamma>,(y:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK" using "1.prems" Hteq by auto
        moreover
        have "\<Gamma>1 \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
        hence "\<Gamma>1 \<subseteq> \<Gamma>" using Hnin by auto
        hence "\<Gamma>1,(y:convT \<rho> \<tau>x) \<subseteq> \<Gamma>,(y:convT \<rho> \<tau>x)" by auto
        ultimately
        show "\<Gamma>1,(y:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK" using wf_env_subset by auto
      qed (insert Hteq "1.prems" TmLetbangSc, auto)
      thus "\<Gamma>1,(y:convT \<rho> \<tau>x);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar y) i e1 : \<tau>1 \<and> e = TmVar i" using Hg1 by auto
    qed
    have Ht1o_neq: "x\<rhd>\<tau> \<notin> \<Gamma>1 \<longrightarrow> x\<noteq>z \<longrightarrow> \<Gamma>1,(z:convT \<rho> \<tau>x);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar y) i e1 : \<tau>1"
      using Htyp1 not_in_env_ty_same_subst[of x "\<Gamma>1,(z:convT \<rho> \<tau>x)"] Hdom1 by (auto simp add: dom_def)
    {
      assume Hin: "x\<rhd>\<tau> \<in> \<Gamma>2"
      moreover
      have Hwf': "\<Gamma>2;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta> \<Gamma>2] by auto
      ultimately
      have Hg1: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" using wf_env_defines_without by auto 
      have Hneq: "x \<noteq> z" proof 
        assume "x = z"
        hence "\<Gamma>2 defines z" using Hin by (auto simp add: dom_def)
        thus False using Hxnotin by (auto simp add: dom_def)
      qed
      have Hnotin: "\<not> (\<Gamma>2\<bar>x) defines x" by (auto simp add: without_def dom_def)
      have Ht2: "\<forall>w q. (\<Gamma>2\<bar>x),(y:\<tau>),(w:\<tau>x),(q:\<tau>1);\<Delta> \<Turnstile> OK \<and> (\<Gamma>2\<bar>x),(y:\<tau>) \<sim> {w\<rhd>\<tau>x},(q:\<tau>1) \<and>
                       {w\<rhd>\<tau>x} \<sim> {q\<rhd>\<tau>1} \<and> w \<notin> (Lx\<union>{x}\<union>{y}) \<and> q \<notin> (Ly\<union>{x}\<union>{y}) \<and> w \<noteq> q \<longrightarrow>
                 (\<Gamma>2\<bar>x),(y:\<tau>),(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar q) (Suc 0) 
                                                      (freshen_tm w (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'" 
      proof -
        {
          fix w q
          assume Hwf_h: "(\<Gamma>2\<bar>x),(y:\<tau>),(w:\<tau>x),(q:\<tau>1);\<Delta> \<Turnstile> OK" and Hcomp1: "(\<Gamma>2\<bar>x),(y:\<tau>) \<sim> {w\<rhd>\<tau>x},(q:\<tau>1)"
            and Hcomp2: "{w\<rhd>\<tau>x} \<sim> {q\<rhd>\<tau>1}" and Hw: "w \<notin> (Lx\<union>{x}\<union>{y})" and Hq: "q \<notin> (Ly\<union>{x}\<union>{y})" and Hneq: "w \<noteq> q"
          moreover
          have Hnin: "\<not> (\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1) defines x" using Hw Hq Hnotin by (auto simp add: dom_def)
          have Hwft: "\<Delta> \<Turnstile>' \<tau>" using poly_env_wf_ty Hwf' Hin by auto
          have "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1),(x:\<tau>);\<Delta> \<Turnstile> OK" apply (rule wf_env_cons[of "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1)" \<Delta> x \<tau>])
            using Hwf_h Hnin Hwft wf_env_subset[of "(\<Gamma>2\<bar>x),(y:\<tau>),(w:\<tau>x),(q:\<tau>1)" \<Delta> "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1)"] by auto
          hence Hwf: "\<Gamma>2,(w:\<tau>x),(q:\<tau>1);\<Delta> \<Turnstile> OK" using Hg1 by (auto simp add: env_permute)
          moreover
          have  "(\<Gamma>2\<bar>x),(x:\<tau>) \<sim> {w\<rhd>\<tau>x},(q:\<tau>1)" using Hcomp1 Hw Hq by (auto simp add: env_compat_def dom_def)
          hence Hcomp1: "\<Gamma>2 \<sim> {w\<rhd>\<tau>x},(q:\<tau>1)" using Hg1 by auto
          ultimately
          have Ht: "(\<Gamma>2\<bar>x),(x:\<tau>),(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar q) (Suc 0) 
                                                     (freshen_tm w (vsubst_tm (TmFreeVar x) (Suc (Suc i)) e2)) : \<tau>'" 
            using Hg1 Htyp2 by auto
          have "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1),(y:\<tau>); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) 
                                                     (freshen_tm w  e2)) : \<tau>'"
          proof (rule IH)
            note Ht
            also 
            have "vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w (vsubst_tm (TmFreeVar x) (Suc (Suc i)) e2)) =
                  vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (freshen_tm w  e2))"
              using subst_subst_swap[of "Suc (Suc i)" 0] by auto
            also
            have "vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (freshen_tm w  e2)) =
                  vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w  e2))"
              using subst_subst_swap[of "Suc (Suc i)" "Suc 0"] by auto;
            finally
            show "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1),(x:\<tau>);\<Delta>;M;Z2 \<turnstile> 
              vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w e2)) : \<tau>'"
              by (auto simp add: env_permute)
          next
            show "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1),(x:\<tau>);\<Delta> \<Turnstile> OK"
              using Hwf Hg1 by (auto simp add: env_permute)
          next
            show "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1) \<sim> {x\<rhd>\<tau>}"
              using Hq Hw Hnotin by (auto simp add: env_compat_def dom_def)
          next
            show "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1),(y:\<tau>);\<Delta> \<Turnstile> OK" using Hwf_h by (auto simp add: env_permute)
          next
            have "\<Gamma>2\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq Hin by (auto simp add: without_def)
            hence "\<Gamma>2\<bar>x \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto
            thus "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1) \<sim> {y\<rhd>\<tau>}" using Hq Hw by (auto simp add: env_compat_def)
          next
            show "x \<notin> dom ((\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1))" using Hnotin Hq Hw by (auto simp add: dom_def)
          next
            show "x \<notin> set (free_tm (vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w e2)))"
              using not_free_subst Hq Hw "1.prems" TmLetbangSc by auto
          next
            show "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w e2) \<longrightarrow> scopeT \<tau> \<noteq> ScFree \<rho>"
              using "1.prems" TmLetbangSc free_dollars_vsubst_simpl by auto
          qed (insert TmLetbangSc, auto)
          also
          have "vsubst_tm (TmFreeVar y) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w  e2)) =
                vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar y) (Suc (Suc i)) (freshen_tm w  e2))"
            using subst_subst_swap[of "Suc (Suc i)" "Suc 0"] by auto;
          also 
          have "vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar y) (Suc (Suc i)) (freshen_tm w  e2)) =
                vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2))" 
            using subst_subst_swap[of "Suc (Suc i)" 0] by auto
          finally
          have "(\<Gamma>2\<bar>x),(y:\<tau>),(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar q) (Suc 0) 
                                                     (freshen_tm w (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
            by (auto simp add: env_permute)
        }
        thus ?thesis by auto
      qed
    }
    hence Ht2l: "x\<rhd>\<tau> \<in> \<Gamma>2 \<longrightarrow> (\<forall>w q. (\<Gamma>2\<bar>x),(y:\<tau>),(w:\<tau>x),(q:\<tau>1);\<Delta> \<Turnstile> OK \<and> (\<Gamma>2\<bar>x),(y:\<tau>) \<sim> {w\<rhd>\<tau>x},(q:\<tau>1) \<and>
                         {w\<rhd>\<tau>x} \<sim> {q\<rhd>\<tau>1} \<and> w \<notin> (Lx\<union>{x}\<union>{y}) \<and> q \<notin> (Ly\<union>{x}\<union>{y}) \<and> w \<noteq> q \<longrightarrow>
                           (\<Gamma>2\<bar>x),(y:\<tau>),(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar q) (Suc 0) 
                                                      (freshen_tm w (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>')"
      by auto
    {
      fix w q
      assume Hwf: "\<Gamma>2,(w:\<tau>x),(q:\<tau>1);\<Delta> \<Turnstile> OK" and Hcomp1: "\<Gamma>2 \<sim> {w\<rhd>\<tau>x},(q:\<tau>1)" and Hcomp2: "{w\<rhd>\<tau>x} \<sim> {q\<rhd>\<tau>1}" 
        and Hw: "w \<notin> (Lx\<union>{x}\<union>{y})" and Hq: "q \<notin> (Ly\<union>{x}\<union>{y})" and Hneq: "w \<noteq> q"
        and Hnin: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      hence Ht: "\<Gamma>2,(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar q) (Suc 0) 
                                                     (freshen_tm w (vsubst_tm (TmFreeVar x) (Suc (Suc i)) e2)) : \<tau>'" 
        using Htyp2 by auto
      also 
      have "vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w (vsubst_tm (TmFreeVar x) (Suc (Suc i)) e2)) =
            vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (freshen_tm w  e2))"
        using subst_subst_swap[of "Suc (Suc i)" 0] by auto
      also
      have "vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (freshen_tm w  e2)) =
            vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w  e2))"
        using subst_subst_swap[of "Suc (Suc i)" "Suc 0"] by auto
      finally
      have "\<Gamma>2,(w:\<tau>x),(q:\<tau>1);\<Delta>;M;Z2 \<turnstile> 
           vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w e2)) : \<tau>'" by auto
      moreover
      have "\<not> \<Gamma>2,(w:\<tau>x),(q:\<tau>1) defines x" using Hnin Hdom2 Hq Hw by auto
      ultimately
      have "\<Gamma>2,(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) 
          (freshen_tm w e2)) : \<tau>'"
        using not_in_env_ty_same_subst[of x "\<Gamma>2,(w:\<tau>x),(q:\<tau>1)"] by (auto simp add: dom_def)
      also
      have "vsubst_tm (TmFreeVar y) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w  e2)) =
            vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar y) (Suc (Suc i)) (freshen_tm w  e2))"
        using subst_subst_swap[of "Suc (Suc i)" "Suc 0"] by auto
      also 
      have "vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar y) (Suc (Suc i)) (freshen_tm w  e2)) =
            vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2))" 
        using subst_subst_swap[of "Suc (Suc i)" 0] by auto
      finally 
      have "\<Gamma>2,(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar q) (Suc 0) 
          (freshen_tm w (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'" by auto
    }
    hence Ht2o: "x\<rhd>\<tau> \<notin> \<Gamma>2 \<longrightarrow> (\<forall>w q. \<Gamma>2,(w:\<tau>x),(q:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {w\<rhd>\<tau>x},(q:\<tau>1) \<and>
                         {w\<rhd>\<tau>x} \<sim> {q\<rhd>\<tau>1} \<and> w \<notin> (Lx\<union>{x}\<union>{y}) \<and> q \<notin> (Ly\<union>{x}\<union>{y}) \<and> w \<noteq> q \<longrightarrow>
                           \<Gamma>2,(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar q) (Suc 0) 
                                                      (freshen_tm w (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>')"
      by auto
    have Hd1: "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e1 \<longrightarrow>  \<rho> \<notin> dom (Zx \<union> Z2)"
      using Hdollar1 free_dollars_vsubst_simpl by auto
    have Hd2: "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2 \<longrightarrow> \<rho> \<notin> dom (Zx \<union> Z1)"
      using Hdollar2 free_dollars_vsubst_simpl by auto
    have Hdomx: "x\<rhd>\<tau> \<notin> \<Gamma>x \<longrightarrow> \<not> \<Gamma>x defines x" proof (auto simp add: dom_def)
      fix b
      assume "(x, b) \<in> \<Gamma>x"
      and Hne: "x\<rhd>\<tau> \<notin> \<Gamma>x"
      then obtain \<tau>' where "b = Term \<tau>'" and Hin: "x\<rhd>\<tau>' \<in> \<Gamma>x" by (cases b, auto)
      hence "x\<rhd>\<tau>' \<in> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
      hence "\<tau>' = \<tau>" using Ht ty_wf_env wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x \<tau> \<tau>'] by auto
      thus False using Hne Hin by auto
    qed
    have "\<not> x free in e" using "1.prems" TmLetbangSc by auto
    hence Hxz: "(x = z \<and> e = TmVar i) \<or> (x \<noteq> z \<and> e = TmFreeVar z)" using Hx_eq proof (cases e, auto) 
      fix n
      assume "(if i = n then TmFreeVar x else TmVar n) = TmFreeVar z"
      thus "x = z" by (cases "i=n", auto)
    next
      fix n
      assume "(if i = n then TmFreeVar x else TmVar n) = TmFreeVar z"
      thus "n = i" by (cases "i=n", auto)
    qed
    {
      assume H: "x\<rhd>\<tau> \<notin> \<Gamma>x" and H1: "x\<rhd>\<tau> \<in> \<Gamma>1" and H2: "x\<rhd>\<tau> \<in> \<Gamma>2"
      have Hy_eq: "y \<noteq> z" proof (cases "y = z")
        assume Heq: "y = z"
        have "\<Gamma>x \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq H by auto
        hence "\<Gamma>x \<subseteq> \<Gamma>" using H by auto
        moreover
        have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>x" using Htyp by auto
        ultimately
        have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>,(y:\<tau>)" by auto
        hence "convT \<rho> \<tau>x = \<tau>" using Heq "1.prems" wf_env_sanity[of "\<Gamma>,(y:\<tau>)" \<Delta> y] by auto
        hence "scopeT \<tau> = ScFree \<rho>" by (cases \<tau>x, auto)
        hence False using "1.prems" TmLetbangSc by auto
        thus "y \<noteq> z" by auto
      qed auto
      have Hwf: "\<Gamma>2;\<Delta> \<Turnstile> OK" and "\<Gamma>1;\<Delta> \<Turnstile> OK" 
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heq2: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" and Heq1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" using wf_env_defines_without H1 H2 by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x)" using H by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x" by (auto simp add: insert_ident)
      have "(\<Gamma>x \<union> ((\<Gamma>1\<bar>x),(y:\<tau>)) \<union> ((\<Gamma>2\<bar>x),(y:\<tau>)));\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let$ \<rho> (vsubst_tm (TmFreeVar y) i e)
        (vsubst_tm (TmFreeVar y) i e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'"
        using Hxz proof rule
        assume "x = z \<and> e = TmVar i"
        hence Hxz_1: "x = z" and Hxz_2: "e = TmVar i" by auto
        hence "\<Gamma>x;\<Delta>;M;Zx \<turnstile> TmFreeVar x : convT \<rho> \<tau>x" using Htyp by auto
        hence False using Hdomx H by (auto simp add: dom_def)
        thus ?thesis by auto
      next
        assume "x \<noteq> z \<and> e = TmFreeVar z"
        hence Hxz_1: "x \<noteq> z" and Hxz_2: "e = TmFreeVar z" by auto
        have "(\<Gamma>1\<bar>x),(x:\<tau>) \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq1 Heq2 Hcompat1 by auto
        hence Hqual: "qualT \<tau> \<preceq> qU" by (auto simp add: env_compat_def)
        have "(\<Gamma>x \<union> ((\<Gamma>1\<bar>x),(y:\<tau>)) \<union> ((\<Gamma>2\<bar>x),(y:\<tau>)));\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let$ \<rho> (TmFreeVar z)
              (vsubst_tm (TmFreeVar y) i e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'" proof rule
          show "\<not> (Z2\<union>Z1) defines \<rho>"
            and "Z1 \<Turnstile> ScopeSet"
            and "Z2 \<Turnstile> ScopeSet"
            and "validScope (scopeT \<tau>1) Z1"
            and "isBangable \<tau>x"
            and "Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>)"
            and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e1 \<longrightarrow>  \<rho> \<notin> dom (Zx \<union> Z2)"
            and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2 \<longrightarrow> \<rho> \<notin> dom (Zx \<union> Z1)"
            and "finite (Lx\<union>{x}\<union>{y})"
            and "finite (Ly\<union>{x}\<union>{y})"
            using Ht\<rho> HZ1 HZ2 Hvalid Hbang HZf Hd1 Hd2 Hlx Hly by auto
        next          
          have "y \<noteq> z" using Hy_eq Hxz_1 by auto
          thus "z \<notin> dom ((\<Gamma>1\<bar>x),(y:\<tau>) \<union> (\<Gamma>2 \<bar> x),(y:\<tau>))" using Hxnotin by (auto simp add: dom_def without_def)
        next
          show "\<Gamma>x;\<Delta>;M;Zx \<turnstile> TmFreeVar z : convT \<rho> \<tau>x" using Htyp by auto
        next
          show "(\<Gamma>1\<bar>x),(y:\<tau>),(z:convT \<rho> \<tau>x);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar y) i e1 : \<tau>1" 
            using Ht1i H1 by (auto simp add: env_permute)
        next
          note Ht2l
          thus "\<forall>xa ya. (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta> \<Turnstile> OK \<and> (\<Gamma>2\<bar>x),(y:\<tau>) \<sim> {xa\<rhd>\<tau>x},(ya:\<tau>1) \<and> {xa\<rhd>\<tau>x} \<sim> {ya\<rhd>\<tau>1} \<and>
                        xa \<notin> Lx \<union> {x} \<union> {y} \<and> ya \<notin> Ly \<union> {x} \<union> {y} \<and> xa \<noteq> ya \<longrightarrow>
                (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar ya) (Suc 0)
                                                      (freshen_tm xa (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
            using H2 by auto
        next
          have "(\<Gamma>1\<bar>x),(x:\<tau>) \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq1 Heq2 Hcompat1 by auto
          thus "(\<Gamma>1\<bar>x),(y:\<tau>) \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" by (auto simp add: env_compat_def)
        next
          have "\<Gamma>x \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq2 Hcompat2 by auto
          thus "\<Gamma>x \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using Hqual by (auto simp add: env_compat_def)
        next
          have "\<Gamma>x \<sim> (\<Gamma>1\<bar>x),(x:\<tau>)" using Heq1 Hcompat3 by auto
          thus "\<Gamma>x \<sim> (\<Gamma>1\<bar>x),(y:\<tau>)" using Hqual by (auto simp add: env_compat_def)
        next
          show "\<Gamma>x \<union> (\<Gamma>1\<bar>x),(y:\<tau>) \<union> (\<Gamma>2\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using prems H_eq by auto
        qed
        thus ?thesis using Hxz_2 by auto
      qed
      hence ?thesis using H_eq HZ_eq TmLetbangSc by auto
    }
    moreover
    {
      assume H: "x\<rhd>\<tau> \<notin> \<Gamma>x" and H1: "x\<rhd>\<tau> \<in> \<Gamma>1" and H2: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      have Hy_eq: "y \<noteq> z" proof (cases "y = z")
        assume Heq: "y = z"
        have "\<Gamma>x \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq H by auto
        hence "\<Gamma>x \<subseteq> \<Gamma>" using H by auto
        moreover
        have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>x" using Htyp by auto
        ultimately
        have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>,(y:\<tau>)" by auto
        hence "convT \<rho> \<tau>x = \<tau>" using Heq "1.prems" wf_env_sanity[of "\<Gamma>,(y:\<tau>)" \<Delta> y] by auto
        hence "scopeT \<tau> = ScFree \<rho>" by (cases \<tau>x, auto)
        hence False using "1.prems" TmLetbangSc by auto
        thus "y \<noteq> z" by auto
      qed auto
      have "\<Gamma>1;\<Delta> \<Turnstile> OK" 
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heq1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" using wf_env_defines_without H1 H2 by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2)" using H H2 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2" by (auto simp add: insert_ident)
      have "(\<Gamma>x \<union> ((\<Gamma>1\<bar>x),(y:\<tau>)) \<union> (\<Gamma>2,(y:\<tau>)));\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let$ \<rho> (vsubst_tm (TmFreeVar y) i e)
        (vsubst_tm (TmFreeVar y) i e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'"
        using Hxz proof rule
        assume "x = z \<and> e = TmVar i"
        hence Hxz_1: "x = z" and Hxz_2: "e = TmVar i" by auto
        hence "\<Gamma>x;\<Delta>;M;Zx \<turnstile> TmFreeVar x : convT \<rho> \<tau>x" using Htyp by auto
        hence False using Hdomx H by (auto simp add: dom_def)
        thus ?thesis by auto
      next
        assume "x \<noteq> z \<and> e = TmFreeVar z"
        hence Hxz_1: "x \<noteq> z" and Hxz_2: "e = TmFreeVar z" by auto
        have "(\<Gamma>x \<union> ((\<Gamma>1\<bar>x),(y:\<tau>)) \<union> \<Gamma>2);\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let$ \<rho> (TmFreeVar z)
              (vsubst_tm (TmFreeVar y) i e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'" proof rule
          show "\<not> (Z2\<union>Z1) defines \<rho>"
            and "Z1 \<Turnstile> ScopeSet"
            and "Z2 \<Turnstile> ScopeSet"
            and "validScope (scopeT \<tau>1) Z1"
            and "isBangable \<tau>x"
            and "Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>)"
            and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e1 \<longrightarrow>  \<rho> \<notin> dom (Zx \<union> Z2)"
            and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2 \<longrightarrow> \<rho> \<notin> dom (Zx \<union> Z1)"
            and "finite (Lx\<union>{x}\<union>{y})"
            and "finite (Ly\<union>{x}\<union>{y})"
            using Ht\<rho> HZ1 HZ2 Hvalid Hbang HZf Hd1 Hd2 Hlx Hly by auto
        next          
          have "y \<noteq> z" using Hy_eq Hxz_1 by auto
          thus "z \<notin> dom ((\<Gamma>1\<bar>x),(y:\<tau>) \<union> \<Gamma>2)" using Hxnotin by (auto simp add: dom_def without_def)
        next
          show "\<Gamma>x;\<Delta>;M;Zx \<turnstile> TmFreeVar z : convT \<rho> \<tau>x" using Htyp by auto
        next
          show "(\<Gamma>1\<bar>x),(y:\<tau>),(z:convT \<rho> \<tau>x);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar y) i e1 : \<tau>1" 
            using Ht1i H1 by (auto simp add: env_permute)
        next
          note Ht2o
          thus "\<forall>xa ya. \<Gamma>2,(xa:\<tau>x),(ya:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {xa\<rhd>\<tau>x},(ya:\<tau>1) \<and> {xa\<rhd>\<tau>x} \<sim> {ya\<rhd>\<tau>1} \<and>
                        xa \<notin> Lx \<union> {x} \<union> {y} \<and> ya \<notin> Ly \<union> {x} \<union> {y} \<and> xa \<noteq> ya \<longrightarrow>
                \<Gamma>2,(xa:\<tau>x),(ya:\<tau>1);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar ya) (Suc 0)
                                                      (freshen_tm xa (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
            using H2 by auto
        next
          have "(\<Gamma>1\<bar>x),(x:\<tau>) \<sim> \<Gamma>2" using Heq1 Hcompat1 by auto
          hence Hc1: "(\<Gamma>1\<bar>x) \<sim> \<Gamma>2" using env_compat_subset by auto
          have "\<Gamma>2 \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
          hence "\<Gamma>2 \<subseteq> \<Gamma>" using H2 by auto
          hence "\<Gamma>2 \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto 
          hence "{y\<rhd>\<tau>} \<sim> \<Gamma>2" using env_compat_symm by auto
          thus "(\<Gamma>1\<bar>x),(y:\<tau>) \<sim> \<Gamma>2" using Hc1 env_compat_union_right[of "(\<Gamma>1\<bar>x)" _ "{y\<rhd>\<tau>}"] by auto
        next
          show "\<Gamma>x \<sim> \<Gamma>2" by assumption
        next
          have "\<Gamma>x \<sim> (\<Gamma>1\<bar>x),(x:\<tau>)" using Heq1 Hcompat3 by auto
          hence Hc: "\<Gamma>x \<sim> (\<Gamma>1\<bar>x)" using env_compat_subset env_compat_symm by blast
          have "\<Gamma>x \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
          hence "\<Gamma>x \<subseteq> \<Gamma>" using H by auto
          hence "\<Gamma>x \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto
          thus "\<Gamma>x \<sim> (\<Gamma>1\<bar>x),(y:\<tau>)" using Hc env_compat_union_left[of _ "(\<Gamma>1\<bar>x)" "{y\<rhd>\<tau>}"] by auto
        next
          show "\<Gamma>x \<union> (\<Gamma>1\<bar>x),(y:\<tau>) \<union> \<Gamma>2;\<Delta> \<Turnstile> OK" using prems H_eq by auto
        qed
        thus ?thesis using Hxz_2 by auto
      qed
      hence ?thesis using H_eq HZ_eq TmLetbangSc by auto
    }
    moreover
    {
      assume H: "x\<rhd>\<tau> \<notin> \<Gamma>x" and H1: "x\<rhd>\<tau> \<notin> \<Gamma>1" and H2: "x\<rhd>\<tau> \<in> \<Gamma>2"
      have Hy_eq: "y \<noteq> z" proof (cases "y = z")
        assume Heq: "y = z"
        have "\<Gamma>x \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq H by auto
        hence "\<Gamma>x \<subseteq> \<Gamma>" using H by auto
        moreover
        have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>x" using Htyp by auto
        ultimately
        have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>,(y:\<tau>)" by auto
        hence "convT \<rho> \<tau>x = \<tau>" using Heq "1.prems" wf_env_sanity[of "\<Gamma>,(y:\<tau>)" \<Delta> y] by auto
        hence "scopeT \<tau> = ScFree \<rho>" by (cases \<tau>x, auto)
        hence False using "1.prems" TmLetbangSc by auto
        thus "y \<noteq> z" by auto
      qed auto
      have "\<Gamma>2;\<Delta> \<Turnstile> OK" 
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heq2: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" using wf_env_defines_without H1 H2 by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>x \<union> \<Gamma>1 \<union> \<Gamma>2\<bar>x),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>x \<union> \<Gamma>1 \<union> \<Gamma>2\<bar>x)" using H H1 H2 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>x \<union> \<Gamma>1 \<union> \<Gamma>2\<bar>x" by (auto simp add: insert_ident)
      have "(\<Gamma>x \<union> \<Gamma>1 \<union> ((\<Gamma>2\<bar>x),(y:\<tau>)));\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let$ \<rho> (vsubst_tm (TmFreeVar y) i e)
        (vsubst_tm (TmFreeVar y) i e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'"
        using Hxz proof rule
        assume "x = z \<and> e = TmVar i"
        hence Hxz_1: "x = z" and Hxz_2: "e = TmVar i" by auto
        hence "\<Gamma>x;\<Delta>;M;Zx \<turnstile> TmFreeVar x : convT \<rho> \<tau>x" using Htyp by auto
        hence False using Hdomx H by (auto simp add: dom_def)
        thus ?thesis by auto
      next
        assume "x \<noteq> z \<and> e = TmFreeVar z"
        hence Hxz_1: "x \<noteq> z" and Hxz_2: "e = TmFreeVar z" by auto
        have "(\<Gamma>x \<union> \<Gamma>1 \<union>((\<Gamma>2\<bar>x),(y:\<tau>)));\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let$ \<rho> (TmFreeVar z)
              (vsubst_tm (TmFreeVar y) i e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'" proof rule
          show "\<not> (Z2\<union>Z1) defines \<rho>"
            and "Z1 \<Turnstile> ScopeSet"
            and "Z2 \<Turnstile> ScopeSet"
            and "validScope (scopeT \<tau>1) Z1"
            and "isBangable \<tau>x"
            and "Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>)"
            and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e1 \<longrightarrow>  \<rho> \<notin> dom (Zx \<union> Z2)"
            and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2 \<longrightarrow> \<rho> \<notin> dom (Zx \<union> Z1)"
            and "finite (Lx\<union>{x}\<union>{y})"
            and "finite (Ly\<union>{x}\<union>{y})"
            using Ht\<rho> HZ1 HZ2 Hvalid Hbang HZf Hd1 Hd2 Hlx Hly by auto
        next          
          have "y \<noteq> z" using Hy_eq Hxz_1 by auto
          thus "z \<notin> dom (\<Gamma>1 \<union> (\<Gamma>2\<bar>x),(y:\<tau>))" using Hxnotin by (auto simp add: dom_def without_def)
        next
          show "\<Gamma>x;\<Delta>;M;Zx \<turnstile> TmFreeVar z : convT \<rho> \<tau>x" using Htyp by auto
        next
          show "\<Gamma>1,(z:convT \<rho> \<tau>x);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar y) i e1 : \<tau>1" 
            using Ht1o_neq Hxz_1 H1 by (auto simp add: env_permute)
        next      
          note Ht2l
          thus "\<forall>xa ya. (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta> \<Turnstile> OK \<and> (\<Gamma>2\<bar>x),(y:\<tau>) \<sim> {xa\<rhd>\<tau>x},(ya:\<tau>1) \<and> {xa\<rhd>\<tau>x} \<sim> {ya\<rhd>\<tau>1} \<and>
                        xa \<notin> Lx \<union> {x} \<union> {y} \<and> ya \<notin> Ly \<union> {x} \<union> {y} \<and> xa \<noteq> ya \<longrightarrow>
                (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar ya) (Suc 0)
                                                      (freshen_tm xa (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
            using H2 by auto
        next
          have "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq2 Hcompat1 by auto
          hence Hc1: "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x)" using env_compat_subset env_compat_symm by blast
          have "\<Gamma>1 \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
          hence "\<Gamma>1 \<subseteq> \<Gamma>" using H1 by auto
          hence "\<Gamma>1 \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto 
          thus "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using Hc1 env_compat_union_left[of \<Gamma>1 _ "{y\<rhd>\<tau>}"] by auto
        next
          have "\<Gamma>x \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq2 Hcompat2 by auto
          hence Hc1: "\<Gamma>x \<sim> (\<Gamma>2\<bar>x)" using env_compat_subset env_compat_symm by blast
          have "\<Gamma>x \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
          hence "\<Gamma>x \<subseteq> \<Gamma>" using H by auto
          hence "\<Gamma>x \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto 
          thus "\<Gamma>x \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using Hc1 env_compat_union_left[of \<Gamma>x _ "{y\<rhd>\<tau>}"] by auto
        next
          show "\<Gamma>x \<sim> \<Gamma>1" by assumption
        next
          show "\<Gamma>x \<union> \<Gamma>1 \<union> (\<Gamma>2\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using H_eq prems by auto
        qed
        thus ?thesis using Hxz_2 by auto
      qed
      hence ?thesis using H_eq HZ_eq TmLetbangSc by auto
    }
    moreover
    {
      assume H: "x\<rhd>\<tau> \<in> \<Gamma>x" and H1: "x\<rhd>\<tau> \<in> \<Gamma>1" and H2: "x\<rhd>\<tau> \<in> \<Gamma>2"
      have "\<Gamma>2;\<Delta> \<Turnstile> OK" and "\<Gamma>1;\<Delta> \<Turnstile> OK" and "\<Gamma>x;\<Delta> \<Turnstile> OK"
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heqx: "\<Gamma>x = (\<Gamma>x\<bar>x),(x:\<tau>)" and Heq2: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" and Heq1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" 
        using wf_env_defines_without H1 H2 H by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>x\<bar>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>x\<bar>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x)" using H by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>x\<bar>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x" by (auto simp add: insert_ident)
      have "((\<Gamma>x\<bar>x),(y:\<tau>) \<union> ((\<Gamma>1\<bar>x),(y:\<tau>)) \<union> ((\<Gamma>2\<bar>x),(y:\<tau>)));\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let$ \<rho> (vsubst_tm (TmFreeVar y) i e)
        (vsubst_tm (TmFreeVar y) i e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'"
        using Hxz proof rule
        assume "x \<noteq> z \<and> e = TmFreeVar z"
        hence Hxz_1: "x \<noteq> z" and Hxz_2: "e = TmFreeVar z" by auto
        have "(\<Gamma>1\<bar>x),(x:\<tau>) \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq1 Heq2 Hcompat1 by auto
        hence Hqual: "qualT \<tau> \<preceq> qU" by (auto simp add: env_compat_def)
        have "(((\<Gamma>x\<bar>x),(y:\<tau>)) \<union> ((\<Gamma>1\<bar>x),(y:\<tau>)) \<union> ((\<Gamma>2\<bar>x),(y:\<tau>)));\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let$ \<rho> (TmFreeVar z)
              (vsubst_tm (TmFreeVar y) i e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'" proof rule
          show "\<not> (Z2\<union>Z1) defines \<rho>"
            and "Z1 \<Turnstile> ScopeSet"
            and "Z2 \<Turnstile> ScopeSet"
            and "validScope (scopeT \<tau>1) Z1"
            and "isBangable \<tau>x"
            and "Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>)"
            and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e1 \<longrightarrow>  \<rho> \<notin> dom (Zx \<union> Z2)"
            and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2 \<longrightarrow> \<rho> \<notin> dom (Zx \<union> Z1)"
            and "finite (Lx\<union>{x}\<union>{y})"
            and "finite (Ly\<union>{x}\<union>{y})"
            using Ht\<rho> HZ1 HZ2 Hvalid Hbang HZf Hd1 Hd2 Hlx Hly by auto
        next                
          have Hy_eq: "y \<noteq> z" proof (cases "y = z")
            assume Heq: "y = z"
            moreover
            have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>x,(x:\<tau>)" using Htyp H\<Gamma>x_eq by auto
            hence "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>" using H\<Gamma>x_eq Hxz_1 by auto
            ultimately
            have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>,(y:\<tau>)" by auto
            hence "convT \<rho> \<tau>x = \<tau>" using Heq "1.prems" wf_env_sanity[of "\<Gamma>,(y:\<tau>)" \<Delta> y] by auto
            hence "scopeT \<tau> = ScFree \<rho>" by (cases \<tau>x, auto)
            hence False using "1.prems" TmLetbangSc by auto
            thus "y \<noteq> z" by auto
          qed
          thus "z \<notin> dom ((\<Gamma>1\<bar>x),(y:\<tau>) \<union> (\<Gamma>2 \<bar> x),(y:\<tau>))" using Hxnotin by (auto simp add: dom_def without_def)
        next
          show "(\<Gamma>x\<bar>x),(y:\<tau>);\<Delta>;M;Zx \<turnstile> TmFreeVar z : convT \<rho> \<tau>x" proof
            have "(\<Gamma>x\<bar>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using H_eq prems by auto
            thus "(\<Gamma>x\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using wf_env_subset by blast
          next
            show "z\<rhd>convT \<rho> \<tau>x \<in> (\<Gamma>x \<bar> x),(y:\<tau>)" using Htyp Hxz_1 by (auto simp add: without_def)
          next
            have "qualG ((\<Gamma>x \<bar> x),(x:\<tau>) \<bar> z) \<preceq> qU" using Htyp Heqx by auto
            thus "qualG ((\<Gamma>x \<bar> x),(y:\<tau>) \<bar> z) \<preceq> qU" using Hqual by (auto simp add: qualGrelL_def without_def)
          next
            show "M;\<Delta> \<Turnstile> OK" using Htyp by auto
          next
            show "minimalValidScope (scopeT (convT \<rho> \<tau>x)) Zx" using Htyp by auto
          qed
        next
          show "(\<Gamma>1\<bar>x),(y:\<tau>),(z:convT \<rho> \<tau>x);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar y) i e1 : \<tau>1" 
            using Ht1i H1 by (auto simp add: env_permute)
        next
          note Ht2l
          thus "\<forall>xa ya. (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta> \<Turnstile> OK \<and> (\<Gamma>2\<bar>x),(y:\<tau>) \<sim> {xa\<rhd>\<tau>x},(ya:\<tau>1) \<and> {xa\<rhd>\<tau>x} \<sim> {ya\<rhd>\<tau>1} \<and>
                        xa \<notin> Lx \<union> {x} \<union> {y} \<and> ya \<notin> Ly \<union> {x} \<union> {y} \<and> xa \<noteq> ya \<longrightarrow>
                (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar ya) (Suc 0)
                                                      (freshen_tm xa (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
            using H2 by auto
        next
          have "(\<Gamma>1\<bar>x),(x:\<tau>) \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq1 Heq2 Hcompat1 by auto
          thus "(\<Gamma>1\<bar>x),(y:\<tau>) \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" by (auto simp add: env_compat_def)
        next
          have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq2 Hcompat2 Heqx by auto
          thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using Hqual by (auto simp add: env_compat_def)
        next
          have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> (\<Gamma>1\<bar>x),(x:\<tau>)" using Heq1 Hcompat3 Heqx by auto
          thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> (\<Gamma>1\<bar>x),(y:\<tau>)" using Hqual by (auto simp add: env_compat_def)
        next
          show "(\<Gamma>x\<bar>x),(y:\<tau>) \<union> (\<Gamma>1\<bar>x),(y:\<tau>) \<union> (\<Gamma>2\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using H_eq prems by auto
        qed
        thus ?thesis using Hxz_2 by auto
      next
        assume "x = z \<and> e = TmVar i"
        hence Hxz_1: "x = z" and Hxz_2: "e = TmVar i" by auto
        have "(\<Gamma>1\<bar>x),(x:\<tau>) \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq1 Heq2 Hcompat1 by auto
        hence Hqual: "qualT \<tau> \<preceq> qU" by (auto simp add: env_compat_def)
        have "\<Gamma>x \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq H by auto
        moreover
        have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>x" using Htyp by auto
        ultimately
        have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>,(x:\<tau>)" by auto
        hence "convT \<rho> \<tau>x = \<tau>" using wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x "convT \<rho> \<tau>x" \<tau>] Hxz_1 Hwf H\<Gamma>x_eq by auto
        hence "scopeT \<tau> = ScFree \<rho>" by (cases \<tau>x, auto)
        hence False using "1.prems" TmLetbangSc by auto
        thus ?thesis by auto
      qed
      hence ?thesis using H_eq HZ_eq TmLetbangSc by auto
    }
    moreover
    {
      assume H: "x\<rhd>\<tau> \<in> \<Gamma>x" and H1: "x\<rhd>\<tau> \<in> \<Gamma>1" and H2: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      have "\<Gamma>1;\<Delta> \<Turnstile> OK" and "\<Gamma>x;\<Delta> \<Turnstile> OK"
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heqx: "\<Gamma>x = (\<Gamma>x\<bar>x),(x:\<tau>)" and Heq1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" 
        using wf_env_defines_without H1 H2 H by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>x\<bar>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>x\<bar>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2)" using H2 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>x\<bar>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2" by (auto simp add: insert_ident)
      have "((\<Gamma>x\<bar>x),(y:\<tau>) \<union> ((\<Gamma>1\<bar>x),(y:\<tau>)) \<union> \<Gamma>2);\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let$ \<rho> (vsubst_tm (TmFreeVar y) i e)
        (vsubst_tm (TmFreeVar y) i e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'"
        using Hxz proof rule
        assume "x \<noteq> z \<and> e = TmFreeVar z"
        hence Hxz_1: "x \<noteq> z" and Hxz_2: "e = TmFreeVar z" by auto
        have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> (\<Gamma>1\<bar>x),(x:\<tau>)" using Heq1 Heqx Hcompat3 by auto
        hence Hqual: "qualT \<tau> \<preceq> qU" by (auto simp add: env_compat_def)
        have "(((\<Gamma>x\<bar>x),(y:\<tau>)) \<union> ((\<Gamma>1\<bar>x),(y:\<tau>)) \<union> \<Gamma>2);\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let$ \<rho> (TmFreeVar z)
              (vsubst_tm (TmFreeVar y) i e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'" proof rule
          show "\<not> (Z2\<union>Z1) defines \<rho>"
            and "Z1 \<Turnstile> ScopeSet"
            and "Z2 \<Turnstile> ScopeSet"
            and "validScope (scopeT \<tau>1) Z1"
            and "isBangable \<tau>x"
            and "Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>)"
            and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e1 \<longrightarrow>  \<rho> \<notin> dom (Zx \<union> Z2)"
            and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2 \<longrightarrow> \<rho> \<notin> dom (Zx \<union> Z1)"
            and "finite (Lx\<union>{x}\<union>{y})"
            and "finite (Ly\<union>{x}\<union>{y})"
            using Ht\<rho> HZ1 HZ2 Hvalid Hbang HZf Hd1 Hd2 Hlx Hly by auto
        next                
          have Hy_eq: "y \<noteq> z" proof (cases "y = z")
            assume Heq: "y = z"
            moreover
            have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>x,(x:\<tau>)" using Htyp H\<Gamma>x_eq by auto
            hence "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>" using H\<Gamma>x_eq Hxz_1 by auto
            ultimately
            have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>,(y:\<tau>)" by auto
            hence "convT \<rho> \<tau>x = \<tau>" using Heq "1.prems" wf_env_sanity[of "\<Gamma>,(y:\<tau>)" \<Delta> y] by auto
            hence "scopeT \<tau> = ScFree \<rho>" by (cases \<tau>x, auto)
            hence False using "1.prems" TmLetbangSc by auto
            thus "y \<noteq> z" by auto
          qed
          thus "z \<notin> dom ((\<Gamma>1\<bar>x),(y:\<tau>) \<union> \<Gamma>2)" using Hxnotin by (auto simp add: dom_def without_def)
        next
          show "(\<Gamma>x\<bar>x),(y:\<tau>);\<Delta>;M;Zx \<turnstile> TmFreeVar z : convT \<rho> \<tau>x" proof
            have "(\<Gamma>x\<bar>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2),(y:\<tau>);\<Delta> \<Turnstile> OK" using H_eq prems by auto
            thus "(\<Gamma>x\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using wf_env_subset by blast
          next
            show "z\<rhd>convT \<rho> \<tau>x \<in> (\<Gamma>x \<bar> x),(y:\<tau>)" using Htyp Hxz_1 by (auto simp add: without_def)
          next
            have "qualG ((\<Gamma>x \<bar> x),(x:\<tau>) \<bar> z) \<preceq> qU" using Htyp Heqx by auto
            thus "qualG ((\<Gamma>x \<bar> x),(y:\<tau>) \<bar> z) \<preceq> qU" using Hqual by (auto simp add: qualGrelL_def without_def)
          next
            show "M;\<Delta> \<Turnstile> OK" using Htyp by auto
          next
            show "minimalValidScope (scopeT (convT \<rho> \<tau>x)) Zx" using Htyp by auto
          qed
        next
          show "(\<Gamma>1\<bar>x),(y:\<tau>),(z:convT \<rho> \<tau>x);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar y) i e1 : \<tau>1" 
            using Ht1i H1 by (auto simp add: env_permute)
        next
          note Ht2o
          thus "\<forall>xa ya. \<Gamma>2,(xa:\<tau>x),(ya:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {xa\<rhd>\<tau>x},(ya:\<tau>1) \<and> {xa\<rhd>\<tau>x} \<sim> {ya\<rhd>\<tau>1} \<and>
                        xa \<notin> Lx \<union> {x} \<union> {y} \<and> ya \<notin> Ly \<union> {x} \<union> {y} \<and> xa \<noteq> ya \<longrightarrow>
                \<Gamma>2,(xa:\<tau>x),(ya:\<tau>1);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar ya) (Suc 0)
                                                      (freshen_tm xa (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
            using H2 by auto
        next
          have "(\<Gamma>1\<bar>x),(x:\<tau>) \<sim> \<Gamma>2" using Heq1 Hcompat1 by auto
          hence Hc1: "(\<Gamma>1\<bar>x) \<sim> \<Gamma>2" using env_compat_subset by auto
          have "\<Gamma>2 \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
          hence "\<Gamma>2 \<subseteq> \<Gamma>" using H2 by auto
          hence "\<Gamma>2 \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto 
          hence "{y\<rhd>\<tau>} \<sim> \<Gamma>2" using env_compat_symm by auto
          thus "(\<Gamma>1\<bar>x),(y:\<tau>) \<sim> \<Gamma>2" using Hc1 env_compat_union_right[of "(\<Gamma>1\<bar>x)" _ "{y\<rhd>\<tau>}"] by auto
        next
          have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> \<Gamma>2" using Hcompat2 Heqx by auto
          thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> \<Gamma>2" using Hqual by (auto simp add: env_compat_def)
        next
          have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> (\<Gamma>1\<bar>x),(x:\<tau>)" using Heq1 Hcompat3 Heqx by auto
          thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> (\<Gamma>1\<bar>x),(y:\<tau>)" using Hqual by (auto simp add: env_compat_def)
        next
          show "(\<Gamma>x\<bar>x),(y:\<tau>) \<union> (\<Gamma>1\<bar>x),(y:\<tau>) \<union> \<Gamma>2;\<Delta> \<Turnstile> OK" using H_eq prems by auto
        qed
        thus ?thesis using Hxz_2 by auto
      next
        assume "x = z \<and> e = TmVar i"
        hence Hxz_1: "x = z" and Hxz_2: "e = TmVar i" by auto
        have "\<Gamma>x \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq H by auto
        moreover
        have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>x" using Htyp by auto
        ultimately
        have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>,(x:\<tau>)" by auto
        hence "convT \<rho> \<tau>x = \<tau>" using wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x "convT \<rho> \<tau>x" \<tau>] Hxz_1 Hwf H\<Gamma>x_eq by auto
        hence "scopeT \<tau> = ScFree \<rho>" by (cases \<tau>x, auto)
        hence False using "1.prems" TmLetbangSc by auto
        thus ?thesis by auto
      qed
      hence ?thesis using H_eq HZ_eq TmLetbangSc by auto
    }
    moreover
    {
      assume H: "x\<rhd>\<tau> \<in> \<Gamma>x" and H1: "x\<rhd>\<tau> \<notin> \<Gamma>1" and H2: "x\<rhd>\<tau> \<in> \<Gamma>2"
      have "\<Gamma>x;\<Delta> \<Turnstile> OK" and "\<Gamma>2;\<Delta> \<Turnstile> OK" 
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heqx: "\<Gamma>x = (\<Gamma>x\<bar>x),(x:\<tau>)" and Heq2: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)"
        using wf_env_defines_without H1 H2 H by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>x\<bar>x \<union> \<Gamma>1 \<union> \<Gamma>2\<bar>x),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>x\<bar>x \<union> \<Gamma>1 \<union> \<Gamma>2\<bar>x)" using H2 H1 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>x\<bar>x \<union> \<Gamma>1 \<union> \<Gamma>2\<bar>x" by (auto simp add: insert_ident)
      have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heqx Heq2 Hcompat1 by auto
      hence Hqual: "qualT \<tau> \<preceq> qU" by (auto simp add: env_compat_def)
      have "((\<Gamma>x\<bar>x),(y:\<tau>) \<union> \<Gamma>1 \<union> (\<Gamma>2\<bar>x),(y:\<tau>));\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let$ \<rho> (vsubst_tm (TmFreeVar y) i e)
        (vsubst_tm (TmFreeVar y) i e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'"
        using Hxz proof rule
        assume "x \<noteq> z \<and> e = TmFreeVar z"
        hence Hxz_1: "x \<noteq> z" and Hxz_2: "e = TmFreeVar z" by auto
        have Hy_eq: "y \<noteq> z" proof (cases "y = z")
          assume Heq: "y = z"
          moreover
          have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>x,(x:\<tau>)" using Htyp H\<Gamma>x_eq by auto
          hence "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>" using H\<Gamma>x_eq Hxz_1 by auto
          ultimately
          have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>,(y:\<tau>)" by auto
          hence "convT \<rho> \<tau>x = \<tau>" using Heq "1.prems" wf_env_sanity[of "\<Gamma>,(y:\<tau>)" \<Delta> y] by auto
          hence "scopeT \<tau> = ScFree \<rho>" by (cases \<tau>x, auto)
          hence False using "1.prems" TmLetbangSc by auto
          thus "y \<noteq> z" by auto
        qed
        have "(((\<Gamma>x\<bar>x),(y:\<tau>)) \<union> \<Gamma>1 \<union> ((\<Gamma>2\<bar>x),(y:\<tau>)));\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let$ \<rho> (TmFreeVar z)
              (vsubst_tm (TmFreeVar y) i e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'" proof rule
          show "\<not> (Z2\<union>Z1) defines \<rho>"
            and "Z1 \<Turnstile> ScopeSet"
            and "Z2 \<Turnstile> ScopeSet"
            and "validScope (scopeT \<tau>1) Z1"
            and "isBangable \<tau>x"
            and "Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>)"
            and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e1 \<longrightarrow>  \<rho> \<notin> dom (Zx \<union> Z2)"
            and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2 \<longrightarrow> \<rho> \<notin> dom (Zx \<union> Z1)"
            and "finite (Lx\<union>{x}\<union>{y})"
            and "finite (Ly\<union>{x}\<union>{y})"
            using Ht\<rho> HZ1 HZ2 Hvalid Hbang HZf Hd1 Hd2 Hlx Hly by auto
        next                
          show "z \<notin> dom (\<Gamma>1 \<union> ((\<Gamma>2\<bar>x),(y:\<tau>)))" using Hy_eq Hxnotin by (auto simp add: dom_def without_def)
        next
          show "(\<Gamma>x\<bar>x),(y:\<tau>);\<Delta>;M;Zx \<turnstile> TmFreeVar z : convT \<rho> \<tau>x" proof
            have "(\<Gamma>x\<bar>x \<union> \<Gamma>1 \<union> \<Gamma>2\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using H_eq prems by auto
            thus "(\<Gamma>x\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using wf_env_subset by blast
          next
            show "z\<rhd>convT \<rho> \<tau>x \<in> (\<Gamma>x \<bar> x),(y:\<tau>)" using Htyp Hxz_1 by (auto simp add: without_def)
          next
            have "qualG ((\<Gamma>x \<bar> x),(x:\<tau>) \<bar> z) \<preceq> qU" using Htyp Heqx by auto
            also have "((\<Gamma>x \<bar> x),(x:\<tau>) \<bar> z) = ((\<Gamma>x \<bar> x)\<bar> z),(x:\<tau>)"
              using Hxz_1 by (auto simp add: without_def dom_def)
            finally have "qualG (((\<Gamma>x \<bar> x)\<bar>z),(x:\<tau>)) \<preceq> qU" by auto
            hence "qualG (((\<Gamma>x \<bar> x)\<bar>z),(y:\<tau>)) \<preceq> qU" by (auto simp add: qualGrelL_def without_def)
            also have "((\<Gamma>x \<bar> x)\<bar> z),(y:\<tau>) = ((\<Gamma>x \<bar> x),(y:\<tau>) \<bar> z)" using Hy_eq by (auto simp add: without_def dom_def)
            finally show "qualG ((\<Gamma>x \<bar> x),(y:\<tau>) \<bar> z) \<preceq> qU" by auto
          next
            show "M;\<Delta> \<Turnstile> OK" using Htyp by auto
          next
            show "minimalValidScope (scopeT (convT \<rho> \<tau>x)) Zx" using Htyp by auto
          qed
        next
          show "\<Gamma>1,(z:convT \<rho> \<tau>x);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar y) i e1 : \<tau>1" 
            using Ht1o_neq H1 Hxz_1 by (auto simp add: env_permute)
        next
          note Ht2l
          thus "\<forall>xa ya. (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta> \<Turnstile> OK \<and> (\<Gamma>2\<bar>x),(y:\<tau>) \<sim> {xa\<rhd>\<tau>x},(ya:\<tau>1) \<and> {xa\<rhd>\<tau>x} \<sim> {ya\<rhd>\<tau>1} \<and>
                        xa \<notin> Lx \<union> {x} \<union> {y} \<and> ya \<notin> Ly \<union> {x} \<union> {y} \<and> xa \<noteq> ya \<longrightarrow>
                (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar ya) (Suc 0)
                                                      (freshen_tm xa (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
            using H2 by auto
        next
          have "\<Gamma> \<sim> {y\<rhd>\<tau>}" by assumption
          thus "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using H_eq Hcompat1 by (auto simp add: env_compat_def dom_def without_def)
        next
          have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> \<Gamma>1" using Heqx Hcompat3 by auto
          hence Hc: "(\<Gamma>x\<bar>x) \<sim> \<Gamma>1" using env_compat_subset env_compat_symm by blast
          have "\<Gamma>1 \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
          hence "\<Gamma>1 \<subseteq> \<Gamma>" using H1 by auto
          hence "\<Gamma>1 \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto
          hence "\<Gamma>1 \<sim> (\<Gamma>x\<bar>x),(y:\<tau>)" using Hc env_compat_union_left[of \<Gamma>1 "(\<Gamma>x\<bar>x)" "{y\<rhd>\<tau>}"] env_compat_symm by auto
          thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> \<Gamma>1" using env_compat_symm by auto
        next
          have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heqx Heq2 Hcompat2 by auto
          thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using Hqual by (auto simp add: env_compat_def)
        next
          show "(\<Gamma>x\<bar>x),(y:\<tau>) \<union> \<Gamma>1 \<union> (\<Gamma>2\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using H_eq prems by auto
        qed
        thus ?thesis using Hxz_2 by auto
      next
        assume "x = z \<and> e = TmVar i"
        hence Hxz_1: "x = z" and Hxz_2: "e = TmVar i" by auto
        have "\<Gamma>x \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq H by auto
        moreover
        have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>x" using Htyp by auto
        ultimately
        have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>,(x:\<tau>)" by auto
        hence "convT \<rho> \<tau>x = \<tau>" using wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x "convT \<rho> \<tau>x" \<tau>] Hxz_1 Hwf H\<Gamma>x_eq by auto
        hence "scopeT \<tau> = ScFree \<rho>" by (cases \<tau>x, auto)
        hence False using "1.prems" TmLetbangSc by auto
        thus ?thesis by auto
      qed
      hence ?thesis using H_eq HZ_eq TmLetbangSc by auto
    }
    moreover
    {
      assume H: "x\<rhd>\<tau> \<in> \<Gamma>x" and H1: "x\<rhd>\<tau> \<notin> \<Gamma>1" and H2: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      have "\<Gamma>x;\<Delta> \<Turnstile> OK"
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heqx: "\<Gamma>x = (\<Gamma>x\<bar>x),(x:\<tau>)"
        using wf_env_defines_without H1 H2 H by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>x\<bar>x \<union> \<Gamma>1 \<union> \<Gamma>2),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>x\<bar>x \<union> \<Gamma>1 \<union> \<Gamma>2)" using H2 H1 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>x\<bar>x \<union> \<Gamma>1 \<union> \<Gamma>2" by (auto simp add: insert_ident)
      have "((\<Gamma>x\<bar>x),(y:\<tau>) \<union> \<Gamma>1 \<union> \<Gamma>2);\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let$ \<rho> (vsubst_tm (TmFreeVar y) i e)
        (vsubst_tm (TmFreeVar y) i e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'"
        using Hxz proof rule
        assume "x \<noteq> z \<and> e = TmFreeVar z"
        hence Hxz_1: "x \<noteq> z" and Hxz_2: "e = TmFreeVar z" by auto
        have Hy_eq: "y \<noteq> z" proof (cases "y = z")
          assume Heq: "y = z"
          moreover
          have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>x,(x:\<tau>)" using Htyp H\<Gamma>x_eq by auto
          hence "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>" using H\<Gamma>x_eq Hxz_1 by auto
          ultimately
          have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>,(y:\<tau>)" by auto
          hence "convT \<rho> \<tau>x = \<tau>" using Heq "1.prems" wf_env_sanity[of "\<Gamma>,(y:\<tau>)" \<Delta> y] by auto
          hence "scopeT \<tau> = ScFree \<rho>" by (cases \<tau>x, auto)
          hence False using "1.prems" TmLetbangSc by auto
          thus "y \<noteq> z" by auto
        qed
        have "(((\<Gamma>x\<bar>x),(y:\<tau>)) \<union> \<Gamma>1 \<union> \<Gamma>2);\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let$ \<rho> (TmFreeVar z)
              (vsubst_tm (TmFreeVar y) i e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'" proof rule
          show "\<not> (Z2\<union>Z1) defines \<rho>"
            and "Z1 \<Turnstile> ScopeSet"
            and "Z2 \<Turnstile> ScopeSet"
            and "validScope (scopeT \<tau>1) Z1"
            and "isBangable \<tau>x"
            and "Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>)"
            and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e1 \<longrightarrow>  \<rho> \<notin> dom (Zx \<union> Z2)"
            and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2 \<longrightarrow> \<rho> \<notin> dom (Zx \<union> Z1)"
            and "finite (Lx\<union>{x}\<union>{y})"
            and "finite (Ly\<union>{x}\<union>{y})"
            using Ht\<rho> HZ1 HZ2 Hvalid Hbang HZf Hd1 Hd2 Hlx Hly by auto
        next                
          show "z \<notin> dom (\<Gamma>1 \<union> \<Gamma>2)" using Hy_eq Hxnotin by (auto simp add: dom_def without_def)
        next
          show "(\<Gamma>x\<bar>x),(y:\<tau>);\<Delta>;M;Zx \<turnstile> TmFreeVar z : convT \<rho> \<tau>x" proof
            have "(\<Gamma>x\<bar>x \<union> \<Gamma>1 \<union> \<Gamma>2),(y:\<tau>);\<Delta> \<Turnstile> OK" using prems H_eq by auto
            thus "(\<Gamma>x\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using wf_env_subset by blast
          next
            show "z\<rhd>convT \<rho> \<tau>x \<in> (\<Gamma>x \<bar> x),(y:\<tau>)" using Htyp Hxz_1 by (auto simp add: without_def)
          next
            have "qualG ((\<Gamma>x \<bar> x),(x:\<tau>) \<bar> z) \<preceq> qU" using Htyp Heqx by auto
            also have "((\<Gamma>x \<bar> x),(x:\<tau>) \<bar> z) = ((\<Gamma>x \<bar> x)\<bar> z),(x:\<tau>)"
              using Hxz_1 by (auto simp add: without_def dom_def)
            finally have "qualG (((\<Gamma>x \<bar> x)\<bar>z),(x:\<tau>)) \<preceq> qU" by auto
            hence "qualG (((\<Gamma>x \<bar> x)\<bar>z),(y:\<tau>)) \<preceq> qU" by (auto simp add: qualGrelL_def without_def)
            also have "((\<Gamma>x \<bar> x)\<bar> z),(y:\<tau>) = ((\<Gamma>x \<bar> x),(y:\<tau>) \<bar> z)" using Hy_eq by (auto simp add: without_def dom_def)
            finally show "qualG ((\<Gamma>x \<bar> x),(y:\<tau>) \<bar> z) \<preceq> qU" by auto
          next
            show "M;\<Delta> \<Turnstile> OK" using Htyp by auto
          next
            show "minimalValidScope (scopeT (convT \<rho> \<tau>x)) Zx" using Htyp by auto
          qed
        next
          show "\<Gamma>1,(z:convT \<rho> \<tau>x);\<Delta>;M;Zf \<turnstile> vsubst_tm (TmFreeVar y) i e1 : \<tau>1" 
            using Ht1o_neq H1 Hxz_1 by (auto simp add: env_permute)
        next
          note Ht2o
          thus "\<forall>xa ya. \<Gamma>2,(xa:\<tau>x),(ya:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {xa\<rhd>\<tau>x},(ya:\<tau>1) \<and> {xa\<rhd>\<tau>x} \<sim> {ya\<rhd>\<tau>1} \<and>
                        xa \<notin> Lx \<union> {x} \<union> {y} \<and> ya \<notin> Ly \<union> {x} \<union> {y} \<and> xa \<noteq> ya \<longrightarrow>
                \<Gamma>2,(xa:\<tau>x),(ya:\<tau>1);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar ya) (Suc 0)
                                                      (freshen_tm xa (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
            using H2 by auto
        next
          show "\<Gamma>1 \<sim> \<Gamma>2" by assumption 
        next
          have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> \<Gamma>1" using Heqx Hcompat3 by auto
          hence Hc: "(\<Gamma>x\<bar>x) \<sim> \<Gamma>1" using env_compat_subset env_compat_symm by blast
          have "\<Gamma>1 \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
          hence "\<Gamma>1 \<subseteq> \<Gamma>" using H1 by auto
          hence "\<Gamma>1 \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto
          hence "\<Gamma>1 \<sim> (\<Gamma>x\<bar>x),(y:\<tau>)" using Hc env_compat_union_left[of \<Gamma>1 "(\<Gamma>x\<bar>x)" "{y\<rhd>\<tau>}"] env_compat_symm by auto
          thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> \<Gamma>1" using env_compat_symm by auto
        next
          have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> \<Gamma>2" using Heqx Hcompat2 by auto
          hence Hc: "(\<Gamma>x\<bar>x) \<sim> \<Gamma>2" using env_compat_subset env_compat_symm by blast
          have "\<Gamma>2 \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
          hence "\<Gamma>2 \<subseteq> \<Gamma>" using H2 by auto
          hence "\<Gamma>2 \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto
          hence "\<Gamma>2 \<sim> (\<Gamma>x\<bar>x),(y:\<tau>)" using Hc env_compat_union_left[of \<Gamma>2 "(\<Gamma>x\<bar>x)" "{y\<rhd>\<tau>}"] env_compat_symm by auto
          thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> \<Gamma>2" using env_compat_symm by auto
        next
          show "(\<Gamma>x\<bar>x),(y:\<tau>) \<union> \<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK" using H_eq prems by auto
        qed
        thus ?thesis using Hxz_2 by auto
      next
        assume "x = z \<and> e = TmVar i"
        hence Hxz_1: "x = z" and Hxz_2: "e = TmVar i" by auto
        have "\<Gamma>x \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq H by auto
        moreover
        have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>x" using Htyp by auto
        ultimately
        have "z\<rhd>convT \<rho> \<tau>x \<in> \<Gamma>,(x:\<tau>)" by auto
        hence "convT \<rho> \<tau>x = \<tau>" using wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x "convT \<rho> \<tau>x" \<tau>] Hxz_1 Hwf H\<Gamma>x_eq by auto
        hence "scopeT \<tau> = ScFree \<rho>" by (cases \<tau>x, auto)
        hence False using "1.prems" TmLetbangSc by auto
        thus ?thesis by auto
      qed
      hence ?thesis using H_eq HZ_eq TmLetbangSc by auto
    }
    moreover
    {
      assume H: "x\<rhd>\<tau> \<notin> \<Gamma>x" and  H1: "x\<rhd>\<tau> \<notin> \<Gamma>1" and H2: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      hence "\<not> \<Gamma>x defines x" and  "\<not> \<Gamma>1 defines x" and "\<not> \<Gamma>2 defines x" using Hdomx Hdom1 Hdom2 by auto
      hence ?thesis using H\<Gamma>x_eq by (auto simp add: dom_def)
    }
    ultimately
    show ?thesis by auto
  next
    case (TmLetbang e e1 e2)
    let ?es = "vsubst_tm (TmFreeVar x) i e"
    let ?e1s = "vsubst_tm (TmFreeVar x) (Suc i) e1"
    let ?e2s = "vsubst_tm (TmFreeVar x) (Suc (Suc i)) e2"
    let ?ey = "vsubst_tm (TmFreeVar y) i e"
    let ?e1y = "vsubst_tm (TmFreeVar y) (Suc i) e1"
    let ?e2y = "vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2"
    have Ht: "\<Gamma>,(x:\<tau>);\<Delta>;M;Z \<turnstile> let! ?es ?e1s in ?e2s : \<tau>'" using 1 TmLetbang by auto
    then obtain \<Gamma>1 \<Gamma>2 \<Gamma>x Z1 Z2 Zx \<tau>x \<tau>1 Lx Ly
      where H\<Gamma>x_eq: "\<Gamma>,(x:\<tau>) = \<Gamma>x \<union> \<Gamma>1 \<union> \<Gamma>2"
      and Hcompat1: "\<Gamma>1 \<sim> \<Gamma>2" and Hcompat2: "\<Gamma>x \<sim> \<Gamma>2" and Hcompat3: "\<Gamma>x \<sim> \<Gamma>1"
      and Hwf: "\<Gamma>x \<union> \<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK"
      and HZ_eq: "Z = Z2 \<union> Z1 \<union> Zx"
      and HZ1: "Z1 \<Turnstile> ScopeSet"
      and HZ2: "Z2 \<Turnstile> ScopeSet"
      and Hvalid: "validScope (scopeT \<tau>1) Z1"
      and Hbang: "isBangable \<tau>x"
      and Htyp: "\<Gamma>x;\<Delta>;M;Zx \<turnstile> ?es : \<tau>x"
      and Htyp1: "\<forall>x \<rho>. \<Gamma>1,(x:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK \<and> \<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>x} \<and> x \<notin> Lx
                             \<and> \<not> \<rho> is free dollar sc in ?e1s \<and> \<not> (Z2\<union>Z1) defines \<rho> \<longrightarrow> (
                             \<Gamma>1,(x:convT \<rho> \<tau>x); \<Delta>; M; Z1 \<turnstile> freshen_tm x (freshen_sc \<rho> ?e1s) : \<tau>1
                             \<or>
                             \<Gamma>1,(x:convT \<rho> \<tau>x); \<Delta>; M; Z1,(\<rho>:\<section>) \<turnstile> freshen_tm x (freshen_sc \<rho> ?e1s) : \<tau>1
                         )"
      and Htyp2: "\<forall>z w. \<Gamma>2,(z:\<tau>x),(w:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {z\<rhd>\<tau>x},(w:\<tau>1) \<and> {z\<rhd>\<tau>x} \<sim> {w\<rhd>\<tau>1} \<and> 
                  z \<notin> Lx \<and> w \<notin> Ly \<and> z \<noteq> w \<longrightarrow>
                  \<Gamma>2,(z:\<tau>x),(w:\<tau>1); \<Delta>;M;Z2 \<turnstile> 
                  vsubst_tm (TmFreeVar w) (Suc 0) (vsubst_tm (TmFreeVar z) 0 ?e2s) : \<tau>'"
      and Hdollar1: "\<forall>\<rho>. \<rho> is free dollar sc in ?e1s \<longrightarrow> \<not> (Z2\<union>Zx) defines \<rho>"
      and Hdollar2: "\<forall>\<rho>. \<rho> is free dollar sc in ?e2s \<longrightarrow> \<not> (Z1\<union>Zx) defines \<rho>"
      and Hdollarx: "\<forall>\<rho>. \<rho> is free dollar sc in ?es \<longrightarrow> \<not> (Z1\<union>Z2) defines \<rho>"
      and Hlx: "finite Lx"
      and Hly: "finite Ly"
      using typing_let_elim[of "\<Gamma>,(x:\<tau>)" \<Delta> M Z ?es ?e1s ?e2s \<tau>']
      by auto;
    {
      assume Hin: "x\<rhd>\<tau> \<in> \<Gamma>x"
      moreover
      have Hwf: "\<Gamma>x;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta> \<Gamma>x] by auto
      ultimately
      have Hg1: "\<Gamma>x = (\<Gamma>x\<bar>x),(x:\<tau>)" using wf_env_defines_without by auto 
      have Hnotin: "\<not> (\<Gamma>x\<bar>x) defines x" by (auto simp add: without_def dom_def)
      have Ht1: "(\<Gamma>x\<bar>x),(y:\<tau>);\<Delta>;M;Zx \<turnstile> vsubst_tm (TmFreeVar y) i e : \<tau>x" proof (rule IH)
        show "(\<Gamma>x\<bar>x),(x:\<tau>);\<Delta>;M;Zx \<turnstile> vsubst_tm (TmFreeVar x) i e : \<tau>x"
          using Htyp Hg1 by auto
      next
        show "(\<Gamma>x\<bar>x),(x:\<tau>);\<Delta> \<Turnstile> OK" using Hg1 Hwf by auto
      next
        show "(\<Gamma>x\<bar>x) \<sim> {x\<rhd>\<tau>}" by (auto simp add: env_compat_def dom_def without_def)
      next
        have "\<Gamma>x\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        thus "(\<Gamma>x\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using "1.prems" wf_env_subset[of "\<Gamma>,(y:\<tau>)" \<Delta> "(\<Gamma>x\<bar>x),(y:\<tau>)"] by auto
      next
        have "\<Gamma>x\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        thus "\<Gamma>x\<bar>x \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset[of \<Gamma> "{y\<rhd>\<tau>}" "(\<Gamma>x\<bar>x)"] by auto
      next
        show "\<not> x free in e" using "1.prems" TmLetbang by auto
      next
        show "\<not> (\<Gamma>x\<bar>x) defines x" using Hnotin by auto
      qed (insert TmLetbang "1.prems", auto)
    }
    hence Htxi: "x\<rhd>\<tau> \<in> \<Gamma>x \<longrightarrow> (\<Gamma>x\<bar>x),(y:\<tau>);\<Delta>;M;Zx \<turnstile> vsubst_tm (TmFreeVar y) i e : \<tau>x"
      by auto
    {
      assume Hin: "x\<rhd>\<tau> \<in> \<Gamma>1"
      moreover
      have Hwf: "\<Gamma>1;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta> \<Gamma>1] by auto
      ultimately
      have Hg1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" using wf_env_defines_without by auto 
      have Hnotin: "\<not> (\<Gamma>1\<bar>x) defines x" by (auto simp add: without_def dom_def)
      have Ht1: "\<forall>z \<rho>. (\<Gamma>1\<bar>x),(y:\<tau>),(z:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK \<and> (\<Gamma>1\<bar>x),(y:\<tau>) \<sim> {z\<rhd>convT \<rho> \<tau>x} \<and> z \<notin> (Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1)
                             \<and> \<not> \<rho> is free dollar sc in ?e1y \<and> \<not> (Z2\<union>Z1) defines \<rho> \<longrightarrow> (
                             (\<Gamma>1\<bar>x),(y:\<tau>),(z:convT \<rho> \<tau>x); \<Delta>; M; Z1 \<turnstile> freshen_tm z (freshen_sc \<rho> ?e1y) : \<tau>1
                             \<or>
                             (\<Gamma>1\<bar>x),(y:\<tau>),(z:convT \<rho> \<tau>x); \<Delta>; M; Z1,(\<rho>:\<section>) \<turnstile> freshen_tm z (freshen_sc \<rho> ?e1y) : \<tau>1
                         )"
      proof (rule, rule, rule)
        fix z \<rho> 
        assume "(\<Gamma>1\<bar>x),(y:\<tau>),(z:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK \<and> (\<Gamma>1\<bar>x),(y:\<tau>) \<sim> {z\<rhd>convT \<rho> \<tau>x} \<and> z \<notin> Lx \<union> {x} \<union> {y} \<union> dom \<Gamma>1 \<and>
          \<rho> \<notin> set (free_let_dollar_scope ?e1y) \<and>
          \<rho> \<notin> Environ.dom (Z2 \<union> Z1)"
        hence Hwf: "(\<Gamma>1\<bar>x),(y:\<tau>),(z:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK" and Hcomp1: "(\<Gamma>1\<bar>x),(y:\<tau>) \<sim> {z\<rhd>convT \<rho> \<tau>x}" 
          and Hz: "z \<notin> (Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1)"
          and Hd: "\<not> \<rho> is free dollar sc in ?e1y" and "\<not> (Z2\<union>Z1) defines \<rho>"
          by auto
        moreover
        have Hnin: "\<not> (\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x) defines x" using Hz Hnotin by (auto simp add: dom_def)
        have Hwft: "\<Delta> \<Turnstile>' \<tau>" using poly_env_wf_ty Hwf Hin by auto
        have Hwf_1: "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(x:\<tau>);\<Delta> \<Turnstile> OK" apply (rule wf_env_cons[of "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x)" \<Delta> x \<tau>])
          using Hwf Hnin Hwft wf_env_subset[of "(\<Gamma>1\<bar>x),(y:\<tau>),(z:convT \<rho> \<tau>x)" \<Delta> "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x)"] by auto
        hence "(\<Gamma>1\<bar>x),(x:\<tau>),(z:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK" by (auto simp add: env_permute)
        hence Hwf: "\<Gamma>1,(z:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK" using Hg1 by auto 
        moreover
        have "\<not> \<rho> is free dollar sc in ?e1s" using Hd free_dollars_vsubst_simpl by auto
        moreover
        have  Hcomp_ih: "(\<Gamma>1\<bar>x),(x:\<tau>) \<sim> {z\<rhd>convT \<rho> \<tau>x}" using Hcomp1 Hz by (auto simp add: env_compat_def dom_def)
        hence Hcomp1: "\<Gamma>1 \<sim> {z\<rhd>convT \<rho> \<tau>x}" using Hg1 by auto
        ultimately
        have "(\<Gamma>1\<bar>x),(x:\<tau>),(z:convT \<rho> \<tau>x); \<Delta>; M; Z1 \<turnstile> freshen_tm z (freshen_sc \<rho> ?e1s) : \<tau>1
          \<or>
          (\<Gamma>1\<bar>x),(x:\<tau>),(z:convT \<rho> \<tau>x); \<Delta>; M; Z1,(\<rho>:\<section>) \<turnstile> freshen_tm z (freshen_sc \<rho> ?e1s) : \<tau>1"
          using Htyp1 Hg1 by auto 
        hence "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(x:\<tau>); \<Delta>; M; Z1 \<turnstile> freshen_tm z (freshen_sc \<rho> ?e1s) : \<tau>1
          \<or>
          (\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(x:\<tau>); \<Delta>; M; Z1,(\<rho>:\<section>) \<turnstile> freshen_tm z (freshen_sc \<rho> ?e1s) : \<tau>1" by (auto simp only: env_permute)
        also have "freshen_tm z (freshen_sc \<rho> ?e1s) = freshen_tm z (vsubst_tm (TmFreeVar x) (Suc i) (freshen_sc \<rho> e1))"
          using subst_subst_tmsc_swap by auto
        also have "freshen_tm z (vsubst_tm (TmFreeVar x) (Suc i) (freshen_sc \<rho> e1)) =
          vsubst_tm (TmFreeVar x) (Suc i) (freshen_tm z (freshen_sc \<rho> e1))"
          using subst_subst_swap[of "0" "Suc i"] by auto
        finally
        have Htyp_new:
          "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(x:\<tau>); \<Delta>; M; Z1 \<turnstile> vsubst_tm (TmFreeVar x) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) : \<tau>1
          \<or>
          (\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(x:\<tau>); \<Delta>; M; Z1,(\<rho>:\<section>) \<turnstile> vsubst_tm (TmFreeVar x) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) : \<tau>1" 
          by (auto simp only: env_permute)
        have "vsubst_tm (TmFreeVar y) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) = 
          freshen_tm z (vsubst_tm (TmFreeVar y) (Suc i) (freshen_sc \<rho> e1))"
          using subst_subst_swap[of "0" "Suc i"] by auto
        hence Hsubst_eq: "vsubst_tm (TmFreeVar y) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) = freshen_tm z (freshen_sc \<rho> ?e1y)"
          using subst_subst_tmsc_swap by auto
        have Hih1: "size (freshen_tm z (freshen_sc \<rho> e1)) < size e_y"
          using TmLetbang by auto
        have Hih3: "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x) \<sim> {x\<rhd>\<tau>}" using Hcomp_ih Hz 
          by (auto simp add: env_compat_def dom_def without_def)
        have "\<Gamma>1\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)  
        hence Hwf_4: "(\<Gamma>1\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using "1.prems" wf_env_subset[of "\<Gamma>,(y:\<tau>)" \<Delta> "(\<Gamma>1\<bar>x),(y:\<tau>)"] by auto
        have "\<Delta> \<Turnstile>' \<tau>x" using Htyp poly_ty_wf_type by auto
        hence "\<Delta> \<Turnstile>' convT \<rho> \<tau>x" by (cases \<tau>x, auto)
        moreover
        have "\<not> (\<Gamma>1\<bar>x),(y:\<tau>) defines z" using Hz by (auto simp add: dom_def without_def)
        ultimately
        have Hih4: "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using Hwf_4 wf_env_cons[of "(\<Gamma>1\<bar>x),(y:\<tau>)" \<Delta> z "convT \<rho> \<tau>x"] 
          by (auto simp add: env_permute)
        have "\<Gamma>1\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq by (auto simp add: without_def)
        hence "\<Gamma>1\<bar>x \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset[of \<Gamma> "{y\<rhd>\<tau>}" "(\<Gamma>1\<bar>x)"] by auto
        moreover
        have "{z\<rhd>convT \<rho> \<tau>x} \<sim> {y\<rhd>\<tau>}" by (cases \<tau>x, auto simp add: env_compat_def) 
        ultimately
        have Hih5: "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x) \<sim> {y\<rhd>\<tau>}" 
          using env_compat_union_right[of "\<Gamma>1\<bar>x" "{y\<rhd>\<tau>}" "{z\<rhd>convT \<rho> \<tau>x}"] by auto
        have "\<not> x free in e1" using "1.prems" TmLetbang by auto
        hence Hih6: "x \<notin> set (free_tm (freshen_tm z (freshen_sc \<rho> e1)))"
          using Hz not_free_subst not_free_subst_sc_term by auto
        have Hih7: "x \<notin> dom ((\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x))" using Hz Hnotin by (auto simp add: dom_def)
        have "\<forall>\<rho>'. \<rho>' is free dollar sc in e1 \<longrightarrow> scopeT \<tau> \<noteq> ScFree \<rho>'" using "1.prems" TmLetbang by auto
        hence Hih8: "\<forall>\<rho>'. \<rho>' is free dollar sc in freshen_tm z (freshen_sc \<rho> e1) \<longrightarrow> scopeT \<tau> \<noteq> ScFree \<rho>'"
          using free_dollars_vsubst by auto
        have "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(y:\<tau>);\<Delta>;M;Z1 \<turnstile> freshen_tm z (freshen_sc \<rho> (vsubst_tm (TmFreeVar y) (Suc i) e1)) : \<tau>1 \<or> 
          (\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(y:\<tau>);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm z (freshen_sc \<rho> (vsubst_tm (TmFreeVar y) (Suc i) e1)) : \<tau>1" 
          using Htyp_new
        proof
          assume Ht: "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(x:\<tau>); \<Delta>; M; Z1 \<turnstile>
            vsubst_tm (TmFreeVar x) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) : \<tau>1"
          have "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(y:\<tau>); \<Delta>; M; Z1 \<turnstile> vsubst_tm (TmFreeVar y) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) : \<tau>1" 
            apply (rule IH) using Hih1 Ht Hwf_1 Hih3 Hih4 Hih5 Hih6 Hih7 Hih8 by auto
         also note Hsubst_eq
         finally
         show ?thesis by auto
        next
          assume Ht: "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(x:\<tau>); \<Delta>; M; Z1,(\<rho>:\<section>) \<turnstile>
            vsubst_tm (TmFreeVar x) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) : \<tau>1"
          have "(\<Gamma>1\<bar>x),(z:convT \<rho> \<tau>x),(y:\<tau>); \<Delta>; M; Z1,(\<rho>:\<section>) \<turnstile> 
            vsubst_tm (TmFreeVar y) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) : \<tau>1" 
            apply (rule IH) using Hih1 Ht Hwf_1 Hih3 Hih4 Hih5 Hih6 Hih7 Hih8 by auto
         also note Hsubst_eq
         finally
         show ?thesis by auto
        qed
        thus "(\<Gamma>1\<bar>x),(y:\<tau>),(z:convT \<rho> \<tau>x);\<Delta>;M;Z1 \<turnstile> freshen_tm z (freshen_sc \<rho> ?e1y) : \<tau>1 \<or> 
          (\<Gamma>1\<bar>x),(y:\<tau>),(z:convT \<rho> \<tau>x);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm z (freshen_sc \<rho> ?e1y) : \<tau>1" 
          by (auto simp only: env_permute)
      qed
    };
    hence Ht1i: "x\<rhd>\<tau> \<in> \<Gamma>1 \<longrightarrow> 
        (\<forall>xa \<rho>. (\<Gamma>1\<bar>x),(y:\<tau>),(xa:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK \<and> (\<Gamma>1\<bar>x),(y:\<tau>) \<sim> {xa\<rhd>convT \<rho> \<tau>x} \<and> xa \<notin> Lx \<union> {x} \<union> {y} \<union> dom \<Gamma>1 \<and>
            \<rho> \<notin> set (free_let_dollar_scope (vsubst_tm (TmFreeVar y) (Suc i) e1)) \<and> \<rho> \<notin> Environ.dom (Z2 \<union> Z1) \<longrightarrow>
            (\<Gamma>1\<bar>x),(y:\<tau>),(xa:convT \<rho> \<tau>x);\<Delta>;M;Z1 \<turnstile> freshen_tm xa (freshen_sc \<rho> (vsubst_tm (TmFreeVar y) (Suc i) e1)) : \<tau>1 \<or>
            (\<Gamma>1\<bar>x),(y:\<tau>),(xa:convT \<rho> \<tau>x);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm xa (freshen_sc \<rho> (vsubst_tm (TmFreeVar y) (Suc i) e1)) : \<tau>1)"
      by auto
    have Hdomx: "x\<rhd>\<tau> \<notin> \<Gamma>x \<longrightarrow> \<not> \<Gamma>x defines x" proof (auto simp add: dom_def)
      fix b
      assume "(x, b) \<in> \<Gamma>x"
      and Hne: "x\<rhd>\<tau> \<notin> \<Gamma>x"
      then obtain \<tau>' where "b = Term \<tau>'" and Hin: "x\<rhd>\<tau>' \<in> \<Gamma>x" by (cases b, auto)
      hence "x\<rhd>\<tau>' \<in> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
      hence "\<tau>' = \<tau>" using Ht ty_wf_env wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x \<tau> \<tau>'] by auto
      thus False using Hne Hin by auto
    qed
    hence Htxo: "x\<rhd>\<tau> \<notin> \<Gamma>x \<longrightarrow> \<Gamma>x;\<Delta>;M;Zx \<turnstile> vsubst_tm (TmFreeVar y) i e : \<tau>x"
      using Htyp not_in_env_ty_same_subst Hdomx by auto
    have Hdom1: "x\<rhd>\<tau> \<notin> \<Gamma>1 \<longrightarrow> \<not> \<Gamma>1 defines x" proof (auto simp add: dom_def)
      fix b
      assume "(x, b) \<in> \<Gamma>1"
      and Hne: "x\<rhd>\<tau> \<notin> \<Gamma>1"
      then obtain \<tau>' where "b = Term \<tau>'" and Hin: "x\<rhd>\<tau>' \<in> \<Gamma>1" by (cases b, auto)
      hence "x\<rhd>\<tau>' \<in> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
      hence "\<tau>' = \<tau>" using Ht ty_wf_env wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x \<tau> \<tau>'] by auto
      thus False using Hne Hin by auto
    qed
    have Hdom2: "x\<rhd>\<tau> \<notin> \<Gamma>2 \<longrightarrow> \<not> \<Gamma>2 defines x" proof (auto simp add: dom_def)
      fix b
      assume "(x, b) \<in> \<Gamma>2"
      and Hne: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      then obtain \<tau>' where "b = Term \<tau>'" and Hin: "x\<rhd>\<tau>' \<in> \<Gamma>2" by (cases b, auto)
      hence "x\<rhd>\<tau>' \<in> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
      hence "\<tau>' = \<tau>" using Ht ty_wf_env wf_env_sanity[of "\<Gamma>,(x:\<tau>)" \<Delta> x \<tau> \<tau>'] by auto
      thus False using Hne Hin by auto
    qed
    {
      fix z \<rho>
      assume Hassm: "x\<rhd>\<tau> \<notin> \<Gamma>1"
        and "\<Gamma>1,(z:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK" and "\<Gamma>1 \<sim> {z\<rhd>convT \<rho> \<tau>x}" and Hz: "z \<notin> Lx \<union> {x} \<union> {y} \<union> Environ.dom \<Gamma>1"
        and "\<rho> \<notin> set (free_let_dollar_scope (vsubst_tm (TmFreeVar y) (Suc i) e1))" and "\<rho> \<notin> Environ.dom (Z2 \<union> Z1)"
      moreover
      hence "\<not> \<rho> is free dollar sc in ?e1s" using free_dollars_vsubst_simpl by auto
      ultimately
      have "\<Gamma>1,(z:convT \<rho> \<tau>x); \<Delta>; M; Z1 \<turnstile> freshen_tm z (freshen_sc \<rho> ?e1s) : \<tau>1
        \<or>
        \<Gamma>1,(z:convT \<rho> \<tau>x); \<Delta>; M; Z1,(\<rho>:\<section>) \<turnstile> freshen_tm z (freshen_sc \<rho> ?e1s) : \<tau>1"
        using Htyp1 by auto
      also have "freshen_tm z (freshen_sc \<rho> ?e1s) = freshen_tm z (vsubst_tm (TmFreeVar x) (Suc i) (freshen_sc \<rho> e1))"
        using subst_subst_tmsc_swap by auto
      also have "freshen_tm z (vsubst_tm (TmFreeVar x) (Suc i) (freshen_sc \<rho> e1)) =
        vsubst_tm (TmFreeVar x) (Suc i) (freshen_tm z (freshen_sc \<rho> e1))"
        using subst_subst_swap[of "0" "Suc i"] by auto
      finally
      have
        "\<Gamma>1,(z:convT \<rho> \<tau>x);\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar y) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) : \<tau>1 \<or>
        \<Gamma>1,(z:convT \<rho> \<tau>x);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> vsubst_tm (TmFreeVar y) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) : \<tau>1"
      proof
        assume "\<Gamma>1,(z:convT \<rho> \<tau>x);\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar x) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) : \<tau>1"
        moreover
        have "\<not> (\<Gamma>1,(z:convT \<rho> \<tau>x)) defines x" using Hdom1 Hassm Hz by (auto simp add: dom_def)
        ultimately
        have "\<Gamma>1,(z:convT \<rho> \<tau>x);\<Delta>;M;Z1 \<turnstile> vsubst_tm (TmFreeVar y) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) : \<tau>1"
          using not_in_env_ty_same_subst Hz by auto
        thus ?thesis by auto
      next
        assume "\<Gamma>1,(z:convT \<rho> \<tau>x);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> vsubst_tm (TmFreeVar x) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) : \<tau>1"
        moreover
        have "\<not> (\<Gamma>1,(z:convT \<rho> \<tau>x)) defines x" using Hdom1 Hassm Hz by (auto simp add: dom_def)
        ultimately
        have "\<Gamma>1,(z:convT \<rho> \<tau>x);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> vsubst_tm (TmFreeVar y) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) : \<tau>1"
          using not_in_env_ty_same_subst Hz by auto
        thus ?thesis by auto
      qed
      also 
      have "vsubst_tm (TmFreeVar y) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) = 
        freshen_tm z (vsubst_tm (TmFreeVar y) (Suc i) (freshen_sc \<rho> e1))"
        using subst_subst_swap[of "0" "Suc i"] by auto
      hence Hsubst_eq: "vsubst_tm (TmFreeVar y) (Suc i) (freshen_tm z (freshen_sc \<rho> e1)) = freshen_tm z (freshen_sc \<rho> ?e1y)"
        using subst_subst_tmsc_swap by auto
      finally
      have
        "\<Gamma>1,(z:convT \<rho> \<tau>x);\<Delta>;M;Z1 \<turnstile> freshen_tm z (freshen_sc \<rho> (vsubst_tm (TmFreeVar y) (Suc i) e1)) : \<tau>1 \<or>
        \<Gamma>1,(z:convT \<rho> \<tau>x);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm z (freshen_sc \<rho> (vsubst_tm (TmFreeVar y) (Suc i) e1)) : \<tau>1"
        by auto
    }
    hence Ht1o: "x\<rhd>\<tau> \<notin> \<Gamma>1 \<longrightarrow> (
      \<forall>xa \<rho>. \<Gamma>1,(xa:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK \<and> \<Gamma>1 \<sim> {xa\<rhd>convT \<rho> \<tau>x} \<and>  xa \<notin> Lx \<union> {x} \<union> {y} \<union> Environ.dom \<Gamma>1 \<and>
      \<rho> \<notin> set (free_let_dollar_scope (vsubst_tm (TmFreeVar y) (Suc i) e1)) \<and> \<rho> \<notin> Environ.dom (Z2 \<union> Z1) \<longrightarrow>
      \<Gamma>1,(xa:convT \<rho> \<tau>x);\<Delta>;M;Z1 \<turnstile> freshen_tm xa (freshen_sc \<rho> (vsubst_tm (TmFreeVar y) (Suc i) e1)) : \<tau>1 \<or>
      \<Gamma>1,(xa:convT \<rho> \<tau>x);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm xa (freshen_sc \<rho> (vsubst_tm (TmFreeVar y) (Suc i) e1)) : \<tau>1)"
      by auto
    {
      assume Hin: "x\<rhd>\<tau> \<in> \<Gamma>2"
      moreover
      have Hwf': "\<Gamma>2;\<Delta> \<Turnstile> OK" using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta> \<Gamma>2] by auto
      ultimately
      have Hg1: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" using wf_env_defines_without by auto 
      have Hnotin: "\<not> (\<Gamma>2\<bar>x) defines x" by (auto simp add: without_def dom_def)
      have Ht2: "\<forall>w q. (\<Gamma>2\<bar>x),(y:\<tau>),(w:\<tau>x),(q:\<tau>1);\<Delta> \<Turnstile> OK \<and> (\<Gamma>2\<bar>x),(y:\<tau>) \<sim> {w\<rhd>\<tau>x},(q:\<tau>1) \<and>
                       {w\<rhd>\<tau>x} \<sim> {q\<rhd>\<tau>1} \<and> w \<notin> (Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1) \<and> q \<notin> (Ly\<union>{x}\<union>{y}) \<and> w \<noteq> q \<longrightarrow>
                 (\<Gamma>2\<bar>x),(y:\<tau>),(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar q) (Suc 0) 
                                                      (freshen_tm w (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'" 
      proof -
        {
          fix w q
          assume Hwf: "(\<Gamma>2\<bar>x),(y:\<tau>),(w:\<tau>x),(q:\<tau>1);\<Delta> \<Turnstile> OK" and Hcomp1: "(\<Gamma>2\<bar>x),(y:\<tau>) \<sim> {w\<rhd>\<tau>x},(q:\<tau>1)"
            and Hcomp2: "{w\<rhd>\<tau>x} \<sim> {q\<rhd>\<tau>1}" and Hw: "w \<notin> (Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1)" and Hq: "q \<notin> (Ly\<union>{x}\<union>{y})" and Hneq: "w \<noteq> q"
          moreover
          have Hnin: "\<not> (\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1) defines x" using Hw Hq Hnotin by (auto simp add: dom_def)
          have Hwft: "\<Delta> \<Turnstile>' \<tau>" using poly_env_wf_ty Hwf' Hin by auto
          have "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1),(x:\<tau>);\<Delta> \<Turnstile> OK" apply (rule wf_env_cons[of "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1)" \<Delta> x \<tau>])
            using Hwf Hnin Hwft wf_env_subset[of "(\<Gamma>2\<bar>x),(y:\<tau>),(w:\<tau>x),(q:\<tau>1)" \<Delta> "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1)"] by auto
          hence Hwf: "\<Gamma>2,(w:\<tau>x),(q:\<tau>1);\<Delta> \<Turnstile> OK" using Hg1 by (auto simp add: env_permute)
          moreover
          have  "(\<Gamma>2\<bar>x),(x:\<tau>) \<sim> {w\<rhd>\<tau>x},(q:\<tau>1)" using Hcomp1 Hw Hq by (auto simp add: env_compat_def dom_def)
          hence Hcomp1: "\<Gamma>2 \<sim> {w\<rhd>\<tau>x},(q:\<tau>1)" using Hg1 by auto
          ultimately
          have Ht: "(\<Gamma>2\<bar>x),(x:\<tau>),(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar q) (Suc 0) 
                                                     (freshen_tm w (vsubst_tm (TmFreeVar x) (Suc (Suc i)) e2)) : \<tau>'" 
            using Hg1 Htyp2 by auto
          have "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1),(y:\<tau>); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) 
                                                     (freshen_tm w  e2)) : \<tau>'"
          proof (rule IH)
            note Ht
            also 
            have "vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w (vsubst_tm (TmFreeVar x) (Suc (Suc i)) e2)) =
                  vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (freshen_tm w  e2))"
              using subst_subst_swap[of "Suc (Suc i)" 0] by auto
            also
            have "vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (freshen_tm w  e2)) =
                  vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w  e2))"
              using subst_subst_swap[of "Suc (Suc i)" "Suc 0"] by auto;
            finally
            show "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1),(x:\<tau>);\<Delta>;M;Z2 \<turnstile> 
              vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w e2)) : \<tau>'"
              by (auto simp add: env_permute)
          next
            show "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1),(x:\<tau>);\<Delta> \<Turnstile> OK"
              using Hwf Hg1 by (auto simp add: env_permute)
          next
            show "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1) \<sim> {x\<rhd>\<tau>}"
              using Hq Hw Hnotin by (auto simp add: env_compat_def dom_def)
          next
            show "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1),(y:\<tau>);\<Delta> \<Turnstile> OK" using Hwf Hg1 by (auto simp add: env_permute)
          next
            have "\<Gamma>2\<bar>x \<subseteq> \<Gamma>" using H\<Gamma>x_eq Hin by (auto simp add: without_def)
            hence "\<Gamma>2\<bar>x \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto
            thus "(\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1) \<sim> {y\<rhd>\<tau>}" using Hq Hw by (auto simp add: env_compat_def)
          next
            show "x \<notin> dom ((\<Gamma>2\<bar>x),(w:\<tau>x),(q:\<tau>1))" using Hnotin Hq Hw by (auto simp add: dom_def)
          next
            show "x \<notin> set (free_tm (vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w e2)))"
              using not_free_subst Hq Hw "1.prems" TmLetbang by auto
          next
            show "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w e2) \<longrightarrow> scopeT \<tau> \<noteq> ScFree \<rho>"
              using "1.prems" TmLetbang free_dollars_vsubst_simpl by auto
          qed (insert TmLetbang, auto)
          also
          have "vsubst_tm (TmFreeVar y) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w  e2)) =
                vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar y) (Suc (Suc i)) (freshen_tm w  e2))"
            using subst_subst_swap[of "Suc (Suc i)" "Suc 0"] by auto
          also 
          have "vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar y) (Suc (Suc i)) (freshen_tm w  e2)) =
                vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2))" 
            using subst_subst_swap[of "Suc (Suc i)" 0] by auto
          finally
          have "(\<Gamma>2\<bar>x),(y:\<tau>),(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar q) (Suc 0) 
                                                     (freshen_tm w (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
            by (auto simp add: env_permute)
        }
        thus ?thesis by auto
      qed
    }
    hence Ht2l: "x\<rhd>\<tau> \<in> \<Gamma>2 \<longrightarrow> (\<forall>w q. (\<Gamma>2\<bar>x),(y:\<tau>),(w:\<tau>x),(q:\<tau>1);\<Delta> \<Turnstile> OK \<and> (\<Gamma>2\<bar>x),(y:\<tau>) \<sim> {w\<rhd>\<tau>x},(q:\<tau>1) \<and>
                         {w\<rhd>\<tau>x} \<sim> {q\<rhd>\<tau>1} \<and> w \<notin> (Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1) \<and> q \<notin> (Ly\<union>{x}\<union>{y}) \<and> w \<noteq> q \<longrightarrow>
                           (\<Gamma>2\<bar>x),(y:\<tau>),(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar q) (Suc 0) 
                                                      (freshen_tm w (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>')"
      by auto
    {
      fix w q
      assume Hwf: "\<Gamma>2,(w:\<tau>x),(q:\<tau>1);\<Delta> \<Turnstile> OK" and Hcomp1: "\<Gamma>2 \<sim> {w\<rhd>\<tau>x},(q:\<tau>1)" and Hcomp2: "{w\<rhd>\<tau>x} \<sim> {q\<rhd>\<tau>1}" 
        and Hw: "w \<notin> (Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1)" and Hq: "q \<notin> (Ly\<union>{x}\<union>{y})" and Hneq: "w \<noteq> q"
        and Hnin: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      hence Ht: "\<Gamma>2,(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar q) (Suc 0) 
                                                     (freshen_tm w (vsubst_tm (TmFreeVar x) (Suc (Suc i)) e2)) : \<tau>'" 
        using Htyp2 by auto
      also 
      have "vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w (vsubst_tm (TmFreeVar x) (Suc (Suc i)) e2)) =
            vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (freshen_tm w  e2))"
        using subst_subst_swap[of "Suc (Suc i)" 0] by auto
      also
      have "vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar x) (Suc (Suc i)) (freshen_tm w  e2)) =
            vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w  e2))"
        using subst_subst_swap[of "Suc (Suc i)" "Suc 0"] by auto
      finally
      have "\<Gamma>2,(w:\<tau>x),(q:\<tau>1);\<Delta>;M;Z2 \<turnstile> 
           vsubst_tm (TmFreeVar x) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w e2)) : \<tau>'" by auto
      moreover
      have "\<not> \<Gamma>2,(w:\<tau>x),(q:\<tau>1) defines x" using Hnin Hdom2 Hq Hw by auto
      ultimately
      have "\<Gamma>2,(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) 
          (freshen_tm w e2)) : \<tau>'"
        using not_in_env_ty_same_subst[of x "\<Gamma>2,(w:\<tau>x),(q:\<tau>1)"] by (auto simp add: dom_def)
      also
      have "vsubst_tm (TmFreeVar y) (Suc (Suc i)) (vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w  e2)) =
            vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar y) (Suc (Suc i)) (freshen_tm w  e2))"
        using subst_subst_swap[of "Suc (Suc i)" "Suc 0"] by auto
      also 
      have "vsubst_tm (TmFreeVar q) (Suc 0) (vsubst_tm (TmFreeVar y) (Suc (Suc i)) (freshen_tm w  e2)) =
            vsubst_tm (TmFreeVar q) (Suc 0) (freshen_tm w (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2))" 
        using subst_subst_swap[of "Suc (Suc i)" 0] by auto
      finally 
      have "\<Gamma>2,(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar q) (Suc 0) 
          (freshen_tm w (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'" by auto
    }
    hence Ht2o: "x\<rhd>\<tau> \<notin> \<Gamma>2 \<longrightarrow> (\<forall>w q. \<Gamma>2,(w:\<tau>x),(q:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {w\<rhd>\<tau>x},(q:\<tau>1) \<and>
                         {w\<rhd>\<tau>x} \<sim> {q\<rhd>\<tau>1} \<and> w \<notin> (Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1) \<and> q \<notin> (Ly\<union>{x}\<union>{y}) \<and> w \<noteq> q \<longrightarrow>
                           \<Gamma>2,(w:\<tau>x),(q:\<tau>1); \<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar q) (Suc 0) 
                                                      (freshen_tm w (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>')"
      by auto
    have Hd1: "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc i) e1 \<longrightarrow>  \<rho> \<notin> dom (Z2 \<union> Zx)"
      using Hdollar1 free_dollars_vsubst_simpl by auto
    have Hd2: "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2 \<longrightarrow> \<rho> \<notin> dom (Z1 \<union> Zx)"
      using Hdollar2 free_dollars_vsubst_simpl by auto
    have Hdx: "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e \<longrightarrow> \<rho> \<notin> dom (Z1 \<union> Z2)"
      using Hdollarx free_dollars_vsubst_simpl by auto
    from Hwf "1.prems" H\<Gamma>x_eq have "\<Gamma>1;\<Delta> \<Turnstile> OK" using wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta> \<Gamma>1] by auto
    hence "finite \<Gamma>1" by (auto simp add: wf_env_def)
    hence Hlx: "finite (Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1)" using dom_finite Hlx by auto
    {
      assume H: "x\<rhd>\<tau> \<notin> \<Gamma>x" and H1: "x\<rhd>\<tau> \<in> \<Gamma>1" and H2: "x\<rhd>\<tau> \<in> \<Gamma>2"
      have Hwf: "\<Gamma>2;\<Delta> \<Turnstile> OK" and Hwf_1:"\<Gamma>1;\<Delta> \<Turnstile> OK" 
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heq2: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" and Heq1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" using wf_env_defines_without H1 H2 by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x)" using H by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x" by (auto simp add: insert_ident)
      have "(\<Gamma>1\<bar>x),(x:\<tau>) \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq1 Heq2 Hcompat1 by auto
      hence Hqual: "qualT \<tau> \<preceq> qU" by (auto simp add: env_compat_def)
      have "(\<Gamma>x \<union> ((\<Gamma>1\<bar>x),(y:\<tau>)) \<union> ((\<Gamma>2\<bar>x),(y:\<tau>)));\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let! (vsubst_tm (TmFreeVar y) i e)
        (vsubst_tm (TmFreeVar y) (Suc i) e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'"
      proof (rule Typing.t_let[where Lx="Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1"])
        show  "Z1 \<Turnstile> ScopeSet"
          and "Z2 \<Turnstile> ScopeSet"
          and "validScope (scopeT \<tau>1) Z1"
          and "isBangable \<tau>x"
          and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc i) e1 \<longrightarrow>  \<rho> \<notin> dom (Z2 \<union> Zx)"
          and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2 \<longrightarrow> \<rho> \<notin> dom (Z1 \<union> Zx)"
          and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e \<longrightarrow> \<rho> \<notin> Environ.dom (Z1 \<union> Z2)"
          and "finite (Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1)"
          and "finite (Ly\<union>{x}\<union>{y})"
          using HZ1 HZ2 Hvalid Hbang Hdx Hd1 Hd2 Hlx Hly by auto
      next
        show "\<Gamma>x;\<Delta>;M;Zx \<turnstile> vsubst_tm (TmFreeVar y) i e : \<tau>x" using Htxo H by auto
      next
        show "\<forall>xa \<rho>. (\<Gamma>1\<bar>x),(y:\<tau>),(xa:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK \<and> (\<Gamma>1\<bar>x),(y:\<tau>) \<sim> {xa\<rhd>convT \<rho> \<tau>x} \<and> xa \<notin> Lx \<union> {x} \<union> {y}\<union>dom \<Gamma>1 \<and>
          \<rho> \<notin> set (free_let_dollar_scope (vsubst_tm (TmFreeVar y) (Suc i) e1)) \<and> \<rho> \<notin> Environ.dom (Z2 \<union> Z1) \<longrightarrow>
          (\<Gamma>1\<bar>x),(y:\<tau>),(xa:convT \<rho> \<tau>x);\<Delta>;M;Z1 \<turnstile> freshen_tm xa (freshen_sc \<rho> (vsubst_tm (TmFreeVar y) (Suc i) e1)) : \<tau>1 \<or>
          (\<Gamma>1\<bar>x),(y:\<tau>),(xa:convT \<rho> \<tau>x);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm xa (freshen_sc \<rho> (vsubst_tm (TmFreeVar y) (Suc i) e1)) : \<tau>1"
          using Ht1i H1 by auto
      next
        note Ht2l
        thus "\<forall>xa ya. (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta> \<Turnstile> OK \<and> (\<Gamma>2\<bar>x),(y:\<tau>) \<sim> {xa\<rhd>\<tau>x},(ya:\<tau>1) \<and> {xa\<rhd>\<tau>x} \<sim> {ya\<rhd>\<tau>1} \<and>
          xa \<notin> Lx \<union> {x} \<union> {y} \<union>dom \<Gamma>1 \<and> ya \<notin> Ly \<union> {x} \<union> {y} \<and> xa \<noteq> ya \<longrightarrow>
          (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar ya) (Suc 0)
          (freshen_tm xa (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
          using H2 by auto
      next
        have "(\<Gamma>1\<bar>x),(x:\<tau>) \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq1 Heq2 Hcompat1 by auto
        thus "(\<Gamma>1\<bar>x),(y:\<tau>) \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" by (auto simp add: env_compat_def)
      next
        have "\<Gamma>x \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq2 Hcompat2 by auto
        thus "\<Gamma>x \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using Hqual by (auto simp add: env_compat_def)
      next
        have "\<Gamma>x \<sim> (\<Gamma>1\<bar>x),(x:\<tau>)" using Heq1 Hcompat3 by auto
        thus "\<Gamma>x \<sim> (\<Gamma>1\<bar>x),(y:\<tau>)" using Hqual by (auto simp add: env_compat_def)
      next
        show "\<Gamma>x \<union> (\<Gamma>1\<bar>x),(y:\<tau>) \<union> (\<Gamma>2\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using H_eq prems by auto
      qed
      hence ?thesis using H_eq HZ_eq TmLetbang by auto
    }
    moreover
    {
      assume H: "x\<rhd>\<tau> \<notin> \<Gamma>x" and H1: "x\<rhd>\<tau> \<in> \<Gamma>1" and H2: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      have "\<Gamma>1;\<Delta> \<Turnstile> OK" 
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heq1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" using wf_env_defines_without H1 H2 by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2)" using H H2 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2" by (auto simp add: insert_ident)
      have "(\<Gamma>x \<union> ((\<Gamma>1\<bar>x),(y:\<tau>)) \<union> \<Gamma>2);\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let! (vsubst_tm (TmFreeVar y) i e)
        (vsubst_tm (TmFreeVar y) (Suc i) e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'"
      proof (rule Typing.t_let[where Lx="Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1"])
        show "Z1 \<Turnstile> ScopeSet"
          and "Z2 \<Turnstile> ScopeSet"
          and "validScope (scopeT \<tau>1) Z1"
          and "isBangable \<tau>x"
          and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc i) e1 \<longrightarrow>  \<rho> \<notin> dom (Z2 \<union> Zx)"
          and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2 \<longrightarrow> \<rho> \<notin> dom (Z1 \<union> Zx)"
          and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e \<longrightarrow> \<rho> \<notin> Environ.dom (Z1 \<union> Z2)"
          and "finite (Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1)"
          and "finite (Ly\<union>{x}\<union>{y})"
          using HZ1 HZ2 Hvalid Hbang Hdx Hd1 Hd2 Hlx Hly by auto
      next          
        show "\<Gamma>x;\<Delta>;M;Zx \<turnstile> vsubst_tm (TmFreeVar y) i e : \<tau>x" using Htxo H by auto
      next
        show "\<forall>xa \<rho>. (\<Gamma>1\<bar>x),(y:\<tau>),(xa:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK \<and> (\<Gamma>1\<bar>x),(y:\<tau>) \<sim> {xa\<rhd>convT \<rho> \<tau>x} \<and> xa \<notin> Lx \<union> {x} \<union> {y}\<union>dom \<Gamma>1 \<and>
          \<rho> \<notin> set (free_let_dollar_scope (vsubst_tm (TmFreeVar y) (Suc i) e1)) \<and> \<rho> \<notin> Environ.dom (Z2 \<union> Z1) \<longrightarrow>
          (\<Gamma>1\<bar>x),(y:\<tau>),(xa:convT \<rho> \<tau>x);\<Delta>;M;Z1 \<turnstile> freshen_tm xa (freshen_sc \<rho> (vsubst_tm (TmFreeVar y) (Suc i) e1)) : \<tau>1 \<or>
          (\<Gamma>1\<bar>x),(y:\<tau>),(xa:convT \<rho> \<tau>x);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm xa (freshen_sc \<rho> (vsubst_tm (TmFreeVar y) (Suc i) e1)) : \<tau>1"
          using Ht1i H1 by auto
      next
        note Ht2o
        thus "\<forall>xa ya. \<Gamma>2,(xa:\<tau>x),(ya:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {xa\<rhd>\<tau>x},(ya:\<tau>1) \<and> {xa\<rhd>\<tau>x} \<sim> {ya\<rhd>\<tau>1} \<and>
          xa \<notin> Lx \<union> {x} \<union> {y} \<union> dom \<Gamma>1 \<and> ya \<notin> Ly \<union> {x} \<union> {y}  \<and> xa \<noteq> ya \<longrightarrow>
          \<Gamma>2,(xa:\<tau>x),(ya:\<tau>1);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar ya) (Suc 0)
          (freshen_tm xa (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
          using H2 by auto
      next
        have "(\<Gamma>1\<bar>x),(x:\<tau>) \<sim> \<Gamma>2" using Heq1 Hcompat1 by auto
        hence Hc1: "(\<Gamma>1\<bar>x) \<sim> \<Gamma>2" using env_compat_subset by auto
        have "\<Gamma>2 \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
        hence "\<Gamma>2 \<subseteq> \<Gamma>" using H2 by auto
        hence "\<Gamma>2 \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto 
        hence "{y\<rhd>\<tau>} \<sim> \<Gamma>2" using env_compat_symm by auto
        thus "(\<Gamma>1\<bar>x),(y:\<tau>) \<sim> \<Gamma>2" using Hc1 env_compat_union_right[of "(\<Gamma>1\<bar>x)" _ "{y\<rhd>\<tau>}"] by auto
      next
        show "\<Gamma>x \<sim> \<Gamma>2" by assumption
      next
        have "\<Gamma>x \<sim> (\<Gamma>1\<bar>x),(x:\<tau>)" using Heq1 Hcompat3 by auto
        hence Hc: "\<Gamma>x \<sim> (\<Gamma>1\<bar>x)" using env_compat_subset env_compat_symm by blast
        have "\<Gamma>x \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
        hence "\<Gamma>x \<subseteq> \<Gamma>" using H by auto
        hence "\<Gamma>x \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto
        thus "\<Gamma>x \<sim> (\<Gamma>1\<bar>x),(y:\<tau>)" using Hc env_compat_union_left[of _ "(\<Gamma>1\<bar>x)" "{y\<rhd>\<tau>}"] by auto
      next
        show "\<Gamma>x \<union> (\<Gamma>1\<bar>x),(y:\<tau>) \<union> \<Gamma>2;\<Delta> \<Turnstile> OK" using H_eq prems by auto
      qed
      hence ?thesis using H_eq HZ_eq TmLetbang by auto
    }
    moreover
    {
      assume H: "x\<rhd>\<tau> \<notin> \<Gamma>x" and H1: "x\<rhd>\<tau> \<notin> \<Gamma>1" and H2: "x\<rhd>\<tau> \<in> \<Gamma>2"
      have "\<Gamma>2;\<Delta> \<Turnstile> OK" 
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heq2: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" using wf_env_defines_without H1 H2 by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>x \<union> \<Gamma>1 \<union> \<Gamma>2\<bar>x),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>x \<union> \<Gamma>1 \<union> \<Gamma>2\<bar>x)" using H H1 H2 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>x \<union> \<Gamma>1 \<union> \<Gamma>2\<bar>x" by (auto simp add: insert_ident)
      have "(\<Gamma>x \<union> \<Gamma>1 \<union> ((\<Gamma>2\<bar>x),(y:\<tau>)));\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let! (vsubst_tm (TmFreeVar y) i e)
        (vsubst_tm (TmFreeVar y) (Suc i) e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'"
      proof (rule Typing.t_let[where Lx="Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1"])
        show "Z1 \<Turnstile> ScopeSet"
          and "Z2 \<Turnstile> ScopeSet"
          and "validScope (scopeT \<tau>1) Z1"
          and "isBangable \<tau>x"
          and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc i) e1 \<longrightarrow>  \<rho> \<notin> dom (Z2 \<union> Zx)"
          and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2 \<longrightarrow> \<rho> \<notin> dom (Z1 \<union> Zx)"
          and "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar y) i e \<longrightarrow> \<rho> \<notin> Environ.dom (Z1 \<union> Z2)"
          and "finite (Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1)"
          and "finite (Ly\<union>{x}\<union>{y})"
          using HZ1 HZ2 Hvalid Hbang Hdx Hd1 Hd2 Hlx Hly by auto
      next
        show "\<Gamma>x;\<Delta>;M;Zx \<turnstile> vsubst_tm (TmFreeVar y) i e : \<tau>x" using Htxo H by auto
      next
        show "\<forall>xa \<rho>. \<Gamma>1,(xa:convT \<rho> \<tau>x);\<Delta> \<Turnstile> OK \<and> \<Gamma>1 \<sim> {xa\<rhd>convT \<rho> \<tau>x} \<and>  xa \<notin> Lx \<union> {x} \<union> {y} \<union> Environ.dom \<Gamma>1 \<and>
           \<rho> \<notin> set (free_let_dollar_scope (vsubst_tm (TmFreeVar y) (Suc i) e1)) \<and> \<rho> \<notin> Environ.dom (Z2 \<union> Z1) \<longrightarrow>
          \<Gamma>1,(xa:convT \<rho> \<tau>x);\<Delta>;M;Z1 \<turnstile> freshen_tm xa (freshen_sc \<rho> (vsubst_tm (TmFreeVar y) (Suc i) e1)) : \<tau>1 \<or>
          \<Gamma>1,(xa:convT \<rho> \<tau>x);\<Delta>;M;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm xa (freshen_sc \<rho> (vsubst_tm (TmFreeVar y) (Suc i) e1)) : \<tau>1"
          using Ht1o H1 by auto
      next      
        note Ht2l
        thus "\<forall>xa ya. (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta> \<Turnstile> OK \<and> (\<Gamma>2\<bar>x),(y:\<tau>) \<sim> {xa\<rhd>\<tau>x},(ya:\<tau>1) \<and> {xa\<rhd>\<tau>x} \<sim> {ya\<rhd>\<tau>1} \<and>
          xa \<notin> Lx \<union> {x} \<union> {y} \<union> dom \<Gamma>1 \<and> ya \<notin> Ly \<union> {x} \<union> {y} \<and> xa \<noteq> ya \<longrightarrow>
          (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar ya) (Suc 0)
            (freshen_tm xa (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
          using H2 by auto
      next
        have "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq2 Hcompat1 by auto
        hence Hc1: "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x)" using env_compat_subset env_compat_symm by blast
        have "\<Gamma>1 \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
        hence "\<Gamma>1 \<subseteq> \<Gamma>" using H1 by auto
        hence "\<Gamma>1 \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto 
        thus "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using Hc1 env_compat_union_left[of \<Gamma>1 _ "{y\<rhd>\<tau>}"] by auto
      next
        have "\<Gamma>x \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq2 Hcompat2 by auto
        hence Hc1: "\<Gamma>x \<sim> (\<Gamma>2\<bar>x)" using env_compat_subset env_compat_symm by blast
        have "\<Gamma>x \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
        hence "\<Gamma>x \<subseteq> \<Gamma>" using H by auto
        hence "\<Gamma>x \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto 
        thus "\<Gamma>x \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using Hc1 env_compat_union_left[of \<Gamma>x _ "{y\<rhd>\<tau>}"] by auto
      next
        show "\<Gamma>x \<sim> \<Gamma>1" by assumption
      next
        show "\<Gamma>x \<union> \<Gamma>1 \<union> (\<Gamma>2\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using "1.prems" H_eq by auto
      qed
      hence ?thesis using H_eq HZ_eq TmLetbang by auto
    }
    moreover
    {
      assume H: "x\<rhd>\<tau> \<in> \<Gamma>x" and H1: "x\<rhd>\<tau> \<in> \<Gamma>1" and H2: "x\<rhd>\<tau> \<in> \<Gamma>2"
      have "\<Gamma>2;\<Delta> \<Turnstile> OK" and "\<Gamma>1;\<Delta> \<Turnstile> OK" and "\<Gamma>x;\<Delta> \<Turnstile> OK"
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heqx: "\<Gamma>x = (\<Gamma>x\<bar>x),(x:\<tau>)" and Heq2: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)" and Heq1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" 
        using wf_env_defines_without H1 H2 H by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>x\<bar>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>x\<bar>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x)" using H by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>x\<bar>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2\<bar>x" by (auto simp add: insert_ident)
      have "(\<Gamma>1\<bar>x),(x:\<tau>) \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq1 Heq2 Hcompat1 by auto
      hence Hqual: "qualT \<tau> \<preceq> qU" by (auto simp add: env_compat_def)
      have "((\<Gamma>x\<bar>x),(y:\<tau>) \<union> ((\<Gamma>1\<bar>x),(y:\<tau>)) \<union> ((\<Gamma>2\<bar>x),(y:\<tau>)));\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let! (vsubst_tm (TmFreeVar y) i e)
        (vsubst_tm (TmFreeVar y) (Suc i) e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'"
      proof (rule Typing.t_let[where Lx="Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1"])
        show "(\<Gamma>x\<bar>x),(y:\<tau>);\<Delta>;M;Zx \<turnstile> vsubst_tm (TmFreeVar y) i e : \<tau>x" using Htxi H by auto
      next
        note Ht2l
        thus "\<forall>xa ya. (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta> \<Turnstile> OK \<and> (\<Gamma>2\<bar>x),(y:\<tau>) \<sim> {xa\<rhd>\<tau>x},(ya:\<tau>1) \<and> {xa\<rhd>\<tau>x} \<sim> {ya\<rhd>\<tau>1} \<and>
          xa \<notin> Lx \<union> {x} \<union> {y} \<union> dom \<Gamma>1 \<and> ya \<notin> Ly \<union> {x} \<union> {y} \<and> xa \<noteq> ya \<longrightarrow>
          (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar ya) (Suc 0)
          (freshen_tm xa (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
          using H2 by auto
      next
        have "(\<Gamma>1\<bar>x),(x:\<tau>) \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq1 Heq2 Hcompat1 by auto
        thus "(\<Gamma>1\<bar>x),(y:\<tau>) \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" by (auto simp add: env_compat_def)
      next
        have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heq2 Hcompat2 Heqx by auto
        thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using Hqual by (auto simp add: env_compat_def)
      next
        have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> (\<Gamma>1\<bar>x),(x:\<tau>)" using Heq1 Hcompat3 Heqx by auto
        thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> (\<Gamma>1\<bar>x),(y:\<tau>)" using Hqual by (auto simp add: env_compat_def)
      next
        show "(\<Gamma>x\<bar>x),(y:\<tau>) \<union> (\<Gamma>1\<bar>x),(y:\<tau>) \<union> (\<Gamma>2\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using "1.prems" H_eq by auto
      qed (insert Htxi Htxo H Ht1i Ht1o H1 Ht2l Ht2o H2 HZ1 HZ2 Hvalid Hbang Hdx Hd1 Hd2 Hlx Hly, auto)
      hence ?thesis using H_eq HZ_eq TmLetbang by auto
    }
    moreover
    {
      assume H: "x\<rhd>\<tau> \<in> \<Gamma>x" and H1: "x\<rhd>\<tau> \<in> \<Gamma>1" and H2: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      have "\<Gamma>1;\<Delta> \<Turnstile> OK" and "\<Gamma>x;\<Delta> \<Turnstile> OK"
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heqx: "\<Gamma>x = (\<Gamma>x\<bar>x),(x:\<tau>)" and Heq1: "\<Gamma>1 = (\<Gamma>1\<bar>x),(x:\<tau>)" 
        using wf_env_defines_without H1 H2 H by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>x\<bar>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>x\<bar>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2)" using H2 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>x\<bar>x \<union> \<Gamma>1\<bar>x \<union> \<Gamma>2" by (auto simp add: insert_ident)
      have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> (\<Gamma>1\<bar>x),(x:\<tau>)" using Heqx Heq1 Hcompat3 by auto
      hence Hqual: "qualT \<tau> \<preceq> qU" by (auto simp add: env_compat_def)
      have "((\<Gamma>x\<bar>x),(y:\<tau>) \<union> ((\<Gamma>1\<bar>x),(y:\<tau>)) \<union> \<Gamma>2);\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let! (vsubst_tm (TmFreeVar y) i e)
        (vsubst_tm (TmFreeVar y) (Suc i) e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'"
      proof (rule Typing.t_let[where Lx="Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1"])
        show "(\<Gamma>x\<bar>x),(y:\<tau>);\<Delta>;M;Zx \<turnstile> vsubst_tm (TmFreeVar y) i e : \<tau>x" using Htxi H by auto
      next
        note Ht2o
        thus "\<forall>xa ya. \<Gamma>2,(xa:\<tau>x),(ya:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {xa\<rhd>\<tau>x},(ya:\<tau>1) \<and> {xa\<rhd>\<tau>x} \<sim> {ya\<rhd>\<tau>1} \<and>
                      xa \<notin> Lx \<union> {x} \<union> {y} \<union> dom \<Gamma>1 \<and> ya \<notin> Ly \<union> {x} \<union> {y} \<and> xa \<noteq> ya \<longrightarrow>
              \<Gamma>2,(xa:\<tau>x),(ya:\<tau>1);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar ya) (Suc 0)
                                                   (freshen_tm xa (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
          using H2 by auto
      next
        have "(\<Gamma>1\<bar>x),(x:\<tau>) \<sim> \<Gamma>2" using Heq1 Hcompat1 by auto
        hence Hc1: "(\<Gamma>1\<bar>x) \<sim> \<Gamma>2" using env_compat_subset by auto
        have "\<Gamma>2 \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
        hence "\<Gamma>2 \<subseteq> \<Gamma>" using H2 by auto
        hence "\<Gamma>2 \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto 
        hence "{y\<rhd>\<tau>} \<sim> \<Gamma>2" using env_compat_symm by auto
        thus "(\<Gamma>1\<bar>x),(y:\<tau>) \<sim> \<Gamma>2" using Hc1 env_compat_union_right[of "(\<Gamma>1\<bar>x)" _ "{y\<rhd>\<tau>}"] by auto
      next
        have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> \<Gamma>2" using Hcompat2 Heqx by auto
        thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> \<Gamma>2" using Hqual by (auto simp add: env_compat_def)
      next
        have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> (\<Gamma>1\<bar>x),(x:\<tau>)" using Heq1 Hcompat3 Heqx by auto
        thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> (\<Gamma>1\<bar>x),(y:\<tau>)" using Hqual by (auto simp add: env_compat_def)
      next
        show "(\<Gamma>x\<bar>x),(y:\<tau>) \<union> (\<Gamma>1\<bar>x),(y:\<tau>) \<union> \<Gamma>2;\<Delta> \<Turnstile> OK" using "1.prems" H_eq by auto
      qed (insert Htxi Htxo H Ht1i Ht1o H1 Ht2l Ht2o H2 HZ1 HZ2 Hvalid Hbang Hdx Hd1 Hd2 Hlx Hly, auto)
      hence ?thesis using H_eq HZ_eq TmLetbang by auto
    }
    moreover
    {
      assume H: "x\<rhd>\<tau> \<in> \<Gamma>x" and H1: "x\<rhd>\<tau> \<notin> \<Gamma>1" and H2: "x\<rhd>\<tau> \<in> \<Gamma>2"
      have "\<Gamma>x;\<Delta> \<Turnstile> OK" and "\<Gamma>2;\<Delta> \<Turnstile> OK" 
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heqx: "\<Gamma>x = (\<Gamma>x\<bar>x),(x:\<tau>)" and Heq2: "\<Gamma>2 = (\<Gamma>2\<bar>x),(x:\<tau>)"
        using wf_env_defines_without H1 H2 H by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>x\<bar>x \<union> \<Gamma>1 \<union> \<Gamma>2\<bar>x),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>x\<bar>x \<union> \<Gamma>1 \<union> \<Gamma>2\<bar>x)" using H2 H1 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>x\<bar>x \<union> \<Gamma>1 \<union> \<Gamma>2\<bar>x" by (auto simp add: insert_ident)
      have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heqx Heq2 Hcompat1 by auto
      hence Hqual: "qualT \<tau> \<preceq> qU" by (auto simp add: env_compat_def)
      have "((\<Gamma>x\<bar>x),(y:\<tau>) \<union> \<Gamma>1 \<union> (\<Gamma>2\<bar>x),(y:\<tau>));\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let! (vsubst_tm (TmFreeVar y) i e)
        (vsubst_tm (TmFreeVar y) (Suc i) e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'"
      proof (rule Typing.t_let[where Lx="Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1"])
        show "(\<Gamma>x\<bar>x),(y:\<tau>);\<Delta>;M;Zx \<turnstile> vsubst_tm (TmFreeVar y) i e : \<tau>x" using Htxi H by auto
      next
        note Ht2l
        thus "\<forall>xa ya. (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta> \<Turnstile> OK \<and> (\<Gamma>2\<bar>x),(y:\<tau>) \<sim> {xa\<rhd>\<tau>x},(ya:\<tau>1) \<and> {xa\<rhd>\<tau>x} \<sim> {ya\<rhd>\<tau>1} \<and>
                        xa \<notin> Lx \<union> {x} \<union> {y} \<union> dom \<Gamma>1 \<and> ya \<notin> Ly \<union> {x} \<union> {y} \<and> xa \<noteq> ya \<longrightarrow>
                (\<Gamma>2\<bar>x),(y:\<tau>),(xa:\<tau>x),(ya:\<tau>1);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar ya) (Suc 0)
                                                      (freshen_tm xa (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
          using H2 by auto
      next
        show "\<Gamma>1 \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using Heq2 Hcompat1 Hqual by (auto simp add: env_compat_def)
      next
        have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> \<Gamma>1" using Heqx Hcompat3 by auto
        hence Hc: "(\<Gamma>x\<bar>x) \<sim> \<Gamma>1" using env_compat_subset env_compat_symm by blast
        have "\<Gamma>1 \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
        hence "\<Gamma>1 \<subseteq> \<Gamma>" using H1 by auto
        hence "\<Gamma>1 \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto
        hence "\<Gamma>1 \<sim> (\<Gamma>x\<bar>x),(y:\<tau>)" using Hc env_compat_union_left[of \<Gamma>1 "(\<Gamma>x\<bar>x)" "{y\<rhd>\<tau>}"] env_compat_symm by auto
        thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> \<Gamma>1" using env_compat_symm by auto
      next
        have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> (\<Gamma>2\<bar>x),(x:\<tau>)" using Heqx Heq2 Hcompat2 by auto
        thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> (\<Gamma>2\<bar>x),(y:\<tau>)" using Hqual by (auto simp add: env_compat_def)
      next
        show "(\<Gamma>x\<bar>x),(y:\<tau>) \<union> \<Gamma>1 \<union> (\<Gamma>2\<bar>x),(y:\<tau>);\<Delta> \<Turnstile> OK" using "1.prems" H_eq by auto 
      qed (insert Htxi Htxo H Ht1i Ht1o H1 Ht2l Ht2o H2 HZ1 HZ2 Hvalid Hbang Hdx Hd1 Hd2 Hlx Hly, auto)
      hence ?thesis using H_eq HZ_eq TmLetbang by auto
    }
    moreover
    {
      assume H: "x\<rhd>\<tau> \<in> \<Gamma>x" and H1: "x\<rhd>\<tau> \<notin> \<Gamma>1" and H2: "x\<rhd>\<tau> \<notin> \<Gamma>2"
      have "\<Gamma>x;\<Delta> \<Turnstile> OK"
        using "1.prems" H\<Gamma>x_eq wf_env_subset[of "\<Gamma>,(x:\<tau>)" \<Delta>] by auto
      hence Heqx: "\<Gamma>x = (\<Gamma>x\<bar>x),(x:\<tau>)"
        using wf_env_defines_without H1 H2 H by auto
      hence "\<Gamma>,(x:\<tau>) = (\<Gamma>x\<bar>x \<union> \<Gamma>1 \<union> \<Gamma>2),(x:\<tau>)" using H\<Gamma>x_eq by auto
      moreover
      have "x\<rhd>\<tau> \<notin> (\<Gamma>x\<bar>x \<union> \<Gamma>1 \<union> \<Gamma>2)" using H2 H1 by (auto simp add: without_def dom_def)
      moreover
      have "\<not> \<Gamma> defines x" by assumption
      hence "x\<rhd>\<tau> \<notin> \<Gamma>" by (auto simp add: without_def dom_def)
      ultimately
      have H_eq: "\<Gamma> = \<Gamma>x\<bar>x \<union> \<Gamma>1 \<union> \<Gamma>2" by (auto simp add: insert_ident)
      have "((\<Gamma>x\<bar>x),(y:\<tau>) \<union> \<Gamma>1 \<union> \<Gamma>2);\<Delta>;M;Z2\<union>Z1\<union>Zx \<turnstile> let! (vsubst_tm (TmFreeVar y) i e)
        (vsubst_tm (TmFreeVar y) (Suc i )e1) in (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2) : \<tau>'"
      proof (rule Typing.t_let[where Lx="Lx\<union>{x}\<union>{y}\<union>dom \<Gamma>1"])
        show "(\<Gamma>x\<bar>x),(y:\<tau>);\<Delta>;M;Zx \<turnstile> vsubst_tm (TmFreeVar y) i e : \<tau>x" using Htxi H by auto
        note Ht2o
        thus "\<forall>xa ya. \<Gamma>2,(xa:\<tau>x),(ya:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {xa\<rhd>\<tau>x},(ya:\<tau>1) \<and> {xa\<rhd>\<tau>x} \<sim> {ya\<rhd>\<tau>1} \<and>
                      xa \<notin> Lx \<union> {x} \<union> {y} \<union> dom \<Gamma>1 \<and> ya \<notin> Ly \<union> {x} \<union> {y} \<and> xa \<noteq> ya \<longrightarrow>
              \<Gamma>2,(xa:\<tau>x),(ya:\<tau>1);\<Delta>;M;Z2 \<turnstile> vsubst_tm (TmFreeVar ya) (Suc 0)
                                                    (freshen_tm xa (vsubst_tm (TmFreeVar y) (Suc (Suc i)) e2)) : \<tau>'"
            using H2 by auto
        next
          show "\<Gamma>1 \<sim> \<Gamma>2" by assumption 
        next
          have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> \<Gamma>1" using Heqx Hcompat3 by auto
          hence Hc: "(\<Gamma>x\<bar>x) \<sim> \<Gamma>1" using env_compat_subset env_compat_symm by blast
          have "\<Gamma>1 \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
          hence "\<Gamma>1 \<subseteq> \<Gamma>" using H1 by auto
          hence "\<Gamma>1 \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto
          hence "\<Gamma>1 \<sim> (\<Gamma>x\<bar>x),(y:\<tau>)" using Hc env_compat_union_left[of \<Gamma>1 "(\<Gamma>x\<bar>x)" "{y\<rhd>\<tau>}"] env_compat_symm by auto
          thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> \<Gamma>1" using env_compat_symm by auto
        next
          have "(\<Gamma>x\<bar>x),(x:\<tau>) \<sim> \<Gamma>2" using Heqx Hcompat2 by auto
          hence Hc: "(\<Gamma>x\<bar>x) \<sim> \<Gamma>2" using env_compat_subset env_compat_symm by blast
          have "\<Gamma>2 \<subseteq> \<Gamma>,(x:\<tau>)" using H\<Gamma>x_eq by auto
          hence "\<Gamma>2 \<subseteq> \<Gamma>" using H2 by auto
          hence "\<Gamma>2 \<sim> {y\<rhd>\<tau>}" using "1.prems" env_compat_subset by auto
          hence "\<Gamma>2 \<sim> (\<Gamma>x\<bar>x),(y:\<tau>)" using Hc env_compat_union_left[of \<Gamma>2 "(\<Gamma>x\<bar>x)" "{y\<rhd>\<tau>}"] env_compat_symm by auto
          thus "(\<Gamma>x\<bar>x),(y:\<tau>) \<sim> \<Gamma>2" using env_compat_symm by auto
        next
          show "(\<Gamma>x\<bar>x),(y:\<tau>) \<union> \<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK" using "1.prems" H_eq by auto
        qed (insert Htxi Htxo H Ht1i Ht1o H1 Ht2l Ht2o H2 HZ1 HZ2 Hvalid Hbang Hdx Hd1 Hd2 Hlx Hly, auto)
        hence ?thesis using H_eq HZ_eq TmLetbang by auto
      }
      moreover
      {
        assume H: "x\<rhd>\<tau> \<notin> \<Gamma>x" and  H1: "x\<rhd>\<tau> \<notin> \<Gamma>1" and H2: "x\<rhd>\<tau> \<notin> \<Gamma>2"
        hence "\<not> \<Gamma>x defines x" and  "\<not> \<Gamma>1 defines x" and "\<not> \<Gamma>2 defines x" using Hdomx Hdom1 Hdom2 by auto
        hence ?thesis using H\<Gamma>x_eq by (auto simp add: dom_def)
      }
      ultimately
      show ?thesis by auto
    qed
  qed

lemma qualG_vsubst_left:
  assumes "qualG (vsubst_ty_env (TyFreeVar a) i \<Gamma>) \<preceq> q"
  shows "qualG (vsubst_ty_env \<phi>s i \<Gamma>) \<preceq> q"
using prems proof (auto simp add: qualGrelL_def vsubst_ty_env_def)
  fix x \<tau>
  assume "\<forall>x \<tau>. (\<exists>\<tau>i. \<tau> = vsubst_ty (TyFreeVar a) i \<tau>i \<and> x\<rhd>\<tau>i \<in> \<Gamma>) \<longrightarrow> qualT \<tau> \<preceq> q"
    and "x\<rhd>\<tau> \<in> \<Gamma>"
  hence "qualT (vsubst_ty (TyFreeVar a) i \<tau>) \<preceq> q" by blast
  moreover
  have "qualT (vsubst_ty (TyFreeVar a) i \<tau>) = qualT (vsubst_ty \<phi>s i \<tau>)" by (cases \<tau>, auto)
  ultimately
  show "qualT (vsubst_ty \<phi>s i \<tau>) \<preceq> q" by auto
qed

lemma qualG_vsubst_left_special:
  assumes "qualG ((vsubst_ty_env (TyFreeVar a) i \<Gamma>)\<bar>y) \<preceq> q" 
  shows "qualG ((vsubst_ty_env \<phi>s i \<Gamma>)\<bar>y) \<preceq> q"
using prems proof (auto simp add: qualGrelL_def vsubst_ty_env_def without_def)
  fix x \<tau>
  assume "\<forall>x \<tau>. (\<exists>\<tau>i. \<tau> = vsubst_ty (TyFreeVar a) i \<tau>i \<and> x\<rhd>\<tau>i \<in> \<Gamma>) \<and> x \<noteq> y \<longrightarrow> qualT \<tau> \<preceq> q"
    and "x \<noteq> y"
    and "x\<rhd>\<tau> \<in> \<Gamma>"
  hence "qualT (vsubst_ty (TyFreeVar a) i \<tau>) \<preceq> q" by blast
  moreover
  have "qualT (vsubst_ty (TyFreeVar a) i \<tau>) = qualT (vsubst_ty \<phi>s i \<tau>)" by (cases \<tau>, auto)
  ultimately
  show"qualT (vsubst_ty \<phi>s i \<tau>) \<preceq> q" by auto
qed

lemma vsubst_ty_free_db_pty:
  assumes "free_db_pty \<phi>s = []"
  shows "set (free_db_pty (vsubst_pty \<phi>s k \<phi>)) = set (free_db_pty \<phi>) - {k}"
using vsubst_ty_free_db_ty prems by (induct \<phi> arbitrary: k taking: 
 "%\<tau>. \<forall>k. free_db_pty \<phi>s = [] \<longrightarrow> set (free_db_ty (vsubst_ty \<phi>s k \<tau>)) = set (free_db_ty \<tau>) - {k}",
  auto simp add: shift_many_sanity)

lemma poly_subst_var_pretype_ok:
  fixes \<phi>s :: pty
  assumes "\<Delta> \<Turnstile>' vsubst_pty (TyFreeVar z) i \<phi>"
  and "\<Delta> \<Turnstile>' \<phi>s"
  shows "\<Delta> \<Turnstile>' vsubst_pty \<phi>s i \<phi>"
proof -
  obtain \<tau> where H: "\<tau> = Qual q \<pi> \<phi>" by auto
  hence "\<Delta> \<Turnstile>' vsubst_ty (TyFreeVar z) i \<tau>" and "\<Delta> \<Turnstile>' \<phi>s" using assms by auto
  hence "\<Delta> \<Turnstile>' vsubst_ty \<phi>s i \<tau>" using poly_subst_var_type_ok by auto
  thus ?thesis using H by auto
qed

lemma poly_wf_pty_strength:
  fixes \<phi> :: pty
  assumes "\<Delta>,(z:\<star>) \<Turnstile>' \<phi>"
  and "\<not> z free in pretype \<phi>"
  shows "\<Delta> \<Turnstile>' \<phi>"
proof -
  obtain \<tau> where H: "\<tau> = Qual q \<pi> \<phi>" by auto
  hence "\<Delta>,(z:\<star>) \<Turnstile>' \<tau>" and "\<not> z free in type \<tau>" using assms by auto
  hence "\<Delta> \<Turnstile>' \<tau>" using poly_wf_ty_strength by auto
  thus ?thesis using H by auto
qed

lemma in_env_exists_sanity:
  fixes \<Gamma> :: env
  assumes "y\<rhd>\<tau> \<in> vsubst_ty_env (TyFreeVar x) i \<Gamma>"
  shows "\<exists>\<tau>'. \<tau> = vsubst_ty (TyFreeVar x) i \<tau>' \<and> y\<rhd>\<tau>' \<in> \<Gamma>"
proof -
  have "(vsubst_ty_env (TyFreeVar x) i \<Gamma>) defines y" using prems by (auto simp add: dom_def)
  hence "\<Gamma> defines y" using vsubst_ty_env_dom_sanity by auto
  then obtain b where "(y, b) \<in> \<Gamma>" using defines_sanity[of y \<Gamma>] by auto
  then obtain \<tau>' where "b = Term \<tau>'" by (cases b)
  thus ?thesis using exI[of _ \<tau>'] using vsubst_ty_env_def prems by auto
qed

lemma subst_subst_equal_tyty:
  assumes "vsubst_ty (TyFreeVar a) i \<tau>1 = vsubst_ty (TyFreeVar a) i \<tau>2"
  and "\<not> a free in type \<tau>1"
  and "\<not> a free in type \<tau>2"
  shows "\<tau>1 = \<tau>2"
using prems proof (induct \<tau>1 arbitrary: i \<tau>2
taking: "%\<phi>1. \<forall>i \<phi>2. vsubst_pty (TyFreeVar a) i \<phi>1 = vsubst_pty (TyFreeVar a) i \<phi>2 \<longrightarrow>
  \<not> a free in pretype \<phi>1 \<longrightarrow> \<not> a free in pretype \<phi>2 \<longrightarrow> \<phi>1 = \<phi>2")
  case (Qual q \<pi> \<phi>) 
  obtain \<phi>2 q2 \<pi>2 where Heq: "\<tau>2 = Qual q2 \<pi>2 \<phi>2" by (cases \<tau>2, auto)
  hence Hq: "q2 = q" and H\<pi>: "\<pi>2 = \<pi>" using Qual by auto
  hence "vsubst_ty (TyFreeVar a) i (Qual q \<pi> \<phi>) = vsubst_ty (TyFreeVar a) i (Qual q \<pi> \<phi>2)"
    using Qual Heq by auto
  hence "vsubst_pty (TyFreeVar a) i \<phi> = vsubst_pty (TyFreeVar a) i \<phi>2" by auto
  moreover
  have "\<not> a free in pretype \<phi>" using Qual by auto
  moreover
  have "\<not> a free in pretype \<phi>2" using Qual Heq by auto
  ultimately
  have "\<phi> = \<phi>2" using Qual by auto
  thus ?case using Heq Hq H\<pi> by auto
next
  case TyUnit thus ?case proof auto
    fix i \<phi>2
    assume "TyUnit = vsubst_pty (TyFreeVar a) i \<phi>2"
    thus "TyUnit = \<phi>2" by (cases \<phi>2, auto simp add: split_if)
  qed
next
  case (TyBase n)
  {
    fix i \<phi>2
    assume "TyBase n = vsubst_pty (TyFreeVar a) i \<phi>2" 
    hence "TyBase n = \<phi>2" by (cases \<phi>2, auto simp add: split_if)
  }
  thus ?case by auto
next
  case (TyFreeVar z) 
  {
    fix i \<phi>2
    assume "TyFreeVar z = vsubst_pty (TyFreeVar a) i \<phi>2"
      and "\<not> a free in pretype (TyFreeVar z)"
    hence "TyFreeVar z = \<phi>2" by (cases \<phi>2, auto simp add: split_if)
  }
  thus ?case by auto
next
  case (TyVar y) 
  { 
    fix i \<phi>2
    assume Heq: "vsubst_pty (TyFreeVar a) i (TyVar y) = vsubst_pty (TyFreeVar a) i \<phi>2"
    and "a \<notin> set (free_pty (TyVar y))" and Hfr: "a \<notin> set (free_pty \<phi>2)"
    have "TyVar y = \<phi>2" proof (cases "y=i")
      case False
      thus ?thesis using Heq by (cases \<phi>2, auto)
    next
      case True
      have "\<phi>2 \<noteq> TyFreeVar a" using Hfr by auto
      moreover have "TyFreeVar a = vsubst_pty (TyFreeVar a) i \<phi>2" using TyVar Heq True by auto
      ultimately show ?thesis using True by (cases \<phi>2, auto simp add: split_if)
    qed
  }
  thus ?case by auto
next
  case (TyArrow \<tau>a X \<tau>r) show ?case proof auto
    fix i \<phi>2
    assume Heq: "vsubst_ty (TyFreeVar a) i \<tau>a \<rightarrow>^X vsubst_ty (TyFreeVar a) i \<tau>r =
            vsubst_pty (TyFreeVar a) i \<phi>2"
      and Ha: "\<not> a free in type \<tau>a" and Hr: "\<not> a free in type \<tau>r" and H\<phi>: "\<not> a free in pretype \<phi>2"
    thus "\<tau>a \<rightarrow>^X \<tau>r = \<phi>2" proof (cases \<phi>2)
      case (TyArrow \<tau>1 X' \<tau>2)
      have Hih1: "vsubst_ty (TyFreeVar a) i \<tau>a = vsubst_ty (TyFreeVar a) i \<tau>1" using TyArrow Heq by auto
      have Hih2: "vsubst_ty (TyFreeVar a) i \<tau>r = vsubst_ty (TyFreeVar a) i \<tau>2" using TyArrow Heq by auto
      have Hg1: "X = X'" using TyArrow Heq by auto
      have Hf1: "\<not> a free in type \<tau>1" and Hf2: "\<not> a free in type \<tau>2" using H\<phi> TyArrow by auto
      have Hg2: "\<tau>a = \<tau>1" using TyArrow Hih1 Hf1 prems by auto
      have Hg3: "\<tau>r = \<tau>2" using TyArrow Hih2 Hf2 prems by auto
      show ?thesis using Hg1 Hg2 Hg3 TyArrow by auto
    qed (auto simp add: split_if)
  qed
next
  case (TyRef \<tau>ins) show ?case proof auto
    fix i \<phi>2
    assume Heq: "TyRef (vsubst_ty (TyFreeVar a) i \<tau>ins) = vsubst_pty (TyFreeVar a) i \<phi>2"
      and Hf: "\<not> a free in type \<tau>ins" and H\<phi>: "\<not> a free in pretype \<phi>2"
    thus "TyRef \<tau>ins = \<phi>2" proof (cases \<phi>2)
      case (TyRef \<tau>2)
      have Hih1: "vsubst_ty (TyFreeVar a) i \<tau>ins = vsubst_ty (TyFreeVar a) i \<tau>2" using TyRef Heq by auto
      have Hf1: "\<not> a free in type \<tau>2" using H\<phi> TyRef by auto
      have Hg2: "\<tau>ins = \<tau>2" using TyRef Hih1 Hf1 prems by auto
      show ?thesis using Hg2 TyRef by auto
    qed (auto simp add: split_if)
  qed
next
  case (TyForall \<tau>ins X) show ?case proof auto
    fix i \<phi>2
    assume Heq: "\<forall>.(vsubst_ty (TyFreeVar a) (Suc i) \<tau>ins)^X = vsubst_pty (TyFreeVar a) i \<phi>2"
      and Hf: "\<not> a free in type \<tau>ins" and H\<phi>: "\<not> a free in pretype \<phi>2"
    thus "\<forall>.\<tau>ins^X = \<phi>2" proof (cases \<phi>2)
      case (TyForall \<tau>2 X2) 
      have Hih1: "vsubst_ty (TyFreeVar a) (Suc i) \<tau>ins = vsubst_ty (TyFreeVar a) (Suc i) \<tau>2" 
        using TyForall Heq by auto
      have Hg1: "X = X2" using TyForall Heq by auto
      have Hf1: "\<not> a free in type \<tau>2" using H\<phi> TyForall by auto
      have Hg2: "\<tau>ins = \<tau>2" using TyForall Hih1 Hf1 prems by auto
      show ?thesis using Hg1 Hg2 TyForall by auto
    qed (auto simp add: split_if)
  qed
qed

lemma free_dollars_vsubst_tmty:
  shows "free_let_dollar_scope e = free_let_dollar_scope (vsubst_tmty \<phi>s i e)"
by (induct e arbitrary: i, auto)

lemma free_dollars_vsubst_tm:
  shows "free_let_dollar_scope e = free_let_dollar_scope (vsubst_tm (TmFreeVar x) i e)"
by (induct e arbitrary: i, auto)

lemma subst_subst_equal_env:
  fixes \<Gamma>1 :: env
  fixes \<Gamma>2 :: env
  assumes H: "vsubst_ty_env (TyFreeVar a) i \<Gamma>1 = vsubst_ty_env (TyFreeVar a) i \<Gamma>2"
  and Hx1: "\<not> a free in environ \<Gamma>1"
  and Hx2: "\<not> a free in environ \<Gamma>2"
  shows "\<Gamma>1 = \<Gamma>2"
proof -
  {
    fix \<Gamma>1 \<Gamma>2
    assume H: "vsubst_ty_env (TyFreeVar a) i \<Gamma>1 = vsubst_ty_env (TyFreeVar a) i \<Gamma>2"
      and Hx1: "\<not> a free in environ \<Gamma>1"
      and Hx2: "\<not> a free in environ \<Gamma>2"
    { fix y b
      assume H': "(y, b) \<in> \<Gamma>1"
      then obtain \<tau>' where Hb: "b = Term \<tau>'" by (cases b, auto)
      hence H': "y\<rhd>\<tau>' \<in> \<Gamma>1" using H' by auto
      hence "(y\<rhd>vsubst_ty (TyFreeVar a) i \<tau>') \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>1"
        using vsubst_ty_env_def by auto
      hence "(y\<rhd>vsubst_ty (TyFreeVar a) i \<tau>') \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>2"
        using H by auto
      then obtain \<tau>2 where Heq: "vsubst_ty (TyFreeVar a) i \<tau>' = vsubst_ty (TyFreeVar a) i \<tau>2" 
        and "y\<rhd>\<tau>2 \<in> \<Gamma>2"
        using in_env_exists_sanity[of y "vsubst_ty (TyFreeVar a) i \<tau>'" a i \<Gamma>2] by auto
      moreover
      hence "\<not> a free in type \<tau>2"
        and  "\<not> a free in type \<tau>'"  using Hx2 Hx1 H' free_ty_env_def by auto
      hence "\<tau>' = \<tau>2" using subst_subst_equal_tyty[of a i \<tau>' \<tau>2] Heq by auto
      ultimately
      have "\<exists>b2. b = b2 \<and> (y,b2) \<in> \<Gamma>2" using exI[of _ "\<tau>2"] Hb
        by auto
    }
  }
  hence H: "\<forall>\<Gamma>1 \<Gamma>2. vsubst_ty_env (TyFreeVar a) i \<Gamma>1 = vsubst_ty_env (TyFreeVar a) i \<Gamma>2 \<longrightarrow>
                \<not> a free in environ \<Gamma>1 \<longrightarrow> \<not> a free in environ \<Gamma>2 \<longrightarrow> 
                (\<forall>y b. (y, b) \<in> \<Gamma>1 \<longrightarrow> (\<exists>b2. (b = b2 \<and> (y ,b2) \<in> \<Gamma>2)))" by blast
  have H1: "\<forall>y b. (y, b) \<in> \<Gamma>1 \<longrightarrow> (\<exists>b2. b = b2 \<and> (y, b2) \<in> \<Gamma>2)" using H assms by blast
  have H2: "\<forall>y b. (y, b) \<in> \<Gamma>2 \<longrightarrow> (\<exists>b2. b = b2 \<and> (y, b2) \<in> \<Gamma>1)" using H assms by blast
  show ?thesis using H1 H2 by auto
qed

consts parametrize_ty :: "name \<Rightarrow> nat \<Rightarrow> ty \<Rightarrow> ty"
consts parametrize_pty :: "name \<Rightarrow> nat \<Rightarrow> pty \<Rightarrow> pty"
primrec
  "parametrize_pty a i TyUnit = TyUnit"
  "parametrize_pty a i (TyBase n) = TyBase n"
  "parametrize_pty a i (TyFreeVar b) = (if a=b then TyVar i else TyFreeVar b)"
  "parametrize_pty a i (TyVar n) = TyVar n"
  "parametrize_pty a i (\<tau>1 \<rightarrow>^X \<tau>2) = (parametrize_ty a i \<tau>1) \<rightarrow>^X (parametrize_ty a i \<tau>2)"
  "parametrize_pty a i (\<forall>.\<tau>^X) = \<forall>.(parametrize_ty a (Suc i) \<tau>)^X"
  "parametrize_pty a i (TyRef \<tau>) = TyRef (parametrize_ty a i \<tau>)"
  "parametrize_ty a i (Qual q \<pi> \<phi>) = Qual q \<pi> (parametrize_pty a i \<phi>)"

constdefs parametrize_ty_env :: "name \<Rightarrow> nat \<Rightarrow> env \<Rightarrow> env"
 "parametrize_ty_env a i \<Gamma> \<equiv> { (x\<rhd>parametrize_ty a i \<tau>) | x \<tau>. (x\<rhd>\<tau>) \<in> \<Gamma>}"

lemma parametrize_not_free:
  shows "\<not> a free in type (parametrize_ty a i \<tau>)"
by (induct \<tau> arbitrary: i taking: "%\<phi>. \<forall>i.  \<not> a free in pretype (parametrize_pty a i \<phi>)", auto)

lemma parametrize_not_free_env:
  shows "\<not> a free in environ (parametrize_ty_env a i \<Gamma>)"
by (auto simp add: parametrize_ty_env_def parametrize_not_free free_ty_env_def)

lemma parametrize_subst_equal:
  assumes "\<not> i \<in> set (free_db_ty \<tau>)"
  shows "\<tau> = vsubst_ty (TyFreeVar a) i (parametrize_ty a i \<tau>)"
proof -
  { 
    fix n :: nat
    assume "size \<tau> = n" and "i \<notin> set (free_db_ty \<tau>)"
    hence "\<tau> = vsubst_ty (TyFreeVar a) i (parametrize_ty a i \<tau>)"
    proof (induct n arbitrary: i \<tau> rule: nat_less_induct)
      fix n i \<tau> 
      assume IH: "\<forall>m<n. \<forall>j \<tau>. size \<tau> = m \<longrightarrow> j \<notin> set (free_db_ty \<tau>) \<longrightarrow>
        \<tau> = vsubst_ty (TyFreeVar a) j (parametrize_ty a j \<tau>)"
        and Hsize: "size \<tau> = n"
        and Hdb: "i \<notin> set (free_db_ty \<tau>)"
      hence IH: "\<And>j \<tau>o. \<lbrakk> size \<tau>o < size \<tau>; j \<notin> set (free_db_ty \<tau>o) \<rbrakk> \<Longrightarrow> 
         \<tau>o = vsubst_ty (TyFreeVar a) j (parametrize_ty a j \<tau>o)"
        by auto
      show "\<tau> = vsubst_ty (TyFreeVar a) i (parametrize_ty a i \<tau>)" proof (cases \<tau>)
        fix q \<pi> \<phi> 
        assume H\<tau>_eq: "\<tau> = Qual q \<pi> \<phi>"
        have "\<phi> = vsubst_pty (TyFreeVar a) i (parametrize_pty a i \<phi>)" proof (cases \<phi>)
          case (TyVar n) 
          thus ?thesis using Hdb H\<tau>_eq by auto
        next
          case (TyArrow \<tau>1 X \<tau>2) 
          have "\<tau>1 = vsubst_ty (TyFreeVar a) i (parametrize_ty a i \<tau>1)"
            using IH[of \<tau>1 i] H\<tau>_eq TyArrow Hdb by auto
          moreover
          have "\<tau>2 = vsubst_ty (TyFreeVar a) i (parametrize_ty a i \<tau>2)"
            using IH[of \<tau>2 i] H\<tau>_eq TyArrow Hdb by auto
          ultimately
          show ?thesis using TyArrow by auto
        next
          case (TyRef \<tau>n) 
          have "\<tau>n = vsubst_ty (TyFreeVar a) i (parametrize_ty a i \<tau>n)"
            using IH[of \<tau>n i] H\<tau>_eq TyRef Hdb by auto
          thus ?thesis using TyRef by auto
        next
          case (TyForall \<tau>n X) 
          have "\<tau>n = vsubst_ty (TyFreeVar a) (Suc i) (parametrize_ty a (Suc i) \<tau>n)"
            using IH[of \<tau>n "Suc i"] H\<tau>_eq TyForall Hdb by (auto simp add: shift_many_sanity)
          thus ?thesis using TyForall by auto
        qed auto
        thus ?thesis using H\<tau>_eq by auto
      qed
    qed
  }
  thus ?thesis using prems by auto
qed

lemma parametrize_ty:
  fixes \<tau> :: ty
  assumes "\<Delta> \<Turnstile>' \<tau>"
  shows "\<exists>\<tau>p. \<not> a free in type \<tau>p \<and> \<tau> = vsubst_ty (TyFreeVar a) i \<tau>p"
proof -
  have Hg1: "\<not> a free in type (parametrize_ty a i \<tau>)"
    using parametrize_not_free by auto
  have Hg2: "\<tau> = vsubst_ty (TyFreeVar a) i (parametrize_ty a i \<tau>)" 
    using parametrize_subst_equal poly_wf_no_free_db_ty assms by auto
  show ?thesis using Hg1 Hg2 by auto
qed

lemma parametrize_ty_env:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  shows "\<exists>\<Gamma>p. \<not> a free in environ \<Gamma>p \<and> \<Gamma> = vsubst_ty_env (TyFreeVar a) i \<Gamma>p"
proof -
  show ?thesis apply (rule exI[of _ "parametrize_ty_env a i \<Gamma>"]) proof
    show "a \<notin> free_ty_env (parametrize_ty_env a i \<Gamma>)" 
      using parametrize_not_free_env by auto
  next
    show "\<Gamma> = vsubst_ty_env (TyFreeVar a) i (parametrize_ty_env a i \<Gamma>)" proof auto
      fix x b
      assume Hin: "(x,b) \<in> \<Gamma>"
      then obtain \<tau> where Heq: "b = Term \<tau>" using prems by (cases b, auto)
      hence "x\<rhd>\<tau> \<in> \<Gamma>" using Hin by auto
      hence "(x\<rhd>parametrize_ty a i \<tau>) \<in> parametrize_ty_env a i \<Gamma>"
        by (auto simp add: parametrize_ty_env_def)
      hence "x\<rhd>(vsubst_ty (TyFreeVar a) i (parametrize_ty a i \<tau>)) \<in>
        vsubst_ty_env (TyFreeVar a) i (parametrize_ty_env a i \<Gamma>)"
        by (auto simp add: vsubst_ty_env_def)
      moreover
      have "\<Delta> \<Turnstile>' \<tau>" using poly_env_wf_ty assms Hin Heq by auto
      hence "free_db_ty \<tau> = []" using poly_wf_no_free_db_ty by auto
      hence "(vsubst_ty (TyFreeVar a) i (parametrize_ty a i \<tau>)) = \<tau>" 
        using parametrize_subst_equal by auto
      ultimately
      show "(x,b) \<in> vsubst_ty_env (TyFreeVar a) i (parametrize_ty_env a i \<Gamma>)"
        using Heq by auto
    next
      fix x b
      assume Hin: "(x,b) \<in> vsubst_ty_env (TyFreeVar a) i (parametrize_ty_env a i \<Gamma>)"
      then obtain \<tau> where Heq: "b = Term \<tau>" using prems by (cases b, auto)
      hence "x\<rhd>\<tau> \<in> vsubst_ty_env (TyFreeVar a) i (parametrize_ty_env a i \<Gamma>)" using Hin by auto
      then obtain \<tau>n where Hn: "x\<rhd>\<tau>n \<in> \<Gamma>" 
        and H\<tau>: "\<tau> = vsubst_ty (TyFreeVar a) i (parametrize_ty a i \<tau>n)"
        by (auto simp add: vsubst_ty_env_def parametrize_ty_env_def)
      hence "\<Delta> \<Turnstile>' \<tau>n" using assms poly_env_wf_ty by auto
      hence "free_db_ty \<tau>n = []" using poly_wf_no_free_db_ty by auto
      hence "\<tau>n = \<tau>" using parametrize_subst_equal H\<tau> by auto
      hence "x\<rhd>\<tau> \<in> \<Gamma>" using Hn by auto
      thus "(x,b) \<in> \<Gamma>" using Heq by auto
    qed
  qed
qed

lemma poly_tenv_cons_pty:
  fixes \<phi> :: pty
  assumes "\<Delta> \<Turnstile>' \<phi>"
  shows "\<Delta>,(z:\<star>) \<Turnstile>' \<phi>"
proof -
  have H: "\<And>\<Delta>'. \<Delta>' \<Turnstile>' TY \<Longrightarrow> \<Delta>',(z:\<star>) \<Turnstile>' TY" using wf_env_ty_def by auto
  show ?thesis using prems by (induct taking: "%\<Delta>. %\<tau>::ty. \<Delta> \<Turnstile>' \<tau> \<longrightarrow> \<Delta>,(z:\<star>) \<Turnstile>' \<tau>",
    auto simp add: H env_permute)
qed

lemma poly_wf_free_defined:
  fixes \<phi> :: pty
  assumes "\<Delta> \<Turnstile>' \<phi>"
  and "a free in pretype \<phi>"
  shows "\<Delta> defines a"
using prems proof (
induct taking: "%\<Delta> \<tau>. \<Delta> \<Turnstile>' \<tau> \<longrightarrow> a free in type \<tau> \<longrightarrow> \<Delta> defines a")
  case (wf_forall \<Delta> L \<tau> X) 
  then obtain b where Hdiff: "b \<noteq> a" and Hnin: "b \<notin> L" using fresh_exists[of "L\<union>{a}"] by auto
  have "a free in type \<tau>" using wf_forall by auto
  hence "a free in type (freshen_ty b \<tau>)" using Hdiff free_subst_ty[of a \<tau> "TyFreeVar b"] by auto
  thus ?case using wf_forall Hdiff Hnin by (auto simp add: dom_def)
qed (auto simp add: dom_def)

(*
lemma poly_subst_var_type_ok_env:
  fixes \<phi>s :: pty
  assumes "vsubst_ty_env (TyFreeVar a) i \<Gamma>;\<Delta>,(a:\<star>) \<Turnstile> OK"
  and "\<Delta> \<Turnstile>' \<phi>s"
  and "\<not> a free in environ \<Gamma>"
  shows "vsubst_ty_env \<phi>s i \<Gamma>;\<Delta> \<Turnstile> OK"
proof -
  { fix x \<tau>1 \<tau>2
    assume H1: "x\<rhd>\<tau>1 \<in> vsubst_ty_env \<phi>s i \<Gamma>"
      and H2: "x\<rhd>\<tau>2 \<in> vsubst_ty_env \<phi>s i \<Gamma>"
    then obtain \<tau>n1 \<tau>n2 
      where Hn1: "\<tau>1 = vsubst_ty \<phi>s i \<tau>n1"
      and Hn1_in: "x\<rhd>\<tau>n1 \<in> \<Gamma>"
      and Hn2: "\<tau>2 = vsubst_ty \<phi>s i \<tau>n2"
      and Hn2_in: "x\<rhd>\<tau>n2 \<in> \<Gamma>"
      by (auto simp add: vsubst_ty_env_def)
    hence Hn1': "x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>n1 \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>"
      and "x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>n2 \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>"
      by (auto simp add: vsubst_ty_env_def)
    hence "vsubst_ty (TyFreeVar a) i \<tau>n1 = vsubst_ty (TyFreeVar a) i \<tau>n2"
      using assms by (auto simp add: wf_env_def)
    moreover
    have Han1: "\<not> a free in type \<tau>n1" and  "\<not> a free in type \<tau>n2" using assms Hn1_in Hn2_in
      by (auto simp add: free_ty_env_def)
    ultimately
    have "\<tau>n1 = \<tau>n2" using subst_subst_equal_tyty by auto
    hence "\<tau>1 = \<tau>2" using Hn1 Hn2 by auto
    moreover
    have Htemp: "\<Delta>,(a:\<star>) \<Turnstile>' \<phi>s" using assms poly_tenv_cons_pty by auto 
    have "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty (TyFreeVar a) i \<tau>n1" using Hn1' assms poly_env_wf_ty by auto
    hence Hwf: "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty \<phi>s i \<tau>n1" 
      using poly_subst_var_type_ok[of "\<Delta>,(a:\<star>)"] assms Htemp by auto
    hence "\<Delta> \<Turnstile>' vsubst_ty \<phi>s i \<tau>n1" proof (cases "a free in pretype \<phi>s")
      case False 
      hence "\<not> a free in type (vsubst_ty \<phi>s i \<tau>n1)" using Han1 not_free_subst_ty by auto
      thus ?thesis using False Hwf poly_wf_ty_strength by auto
    next
      case True
      hence "\<Delta> defines a" using poly_wf_free_defined assms by auto
      hence "\<Delta>,(a:\<star>) = \<Delta>" proof (auto simp add: dom_def)
        fix b
        assume "(a, b) \<in> \<Delta>"
        thus "a\<rhd>\<star> \<in> \<Delta>" by (cases b, auto)
      qed
      thus ?thesis using Hwf by auto
    qed
    ultimately
    have "\<tau>1 = \<tau>2 \<and> \<Delta> \<Turnstile>' \<tau>1" using Hn1 by auto
  }
  moreover
  have "finite (vsubst_ty_env (TyFreeVar a) i \<Gamma>)" using assms by (auto simp add: wf_env_def)
  hence "finite \<Gamma>" using vsubst_env_finite by auto
  hence "finite (vsubst_ty_env \<phi>s i \<Gamma>)" using vsubst_env_finite by auto
  moreover
  have "\<Delta>,(a:\<star>) \<Turnstile>' TY" using assms poly_env_wf_env_ty by auto
  hence "\<Delta> \<Turnstile>' TY" by (auto simp add: wf_env_ty_def)
  ultimately
  show ?thesis by (auto simp only: wf_env_def)
qed
*)

lemma env_subst_eq_union_eq:
  assumes "vsubst_ty_env (TyFreeVar a) i \<Gamma> = 
    vsubst_ty_env (TyFreeVar a) i \<Gamma>1 \<union> vsubst_ty_env (TyFreeVar a) i \<Gamma>2"
  and "vsubst_ty_env (TyFreeVar a) i \<Gamma>;\<Delta> \<Turnstile> OK"
  and "vsubst_ty_env (TyFreeVar a) i \<Gamma>1;\<Delta> \<Turnstile> OK"
  and "vsubst_ty_env (TyFreeVar a) i \<Gamma>2;\<Delta> \<Turnstile> OK"
  and "\<not> a free in environ \<Gamma>"
  and "\<not> a free in environ \<Gamma>1"
  and "\<not> a free in environ \<Gamma>2"
  shows "vsubst_ty_env \<phi>s i \<Gamma> = vsubst_ty_env \<phi>s i \<Gamma>1 \<union> vsubst_ty_env \<phi>s i \<Gamma>2"
proof auto
  fix x b
  assume Hin: "(x, b) \<in> vsubst_ty_env \<phi>s i \<Gamma>"
  and Hnotin: "(x, b) \<notin> vsubst_ty_env \<phi>s i \<Gamma>2"
  then obtain \<tau> where H: "b = Term \<tau>" using assms by (cases b, auto)
  obtain \<tau>n where H\<tau>: "\<tau> = vsubst_ty \<phi>s i \<tau>n" and Hn: "x\<rhd>\<tau>n \<in> \<Gamma>"
    using Hin H by (auto simp add: vsubst_ty_env_def)
  hence "(x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>n) \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>" 
    by (auto simp add: vsubst_ty_env_def)
  hence "(x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>n) \<in> 
     vsubst_ty_env (TyFreeVar a) i \<Gamma>1 \<union> vsubst_ty_env (TyFreeVar a) i \<Gamma>2" using assms by auto
  hence "(x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>n) \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>1 \<or>
    (x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>n) \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>2" by auto
  thus "(x, b) \<in> vsubst_ty_env \<phi>s i \<Gamma>1" proof 
    assume "(x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>n) \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>1"
    then obtain \<tau>j where Hsubst: "vsubst_ty (TyFreeVar a) i \<tau>n = vsubst_ty (TyFreeVar a) i \<tau>j"
      and Hj: "x\<rhd>\<tau>j \<in> \<Gamma>1" by (auto simp add: vsubst_ty_env_def)
    moreover
    have "\<not> a free in environ \<Gamma>1" by assumption
    hence "\<not> a free in type \<tau>j" using Hj by (auto simp add: free_ty_env_def)
    moreover
    have "\<not> a free in environ \<Gamma>" by assumption
    hence "\<not> a free in type \<tau>n" using Hn by (auto simp add: free_ty_env_def)
    ultimately
    have "\<tau>n = \<tau>j" using subst_subst_equal_tyty by auto
    thus "(x, b) \<in> vsubst_ty_env \<phi>s i \<Gamma>1" using Hin H H\<tau> Hj by (auto simp add: vsubst_ty_env_def)
  next
    assume "(x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>n) \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>2"
    then obtain \<tau>j where Hsubst: "vsubst_ty (TyFreeVar a) i \<tau>n = vsubst_ty (TyFreeVar a) i \<tau>j"
      and Hj: "x\<rhd>\<tau>j \<in> \<Gamma>2" by (auto simp add: vsubst_ty_env_def)
    moreover
    have "\<not> a free in environ \<Gamma>2" by assumption
    hence "\<not> a free in type \<tau>j" using Hj by (auto simp add: free_ty_env_def)
    moreover
    have "\<not> a free in environ \<Gamma>" by assumption
    hence "\<not> a free in type \<tau>n" using Hn by (auto simp add: free_ty_env_def)
    ultimately
    have "\<tau>n = \<tau>j" using subst_subst_equal_tyty by auto
    thus "(x, b) \<in> vsubst_ty_env \<phi>s i \<Gamma>1" using Hin H H\<tau> Hj Hnotin by (auto simp add: vsubst_ty_env_def)
  qed
next
  fix x b
  assume Hin: "(x, b) \<in> vsubst_ty_env \<phi>s i \<Gamma>1"
  then obtain \<tau> where H: "b = Term \<tau>" using assms by (cases b, auto)
  obtain \<tau>n where H\<tau>: "\<tau> = vsubst_ty \<phi>s i \<tau>n" and Hn: "x\<rhd>\<tau>n \<in> \<Gamma>1"
    using Hin H by (auto simp add: vsubst_ty_env_def)
  hence "(x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>n) \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>1" 
    by (auto simp add: vsubst_ty_env_def)
  hence "(x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>n) \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>" using assms by auto
  then obtain \<tau>j where Hsubst: "vsubst_ty (TyFreeVar a) i \<tau>n = vsubst_ty (TyFreeVar a) i \<tau>j"
    and Hj: "x\<rhd>\<tau>j \<in> \<Gamma>" by (auto simp add: vsubst_ty_env_def)
  moreover
  have "\<not> a free in environ \<Gamma>" by assumption
  hence "\<not> a free in type \<tau>j" using Hj by (auto simp add: free_ty_env_def)
  moreover
  have "\<not> a free in environ \<Gamma>1" by assumption
  hence "\<not> a free in type \<tau>n" using Hn by (auto simp add: free_ty_env_def)
  ultimately
  have "\<tau>n = \<tau>j" using subst_subst_equal_tyty by auto
  thus "(x, b) \<in> vsubst_ty_env \<phi>s i \<Gamma>" using Hin H H\<tau> Hj by (auto simp add: vsubst_ty_env_def)
next
  fix x b
  assume Hin: "(x, b) \<in> vsubst_ty_env \<phi>s i \<Gamma>2"
  then obtain \<tau> where H: "b = Term \<tau>" using assms by (cases b, auto)
  obtain \<tau>n where H\<tau>: "\<tau> = vsubst_ty \<phi>s i \<tau>n" and Hn: "x\<rhd>\<tau>n \<in> \<Gamma>2"
    using Hin H by (auto simp add: vsubst_ty_env_def)
  hence "(x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>n) \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>2"
    by (auto simp add: vsubst_ty_env_def)
  hence "(x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>n) \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>" using assms by auto
  then obtain \<tau>j where Hsubst: "vsubst_ty (TyFreeVar a) i \<tau>n = vsubst_ty (TyFreeVar a) i \<tau>j"
    and Hj: "x\<rhd>\<tau>j \<in> \<Gamma>" by (auto simp add: vsubst_ty_env_def)
  moreover
  have "\<not> a free in environ \<Gamma>" by assumption
  hence "\<not> a free in type \<tau>j" using Hj by (auto simp add: free_ty_env_def)
  moreover
  have "\<not> a free in environ \<Gamma>2" by assumption
  hence "\<not> a free in type \<tau>n" using Hn by (auto simp add: free_ty_env_def)
  ultimately
  have "\<tau>n = \<tau>j" using subst_subst_equal_tyty by auto
  thus "(x, b) \<in> vsubst_ty_env \<phi>s i \<Gamma>" using Hin H H\<tau> Hj by (auto simp add: vsubst_ty_env_def)
qed

lemma env_subst_compat:
  assumes "vsubst_ty_env (TyFreeVar a) i \<Gamma>1 \<sim> vsubst_ty_env (TyFreeVar a) i \<Gamma>2"
  and Hwf: "vsubst_ty_env (TyFreeVar a) i \<Gamma>1 \<union> vsubst_ty_env (TyFreeVar a) i \<Gamma>2 ; \<Delta> \<Turnstile> OK"
  shows "vsubst_ty_env \<phi>s i \<Gamma>1 \<sim> vsubst_ty_env \<phi>s i \<Gamma>2"
using prems proof (auto simp add: env_compat_def)
  fix x \<tau> 
  assume H: "\<forall>x \<tau>. x\<rhd>\<tau> \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>1 \<and> x\<rhd>\<tau> \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>2 \<longrightarrow>
                 qualT \<tau> \<preceq> qU"
    and H1: "x\<rhd>\<tau> \<in> vsubst_ty_env \<phi>s i \<Gamma>1"
    and H2: "x\<rhd>\<tau> \<in> vsubst_ty_env \<phi>s i \<Gamma>2"
  then obtain \<tau>n1 where H\<tau>1: "\<tau> = vsubst_ty \<phi>s i \<tau>n1" and Hn1: "x\<rhd>\<tau>n1 \<in> \<Gamma>1"
    by (auto simp add: vsubst_ty_env_def)
  hence Hin1: "(x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>n1) \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>1" 
    by (auto simp add: vsubst_ty_env_def)
  then obtain \<tau>n2 where H\<tau>2: "\<tau> = vsubst_ty \<phi>s i \<tau>n2" and Hn2: "x\<rhd>\<tau>n2 \<in> \<Gamma>2"
    using H2 by (auto simp add: vsubst_ty_env_def)
  hence Hin2: "(x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>n2) \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>2" 
    by (auto simp add: vsubst_ty_env_def)
  hence "vsubst_ty (TyFreeVar a) i \<tau>n2 = vsubst_ty (TyFreeVar a) i \<tau>n1" using Hwf Hin1 Hin2 
    by (auto simp add: wf_env_def)
  hence "qualT (vsubst_ty (TyFreeVar a) i \<tau>n1) \<preceq> qU" using H Hin1 Hin2 by auto
  hence "qualT \<tau>n1 \<preceq> qU" by (cases \<tau>n1, auto)
  thus "qualT \<tau> \<preceq> qU" using H\<tau>1 by (cases \<tau>n1, auto)
qed

lemma free_tmty_subst_equal:
  shows "set (free_tmty e) = set (free_tmty (vsubst_tm (TmFreeVar x) i e))"
by (induct e arbitrary: i, auto)

lemma vsubst_convT_equal:
  "convT \<rho> (vsubst_ty (TyFreeVar a) i \<tau>) = vsubst_ty (TyFreeVar a) i (convT \<rho> \<tau>)"
by (cases \<tau>, auto)


(* closed types *)

typedef cpty = "{\<phi> | \<phi>. set (free_db_pty \<phi>) = {}}"
   by (auto simp add: exI[of _ "TyUnit"])
(*typedef cty = "{\<tau> | \<tau>. set (free_db_ty \<tau>) = {}}"
   by (auto simp add: exI[of _ "Qual q \<pi> TyUnit"])*)

inductive
  subst_func :: "(cpty \<Rightarrow> pty) \<Rightarrow> bool"
and
  subst_func_ty :: "(cpty \<Rightarrow> ty) \<Rightarrow> bool"
  where
    subst_func_ty_id[intro!]   : "\<lbrakk> subst_func f \<rbrakk> \<Longrightarrow> subst_func_ty (%t. Qual q \<pi> (f t))"
  | subst_func_id[intro!]      : "subst_func (%t. Rep_cpty t)"
  | subst_func_unit[intro!]    : "subst_func (%t. TyUnit)"
  | subst_func_base[intro!]    : "subst_func (%t. TyBase n)"
  | subst_func_freevar[intro!] : "subst_func (%t. TyFreeVar a)"
  | subst_func_var[intro!]     : "subst_func (%t. TyVar k)"
  | subst_func_arrow[intro!]   : "\<lbrakk> subst_func_ty f; subst_func_ty g \<rbrakk> \<Longrightarrow> 
                                    subst_func (%t. f t \<rightarrow>^X g t)"
  | subst_func_forall[intro!]  : "\<lbrakk> subst_func_ty f \<rbrakk> \<Longrightarrow> subst_func (%t. \<forall>.f t^X)"
  | subst_func_ref[intro!]     : "\<lbrakk> subst_func_ty f \<rbrakk> \<Longrightarrow> subst_func (%t. TyRef (f t))"

declare Abs_cpty_inverse[simp]
declare Rep_cpty_inverse[simp]

consts not_free_in_func_pty :: "name \<Rightarrow> (cpty \<Rightarrow> pty) \<Rightarrow> bool"
defs not_free_in_func_pty[simp]:
  "not_free_in_func_pty a f \<equiv> (\<forall>c\<phi>. \<not> a free in pretype (Rep_cpty c\<phi>) \<longrightarrow> \<not> a free in pretype f c\<phi>)"

consts not_free_in_func :: "name \<Rightarrow> (cpty \<Rightarrow> ty) \<Rightarrow> bool"
defs not_free_in_func[simp]:
  "not_free_in_func a f \<equiv> (\<forall>c\<phi> . \<not> a free in pretype (Rep_cpty c\<phi>) \<longrightarrow> \<not> a free in type f c\<phi>)"

lemma subst_func_const:
  shows "subst_func_ty (%ct. \<tau>)"
by (induct \<tau> taking: "%\<phi>. subst_func (%\<phi>c. \<phi>)", auto)

lemma subst_func_const_pty:
  shows "subst_func (%c\<phi>. \<phi>)"
by (induct \<phi> taking: "%\<tau>. subst_func_ty (%c\<tau>. \<tau>)", auto)


lemma free_var_ty_sanity:
  shows "\<not> (\<forall>\<phi>c. a free in pretype (Rep_cpty \<phi>c))"
proof
  assume "\<forall>\<phi>c. a free in pretype (Rep_cpty \<phi>c)"
  hence "a free in pretype (Rep_cpty (Abs_cpty TyUnit))" by blast
  moreover
  have "\<not> a free in pretype TyUnit" by auto
  moreover
  have "set (free_db_pty TyUnit) = {}" by auto
  hence "TyUnit \<in> cpty" by (auto simp add: cpty_def)
  hence "Rep_cpty (Abs_cpty TyUnit) = TyUnit" by auto
  ultimately
  show False by auto
qed

lemma subst_func_equal_ty:
  assumes "subst_func_ty f"
  and "subst_func_ty g"
  and "not_free_in_func a f"
  and "not_free_in_func a g"
  and "f (Abs_cpty (TyFreeVar a)) = g (Abs_cpty (TyFreeVar a))"
  shows "f \<phi> = g \<phi>"
proof -
  have "set (free_db_pty (TyFreeVar a)) = {}" by auto
  hence "TyFreeVar a \<in> cpty" by (auto simp add: cpty_def)
  hence "Rep_cpty (Abs_cpty (TyFreeVar a)) = TyFreeVar a" by auto
  from prems this show ?thesis proof (induct f arbitrary: g \<phi>
    taking: "%f.\<forall>g \<phi>. subst_func f \<longrightarrow> subst_func g \<longrightarrow> not_free_in_func_pty a f \<longrightarrow>
    not_free_in_func_pty a g \<longrightarrow> f (Abs_cpty (TyFreeVar a)) = g (Abs_cpty (TyFreeVar a)) \<longrightarrow>
    f \<phi> = g \<phi>")
    case subst_func_unit thus ?case proof rule+
      fix g \<phi> 
      assume "subst_func g" and "TyUnit = g (Abs_cpty (TyFreeVar a))"
      thus "TyUnit = g \<phi>" by (cases g, auto simp add: cpty_def)
    qed
  next
    case (subst_func_base n) thus ?case proof rule+
      fix g \<phi> 
      assume "subst_func g" and "TyBase n = g (Abs_cpty (TyFreeVar a))"
      thus "TyBase n = g \<phi>" by (cases g, auto simp add: cpty_def)
    qed
  next
    case (subst_func_freevar b) thus ?case proof rule+
      fix g \<phi> 
      assume "subst_func g" and "TyFreeVar b = g (Abs_cpty (TyFreeVar a))"
      and "not_free_in_func_pty a (\<lambda>t. TyFreeVar b)"
      thus "TyFreeVar b = g \<phi>" by (cases g, auto simp add: cpty_def free_var_ty_sanity)
    qed
  next
    case (subst_func_var k) thus ?case proof rule+ 
      fix g \<phi>
      assume "subst_func g" and "TyVar k = g (Abs_cpty (TyFreeVar a))"
      thus "TyVar k = g \<phi>" by (cases g, auto simp add: cpty_def)
    qed
  next
    case (subst_func_ty_id f q \<pi> g \<phi>) 
    have "not_free_in_func a g" by assumption
    from subst_func_ty_id.prems this show ?case proof (cases g, auto simp del: not_free_in_func)
      fix g
      assume "not_free_in_func a (\<lambda>a. Qual q \<pi> (g a))"
        and Hsubst:"subst_func g"
        and Heq: "f (Abs_cpty (TyFreeVar a)) = g (Abs_cpty (TyFreeVar a))"
      have IH: "\<forall>g \<phi>. subst_func f \<longrightarrow>
        subst_func g \<longrightarrow>
        not_free_in_func_pty a f \<longrightarrow>
        not_free_in_func_pty a g \<longrightarrow>
        f (Abs_cpty (TyFreeVar a)) = g (Abs_cpty (TyFreeVar a)) \<longrightarrow> f \<phi> = g \<phi>" 
        and "subst_func f"
        by assumption
      moreover
      have "not_free_in_func a (\<lambda>a. Qual q \<pi> (f a))" by assumption
      hence "not_free_in_func_pty a f" by auto 
      moreover
      have "not_free_in_func a (\<lambda>a. Qual q \<pi> (g a))" by assumption
      hence "not_free_in_func_pty a g" by auto 
      ultimately
      show "f \<phi> = g \<phi>" using Hsubst Heq by blast
    qed
  next
    case subst_func_id thus ?case proof rule+
      fix g \<phi>
      assume Hsubst: "subst_func g" 
        and Hfree: "not_free_in_func_pty a g" 
        and Heq: "Rep_cpty (Abs_cpty (TyFreeVar a)) = g (Abs_cpty (TyFreeVar a))"
      hence Heq: "TyFreeVar a = g (Abs_cpty (TyFreeVar a))" by (auto simp add: cpty_def)
      show "Rep_cpty \<phi> = g \<phi>" using Hsubst Heq Hfree proof (cases g)
        case (subst_func_freevar a')
        have "a \<noteq> a'" using prems not_free_in_func_pty[of a g] by (auto simp add: free_var_ty_sanity)
        hence False using Heq subst_func_freevar by auto
        thus ?thesis by auto
      qed auto
    qed
  next
    case (subst_func_ref fr) show ?case proof rule+
      fix g \<phi> 
      assume "subst_func g"
        and "subst_func (\<lambda>t. TyRef (fr t))"
        and Hfree_fr: "not_free_in_func_pty a (\<lambda>t. TyRef (fr t))"
        and "not_free_in_func_pty a g"
        and "TyRef (fr (Abs_cpty (TyFreeVar a))) = g (Abs_cpty (TyFreeVar a))"
      thus "TyRef (fr \<phi>) = g \<phi>" proof (cases g, auto simp add: cpty_def del: not_free_in_func)
        fix g
        assume Hfree: "\<forall>c\<phi>. a \<notin> set (free_pty (Rep_cpty c\<phi>)) \<longrightarrow> a \<notin> set (free_ty (g c\<phi>))"
          and "subst_func_ty g"
          and "fr (Abs_cpty (TyFreeVar a)) = g (Abs_cpty (TyFreeVar a))"
        moreover
        have "not_free_in_func a fr" using Hfree_fr by auto
        moreover
        have "Rep_cpty (Abs_cpty (TyFreeVar a)) = TyFreeVar a" by (auto simp add: cpty_def)
        moreover
        have "not_free_in_func a g" using Hfree by auto
        ultimately
        show "fr \<phi> = g \<phi>" using subst_func_ref.hyps by blast
      qed
    qed 
  next
    case (subst_func_forall fr X) show ?case proof rule+
      fix g \<phi> 
      assume "subst_func g"
        and "subst_func (\<lambda>t. \<forall>.fr t^X)"
        and Hfree_fr: "not_free_in_func_pty a (\<lambda>t. \<forall>.fr t^X)"
        and "not_free_in_func_pty a g"
        and "\<forall>. (fr (Abs_cpty (TyFreeVar a)))^X = g (Abs_cpty (TyFreeVar a))"
      thus "\<forall>. fr \<phi>^X = g \<phi>" proof (cases g, auto simp add: cpty_def del: not_free_in_func)
        fix g
        assume Hfree: "\<forall>c\<phi>. a \<notin> set (free_pty (Rep_cpty c\<phi>)) \<longrightarrow> a \<notin> set (free_ty (g c\<phi>))"
          and "subst_func_ty g"
          and "fr (Abs_cpty (TyFreeVar a)) = g (Abs_cpty (TyFreeVar a))"
        moreover
        have "not_free_in_func a fr" using Hfree_fr by auto
        moreover
        have "Rep_cpty (Abs_cpty (TyFreeVar a)) = TyFreeVar a" by (auto simp add: cpty_def)
        moreover
        have "not_free_in_func a g" using Hfree by auto
        ultimately
        show "fr \<phi> = g \<phi>" using subst_func_forall.hyps by blast
      qed
    qed 
  next
    case (subst_func_arrow f1 f2 X) show ?case proof rule+
      fix g \<phi> 
      assume "subst_func g"
        and "subst_func (\<lambda>t. f1 t \<rightarrow>^X f2 t)"
        and Hfree_fr: "not_free_in_func_pty a (\<lambda>t. f1 t \<rightarrow>^X f2 t)"
        and "not_free_in_func_pty a g"
        and "f1 (Abs_cpty (TyFreeVar a)) \<rightarrow>^X f2 (Abs_cpty (TyFreeVar a)) = g (Abs_cpty (TyFreeVar a))"
      thus "f1 \<phi> \<rightarrow>^X f2 \<phi> = g \<phi>" proof (cases g, auto simp add: cpty_def del: not_free_in_func)
        fix g foo
        assume Hfree: "\<forall>c\<phi>. a \<notin> set (free_pty (Rep_cpty c\<phi>)) \<longrightarrow>
          a \<notin> set (free_ty (g c\<phi>)) \<and> a \<notin> set (free_ty (foo c\<phi>))"
          and "subst_func_ty g"
          and "f1 (Abs_cpty (TyFreeVar a)) = g (Abs_cpty (TyFreeVar a))"
        moreover
        have "not_free_in_func a f1" using Hfree_fr by auto
        moreover
        have "Rep_cpty (Abs_cpty (TyFreeVar a)) = TyFreeVar a" by (auto simp add: cpty_def)
        moreover
        have "not_free_in_func a g" using Hfree by auto
        ultimately
        show "f1 \<phi> = g \<phi>" using subst_func_arrow.hyps by blast
      next
        fix foo g
        assume Hfree: "\<forall>c\<phi>. a \<notin> set (free_pty (Rep_cpty c\<phi>)) \<longrightarrow>
          a \<notin> set (free_ty (foo c\<phi>)) \<and> a \<notin> set (free_ty (g c\<phi>))"
          and "subst_func_ty g"
          and "f2 (Abs_cpty (TyFreeVar a)) = g (Abs_cpty (TyFreeVar a))"
        moreover
        have "not_free_in_func a f2" using Hfree_fr by auto
        moreover
        have "Rep_cpty (Abs_cpty (TyFreeVar a)) = TyFreeVar a" by (auto simp add: cpty_def)
        moreover
        have "not_free_in_func a g" using Hfree by auto
        ultimately
        show "f2 \<phi> = g \<phi>" using subst_func_arrow.hyps by blast
      qed
    qed
  qed
qed
   
lemma subst_func_subst:
  assumes "subst_func_ty g"
  and "subst_func f"
  shows "subst_func_ty (%t. vsubst_ty (f t) i (g t))"
using prems proof (induct g arbitrary: i f
taking: "%g. \<forall>i f. subst_func g \<longrightarrow> subst_func f \<longrightarrow> subst_func (%t. vsubst_pty (f t) i (g t))", auto)
  fix i f
  { fix t
    have "set (free_db_pty (Rep_cpty t)) = {}" using cpty_def Rep_cpty by auto
    hence "i \<notin> set (free_db_pty (Rep_cpty t))" by auto
    hence "vsubst_pty (f t) i (Rep_cpty t) = Rep_cpty t" by (rule poly_not_in_free_db_same_subst_pty)
  }
  thus "subst_func (%t. vsubst_pty (f t) i (Rep_cpty t))" by auto
next
  fix k i f
  assume "subst_func f"
  show "subst_func (%t. if k = i then f t else TyVar k)" by (cases "k = i", auto)
qed

lemma subst_func_subst_pty:
  assumes "subst_func g"
  and "subst_func f"
  shows "subst_func (%t. vsubst_pty (f t) i (g t))"
using prems proof (induct g arbitrary: i f
  taking: "%g. \<forall>i f. subst_func_ty g \<longrightarrow> subst_func f \<longrightarrow> subst_func_ty (%t. vsubst_ty (f t) i (g t))", 
  auto)
  fix q \<pi> i f
  { fix t
    have "set (free_db_pty (Rep_cpty t)) = {}" using cpty_def Rep_cpty by auto
    hence "i \<notin> set (free_db_pty (Rep_cpty t))" by auto
    hence "vsubst_pty (f t) i (Rep_cpty t) = Rep_cpty t" by (rule poly_not_in_free_db_same_subst_pty)
  }
  thus "subst_func (%t. vsubst_pty (f t) i (Rep_cpty t))" by auto
next
  fix k i f
  assume "subst_func f"
  show "subst_func (%t. if k = i then f t else TyVar k)" by (cases "k = i", auto)
qed

lemma vsubst_equal_sanity:
  assumes "vsubst_ty (vsubst_pty (TyFreeVar a) i \<phi>) 0 (vsubst_ty (TyFreeVar a) (Suc i) \<tau>d) =
           vsubst_ty (TyFreeVar a) i \<tau>"
  and "\<not> a free in type \<tau>"
  and "\<not> a free in pretype \<phi>"
  and "\<not> a free in type \<tau>d"
  and "set (free_db_ty \<tau>) \<subseteq> {i}"
  and "set (free_db_pty \<phi>) \<subseteq> {i}"
  and "set (free_db_ty \<tau>d) \<subseteq> {0, Suc i}"
  and "set (free_db_pty \<phi>s) = {}"
  shows "vsubst_ty (vsubst_pty \<phi>s i \<phi>) 0 (vsubst_ty \<phi>s (Suc i) \<tau>d) = vsubst_ty \<phi>s i \<tau>"
proof -
  let ?f = "%c\<phi>. vsubst_ty (vsubst_pty (Rep_cpty c\<phi>) i \<phi>) 0 (vsubst_ty (Rep_cpty c\<phi>) (Suc i) \<tau>d)"
  let ?g = "%c\<phi>. vsubst_ty (Rep_cpty c\<phi>) i \<tau>"
  let ?\<phi>s' = "Rep_cpty (Abs_cpty \<phi>s)"
  have "subst_func_ty (%c\<phi>. vsubst_ty (Rep_cpty c\<phi>) (Suc i) \<tau>d)"
    using subst_func_subst[of "%ct. \<tau>d"] by (auto simp add: subst_func_const)
  moreover
  have "subst_func (%c\<phi>. vsubst_pty (Rep_cpty c\<phi>) i \<phi>)" apply (rule subst_func_subst_pty[of "%ct. \<phi>"]) 
    by (auto simp add: subst_func_const_pty)
  ultimately
  have "subst_func_ty ?f" using subst_func_subst by auto
  moreover
  have "subst_func_ty ?g" using subst_func_subst[of "%ct. \<tau>"] by (auto simp add: subst_func_const)
  moreover
  have "not_free_in_func a ?f" using prems by (auto simp add: not_free_subst_ty not_free_subst_pty)
  moreover
  have "not_free_in_func a ?g" using prems by (auto simp add: not_free_subst_ty) 
  moreover
  have "set (free_db_pty (TyFreeVar a)) = {}" by auto
  hence "TyFreeVar a \<in> cpty" by (auto simp add: cpty_def)
  hence "Rep_cpty (Abs_cpty (TyFreeVar a)) = TyFreeVar a" by auto
  hence "?f (Abs_cpty (TyFreeVar a)) = ?g (Abs_cpty (TyFreeVar a))" by auto
  ultimately
  have "vsubst_ty (vsubst_pty ?\<phi>s' i \<phi>) 0 (vsubst_ty ?\<phi>s' (Suc i) \<tau>d) = vsubst_ty ?\<phi>s' i \<tau>"
    by (rule subst_func_equal_ty)
  moreover
  have "\<phi>s \<in> cpty" using prems by (auto simp add: cpty_def)
  hence "?\<phi>s' = \<phi>s" by auto
  ultimately
  show ?thesis by auto
qed

(*
consts convert_ty:: "nat \<Rightarrow> nat \<Rightarrow> ty \<Rightarrow> ty"
consts convert_pty:: "nat \<Rightarrow> nat \<Rightarrow> pty \<Rightarrow> pty"
primrec
  "convert_ty j i (Qual q \<pi> \<phi>) = Qual q \<pi> (convert_pty j i \<phi>)"
  "convert_pty j i TyUnit = TyUnit"
  "convert_pty j i (TyBase n) = (TyBase n)"
  "convert_pty j i (TyFreeVar b) = TyFreeVar b"
  "convert_pty j i (TyVar n) = (if n=i then (TyVar j) else (TyVar n))"
  "convert_pty j i (\<tau>1 \<rightarrow>^X \<tau>2) = (convert_ty j i \<tau>1) \<rightarrow>^X (convert_ty j i \<tau>2)"
  "convert_pty j i (\<forall>.\<tau>^X) = \<forall>.(convert_ty (Suc j) (Suc i) \<tau>)^X"
  "convert_pty j i (TyRef \<tau>) = TyRef (convert_ty j i \<tau>)"

lemma not_free_convert_ty:
  assumes "\<not> a free in type \<tau>"
  shows "\<not> a free in type convert_ty j i \<tau>"
using prems by (induct \<tau> arbitrary: i j
  taking: "%\<phi>. \<forall>i j. \<not> a free in pretype \<phi> \<longrightarrow> \<not> a free in pretype (convert_pty j i \<phi>)", auto)

lemma subst_convert_subst_swap:
  assumes "i \<noteq> k"
  and "j \<noteq> k"
  and "\<forall>k. vsubst_pty (TyFreeVar a) k \<phi>s = \<phi>s"
  shows "vsubst_ty \<phi>s i (convert_ty i j (vsubst_ty (TyFreeVar a) k \<tau>)) =
         vsubst_ty (TyFreeVar a) k (vsubst_ty \<phi>s i (convert_ty i j \<tau>))"
using prems by (induct \<tau> arbitrary: i j k
  taking: "%\<phi>. \<forall>i j k. i \<noteq> k \<longrightarrow> j \<noteq> k \<longrightarrow> (\<forall>k. vsubst_pty (TyFreeVar a) k \<phi>s = \<phi>s) \<longrightarrow>
         vsubst_pty \<phi>s i (convert_pty i j (vsubst_pty (TyFreeVar a) k \<phi>)) = 
         vsubst_pty (TyFreeVar a) k (vsubst_pty \<phi>s i (convert_pty i j \<phi>))", auto)

constdefs convert_ty_env:: "nat \<Rightarrow> nat \<Rightarrow> env \<Rightarrow> env"
  "convert_ty_env j i \<Gamma> \<equiv> { (x\<rhd>convert_ty j i \<tau>) | x \<tau>. (x \<rhd> \<tau>) \<in> \<Gamma> }"
*)

lemma subst_convert_ty:
  fixes \<tau>f :: ty
  fixes \<phi>s :: pty
  assumes "\<Delta> \<Turnstile>' \<tau>f"
  and "\<tau>f = vsubst_ty (TyFreeVar a) i \<tau>"
  and "\<Delta> \<Turnstile>' \<phi>s"
  shows "vsubst_ty \<phi>s i \<tau> = vsubst_ty \<phi>s j (convert_ty j i \<tau>)"
using prems proof (induct arbitrary: \<tau> i j
taking: "%\<Delta> \<phi>f. \<forall>\<phi> i j. \<Delta> \<Turnstile>' \<phi>f \<longrightarrow> \<phi>f = vsubst_pty (TyFreeVar a) i \<phi> \<longrightarrow> \<Delta> \<Turnstile>' \<phi>s \<longrightarrow>
      vsubst_pty \<phi>s i \<phi> = vsubst_pty \<phi>s j (convert_pty j i \<phi>)")
  case (wf_ty \<Delta> \<phi> q \<pi> \<tau> i j) thus ?case by (cases \<tau>, auto)
next
  case (wf_unit \<Delta>) show ?case proof rule+
    fix \<phi> i j
    assume "TyUnit = vsubst_pty (TyFreeVar a) i \<phi>"
    thus "vsubst_pty \<phi>s i \<phi> = vsubst_pty \<phi>s j (convert_pty j i \<phi>)" by (cases \<phi>, auto)
  qed
next
  case (wf_base \<Delta> n) show ?case proof rule+
    fix \<phi> i j
    assume "TyBase n = vsubst_pty (TyFreeVar a) i \<phi>"
    thus "vsubst_pty \<phi>s i \<phi> = vsubst_pty \<phi>s j (convert_pty j i \<phi>)" by (cases \<phi>, auto)
  qed
next
  case (wf_tvar c \<Delta>) show ?case proof rule+
    fix \<phi> i j
    assume "TyFreeVar c = vsubst_pty (TyFreeVar a) i \<phi>"
    thus "vsubst_pty \<phi>s i \<phi> = vsubst_pty \<phi>s j (convert_pty j i \<phi>)" by (cases \<phi>, auto)
  qed
next
  case (wf_arrow \<Delta> \<tau>1 \<tau>2 Z) show ?case proof rule+
    fix \<phi> i j
    assume H: "\<tau>1 \<rightarrow>^Z \<tau>2 = vsubst_pty (TyFreeVar a) i \<phi>"
      and "\<Delta> \<Turnstile>' \<tau>1 \<rightarrow>^Z \<tau>2"
      and Hwf: "\<Delta> \<Turnstile>' \<phi>s"
    thus "vsubst_pty \<phi>s i \<phi> = vsubst_pty \<phi>s j (convert_pty j i \<phi>)"  proof (cases \<phi>)
      case (TyArrow \<tau>1n Zn \<tau>2n)
      hence "vsubst_ty \<phi>s i \<tau>1n = vsubst_ty \<phi>s j (convert_ty j i \<tau>1n)" using H wf_arrow Hwf by force
      moreover
      have "vsubst_ty \<phi>s i \<tau>2n = vsubst_ty \<phi>s j (convert_ty j i \<tau>2n)" using H wf_arrow Hwf TyArrow
        by force
      ultimately
      show "vsubst_pty \<phi>s i \<phi> = vsubst_pty \<phi>s j (convert_pty j i \<phi>)" using H TyArrow by auto
    qed auto
  qed
next
  case (wf_ref \<Delta> \<tau>r) show ?case proof rule+
    fix \<phi> i j
    assume H: "TyRef \<tau>r = vsubst_pty (TyFreeVar a) i \<phi>"
      and "\<Delta> \<Turnstile>' TyRef \<tau>r"
      and Hwf: "\<Delta> \<Turnstile>' \<phi>s"
    thus "vsubst_pty \<phi>s i \<phi> = vsubst_pty \<phi>s j (convert_pty j i \<phi>)"  proof (cases \<phi>)
      case (TyRef \<tau>r)
      hence "vsubst_ty \<phi>s i \<tau>r = vsubst_ty \<phi>s j (convert_ty j i \<tau>r)" using H wf_ref Hwf by force
      thus "vsubst_pty \<phi>s i \<phi> = vsubst_pty \<phi>s j (convert_pty j i \<phi>)" using H TyRef by auto
    qed auto
  qed
next
  case (wf_forall \<Delta> L \<tau>ins X) show ?case proof rule+
    fix  \<phi> i j
    assume Hwf: "\<Delta> \<Turnstile>' \<forall>.\<tau>ins^X"
      and Heq: "\<forall>.\<tau>ins^X = vsubst_pty (TyFreeVar a) i \<phi>"
      and "\<Delta> \<Turnstile>' \<phi>s"
    thus "vsubst_pty \<phi>s i \<phi> = vsubst_pty \<phi>s j (convert_pty j i \<phi>)"  proof (cases \<phi>)
      case (TyForall \<tau>ins2 X2) 
      have IH: "\<forall>d. d \<notin> L \<longrightarrow> \<Delta>,(d:\<star>) \<Turnstile>' freshen_ty d \<tau>ins \<and> 
          (\<forall>\<tau>f i j.
           freshen_ty d \<tau>ins = vsubst_ty (TyFreeVar a) i \<tau>f \<longrightarrow>
           \<Delta>,(d:\<star>) \<Turnstile>' \<phi>s \<longrightarrow>
           (vsubst_ty \<phi>s i \<tau>f = vsubst_ty \<phi>s j (convert_ty j i \<tau>f)))" using wf_forall by auto
      have "finite L" by assumption
      then obtain d where Hd1: "\<not> d free in type (vsubst_ty \<phi>s (Suc i) \<tau>ins2)" 
        and Hd2: "\<not> d free in type (vsubst_ty \<phi>s (Suc j) (convert_ty (Suc j) (Suc i) \<tau>ins2))"
        and Hd3: "d \<notin> L"
        using
          fresh_exists[of "L\<union>(set (free_ty (vsubst_ty \<phi>s (Suc i) \<tau>ins2)))\<union>
          (set (free_ty (vsubst_ty \<phi>s (Suc j) (convert_ty (Suc j) (Suc i) \<tau>ins2))))"]  
        by auto
      have Hwf_prem: "\<Delta>,(d:\<star>) \<Turnstile>' \<phi>s" using prems poly_tenv_cons_pty by auto 
      from TyForall Heq have "\<tau>ins = vsubst_ty (TyFreeVar a) (Suc i) \<tau>ins2" by auto
      hence "freshen_ty d \<tau>ins = freshen_ty d (vsubst_ty (TyFreeVar a) (Suc i) \<tau>ins2)" by auto
      also 
      have "freshen_ty d (vsubst_ty (TyFreeVar a) (Suc i) \<tau>ins2) = 
            vsubst_ty (TyFreeVar a) (Suc i) (freshen_ty d \<tau>ins2)" 
        using subst_subst_ty_swap[of 0 "Suc i"] by auto
      finally
      have
       "vsubst_ty \<phi>s (Suc i) (freshen_ty d \<tau>ins2) = 
             vsubst_ty \<phi>s (Suc j) (convert_ty (Suc j) (Suc i) (freshen_ty d \<tau>ins2))" 
        using IH Hwf_prem Hd3 by auto
      also
      have "free_db_pty \<phi>s = []" using poly_wf_no_free_db_pty prems by auto
      hence H\<phi>s_subst: "\<forall>k. vsubst_pty (TyFreeVar d) k \<phi>s = \<phi>s"
        using poly_no_free_db_vars_same_subst_pty by auto
      hence "vsubst_ty \<phi>s (Suc i) (freshen_ty d \<tau>ins2) = freshen_ty d (vsubst_ty \<phi>s (Suc i) \<tau>ins2)"
        using subst_subst_ty_swap[of 0 "Suc i" d] by auto
(*      also have "convert_ty (Suc j) (Suc i) (freshen_ty d \<tau>ins2) =
                 freshen_ty d (convert_ty (Suc j) (Suc i) \<tau>ins2)" 
        using subst_convert_subst_swap*)
      also have "vsubst_ty \<phi>s (Suc j) (convert_ty (Suc j) (Suc i) (freshen_ty d \<tau>ins2)) = 
                 freshen_ty d (vsubst_ty \<phi>s (Suc j) (convert_ty (Suc j) (Suc i) \<tau>ins2))"
        using subst_convert_subst_swap[of "Suc j" 0 "Suc i" d] using H\<phi>s_subst by auto
      finally 
      have "freshen_ty d (vsubst_ty \<phi>s (Suc i) \<tau>ins2) =
        freshen_ty d (vsubst_ty \<phi>s (Suc j) (convert_ty (Suc j) (Suc i) \<tau>ins2))" by auto
      moreover
      have "\<not> d free in type (vsubst_ty \<phi>s (Suc i) \<tau>ins2)" using Hd1 by auto
      moreover
      have "\<not> d free in type (vsubst_ty \<phi>s (Suc j) (convert_ty (Suc j) (Suc i) \<tau>ins2))" 
        using Hd2 by auto
      ultimately
      have "vsubst_ty \<phi>s (Suc i) \<tau>ins2 = vsubst_ty \<phi>s (Suc j) (convert_ty (Suc j) (Suc i) \<tau>ins2)" 
        using subst_subst_equal_tyty by auto
      hence "vsubst_pty \<phi>s i (\<forall>.\<tau>ins2^X2) = vsubst_pty \<phi>s j (convert_pty j i (\<forall>.\<tau>ins2^X2))" by auto
      thus "vsubst_pty \<phi>s i \<phi> = vsubst_pty \<phi>s j (convert_pty j i \<phi>)" using TyForall by auto
    qed auto
  qed
qed

lemma poly_wf_env_wf_ty:
  fixes \<tau> :: ty
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "x\<rhd>\<tau> \<in> \<Gamma>"
  shows "\<Delta> \<Turnstile>' \<tau>"
using prems by (auto simp add: wf_env_def)

lemma subst_convert_env:
  fixes \<phi>s :: pty
  assumes "\<Gamma>';\<Delta> \<Turnstile> OK"
  and "\<Gamma>' = vsubst_ty_env (TyFreeVar x) i \<Gamma>"
  and "\<Delta> \<Turnstile>' \<phi>s"
  shows "vsubst_ty_env \<phi>s i \<Gamma> = vsubst_ty_env \<phi>s j (convert_ty_env j i \<Gamma>)"
proof (auto simp add: vsubst_ty_env_def convert_ty_env_def)
  fix y \<tau>
  assume "(y\<rhd>\<tau>) \<in> \<Gamma>"
  hence "(y\<rhd>(vsubst_ty (TyFreeVar x) i \<tau>)) \<in> \<Gamma>'" using prems by (auto simp add: vsubst_ty_env_def)
  hence "\<Delta> \<Turnstile>' vsubst_ty (TyFreeVar x) i \<tau>" using prems poly_wf_env_wf_ty by auto
  hence "vsubst_ty \<phi>s i \<tau> = vsubst_ty \<phi>s j (convert_ty j i \<tau>)" and
    "(y\<rhd>\<tau>) \<in> \<Gamma>" using prems by (auto simp add: subst_convert_ty)
  thus "\<exists>\<tau>''. vsubst_ty \<phi>s i \<tau> = vsubst_ty \<phi>s j \<tau>'' \<and> (\<exists>\<tau>. \<tau>'' = convert_ty j i \<tau> \<and> (y\<rhd>\<tau>) \<in> \<Gamma>)"
    by auto
next
  fix y \<tau>
  assume "(y\<rhd>\<tau>) \<in> \<Gamma>"
  hence "(y\<rhd>(vsubst_ty (TyFreeVar x) i \<tau>)) \<in> \<Gamma>'" using prems by (auto simp add: vsubst_ty_env_def)
  hence "\<Delta> \<Turnstile>' vsubst_ty (TyFreeVar x) i \<tau>" using prems poly_wf_env_wf_ty by auto
  hence "vsubst_ty \<phi>s i \<tau> = vsubst_ty \<phi>s j (convert_ty j i \<tau>)"
    using prems by (auto simp add: subst_convert_ty)
  hence "vsubst_ty \<phi>s j (convert_ty j i \<tau>) = vsubst_ty \<phi>s i \<tau>" and "(y\<rhd>\<tau>) \<in> \<Gamma>"
    using prems by auto
  thus "\<exists>\<tau>''. vsubst_ty \<phi>s j (convert_ty j i \<tau>) = vsubst_ty \<phi>s i \<tau>'' \<and> (y\<rhd>\<tau>'') \<in> \<Gamma>" by auto
qed

lemma not_free_convert_env:
  assumes "\<not> x free in environ \<Gamma>"
  shows "\<not> x free in environ (convert_ty_env i j \<Gamma>)"
proof (auto simp add: free_ty_env_def convert_ty_env_def)
  fix y \<tau>
  assume "x free in type convert_ty i j \<tau>"
    and "(y\<rhd>\<tau>) \<in> \<Gamma>"
  hence "\<not> x free in type \<tau>" using prems by (auto simp add: free_ty_env_def)
  hence "\<not> x free in type convert_ty i j \<tau>" by (rule not_free_convert_ty)
  thus False by contradiction
qed

lemma subst_pty_ty_complicated:
  assumes "\<forall>j \<rho>. vsubst_sc_pty (ScFree \<rho>) j \<phi>s = \<phi>s"
  and "\<forall>j \<rho>. vsubst_sc_pty (ScFree \<rho>) j (vsubst_pty (TyFreeVar a) i \<phi>) = vsubst_pty (TyFreeVar a) i \<phi>"
  shows "vsubst_sc_pty (ScFree \<rho>) j (vsubst_pty \<phi>s i \<phi>) = vsubst_pty \<phi>s i \<phi>"
using prems by (induct \<phi> arbitrary: i j taking:
"%\<tau>. \<forall>i j. (\<forall>j \<rho>. vsubst_sc_pty (ScFree \<rho>) j \<phi>s = \<phi>s) \<longrightarrow> 
          (\<forall>j \<rho>. vsubst_sc_ty (ScFree \<rho>) j (vsubst_ty (TyFreeVar a) i \<tau>) = vsubst_ty (TyFreeVar a) i \<tau>)
            \<longrightarrow> 
        (vsubst_sc_ty (ScFree \<rho>) j (vsubst_ty \<phi>s i \<tau>) = vsubst_ty \<phi>s i \<tau>)", auto)



lemma substitution_pty:
  fixes n :: nat
  fixes \<phi>s :: pty
  assumes "size e = n"
  and "(vsubst_ty_env \<phi>s i \<Gamma>);\<Delta> \<Turnstile> OK"
  and "(vsubst_ty_env \<phi>s i M);\<Delta> \<Turnstile> OK"
  and "\<Delta> \<Turnstile>' \<phi>s"
(*  and "Zs \<Turnstile> \<phi>s"
  and "\<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> \<rho> \<notin> Environ.dom Zs"*)
  and "vsubst_ty_env (TyFreeVar a) i \<Gamma>;\<Delta>,(a:\<star>);
       vsubst_ty_env (TyFreeVar a) i M;Z \<turnstile>
          vsubst_tmty (TyFreeVar a) i e : vsubst_ty (TyFreeVar a) i \<tau>"
  and "\<not> a free in term e"
  and "\<not> a free in type \<tau>"
  and H\<Gamma>free: "\<not> a free in environ \<Gamma>"
  and HMfree: "\<not> a free in environ M"
  and "\<not> \<Delta> defines a"
  and "\<forall>i \<rho>. vsubst_sc_pty (ScFree \<rho>) i \<phi>s = \<phi>s"
  shows "vsubst_ty_env \<phi>s i \<Gamma>;\<Delta>;vsubst_ty_env \<phi>s i M;Z \<turnstile> vsubst_tmty \<phi>s i e : vsubst_ty \<phi>s i \<tau>"
using prems proof (induct n arbitrary: e \<tau> i \<tau>' \<Gamma> \<Delta> M Z rule: nat_less_induct)
  case (1 n_y e_y \<tau> i \<Gamma> \<Delta> M Z)
  let ?\<Gamma>x = "vsubst_ty_env (TyFreeVar a) i \<Gamma>"
  let ?\<Gamma>t = "vsubst_ty_env \<phi>s i \<Gamma>"
  let ?Mx = "vsubst_ty_env (TyFreeVar a) i M"
  let ?Mt = "vsubst_ty_env \<phi>s i M"
  have H\<Gamma>free: "\<not> a free in environ \<Gamma>" by assumption
  from 1 have IH:
    "\<And>\<Gamma> M \<Delta> \<tau> Z e i. 
     \<lbrakk> size e < size e_y; vsubst_ty_env \<phi>s i \<Gamma>;\<Delta> \<Turnstile> OK; vsubst_ty_env \<phi>s i M;\<Delta> \<Turnstile> OK; \<Delta> \<Turnstile>' \<phi>s; (*Zs \<Turnstile> \<phi>s;
       \<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> \<rho> \<notin> Environ.dom Zs;*)
       vsubst_ty_env (TyFreeVar a) i \<Gamma>;\<Delta>,(a:\<star>);vsubst_ty_env (TyFreeVar a) i M;Z \<turnstile> 
                    vsubst_tmty (TyFreeVar a) i e : vsubst_ty (TyFreeVar a) i \<tau>;
      a \<notin> set (free_tmty e); a \<notin> set (free_ty \<tau>); a \<notin> free_ty_env \<Gamma>;
      \<forall>i \<rho>. vsubst_sc_pty (ScFree \<rho>) i \<phi>s = \<phi>s;
      a \<notin> free_ty_env M; a \<notin> dom \<Delta> \<rbrakk> \<Longrightarrow>
      vsubst_ty_env \<phi>s i \<Gamma>;\<Delta>;vsubst_ty_env \<phi>s i M;Z \<turnstile> vsubst_tmty \<phi>s i e : vsubst_ty \<phi>s i \<tau>"
    by auto
  show ?case proof (cases e_y)

    (* unit case *)
    case (TmUnit q \<pi>)
    hence Ht: "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> TmUnit q \<pi> : vsubst_ty (TyFreeVar a) i \<tau>" using 1 by auto
    hence Heq: "vsubst_ty (TyFreeVar a) i \<tau> = Qual q \<pi> TyUnit" by auto
    then obtain \<phi> where H\<phi>: "\<tau> = Qual q \<pi> \<phi>" by (cases \<tau>, auto)
    hence "\<phi> = TyUnit" using Heq by (cases \<phi>, auto simp add: trivial_case)
    hence H\<tau>: "\<tau> = Qual q \<pi> TyUnit" using H\<phi> by auto
    have Ht: "?\<Gamma>t;\<Delta>;?Mt;Z \<turnstile> TmUnit q \<pi>  : Qual q \<pi> TyUnit"
    proof 
      have "qualG vsubst_ty_env (TyFreeVar a) i \<Gamma> \<preceq> qU" using Ht by auto
      thus "qualG vsubst_ty_env \<phi>s i \<Gamma> \<preceq> qU" using qualG_vsubst_left by auto
    next
      show "minimalValidScope \<pi> Z" using Ht by auto
    qed 
    show ?thesis using Ht H\<tau> TmUnit by auto
  next;

    (* free variable case *)
    case (TmFreeVar y)
    hence Ht: "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> TmFreeVar y : vsubst_ty (TyFreeVar a) i \<tau>"
      using 1 by auto
    hence Hx': "?\<Gamma>x = vsubst_ty_env (TyFreeVar a) i (\<Gamma>,(y:\<tau>))"
      by (auto simp add: vsubst_ty_env_def)
    moreover
    have "\<not> a free in environ (\<Gamma>,(y:\<tau>))" using "1.prems" by (auto simp add: free_ty_env_def)
    moreover
    hence "\<not> a free in environ \<Gamma>" using prems by auto
    ultimately
    have "\<Gamma> = \<Gamma>,(y:\<tau>)" using subst_subst_equal_env[of a i \<Gamma> "\<Gamma>,(y:\<tau>)"] by auto
    hence Heq: "?\<Gamma>t = vsubst_ty_env \<phi>s i \<Gamma>,(y:vsubst_ty \<phi>s i \<tau>)" by (auto simp add: vsubst_ty_env_def)
    have "qualG ((vsubst_ty_env (TyFreeVar a) i \<Gamma>)\<bar>y) \<preceq> qU" using Ht by auto
    hence Hqual: "qualG ((vsubst_ty_env \<phi>s i \<Gamma>)\<bar>y) \<preceq> qU" using qualG_vsubst_left_special by auto 
    moreover
    have "minimalValidScope (scopeT (vsubst_ty (TyFreeVar a) i \<tau>)) Z" using Ht by auto
    hence "minimalValidScope (scopeT (vsubst_ty \<phi>s i \<tau>)) Z" by (cases \<tau>, auto)
    ultimately
    show ?thesis using prems Heq by auto
  next;

    (* de'Brujin case *)
    case (TmVar j)
    hence "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> vsubst_tmty (TyFreeVar a) i (TmVar j) : vsubst_ty (TyFreeVar a) i \<tau>"
      using prems by auto
    hence "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> TmVar j : vsubst_ty (TyFreeVar a) i \<tau>" by auto
    thus ?thesis by (cases, auto)
  next;
    
    (* Term application case *)
    case (TmApp e1 e2)
    let ?e1s = "vsubst_tmty (TyFreeVar a) i e1"
    let ?e2s = "vsubst_tmty (TyFreeVar a) i e2"
    let ?e1\<phi> = "vsubst_tmty \<phi>s i e1"
    let ?e2\<phi> = "vsubst_tmty \<phi>s i e2"
    have "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> ?e1s \<cdot> ?e2s : vsubst_ty (TyFreeVar a) i \<tau>" using 1 TmApp by auto
    then obtain X Z1 Z2 Zmin \<Gamma>1 \<Gamma>2 \<tau>arg q
      where H\<Gamma>x_eq: "?\<Gamma>x = \<Gamma>1 \<union> \<Gamma>2"
      and HZ_eq: "Z = Z1 \<union> Z2 \<union> filterScFreeX X \<union> Zmin"
      and Hcompat: "\<Gamma>1 \<sim> \<Gamma>2" and Hok: "\<Gamma>1 \<union> \<Gamma>2;\<Delta>,(a:\<star>) \<Turnstile> OK"
      and Ht1: "\<Gamma>1;\<Delta>,(a:\<star>);?Mx;Z1 \<turnstile> ?e1s : Qual q ScBot (\<tau>arg \<rightarrow>^X (vsubst_ty (TyFreeVar a) i \<tau>))"
      and Ht2: "\<Gamma>2;\<Delta>,(a:\<star>);?Mx;Z2 \<turnstile> ?e2s : \<tau>arg"
      and Hdollar1: "\<forall>\<rho>. \<rho> is free dollar sc in ?e1s \<longrightarrow> \<rho> \<notin> dom (Z2 \<union> filterScFreeX X \<union> Zmin)"
      and Hdollar2: "\<forall>\<rho>. \<rho> is free dollar sc in ?e2s \<longrightarrow> \<rho> \<notin> dom (Z1 \<union> filterScFreeX X \<union> Zmin)" 
      and Hmin: "minimalValidScope (scopeT (vsubst_ty (TyFreeVar a) i \<tau>)) Zmin"
      and Hscopeset:"filterScFreeX X \<Turnstile> ScopeSet" 
      using typing_app_elim[of "?\<Gamma>x" "\<Delta>,(a:\<star>)" "?Mx" Z "?e1s" "?e2s" "vsubst_ty (TyFreeVar a) i \<tau>"]
      by auto
    have "\<Delta>,(a:\<star>) \<Turnstile>' \<tau>arg" using Ht2 poly_ty_wf_type by auto
    then obtain \<tau>n where Hn: "\<tau>arg = vsubst_ty (TyFreeVar a) i \<tau>n"
      and Hfree: "\<not> a free in type \<tau>n" 
      using parametrize_ty[of "\<Delta>,(a:\<star>)" \<tau>arg a i] by auto
    have Hwf1: "\<Gamma>1;\<Delta>,(a:\<star>) \<Turnstile> OK" using Ht1 ty_wf_env by auto
    then obtain \<Gamma>r1
      where Hr1: "\<Gamma>1 = vsubst_ty_env (TyFreeVar a) i \<Gamma>r1"
      and Hr1_free: "\<not> a free in environ \<Gamma>r1"
      using parametrize_ty_env by blast
    have Hwf2: "\<Gamma>2;\<Delta>,(a:\<star>) \<Turnstile> OK" using Ht2 ty_wf_env by auto
    then obtain \<Gamma>r2
      where Hr2: "\<Gamma>2 = vsubst_ty_env (TyFreeVar a) i \<Gamma>r2"
      and Hr2_free: "\<not> a free in environ \<Gamma>r2"
      using parametrize_ty_env by blast
    have "?\<Gamma>x = vsubst_ty_env (TyFreeVar a) i (\<Gamma>r1\<union>\<Gamma>r2)" 
      using Hr1 Hr2 H\<Gamma>x_eq by (auto simp add: vsubst_ty_env_def)
    moreover
    have "\<not> a free in environ (\<Gamma>r1\<union>\<Gamma>r2)" using Hr1_free Hr2_free 
      apply (auto simp add: free_ty_env_def) by blast
    ultimately
    have "\<Gamma> = \<Gamma>r1 \<union> \<Gamma>r2" using H\<Gamma>free subst_subst_equal_env[of a i \<Gamma> "\<Gamma>r1\<union>\<Gamma>r2"] by fast
    hence "vsubst_ty_env \<phi>s i (\<Gamma>r1\<union>\<Gamma>r2);\<Delta> \<Turnstile> OK" using "1.prems" by auto
    also have "vsubst_ty_env \<phi>s i (\<Gamma>r1\<union>\<Gamma>r2) = vsubst_ty_env \<phi>s i \<Gamma>r1 \<union> vsubst_ty_env \<phi>s i \<Gamma>r2"
      by (auto simp add: vsubst_ty_env_def)
    finally
    have Hwf_union: "vsubst_ty_env \<phi>s i \<Gamma>r1 \<union> vsubst_ty_env \<phi>s i \<Gamma>r2;\<Delta> \<Turnstile> OK" by auto
    have Ht1: "(vsubst_ty_env \<phi>s i \<Gamma>r1);\<Delta>;?Mt;Z1 \<turnstile> 
                 ?e1\<phi> : (vsubst_ty \<phi>s i (Qual q ScBot(\<tau>n \<rightarrow>^X \<tau>)))"
      proof (rule IH[of e1 i \<Gamma>r1 \<Delta> M Z1 "Qual q ScBot (\<tau>n \<rightarrow>^X \<tau>)"])
      show "size e1 < size e_y" using TmApp by auto
    next
      show "vsubst_ty_env \<phi>s i \<Gamma>r1;\<Delta> \<Turnstile> OK" 
        using Hwf_union wf_env_subset by auto 
    next
      show "?Mt;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<phi>s" (*and "Zs \<Turnstile> \<phi>s"*) by assumption
    next
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1;\<Delta>,(a:\<star>);?Mx;Z1 \<turnstile> 
            ?e1s : vsubst_ty (TyFreeVar a) i (Qual q ScBot (\<tau>n \<rightarrow>^X \<tau>))" 
        using Hr1 Ht1 Hn by auto
    next
      show "a \<notin> set (free_tmty e1)" and "a \<notin> set (free_ty (Qual q ScBot (\<tau>n \<rightarrow>^X \<tau>)))"
        and "a \<notin> free_ty_env \<Gamma>r1" and "a \<notin> Environ.dom \<Delta>" 
        using "1.prems" Hfree TmApp Hr1_free by auto
(*    next
      show "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom Zs"
        using TmApp "1.prems" by (auto simp add: dom_def)
*)
    qed 
    have Ht2: "(vsubst_ty_env \<phi>s i \<Gamma>r2);\<Delta>;?Mt;Z2 \<turnstile> ?e2\<phi> : vsubst_ty \<phi>s i \<tau>n" 
    proof (rule IH[of e2 i \<Gamma>r2 \<Delta> M Z2 \<tau>n])
      show "size e2 < size e_y" using TmApp by auto
    next  
      show "vsubst_ty_env \<phi>s i \<Gamma>r2;\<Delta> \<Turnstile> OK" 
        using Hwf_union wf_env_subset by auto
    next
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma>r2;\<Delta>,(a:\<star>);?Mx;Z2 \<turnstile> 
            ?e2s : vsubst_ty (TyFreeVar a) i \<tau>n" 
        using Hr2 Ht2 Hn by auto
    next
      show "a \<notin> set (free_tmty e2)" and "a \<notin> set (free_ty \<tau>n)"
        and "a \<notin> free_ty_env \<Gamma>r2" and "a \<notin> Environ.dom \<Delta>" 
        using "1.prems" Hfree TmApp Hr2_free by auto
(*    next
      show "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom Zs"
        using TmApp "1.prems" by (auto simp add: dom_def)
*)
    qed
    have Hok1: "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1;\<Delta>,(a:\<star>) \<Turnstile> OK"  
      using Hok Hr1 wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2" _ \<Gamma>1] by auto
    have Hok2: "vsubst_ty_env (TyFreeVar a) i \<Gamma>r2;\<Delta>,(a:\<star>) \<Turnstile> OK"  
      using Hok Hr2 wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2" _ \<Gamma>2] by auto
    have Hfree: "\<not> a free in environ \<Gamma>" by assumption 
    have Hfoo:  "vsubst_ty_env \<phi>s i \<Gamma> = (vsubst_ty_env \<phi>s i \<Gamma>r1) \<union> vsubst_ty_env \<phi>s i \<Gamma>r2"
      proof (rule env_subst_eq_union_eq[of a i \<Gamma> \<Gamma>r1 \<Gamma>r2 "\<Delta>,(a:\<star>)" \<phi>s])
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma> =
       vsubst_ty_env (TyFreeVar a) i \<Gamma>r1 \<union> vsubst_ty_env (TyFreeVar a) i \<Gamma>r2" 
        using H\<Gamma>x_eq Hr1 Hr2 by auto
    next
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma>;\<Delta>,(a:\<star>) \<Turnstile> OK" using prems by auto
    next
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1;\<Delta>,(a:\<star>) \<Turnstile> OK" 
        and "vsubst_ty_env (TyFreeVar a) i \<Gamma>r2;\<Delta>,(a:\<star>) \<Turnstile> OK"
        using Hok1 Hok2 by auto
    next
      show "\<not> a free in environ \<Gamma>" and "\<not> a free in environ \<Gamma>r1" and "\<not> a free in environ \<Gamma>r2"
        using Hfree Hr1_free Hr2_free by auto
    qed
(*
    have Hdollar_y: "\<forall>\<rho>. \<rho> is free dollar sc in e_y \<longrightarrow> \<rho> \<notin> Environ.dom Zs" by assumption
    hence Hdollar_1s: "\<forall>\<rho>. \<rho> is free dollar sc in ?e1\<phi> \<longrightarrow> \<rho> \<notin> Environ.dom Zs" 
      using free_dollars_vsubst_tmty TmApp by auto
*)
    have "free_let_dollar_scope ?e1\<phi> = free_let_dollar_scope ?e1s"
      using free_dollars_vsubst_tmty by auto
    hence Hdollar1: "\<forall>\<rho>. \<rho> is free dollar sc in ?e1\<phi> \<longrightarrow> \<rho> \<notin> dom (Z2\<union>filterScFreeX X\<union>Zmin)"
      using (*Hdollar_1s*) Hdollar1 by (auto simp add: dom_def)
(*    have Hdollar_2s: "\<forall>\<rho>. \<rho> is free dollar sc in ?e2\<phi> \<longrightarrow> \<rho> \<notin> Environ.dom Zs" 
      using Hdollar_y free_dollars_vsubst_tmty TmApp by auto
*)
    have "free_let_dollar_scope ?e2\<phi> = free_let_dollar_scope ?e2s"
      using free_dollars_vsubst_tmty by auto
    hence Hdollar2: "\<forall>\<rho>. \<rho> is free dollar sc in ?e2\<phi> \<longrightarrow> \<rho> \<notin> dom (Z1\<union>filterScFreeX X\<union>Zmin)" 
      using (*Hdollar_2s*) Hdollar2 by (auto simp add: dom_def)
    have Hcompat: "vsubst_ty_env \<phi>s i \<Gamma>r1 \<sim> vsubst_ty_env \<phi>s i \<Gamma>r2" 
      using env_subst_compat Hcompat Hok Hr1 Hr2 by auto
    have Hok: "vsubst_ty_env \<phi>s i \<Gamma>r1 \<union> vsubst_ty_env \<phi>s i \<Gamma>r2;\<Delta> \<Turnstile> OK" 
      using Hfoo "1.prems" by auto
    have Hmin: "minimalValidScope (scopeT (vsubst_ty \<phi>s i \<tau>)) Zmin" using Hmin by (cases \<tau>, auto)
    have Htyp: "vsubst_ty_env \<phi>s i \<Gamma>r1 \<union> vsubst_ty_env \<phi>s i \<Gamma>r2;\<Delta>;?Mt;Z1\<union>Z2\<union>filterScFreeX X\<union>Zmin \<turnstile> 
      ?e1\<phi>\<cdot>?e2\<phi> : vsubst_ty \<phi>s i \<tau>"
      apply (rule t_app)
      using Ht1 Ht2 Hdollar1 Hdollar2 Hcompat Hok Hmin Hscopeset by auto 
    thus ?thesis using TmApp Htyp Hfoo HZ_eq by auto
  next;

    (* new case *)
    case (TmNew er)
    let ?ers = "vsubst_tmty (TyFreeVar a) i er"
    let ?er\<phi> = "vsubst_tmty \<phi>s i er"
    have "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> TmNew ?ers : vsubst_ty (TyFreeVar a) i \<tau>"
      using "1.prems" TmNew by auto
    then obtain \<tau>r where H\<tau>: "vsubst_ty (TyFreeVar a) i \<tau> = Qual qL ScBot (TyRef \<tau>r)"
      and Ht: "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> ?ers : \<tau>r" by auto
    from H\<tau> obtain \<phi>f where Hf: "\<tau> = Qual qL ScBot \<phi>f" 
      and "Qual qL ScBot (vsubst_pty (TyFreeVar a) i \<phi>f) = Qual qL ScBot (TyRef \<tau>r)"
      by (cases \<tau>, auto)
    then obtain \<tau>f where "\<phi>f = TyRef \<tau>f" by (cases \<phi>f, auto simp add: trivial_case)
    hence Hf: "\<tau> = Qual qL ScBot (TyRef \<tau>f)" using Hf by auto
    hence Ht: "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> ?ers : vsubst_ty (TyFreeVar a) i \<tau>f" using H\<tau> Ht by auto
    have Ht1: "(vsubst_ty_env \<phi>s i \<Gamma>);\<Delta>;?Mt;Z \<turnstile> ?er\<phi> : vsubst_ty \<phi>s i \<tau>f"
    proof (rule IH[of er i \<Gamma> \<Delta> M Z \<tau>f])
      show "size er < size e_y" using TmNew by auto
    next
      show "vsubst_ty_env \<phi>s i \<Gamma>;\<Delta> \<Turnstile> OK" and "?Mt;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<phi>s" (*and "Zs \<Turnstile> \<phi>s"*)
        by assumption
    next
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma>;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> 
            ?ers : vsubst_ty (TyFreeVar a) i \<tau>f" 
        using Ht H\<tau> by auto
    next
      show "a \<notin> set (free_tmty er)" and "a \<notin> set (free_ty \<tau>f)"
        and "a \<notin> free_ty_env \<Gamma>" and "a \<notin> Environ.dom \<Delta>" 
        using "1.prems" TmNew Hf by auto
(*    next
      show "\<forall>\<rho>. \<rho> is free dollar sc in er \<longrightarrow> \<rho> \<notin> Environ.dom Zs"
        using TmNew "1.prems" by (auto simp add: dom_def)
*)
    qed
    have Htyp: "vsubst_ty_env \<phi>s i \<Gamma>;\<Delta>;?Mt;Z \<turnstile> TmNew ?er\<phi> : vsubst_ty \<phi>s i \<tau>"
      using Ht1 Hf by auto
    show ?thesis using TmNew Htyp by auto
  next;

    (* assign case *)
    case (TmAssign e1 e2)
    let ?e1s = "vsubst_tmty (TyFreeVar a) i e1"
    let ?e2s = "vsubst_tmty (TyFreeVar a) i e2"
    let ?e1\<phi> = "vsubst_tmty \<phi>s i e1"
    let ?e2\<phi> = "vsubst_tmty \<phi>s i e2"
    have "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> ?e1s := ?e2s : vsubst_ty (TyFreeVar a) i \<tau>" using 1 TmAssign by auto
    then obtain Z1 Z2 \<Gamma>1 \<Gamma>2 \<tau>arg \<pi> 
      where H\<Gamma>x_eq: "?\<Gamma>x = \<Gamma>1 \<union> \<Gamma>2"
      and HZ_eq: "Z = Z1 \<union> Z2"
      and Hcompat: "\<Gamma>1 \<sim> \<Gamma>2" and Hok: "\<Gamma>1 \<union> \<Gamma>2;\<Delta>,(a:\<star>) \<Turnstile> OK"
      and Ht1: "\<Gamma>1;\<Delta>,(a:\<star>);?Mx;Z1 \<turnstile> ?e1s : Qual qU \<pi> (TyRef \<tau>arg)"
      and Ht2: "\<Gamma>2;\<Delta>,(a:\<star>);?Mx;Z2 \<turnstile> ?e2s : \<tau>arg"
      and Hdollar1: "\<forall>\<rho>. \<rho> is free dollar sc in ?e1s \<longrightarrow> \<rho> \<notin> Environ.dom Z2"
      and Hdollar2: "\<forall>\<rho>. \<rho> is free dollar sc in ?e2s \<longrightarrow> \<rho> \<notin> Environ.dom Z1" 
      and Hunit: "vsubst_ty (TyFreeVar a) i \<tau> = Qual qU ScBot TyUnit"
      and Hqual: "qualT \<tau>arg \<preceq> qU"
      using typing_assign_elim[of "?\<Gamma>x" "\<Delta>,(a:\<star>)" "?Mx" Z "?e1s" "?e2s" "vsubst_ty (TyFreeVar a) i \<tau>"]
      by auto
    have "\<Delta>,(a:\<star>) \<Turnstile>' \<tau>arg" using Ht2 poly_ty_wf_type by auto
    then obtain \<tau>n where Hn: "\<tau>arg = vsubst_ty (TyFreeVar a) i \<tau>n"
      and Hfree: "\<not> a free in type \<tau>n" 
      using parametrize_ty[of "\<Delta>,(a:\<star>)" \<tau>arg a i] by auto
    have Hwf1: "\<Gamma>1;\<Delta>,(a:\<star>) \<Turnstile> OK" using Ht1 ty_wf_env by auto
    then obtain \<Gamma>r1
      where Hr1: "\<Gamma>1 = vsubst_ty_env (TyFreeVar a) i \<Gamma>r1"
      and Hr1_free: "\<not> a free in environ \<Gamma>r1"
      using parametrize_ty_env by blast
    have Hwf2: "\<Gamma>2;\<Delta>,(a:\<star>) \<Turnstile> OK" using Ht2 ty_wf_env by auto
    then obtain \<Gamma>r2
      where Hr2: "\<Gamma>2 = vsubst_ty_env (TyFreeVar a) i \<Gamma>r2"
      and Hr2_free: "\<not> a free in environ \<Gamma>r2"
      using parametrize_ty_env by blast
    have "?\<Gamma>x = vsubst_ty_env (TyFreeVar a) i (\<Gamma>r1\<union>\<Gamma>r2)" 
      using Hr1 Hr2 H\<Gamma>x_eq by (auto simp add: vsubst_ty_env_def)
    moreover
    have "\<not> a free in environ (\<Gamma>r1\<union>\<Gamma>r2)" using Hr1_free Hr2_free 
      apply (auto simp add: free_ty_env_def) by blast
    ultimately
    have "\<Gamma> = \<Gamma>r1 \<union> \<Gamma>r2" using H\<Gamma>free subst_subst_equal_env[of a i \<Gamma> "\<Gamma>r1\<union>\<Gamma>r2"] by fast
    hence "vsubst_ty_env \<phi>s i (\<Gamma>r1\<union>\<Gamma>r2);\<Delta> \<Turnstile> OK" using "1.prems" by auto
    also have "vsubst_ty_env \<phi>s i (\<Gamma>r1\<union>\<Gamma>r2) = vsubst_ty_env \<phi>s i \<Gamma>r1 \<union> vsubst_ty_env \<phi>s i \<Gamma>r2"
      by (auto simp add: vsubst_ty_env_def)
    finally
    have Hwf_union: "vsubst_ty_env \<phi>s i \<Gamma>r1 \<union> vsubst_ty_env \<phi>s i \<Gamma>r2;\<Delta> \<Turnstile> OK" by auto
    have Ht1: "(vsubst_ty_env \<phi>s i \<Gamma>r1);\<Delta>;?Mt;Z1 \<turnstile> ?e1\<phi> : vsubst_ty \<phi>s i (Qual qU \<pi> (TyRef \<tau>n))"
    proof (rule IH[of e1 i \<Gamma>r1 \<Delta> M Z1 "Qual qU \<pi> (TyRef \<tau>n)" ])
      show "size e1 < size e_y" using TmAssign by auto
    next
        show "vsubst_ty_env \<phi>s i \<Gamma>r1;\<Delta> \<Turnstile> OK" 
        using Hwf_union wf_env_subset by auto
    next
      show "?Mt;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<phi>s" (*and "Zs \<Turnstile> \<phi>s"*) by assumption
    next
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1;\<Delta>,(a:\<star>);?Mx;Z1 \<turnstile> 
            ?e1s : vsubst_ty (TyFreeVar a) i (Qual qU \<pi> (TyRef \<tau>n))" 
        using Hr1 Ht1 Hn by auto
    next
      show "a \<notin> set (free_tmty e1)" and "a \<notin> set (free_ty (Qual qU \<pi> (TyRef \<tau>n)))"
        and "a \<notin> free_ty_env \<Gamma>r1" and "a \<notin> Environ.dom \<Delta>" 
        using "1.prems" Hfree TmAssign Hr1_free by auto
(*    next
      show "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom Zs"
        using TmAssign "1.prems" by (auto simp add: dom_def)
*)
    qed
    have Ht2: "(vsubst_ty_env \<phi>s i \<Gamma>r2);\<Delta>;?Mt;Z2 \<turnstile> ?e2\<phi> : vsubst_ty \<phi>s i \<tau>n" 
    proof (rule IH[of e2 i \<Gamma>r2 \<Delta> M Z2 \<tau>n])
      show "size e2 < size e_y" using TmAssign by auto
    next  
      show "vsubst_ty_env \<phi>s i \<Gamma>r2;\<Delta> \<Turnstile> OK" 
        using Hwf_union wf_env_subset by auto
    next
      show "?Mt;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<phi>s" (*and "Zs \<Turnstile> \<phi>s"*) by assumption
    next
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma>r2;\<Delta>,(a:\<star>);?Mx;Z2 \<turnstile> 
            ?e2s : vsubst_ty (TyFreeVar a) i \<tau>n" 
        using Hr2 Ht2 Hn by auto
    next
      show "a \<notin> set (free_tmty e2)" and "a \<notin> set (free_ty \<tau>n)"
        and "a \<notin> free_ty_env \<Gamma>r2" and "a \<notin> Environ.dom \<Delta>" 
        using "1.prems" Hfree TmAssign Hr2_free by auto
(*    next
      show "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom Zs"
        using TmAssign "1.prems" by (auto simp add: dom_def)
*)
    qed
    have Hok1: "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1;\<Delta>,(a:\<star>) \<Turnstile> OK"  
      using Hok Hr1 wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2" _ \<Gamma>1] by auto
    have Hok2: "vsubst_ty_env (TyFreeVar a) i \<Gamma>r2;\<Delta>,(a:\<star>) \<Turnstile> OK"  
      using Hok Hr2 wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2" _ \<Gamma>2] by auto
    have Hfree: "\<not> a free in environ \<Gamma>" by assumption 
    have Hfoo:  "vsubst_ty_env \<phi>s i \<Gamma> = (vsubst_ty_env \<phi>s i \<Gamma>r1) \<union> vsubst_ty_env \<phi>s i \<Gamma>r2"
      proof (rule env_subst_eq_union_eq[of a i \<Gamma> \<Gamma>r1 \<Gamma>r2 "\<Delta>,(a:\<star>)" \<phi>s])
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma> =
       vsubst_ty_env (TyFreeVar a) i \<Gamma>r1 \<union> vsubst_ty_env (TyFreeVar a) i \<Gamma>r2" 
        using H\<Gamma>x_eq Hr1 Hr2 by auto
    next
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma>;\<Delta>,(a:\<star>) \<Turnstile> OK" using prems by auto
    next
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1;\<Delta>,(a:\<star>) \<Turnstile> OK" 
        and "vsubst_ty_env (TyFreeVar a) i \<Gamma>r2;\<Delta>,(a:\<star>) \<Turnstile> OK"
        using Hok1 Hok2 by auto
    next
      show "\<not> a free in environ \<Gamma>" and "\<not> a free in environ \<Gamma>r1" and "\<not> a free in environ \<Gamma>r2"
        using Hfree Hr1_free Hr2_free by auto
    qed
(*    have Hdollar_y: "\<forall>\<rho>. \<rho> is free dollar sc in e_y \<longrightarrow> \<rho> \<notin> Environ.dom Zs" by assumption
    hence Hdollar_1s: "\<forall>\<rho>. \<rho> is free dollar sc in ?e1\<phi> \<longrightarrow> \<rho> \<notin> Environ.dom Zs" 
      using free_dollars_vsubst_tmty TmAssign by auto
*)
    have "free_let_dollar_scope ?e1\<phi> = free_let_dollar_scope ?e1s"
      using free_dollars_vsubst_tmty by auto
    hence Hdollar1: "\<forall>\<rho>. \<rho> is free dollar sc in ?e1\<phi> \<longrightarrow> \<rho> \<notin> Environ.dom Z2"
      using Hdollar1 by (auto simp add: dom_def)
(*
    have Hdollar_2s: "\<forall>\<rho>. \<rho> is free dollar sc in ?e2\<phi> \<longrightarrow> \<rho> \<notin> Environ.dom Zs" 
      using Hdollar_y free_dollars_vsubst_tmty TmAssign by auto
*)
    have "free_let_dollar_scope ?e2\<phi> = free_let_dollar_scope ?e2s"
      using free_dollars_vsubst_tmty by auto
    hence Hdollar2: "\<forall>\<rho>. \<rho> is free dollar sc in ?e2\<phi> \<longrightarrow> \<rho> \<notin> Environ.dom Z1" 
      using Hdollar2 by (auto simp add: dom_def)
    have Hcompat: "vsubst_ty_env \<phi>s i \<Gamma>r1 \<sim> vsubst_ty_env \<phi>s i \<Gamma>r2" 
      using env_subst_compat Hcompat Hok Hr1 Hr2 by auto
    have Hok: "vsubst_ty_env \<phi>s i \<Gamma>r1 \<union> vsubst_ty_env \<phi>s i \<Gamma>r2;\<Delta> \<Turnstile> OK" 
      using Hfoo "1.prems" by auto
    have Hunit: "\<tau> = Qual qU ScBot TyUnit" using Hunit proof (cases \<tau>, auto)
      fix \<phi> 
      assume "vsubst_pty (TyFreeVar a) i \<phi> = TyUnit" 
      thus "\<phi> = TyUnit" by (cases \<phi>, auto simp add: trivial_case)
    qed
    have "qualT \<tau>arg = qualT \<tau>n" using Hn by (cases \<tau>n, auto)
    hence Hqual: "qualT (vsubst_ty \<phi>s i \<tau>n) \<preceq> qU" using Hqual by (cases \<tau>n, auto)
    have Htyp: "vsubst_ty_env \<phi>s i \<Gamma>r1 \<union> vsubst_ty_env \<phi>s i \<Gamma>r2;\<Delta>;?Mt;Z1\<union>Z2 \<turnstile> 
      ?e1\<phi> := ?e2\<phi> : vsubst_ty \<phi>s i \<tau>"
      using Ht1 Ht2 Hdollar1 Hdollar2 Hcompat Hok Hqual Hunit t_assign[where \<tau>="vsubst_ty \<phi>s i \<tau>n"]
    by auto 
    show ?thesis using TmAssign Htyp Hfoo HZ_eq by auto
  next;

    (* swap case *)
    case (TmSwap e1 e2)
    let ?e1s = "vsubst_tmty (TyFreeVar a) i e1"
    let ?e2s = "vsubst_tmty (TyFreeVar a) i e2"
    let ?e1\<phi> = "vsubst_tmty \<phi>s i e1"
    let ?e2\<phi> = "vsubst_tmty \<phi>s i e2"
    have "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> ?e1s :=: ?e2s : vsubst_ty (TyFreeVar a) i \<tau>" using 1 TmSwap by auto
    then obtain Z1 Z2 \<Gamma>1 \<Gamma>2 \<pi> 
      where H\<Gamma>x_eq: "?\<Gamma>x = \<Gamma>1 \<union> \<Gamma>2"
      and HZ_eq: "Z = Z2 \<union> Z1"
      and Hcompat: "\<Gamma>1 \<sim> \<Gamma>2" and Hok: "\<Gamma>1 \<union> \<Gamma>2;\<Delta>,(a:\<star>) \<Turnstile> OK"
      and Ht1: "\<Gamma>1;\<Delta>,(a:\<star>);?Mx;Z1 \<turnstile> ?e1s : Qual qU \<pi> (TyRef (vsubst_ty (TyFreeVar a) i \<tau>))"
      and Ht2: "\<Gamma>2;\<Delta>,(a:\<star>);?Mx;Z2 \<turnstile> ?e2s : vsubst_ty (TyFreeVar a) i \<tau>"
      and Hdollar1: "\<forall>\<rho>. \<rho> is free dollar sc in ?e1s \<longrightarrow> \<rho> \<notin> Environ.dom Z2"
      and Hdollar2: "\<forall>\<rho>. \<rho> is free dollar sc in ?e2s \<longrightarrow> \<rho> \<notin> Environ.dom Z1" 
      and Hqual: "qualT (vsubst_ty (TyFreeVar a) i \<tau>) = qL"
      using typing_swap_elim[of "?\<Gamma>x" "\<Delta>,(a:\<star>)" "?Mx" Z "?e1s" "?e2s" "vsubst_ty (TyFreeVar a) i \<tau>"]
      by auto
    have Hwf1: "\<Gamma>1;\<Delta>,(a:\<star>) \<Turnstile> OK" using Ht1 ty_wf_env by auto
    then obtain \<Gamma>r1
      where Hr1: "\<Gamma>1 = vsubst_ty_env (TyFreeVar a) i \<Gamma>r1"
      and Hr1_free: "\<not> a free in environ \<Gamma>r1"
      using parametrize_ty_env by blast
    have Hwf2: "\<Gamma>2;\<Delta>,(a:\<star>) \<Turnstile> OK" using Ht2 ty_wf_env by auto
    then obtain \<Gamma>r2
      where Hr2: "\<Gamma>2 = vsubst_ty_env (TyFreeVar a) i \<Gamma>r2"
      and Hr2_free: "\<not> a free in environ \<Gamma>r2"
      using parametrize_ty_env by blast
    have "?\<Gamma>x = vsubst_ty_env (TyFreeVar a) i (\<Gamma>r1\<union>\<Gamma>r2)" 
      using Hr1 Hr2 H\<Gamma>x_eq by (auto simp add: vsubst_ty_env_def)
    moreover
    have "\<not> a free in environ (\<Gamma>r1\<union>\<Gamma>r2)" using Hr1_free Hr2_free 
      apply (auto simp add: free_ty_env_def) by blast
    ultimately
    have "\<Gamma> = \<Gamma>r1 \<union> \<Gamma>r2" using H\<Gamma>free subst_subst_equal_env[of a i \<Gamma> "\<Gamma>r1\<union>\<Gamma>r2"] by fast
    hence "vsubst_ty_env \<phi>s i (\<Gamma>r1\<union>\<Gamma>r2);\<Delta> \<Turnstile> OK" using "1.prems" by auto
    also have "vsubst_ty_env \<phi>s i (\<Gamma>r1\<union>\<Gamma>r2) = vsubst_ty_env \<phi>s i \<Gamma>r1 \<union> vsubst_ty_env \<phi>s i \<Gamma>r2"
      by (auto simp add: vsubst_ty_env_def)
    finally
    have Hwf_union: "vsubst_ty_env \<phi>s i \<Gamma>r1 \<union> vsubst_ty_env \<phi>s i \<Gamma>r2;\<Delta> \<Turnstile> OK" by auto
    have Ht1: "(vsubst_ty_env \<phi>s i \<Gamma>r1);\<Delta>;?Mt;Z1 \<turnstile> ?e1\<phi> : vsubst_ty \<phi>s i (Qual qU \<pi> (TyRef \<tau>))"
    proof (rule IH[of e1 i \<Gamma>r1 \<Delta> M Z1 "Qual qU \<pi> (TyRef \<tau>)"])
      show "size e1 < size e_y" using TmSwap by auto
    next
      show "vsubst_ty_env \<phi>s i \<Gamma>r1;\<Delta> \<Turnstile> OK" 
        using Hwf_union wf_env_subset by auto
    next
      show "?Mt;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<phi>s" (*and "Zs \<Turnstile> \<phi>s"*) by assumption
    next
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1;\<Delta>,(a:\<star>);?Mx;Z1 \<turnstile> 
            ?e1s : vsubst_ty (TyFreeVar a) i (Qual qU \<pi> (TyRef \<tau>))" 
        using Hr1 Ht1 by auto
    next
      show "a \<notin> set (free_tmty e1)" and "a \<notin> set (free_ty (Qual qU \<pi> (TyRef \<tau>)))"
        and "a \<notin> free_ty_env \<Gamma>r1" and "a \<notin> Environ.dom \<Delta>" 
        using "1.prems" TmSwap Hr1_free by auto
(*    next
      show "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom Zs"
        using TmSwap "1.prems" by (auto simp add: dom_def)
*)
    qed
    have Ht2: "(vsubst_ty_env \<phi>s i \<Gamma>r2);\<Delta>;?Mt;Z2 \<turnstile> ?e2\<phi> : vsubst_ty \<phi>s i \<tau>"
    proof (rule IH[of e2 i \<Gamma>r2 \<Delta> M Z2 \<tau>])
      show "size e2 < size e_y" using TmSwap by auto
    next  
      show "vsubst_ty_env \<phi>s i \<Gamma>r2;\<Delta> \<Turnstile> OK" 
        using Hwf_union wf_env_subset by auto
    next
      show "?Mt;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<phi>s" (*and "Zs \<Turnstile> \<phi>s"*) by assumption
    next
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma>r2;\<Delta>,(a:\<star>);?Mx;Z2 \<turnstile> 
            ?e2s : vsubst_ty (TyFreeVar a) i \<tau>" 
        using Hr2 Ht2 by auto
    next
      show "a \<notin> set (free_tmty e2)" and "a \<notin> set (free_ty \<tau>)"
        and "a \<notin> free_ty_env \<Gamma>r2" and "a \<notin> Environ.dom \<Delta>" 
        using "1.prems" TmSwap Hr2_free by auto
(*
    next
      show "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom Zs"
        using TmSwap "1.prems" by (auto simp add: dom_def)
*)
    qed
    have Hok1: "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1;\<Delta>,(a:\<star>) \<Turnstile> OK"  
      using Hok Hr1 wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2" _ \<Gamma>1] by auto
    have Hok2: "vsubst_ty_env (TyFreeVar a) i \<Gamma>r2;\<Delta>,(a:\<star>) \<Turnstile> OK"  
      using Hok Hr2 wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2" _ \<Gamma>2] by auto
    have Hfree: "\<not> a free in environ \<Gamma>" by assumption 
    have Hfoo:  "vsubst_ty_env \<phi>s i \<Gamma> = (vsubst_ty_env \<phi>s i \<Gamma>r1) \<union> vsubst_ty_env \<phi>s i \<Gamma>r2"
      proof (rule env_subst_eq_union_eq[of a i \<Gamma> \<Gamma>r1 \<Gamma>r2 "\<Delta>,(a:\<star>)" \<phi>s])
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma> =
       vsubst_ty_env (TyFreeVar a) i \<Gamma>r1 \<union> vsubst_ty_env (TyFreeVar a) i \<Gamma>r2" 
        using H\<Gamma>x_eq Hr1 Hr2 by auto
    next
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma>;\<Delta>,(a:\<star>) \<Turnstile> OK" using prems by auto
    next
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1;\<Delta>,(a:\<star>) \<Turnstile> OK" 
        and "vsubst_ty_env (TyFreeVar a) i \<Gamma>r2;\<Delta>,(a:\<star>) \<Turnstile> OK"
        using Hok1 Hok2 by auto
    next
      show "\<not> a free in environ \<Gamma>" and "\<not> a free in environ \<Gamma>r1" and "\<not> a free in environ \<Gamma>r2"
        using Hfree Hr1_free Hr2_free by auto
    qed
(*
    have Hdollar_y: "\<forall>\<rho>. \<rho> is free dollar sc in e_y \<longrightarrow> \<rho> \<notin> Environ.dom Zs" by assumption
    hence Hdollar_1s: "\<forall>\<rho>. \<rho> is free dollar sc in ?e1\<phi> \<longrightarrow> \<rho> \<notin> Environ.dom Zs" 
      using free_dollars_vsubst_tmty TmSwap by auto
*)
    have "free_let_dollar_scope ?e1\<phi> = free_let_dollar_scope ?e1s"
      using free_dollars_vsubst_tmty by auto
    hence Hdollar1: "\<forall>\<rho>. \<rho> is free dollar sc in ?e1\<phi> \<longrightarrow> \<rho> \<notin> Environ.dom Z2"
      using Hdollar1 by (auto simp add: dom_def)
(*
    have Hdollar_2s: "\<forall>\<rho>. \<rho> is free dollar sc in ?e2\<phi> \<longrightarrow> \<rho> \<notin> Environ.dom Zs" 
      using Hdollar_y free_dollars_vsubst_tmty TmSwap by auto
*)
    have "free_let_dollar_scope ?e2\<phi> = free_let_dollar_scope ?e2s"
      using free_dollars_vsubst_tmty by auto
    hence Hdollar2: "\<forall>\<rho>. \<rho> is free dollar sc in ?e2\<phi> \<longrightarrow> \<rho> \<notin> Environ.dom Z1" 
      using Hdollar2 by (auto simp add: dom_def)
    have Hcompat: "vsubst_ty_env \<phi>s i \<Gamma>r1 \<sim> vsubst_ty_env \<phi>s i \<Gamma>r2" 
      using env_subst_compat Hcompat Hok Hr1 Hr2 by auto
    have Hok: "vsubst_ty_env \<phi>s i \<Gamma>r1 \<union> vsubst_ty_env \<phi>s i \<Gamma>r2;\<Delta> \<Turnstile> OK" 
      using Hfoo "1.prems" by auto
    have Hqual: "qualT (vsubst_ty \<phi>s i \<tau>) = qL" using Hqual by (cases \<tau>, auto)
    hence Hqual: "qL \<preceq> qualT (vsubst_ty \<phi>s i \<tau>) " by auto 
    have Htyp: "vsubst_ty_env \<phi>s i \<Gamma>r1 \<union> vsubst_ty_env \<phi>s i \<Gamma>r2;\<Delta>;?Mt;Z2\<union>Z1 \<turnstile> 
      ?e1\<phi> :=: ?e2\<phi> : vsubst_ty \<phi>s i \<tau>"
      apply (rule t_swap) using Ht1 Ht2 Hdollar1 Hdollar2 Hcompat Hok Hqual by auto
    show ?thesis using TmSwap Htyp HZ_eq Hfoo by auto
  next;

    (* free case *)
    case (TmFree er)
    let ?ers = "vsubst_tmty (TyFreeVar a) i er"
    let ?er\<phi> = "vsubst_tmty \<phi>s i er"
    have "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> TmFree ?ers : vsubst_ty (TyFreeVar a) i \<tau>"
      using "1.prems" TmFree by auto
    then obtain Z\<tau> Ze 
      where Ht: "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Ze \<turnstile> ?ers : vsubst_ty (TyFreeVar a) i (Qual qL ScBot (TyRef \<tau>))"
      and Hmin: "minimalValidScope (scopeT (vsubst_ty (TyFreeVar a) i \<tau>)) Z\<tau>"
      and HZ_eq: "Z = Ze \<union> Z\<tau>"
      and Hdollar: "\<forall>\<rho>. \<rho> is free dollar sc in ?ers \<longrightarrow> \<not> Z\<tau> defines \<rho>"
      using typing_free_elim[of "?\<Gamma>x" "\<Delta>,(a:\<star>)" "?Mx" Z "?ers" "vsubst_ty (TyFreeVar a) i \<tau>"] by auto
    have Ht1: "(vsubst_ty_env \<phi>s i \<Gamma>);\<Delta>;?Mt;Ze \<turnstile> ?er\<phi> : vsubst_ty \<phi>s i (Qual qL ScBot (TyRef \<tau>))"
      proof (rule IH[of er i \<Gamma> \<Delta> M Ze "Qual qL ScBot (TyRef \<tau>)"])
      show "size er < size e_y" using TmFree by auto
    next
      show "vsubst_ty_env \<phi>s i \<Gamma>;\<Delta> \<Turnstile> OK" and "?Mt;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<phi>s" (*and "Zs \<Turnstile> \<phi>s" *)
        by assumption
    next
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma>;\<Delta>,(a:\<star>);?Mx;Ze \<turnstile> 
            ?ers : vsubst_ty (TyFreeVar a) i (Qual qL ScBot (TyRef \<tau>))" 
        using Ht by auto
    next
      show "a \<notin> set (free_tmty er)" and "a \<notin> set (free_ty (Qual qL ScBot (TyRef \<tau>)))"
        and "a \<notin> free_ty_env \<Gamma>" and "a \<notin> Environ.dom \<Delta>" 
        using "1.prems" TmFree by auto
(*    next
      show "\<forall>\<rho>. \<rho> is free dollar sc in er \<longrightarrow> \<rho> \<notin> Environ.dom Zs"
        using TmFree "1.prems" by (auto simp add: dom_def)
*)
    qed
    have Hmin: "minimalValidScope (scopeT (vsubst_ty \<phi>s i \<tau>)) Z\<tau>" using Hmin by (cases \<tau>, auto)
(*    have Hdollar_y: "\<forall>\<rho>. \<rho> is free dollar sc in e_y \<longrightarrow> \<rho> \<notin> Environ.dom Zs" by assumption
    hence Hdollar_1s: "\<forall>\<rho>. \<rho> is free dollar sc in ?er\<phi> \<longrightarrow> \<rho> \<notin> Environ.dom Zs" 
      using free_dollars_vsubst_tmty TmFree by auto
*)
    have "free_let_dollar_scope ?er\<phi> = free_let_dollar_scope ?ers"
      using free_dollars_vsubst_tmty by auto
    hence Hdollar: "\<forall>\<rho>. \<rho> is free dollar sc in ?er\<phi> \<longrightarrow> \<not> Z\<tau> defines \<rho>"
      using Hdollar by (auto simp add: dom_def)
    have Htyp: "vsubst_ty_env \<phi>s i \<Gamma>;\<Delta>;?Mt;Ze\<union>Z\<tau> \<turnstile> TmFree ?er\<phi> : vsubst_ty \<phi>s i \<tau>"
      using Ht1 Hdollar Hmin by auto
    show ?thesis using TmFree Htyp HZ_eq by auto
  next;

    (* term abstraction case *)
    case (TmAbs q \<pi> \<tau>arg er)
    hence Ht: "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> \<lambda>^q^\<pi>[vsubst_ty (TyFreeVar a) i \<tau>arg]. vsubst_tmty (TyFreeVar a) i er
                          : vsubst_ty (TyFreeVar a) i \<tau>"
      using 1 by auto
    obtain X \<tau>2 Lx
      where H\<tau>: "vsubst_ty (TyFreeVar a) i \<tau> = Qual q \<pi> ((vsubst_ty (TyFreeVar a) i \<tau>arg) \<rightarrow>^X \<tau>2)"
      and Ht: "\<forall>x. ?\<Gamma>x,(x:vsubst_ty (TyFreeVar a) i \<tau>arg);\<Delta>,(a:\<star>) \<Turnstile> OK \<and>
               ?\<Gamma>x \<sim> {x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>arg} \<and> x \<notin> Lx \<longrightarrow>
          ?\<Gamma>x,(x:vsubst_ty (TyFreeVar a) i \<tau>arg);\<Delta>,(a:\<star>);?Mx;filterScFreeX X \<turnstile> 
           vsubst_tm (TmFreeVar x) 0 (vsubst_tmty (TyFreeVar a) i er) : \<tau>2"
      and Hqual: "qualG ?\<Gamma>x \<preceq> q" and Hwf: "?\<Gamma>x;\<Delta>,(a:\<star>) \<Turnstile> OK" and HwfM: "?Mx;\<Delta>,(a:\<star>) \<Turnstile> OK"
      and HX: "filterScVarX X = {}" 
      and H\<Delta>: "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty (TyFreeVar a) i \<tau>arg"
      and Hmin: "minimalValidScope \<pi> Z"
      and Hfin: "finite Lx" 
      apply (rule typing_abs_elim[of "?\<Gamma>x" "\<Delta>,(a:\<star>)" "?Mx" Z q \<pi> "vsubst_ty (TyFreeVar a) i \<tau>arg" 
        "vsubst_tmty (TyFreeVar a) i er" "vsubst_ty (TyFreeVar a) i \<tau>"]) using Ht by auto
    have Harg_free: "\<not> a free in type \<tau>arg" using prems by auto
    have "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty (TyFreeVar a) i \<tau>" using "1.prems" poly_ty_wf_type by auto
    hence "\<Delta>,(a:\<star>) \<Turnstile>' \<tau>2" using H\<tau> by auto
    then obtain \<tau>r where Htr: "\<tau>2 = vsubst_ty (TyFreeVar a) i \<tau>r" and Hfx: "\<not> a free in type \<tau>r"
      using parametrize_ty[of "\<Delta>,(a:\<star>)" \<tau>2] by blast
    moreover
    have "\<not> a free in type \<tau>" by assumption
    ultimately
    have H\<tau>_eq: "\<tau> = Qual q \<pi> (\<tau>arg \<rightarrow>^X \<tau>r)" using H\<tau> Harg_free subst_subst_equal_tyty by auto
    have "?\<Gamma>t;\<Delta>;?Mt;Z \<turnstile> \<lambda>^q^\<pi>[vsubst_ty \<phi>s i \<tau>arg].vsubst_tmty \<phi>s i er :
                         Qual q \<pi> (vsubst_ty \<phi>s i \<tau>arg \<rightarrow>^X vsubst_ty \<phi>s i \<tau>r)"
    proof (rule t_abs[where L="Lx\<union>dom ?\<Gamma>x"])
      show "\<forall>x. ?\<Gamma>t,(x:vsubst_ty \<phi>s i \<tau>arg);\<Delta> \<Turnstile> OK \<and> ?\<Gamma>t \<sim> {x\<rhd>vsubst_ty \<phi>s i \<tau>arg} \<and> x \<notin> (Lx\<union>dom ?\<Gamma>x)
        \<longrightarrow>
        ?\<Gamma>t,(x:vsubst_ty \<phi>s i \<tau>arg);\<Delta>;?Mt;filterScFreeX X \<turnstile> freshen_tm x (vsubst_tmty \<phi>s i er) : 
                                                             vsubst_ty \<phi>s i \<tau>r" 
      proof (auto simp del: filterScFreeX)
        fix x
        assume Hwf: "vsubst_ty_env \<phi>s i \<Gamma>,(x:vsubst_ty \<phi>s i \<tau>arg);\<Delta> \<Turnstile> OK"
          and Hcompat: "vsubst_ty_env \<phi>s i \<Gamma> \<sim> {x\<rhd>vsubst_ty \<phi>s i \<tau>arg}"
          and Hfresh: "x \<notin> dom ?\<Gamma>x"
          and Hlx: "x \<notin> Lx"
        have Ht: "vsubst_ty_env \<phi>s i (\<Gamma>,(x:\<tau>arg));\<Delta>;?Mt;filterScFreeX X \<turnstile> 
           vsubst_tmty \<phi>s i (vsubst_tm (TmFreeVar x) 0 er) : vsubst_ty \<phi>s i \<tau>r" 
        proof (rule IH[of "vsubst_tm (TmFreeVar x) 0 er" i "\<Gamma>,(x:\<tau>arg)" \<Delta> M "filterScFreeX X" \<tau>r])
          show "size (vsubst_tm (TmFreeVar x) 0 er) < size e_y" using TmAbs by auto
        next
          note Hwf
          also have "?\<Gamma>t,(x: vsubst_ty \<phi>s i \<tau>arg) = vsubst_ty_env \<phi>s i (\<Gamma>,(x:\<tau>arg))"
            by (auto simp add: vsubst_ty_env_def)
          finally
          show "vsubst_ty_env \<phi>s i (\<Gamma>,(x:\<tau>arg));\<Delta> \<Turnstile> OK" by auto
        next
          show "vsubst_ty_env \<phi>s i M;\<Delta> \<Turnstile> OK" by assumption
        next
          show "\<Delta> \<Turnstile>' \<phi>s" by assumption
        next
(*          show "Zs \<Turnstile> \<phi>s" by assumption
        next
          have "\<forall>\<rho>. \<rho> is free dollar sc in e_y \<longrightarrow> \<rho> \<notin> Environ.dom Zs" by assumption
          thus "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tm (TmFreeVar x) 0 er \<longrightarrow> \<rho> \<notin> Environ.dom Zs"
            using free_dollars_vsubst_tm TmAbs by auto
*)
        next
          show "\<not> a free in term (vsubst_tm (TmFreeVar x) 0 er)" 
            using free_tmty_subst_equal "1.prems" TmAbs by auto
        next
          show "\<not> a free in type \<tau>r" using Hfx by auto
        next
          have "\<not> a free in environ \<Gamma>" by assumption
          thus "\<not> a free in environ (\<Gamma>,(x:\<tau>arg))" using Harg_free by (auto simp add: free_ty_env_def)
        next
          show "\<not> \<Delta> defines a" by assumption
        next
          have "?\<Gamma>x;\<Delta>,(a:\<star>) \<Turnstile> OK" by assumption
          hence "?\<Gamma>x,(x:vsubst_ty (TyFreeVar a) i \<tau>arg);\<Delta>,(a:\<star>) \<Turnstile> OK"
            and "?\<Gamma>x \<sim> {x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>arg}"
            and "x \<notin> Lx" 
            using Hfresh Hlx H\<Delta> by (auto simp add: dom_def wf_env_def env_compat_def)
          hence "?\<Gamma>x,(x: vsubst_ty (TyFreeVar a) i \<tau>arg);\<Delta>,(a:\<star>);?Mx;filterScFreeX X \<turnstile>
            vsubst_tm (TmFreeVar x) 0 (vsubst_tmty (TyFreeVar a) i er) : vsubst_ty (TyFreeVar a) i \<tau>r"
            using Ht Htr by auto
          also have "vsubst_tm (TmFreeVar x) 0 (vsubst_tmty (TyFreeVar a) i er) = 
            vsubst_tmty (TyFreeVar a) i (vsubst_tm (TmFreeVar x) 0 er)" 
            using subst_subst_tytm_swap by auto
          also have "?\<Gamma>x,(x: vsubst_ty (TyFreeVar a) i \<tau>arg) = vsubst_ty_env (TyFreeVar a) i (\<Gamma>,(x:\<tau>arg))"
            by (auto simp add: vsubst_ty_env_def)
          finally show "vsubst_ty_env (TyFreeVar a) i (\<Gamma>,(x:\<tau>arg));\<Delta>,(a:\<star>);?Mx;filterScFreeX X \<turnstile>
            vsubst_tmty (TyFreeVar a) i (vsubst_tm (TmFreeVar x) 0 er) : vsubst_ty (TyFreeVar a) i \<tau>r"
            by auto
        qed
        also
        have "vsubst_tmty \<phi>s i (vsubst_tm (TmFreeVar x) 0 er) =
              vsubst_tm (TmFreeVar x) 0 (vsubst_tmty \<phi>s i er)" 
          using subst_subst_tytm_swap[of x] by auto
        also
        have "vsubst_ty_env \<phi>s i (\<Gamma>,(x:\<tau>arg)) = ?\<Gamma>t,(x:vsubst_ty \<phi>s i \<tau>arg)"
          by (auto simp add: vsubst_ty_env_def)
        finally
        show Ht: "?\<Gamma>t,(x:vsubst_ty \<phi>s i \<tau>arg);\<Delta>;?Mt;filterScFreeX X \<turnstile> 
           vsubst_tm (TmFreeVar x) 0 (vsubst_tmty \<phi>s i er) : vsubst_ty \<phi>s i \<tau>r" by auto
      qed
    next
      show "qualG vsubst_ty_env \<phi>s i \<Gamma> \<preceq> q" using Hqual qualG_vsubst_left by auto
    next
      show "vsubst_ty_env \<phi>s i \<Gamma>;\<Delta> \<Turnstile> OK" 
        and "vsubst_ty_env \<phi>s i M;\<Delta> \<Turnstile> OK"
        by assumption
    next
      show 
        "filterScVarX X = {}" and "minimalValidScope \<pi> Z" using Hmin HX by auto
    next
      have "vsubst_ty_env \<phi>s i \<Gamma>;\<Delta> \<Turnstile> OK" by assumption 
      moreover
      have "\<Delta>,(a:\<star>) \<Turnstile>' \<phi>s" using "1.prems" poly_tenv_cons_pty by auto
      hence Hwf: "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty \<phi>s i \<tau>arg" 
        using H\<Delta> poly_subst_var_type_ok[of "\<Delta>,(a:\<star>)"] assms by auto
      show "\<Delta> \<Turnstile>' vsubst_ty \<phi>s i \<tau>arg" proof (cases "a free in pretype \<phi>s")
        case False 
        hence "\<not> a free in type (vsubst_ty \<phi>s i \<tau>arg)" using not_free_subst_ty Harg_free by auto
        moreover
        have "vsubst_ty (TyFreeVar a) i \<tau> = vsubst_ty (TyFreeVar a) i (Qual q \<pi> (\<tau>arg \<rightarrow>^X \<tau>r))"
          using H\<tau>  Htr by auto
        ultimately
        show ?thesis using False Hwf poly_wf_ty_strength by auto
      next
        case True
        hence "\<Delta> defines a" using poly_wf_free_defined assms by auto
        hence "\<Delta>,(a:\<star>) = \<Delta>" proof (auto simp add: dom_def)
          fix b
          assume "(a, b) \<in> \<Delta>"
          thus "a\<rhd>\<star> \<in> \<Delta>" by (cases b, auto)
        qed
        thus ?thesis using Hwf by auto
      qed
    next
      have "?\<Gamma>x;\<Delta>,(a:\<star>) \<Turnstile> OK" by assumption
      hence "finite ?\<Gamma>x" by (auto simp add: wf_env_def)
      hence "finite (dom ?\<Gamma>x)" using dom_finite by auto
      thus "finite (Lx \<union> dom ?\<Gamma>x)" by auto
    qed
    thus ?thesis using TmAbs H\<tau>_eq by auto
next;

    (* type application case *)
    case (TmTapp e \<phi>)
    let ?er = "vsubst_tmty (TyFreeVar a) i e"
    let ?\<phi>r = "vsubst_pty (TyFreeVar a) i \<phi>"
    let ?et = "vsubst_tmty \<phi>s i e"
    let ?\<phi>t = "vsubst_pty \<phi>s i \<phi>"
    have Ht1: "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> ?er\<langle>?\<phi>r\<rangle> : vsubst_ty (TyFreeVar a) i \<tau>" using TmTapp "1.prems" by auto
    then obtain Ze Z\<tau> \<tau>e q X 
      where HZ_eq: "Z = Ze \<union> filterScFreeX X \<union> Z\<tau>"
      and H\<tau>: "vsubst_ty (TyFreeVar a) i \<tau> = vsubst_ty ?\<phi>r 0 \<tau>e"
      and Htyp: "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Ze \<turnstile> ?er : Qual q ScBot (\<forall>.\<tau>e^X)"
      and Hmin: "minimalValidScope (scopeT \<tau>e) Z\<tau>"
      and Hscopeset: "filterScFreeX X \<Turnstile> ScopeSet"
      and Hdollar: "\<forall>\<rho>. \<rho> is free dollar sc in ?er \<longrightarrow> \<rho> \<notin> dom (filterScFreeX X \<union> Z\<tau>)"
      and Hnew: "\<forall>j \<rho>. vsubst_sc_pty (ScFree \<rho>) j (vsubst_pty (TyFreeVar a) i \<phi>) = vsubst_pty (TyFreeVar a) i \<phi>"
      using typing_tyapp_elim[of "?\<Gamma>x" "\<Delta>,(a:\<star>)" "?Mx" Z "?er" "?\<phi>r" "vsubst_ty (TyFreeVar a) i \<tau>"]
      by auto
    have Hwf_pty: "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_pty (TyFreeVar a) i \<phi>" using Ht1 by auto 
    have "\<Delta>,(a:\<star>) \<Turnstile>' Qual q ScBot (\<forall>.\<tau>e^X)"  using Htyp poly_ty_wf_type by auto
    then obtain \<tau>n' where Hn: "Qual q ScBot (\<forall>.\<tau>e^X) = vsubst_ty (TyFreeVar a) i \<tau>n'"
      and  Hfree: "\<not> a free in type \<tau>n'" 
      using parametrize_ty[of "\<Delta>,(a:\<star>)" "Qual q ScBot (\<forall>.\<tau>e^X)" a] by auto
    then obtain \<tau>n where Hn2: "\<tau>n' = Qual q ScBot (\<forall>.\<tau>n^X)" proof (cases \<tau>n', auto)
      fix \<phi>i
      assume "\<And>\<tau>n. \<phi>i = \<forall>.\<tau>n^X \<Longrightarrow> thesis"
        and "\<forall>.\<tau>e^X = vsubst_pty (TyFreeVar a) i \<phi>i"
      thus "thesis" by (cases \<phi>i, auto simp add: split_if)
    qed
    hence Hr: "\<tau>e = vsubst_ty (TyFreeVar a) (Suc i) \<tau>n"
      and Hr_free: "\<not> a free in type \<tau>n"
      using Hn Hfree by auto
    hence Heqt: "vsubst_ty (TyFreeVar a) i \<tau> = vsubst_ty ?\<phi>r 0 (vsubst_ty (TyFreeVar a) (Suc i) \<tau>n)"
      using Hr H\<tau> by auto
(*    have H\<tau>_eq: "vsubst_ty \<phi>s i \<tau> = vsubst_ty ?\<phi>t 0 (vsubst_ty \<phi>s (Suc i) \<tau>n)" *)
    have H\<tau>_eq: "vsubst_ty ?\<phi>t 0 (vsubst_ty \<phi>s (Suc i) \<tau>n) = vsubst_ty \<phi>s i \<tau>" 
    proof (rule vsubst_equal_sanity)
      have "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty (TyFreeVar a) i \<tau>" using Ht1 poly_ty_wf_type by auto
      hence "set (free_db_ty (vsubst_ty (TyFreeVar a) i \<tau>)) = {}" 
        using poly_wf_no_free_db_ty by auto
      thus A1: "set (free_db_ty \<tau>) \<subseteq> {i}" using vsubst_ty_free_db_ty by auto
    next
      have "\<Delta>,(a:\<star>) \<Turnstile>'
        vsubst_ty (vsubst_pty (TyFreeVar a) i \<phi>) 0 (vsubst_ty (TyFreeVar a) (Suc i) \<tau>n)"
        using Heqt Ht1 poly_ty_wf_type by auto
      hence "set (free_db_ty
        (vsubst_ty (vsubst_pty (TyFreeVar a) i \<phi>) 0 (vsubst_ty (TyFreeVar a) (Suc i) \<tau>n))) = {}"
        using poly_wf_no_free_db_ty by auto
      moreover
      have "set (free_db_ty
        (vsubst_ty (vsubst_pty (TyFreeVar a) i \<phi>) 0 (vsubst_ty (TyFreeVar a) (Suc i) \<tau>n)))
        = set (free_db_ty ((vsubst_ty (TyFreeVar a) (Suc i) \<tau>n))) - {0}"
        using vsubst_ty_free_db_ty[of "vsubst_pty (TyFreeVar a) i \<phi>"] 
          poly_wf_no_free_db_pty[of _ "vsubst_pty (TyFreeVar a) i \<phi>"] prems Hwf_pty by auto 
      ultimately
      show A3: "set (free_db_ty \<tau>n) \<subseteq> {0, Suc i}" using vsubst_ty_free_db_ty by auto
    next
      have "set (free_db_pty (vsubst_pty (TyFreeVar a) i \<phi>)) = {}"
        using poly_wf_no_free_db_pty Hwf_pty by auto
      moreover
      have "set (free_db_pty (vsubst_pty (TyFreeVar a) i \<phi>)) = set (free_db_pty \<phi>) - {i}"
        using vsubst_ty_free_db_pty by auto
      ultimately
      show A2: "set (free_db_pty \<phi>) \<subseteq> {i}" by auto
    next
      show "a \<notin> set (free_ty \<tau>)" and "a \<notin> set (free_pty \<phi>)" and  "a \<notin> set (free_ty \<tau>n)"
        using Hfree Hr_free "1.prems" TmTapp by auto
    next
      show "set (free_db_pty \<phi>s) = {}" using prems poly_wf_no_free_db_pty by auto
    next
      show "vsubst_ty (vsubst_pty (TyFreeVar a) i \<phi>) 0 (vsubst_ty (TyFreeVar a) (Suc i) \<tau>n) =
            vsubst_ty (TyFreeVar a) i \<tau>"
        using Heqt by auto
    qed
    have "?\<Gamma>t;\<Delta>;?Mt;Ze\<union>filterScFreeX X\<union>Z\<tau> \<turnstile> ?et\<langle>?\<phi>t\<rangle> : vsubst_ty ?\<phi>t 0 (vsubst_ty \<phi>s (Suc i) \<tau>n)"
    proof rule
      have "?\<Gamma>t;\<Delta>;?Mt;Ze \<turnstile> vsubst_tmty \<phi>s i e : vsubst_ty \<phi>s i \<tau>n'" proof (rule IH)
        show "size e < size e_y" using TmTapp by auto
      next
        show "vsubst_ty_env \<phi>s i \<Gamma>;\<Delta> \<Turnstile> OK" 
          and "vsubst_ty_env \<phi>s i M;\<Delta> \<Turnstile> OK"
          and "\<Delta> \<Turnstile>' \<phi>s" (*and "Zs \<Turnstile> \<phi>s"*)
          and "a \<notin> free_ty_env \<Gamma>"
          and "a \<notin> Environ.dom \<Delta>"
          by assumption
(*      next
        show "\<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> \<rho> \<notin> Environ.dom Zs" using "1.prems" TmTapp by auto
*)
      next
        note Htyp
        also note Hn
        finally
        show "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Ze \<turnstile> vsubst_tmty (TyFreeVar a) i e : vsubst_ty (TyFreeVar a) i \<tau>n'"
          by auto
      next
        show "a \<notin> set (free_tmty e)" using "1.prems" TmTapp by auto
      next
        show "a \<notin> set (free_ty \<tau>n')" using Hfree by auto
      qed
      thus "?\<Gamma>t;\<Delta>;?Mt;Ze \<turnstile> vsubst_tmty \<phi>s i e : Qual q ScBot (\<forall>.vsubst_ty \<phi>s (Suc i) \<tau>n^X)" 
        using Hn2 by auto
    next
      have H_f: "\<not> a free in pretype \<phi>" using "1.prems" TmTapp by auto
      obtain \<tau>foo where Hfoo: "\<tau>foo = Qual q \<pi> \<phi>s" by auto
      hence "\<Delta> \<Turnstile>' \<tau>foo" using "1.prems" by auto
      hence "\<Delta>,(a:\<star>) \<Turnstile>' \<tau>foo" using poly_tenv_cons by auto
      hence "\<Delta>,(a:\<star>) \<Turnstile>' \<phi>s" using Hfoo by auto
      hence H: "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_pty \<phi>s i \<phi>" using Hwf_pty poly_subst_var_pretype_ok by auto 
      thus "\<Delta> \<Turnstile>' vsubst_pty \<phi>s i \<phi>" proof (cases "a free in pretype \<phi>s")
        case False 
        hence "\<not> a free in pretype (vsubst_pty \<phi>s i \<phi>)" using not_free_subst_pty H_f by auto
        thus ?thesis using H poly_wf_pty_strength by auto
      next
        case True
        hence "\<Delta> defines a" using poly_wf_free_defined assms by auto
        hence "\<Delta>,(a:\<star>) = \<Delta>" proof (auto simp add: dom_def)
          fix b
          assume "(a, b) \<in> \<Delta>"
          thus "a\<rhd>\<star> \<in> \<Delta>" by (cases b, auto)
        qed
        thus ?thesis using H by auto
      qed
    next
      show "minimalValidScope (scopeT (vsubst_ty \<phi>s (Suc i) \<tau>n)) Z\<tau>" using Hmin Hr
      by (cases \<tau>n, auto)
    next
(*
      have "\<forall>\<rho>. \<rho> is free dollar sc in e_y \<longrightarrow> \<rho> \<notin> Environ.dom Zs" by assumption
      hence Hdollar_1s: "\<forall>\<rho>. \<rho> is free dollar sc in ?er \<longrightarrow> \<rho> \<notin> Environ.dom Zs" 
        using free_dollars_vsubst_tmty TmTapp by auto
*)
      have "free_let_dollar_scope ?er = free_let_dollar_scope ?et"
        using free_dollars_vsubst_tmty by auto
      thus "\<forall>\<rho>. \<rho> is free dollar sc in ?et \<longrightarrow> \<not> (filterScFreeX X \<union> Z\<tau>) defines \<rho>"
        using Hdollar by (auto simp add: dom_def)
    next
      have "\<forall>j \<rho>. vsubst_sc_pty (ScFree \<rho>) j \<phi>s = \<phi>s" by assumption 
      thus "\<forall>j \<rho>. vsubst_sc_pty (ScFree \<rho>) j (vsubst_pty \<phi>s i \<phi>) = vsubst_pty \<phi>s i \<phi>" 
        using Hnew subst_pty_ty_complicated by auto
    qed
    thus ?thesis using H\<tau>_eq HZ_eq TmTapp by auto 
  next;

    (* type abstraction case *)
    case (TmProd q \<pi> e)
    let ?ex = "vsubst_tmty (TyFreeVar a) (Suc i) e"
    let ?et = "vsubst_tmty \<phi>s (Suc i) e"
    have Ht: "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> \<Lambda>^q^\<pi>[*]. vsubst_tmty (TyFreeVar a) (Suc i) e :
                            vsubst_ty (TyFreeVar a) i \<tau>"
      using prems by auto
    then obtain X \<tau>b L
      where H\<tau>': "vsubst_ty (TyFreeVar a) i \<tau> = Qual q \<pi> (\<forall>.\<tau>b^X)"
      and Hvalue: "value ?ex"
      and Hqual: "qualG ?\<Gamma>x \<preceq> q"
      and Htyp: "\<forall>c. c \<notin> L \<longrightarrow> ?\<Gamma>x;\<Delta>,(a:\<star>),(c:\<star>);?Mx;filterScFreeX X \<turnstile>
               vsubst_tmty (TyFreeVar c) 0 ?ex : vsubst_ty (TyFreeVar c) 0 \<tau>b"
      and "minimalValidScope \<pi> Z"  
      and Hfin: "finite L"
      using typing_poly_elim[of "?\<Gamma>x" "\<Delta>,(a:\<star>)" "?Mx" Z q \<pi> "?ex" "vsubst_ty (TyFreeVar a) i \<tau>"]
      by force
    obtain \<tau>r where H\<tau>: "\<tau> = Qual q \<pi> (\<forall>.\<tau>r^X)" using H\<tau>' proof (cases \<tau>, auto)
      fix \<phi>
      assume "\<And>\<tau>r. \<phi> = \<forall>.\<tau>r^X \<Longrightarrow> thesis"
        and "vsubst_pty (TyFreeVar a) i \<phi> = \<forall>.\<tau>b^X"
      thus thesis by (cases \<phi>, auto simp add: trivial_case)
    qed
    let ?Mi = "convert_ty_env (Suc i) i M"
    let ?Mxi = "vsubst_ty_env (TyFreeVar a) (Suc i) ?Mi"
    let ?\<Gamma>i = "convert_ty_env (Suc i) i \<Gamma>"
    let ?\<Gamma>xi = "vsubst_ty_env (TyFreeVar a) (Suc i) ?\<Gamma>i"
    let ?Mti = "vsubst_ty_env \<phi>s (Suc i) ?Mi"
    let ?\<Gamma>ti = "vsubst_ty_env \<phi>s(Suc i) ?\<Gamma>i"
    have Htyp: "\<forall>c. c \<notin> L \<longrightarrow> ?\<Gamma>x;\<Delta>,(a:\<star>),(c:\<star>);?Mx;filterScFreeX X \<turnstile>
               vsubst_tmty (TyFreeVar c) 0 ?ex :
               vsubst_ty (TyFreeVar c) 0 (vsubst_ty (TyFreeVar a) (Suc i) \<tau>r)" using Htyp H\<tau> H\<tau>' by auto
    have "?\<Gamma>t;\<Delta>;?Mt;Z \<turnstile> \<Lambda>^q^\<pi>[*]. ?et : Qual q \<pi> (\<forall>.vsubst_ty \<phi>s (Suc i) \<tau>r^X)" 
    proof rule
      show "minimalValidScope \<pi> Z" by assumption
    next
      show "?\<Gamma>t;\<Delta> \<Turnstile> OK" and "?Mt;\<Delta> \<Turnstile> OK" by assumption
    next
      show "qualG ?\<Gamma>t \<preceq> q" using Hqual qualG_vsubst_left by auto
    next
      show "value vsubst_tmty \<phi>s (Suc i) e" using Hvalue by (induct e arbitrary: i, auto)
    next
      have Hok1: "?\<Gamma>x;\<Delta>,(a:\<star>) \<Turnstile> OK" and Hok2: "?Mx;\<Delta>,(a:\<star>) \<Turnstile> OK" using prems by auto
      have "\<Delta> \<Turnstile>' TY" using prems poly_env_wf_env_ty by auto
      hence "\<Delta>,(a:\<star>) \<Turnstile>' TyFreeVar a" by (auto simp add: wf_env_ty_def)
      hence Hgisx: "?\<Gamma>x = ?\<Gamma>xi" and Hmisx: "?Mx = ?Mxi" using subst_convert_env Hok1 Hok2 by auto
      have "\<Delta> \<Turnstile>' \<phi>s" by assumption
      hence "\<Delta>,(a:\<star>) \<Turnstile>' \<phi>s" using poly_tenv_cons_pty by auto
      hence Hgist: "?\<Gamma>t = ?\<Gamma>ti" and Hmist: "?Mt = ?Mti" using subst_convert_env Hok1 Hok2 by auto
      { 
        fix c
        assume Hnin: "c \<notin> L\<union>{a}" 
        have "?\<Gamma>ti;\<Delta>,(c:\<star>);?Mti;filterScFreeX X \<turnstile> vsubst_tmty \<phi>s (Suc i)
              (freshen_tmty c e) : vsubst_ty \<phi>s (Suc i) (freshen_ty c \<tau>r)" 
        proof (rule IH[of "freshen_tmty c e" "Suc i" "?\<Gamma>i" "\<Delta>,(c:\<star>)" "?Mi" "filterScFreeX X"])
          show "size (freshen_tmty c e) < size e_y" using prems by auto
        next
          show "?\<Gamma>ti;\<Delta>,(c:\<star>) \<Turnstile> OK" using Hgist prems poly_env_ty_cons by auto
        next
          show "?Mti;\<Delta>,(c:\<star>) \<Turnstile> OK" using Hmist prems poly_env_ty_cons by auto
        next
          show "\<Delta>,(c:\<star>) \<Turnstile>' \<phi>s" using prems poly_tenv_cons_pty by auto
(*        next
          show "Zs \<Turnstile> \<phi>s" by assumption
        next
          have "\<forall>\<rho>. \<rho> is free dollar sc in e_y \<longrightarrow> \<rho> \<notin> Environ.dom Zs" by assumption
          thus Hdollar_1s: "\<forall>\<rho>. \<rho> is free dollar sc in (freshen_tmty c e) \<longrightarrow> \<rho> \<notin> Environ.dom Zs" 
            using free_dollars_vsubst_tmty TmProd by auto
*)
        next
          show "a \<notin> free_ty_env (convert_ty_env (Suc i) i \<Gamma>)" using prems not_free_convert_env
            by auto
        next            
          show "a \<notin> free_ty_env (convert_ty_env (Suc i) i M)" using prems not_free_convert_env
            by auto
        next
          show "a \<notin> Environ.dom (\<Delta>,(c:\<star>))" using prems by (auto simp add: dom_def)
        next
          show "\<not> a free in term (freshen_tmty c e)" using not_free_subst_tmty prems by auto 
        next
          show "\<not> a free in type (freshen_ty c \<tau>r)" using not_free_subst_ty prems by auto
        next
          have "?\<Gamma>x;\<Delta>,(c:\<star>),(a:\<star>);?Mx;filterScFreeX X \<turnstile> 
                   freshen_tmty c (vsubst_tmty (TyFreeVar a) (Suc i) e) :
                   freshen_ty c (vsubst_ty (TyFreeVar a) (Suc i) \<tau>r)" 
            using Htyp Hnin by (auto simp add: env_permute)
          also have "freshen_tmty c (vsubst_tmty (TyFreeVar a) (Suc i) e) = 
                   vsubst_tmty (TyFreeVar a) (Suc i) (freshen_tmty c e)" 
            using subst_subst_tyty_swap[of 0 "Suc i"] by auto
          also have "freshen_ty c (vsubst_ty (TyFreeVar a) (Suc i) \<tau>r) =
                   vsubst_ty (TyFreeVar a) (Suc i) (freshen_ty c \<tau>r)" 
            using subst_subst_ty_swap[of 0 "Suc i"] by auto
          also note Hgisx
          also note Hmisx
          finally
          show "?\<Gamma>xi;\<Delta>,(c:\<star>),(a:\<star>);?Mxi;filterScFreeX X \<turnstile> 
                   vsubst_tmty (TyFreeVar a) (Suc i) (freshen_tmty c e) :
                   vsubst_ty (TyFreeVar a) (Suc i) (freshen_ty c \<tau>r)" by auto
        qed
        also
        have H\<phi>s_k: "\<forall>k. vsubst_pty (TyFreeVar c) k \<phi>s = \<phi>s" 
          using poly_no_free_db_vars_same_subst_pty poly_wf_no_free_db_pty prems by auto
        hence "vsubst_tmty \<phi>s (Suc i) (freshen_tmty c e) = freshen_tmty c (vsubst_tmty \<phi>s (Suc i) e)"
          using subst_subst_tyty_swap[of "Suc i" 0 c] by auto
        also
        have "vsubst_ty \<phi>s (Suc i) (freshen_ty c \<tau>r) = freshen_ty c (vsubst_ty \<phi>s (Suc i) \<tau>r)"
          using subst_subst_ty_swap H\<phi>s_k by auto
        finally
        have "?\<Gamma>t; \<Delta>,(c:\<star>);?Mt;filterScFreeX X \<turnstile> freshen_tmty c (vsubst_tmty \<phi>s (Suc i) e) : 
                                                     freshen_ty c (vsubst_ty \<phi>s (Suc i) \<tau>r)"
          using Hgist Hmist
          by auto
      }
      thus "\<forall>c. c \<notin> L\<union>{a} \<longrightarrow> ?\<Gamma>t;\<Delta>,(c:\<star>);?Mt;filterScFreeX X \<turnstile> freshen_tmty c (vsubst_tmty \<phi>s (Suc i) e) : 
                                                              freshen_ty c (vsubst_ty \<phi>s (Suc i) \<tau>r)" 
        by auto
    next
      show "finite (L\<union>{a})" using prems by auto
    qed
    thus ?thesis using TmProd H\<tau> by auto
  next;

    (* deref case *)
    case (TmDeref er)
    let ?ers = "vsubst_tmty (TyFreeVar a) i er"
    let ?er\<phi> = "vsubst_tmty \<phi>s i er"
    have "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> TmDeref ?ers : vsubst_ty (TyFreeVar a) i \<tau>"
      using "1.prems" TmDeref by auto
    then obtain \<pi>r Ze Z\<tau>
      where Ht: "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Ze \<turnstile> ?ers : vsubst_ty (TyFreeVar a) i (Qual qU \<pi>r (TyRef \<tau>))"
      and "qualT (vsubst_ty (TyFreeVar a) i \<tau>) \<preceq> qU" 
      and Hmin: "minimalValidScope (scopeT (vsubst_ty (TyFreeVar a) i \<tau>)) Z\<tau>"
      and HZ_eq: "Z = Ze \<union> Z\<tau>"
      and Hdollar: "\<forall>\<rho>. \<rho> is free dollar sc in ?ers \<longrightarrow> \<not> Z\<tau> defines \<rho>"
      by auto  
    hence Hqual: "qualT \<tau> \<preceq> qU" by (cases \<tau>, auto)
    have Ht1: "(vsubst_ty_env \<phi>s i \<Gamma>);\<Delta>;?Mt;Ze \<turnstile> ?er\<phi> : vsubst_ty \<phi>s i (Qual qU \<pi>r (TyRef \<tau>))"
    proof (rule IH[of er i \<Gamma> \<Delta> M Ze "Qual qU \<pi>r (TyRef \<tau>)"])
      show "size er < size e_y" using TmDeref by auto
    next
      show "vsubst_ty_env \<phi>s i \<Gamma>;\<Delta> \<Turnstile> OK" and "?Mt;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<phi>s" (*and "Zs \<Turnstile> \<phi>s" *)
        by assumption
    next
      show "vsubst_ty_env (TyFreeVar a) i \<Gamma>;\<Delta>,(a:\<star>);?Mx;Ze \<turnstile> 
            ?ers : vsubst_ty (TyFreeVar a) i (Qual qU \<pi>r (TyRef \<tau>))" 
        using Ht by auto
    next
      show "a \<notin> set (free_tmty er)" and "a \<notin> set (free_ty (Qual qU \<pi>r (TyRef \<tau>)))"
        and "a \<notin> free_ty_env \<Gamma>" and "a \<notin> Environ.dom \<Delta>" 
        using "1.prems" TmDeref by auto
(*    next
      show "\<forall>\<rho>. \<rho> is free dollar sc in er \<longrightarrow> \<rho> \<notin> Environ.dom Zs"
        using TmDeref "1.prems" by (auto simp add: dom_def)
*)
    qed
    have Hqual: "qualT (vsubst_ty \<phi>s i \<tau>) \<preceq> qU" using Hqual by (cases \<tau>, auto)
    have Hmin: "minimalValidScope (scopeT (vsubst_ty \<phi>s i \<tau>)) Z\<tau>" using Hmin by (cases \<tau>, auto)
(*
    have Hdollar_y: "\<forall>\<rho>. \<rho> is free dollar sc in e_y \<longrightarrow> \<rho> \<notin> Environ.dom Zs" by assumption
    hence Hdollar_1s: "\<forall>\<rho>. \<rho> is free dollar sc in ?er\<phi> \<longrightarrow> \<rho> \<notin> Environ.dom Zs" 
      using free_dollars_vsubst_tmty TmDeref by auto
*)
    have "free_let_dollar_scope ?er\<phi> = free_let_dollar_scope ?ers"
      using free_dollars_vsubst_tmty by auto
    hence Hdollar: "\<forall>\<rho>. \<rho> is free dollar sc in ?er\<phi> \<longrightarrow> \<not> Z\<tau> defines \<rho>"
      using Hdollar by (auto simp add: dom_def)
    have Htyp: "vsubst_ty_env \<phi>s i \<Gamma>;\<Delta>;?Mt;Ze\<union>Z\<tau> \<turnstile> TmDeref ?er\<phi> : vsubst_ty \<phi>s i \<tau>"
      using Ht1 Hqual Hmin Hdollar by auto
    show ?thesis using TmDeref Htyp HZ_eq by auto
  next;

    (* loc case *)
    case (TmLoc q \<pi> l)
    hence Ht: "vsubst_ty_env (TyFreeVar a) i \<Gamma>;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> 
                 TmLoc q \<pi> l : vsubst_ty (TyFreeVar a) i \<tau>"
      using prems by auto
    then obtain \<tau>r
      where H\<tau>r: "vsubst_ty (TyFreeVar a) i \<tau> = Qual q \<pi> (TyRef \<tau>r)"
      and "?\<Gamma>x;\<Delta>,(a:\<star>) \<Turnstile> OK" and Hqual: "qualG ?\<Gamma>x \<preceq> qU" and "?Mx;\<Delta>,(a:\<star>) \<Turnstile> OK" 
      and Hin: "(l\<rhd>\<tau>r) \<in> ?Mx" and Hmin: "minimalValidScope \<pi> Z"
      by auto
    obtain \<phi>n where "\<tau> = Qual q \<pi> \<phi>n" using H\<tau>r by (cases \<tau>, auto)
    then obtain \<tau>n where Hn: "\<tau> = Qual q \<pi> (TyRef \<tau>n)" 
      using H\<tau>r by (cases \<phi>n, auto simp add: trivial_case)
    hence Hx': "?Mx = vsubst_ty_env (TyFreeVar a) i (M,(l:\<tau>n))"
      using Ht by (auto simp add: vsubst_ty_env_def)
    moreover
    have "\<not> a free in type \<tau>n" and "\<not> a free in environ M" using "1.prems" Hn by auto
    hence "\<not> a free in environ (M,(l:\<tau>n))" by (auto simp add: free_ty_env_def)
    moreover
    hence "\<not> a free in environ M" using prems by auto
    ultimately
    have "M = M,(l:\<tau>n)" using subst_subst_equal_env[of a i M "M,(l:\<tau>n)"] by auto
    hence Heq: "?Mt = vsubst_ty_env \<phi>s i M,(l:vsubst_ty \<phi>s i \<tau>n)" by (auto simp add: vsubst_ty_env_def)
    have "qualG ?\<Gamma>x \<preceq> qU" using Ht by auto
    hence Hqual: "qualG ?\<Gamma>t \<preceq> qU" using qualG_vsubst_left by auto
    have "?\<Gamma>t;\<Delta> \<Turnstile> OK" and "?Mt;\<Delta> \<Turnstile> OK" by assumption
    hence "?\<Gamma>t;\<Delta>;?Mt;Z \<turnstile> TmLoc q \<pi> l : Qual q \<pi> (TyRef (vsubst_ty \<phi>s i \<tau>n))" using Hqual Hmin Heq by auto
    thus ?thesis using Hn TmLoc by auto 
  next;

    (* letbang' case *)
    case (TmLetbangSc \<rho> e e1 e2)
    let ?es = "vsubst_tmty (TyFreeVar a) i e"
    let ?e1s = "vsubst_tmty (TyFreeVar a) i e1"
    let ?e2s = "vsubst_tmty (TyFreeVar a) i e2"
    let ?e1\<phi> = "vsubst_tmty \<phi>s i e1"
    let ?e2\<phi> = "vsubst_tmty \<phi>s i e2"
    have "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> let$ \<rho> ?es ?e1s in ?e2s : vsubst_ty (TyFreeVar a) i \<tau>" 
      using 1 TmLetbangSc by auto
    then obtain \<Gamma>1 \<Gamma>2 \<Gamma>x Z1 Z2 Zx x \<tau>x \<tau>1 Zf Lx Ly
      where H\<Gamma>x_eq: "?\<Gamma>x = \<Gamma>x \<union> \<Gamma>1 \<union> \<Gamma>2"
      and Hcompat1: "\<Gamma>1 \<sim> \<Gamma>2" and Hcompat2: "\<Gamma>x \<sim> \<Gamma>2" and Hcompat3: "\<Gamma>x \<sim> \<Gamma>1"
      and Hwf: "\<Gamma>x \<union> \<Gamma>1 \<union> \<Gamma>2;\<Delta>,(a:\<star>) \<Turnstile> OK"
      and HZ_eq: "Z = Z2 \<union> Z1 \<union> Zx"
      and Ht\<rho>: "\<not> (Z2\<union>Z1) defines \<rho>"
      and HZ1: "Z1 \<Turnstile> ScopeSet"
      and HZ2: "Z2 \<Turnstile> ScopeSet"
      and Hvalid: "validScope (scopeT \<tau>1) Z1"
      and Hbang: "isBangable \<tau>x"
      and Htyp: "\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Zx \<turnstile> TmFreeVar x : convT \<rho> \<tau>x"
      and Hxnotin: "\<not> (\<Gamma>1\<union>\<Gamma>2) defines x"
      and Htyp1: "\<Gamma>1,(x:convT \<rho> \<tau>x); \<Delta>,(a:\<star>); ?Mx; Zf \<turnstile> vsubst_tmty (TyFreeVar a) i e1 : \<tau>1"
      and Htyp2: "\<forall>x y. \<Gamma>2,(x:\<tau>x),(y:\<tau>1);\<Delta>,(a:\<star>) \<Turnstile> OK \<and> \<Gamma>2 \<sim> {x\<rhd>\<tau>x},(y:\<tau>1) \<and> {x\<rhd>\<tau>x} \<sim> {y\<rhd>\<tau>1} \<and> 
                  x \<notin> Lx \<and> y \<notin> Ly \<and> x \<noteq> y \<longrightarrow>
                  \<Gamma>2,(x:\<tau>x),(y:\<tau>1); \<Delta>,(a:\<star>);?Mx;Z2 \<turnstile> 
                  vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 (vsubst_tmty (TyFreeVar a) i e2)) :
                  vsubst_ty (TyFreeVar a) i \<tau>"
      and Hdollar1: "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tmty (TyFreeVar a) i e1) \<longrightarrow> \<not> (Zx\<union>Z2) defines \<rho>"
      and Hdollar2: "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tmty (TyFreeVar a) i e2) \<longrightarrow> \<not> (Zx\<union>Z1) defines \<rho>"
      and Hx_eq: "vsubst_tmty (TyFreeVar a) i e = TmFreeVar x"
      and HZf: "Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>)"
      and Hlx: "finite Lx"
      and Hly: "finite Ly"
      using typing_letsc_elim[of "?\<Gamma>x" "\<Delta>,(a:\<star>)" "?Mx" Z \<rho> ?es ?e1s ?e2s "vsubst_ty (TyFreeVar a) i \<tau>"]
      by auto
    have Hx_eq: "e = TmFreeVar x" using Hx_eq by (cases e, auto)
    have Hwfx: "\<Gamma>x;\<Delta>,(a:\<star>) \<Turnstile> OK" using Hwf wf_env_subset by auto
    then obtain \<Gamma>rx
      where Hrx: "\<Gamma>x = vsubst_ty_env (TyFreeVar a) i \<Gamma>rx"
      and Hrx_free: "\<not> a free in environ \<Gamma>rx"
      using parametrize_ty_env by blast
    have Hwf1: "\<Gamma>1;\<Delta>,(a:\<star>) \<Turnstile> OK" using Hwf wf_env_subset by auto
    then obtain \<Gamma>r1
      where Hr1: "\<Gamma>1 = vsubst_ty_env (TyFreeVar a) i \<Gamma>r1"
      and Hr1_free: "\<not> a free in environ \<Gamma>r1"
      using parametrize_ty_env by blast
    have Hwf2: "\<Gamma>2;\<Delta>,(a:\<star>) \<Turnstile> OK" using Hwf wf_env_subset by auto
    then obtain \<Gamma>r2
      where Hr2: "\<Gamma>2 = vsubst_ty_env (TyFreeVar a) i \<Gamma>r2"
      and Hr2_free: "\<not> a free in environ \<Gamma>r2"
      using parametrize_ty_env by blast
    have "\<not> a free in environ \<Gamma>" by assumption
    moreover
    have "\<not> a free in environ (\<Gamma>rx \<union> \<Gamma>r1)" using Hrx_free Hr1_free
      apply (auto simp add: free_ty_env_def) by force 
    hence "\<not> a free in environ (\<Gamma>rx \<union> \<Gamma>r1 \<union> \<Gamma>r2)" using Hr2_free
      apply (auto simp add: free_ty_env_def) by force 
    moreover
    have "vsubst_ty_env (TyFreeVar a) i \<Gamma> = vsubst_ty_env (TyFreeVar a) i (\<Gamma>rx \<union> \<Gamma>r1 \<union> \<Gamma>r2)"
      using Hrx Hr1 Hr2 H\<Gamma>x_eq by (auto simp add: vsubst_ty_env_def)
    ultimately
    have H\<Gamma>x_eq: "\<Gamma> = \<Gamma>rx \<union> \<Gamma>r1 \<union> \<Gamma>r2" using subst_subst_equal_env[of a i \<Gamma> "\<Gamma>rx\<union>\<Gamma>r1\<union>\<Gamma>r2"] by auto
    have "\<Delta>,(a:\<star>) \<Turnstile>' convT \<rho> \<tau>x" using Htyp poly_ty_wf_type by auto
    hence "\<Delta>,(a:\<star>) \<Turnstile>' \<tau>x" by (cases \<tau>x, auto)
    then obtain \<tau>rx
      where H\<tau>rx:  "\<tau>x = vsubst_ty (TyFreeVar a) i \<tau>rx" and H\<tau>rx_free: "\<not> a free in type \<tau>rx"
      using parametrize_ty by blast
    have "\<Delta>,(a:\<star>) \<Turnstile>' \<tau>1" using Htyp1 poly_ty_wf_type by auto
    then obtain \<tau>r1
      where H\<tau>r1:  "\<tau>1 = vsubst_ty (TyFreeVar a) i \<tau>r1" and H\<tau>r1_free: "\<not> a free in type \<tau>r1"
      using parametrize_ty by blast
    have "(vsubst_ty_env \<phi>s i \<Gamma>rx) \<union> (vsubst_ty_env \<phi>s i \<Gamma>r1) \<union> (vsubst_ty_env \<phi>s i \<Gamma>r2)
          ;\<Delta>;?Mt;Z2\<union>Z1\<union>Zx \<turnstile> let$ \<rho> (TmFreeVar x) ?e1\<phi> in ?e2\<phi> : vsubst_ty \<phi>s i \<tau>" 
    proof (rule t_let'[where Lx="Lx\<union>dom \<Gamma>2" and Ly="Ly\<union>dom \<Gamma>2"])
      have "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1 \<sim> vsubst_ty_env (TyFreeVar a) i \<Gamma>r2"
        using Hcompat1 Hr1 Hr2 by auto
      moreover
      have "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1 \<union> vsubst_ty_env (TyFreeVar a) i \<Gamma>r2;\<Delta>,(a:\<star>) \<Turnstile> OK"
        using Hwf Hr1 Hr2 wf_env_subset by blast
      ultimately
      show "vsubst_ty_env \<phi>s i \<Gamma>r1 \<sim> vsubst_ty_env \<phi>s i \<Gamma>r2"
        using env_subst_compat by auto
      next
      have "vsubst_ty_env (TyFreeVar a) i \<Gamma>rx \<sim> vsubst_ty_env (TyFreeVar a) i \<Gamma>r2"
        using Hcompat2 Hrx Hr2 by auto
      moreover
      have "vsubst_ty_env (TyFreeVar a) i \<Gamma>rx \<union> vsubst_ty_env (TyFreeVar a) i \<Gamma>r2;\<Delta>,(a:\<star>) \<Turnstile> OK"
        using Hwf Hrx Hr2 wf_env_subset by blast
      ultimately
      show "vsubst_ty_env \<phi>s i \<Gamma>rx \<sim> vsubst_ty_env \<phi>s i \<Gamma>r2"
        using env_subst_compat by auto
    next
      have "vsubst_ty_env (TyFreeVar a) i \<Gamma>rx \<sim> vsubst_ty_env (TyFreeVar a) i \<Gamma>r1"
        using Hcompat3 Hrx Hr1 by auto
      moreover
      have "vsubst_ty_env (TyFreeVar a) i \<Gamma>rx \<union> vsubst_ty_env (TyFreeVar a) i \<Gamma>r1;\<Delta>,(a:\<star>) \<Turnstile> OK"
        using Hwf Hrx Hr1 wf_env_subset by blast
      ultimately
      show "vsubst_ty_env \<phi>s i \<Gamma>rx \<sim> vsubst_ty_env \<phi>s i \<Gamma>r1"
        using env_subst_compat by auto
    next
        have "?\<Gamma>t;\<Delta> \<Turnstile> OK" by assumption
        also
        have "?\<Gamma>t = (vsubst_ty_env \<phi>s i \<Gamma>rx) \<union> (vsubst_ty_env \<phi>s i \<Gamma>r1) \<union> (vsubst_ty_env \<phi>s i \<Gamma>r2)"
           using H\<Gamma>x_eq by (auto simp add: vsubst_ty_env_def)
        finally
        show "vsubst_ty_env \<phi>s i \<Gamma>rx \<union> vsubst_ty_env \<phi>s i \<Gamma>r1 \<union> vsubst_ty_env \<phi>s i \<Gamma>r2;\<Delta> \<Turnstile> OK"
          by auto
    next
      show "\<rho> \<notin> dom (Z2 \<union> Z1)" and "Z1 \<Turnstile> ScopeSet" and "Z2 \<Turnstile> ScopeSet" and "Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>)"
        by assumption
    next
      have "validScope (scopeT \<tau>r1) Z1" using Hvalid H\<tau>r1 by (cases \<tau>r1, auto)
      thus "validScope (scopeT (vsubst_ty \<phi>s i \<tau>r1)) Z1" by (cases \<tau>r1, auto)
    next      
      show "isBangable (vsubst_ty \<phi>s i \<tau>rx)" using Hbang H\<tau>rx proof (cases \<tau>rx, auto)
        fix q \<pi> \<phi>
        assume "isBangable (Qual q \<pi> (vsubst_pty (TyFreeVar a) i \<phi>))"
        thus "isBangable (Qual q \<pi> (vsubst_pty \<phi>s i \<phi>))" by (cases \<phi>, auto simp add: split_if)
      qed
    next
      have "vsubst_ty_env \<phi>s i \<Gamma>rx;\<Delta>;?Mt;Zx \<turnstile> 
             vsubst_tmty \<phi>s i (TmFreeVar x) : vsubst_ty \<phi>s i (convT \<rho> \<tau>rx)" 
      proof (rule IH)
        show "size (TmFreeVar x) < size e_y" using TmLetbangSc by auto
      next
        have "?\<Gamma>t;\<Delta> \<Turnstile> OK" by assumption
        thus "vsubst_ty_env \<phi>s i \<Gamma>rx;\<Delta> \<Turnstile> OK" 
          using H\<Gamma>x_eq wf_env_subset[of "?\<Gamma>t" \<Delta> "vsubst_ty_env \<phi>s i \<Gamma>rx"]
          by (auto simp add: vsubst_ty_env_def)
      next
        show "vsubst_ty_env \<phi>s i M;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<phi>s" 
          and "a \<notin> free_ty_env \<Gamma>rx" and "a \<notin> Environ.dom \<Delta>" 
          by assumption
      next
(*        show "\<forall>\<rho>. \<rho> is free dollar sc in TmFreeVar x \<longrightarrow> \<rho> \<notin> Environ.dom Zs" by auto
      next*)
        note Htyp
        also note Hrx
        also note H\<tau>rx
        also have "convT \<rho> (vsubst_ty (TyFreeVar a) i \<tau>rx) = vsubst_ty (TyFreeVar a) i (convT \<rho> \<tau>rx)"
          by (cases \<tau>rx, auto)
        finally
        show "vsubst_ty_env (TyFreeVar a) i \<Gamma>rx;\<Delta>,(a:\<star>);?Mx;Zx \<turnstile> 
                vsubst_tmty (TyFreeVar a) i (TmFreeVar x) : vsubst_ty (TyFreeVar a) i (convT \<rho> \<tau>rx)"
          by auto
      next
        show "a \<notin> set (free_tmty (TmFreeVar x))" by auto
      next
        show "a \<notin> set (free_ty (convT \<rho> \<tau>rx))" using H\<tau>rx_free by (cases \<tau>rx, auto)
      qed
      also have "vsubst_ty \<phi>s i (convT \<rho> \<tau>rx) = convT \<rho> (vsubst_ty \<phi>s i \<tau>rx)" 
        by (cases \<tau>rx, auto)
      finally
      show "vsubst_ty_env \<phi>s i \<Gamma>rx;\<Delta>;?Mt;Zx \<turnstile> TmFreeVar x : convT \<rho> (vsubst_ty \<phi>s i \<tau>rx)"
        by auto
    next
      show "x \<notin> dom (vsubst_ty_env \<phi>s i \<Gamma>r1 \<union> vsubst_ty_env \<phi>s i \<Gamma>r2)"
        using Hxnotin Hr1 Hr2 by (auto simp add: dom_def vsubst_ty_env_def)
    next
      show "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tmty \<phi>s i e1 \<longrightarrow> \<rho> \<notin> Environ.dom (Zx \<union> Z2)"
        using Hdollar1 poly_free_dollars_tmty_equal by auto
    next
      show "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tmty \<phi>s i e2 \<longrightarrow> \<rho> \<notin> Environ.dom (Zx \<union> Z1)"
        using Hdollar2 poly_free_dollars_tmty_equal by auto
    next
      have "vsubst_ty_env \<phi>s i (\<Gamma>r1,(x:convT \<rho> \<tau>rx));\<Delta>;?Mt;Zf \<turnstile>
        vsubst_tmty \<phi>s i e1 : vsubst_ty \<phi>s i \<tau>r1" 
      proof (rule IH)
        show "size e1 < size e_y" using TmLetbangSc by auto
      next
        have "?\<Gamma>t;\<Delta> \<Turnstile> OK" by assumption
        moreover
        have "vsubst_ty_env \<phi>s i \<Gamma>r1 \<subseteq> ?\<Gamma>t" using H\<Gamma>x_eq by (auto simp add: vsubst_ty_env_def)
        ultimately
        have Hok: "vsubst_ty_env \<phi>s i \<Gamma>r1;\<Delta> \<Turnstile> OK" using wf_env_subset by auto
        have "\<Delta>,(a:\<star>) \<Turnstile>' convT \<rho> \<tau>x" using Htyp poly_ty_wf_type by auto
        hence "\<Delta>,(a:\<star>) \<Turnstile>' convT \<rho> (vsubst_ty (TyFreeVar a) i \<tau>rx)" using H\<tau>rx by auto
        hence "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty (TyFreeVar a) i \<tau>rx" by (cases \<tau>rx, auto)
        moreover
        have "\<Delta> \<Turnstile>' \<phi>s" by assumption
        hence "\<Delta>,(a:\<star>) \<Turnstile>' \<phi>s" using poly_tenv_cons_pty by auto
        ultimately
        have "\<Delta>,(a:\<star>) \<Turnstile>' vsubst_ty \<phi>s i \<tau>rx" using poly_subst_var_type_ok by auto
        moreover
        have "\<not> a free in pretype \<phi>s" proof cases 
          assume H: "a free in pretype \<phi>s"
          have "\<Delta> \<Turnstile>' \<phi>s" and Hf: "\<not> \<Delta> defines a" by assumption
          hence "\<Delta> defines a" using poly_wf_free_defined H by auto
          thus ?thesis using Hf by auto
        qed
        hence "\<not> a free in type (vsubst_ty \<phi>s i \<tau>rx)" using H\<tau>rx_free not_free_subst_ty by auto
        ultimately
        have "\<Delta> \<Turnstile>' vsubst_ty \<phi>s i \<tau>rx" using poly_wf_ty_strength by auto
        hence Hwf_t: "\<Delta> \<Turnstile>' vsubst_ty \<phi>s i (convT \<rho> \<tau>rx)" by (cases \<tau>rx, auto)
        have "\<not> (\<Gamma>1\<union>\<Gamma>2) defines x" by assumption
        hence "\<not> \<Gamma>1 defines x" by (auto simp add: dom_def)
        hence "\<not> \<Gamma>r1 defines x" using Hr1 proof (cases, auto simp add: dom_def)
          fix b
          assume Hf: "\<forall>b. (x, b) \<notin> vsubst_ty_env (TyFreeVar a) i \<Gamma>r1"
            and "(x, b) \<in> \<Gamma>r1"
          then obtain \<tau>r where Hin: "x\<rhd>\<tau>r \<in> \<Gamma>r1" by (cases b, auto)
          hence "x\<rhd>vsubst_ty (TyFreeVar a) i \<tau>r \<in> vsubst_ty_env (TyFreeVar a) i \<Gamma>r1"
            by (auto simp add: vsubst_ty_env_def)
          thus False using Hf by auto
        qed
        hence Hfree: "\<not> (vsubst_ty_env \<phi>s i \<Gamma>r1) defines x"
          by (auto simp add: vsubst_ty_env_def dom_def)
        have "vsubst_ty_env \<phi>s i \<Gamma>r1,(x:vsubst_ty \<phi>s i (convT \<rho> \<tau>rx));\<Delta> \<Turnstile> OK"
          using Hok Hwf_t wf_env_cons Hfree by auto
        also have "vsubst_ty_env \<phi>s i \<Gamma>r1,(x:vsubst_ty \<phi>s i (convT \<rho> \<tau>rx)) =
          vsubst_ty_env \<phi>s i (\<Gamma>r1,(x:convT \<rho> \<tau>rx))" by (auto simp add: vsubst_ty_env_def)
        finally
        show "vsubst_ty_env \<phi>s i (\<Gamma>r1,(x:convT \<rho> \<tau>rx));\<Delta> \<Turnstile> OK" by auto 
      next
        show "vsubst_ty_env \<phi>s i M;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<phi>s" and "a \<notin> dom \<Delta>"
          by assumption
      next
(*        show "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom Zs" using "1.prems" TmLetbangSc by auto
      next*)
        note Htyp1
        also note Hr1
        also note H\<tau>r1
        also note H\<tau>rx
        also have "convT \<rho> (vsubst_ty (TyFreeVar a) i \<tau>rx) = vsubst_ty (TyFreeVar a) i (convT \<rho> \<tau>rx)"
          by (cases \<tau>rx, auto)
        also have "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1,(x:vsubst_ty (TyFreeVar a) i (convT \<rho> \<tau>rx)) =
          vsubst_ty_env (TyFreeVar a) i (\<Gamma>r1,(x:convT \<rho> \<tau>rx))"
          by (auto simp add: vsubst_ty_env_def)
        finally
        show "vsubst_ty_env (TyFreeVar a) i (\<Gamma>r1,(x:convT \<rho> \<tau>rx));\<Delta>,(a:\<star>);?Mx;Zf \<turnstile> 
          vsubst_tmty (TyFreeVar a) i e1 : vsubst_ty (TyFreeVar a) i \<tau>r1" 
          by auto
      next
        show "a \<notin> set (free_tmty e1)" using "1.prems" TmLetbangSc by auto
      next
        show "a \<notin> set (free_ty \<tau>r1)" using H\<tau>r1_free by auto
      next
        have "\<not> a free in type (convT \<rho> \<tau>rx)" using H\<tau>rx_free by (cases \<tau>rx, auto)
        thus "a \<notin> free_ty_env (\<Gamma>r1,(x:convT \<rho> \<tau>rx))"
          using Hr1_free by (auto simp add: free_ty_env_def)
      qed
      also
      have "vsubst_ty_env \<phi>s i (\<Gamma>r1,(x:convT \<rho> \<tau>rx)) =
        vsubst_ty_env \<phi>s i \<Gamma>r1,(x:vsubst_ty \<phi>s i (convT \<rho>  \<tau>rx))" by (auto simp add: vsubst_ty_env_def)
      also 
      have "vsubst_ty \<phi>s i (convT \<rho>  \<tau>rx) = convT \<rho> (vsubst_ty \<phi>s i \<tau>rx)" 
        by (cases \<tau>rx, auto)
      finally
      show "vsubst_ty_env \<phi>s i \<Gamma>r1,(x:convT \<rho> (vsubst_ty \<phi>s i \<tau>rx));\<Delta>;?Mt;Zf \<turnstile>
        vsubst_tmty \<phi>s i e1 : vsubst_ty \<phi>s i \<tau>r1" by auto
    next
      show "\<forall>x y. vsubst_ty_env \<phi>s i \<Gamma>r2,(x:vsubst_ty \<phi>s i \<tau>rx),(y:vsubst_ty \<phi>s i \<tau>r1);\<Delta> \<Turnstile> OK \<and>
        vsubst_ty_env \<phi>s i \<Gamma>r2 \<sim> {x\<rhd>vsubst_ty \<phi>s i \<tau>rx},(y:vsubst_ty \<phi>s i \<tau>r1) \<and> 
        {x\<rhd>vsubst_ty \<phi>s i \<tau>rx} \<sim> {y\<rhd>vsubst_ty \<phi>s i \<tau>r1} \<and> x \<notin> (Lx\<union>dom \<Gamma>2) \<and> y \<notin> (Ly\<union>dom \<Gamma>2) \<and> x \<noteq> y \<longrightarrow>
        vsubst_ty_env \<phi>s i \<Gamma>r2,(x:vsubst_ty \<phi>s i \<tau>rx),(y:vsubst_ty \<phi>s i \<tau>r1);\<Delta>;?Mt;Z2 \<turnstile>
        vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x (vsubst_tmty \<phi>s i e2)) : vsubst_ty \<phi>s i \<tau>"
      proof auto
        fix x y
        assume Hwf_\<phi>: "vsubst_ty_env \<phi>s i \<Gamma>r2,(x:vsubst_ty \<phi>s i \<tau>rx),(y:vsubst_ty \<phi>s i \<tau>r1);\<Delta> \<Turnstile> OK"
          and "vsubst_ty_env \<phi>s i \<Gamma>r2 \<sim> {x\<rhd>vsubst_ty \<phi>s i \<tau>rx},(y:vsubst_ty \<phi>s i \<tau>r1)"
          and "{x\<rhd>vsubst_ty \<phi>s i \<tau>rx} \<sim> {y\<rhd>vsubst_ty \<phi>s i \<tau>r1}"
          and Hx_lx: "x \<notin> Lx" and Hy_ly: "y \<notin> Ly"
          and Hx_fresh: "x \<notin> Environ.dom \<Gamma>2"
          and Hy_fresh: "y \<notin> Environ.dom \<Gamma>2"
          and Hxy_diff: "x \<noteq> y"
        have "vsubst_ty_env \<phi>s i (\<Gamma>r2,(x:\<tau>rx),(y:\<tau>r1));\<Delta>;?Mt;Z2 \<turnstile> 
          vsubst_tmty \<phi>s i (vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 e2)) : 
          vsubst_ty \<phi>s i \<tau>"
        proof (rule IH)
          show "size (vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 e2)) < size e_y"
            using TmLetbangSc by auto
        next
          show "vsubst_ty_env \<phi>s i M;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<phi>s" 
            and "a \<notin> free_ty_env M" and "a \<notin> Environ.dom \<Delta>"
            and "a \<notin> set (free_ty \<tau>)"
            by assumption
        next
          have "\<not> a free in term e2" using TmLetbangSc prems by auto
          thus "a \<notin> set (free_tmty (vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 e2)))"
            using free_tmty_subst_equal by auto
        next
          show "a \<notin> free_ty_env (\<Gamma>r2,(x:\<tau>rx),(y:\<tau>r1))"
            using Hr2_free H\<tau>rx_free H\<tau>r1_free by (auto simp add: free_ty_env_def)
        next
          note Hwf_\<phi>
          also have "vsubst_ty_env \<phi>s i \<Gamma>r2,(x:vsubst_ty \<phi>s i \<tau>rx),(y:vsubst_ty \<phi>s i \<tau>r1) =
            vsubst_ty_env \<phi>s i (\<Gamma>r2,(x:\<tau>rx),(y:\<tau>r1))"
            by (auto simp add: vsubst_ty_env_def)
          finally
          show "vsubst_ty_env \<phi>s i (\<Gamma>r2,(x:\<tau>rx),(y:\<tau>r1));\<Delta> \<Turnstile> OK" by auto 
        next
          have Hwf: "\<Gamma>2;\<Delta>,(a:\<star>) \<Turnstile> OK" using Hwf wf_env_subset by auto
          have "\<Delta>,(a:\<star>) \<Turnstile>' convT \<rho> \<tau>x" using prems poly_ty_wf_type by auto
          hence H\<tau>x: "\<Delta>,(a:\<star>) \<Turnstile>' \<tau>x" by (cases \<tau>x, auto)
          have H\<tau>1: "\<Delta>,(a:\<star>) \<Turnstile>' \<tau>1" using prems poly_ty_wf_type by auto
          have "\<Gamma>2,(x:\<tau>x);\<Delta>,(a:\<star>) \<Turnstile> OK" using wf_env_cons Hx_fresh Hwf H\<tau>x by auto
          hence "\<Gamma>2,(x:\<tau>x),(y:\<tau>1);\<Delta>,(a:\<star>) \<Turnstile> OK" 
            using wf_env_cons[of "\<Gamma>2,(x:\<tau>x)" "\<Delta>,(a:\<star>)" y] Hy_fresh Hxy_diff Hwf H\<tau>1
            by (auto simp add: dom_def)
          moreover
          have "\<Gamma>2 \<sim> {x\<rhd>\<tau>x},(y:\<tau>1) \<and> {x\<rhd>\<tau>x} \<sim> {y\<rhd>\<tau>1}" using Hy_fresh Hx_fresh Hxy_diff 
            by (auto simp add: dom_def env_compat_def)
          moreover
          have "x \<notin> Lx \<and> y \<notin> Ly \<and> x \<noteq> y" using Hx_lx Hy_ly Hxy_diff by auto
          ultimately
          have "\<Gamma>2,(x:\<tau>x),(y:\<tau>1);\<Delta>,(a:\<star>);?Mx;Z2 \<turnstile>
                   vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 (vsubst_tmty (TyFreeVar a) i e2)) 
                   : vsubst_ty (TyFreeVar a) i \<tau>" using Htyp2 by auto
          also note Hr2 
          also note H\<tau>rx
          also note H\<tau>r1
          also
          have "vsubst_ty_env (TyFreeVar a) i \<Gamma>r2,(
            x:vsubst_ty (TyFreeVar a) i \<tau>rx),(y:vsubst_ty (TyFreeVar a) i \<tau>r1) =
            vsubst_ty_env (TyFreeVar a) i (\<Gamma>r2,(x:\<tau>rx),(y:\<tau>r1))"
            by (auto simp add: vsubst_ty_env_def)
          also
          have "vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 (vsubst_tmty (TyFreeVar a) i e2)) =
            vsubst_tmty (TyFreeVar a) i (vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 e2))"
            using subst_subst_tmty_swap by auto
          finally
          show "vsubst_ty_env (TyFreeVar a) i (\<Gamma>r2,(x:\<tau>rx),(y:\<tau>r1));\<Delta>,(a:\<star>);?Mx;Z2 \<turnstile> 
                  vsubst_tmty (TyFreeVar a) i (vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 e2)) :
                  vsubst_ty (TyFreeVar a) i \<tau>" by auto
        qed
        also 
        have "vsubst_ty_env \<phi>s i (\<Gamma>r2,(x:\<tau>rx),(y:\<tau>r1)) =
          vsubst_ty_env \<phi>s i \<Gamma>r2,(x:vsubst_ty \<phi>s i \<tau>rx),(y:vsubst_ty \<phi>s i \<tau>r1)"
          by (auto simp add: vsubst_ty_env_def)
        also 
        have "vsubst_tmty \<phi>s i (vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 e2)) =
          vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 (vsubst_tmty \<phi>s i e2))"
          using subst_subst_tytm_swap by auto
        finally
        show "vsubst_ty_env \<phi>s i \<Gamma>r2,(x:vsubst_ty \<phi>s i \<tau>rx),(y:vsubst_ty \<phi>s i \<tau>r1);\<Delta>;
          vsubst_ty_env \<phi>s i M;Z2 \<turnstile> 
          vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 (vsubst_tmty \<phi>s i e2)) : 
          vsubst_ty \<phi>s i \<tau>"
          by auto
      qed
    next
      have "\<Gamma>2;\<Delta>,(a:\<star>) \<Turnstile> OK" using Hwf wf_env_subset by auto
      hence "finite (dom \<Gamma>2)" using dom_finite by (auto simp add: wf_env_def)
      thus "finite (Lx \<union> dom \<Gamma>2)" and "finite (Ly \<union> dom \<Gamma>2)" using Hlx Hly by auto
    qed
    also have "(vsubst_ty_env \<phi>s i \<Gamma>rx) \<union> (vsubst_ty_env \<phi>s i \<Gamma>r1) \<union> (vsubst_ty_env \<phi>s i \<Gamma>r2) = ?\<Gamma>t"
      using H\<Gamma>x_eq by (auto simp add: vsubst_ty_env_def)
    finally
    show ?thesis using TmLetbangSc Hx_eq HZ_eq by auto
  next;
      
    (* letbang case *)
      case (TmLetbang e e1 e2)
    let ?es = "vsubst_tmty (TyFreeVar a) i e"
    let ?e1s = "vsubst_tmty (TyFreeVar a) i e1"
    let ?e2s = "vsubst_tmty (TyFreeVar a) i e2"
    let ?e\<phi> = "vsubst_tmty \<phi>s i e"
    let ?e1\<phi> = "vsubst_tmty \<phi>s i e1"
    let ?e2\<phi> = "vsubst_tmty \<phi>s i e2"
    have "?\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Z \<turnstile> let! ?es ?e1s in ?e2s : vsubst_ty (TyFreeVar a) i \<tau>" 
      using 1 TmLetbang by auto
    then obtain \<Gamma>1 \<Gamma>2 \<Gamma>x Z1 Z2 Zx x \<tau>x \<tau>1 Lx Ly
      where H\<Gamma>x_eq: "?\<Gamma>x = \<Gamma>x \<union> \<Gamma>1 \<union> \<Gamma>2"
      and Hcompat1: "\<Gamma>1 \<sim> \<Gamma>2" and Hcompat2: "\<Gamma>x \<sim> \<Gamma>2" and Hcompat3: "\<Gamma>x \<sim> \<Gamma>1"
      and Hwf: "\<Gamma>x \<union> \<Gamma>1 \<union> \<Gamma>2;\<Delta>,(a:\<star>) \<Turnstile> OK"
      and HZ_eq: "Z = Z2 \<union> Z1 \<union> Zx"
      and HZ1: "Z1 \<Turnstile> ScopeSet"
      and HZ2: "Z2 \<Turnstile> ScopeSet"
      and Hvalid: "validScope (scopeT \<tau>1) Z1"
      and Hbang: "isBangable \<tau>x"
      and Htyp: "\<Gamma>x;\<Delta>,(a:\<star>);?Mx;Zx \<turnstile> ?es : \<tau>x"
      and Htyp1: "\<forall>x \<rho>. \<Gamma>1,(x:convT \<rho> \<tau>x);\<Delta>,(a:\<star>) \<Turnstile> OK \<and> \<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>x} \<and> x \<notin> Lx \<and>
                   \<not> \<rho> is free dollar sc in ?e1s \<and> \<not> (Z2\<union>Z1) defines \<rho> \<longrightarrow> 
                   ( \<Gamma>1,(x:convT \<rho> \<tau>x); \<Delta>,(a:\<star>); ?Mx; Z1 \<turnstile> freshen_tm x (freshen_sc \<rho> ?e1s) : \<tau>1
                     \<or>
                     \<Gamma>1,(x:convT \<rho> \<tau>x); \<Delta>,(a:\<star>); ?Mx; Z1,(\<rho>:\<section>) \<turnstile> freshen_tm x (freshen_sc \<rho> ?e1s) : \<tau>1)"
      and Htyp2: "\<forall>x y. \<Gamma>2,(x:\<tau>x),(y:\<tau>1);\<Delta>,(a:\<star>) \<Turnstile> OK \<and> \<Gamma>2 \<sim> {x\<rhd>\<tau>x},(y:\<tau>1) \<and> {x\<rhd>\<tau>x} \<sim> {y\<rhd>\<tau>1} 
                  \<and> x \<notin> Lx \<and> y \<notin> Ly \<and> x \<noteq> y \<longrightarrow>
                  \<Gamma>2,(x:\<tau>x),(y:\<tau>1); \<Delta>,(a:\<star>);?Mx;Z2 \<turnstile> 
                  vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 (vsubst_tmty (TyFreeVar a) i e2)) :
                  vsubst_ty (TyFreeVar a) i \<tau>"
      and HdollarX: "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tmty (TyFreeVar a) i e) \<longrightarrow> \<not> (Z1\<union>Z2) defines \<rho>"
      and Hdollar1: "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tmty (TyFreeVar a) i e1) \<longrightarrow> \<not> (Z2\<union>Zx) defines \<rho>"
      and Hdollar2: "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tmty (TyFreeVar a) i e2) \<longrightarrow> \<not> (Z1\<union>Zx) defines \<rho>"
      and Hlx: "finite Lx"
      and Hly: "finite Ly"
      using typing_let_elim[of "?\<Gamma>x" "\<Delta>,(a:\<star>)" "?Mx" Z ?es ?e1s ?e2s "vsubst_ty (TyFreeVar a) i \<tau>"]
      by auto
    have Hwfx: "\<Gamma>x;\<Delta>,(a:\<star>) \<Turnstile> OK" using Hwf wf_env_subset by auto
    then obtain \<Gamma>rx
      where Hrx: "\<Gamma>x = vsubst_ty_env (TyFreeVar a) i \<Gamma>rx"
      and Hrx_free: "\<not> a free in environ \<Gamma>rx"
      using parametrize_ty_env by blast
    have Hwf1: "\<Gamma>1;\<Delta>,(a:\<star>) \<Turnstile> OK" using Hwf wf_env_subset by auto
    then obtain \<Gamma>r1
      where Hr1: "\<Gamma>1 = vsubst_ty_env (TyFreeVar a) i \<Gamma>r1"
      and Hr1_free: "\<not> a free in environ \<Gamma>r1"
      using parametrize_ty_env by blast
    have Hwf2: "\<Gamma>2;\<Delta>,(a:\<star>) \<Turnstile> OK" using Hwf wf_env_subset by auto
    then obtain \<Gamma>r2
      where Hr2: "\<Gamma>2 = vsubst_ty_env (TyFreeVar a) i \<Gamma>r2"
      and Hr2_free: "\<not> a free in environ \<Gamma>r2"
      using parametrize_ty_env by blast
    have "\<not> a free in environ \<Gamma>" by assumption
    moreover
    have "\<not> a free in environ (\<Gamma>rx \<union> \<Gamma>r1)" using Hrx_free Hr1_free
      apply (auto simp add: free_ty_env_def) by force 
    hence "\<not> a free in environ (\<Gamma>rx \<union> \<Gamma>r1 \<union> \<Gamma>r2)" using Hr2_free
      apply (auto simp add: free_ty_env_def) by force 
    moreover
    have "vsubst_ty_env (TyFreeVar a) i \<Gamma> = vsubst_ty_env (TyFreeVar a) i (\<Gamma>rx \<union> \<Gamma>r1 \<union> \<Gamma>r2)"
      using Hrx Hr1 Hr2 H\<Gamma>x_eq by (auto simp add: vsubst_ty_env_def)
    ultimately
    have H\<Gamma>x_eq: "\<Gamma> = \<Gamma>rx \<union> \<Gamma>r1 \<union> \<Gamma>r2" using subst_subst_equal_env[of a i \<Gamma> "\<Gamma>rx\<union>\<Gamma>r1\<union>\<Gamma>r2"] by auto
    have "\<Delta>,(a:\<star>) \<Turnstile>' \<tau>x" using Htyp poly_ty_wf_type by auto
    then obtain \<tau>rx
      where H\<tau>rx:  "\<tau>x = vsubst_ty (TyFreeVar a) i \<tau>rx" and H\<tau>rx_free: "\<not> a free in type \<tau>rx"
      using parametrize_ty by blast 
    have Hwf_1: "\<Gamma>1;\<Delta>,(a:\<star>) \<Turnstile> OK" using Hwf wf_env_subset by auto
    have Hwf_tx: "\<Delta>,(a:\<star>) \<Turnstile>' \<tau>x" using Htyp poly_ty_wf_type by auto 
    {
      fix x \<rho>
      assume Hx: "\<not> \<Gamma>1 defines x" and Hlx: "x \<notin> Lx" 
        and Hr: "\<rho> \<notin> set (free_let_dollar_scope (vsubst_tmty \<phi>s i e1))"
        and "\<not> (Z2 \<union> Z1) defines \<rho>"
      moreover      
      have Hwf_tx: "\<Delta>,(a:\<star>) \<Turnstile>' convT \<rho> \<tau>x" using Hwf_tx by (cases \<tau>x, auto)
      have "\<Gamma>1,(x:convT \<rho> \<tau>x);\<Delta>,(a:\<star>) \<Turnstile> OK" apply (rule wf_env_cons) using Hwf_1 Hwf_tx Hx by auto
      moreover
      have "\<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>x}" using Hx by (auto simp add: env_compat_def dom_def)
      moreover
      have "\<rho> \<notin> set (free_let_dollar_scope (vsubst_tmty (TyFreeVar a) i e1))"
        using Hr free_dollars_vsubst_tmty by auto
      ultimately
      have "\<Gamma>1,(x:convT \<rho> \<tau>x);\<Delta>,(a:\<star>);vsubst_ty_env (TyFreeVar a) i M;Z1 \<turnstile> 
        freshen_tm x (freshen_sc \<rho> (vsubst_tmty (TyFreeVar a) i e1)) : \<tau>1 \<or>
        \<Gamma>1,(x:convT \<rho> \<tau>x);\<Delta>,(a:\<star>);vsubst_ty_env (TyFreeVar a) i M;Z1,(\<rho>:\<section>) \<turnstile> 
        freshen_tm x (freshen_sc \<rho> (vsubst_tmty (TyFreeVar a) i e1)) : \<tau>1"
        using Htyp1 by auto
    }
    hence Htyp1_o: "\<forall>x \<rho>. \<not> \<Gamma>1 defines x \<and> x \<notin> Lx \<and> \<rho> \<notin> set (free_let_dollar_scope (vsubst_tmty \<phi>s i e1))
      \<and> \<not> (Z2 \<union> Z1) defines \<rho> \<longrightarrow>
      \<Gamma>1,(x:convT \<rho> \<tau>x);\<Delta>,(a:\<star>);vsubst_ty_env (TyFreeVar a) i M;Z1 \<turnstile> 
        freshen_tm x (freshen_sc \<rho> (vsubst_tmty (TyFreeVar a) i e1)) : \<tau>1 \<or>
        \<Gamma>1,(x:convT \<rho> \<tau>x);\<Delta>,(a:\<star>);vsubst_ty_env (TyFreeVar a) i M;Z1,(\<rho>:\<section>) \<turnstile> 
        freshen_tm x (freshen_sc \<rho> (vsubst_tmty (TyFreeVar a) i e1)) : \<tau>1" by auto
    have "Z1 \<union> Z2 \<Turnstile> ScopeSet" using HZ1 HZ2 by (auto simp add: wf_scopeset_def)
    then obtain x_t \<rho>_t where "\<Gamma>1,(x_t:convT \<rho>_t \<tau>x);\<Delta>,(a:\<star>) \<Turnstile> OK \<and> x_t \<notin> Lx \<and>  \<Gamma>1 \<sim>{x_t\<rhd>convT \<rho>_t \<tau>x} \<and>
      \<rho>_t \<notin> set (free_let_dollar_scope (vsubst_tmty (TyFreeVar a) i e1)) \<and>
      \<not> (Z2 \<union> Z1) defines \<rho>_t"
      using new_scope_let[of Z1 Z2 Lx \<Gamma>1 "\<Delta>,(a:\<star>)" \<tau>x] Hwf_1 Hwf_tx Hlx by blast
    hence "\<Gamma>1,(x_t:convT \<rho>_t \<tau>x);\<Delta>,(a:\<star>);vsubst_ty_env (TyFreeVar a) i M;Z1 \<turnstile> 
        freshen_tm x_t (freshen_sc \<rho>_t (vsubst_tmty (TyFreeVar a) i e1)) : \<tau>1 \<or>
        \<Gamma>1,(x_t:convT \<rho>_t \<tau>x);\<Delta>,(a:\<star>);vsubst_ty_env (TyFreeVar a) i M;Z1,(\<rho>_t:\<section>) \<turnstile> 
        freshen_tm x_t (freshen_sc \<rho>_t (vsubst_tmty (TyFreeVar a) i e1)) : \<tau>1" using Htyp1 by auto
    hence Hwf_t1: "\<Delta>,(a:\<star>) \<Turnstile>' \<tau>1" using poly_ty_wf_type by auto
    then obtain \<tau>r1
      where H\<tau>r1:  "\<tau>1 = vsubst_ty (TyFreeVar a) i \<tau>r1" and H\<tau>r1_free: "\<not> a free in type \<tau>r1"
      using parametrize_ty by blast
    have "(vsubst_ty_env \<phi>s i \<Gamma>rx) \<union> (vsubst_ty_env \<phi>s i \<Gamma>r1) \<union> (vsubst_ty_env \<phi>s i \<Gamma>r2)
          ;\<Delta>;?Mt;Z2\<union>Z1\<union>Zx \<turnstile> let! ?e\<phi> ?e1\<phi> in ?e2\<phi> : vsubst_ty \<phi>s i \<tau>" proof rule
      have "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1 \<sim> vsubst_ty_env (TyFreeVar a) i \<Gamma>r2"
        using Hcompat1 Hr1 Hr2 by auto
      moreover
      have "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1 \<union> vsubst_ty_env (TyFreeVar a) i \<Gamma>r2;\<Delta>,(a:\<star>) \<Turnstile> OK"
        using Hwf Hr1 Hr2 wf_env_subset by blast
      ultimately
      show "vsubst_ty_env \<phi>s i \<Gamma>r1 \<sim> vsubst_ty_env \<phi>s i \<Gamma>r2"
        using env_subst_compat by auto
      next
      have "vsubst_ty_env (TyFreeVar a) i \<Gamma>rx \<sim> vsubst_ty_env (TyFreeVar a) i \<Gamma>r2"
        using Hcompat2 Hrx Hr2 by auto
      moreover
      have "vsubst_ty_env (TyFreeVar a) i \<Gamma>rx \<union> vsubst_ty_env (TyFreeVar a) i \<Gamma>r2;\<Delta>,(a:\<star>) \<Turnstile> OK"
        using Hwf Hrx Hr2 wf_env_subset by blast
      ultimately
      show "vsubst_ty_env \<phi>s i \<Gamma>rx \<sim> vsubst_ty_env \<phi>s i \<Gamma>r2"
        using env_subst_compat by auto
    next
      have "vsubst_ty_env (TyFreeVar a) i \<Gamma>rx \<sim> vsubst_ty_env (TyFreeVar a) i \<Gamma>r1"
        using Hcompat3 Hrx Hr1 by auto
      moreover
      have "vsubst_ty_env (TyFreeVar a) i \<Gamma>rx \<union> vsubst_ty_env (TyFreeVar a) i \<Gamma>r1;\<Delta>,(a:\<star>) \<Turnstile> OK"
        using Hwf Hrx Hr1 wf_env_subset by blast
      ultimately
      show "vsubst_ty_env \<phi>s i \<Gamma>rx \<sim> vsubst_ty_env \<phi>s i \<Gamma>r1"
        using env_subst_compat by auto
    next
        have "?\<Gamma>t;\<Delta> \<Turnstile> OK" by assumption
        also
        have "?\<Gamma>t = (vsubst_ty_env \<phi>s i \<Gamma>rx) \<union> (vsubst_ty_env \<phi>s i \<Gamma>r1) \<union> (vsubst_ty_env \<phi>s i \<Gamma>r2)"
           using H\<Gamma>x_eq by (auto simp add: vsubst_ty_env_def)
        finally
        show "vsubst_ty_env \<phi>s i \<Gamma>rx \<union> vsubst_ty_env \<phi>s i \<Gamma>r1 \<union> vsubst_ty_env \<phi>s i \<Gamma>r2;\<Delta> \<Turnstile> OK"
          by auto
    next
      show "Z1 \<Turnstile> ScopeSet" and "Z2 \<Turnstile> ScopeSet" by assumption
    next
      show "isBangable (vsubst_ty \<phi>s i \<tau>rx)" using Hbang H\<tau>rx proof (cases \<tau>rx, auto)
        fix q \<pi> \<phi>
        assume "isBangable (Qual q \<pi> (vsubst_pty (TyFreeVar a) i \<phi>))"
        thus "isBangable (Qual q \<pi> (vsubst_pty \<phi>s i \<phi>))" by (cases \<phi>, auto simp add: split_if)
      qed
    next
      show "vsubst_ty_env \<phi>s i \<Gamma>rx;\<Delta>;?Mt;Zx \<turnstile> vsubst_tmty \<phi>s i e : vsubst_ty \<phi>s i \<tau>rx" 
      proof (rule IH)
        show "size e < size e_y" using TmLetbang by auto
      next
        have "?\<Gamma>t;\<Delta> \<Turnstile> OK" by assumption
        thus "vsubst_ty_env \<phi>s i \<Gamma>rx;\<Delta> \<Turnstile> OK" 
          using H\<Gamma>x_eq wf_env_subset[of "?\<Gamma>t" \<Delta> "vsubst_ty_env \<phi>s i \<Gamma>rx"]
          by (auto simp add: vsubst_ty_env_def)
      next
        show "vsubst_ty_env \<phi>s i M;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<phi>s" 
          and "a \<notin> free_ty_env \<Gamma>rx" and "a \<notin> Environ.dom \<Delta>" 
          by assumption
      next
        note Htyp
        also note Hrx
        also note H\<tau>rx
        finally
        show "vsubst_ty_env (TyFreeVar a) i \<Gamma>rx;\<Delta>,(a:\<star>);?Mx;Zx \<turnstile> 
                vsubst_tmty (TyFreeVar a) i e : vsubst_ty (TyFreeVar a) i \<tau>rx"
          by auto
      next
        show "a \<notin> set (free_tmty e)" using "1.prems" TmLetbang by auto
      next
        show "a \<notin> set (free_ty \<tau>rx)" using H\<tau>rx_free by auto
      qed
   next
     show "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tmty \<phi>s i e1 \<longrightarrow> \<rho> \<notin> Environ.dom (Z2 \<union> Zx)"
       using Hdollar1 poly_free_dollars_tmty_equal by auto
   next
     show "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tmty \<phi>s i e2 \<longrightarrow> \<rho> \<notin> Environ.dom (Z1 \<union> Zx)"
       using Hdollar2 poly_free_dollars_tmty_equal by auto
   next
     show "\<forall>\<rho>. \<rho> is free dollar sc in vsubst_tmty \<phi>s i e \<longrightarrow> \<rho> \<notin> Environ.dom (Z1 \<union> Z2)"
       using HdollarX poly_free_dollars_tmty_equal by auto
   next
     {
       fix x \<rho>
       {
         assume Hwf_\<phi>: "vsubst_ty_env \<phi>s i \<Gamma>r1,(x:convT \<rho> (vsubst_ty \<phi>s i \<tau>rx));\<Delta> \<Turnstile> OK"
           and "vsubst_ty_env \<phi>s i \<Gamma>r1 \<sim> {x\<rhd>convT \<rho> (vsubst_ty \<phi>s i \<tau>rx)}" 
           and "x \<notin> (Lx\<union>dom \<Gamma>2\<union>dom \<Gamma>1)"
           and "\<rho> \<notin> set (free_let_dollar_scope (vsubst_tmty \<phi>s i e1))"
           and "\<rho> \<notin> Environ.dom (Z2 \<union> Z1)"
         hence "\<Gamma>1,(x:convT \<rho> \<tau>x);\<Delta>,(a:\<star>);?Mx;Z1 \<turnstile> 
                 freshen_tm x (freshen_sc \<rho> (vsubst_tmty (TyFreeVar a) i e1)) : \<tau>1
               \<or>
               \<Gamma>1,(x:convT \<rho> \<tau>x);\<Delta>,(a:\<star>);?Mx;Z1,(\<rho>:\<section>) \<turnstile> 
                 freshen_tm x (freshen_sc \<rho> (vsubst_tmty (TyFreeVar a) i e1)) : \<tau>1"
           using Htyp1_o Hr1 by auto
         also note Hr1
         also note H\<tau>r1
         also note H\<tau>rx
         finally
         have Htyp1:
           "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1,(x:convT \<rho> (vsubst_ty (TyFreeVar a) i \<tau>rx));\<Delta>,(a:\<star>);?Mx;Z1 \<turnstile> 
                 freshen_tm x (freshen_sc \<rho> (vsubst_tmty (TyFreeVar a) i e1)) : vsubst_ty (TyFreeVar a) i \<tau>r1
               \<or>
               vsubst_ty_env (TyFreeVar a) i \<Gamma>r1,(x:convT \<rho> (vsubst_ty (TyFreeVar a) i \<tau>rx));\<Delta>,(a:\<star>);?Mx;Z1,(\<rho>:\<section>) \<turnstile> 
                 freshen_tm x (freshen_sc \<rho> (vsubst_tmty (TyFreeVar a) i e1)) : vsubst_ty (TyFreeVar a) i \<tau>r1"
           by auto
         have "vsubst_ty_env \<phi>s i (\<Gamma>r1,(x:convT \<rho> \<tau>rx)) = 
             vsubst_ty_env \<phi>s i \<Gamma>r1,(x:vsubst_ty \<phi>s i (convT \<rho> \<tau>rx))" 
           by (auto simp add: vsubst_ty_env_def)
         hence Hg_eq: "vsubst_ty_env \<phi>s i (\<Gamma>r1,(x:convT \<rho> \<tau>rx)) = 
             vsubst_ty_env \<phi>s i \<Gamma>r1,(x:convT \<rho> (vsubst_ty \<phi>s i \<tau>rx))" 
           by (cases \<tau>rx, auto)
         have Hscfree: "\<forall>i \<rho>. vsubst_sc_pty (ScFree \<rho>) i \<phi>s = \<phi>s" by assumption
         have "vsubst_tmty \<phi>s i (freshen_tm x (freshen_sc \<rho> e1)) =
              freshen_tm x (vsubst_tmty \<phi>s i (freshen_sc \<rho> e1))" 
           using subst_subst_tytm_swap by auto
         also have "vsubst_tmty \<phi>s i (freshen_sc \<rho> e1) = freshen_sc \<rho> (vsubst_tmty \<phi>s i e1)"
           using subst_subst_tmtysc_swap Hscfree by auto
         finally
         have He_eq: "vsubst_tmty \<phi>s i (freshen_tm x (freshen_sc \<rho> e1)) =
              freshen_tm x (freshen_sc \<rho> (vsubst_tmty \<phi>s i e1))"  by auto
         have "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1,(x:vsubst_ty (TyFreeVar a) i (convT \<rho> \<tau>rx)) =
             vsubst_ty_env (TyFreeVar a) i (\<Gamma>r1,(x:convT \<rho> \<tau>rx))"
           by (auto simp add: vsubst_ty_env_def)
         hence Hg_eq2: "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1,(x:convT \<rho> (vsubst_ty (TyFreeVar a) i \<tau>rx)) =
             vsubst_ty_env (TyFreeVar a) i (\<Gamma>r1,(x:convT \<rho> \<tau>rx))" 
           by (cases \<tau>rx, auto)
         have "freshen_tm x (freshen_sc \<rho> (vsubst_tmty (TyFreeVar a) i e1)) =
               freshen_tm x (vsubst_tmty (TyFreeVar a) i (freshen_sc \<rho> e1))"
           using subst_subst_tmtysc_swap by auto
         also have "freshen_tm x (vsubst_tmty (TyFreeVar a) i (freshen_sc \<rho> e1)) =
                vsubst_tmty (TyFreeVar a) i (freshen_tm x (freshen_sc \<rho> e1))"
           using subst_subst_tytm_swap by auto
         finally
         have He_eq2: "freshen_tm x (freshen_sc \<rho> (vsubst_tmty (TyFreeVar a) i e1)) =
             vsubst_tmty (TyFreeVar a) i (freshen_tm x (freshen_sc \<rho> e1))" 
           by auto
         have Hih1: "size (freshen_tm x (freshen_sc \<rho> e1)) < size e_y" 
           using TmLetbang vsubst_var_same_size vsubst_sc_var_same_size by auto
         note Hwf_\<phi>
         hence Hih2: "vsubst_ty_env \<phi>s i (\<Gamma>r1,(x:convT \<rho> \<tau>rx));\<Delta> \<Turnstile> OK" using Hg_eq by auto 
         have Hih3: "vsubst_ty_env \<phi>s i M;\<Delta> \<Turnstile> OK" by assumption
         have Hih4: "\<Delta> \<Turnstile>' \<phi>s" by assumption
         have "a \<notin> set (free_tmty e1)" using "1.prems" TmLetbang by auto
         hence Hih6: "a \<notin> set (free_tmty (freshen_tm x (freshen_sc \<rho> e1)))" 
           using not_free_subst_tmty not_free_subst_tm not_free_subst_sc by auto
         have Hih7: "a \<notin> set (free_ty \<tau>r1)" using H\<tau>r1_free by auto
         have "\<not> a free in type (convT \<rho> \<tau>rx)" using H\<tau>rx_free 
           by (cases \<tau>rx, auto)
         hence Hih8: "a \<notin> free_ty_env (\<Gamma>r1,(x:convT \<rho> \<tau>rx))" 
           using Hr1_free by (auto simp add: free_ty_env_def)
         have Hih9_pre: "\<forall>i \<rho>. vsubst_sc_pty (ScFree \<rho>) i \<phi>s = \<phi>s" by assumption
         have Hih9: "a \<notin> free_ty_env M" by assumption
         have Hih10: "a \<notin> Environ.dom \<Delta>" by assumption
         have "vsubst_ty_env \<phi>s i \<Gamma>r1,(x:convT \<rho> (vsubst_ty \<phi>s i \<tau>rx));\<Delta>;?Mt;Z1 \<turnstile> 
                 freshen_tm x (freshen_sc \<rho> (vsubst_tmty \<phi>s i e1)) : vsubst_ty \<phi>s i \<tau>r1
               \<or>
               vsubst_ty_env \<phi>s i \<Gamma>r1,(x:convT \<rho> (vsubst_ty \<phi>s i \<tau>rx));\<Delta>;?Mt;Z1,(\<rho>:\<section>) \<turnstile> 
                 freshen_tm x (freshen_sc \<rho> (vsubst_tmty \<phi>s i e1)) : vsubst_ty \<phi>s i \<tau>r1"
           using Htyp1
         proof
           assume "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1,(x:convT \<rho> (vsubst_ty (TyFreeVar a) i \<tau>rx));\<Delta>,(a:\<star>);?Mx;Z1 \<turnstile>
             freshen_tm x (freshen_sc \<rho> (vsubst_tmty (TyFreeVar a) i e1)) : vsubst_ty (TyFreeVar a) i \<tau>r1"
           also note Hg_eq2
           also note He_eq2
           finally
           have Hih5: " vsubst_ty_env (TyFreeVar a) i (\<Gamma>r1,(x:convT \<rho> \<tau>rx));\<Delta>,(a:\<star>);
             vsubst_ty_env (TyFreeVar a) i M;Z1 \<turnstile>
             vsubst_tmty (TyFreeVar a) i (freshen_tm x (freshen_sc \<rho> e1)) : vsubst_ty (TyFreeVar a) i \<tau>r1" 
             by auto
           have "vsubst_ty_env \<phi>s i (\<Gamma>r1,(x:convT \<rho> \<tau>rx));\<Delta>;?Mt;Z1 \<turnstile> 
                 vsubst_tmty \<phi>s i (freshen_tm x (freshen_sc \<rho> e1)) : vsubst_ty \<phi>s i \<tau>r1"
             apply (rule IH) using  Hih1 Hih2 Hih3 Hih4 Hih5 Hih6 Hih7 Hih8 Hih9_pre Hih9 Hih10 by auto
           also note Hg_eq
           also note He_eq
           finally
           show ?thesis by auto 
         next
           assume "vsubst_ty_env (TyFreeVar a) i \<Gamma>r1,(x:convT \<rho> (vsubst_ty (TyFreeVar a) i \<tau>rx));\<Delta>,(a:\<star>);
             ?Mx;Z1,(\<rho>:\<section>) \<turnstile> 
                 freshen_tm x (freshen_sc \<rho> (vsubst_tmty (TyFreeVar a) i e1)) : vsubst_ty (TyFreeVar a) i \<tau>r1"
           also note Hg_eq2
           also note He_eq2
           finally
           have Hih5: " vsubst_ty_env (TyFreeVar a) i (\<Gamma>r1,(x:convT \<rho> \<tau>rx));\<Delta>,(a:\<star>);
             vsubst_ty_env (TyFreeVar a) i M;Z1,(\<rho>:\<section>) \<turnstile>
             vsubst_tmty (TyFreeVar a) i (freshen_tm x (freshen_sc \<rho> e1)) : vsubst_ty (TyFreeVar a) i \<tau>r1" 
             by auto
           have "vsubst_ty_env \<phi>s i (\<Gamma>r1,(x:convT \<rho> \<tau>rx));\<Delta>;?Mt;Z1,(\<rho>:\<section>) \<turnstile> 
                 vsubst_tmty \<phi>s i (freshen_tm x (freshen_sc \<rho> e1)) : vsubst_ty \<phi>s i \<tau>r1" 
             apply (rule IH) using  Hih1 Hih2 Hih3 Hih4 Hih5 Hih6 Hih7 Hih8 Hih9_pre Hih9 Hih10 by auto
           also note Hg_eq
           also note He_eq
           finally
           show ?thesis by auto 
         qed
       }
       hence "vsubst_ty_env \<phi>s i \<Gamma>r1,(x:convT \<rho> (vsubst_ty \<phi>s i \<tau>rx));\<Delta> \<Turnstile> OK \<and>
                  vsubst_ty_env \<phi>s i \<Gamma>r1 \<sim> {x\<rhd>convT \<rho> (vsubst_ty \<phi>s i \<tau>rx)} \<and> x \<notin> (Lx\<union>dom \<Gamma>2\<union>dom \<Gamma>1) \<and>
                  \<rho> \<notin> set (free_let_dollar_scope (vsubst_tmty \<phi>s i e1)) \<and> \<rho> \<notin> Environ.dom (Z2 \<union> Z1) \<longrightarrow>
                    vsubst_ty_env \<phi>s i \<Gamma>r1,(x:convT \<rho> (vsubst_ty \<phi>s i \<tau>rx));\<Delta>;?Mt;Z1 \<turnstile> 
                      freshen_tm x (freshen_sc \<rho> (vsubst_tmty \<phi>s i e1)) : vsubst_ty \<phi>s i \<tau>r1
                    \<or>
                    vsubst_ty_env \<phi>s i \<Gamma>r1,(x:convT \<rho> (vsubst_ty \<phi>s i \<tau>rx));\<Delta>;?Mt;Z1,(\<rho>:\<section>) \<turnstile> 
                      freshen_tm x (freshen_sc \<rho> (vsubst_tmty \<phi>s i e1)) : vsubst_ty \<phi>s i \<tau>r1"
         by auto
     }
     thus "\<forall>x \<rho>. vsubst_ty_env \<phi>s i \<Gamma>r1,(x:convT \<rho> (vsubst_ty \<phi>s i \<tau>rx));\<Delta> \<Turnstile> OK \<and>
                  vsubst_ty_env \<phi>s i \<Gamma>r1 \<sim> {x\<rhd>convT \<rho> (vsubst_ty \<phi>s i \<tau>rx)} \<and> x \<notin> (Lx\<union>dom \<Gamma>2\<union>dom \<Gamma>1) \<and>
                  \<rho> \<notin> set (free_let_dollar_scope (vsubst_tmty \<phi>s i e1)) \<and> \<rho> \<notin> Environ.dom (Z2 \<union> Z1) \<longrightarrow>
                    vsubst_ty_env \<phi>s i \<Gamma>r1,(x:convT \<rho> (vsubst_ty \<phi>s i \<tau>rx));\<Delta>;?Mt;Z1 \<turnstile> 
                      freshen_tm x (freshen_sc \<rho> (vsubst_tmty \<phi>s i e1)) : vsubst_ty \<phi>s i \<tau>r1
                    \<or>
                    vsubst_ty_env \<phi>s i \<Gamma>r1,(x:convT \<rho> (vsubst_ty \<phi>s i \<tau>rx));\<Delta>;?Mt;Z1,(\<rho>:\<section>) \<turnstile> 
                      freshen_tm x (freshen_sc \<rho> (vsubst_tmty \<phi>s i e1)) : vsubst_ty \<phi>s i \<tau>r1" by auto 
   next
     show "\<forall>x y. vsubst_ty_env \<phi>s i \<Gamma>r2,(x:vsubst_ty \<phi>s i \<tau>rx),(y:vsubst_ty \<phi>s i \<tau>r1);\<Delta> \<Turnstile> OK \<and>
        vsubst_ty_env \<phi>s i \<Gamma>r2 \<sim> {x\<rhd>vsubst_ty \<phi>s i \<tau>rx},(y:vsubst_ty \<phi>s i \<tau>r1) \<and> 
        {x\<rhd>vsubst_ty \<phi>s i \<tau>rx} \<sim> {y\<rhd>vsubst_ty \<phi>s i \<tau>r1} \<and> x \<notin> (Lx\<union>dom \<Gamma>2\<union>dom \<Gamma>1) \<and> y \<notin> (Ly\<union>dom \<Gamma>2) 
       \<and> x \<noteq> y \<longrightarrow>
        vsubst_ty_env \<phi>s i \<Gamma>r2,(x:vsubst_ty \<phi>s i \<tau>rx),(y:vsubst_ty \<phi>s i \<tau>r1);\<Delta>;?Mt;Z2 \<turnstile>
        vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x (vsubst_tmty \<phi>s i e2)) : vsubst_ty \<phi>s i \<tau>"
      proof auto
        fix x y
        assume Hwf_\<phi>: "vsubst_ty_env \<phi>s i \<Gamma>r2,(x:vsubst_ty \<phi>s i \<tau>rx),(y:vsubst_ty \<phi>s i \<tau>r1);\<Delta> \<Turnstile> OK"
          and "vsubst_ty_env \<phi>s i \<Gamma>r2 \<sim> {x\<rhd>vsubst_ty \<phi>s i \<tau>rx},(y:vsubst_ty \<phi>s i \<tau>r1)"
          and "{x\<rhd>vsubst_ty \<phi>s i \<tau>rx} \<sim> {y\<rhd>vsubst_ty \<phi>s i \<tau>r1}"
          and Hx_lx: "x \<notin> Lx" and Hy_ly: "y \<notin> Ly"
          and Hx_fresh: "x \<notin> Environ.dom \<Gamma>2"
          and Hy_fresh: "y \<notin> Environ.dom \<Gamma>2"
          and Hxy_diff: "x \<noteq> y"
        have "vsubst_ty_env \<phi>s i (\<Gamma>r2,(x:\<tau>rx),(y:\<tau>r1));\<Delta>;?Mt;Z2 \<turnstile> 
          vsubst_tmty \<phi>s i (vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 e2)) : 
          vsubst_ty \<phi>s i \<tau>"
        proof (rule IH)
          show "size (vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 e2)) < size e_y"
            using TmLetbang by auto
        next
          show "vsubst_ty_env \<phi>s i M;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<phi>s" 
            and "a \<notin> free_ty_env M" and "a \<notin> Environ.dom \<Delta>"
            and "a \<notin> set (free_ty \<tau>)"
            by assumption
        next
          have "\<not> a free in term e2" using TmLetbang prems by auto
          thus "a \<notin> set (free_tmty (vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 e2)))"
            using free_tmty_subst_equal by auto
        next
          show "a \<notin> free_ty_env (\<Gamma>r2,(x:\<tau>rx),(y:\<tau>r1))"
            using Hr2_free H\<tau>rx_free H\<tau>r1_free by (auto simp add: free_ty_env_def)
        next
          note Hwf_\<phi>
          also have "vsubst_ty_env \<phi>s i \<Gamma>r2,(x:vsubst_ty \<phi>s i \<tau>rx),(y:vsubst_ty \<phi>s i \<tau>r1) =
            vsubst_ty_env \<phi>s i (\<Gamma>r2,(x:\<tau>rx),(y:\<tau>r1))"
            by (auto simp add: vsubst_ty_env_def)
          finally
          show "vsubst_ty_env \<phi>s i (\<Gamma>r2,(x:\<tau>rx),(y:\<tau>r1));\<Delta> \<Turnstile> OK" by auto 
        next
          have Hwf: "\<Gamma>2;\<Delta>,(a:\<star>) \<Turnstile> OK" using Hwf wf_env_subset by auto
          have "\<Gamma>2,(x:\<tau>x);\<Delta>,(a:\<star>) \<Turnstile> OK" using wf_env_cons Hx_fresh Hwf Hwf_tx by auto
          hence "\<Gamma>2,(x:\<tau>x),(y:\<tau>1);\<Delta>,(a:\<star>) \<Turnstile> OK" 
            using wf_env_cons[of "\<Gamma>2,(x:\<tau>x)" "\<Delta>,(a:\<star>)" y] Hy_fresh Hxy_diff Hwf Hwf_t1
            by (auto simp add: dom_def)
          moreover
          have "\<Gamma>2 \<sim> {x\<rhd>\<tau>x},(y:\<tau>1) \<and> {x\<rhd>\<tau>x} \<sim> {y\<rhd>\<tau>1}" using Hy_fresh Hx_fresh Hxy_diff 
            by (auto simp add: dom_def env_compat_def)
          moreover
          have "x \<notin> Lx \<and> y \<notin> Ly \<and> x \<noteq> y" using Hx_lx Hy_ly Hxy_diff by auto
          ultimately
          have "\<Gamma>2,(x:\<tau>x),(y:\<tau>1);\<Delta>,(a:\<star>);?Mx;Z2 \<turnstile>
                   vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 (vsubst_tmty (TyFreeVar a) i e2)) 
                   : vsubst_ty (TyFreeVar a) i \<tau>" using Htyp2 by auto
          also note Hr2 
          also note H\<tau>rx
          also note H\<tau>r1
          also
          have "vsubst_ty_env (TyFreeVar a) i \<Gamma>r2,(
            x:vsubst_ty (TyFreeVar a) i \<tau>rx),(y:vsubst_ty (TyFreeVar a) i \<tau>r1) =
            vsubst_ty_env (TyFreeVar a) i (\<Gamma>r2,(x:\<tau>rx),(y:\<tau>r1))"
            by (auto simp add: vsubst_ty_env_def)
          also
          have "vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 (vsubst_tmty (TyFreeVar a) i e2)) =
            vsubst_tmty (TyFreeVar a) i (vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 e2))"
            using subst_subst_tmty_swap by auto
          finally
          show "vsubst_ty_env (TyFreeVar a) i (\<Gamma>r2,(x:\<tau>rx),(y:\<tau>r1));\<Delta>,(a:\<star>);?Mx;Z2 \<turnstile> 
                  vsubst_tmty (TyFreeVar a) i (vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 e2)) :
                  vsubst_ty (TyFreeVar a) i \<tau>" by auto
        qed
        also 
        have "vsubst_ty_env \<phi>s i (\<Gamma>r2,(x:\<tau>rx),(y:\<tau>r1)) =
          vsubst_ty_env \<phi>s i \<Gamma>r2,(x:vsubst_ty \<phi>s i \<tau>rx),(y:vsubst_ty \<phi>s i \<tau>r1)"
          by (auto simp add: vsubst_ty_env_def)
        also 
        have "vsubst_tmty \<phi>s i (vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 e2)) =
          vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 (vsubst_tmty \<phi>s i e2))"
          using subst_subst_tytm_swap by auto
        finally
        show "vsubst_ty_env \<phi>s i \<Gamma>r2,(x:vsubst_ty \<phi>s i \<tau>rx),(y:vsubst_ty \<phi>s i \<tau>r1);\<Delta>;
          vsubst_ty_env \<phi>s i M;Z2 \<turnstile> 
          vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 (vsubst_tmty \<phi>s i e2)) : 
          vsubst_ty \<phi>s i \<tau>"
          by auto
      qed
    next
      have "validScope (scopeT (vsubst_ty (TyFreeVar a) i \<tau>r1)) Z1" using Hvalid H\<tau>r1 by auto
      thus "validScope (scopeT (vsubst_ty \<phi>s i \<tau>r1)) Z1" by (cases \<tau>r1, auto)
    next
      have "\<Gamma>2;\<Delta>,(a:\<star>) \<Turnstile> OK" using Hwf wf_env_subset by auto
      hence Hf2: "finite (dom \<Gamma>2)" using dom_finite by (auto simp add: wf_env_def)
      have "\<Gamma>1;\<Delta>,(a:\<star>) \<Turnstile> OK" using Hwf wf_env_subset by auto
      hence Hf1: "finite (dom \<Gamma>1)" using dom_finite by (auto simp add: wf_env_def)
      thus "finite (Lx \<union> dom \<Gamma>2 \<union> dom \<Gamma>1)" and "finite (Ly \<union> dom \<Gamma>2)" using Hlx Hly Hf1 Hf2 by auto
    qed
    also have "(vsubst_ty_env \<phi>s i \<Gamma>rx) \<union> (vsubst_ty_env \<phi>s i \<Gamma>r1) \<union> (vsubst_ty_env \<phi>s i \<Gamma>r2) = ?\<Gamma>t"
     using H\<Gamma>x_eq by (auto simp add: vsubst_ty_env_def)
   finally
   show ?thesis using TmLetbang HZ_eq by auto
 qed
qed

end
