theory Safety
imports Metatheory StoreTyping MemTyping PreType

begin

lemma free_scope_store_cons_empty:
  assumes "scopeV v = ScBot"
  shows "free_scope_store (S,(z\<mapsto>v)) = free_scope_store S"
proof -
  have "free_scope_store (S,(z\<mapsto>v)) =
        free_scope_store S \<union> set (free_scope_scope (scopeV v))"
    using free_scope_store_def by auto
  also have "free_scope_scope (scopeV v) = []" using prems by auto
  finally show ?thesis by auto
qed

lemma free_scope_store_without_empty:
  assumes "scopeV v = ScBot"
  and "z\<mapsto>v \<in> S"
  and "S \<Turnstile> Store"
  shows "free_scope_store S = free_scope_store (S\<bar>z)"
proof -
  have "free_scope_store ((S\<bar>z),(z\<mapsto>v)) = free_scope_store (S\<bar>z)"
    using free_scope_store_cons_empty[of _ _ "S\<bar>z"] prems by auto
  also have "(S\<bar>z),(z\<mapsto>v) = S" using wf_store_defines_without prems by auto
  finally show ?thesis by auto
qed

lemma ty_valid_scopeV:
  assumes "\<Gamma>;\<Delta>;M;Z \<turnstile> v : \<tau>"
  and "value v"
  shows "validScope (scopeV v) Z"
proof -
  have "validScope (scopeT \<tau>) Z" using ty_valid_scope assms by auto
  moreover
  have "scopeV v = scopeT \<tau>" using ty_scope_sanity assms by auto
  ultimately
  show "validScope (scopeV v) Z" by auto
qed

lemma valid_scope_equal_cons:
  assumes "free_scope_store S = dom Zs"
  and "validScope (scopeV v) Zs"
  shows "free_scope_store (S,(z\<mapsto>v)) = dom Zs"
using prems proof (cases "scopeV v")
  case ScBot
  hence "free_scope_store S = free_scope_store (S,(z\<mapsto>v))"
    by (auto simp add: free_scope_store_def)
  thus ?thesis using assms by auto
next
  case (ScFree \<rho>)
  have Heq: "free_scope_store (S,(z\<mapsto>v)) =
  free_scope_store S \<union> (set (free_scope_scope (scopeV v)))"
    by (auto simp add: free_scope_store_def)
  hence "free_scope_store (S,(z\<mapsto>v)) = (free_scope_store S) \<union> {\<rho>}" using ScFree by auto
  hence "free_scope_store (S,(z\<mapsto>v)) = (dom Zs) \<union> {\<rho>}" using assms by auto
  moreover
  have "Zs defines \<rho>" using assms ScFree by (auto simp add: dom_def)
  hence "dom Zs = dom Zs \<union> {\<rho>}" by auto
  ultimately
  show ?thesis by auto
qed auto

lemma validScope_weaken:
  assumes "validScope \<pi> Z"
  and "Z \<subseteq> Z'"
  shows "validScope \<pi> Z'"
using prems by (cases \<pi>, auto simp add: dom_def)

lemma value_preservation:
  assumes "value v"
  and "M;\<Delta>;Zs;Zl \<Turnstile> S : (\<Gamma>s \<union> \<Gamma>m)"
  and Has: "\<Gamma>m;\<Delta> \<Turnstile>' \<mu> : M"
  and Ht: "\<Gamma>e;\<Delta>;{};Ze \<turnstile> v : \<tau>"
  and "\<Gamma>e \<union> \<Gamma>r \<subseteq> \<Gamma>s"
  and "\<Gamma>e \<sim> \<Gamma>r"
  and "\<Gamma>e \<union> \<Gamma>r;\<Delta> \<Turnstile> OK"
  and "\<Gamma>s \<sim> \<Gamma>m"
  and "\<not> S defines z"
  and "Ze \<subseteq> Zl"
  and "Zr \<subseteq> Zl"
  and "Zl \<subseteq> Zs"
  and "free_scope_store S = dom Zs"
  and "INV2 (filterStore S Zl)"
  and "INV3 S"
  and "INV4 (filterStore S Zl)"
  and "INV5 S"
  and "INV6 v Zr"
  and "INV9 v"
  shows "\<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'.
          (M';\<Delta>;Zs';Zl' \<Turnstile> S,(z\<mapsto>v) : (\<Gamma>s' \<union> \<Gamma>m')) \<and>
          (\<Gamma>m';\<Delta> \<Turnstile>' \<mu> : M') \<and>
          (\<Gamma>e';\<Delta>;{};Ze' \<turnstile> TmFreeVar z : \<tau>) \<and>
          (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
          (\<Gamma>e' \<sim> \<Gamma>r) \<and>
          (\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK) \<and>
          (\<Gamma>s' \<sim> \<Gamma>m') \<and>
          (Ze' \<subseteq> Zl') \<and> (Zl' \<subseteq> Zs') \<and> (Zs \<subseteq> Zs') \<and> (Zr \<subseteq> Zl') \<and>
          (free_scope_store (S,(z\<mapsto>v)) = dom Zs') \<and>
          INV2 (filterStore (S,(z\<mapsto>v)) Zl') \<and> INV3 (S,(z\<mapsto>v)) \<and>
          INV4 (filterStore (S,(z\<mapsto>v)) Zl') \<and> INV5 (S,(z\<mapsto>v)) \<and> INV6 (TmFreeVar z) Zr \<and>
          INV7 v (TmFreeVar z) Zs \<and> INV8 Ze' Ze (TmFreeVar z) \<and> INV9 (TmFreeVar z)"
proof -
  have H: "validScope (scopeT \<tau>) Ze" using assms ty_valid_scope by auto
  then obtain Ze' where Hmin: "minimalValidScope (scopeT \<tau>) Ze'" and Hsub: "Ze' \<subseteq> Ze"
  proof (cases "scopeT \<tau>")
    case ScBot
    have "minimalValidScope ScBot {}" and "{} \<subseteq> Ze" by (auto simp add: minimalValidScope_def)
    thus ?thesis using ScBot prems by auto
  next
    case (ScFree \<rho>)
    have "minimalValidScope (ScFree \<rho>) ({},(\<rho>:\<section>))" by (auto simp add: minimalValidScope_def)
    moreover
    have "Ze defines \<rho>" using H ScFree by auto
    hence "(\<rho>, EmptyScope) \<in> Ze" proof (auto simp add: dom_def)
      fix b
      assume "(\<rho>, b) \<in> Ze"
      thus "(\<rho>, EmptyScope) \<in> Ze" by (cases b, auto)
    qed
    hence "({},(\<rho>:\<section>)) \<subseteq> Ze" by auto
    ultimately
    show ?thesis using prems by auto
  qed auto
  have "\<Delta> \<Turnstile>' TY" and H\<tau>: "\<Delta> \<Turnstile>' \<tau>" using assms poly_ty_wf_type poly_wf_env_ty by auto
  hence Hg1: "{z\<rhd>\<tau>};\<Delta>;{};Ze' \<turnstile> TmFreeVar z : \<tau>" using memTyping_wf_loc
    by (auto simp add: wf_env_def without_def qualGrelL_def)
  have Hnotin: "\<not> (\<Gamma>s\<union>\<Gamma>m) defines z" using prems store_dom_not_in by auto
  hence "\<not> \<Gamma>s defines z" by (auto simp add: dom_def)
  hence Hgr: "\<not> \<Gamma>r defines z" using prems by (auto simp add: dom_def)
  hence Hg2: "{z\<rhd>\<tau>} \<sim> \<Gamma>r" by (auto simp add: env_compat_def dom_def)
  have "(filterStore S Zl) \<subseteq> S" by (auto simp add: filterStore_def)
  hence "\<not> (filterStore S Zl) defines z" using prems by (auto simp add: dom_def)
  hence "INV2 ((filterStore S Zl),(z\<mapsto>v))"
    and "INV4 ((filterStore S Zl),(z\<mapsto>v))"
    using inv4_cons[of "filterStore S Zl" z v] ty_no_loc
      inv2_cons[of "filterStore S Zl" z v] ty_no_locs[of \<Gamma>e _ Ze v \<tau>] assms
    by auto
  moreover
  have Hfilter_sub: "filterStore (S,(z\<mapsto>v)) Zl \<subseteq> (filterStore S Zl),(z\<mapsto>v)"
    by (auto simp add: filterStore_def)
  ultimately
  have Hg4: "INV2 (filterStore (S,(z\<mapsto>v)) Zl)" and Hg15: "INV4 (filterStore (S,(z\<mapsto>v)) Zl)"
    by (auto simp only: inv2_subset inv4_subset)
  let ?\<Gamma>s' = "(\<Gamma>s - lin \<Gamma>e)"
  have Heq: "?\<Gamma>s' \<union> \<Gamma>e = \<Gamma>s" using prems lin_def by auto
  hence Heqm: "?\<Gamma>s' \<union> \<Gamma>m \<union> \<Gamma>e = \<Gamma>s \<union> \<Gamma>m" using prems lin_def by auto
  have A1: "\<Gamma>e;\<Delta>;M;Ze \<turnstile> v : \<tau>"
    using prems storeTyping_wf_mem weakening_m[of \<Gamma>e _ "{}" _ v \<tau> M] by auto
  have "\<Gamma>m \<sim> \<Gamma>e" using prems env_compat_subset[of \<Gamma>s \<Gamma>m \<Gamma>e] env_compat_symm
    by auto
  moreover
  have "\<Gamma>s - lin \<Gamma>e \<sim> \<Gamma>e" using prems
  proof (auto simp add: env_compat_def lin_def)
    fix x \<tau>
    assume "qualT \<tau> \<noteq> qL"
    thus "qualT \<tau> \<preceq> qU" by (cases "qualT \<tau>", auto)
  qed
  ultimately
  have A2: "(?\<Gamma>s' \<union> \<Gamma>m) \<sim> \<Gamma>e" using prems by (auto simp add: env_compat_def)
  have "un \<Gamma>e \<subseteq> un \<Gamma>s" using prems by (auto simp add: un_def)
  moreover
  have "un (\<Gamma>s - lin \<Gamma>e \<union> \<Gamma>m) = un \<Gamma>s \<union> un \<Gamma>m"
    by (auto simp add: un_def lin_def)
  ultimately
  have A3:  "un \<Gamma>e \<subseteq> un (\<Gamma>s - lin \<Gamma>e \<union> \<Gamma>m)" by auto
  have "validScope (scopeV v) Ze" using prems ty_valid_scopeV by auto
  hence A4: "validScope (scopeV v) Zl" using validScope_weaken prems by auto
  have "M;\<Delta>;Zs;Zl \<Turnstile> S,(z\<mapsto>v) : (?\<Gamma>s' \<union> \<Gamma>m),(z:\<tau>)"
    apply (rule StoreTyping.st_cons[where Z=Ze])
    using A1 A2 A3 A4 Heqm storeTyping_wf_store[of M] storeTyping_wf_env[of M] prems by auto
  hence Hg5: "M;\<Delta>;Zs;Zl \<Turnstile> S,(z\<mapsto>v) : (?\<Gamma>s',(z:\<tau>) \<union> \<Gamma>m)" by auto
  have "\<Gamma>r \<sim> \<Gamma>e" using prems env_compat_symm by auto
  hence "\<Gamma>r \<inter> lin \<Gamma>e = {}" proof (auto simp add: env_compat_def lin_def)
    fix x \<tau>
    assume "\<forall>x \<tau>. (x, Term \<tau>) \<in> \<Gamma>r \<and> (x, Term \<tau>) \<in> \<Gamma>e \<longrightarrow> qualT \<tau> \<preceq> qU"
      and "(x\<rhd>\<tau>) \<in> \<Gamma>r"
      and "(x\<rhd>\<tau>) \<in> \<Gamma>e"
      and "qualT \<tau> = qL"
    hence "qualT \<tau> \<preceq> qU" by force
    thus False using prems by auto
  qed
  hence Hg6: "({z\<rhd>\<tau>} \<union> \<Gamma>r) \<subseteq> ?\<Gamma>s',(z:\<tau>)" using prems by auto
  have "?\<Gamma>s' \<sim> \<Gamma>m" using prems env_compat_subset[of \<Gamma>s \<Gamma>m ?\<Gamma>s'] by auto
  moreover
  have "\<not> \<Gamma>m defines z"
    using prems store_dom_not_in[where \<Gamma>="\<Gamma>s\<union>\<Gamma>m"] by (auto simp add: dom_def)
  hence "{z\<rhd>\<tau>} \<sim> \<Gamma>m" by (auto simp add: env_compat_def dom_def)
  ultimately
  have Hg7: "?\<Gamma>s',(z:\<tau>) \<sim> \<Gamma>m"
    using env_compat_union_right[of "?\<Gamma>s'" \<Gamma>m "{z\<rhd>\<tau>}"] by auto
  have "\<Gamma>r \<subseteq> \<Gamma>s\<union>\<Gamma>m" using prems by auto
  hence "\<Gamma>r;\<Delta> \<Turnstile> OK"
    using prems wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" _ \<Gamma>r] storeTyping_wf_env by auto
  hence Hg9: "{z\<rhd>\<tau>} \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" using Hgr wf_env_cons H\<tau> by auto
  have Hg10: "INV3 (S,(z\<mapsto>v))" using prems ty_no_locs by auto
  have Hg11: "Ze' \<subseteq> Zl" using Hsub assms by auto
  have Hg12: "Zl \<subseteq> Zs" and Hg14: "Zr \<subseteq> Zl" by assumption
  have Hassm: "free_scope_store S = dom Zs" by assumption
  moreover
  have "scopeV v = scopeT \<tau>" using prems ty_scope_sanity by auto
  hence "minimalValidScope (scopeV v) Ze'" using Hmin by auto
  hence "validScope (scopeV v) Ze'" using minimal_is_valid by auto
  hence Hvalid: "validScope (scopeV v) Zs" using validScope_weaken Hg12 Hg11 by auto
  ultimately
  have Hg13: "free_scope_store (S,(z\<mapsto>v)) = dom Zs" using valid_scope_equal_cons by auto
  have "\<not> hasDollars v" using INV9_elim assms by auto
  hence Hg16: "INV5 (S,(z\<mapsto>v))" using assms by (auto simp add: INV5)
  have Hg17: "INV6 (TmFreeVar z) Zr" by auto
  have Hg18: "INV7 v (TmFreeVar z) Zs" by auto
  have Hg19: "INV8 Ze' Ze (TmFreeVar z)" using Hsub by (auto simp add: dom_def)
  show ?thesis
    apply (rule exI[of _ "{z\<rhd>\<tau>}"], rule exI[of _ "?\<Gamma>s',(z:\<tau>)"])
    apply (rule exI[of _ \<Gamma>m], rule exI[of _ M])
    apply (rule exI[of _ Zs], rule exI[of _ Ze'], rule exI[of _ Zl])
    using Hg1 Hg2 Hg4 Hg5 Hg6 Hg7 Hg9 Hg10 Has Hg11 Hg12 Hg13 Hg14 Hg15 Hg16 Hg17 Hg18 Hg19
    by auto
qed

lemma storeTyping_scope_sanity:
  assumes "M;\<Delta>;Z;Zl \<Turnstile> S : \<Gamma>s"
  and "(x\<mapsto>v) \<in> S"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>s"
  shows "scopeV v = scopeT \<tau>"
using prems proof induct
  case (st_cons \<Delta> Z Zs M S \<Gamma>s \<Gamma>c \<sigma> w z Zl)
  show ?case proof (cases "x=z")
    case True
    have H\<sigma>: "\<Delta> \<Turnstile>' \<sigma>" using st_cons poly_ty_wf_type by auto
    have "\<not> (\<Gamma>s\<union>\<Gamma>c) defines z" using prems store_dom_not_in by auto
    hence "\<not> \<Gamma>s defines z" by (auto simp add: dom_def)
    moreover
    have "\<Gamma>s;\<Delta> \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s\<union>\<Gamma>c" _ \<Gamma>s] by auto
    ultimately
    have "\<Gamma>s,(z:\<sigma>);\<Delta> \<Turnstile> OK" using wf_env_cons H\<sigma> by auto
    hence Hteq: "\<tau> = \<sigma>" using True wf_env_def[of "\<Gamma>s,(z:\<sigma>)"] st_cons.prems
      by fast
    have "S,(z\<mapsto>w) \<Turnstile> Store" using st_cons wf_store_cons by auto
    hence Hweq: "v = w" using True wf_store_def[of "S,(z\<mapsto>w)"] st_cons.prems
      by fast
    have "un \<Gamma>s \<subseteq> un (\<Gamma>s,(x:\<tau>))" by (auto simp add: un_def)
    hence "un \<Gamma>c \<subseteq> un (\<Gamma>s,(x:\<tau>))" using st_cons by auto
    thus ?thesis using st_cons Hteq Hweq
      by (auto simp add: ty_qual_sanity ty_scope_sanity)
  next
    case False
    hence Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>s" using st_cons by auto
    have "(x\<mapsto>v) \<in> S" and "(x\<rhd>\<tau>) \<in> (\<Gamma>s \<union> \<Gamma>c)"
      using False st_cons by (auto simp add: without_def dom_def)
    thus "scopeV v = scopeT \<tau>" using st_cons by auto
  qed
next
  case (st_dummy \<Delta> Z Zl M S \<Gamma>s \<sigma> w z)
  show ?case proof (cases "x=z")
    case True
    hence "\<not> \<Gamma>s defines z" using st_dummy store_dom_not_in by (auto simp add: dom_def)
    moreover
    have "\<Gamma>s;\<Delta> \<Turnstile> OK" using prems storeTyping_wf_env by auto
    ultimately
    have "\<Gamma>s,(z:\<sigma>);\<Delta> \<Turnstile> OK" using wf_env_cons st_dummy by auto
    hence Hteq: "\<tau> = \<sigma>" using True wf_env_def[of "\<Gamma>s,(z:\<sigma>)"] st_dummy.prems
      by fast
    have "S,(z\<mapsto>w) \<Turnstile> Store" using st_dummy wf_store_cons storeTyping_wf_store by auto
    hence Hweq: "v = w" using True wf_store_def[of "S,(z\<mapsto>w)"] st_dummy.prems
      by fast
    thus ?thesis using Hweq Hteq st_dummy by auto
  next
    case False
    hence Hin: "(x\<rhd>\<tau>) \<in> \<Gamma>s" using st_dummy by auto
    have "(x\<mapsto>v) \<in> S" using False st_dummy by (auto simp add: without_def dom_def)
    thus ?thesis using Hin st_dummy by auto
  qed
qed auto

theorem progress:
  assumes "\<Gamma>e;\<Delta>;{};Ze \<turnstile> e : \<tau>"
  and "M;\<Delta>;Zs;Zl \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)"
  and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu> : M"
  and "\<Gamma>e \<subseteq> \<Gamma>s"
  and "Ze \<subseteq> Zl"
  and "Zl \<subseteq> Zs"
  and "free_scope_store S = dom Zs"
  shows "not_stuck e S \<mu>"
proof -
  { fix Me
    assume Ht: "\<Gamma>e; \<Delta>; Me; Ze \<turnstile> e : \<tau>"
      and "Me = {}"
      and "M;\<Delta>;Zs;Zl \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)"
      and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu> : M"
      and "\<Gamma>e \<subseteq> \<Gamma>s"
      and "Ze \<subseteq> Zl"
    moreover have "S \<Turnstile> Store" using prems storeTyping_wf_store by auto
    moreover have "\<mu> \<Turnstile> Mem" using prems memTyping_wf_mem by auto
    moreover have "\<forall>l. \<not> l free loc in e" using prems ty_no_locs by auto
    ultimately
    have "isVar e \<or> (\<exists>e' S' \<mu>'. S;\<mu>;e\<hookrightarrow>S';\<mu>';e')"
    proof (induct arbitrary: Zl)
      case (t_unit \<Delta> Z M \<Gamma> q \<pi>)
      have "value (TmUnit q \<pi>)" by auto
      moreover
      have "finite (dom S)"
        using prems storeTyping_wf_store[of _ _ _ _ S] wf_store_def[of S] dom_finite[of S] by auto
      then obtain z where "z \<notin> dom S"
        using prems fresh_exists[of "dom S"] by auto
      ultimately
      have H: "S;\<mu>;TmUnit q \<pi> \<hookrightarrow> S,(z\<mapsto>TmUnit q \<pi>);\<mu>;TmFreeVar z" using t_unit by auto
      show ?case apply (rule disjI2, rule exI, rule exI, rule exI) using H by auto
    next
      case t_var thus ?case by auto
    next
      case (t_abs \<Delta> X Z M \<Gamma> \<tau>1 \<tau>2 e_b q \<pi>)
      have "value (\<lambda>^q^\<pi>[\<tau>1]. e_b)" by auto
      moreover
      have "finite (dom S)"
        using prems storeTyping_wf_store[of _ _ _ _ S] wf_store_def[of S] dom_finite[of S] by auto
      then obtain z where "z \<notin> dom S"
        using prems fresh_exists[of "dom S"] by auto
      ultimately
      have H: "S;\<mu>;(\<lambda>^q^\<pi>[\<tau>1]. e_b) \<hookrightarrow> S,(z\<mapsto>(\<lambda>^q^\<pi>[\<tau>1]. e_b));\<mu>;TmFreeVar z"
        using t_abs by auto
      show ?case apply (rule disjI2, rule exI, rule exI, rule exI) using H by auto
    next
      case (t_app \<Delta> X Z1 Z2 Z Mz \<Gamma>1 \<Gamma>2 \<tau>1 \<tau>2 e1 e2 q)
      moreover
      have "\<forall>l. \<not> l free loc in e1" using t_app by auto
      ultimately
      have "isVar e1 \<or> (\<exists>e' S' \<mu>'. S;\<mu>;e1\<hookrightarrow>S';\<mu>';e')"
        by auto
      thus ?case proof
        assume V1: "isVar e1"
        then obtain x where Hx: "e1 = TmFreeVar x" by auto
        have "\<forall>l. \<not> l free loc in e2" using t_app by auto
        hence "isVar e2 \<or> (\<exists>e' S' \<mu>'. S;\<mu>;e2\<hookrightarrow>S';\<mu>';e')" using t_app by auto
        thus ?case proof
          assume V2: "isVar e2"
          then obtain y where Hy: "e2 = TmFreeVar y" by auto
          have "x \<in> dom \<Gamma>s" using Hx t_app by (auto simp add: dom_def)
          hence "x \<in> dom (\<Gamma>s\<union>\<Gamma>m)" by (auto simp add: dom_def)
          hence "x \<in> dom S" using t_app by (auto simp add: store_dom_in)
          moreover have Hstore: "S \<Turnstile> Store" by assumption
          ultimately
          obtain v where Hsin: "(x\<mapsto>v) \<in> S" and Hv: "value v"
            using store_ok_exists_value by auto
          moreover
          have "(x\<rhd>Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)) \<in> \<Gamma>1" using Hx t_app by auto
          hence Hxing: "(x\<rhd>Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)) \<in> \<Gamma>s" using t_app by auto
          moreover
          have "M;\<Delta>;Zs;Zl \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
          moreover
          have Htvalid: "validScope (scopeT (Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2))) Zl"
            using t_app ty_valid_scope by auto
          have "M;\<Delta>;Zs;Zl \<Turnstile> S:\<Gamma>s \<union> \<Gamma>m" by assumption
          hence "scopeV v = scopeT (Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2))"
            apply (rule storeTyping_scope_sanity) using Hsin Hxing by (auto simp add: Un_ac)
          hence "validScope (scopeV v) Zl" using Htvalid by auto
          ultimately
          obtain \<Gamma>' Ze' where Ht: "\<Gamma>';\<Delta>;M;Ze' \<turnstile> v : Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)"
            using storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x v "Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)"]
            by auto
          then obtain tm where Hv: "v = \<lambda>^q^ScBot[\<tau>1]. tm" using Hv by (cases v, auto)
          obtain S' where "S;x\<Down>S';\<lambda>^q^ScBot[\<tau>1].tm" using Hsin Hv by auto
          hence "S;\<mu>;TmFreeVar x \<cdot> TmFreeVar y \<hookrightarrow> S';\<mu>;freshen_tm y tm" using t_app by blast
          hence H: "S;\<mu>;e1\<cdot>e2 \<hookrightarrow> S';\<mu>;freshen_tm y tm" using Hx Hy by auto
          show ?thesis apply (rule disjI2, rule exI[of _ "freshen_tm y tm"], rule exI[of _ S'],
            rule exI[of _ \<mu>])
            using H by auto
        next
          assume "\<exists>e' S' \<mu>'. S;\<mu>;e2\<hookrightarrow>S';\<mu>';e'"
          then obtain e2' S' \<mu>' where H2: "S;\<mu>;e2 \<hookrightarrow> S';\<mu>';e2'" by auto
          hence H: "S;\<mu>;(e1 \<cdot> e2) \<hookrightarrow> S';\<mu>';(e1 \<cdot> e2')" using V1 H2 by auto
          show ?thesis apply (rule disjI2, rule exI[of _ "e1\<cdot>e2'"], rule exI, rule exI)
            using H by auto
        qed
      next
        assume "\<exists>e' S' \<mu>'. S;\<mu>;e1\<hookrightarrow>S';\<mu>';e'"
        then obtain e1' S' \<mu>' where H1: "S;\<mu>;e1 \<hookrightarrow> S';\<mu>';e1'" by auto
        hence H: "S;\<mu>;(e1 \<cdot> e2) \<hookrightarrow> S';\<mu>';(e1' \<cdot> e2)" using H1 by auto
        show ?thesis apply (rule disjI2, rule exI[of _ "e1'\<cdot>e2"], rule exI, rule exI)
          using H by auto
      qed
    next
      case (t_let \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> M Z1 Z2 Z e e1 e2 \<tau> \<tau>1 \<tau>2)
      moreover
      have "\<forall>l. \<not> l free loc in e" using t_let by auto
      ultimately
      have H: "isVar e \<or> (\<exists>e' S' \<mu>'. S;\<mu>;e\<hookrightarrow>S';\<mu>';e')"
        by auto
      show ?case using H proof
        assume V1: "isVar e"
        then obtain x where Hx: "e = TmFreeVar x" by auto
        have "x \<in> dom \<Gamma>s" using Hx t_let by (auto simp add: dom_def)
        hence "x \<in> dom (\<Gamma>s\<union>\<Gamma>m)" by (auto simp add: dom_def)
        hence "x \<in> dom S" using t_let by (auto simp add: store_dom_in)
        moreover have Hstore: "S \<Turnstile> Store" by assumption
        ultimately
        obtain v where Hsin: "(x\<mapsto>v) \<in> S" and Hv: "value v"
          using store_ok_exists_value by blast
        then obtain S' where Hs': "S;x\<Down>S';v" by auto
        have Hfin: "finite S" using Hstore by (auto simp add: wf_store_def)
        have "finite (free_scope_store S)" proof -
          let ?f = "%m. case m of (Pair x b) \<Rightarrow>
            case b of (StTerm v) \<Rightarrow>
                    set (free_scope_scope (scopeV v))"
          let ?A = "{F. \<exists>x v. F = set (free_scope_scope (scopeV v)) \<and> (x\<mapsto>v) \<in> S}"
          have "finite (\<Union>{F. \<exists>x v. F = set (free_scope_scope (scopeV v)) \<and> (x\<mapsto>v) \<in> S})"
            using Hfin finite_surj[of S ?A ?f] by force
          thus ?thesis using Hfin finite_Union by (auto simp add: free_scope_store_def)
        qed
        then obtain \<rho> where Hr: "\<not> \<rho> free sc in st S" using fresh_exists[of "free_scope_store S"]
          by auto
        have H:
          "S;\<mu>; let! TmFreeVar x e1 in e2 \<hookrightarrow>
          S',(x\<mapsto>convV \<rho> v);\<mu>;let$ \<rho> (TmFreeVar x) (freshen_tm x (freshen_sc \<rho> e1)) in e2"
          apply (rule Semantics.e_let_p'[of S \<mu> \<rho> x S' v e1 e2]) using Hs' Hr Hx by auto
        show ?thesis apply (rule disjI2,
          rule exI[of _ "let$ \<rho> (TmFreeVar x) (freshen_tm x (freshen_sc \<rho> e1)) in e2"],
          rule exI, rule exI)
          using H Hx by auto
      next
        assume "\<exists>e' S' \<mu>'. S;\<mu>;e\<hookrightarrow>S';\<mu>';e'"
        then obtain e' S' \<mu>' where H1: "S;\<mu>;e \<hookrightarrow> S';\<mu>';e'" by auto
        have H: "S;\<mu>;let! e e1 in e2 \<hookrightarrow> S';\<mu>';let! e' e1 in e2" using H1 by auto
        show ?thesis apply (rule disjI2, rule exI[of _ "let! e' e1 in e2"], rule exI, rule exI)
          using H by auto
      qed
    next
      case (t_let' \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> Ma Z1 Z2 Z \<rho> x e1 e2 \<tau> \<tau>1 \<tau>2 Zf)
      have "\<forall>l. \<not> l free loc in e1" using t_let' by auto
      moreover
      have Hstore: "S \<Turnstile> Store" using storeTyping_wf_store t_let' by auto
      have "x \<in> dom \<Gamma>s" using t_let' by (auto simp add: dom_def)
      hence "x \<in> dom (\<Gamma>s\<union>\<Gamma>m)" by (auto simp add: dom_def)
      hence "x \<in> dom S" using t_let' by (auto simp add: store_dom_in)
      then obtain v where Hsin: "(x\<mapsto>v) \<in> S" and Hv: "value v"
        using Hstore store_ok_exists_value by blast
      then obtain S' where Hs': "S;x\<Down>S';v" by auto
      have "(x\<rhd>convT \<rho> \<tau>) \<in> \<Gamma>" using t_let'.hyps by auto
      hence Hxing: "(x\<rhd>convT \<rho> \<tau>) \<in> \<Gamma>s" using t_let'.prems by auto
      hence Hxing': "(x\<rhd>convT \<rho> \<tau>) \<in> \<Gamma>s\<union>\<Gamma>m" using t_let'.prems by auto
      have Htvalid': "validScope (scopeT (convT \<rho> \<tau>)) Z" using ty_valid_scope t_let' by auto
      hence Htvalid: "validScope (scopeT (convT \<rho> \<tau>)) Zl"
        apply (rule valid_scope_subset) using t_let' by auto
      have Hrinz: "(\<rho>, EmptyScope) \<in> Z" using Htvalid'
      proof (cases \<tau>, cases "convT \<rho> \<tau>",  auto simp add: dom_def)
        fix b ba
        assume "(\<rho>, b) \<in> Z"
        and "(\<rho>, ba) \<in> Z"
        thus ?thesis by (cases b, cases ba, auto)
      qed
      have "M;\<Delta>;Zs;Zl \<Turnstile> S:\<Gamma>s \<union> \<Gamma>m" by assumption
      hence "scopeV v = scopeT (convT \<rho> \<tau>)"
        apply (rule storeTyping_scope_sanity) using Hsin Hxing by (auto simp add: Un_ac)
      hence Hvalid: "validScope (scopeV v) Zl" using Htvalid by (cases \<tau>, auto)
      then obtain \<Gamma>' Z' where Htm: "\<Gamma>';\<Delta>;M;Z' \<turnstile> v : convT \<rho> \<tau>"
        and Hqual: "qualV v = qualT (convT \<rho> \<tau>)" and Hscope: "scopeV v = scopeT (convT \<rho> \<tau>)"
        using t_let'.prems Hsin Hxing' Hvalid storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x v]
        by best
      have "\<Gamma>1,(x:convT \<rho> \<tau>) \<subseteq> \<Gamma>s" using t_let'.prems Hxing by auto
      moreover
      hence Hsub: "Z1,(\<rho>:\<section>) \<subseteq> Zl" using Hrinz t_let'.prems by auto
      hence "Zf \<subseteq> Zl" using Hsub t_let' by auto
      ultimately
      have H: "isVar e1 \<or> (\<exists>e' S' \<mu>'. S;\<mu>; e1\<hookrightarrow>S';\<mu>';e')"
        using t_let' by auto
      show ?case using H proof
        assume "\<exists>e1' S' \<mu>'. S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1'"
        then obtain e1' S' \<mu>'
          where Heval: "S;\<mu>;e1 \<hookrightarrow> S';\<mu>'; e1'"
          by auto
        have H: "S;\<mu>;let$ \<rho> (TmFreeVar x) e1 in e2 \<hookrightarrow> S';\<mu>';let$ \<rho> TmFreeVar x e1' in e2"
          using Heval by auto
        show ?thesis apply (rule disjI2, rule exI, rule exI, rule exI)
          using H by auto
      next
        assume V1: "isVar e1"
        then obtain y where Hy: "e1 = TmFreeVar y" by auto
        have "finite S" using Hstore by (auto simp add: wf_store_def)
        then obtain z where Hz: "\<not> S defines z" using dom_finite fresh_exists[of "dom S"] by auto
        have Hqual: "qualV v = qU" and Hscope: "scopeV v = ScFree \<rho>"
          using Hqual Hscope by (cases \<tau>, cases "qualT (convT \<rho> \<tau>)", auto)+
        have H: "S;\<mu>; let$ \<rho> (TmFreeVar x) (TmFreeVar y) in e2 \<hookrightarrow>
          S,(z\<mapsto>convRevV v);\<mu>; vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm z e2)"
          using Semantics.e_let_c[of S \<mu> x S' v z \<rho> y e2] Hstore Hs' Hz t_let'.prems
          by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>; let$ \<rho> TmFreeVar x e1 in e2 \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H Hy by auto
        thus ?thesis by auto
      qed
    next
      case (t_new \<Delta> Z Ma \<Gamma> \<tau> e)
      moreover have "\<forall>l. \<not> l free loc in e" using t_new by auto
      ultimately have "isVar e \<or> makes_progress e S \<mu>" using t_new by auto
      thus ?case proof
        assume "isVar e"
        obtain x where Hx: "e = TmFreeVar x" by (rule isVar_var_elim)
        have "finite (dom S)"
          using prems storeTyping_wf_store[of _ _  _ _ S] wf_store_def[of S] dom_finite[of S] by auto
        then obtain z where Hz: "\<not> S defines z" using fresh_exists[of "dom S"]
          by (auto simp add: dom_finite)
        have Hfin1: "finite (dom \<mu>)"
          using memTyping_wf_mem[of _ _ \<mu>] wf_mem_def[of \<mu>] dom_finite[of \<mu>] prems by auto
        have Hfin: "finite S" using t_new.prems by (auto simp add: wf_store_def)
        have "finite (free_loc_store S)" proof -
          let ?f = "%m. case m of (Pair x b) \<Rightarrow>
            case b of (StTerm v) \<Rightarrow>
                    set (free_loc v)"
          let ?A = "{F. \<exists>x v. F = set (free_loc v) \<and> (x\<mapsto>v) \<in> S}"
          have "finite (\<Union>{F. \<exists>x v. F = set (free_loc v) \<and> (x\<mapsto>v) \<in> S})"
            using Hfin finite_surj[of S ?A ?f] by force
          thus ?thesis using Hfin finite_Union by (auto simp add: free_loc_store_def)
        qed
        then obtain l where Hl: "\<not> \<mu> defines l" and Hl': "\<not> l free loc in st S"
          using fresh_exists[of "dom \<mu> \<union> free_loc_store S"] Hfin1
          by auto
        have "\<Gamma>s defines x" using prems by (auto simp add: dom_def)
        hence "(\<Gamma>s\<union>\<Gamma>m) defines x" by (auto simp add: dom_def)
        hence "S defines x" using prems by (auto simp add: store_dom_in)
        moreover have "S \<Turnstile> Store" by assumption
        ultimately
        obtain v where "(x\<mapsto>v) \<in> S"
          using prems store_ok_exists_value[of S x] by auto
        hence H: "S;\<mu>;TmNew (TmFreeVar x) \<hookrightarrow>
               S,(z\<mapsto>TmLoc qL ScBot l);\<mu>,(l\<mapsto>TmFreeVar x);TmFreeVar z"
          using t_new Hx Hz Hl Hl' e_new_c[of S \<mu> l z] by blast
        have "\<exists>e' S' \<mu>'. S;\<mu>; TmNew (TmFreeVar x) \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?case using Hx by auto
      next
        assume "makes_progress e S \<mu>"
        then obtain e' S' \<mu>' where "S;\<mu>;e \<hookrightarrow> S';\<mu>';e'" by auto
        hence H: "S;\<mu>;TmNew e \<hookrightarrow> S';\<mu>';TmNew e'" by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>; TmNew e \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?case by auto
      qed
    next
      case (t_assign \<Delta> Z1 Z2 Ma \<Gamma>1 \<Gamma>2 \<tau> e1 e2 \<pi>)
      moreover have "\<forall>l. \<not> l free loc in e1" using t_assign by auto
      ultimately have "isVar e1 \<or> makes_progress e1 S \<mu>" using t_assign by auto
      thus ?case proof
        assume "makes_progress e1 S \<mu>"
        then obtain e1' S' \<mu>' where "S;\<mu>;e1 \<hookrightarrow> S';\<mu>';e1'" by auto
        hence H: "S;\<mu>;e1 := e2 \<hookrightarrow> S';\<mu>';e1' := e2" by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>; e1:=e2 \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?thesis by auto
      next
        assume "isVar e1"
        then obtain x where Hx: "e1 = TmFreeVar x" by (rule isVar_var_elim)
        moreover have "\<forall>l. \<not> l free loc in e2" using t_assign by auto
        ultimately have "isVar e2 \<or> makes_progress e2 S \<mu>" using t_assign by auto
        thus ?case proof
          assume "makes_progress e2 S \<mu>"
          then obtain e2' S' \<mu>' where "S;\<mu>;e2 \<hookrightarrow> S';\<mu>';e2'" by auto
          hence H: "S;\<mu>;(TmFreeVar x) := e2 \<hookrightarrow> S';\<mu>';(TmFreeVar x) := e2'" by auto
          have "\<exists>e' S' \<mu>'. S;\<mu>;TmFreeVar x := e2 \<hookrightarrow> S';\<mu>';e'"
            apply (rule exI)+ using H by auto
          thus ?case using Hx by auto
        next
          assume "isVar e2"
          obtain y where Hy: "e2 = TmFreeVar y" by (rule isVar_var_elim)
          have Htvalid': "validScope (scopeT (Qual qU \<pi> (TyRef \<tau>))) Z1"
            apply (rule ty_valid_scope) using t_assign by auto
          hence Htvalid: "validScope (scopeT (Qual qU \<pi> (TyRef \<tau>))) Zl"
            apply (rule valid_scope_subset) using t_assign by auto
          have Hstore: "S \<Turnstile> Store" by assumption
          moreover
          have "x \<in> dom \<Gamma>s" using Hx t_assign by (auto simp add: dom_def)
          hence "x \<in> dom (\<Gamma>s\<union>\<Gamma>m)" by (auto simp add: dom_def)
          hence "x \<in> dom S" using t_assign store_dom_in by auto
          ultimately
          obtain v where Hsin: "(x\<mapsto>v) \<in> S" and Hv: "value v"
            using store_ok_exists_value by auto
          then obtain S' where Hd1: "S;x\<Down>S';v" by auto
          moreover
          have Hxing: "(x\<rhd>Qual qU \<pi> (TyRef \<tau>)) \<in> \<Gamma>s" using Hx t_assign by auto
          moreover
          have "M;\<Delta>;Zs;Zl \<Turnstile> S : \<Gamma>s\<union>\<Gamma>m" by assumption
          moreover
          hence "scopeV v = scopeT (Qual qU \<pi> (TyRef \<tau>))"
            apply (rule storeTyping_scope_sanity) using Hxing Hsin by auto
          hence "validScope (scopeV v) Zl" using Htvalid by auto
          ultimately
          obtain \<Gamma>' Z' where "\<Gamma>';\<Delta>;M;Z' \<turnstile> v : Qual qU \<pi> (TyRef \<tau>)"
            using storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x v "Qual qU \<pi> (TyRef \<tau>)"] Hsin
            by auto
          then obtain l where Hv: "v = TmLoc qU \<pi> l"
            using Hv by (cases v, auto)
          have H:"S;\<mu>;TmAssign (TmFreeVar x) (TmFreeVar y) \<hookrightarrow>
                            S';update \<mu> l (TmFreeVar y); TmUnit qU ScBot"
            using t_assign Hd1 Hv by auto
          have "\<exists>e' S' \<mu>'. S;\<mu>;TmFreeVar x := TmFreeVar y \<hookrightarrow> S';\<mu>';e'"
            apply (rule exI)+ using H by auto
          thus ?thesis using Hx Hy by auto
        qed
      qed
    next
      case (t_swap \<Delta> Z1 Z2 Ma \<Gamma>1 \<Gamma>2 \<tau> e1 e2 \<pi>)
      moreover have "\<forall>l. \<not> l free loc in e1" using t_swap by auto
      ultimately have "isVar e1 \<or> makes_progress e1 S \<mu>" using t_swap by auto
      thus ?case proof
        assume "makes_progress e1 S \<mu>"
        then obtain e1' S' \<mu>' where "S;\<mu>;e1 \<hookrightarrow> S';\<mu>';e1'" by auto
        hence H: "S;\<mu>;e1 :=: e2 \<hookrightarrow> S';\<mu>';e1' :=: e2" by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>; e1 :=: e2 \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?case by auto
      next
        assume "isVar e1"
        then obtain x where Hx: "e1 = TmFreeVar x" by (rule isVar_var_elim)
        moreover have "\<forall>l. \<not> l free loc in e2" using t_swap by auto
        ultimately have "isVar e2 \<or> makes_progress e2 S \<mu>" using t_swap by auto
        thus ?case proof
          assume "makes_progress e2 S \<mu>"
          then obtain e2' S' \<mu>' where "S;\<mu>;e2 \<hookrightarrow> S';\<mu>';e2'" by auto
          hence H: "S;\<mu>;(TmFreeVar x) :=: e2 \<hookrightarrow> S';\<mu>';(TmFreeVar x) :=: e2'"
            by auto
          have "\<exists>e' S' \<mu>'. S;\<mu>;TmFreeVar x :=: e2 \<hookrightarrow> S';\<mu>';e'"
            apply (rule exI)+ using H by auto
          thus ?case using Hx by auto
        next
          assume "isVar e2"
          obtain y where Hy: "e2 = TmFreeVar y" by (rule isVar_var_elim)
          have Htvalid': "validScope (scopeT (Qual qU \<pi> (TyRef \<tau>))) Z1"
            apply (rule ty_valid_scope) using t_swap by auto
          hence Htvalid: "validScope (scopeT (Qual qU \<pi> (TyRef \<tau>))) Zl"
            apply (rule valid_scope_subset) using t_swap by auto
          have Hstore: "S \<Turnstile> Store" by assumption
          moreover
          have "x \<in> dom \<Gamma>s" using Hx t_swap by (auto simp add: dom_def)
          hence "x \<in> dom (\<Gamma>s\<union>\<Gamma>m)" by (auto simp add: dom_def)
          hence "x \<in> dom S" using t_swap store_dom_in by auto
          ultimately
          obtain v where Hsin: "(x\<mapsto>v) \<in> S" and Hv: "value v"
            using store_ok_exists_value by auto
          then obtain S' where Hd1: "S;x\<Down>S';v" by auto
          moreover
          have Hxing: "(x\<rhd>Qual qU \<pi> (TyRef \<tau>)) \<in> \<Gamma>s" using Hx t_swap by auto
          moreover
          have "M;\<Delta>;Zs;Zl \<Turnstile> S : \<Gamma>s\<union>\<Gamma>m" by assumption
          hence "scopeV v = scopeT (Qual qU \<pi> (TyRef \<tau>))"
            apply (rule storeTyping_scope_sanity) using Hxing Hsin by auto
          hence Hvalid: "validScope (scopeV v) Zl" using Htvalid by auto
          ultimately
          obtain \<Gamma>' Z' where "\<Gamma>';\<Delta>;M;Z' \<turnstile> v : Qual qU \<pi> (TyRef \<tau>)"
            using storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x v "Qual qU \<pi> (TyRef \<tau>)"] Hsin t_swap
            by auto
          then obtain l where Hv: "v = TmLoc qU \<pi> l" and HlinM: "l\<rhd>\<tau> \<in> M"
            using Hv by (cases v, auto)
          hence "\<mu> defines l" using t_swap mem_dom_in[of \<Gamma>m \<Delta> \<mu> M l] by (auto simp add: dom_def)
          then obtain z where "(l\<mapsto>TmFreeVar z) \<in> \<mu>"
            using t_swap mem_ok_exists_var[of \<mu> l] by auto
          hence H: "S;\<mu>;(TmFreeVar x) :=: (TmFreeVar y) \<hookrightarrow>
                            S';update \<mu> l (TmFreeVar y); TmFreeVar z"
            using t_swap Hd1 Hv by auto
          have "\<exists>e' S' \<mu>'. S;\<mu>;TmFreeVar x :=: TmFreeVar y \<hookrightarrow> S';\<mu>';e'"
            apply (rule exI)+ using H by auto
          thus ?thesis using Hx Hy by auto
        qed
      qed
    next
      case (t_deref \<pi> \<Delta> Z Z\<tau> Ma \<Gamma> \<tau> e)
      moreover have "\<forall>l. \<not> l free loc in e" using t_deref by auto
      ultimately have "isVar e \<or> makes_progress e S \<mu>" using t_deref by auto
      thus ?case proof
        assume "isVar e"
        then obtain x where Hx: "e = TmFreeVar x" by auto
        have Htvalid': "validScope (scopeT (Qual qU \<pi> (TyRef \<tau>))) Z"
          apply (rule ty_valid_scope) using t_deref by auto
        hence Htvalid: "validScope (scopeT (Qual qU \<pi> (TyRef \<tau>))) Zl"
          apply (rule valid_scope_subset) using t_deref by auto
        have "x \<in> dom \<Gamma>s" using Hx t_deref by (auto simp add: dom_def)
        hence "x \<in> dom (\<Gamma>s\<union>\<Gamma>m)" by (auto simp add: dom_def)
        hence "x \<in> dom S" using t_deref store_dom_in by auto
        then obtain v where Hsin: "(x\<mapsto>v) \<in> S" and Hval: "value v"
          using store_ok_exists_value[of S x] t_deref by auto
        then obtain S' where Hs': "S;x\<Down>S';v" by auto
        have Hgin: "(x\<rhd>Qual qU \<pi> (TyRef \<tau>)) \<in> \<Gamma>s" using Hx t_deref by auto
        have "M;\<Delta>;Zs;Zl \<Turnstile> S : \<Gamma>s\<union>\<Gamma>m" by assumption
        hence "scopeV v = scopeT (Qual qU \<pi> (TyRef \<tau>))"
          apply (rule storeTyping_scope_sanity) using Hgin Hsin by auto
        hence Hvalid: "validScope (scopeV v) Zl" using Htvalid by auto
        then obtain \<Gamma>' Z' where Htm: "\<Gamma>';\<Delta>;M;Z' \<turnstile> v : Qual qU \<pi> (TyRef \<tau>)"
          using t_deref Hsin Hgin
            storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x v "Qual qU \<pi> (TyRef \<tau>)"]
          by auto
        then obtain l where Hv: "v = TmLoc qU \<pi> l"
          using Hval by (cases v, auto)
        hence HlinM: "(l\<rhd>\<tau>) \<in> M" using Htm by auto
        hence "l \<in> dom \<mu>" using mem_dom_in[of \<Gamma>m \<Delta> \<mu> M l] t_deref by (auto simp add: dom_def)
        then obtain z where "(l\<mapsto>TmFreeVar z) \<in> \<mu>"
          using t_deref mem_ok_exists_var[of \<mu> l] by auto
        hence H: "S;\<mu>;TmDeref (TmFreeVar x) \<hookrightarrow> S';\<mu>;TmFreeVar z"
          using t_deref Hs' Hv by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>;TmDeref (TmFreeVar x) \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?thesis using Hx by auto
      next
        assume "makes_progress e S \<mu>"
        then obtain e' S' \<mu>' where "S;\<mu>;e \<hookrightarrow> S';\<mu>';e'" by auto
        hence H: "S;\<mu>;TmDeref e \<hookrightarrow> S';\<mu>';TmDeref e'" by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>;TmDeref e \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?thesis by auto
      qed
    next
      case (t_free \<Delta> Z Z\<tau> Ma \<Gamma> \<tau> e)
      moreover have "\<forall>l. \<not> l free loc in e" using t_free by auto
      ultimately have "isVar e \<or> makes_progress e S \<mu>" using t_free by auto
      thus ?case proof
        assume "isVar e"
        then obtain x where Hx: "e = TmFreeVar x" by auto
        have Htvalid': "validScope (scopeT (Qual qL ScBot (TyRef \<tau>))) Z"
          apply (rule ty_valid_scope) using t_free by auto
        hence Htvalid: "validScope (scopeT (Qual qL ScBot (TyRef \<tau>))) Zl"
          apply (rule valid_scope_subset) using t_free by auto
        have "x \<in> dom \<Gamma>s" using Hx t_free by (auto simp add: dom_def)
        hence "x \<in> dom (\<Gamma>s\<union>\<Gamma>m)" by (auto simp add: dom_def)
        hence "x \<in> dom S" using t_free store_dom_in by auto
        then obtain v where Hsin: "(x\<mapsto>v) \<in> S" and Hval: "value v"
          using store_ok_exists_value[of S x] t_free by auto
        then obtain S' where Hs': "S;x\<Down>S';v" by auto
        have Hgin: "(x\<rhd>Qual qL ScBot (TyRef \<tau>)) \<in> \<Gamma>s" using Hx t_free by auto
        have "M;\<Delta>;Zs;Zl \<Turnstile> S : \<Gamma>s\<union>\<Gamma>m" by assumption
        hence "scopeV v = scopeT (Qual qL ScBot (TyRef \<tau>))"
          apply (rule storeTyping_scope_sanity) using Hgin Hsin by auto
        hence Hvalid: "validScope (scopeV v) Zl" using Htvalid by auto
        then obtain \<Gamma>' Z' where Htm: "\<Gamma>';\<Delta>;M;Z' \<turnstile> v : Qual qL ScBot (TyRef \<tau>)"
          using t_free Hsin Hgin storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x v] by best
        then obtain l where Hv: "v = TmLoc qL ScBot l"
          using Hval by (cases v, auto)
        hence HlinM: "l\<rhd>\<tau> \<in> M" using Htm by auto
        hence "l \<in> dom \<mu>" using mem_dom_in[of \<Gamma>m \<Delta> \<mu> M l] t_free by (auto simp add: dom_def)
        then obtain y where "(l\<mapsto>TmFreeVar y) \<in> \<mu>"
          using t_free mem_ok_exists_var[of \<mu> l] by auto
        hence H: "S;\<mu>;TmFree (TmFreeVar x) \<hookrightarrow> S';\<mu>\<bar>l;TmFreeVar y"
          using t_free Hs' Hv e_free_c[of S \<mu>] by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>;TmFree (TmFreeVar x) \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?thesis using Hx by auto
      next
        assume "makes_progress e S \<mu>"
        then obtain e' S' \<mu>' where "S;\<mu>;e \<hookrightarrow> S';\<mu>';e'" by auto
        hence H: "S;\<mu>;TmFree e \<hookrightarrow> S';\<mu>';TmFree e'" by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>;TmFree e \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?thesis by auto
      qed
    next
      case (t_loc \<Delta> Z Ma \<Gamma> \<tau> l \<pi> q)
      have "value (TmLoc q \<pi> l)" by auto
      moreover
      have "finite (dom S)"
        using prems storeTyping_wf_store[of _ _ _ _ S] wf_store_def[of S] dom_finite[of S] by auto
      then obtain z where "z \<notin> dom S" using fresh_exists[of "dom S"] by auto
      ultimately
      have H: "S;\<mu>;(TmLoc q \<pi> l) \<hookrightarrow> S,(z\<mapsto>(TmLoc q \<pi> l));\<mu>;TmFreeVar z"
        using t_loc by auto
      have "\<exists>e' S' \<mu>'. S;\<mu>;TmLoc q \<pi> l \<hookrightarrow> S';\<mu>';e'"
        apply (rule exI)+ using H by auto
      thus ?case by auto
    next
      case (t_poly \<Delta> Z Z' M \<Gamma> \<tau> v q \<pi>)
      hence "value (\<Lambda>^q^\<pi>[*].v)" by auto
      moreover
      have "finite (dom S)"
        using prems storeTyping_wf_store[of _ _ _ _ S] wf_store_def[of S] dom_finite[of S] by auto
      then obtain z where "z \<notin> dom S" using fresh_exists[of "dom S"] by auto
      ultimately
      have H: "S;\<mu>;(\<Lambda>^q^\<pi>[*].v) \<hookrightarrow> S,(z\<mapsto>(\<Lambda>^q^\<pi>[*].v));\<mu>;TmFreeVar z"
        using t_poly by auto
      have "\<exists>e' S' \<mu>'. S;\<mu>;\<Lambda>^q^\<pi>[*].v \<hookrightarrow> S';\<mu>';e'"
        apply (rule exI)+ using H by auto
      thus ?case by auto
    next
      case (t_ptyapp \<Delta> Z Z\<phi> Ma \<Gamma> \<tau> \<phi> e q X)
      moreover have "\<forall>l. \<not> l free loc in e" using t_ptyapp by auto
      ultimately have "isVar e \<or> makes_progress e S \<mu>" using t_ptyapp by auto
      thus ?case proof
        assume "isVar e"
        then obtain x where Hx: "e = TmFreeVar x" by auto
        have Htvalid': "validScope (scopeT (Qual q ScBot (\<forall>.\<tau>^X))) Z"
          apply (rule ty_valid_scope) using t_ptyapp by auto
        hence Htvalid: "validScope (scopeT (Qual q ScBot (\<forall>.\<tau>^X))) Zl"
          apply (rule valid_scope_subset) using t_ptyapp by auto
        have "x \<in> dom \<Gamma>s" using Hx t_ptyapp by (auto simp add: dom_def)
        hence "x \<in> dom (\<Gamma>s\<union>\<Gamma>m)" by (auto simp add: dom_def)
        hence "x \<in> dom S" using t_ptyapp store_dom_in by auto
        then obtain v where Hsin: "(x\<mapsto>v) \<in> S" and Hval: "value v"
          using store_ok_exists_value[of S x] t_ptyapp by auto
        then obtain S' where Hs': "S;x\<Down>S';v" by auto
        have Hgin: "(x\<rhd>Qual q ScBot (\<forall>.\<tau>^X)) \<in> \<Gamma>s" using Hx t_ptyapp by auto
        have "M;\<Delta>;Zs;Zl \<Turnstile> S : \<Gamma>s\<union>\<Gamma>m" by assumption
        hence "scopeV v = scopeT (Qual q ScBot (\<forall>.\<tau>^X))"
          apply (rule storeTyping_scope_sanity) using Hgin Hsin by auto
        hence Hvalid: "validScope (scopeV v) Zl" using Htvalid by auto
        then obtain \<Gamma>' Z' where Htm: "\<Gamma>';\<Delta>;M;Z' \<turnstile> v : Qual q ScBot (\<forall>.\<tau>^X)"
          using t_ptyapp Hsin Hgin
            storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x v "Qual q ScBot (\<forall>.\<tau>^X)"]
          by auto
        then obtain ef where Hv: "v = \<Lambda>^q^ScBot[*].ef"
          using Hval by (cases v, auto)
        hence H: "S;\<mu>;(TmFreeVar x)\<langle>\<phi>\<rangle> \<hookrightarrow> S';\<mu>;vsubst_tmty \<phi> 0 ef"
          using t_ptyapp Hs' Hv by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>;(TmFreeVar x)\<langle>\<phi>\<rangle> \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?thesis using Hx by auto
      next
        assume "makes_progress e S \<mu>"
        then obtain e' S' \<mu>' where "S;\<mu>;e \<hookrightarrow> S';\<mu>';e'" by auto
        hence H: "S;\<mu>;TmDeref e \<hookrightarrow> S';\<mu>';TmDeref e'" by auto
        have "\<exists>e' S' \<mu>'. S;\<mu>;e\<langle>\<phi>\<rangle> \<hookrightarrow> S';\<mu>';e'"
          apply (rule exI)+ using H by auto
        thus ?thesis by auto
      qed
    qed
  }
  with prems show ?thesis by auto
qed

theorem preservation:
  assumes "\<Gamma>e;\<Delta>;Mz;Ze \<turnstile> e : \<tau>"
  and "Mz = {}"
  and "M;\<Delta>;Zs;Zl \<Turnstile> S : (\<Gamma>s \<union> \<Gamma>m)"
  and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu> : M"
  and "S;\<mu>;e \<hookrightarrow> S';\<mu>';e'"
  and "\<Gamma>e \<union> \<Gamma>r \<subseteq> \<Gamma>s"
  and "\<Gamma>e \<sim> \<Gamma>r"
  and "\<Gamma>e \<union> \<Gamma>r;\<Delta> \<Turnstile> OK"
  and "\<Gamma>s \<sim> \<Gamma>m"
  and "Ze \<subseteq> Zl" 
  and "Zr \<subseteq> Zl"
  and "Zl \<subseteq> Zs"
  and "free_scope_store S = dom Zs"
  and "INV2 (filterStore S Zl)"
  and "INV3 S"
  and "INV4 (filterStore S Zl)" 
  and "INV5 S" 
  and "INV6 e Zr"
  and "INV9 e"
  shows "\<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'.
          (M';\<Delta>;Zs';Zl' \<Turnstile> S' : (\<Gamma>s' \<union> \<Gamma>m')) \<and>
          (\<Gamma>m';\<Delta> \<Turnstile>' \<mu>' : M') \<and>
          (\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e' : \<tau>) \<and>
          (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
          (\<Gamma>e' \<sim> \<Gamma>r) \<and>
          (\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK) \<and>
          (\<Gamma>s' \<sim> \<Gamma>m') \<and>
          (Ze' \<subseteq> Zl') \<and> (Zl' \<subseteq> Zs') \<and> (Zs \<subseteq> Zs') \<and> (Zr \<subseteq> Zl') \<and>
          (free_scope_store S' = dom Zs') \<and>
          INV2 (filterStore S' Zl') \<and> INV3 S' \<and>
          INV4 (filterStore S' Zl') \<and> INV5 S' \<and> INV6 e' Zr \<and>
          INV7 e e' Zs \<and> INV8 Ze' Ze e' \<and> INV9 e'"
proof -
have "\<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK" using prems storeTyping_wf_env by auto
from prems this show ?thesis 
proof (induct arbitrary: e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M Zs Zr);

      (********************* application cases *********************)
  case (t_app \<Delta> X Z1 Z2 Z Mz \<Gamma>1 \<Gamma>2 \<tau>1 \<tau>2 e1 e2 q e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M Zs)
  have "S;\<mu>;TmApp e1 e2 \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_app_p1 Sa Sa' \<mu>a \<mu>a' e1a e1' e2a)
    from t_app
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r Zs Zr.
          \<lbrakk> Mz = {}; M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M;  S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1';
           \<Gamma>1 \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma>1 \<sim> \<Gamma>r; \<Gamma>1 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; 
           Z1 \<subseteq> Zl; Zr \<subseteq> Zl; Zl \<subseteq> Zs;
           free_scope_store S = dom Zs; INV2 (filterStore S Zl); INV3 S;
           INV4 (filterStore S Zl); INV5 S; INV6 e1 Zr; INV9 e1; 
           \<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK \<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'.
            (M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e1' : Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (Ze' \<subseteq> Zl') \<and> (Zl' \<subseteq> Zs') \<and> (Zs \<subseteq> Zs') \<and> (Zr \<subseteq> Zl') \<and>
            (free_scope_store S' = Environ.dom Zs') \<and>
            (INV2 (filterStore S' Zl')) \<and> (INV3 S') \<and>
            (INV4 (filterStore S' Zl')) \<and> (INV5 S') \<and>
            (INV6 e1' Zr) \<and> (INV7 e1 e1' Zs) \<and> (INV8 Ze' Z1 e1') \<and> (INV9 e1')"
      using t_app.hyps by auto
    have Hsubeq: "\<Gamma>1 \<union> (\<Gamma>r\<union>\<Gamma>2) = \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    obtain \<Gamma>1' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'
      where Hg1: "M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>1';\<Delta>;{};Ze' \<turnstile> e1' : Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)"
      and H4_1: "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2) \<subseteq> \<Gamma>s'"
      and H5_1_1: "\<Gamma>1' \<sim> (\<Gamma>r\<union>\<Gamma>2)"
      and H6_1: "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2);\<Delta> \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg10: "INV2 (filterStore S' Zl')"
      and Hg11: "INV3 S'"
      and Hg12: "Zs \<subseteq> Zs'"
      and Hg13: "Ze' \<subseteq> Zl'"
      and Hg14: "free_scope_store S' = dom Zs'"
      and Hg15g: "Zl' \<subseteq> Zs'"
      and Hg15: "Zr\<union>Z2\<union> filterScFreeX X \<union> Z \<subseteq> Zl'"
      and Hg16: "INV4 (filterStore S' Zl')"
      and Hg17: "INV5 S'"
      and Hg18: "INV6 e1' (Zr\<union>Z2\<union> filterScFreeX X \<union> Z)"
      and Hg19: "INV7 e1 e1' Zs"
      and Hg20: "INV8 Ze' Z1 e1'"
      and Hg21: "INV9 e1'"
      using IH[of M Zs S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' "\<Gamma>r\<union>\<Gamma>2" "Zr\<union>Z2\<union>filterScFreeX X \<union> Z"] proof 
      show  "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M" and "\<Gamma>s \<sim> \<Gamma>m" 
        and "\<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK" and "INV2 (filterStore S Zl)" and "Mz = {}" and "INV3 S"
        and "INV4 (filterStore S Zl)" and "INV5 S" 
        and " free_scope_store S = dom Zs" and "Zl \<subseteq> Zs" by assumption
    next
      show "S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1'" using e_app_p1 by auto
    next
      have "\<Gamma>1 \<sim> \<Gamma>r" using t_app env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>1] by auto
      thus "\<Gamma>1 \<sim> (\<Gamma>r\<union>\<Gamma>2)" using env_compat_union_left[of \<Gamma>1 \<Gamma>r \<Gamma>2] t_app by auto
    next
      show "\<Gamma>1 \<union> (\<Gamma>r\<union>\<Gamma>2);\<Delta> \<Turnstile> OK" using Hsubeq t_app by auto
    next
      show "\<Gamma>1 \<union> (\<Gamma>r \<union> \<Gamma>2) \<subseteq> \<Gamma>s" using Hsubeq t_app by auto
    next
      show "Z1 \<subseteq> Zl" using t_app by (auto simp add: Un_ac)
    next
      show "Zr \<union> Z2\<union> filterScFreeX X \<union> Z \<subseteq> Zl" using t_app by (auto simp add: Un_ac)
    next
      have "\<forall>x. x is free dollar sc in e1 \<longrightarrow> x \<notin> Environ.dom (Z2\<union>filterScFreeX X \<union> Z)" 
        and "INV6 (e1\<cdot> e2) Zr" by assumption
      thus "INV6 e1 (Zr \<union> Z2\<union> filterScFreeX X \<union> Z)" by (auto simp add: dom_def)
    next
      have "INV9 e1\<cdot>e2" by assumption
      thus "INV9 e1" by auto
    qed auto
    have H5_1: "\<Gamma>1' \<sim> \<Gamma>r" using H5_1_1 env_compat_def by auto
    have "\<Gamma>1' \<sim> \<Gamma>2" using H5_1_1 env_compat_def by auto moreover
    have "\<Gamma>1' \<union> \<Gamma>2;\<Delta> \<Turnstile> OK" 
      using H6_1 wf_env_subset[of "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2)" \<Delta> "\<Gamma>1' \<union> \<Gamma>2"]
      by auto moreover
    have "\<Gamma>1';\<Delta>;{};Ze' \<turnstile> e1' : Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)" using Ht . moreover
    have "\<Gamma>2;\<Delta>;{};Z2 \<turnstile> e2 : \<tau>1" using t_app by auto 
    moreover
    have "minimalValidScope (scopeT \<tau>2) Z" and "filterScFreeX X \<Turnstile> ScopeSet" by assumption
    moreover
    have Hdol: "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> dom (Z2 \<union> filterScFreeX X \<union> Z) " 
      by assumption
    have "Z1\<union>Z2\<union> filterScFreeX X \<union> Z \<subseteq> Zl" and "Zl \<subseteq> Zs" by assumption
    hence "Z2\<union> filterScFreeX X \<union> Z \<subseteq> Zs" by auto
    hence "\<forall>\<rho>. \<rho> is free dollar sc in e1' \<longrightarrow> \<rho> \<notin> Environ.dom (Z2 \<union> filterScFreeX X \<union> Z)" 
      using eval_preserves_let_dollars[of e1 "Z2\<union> filterScFreeX X\<union>Z" Zs e1'] Hg19 Hdol
      by auto
    moreover
    {
      have Hdol1: "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom (Z1 \<union> filterScFreeX X \<union> Z)" 
        and "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom (Z2 \<union> filterScFreeX X \<union> Z)" 
        by assumption
      moreover
      note Hg19 Hg20
      moreover
      have "Z1\<union>Z2\<union> filterScFreeX X \<union> Z  \<subseteq> Zl" and "Zl \<subseteq> Zs" by assumption
      hence "Z1\<union>Z2 \<subseteq> Zs" by auto
      moreover
      have "\<Gamma>2; \<Delta>; Mz; Z2 \<turnstile> e2 : \<tau>1" by assumption
      moreover
      have "INV9 e2" using t_app.prems by auto
      ultimately
      have "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom Ze'"
        using eval_preserves_let_dollars_scope[of e2 Z1 e1 Z2 e1' Zs Ze' \<Gamma>2 _ Mz \<tau>1] 
        by (auto simp add: dom_def)
      hence "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> dom (Ze'\<union>filterScFreeX X\<union>Z)"
        using Hdol1 by (auto simp add: dom_def)
    }
    ultimately
    have Hg3: "\<Gamma>1' \<union> \<Gamma>2;\<Delta>;{};(Ze'\<union>Z2\<union> filterScFreeX X \<union> Z) \<turnstile> e1' \<cdot> e2 : \<tau>2" 
      using Typing.t_app by auto
    have Hg4: "\<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>s'" using H4_1 by auto
    have "\<Gamma>2 \<sim> \<Gamma>r" using t_app env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>2] by auto
    hence Hg5: "(\<Gamma>1' \<union> \<Gamma>2) \<sim> \<Gamma>r"
      using H5_1 env_compat_union_right[of \<Gamma>1' \<Gamma>r \<Gamma>2] by auto
    have "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2) = \<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    hence Hg6: "\<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" using H6_1 by auto
    have "Z1 \<union> Z2 \<union> filterScFreeX X \<union> Z \<subseteq> Zl" by assumption 
    hence Hg13: "Ze' \<union> Z2 \<union> filterScFreeX X \<union> Z \<subseteq> Zl'" using Hg15 Hg13 by auto
    have "INV6 (e1\<cdot>e2) Zr" by assumption
    hence Hg18: "INV6 (e1'\<cdot>e2) Zr" using Hg18 by (auto simp add: dom_def)
    have Hg19: "INV7 (e1\<cdot>e2) (e1'\<cdot>e2) Zs" using Hg19 by (auto simp add: dom_def)
    have Hg20: "INV8 (Ze' \<union> Z2\<union> filterScFreeX X \<union> Z) (Z1 \<union> Z2\<union> filterScFreeX X \<union> Z) (e1'\<cdot>e2)" 
      using Hg20 by (auto simp add: dom_def)
    have "INV9 e1\<cdot>e2" by assumption
    hence Hg21: "INV9 e1'\<cdot>e2" using Hg21 by auto
    show ?thesis
      apply (rule exI[of _ "\<Gamma>1'\<union>\<Gamma>2"], rule exI[of _ \<Gamma>s'])
      apply (rule exI[of _ \<Gamma>m'], rule exI[of _ M'])
      apply (rule exI[of _ Zs'], rule exI[of _ "Ze' \<union> Z2 \<union> filterScFreeX X \<union> Z"], rule exI[of _ Zl'])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg12 Hg13 Hg14 Hg15 Hg15g Hg16 Hg17 Hg18 Hg19 
        Hg20 Hg21 e_app_p1 
      by auto
  next
    case (e_app_p2 Sa Sa' \<mu>a \<mu>a' e1a e2' e2a)
    from t_app
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r Zs Zr.
          \<lbrakk> Mz = {}; M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M;  S;\<mu>;e2\<hookrightarrow>S';\<mu>';e2';
           \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma>2 \<sim> \<Gamma>r; \<Gamma>2 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; 
           Z2 \<subseteq> Zl; Zr \<subseteq> Zl; Zl \<subseteq> Zs;
           free_scope_store S = dom Zs; INV2 (filterStore S Zl); INV3 S;
           INV4 (filterStore S Zl); INV5 S; INV6 e2 Zr; INV9 e2;
           \<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK \<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'.
            (M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e2' : \<tau>1) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (Ze' \<subseteq> Zl') \<and> (Zl' \<subseteq> Zs') \<and> (Zs \<subseteq> Zs') \<and> (Zr \<subseteq> Zl') \<and>
            (free_scope_store S' = Environ.dom Zs') \<and>
            (INV2 (filterStore S' Zl')) \<and> (INV3 S') \<and>
            (INV4 (filterStore S' Zl')) \<and> (INV5 S') \<and>
            (INV6 e2' Zr) \<and> (INV7 e2 e2' Zs) \<and> (INV8 Ze' Z2 e2') \<and> (INV9 e2')"
      using t_app.hyps by auto
    have Hsubeq: "\<Gamma>2 \<union> (\<Gamma>r\<union>\<Gamma>1) = \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    obtain \<Gamma>2' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'
      where Hg1: "M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>2';\<Delta>;{};Ze' \<turnstile> e2' : \<tau>1"
      and H4_1: "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1) \<subseteq> \<Gamma>s'"
      and H5_1_1: "\<Gamma>2' \<sim> (\<Gamma>r\<union>\<Gamma>1)"
      and H6_1: "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1);\<Delta> \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg10: "INV2 (filterStore S' Zl')"
      and Hg11: "INV3 S'"
      and Hg12: "Zs \<subseteq> Zs'"
      and Hg13: "Ze' \<subseteq> Zl'"
      and Hg14: "free_scope_store S' = dom Zs'"
      and Hg15g: "Zl' \<subseteq> Zs'"
      and Hg15: "Zr\<union>Z1\<union> filterScFreeX X \<union> Z \<subseteq> Zl'"
      and Hg16: "INV4 (filterStore S' Zl')"
      and Hg17: "INV5 S'"
      and Hg18: "INV6 e2' (Zr\<union>Z1\<union> filterScFreeX X \<union> Z)"
      and Hg19: "INV7 e2 e2' Zs"
      and Hg20: "INV8 Ze' Z2 e2'"
      and Hg21: "INV9 e2'"
      using IH[of M Zs S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' "\<Gamma>r\<union>\<Gamma>1" "Zr\<union>Z1\<union> filterScFreeX X \<union> Z"] proof
      show  "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M" and "\<Gamma>s \<sim> \<Gamma>m" 
        and "\<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK" and "INV2 (filterStore S Zl)" and "Mz = {}" and "INV3 S"
        and "INV4 (filterStore S Zl)" and "INV5 S" 
        and " free_scope_store S = dom Zs" and "Zl \<subseteq> Zs" by assumption
    next
      show "S;\<mu>;e2\<hookrightarrow>S';\<mu>';e2'" using e_app_p2 by auto
    next
      have "\<Gamma>2 \<sim> \<Gamma>r" using t_app env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>2] by auto
      thus "\<Gamma>2 \<sim> (\<Gamma>r\<union>\<Gamma>1)"
      using t_app env_compat_union_left[of \<Gamma>2 \<Gamma>r \<Gamma>1] 
        env_compat_symm[of \<Gamma>1 \<Gamma>2] by auto
    next
      show "\<Gamma>2 \<union> (\<Gamma>r\<union>\<Gamma>1);\<Delta> \<Turnstile> OK" using Hsubeq t_app by auto
    next
      show "\<Gamma>2 \<union> (\<Gamma>r \<union> \<Gamma>1) \<subseteq> \<Gamma>s" using Hsubeq t_app by auto
    next
      show "Z2 \<subseteq> Zl" using t_app by (auto simp add: Un_ac)
    next
      show "Zr \<union> Z1 \<union> filterScFreeX X \<union> Z \<subseteq> Zl" using t_app.prems by (auto simp add: Un_ac)
    next
      have "\<forall>x. x is free dollar sc in e2 \<longrightarrow> x \<notin> Environ.dom (Z1\<union> filterScFreeX X \<union> Z)"
        and "INV6 (e1\<cdot>e2) Zr" by assumption
      thus "INV6 e2 (Zr \<union> Z1\<union> filterScFreeX X \<union> Z)" by (auto simp add: dom_def)
    next
      have "INV9 e1\<cdot>e2" by assumption
      thus "INV9 e2" by auto
    next 
    qed auto
    have H5_1: "\<Gamma>2' \<sim> \<Gamma>r" using H5_1_1 env_compat_def by auto
    have "\<Gamma>1 \<sim> \<Gamma>2'" using H5_1_1 env_compat_def by auto moreover
    have "\<Gamma>1 \<union> \<Gamma>2';\<Delta> \<Turnstile> OK" 
      using H6_1 wf_env_subset[of "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1)" _ "\<Gamma>1 \<union> \<Gamma>2'"]
      by auto moreover
    have  "\<Gamma>1;\<Delta>;{};Z1 \<turnstile> e1 : Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)" using t_app by auto
    moreover
    have  "\<Gamma>2';\<Delta>;{};Ze' \<turnstile> e2' : \<tau>1" using Ht by auto
    moreover
      have Hdol: "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> dom (Z1\<union>filterScFreeX X\<union>Z)" 
        by assumption
    have "Z1\<union>Z2\<union> filterScFreeX X \<union> Z \<subseteq> Zl" and "Zl \<subseteq> Zs" by assumption
    hence "Z1\<union>Z2\<union> filterScFreeX X \<union> Z \<subseteq> Zs" by auto
    hence "\<forall>\<rho>. \<rho> is free dollar sc in e2' \<longrightarrow> \<rho> \<notin> Environ.dom (Z1\<union>filterScFreeX X\<union>Z)"
      using eval_preserves_let_dollars[of e2 _ Zs e2'] Hg19 Hdol by auto
    moreover
    {
      have Hdol2: "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> dom (Z2\<union>filterScFreeX X\<union>Z)"
        and "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> dom (Z1\<union>filterScFreeX X\<union>Z)" 
        by assumption
      moreover
      note Hg19 Hg20
      moreover
      have "Z1\<union>Z2\<union> filterScFreeX X \<union> Z \<subseteq> Zl" and "Zl \<subseteq> Zs" by assumption
      hence "Z1\<union>Z2 \<subseteq> Zs" by auto
      moreover
      have "\<Gamma>1; \<Delta>; Mz; Z1 \<turnstile> e1 : Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)" by assumption
      moreover
      have "INV9 e1" using t_app.prems by auto      
      ultimately
      have "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom Ze'"  
        using eval_preserves_let_dollars_scope[of e1 Z2 e2 Z1 e2' Zs Ze' \<Gamma>1 _ Mz] 
        by (auto simp add: dom_def)
      hence "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> dom (Ze'\<union>filterScFreeX X\<union>Z)"
        using Hdol2 by (auto simp add: dom_def)
    }
    moreover
    have "minimalValidScope (scopeT \<tau>2) Z" and "filterScFreeX X \<Turnstile> ScopeSet" by assumption    
    ultimately
    have Hg3: "\<Gamma>1 \<union> \<Gamma>2';\<Delta>;{};Z1\<union>Ze'\<union> filterScFreeX X \<union> Z \<turnstile> e1 \<cdot> e2' : \<tau>2"
      using Typing.t_app t_app by auto
    have Hg4: "\<Gamma>2' \<union> \<Gamma>1 \<union> \<Gamma>r \<subseteq> \<Gamma>s'" using H4_1 by auto
    have "\<Gamma>1 \<sim> \<Gamma>r" using t_app env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>1] by auto
    hence Hg5: "(\<Gamma>1 \<union> \<Gamma>2') \<sim> \<Gamma>r"
      using H5_1 env_compat_union_right[of \<Gamma>1 \<Gamma>r \<Gamma>2'] by auto
    have "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1) = \<Gamma>1 \<union> \<Gamma>2' \<union> \<Gamma>r" by auto
    hence Hg6: "\<Gamma>1 \<union> \<Gamma>2' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" using H6_1 by auto
    have "Z1 \<union> Z2 \<union> filterScFreeX X \<union> Z \<subseteq> Zl" by assumption
    hence Hg13: "Z1 \<union> Ze' \<union> filterScFreeX X \<union> Z \<subseteq> Zl'" using Hg15 Hg13 by auto
    have "INV6 (e1\<cdot>e2) Zr" by assumption
    hence Hg18: "INV6 (e1\<cdot>e2') Zr" using Hg18 by (auto simp add: dom_def)
    have Hg19: "INV7 (e1\<cdot>e2) (e1\<cdot>e2') Zs" using Hg19 by (auto simp add: dom_def)
    have Hg20: "INV8 (Z1 \<union> Ze'\<union> filterScFreeX X \<union> Z) (Z1 \<union> Z2\<union> filterScFreeX X \<union> Z) (e1\<cdot>e2')" 
      using Hg20 by (auto simp add: dom_def)
    have "INV9 e1\<cdot>e2" by assumption
    hence Hg21: "INV9 e1\<cdot>e2'" using Hg21 by auto
    show ?thesis
      apply (rule exI[of _ "\<Gamma>1\<union>\<Gamma>2'"], rule exI[of _ \<Gamma>s'])
      apply (rule exI[of _ \<Gamma>m'], rule exI[of _ M'])
      apply (rule exI[of _ Zs'], rule exI[of _ "Z1 \<union> Ze'\<union> filterScFreeX X \<union> Z"], rule exI[of _ Zl'])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg12 Hg13 Hg14 Hg15 Hg15g Hg16 Hg17 Hg18 Hg19 
        Hg20 Hg21 e_app_p2 
      by auto
  next
    case (e_beta Sa Sa' \<mu> \<tau> e qz y z \<pi>)
    {
      assume Hu: "qz = qU"
      hence Heq: "S = S'" and Hzins: "z\<mapsto>\<lambda>^qz^\<pi>[\<tau>].e \<in> S"
        using e_beta t_app by auto
      have Hz12l: "Z1\<union>Z2\<union> filterScFreeX X \<union> Z \<subseteq> Zl" by assumption
      have "\<Gamma>1;\<Delta>;Mz;Z1 \<turnstile> e1 : Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)" by assumption
      hence "validScope (scopeT (Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2))) Z1" by (rule ty_valid_scope) 
      hence Htvalid: "validScope (scopeT (Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2))) Zl" 
        apply (rule valid_scope_subset) using Hz12l by auto
      have "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" using e_beta t_app by auto
      also have Hgmeq: "\<Gamma>s \<union> \<Gamma>m = \<Gamma>s \<union> (un \<Gamma>m) \<union> \<Gamma>m" 
        by (auto simp add: un_def)
      finally have Hg1: "M;\<Delta>;Zs;Zl \<Turnstile> S:((\<Gamma>s \<union> (un \<Gamma>m)) \<union> \<Gamma>m)" by auto
      have  Hg2: "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M"  using e_beta t_app by auto
      have Hzing: "z\<rhd>(Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)) \<in> \<Gamma>1" 
        using t_app e_beta by auto
      hence Hzing: "z\<rhd>(Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)) \<in> \<Gamma>s\<union>\<Gamma>m" 
        using t_app e_beta by auto
      have Hyg2: "y\<rhd>\<tau>1 \<in> \<Gamma>2" using t_app e_beta by auto
      hence Hying: "y\<rhd>\<tau>1 \<in> \<Gamma>s" using t_app e_beta by auto
      have Htr: "M;\<Delta>;Zs;Zl \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
      hence "scopeV (\<lambda>^qz^\<pi>[\<tau>].e) = scopeT (Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2))"
        apply (rule storeTyping_scope_sanity) using Hzing Hzins by auto
      hence Hvalidv: "validScope (scopeV (\<lambda>^qz^\<pi>[\<tau>].e)) Zl" using Htvalid by auto
      have "\<exists>\<Gamma>v Znew. \<Gamma>v;\<Delta>;M;Znew \<turnstile> \<lambda>^qz^\<pi>[\<tau>].e : Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)
                      \<and> un \<Gamma>v \<subseteq> un (\<Gamma>s \<union> \<Gamma>m) \<and> Znew \<subseteq> Zs"
        using storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" z " \<lambda>^qz^\<pi>[\<tau>].e"
                                 "Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)"] Hzins Hzing Htr Hvalidv
        by auto
      then obtain \<Gamma>v Znew
        where Ht: "\<Gamma>v;\<Delta>;M;Znew \<turnstile> \<lambda>^qz^\<pi>[\<tau>].e : Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)"
        and Hun: "un \<Gamma>v \<subseteq> un (\<Gamma>s \<union> \<Gamma>m)"
        and Hzsub: "Znew \<subseteq> Zs"
        by blast
      hence Hqeq: "qz = q" and Hteq: "\<tau> = \<tau>1"  by auto
      hence HqualG: "qualG \<Gamma>v \<preceq> qU" using Hu Ht by auto
      hence Hunv: "un \<Gamma>v = \<Gamma>v" using equal_unrestricted by auto
      from Ht obtain L where Ht_b: "\<forall>x. \<Gamma>v,(x:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>v \<sim> {x\<rhd>\<tau>1} \<and> x \<notin> L \<longrightarrow>
                           \<Gamma>v,(x:\<tau>1);\<Delta>;M;(filterScFreeX X) \<turnstile> freshen_tm x e : \<tau>2"
        and Hfin: "finite L"
        by auto
      have "\<Gamma>v;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<tau>1" using ty_wf_env Ht by auto
      moreover
      hence "finite (dom \<Gamma>v)" using dom_finite by (auto simp add: wf_env_def)
      ultimately
      obtain k where Hk: "\<Gamma>v,(k:\<tau>1);\<Delta> \<Turnstile> OK \<and> k \<notin> L \<and> \<Gamma>v \<sim> {k\<rhd>\<tau>1} \<and> k \<notin> set (free_tm e) \<and> k \<notin> dom \<Gamma>v" 
        using new_variable_abs[where Lx="L\<union>dom \<Gamma>v\<union>(set (free_tm e))"] Hfin by blast
      hence Htk: "\<Gamma>v,(k:\<tau>1);\<Delta>;M;(filterScFreeX X) \<turnstile> freshen_tm k e : \<tau>2" using Ht_b by auto
      have "\<Gamma>s \<union> \<Gamma>v \<subseteq> \<Gamma>s \<union> \<Gamma>m" using Hunv Hun by (auto simp add: un_def)
      hence "\<Gamma>s \<union> \<Gamma>v;\<Delta> \<Turnstile> OK" using wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" _ "\<Gamma>s\<union>\<Gamma>v"] t_app 
        by auto
      hence Hy_wf: "\<Gamma>v,(y:\<tau>1);\<Delta> \<Turnstile> OK" using wf_env_union_sanity_cons Hying by auto
      have Hy_comp: "\<Gamma>v \<sim> {(y\<rhd>\<tau>1)}" using HqualG
        by (cases "qualT \<tau>1", auto simp add: env_compat_def 
          dom_def qualGrelL_def qualT_unrestricted)
      have Hg10: "INV2 (filterStore S Zl)"  and Hg11: "INV3 S"
        and Hg16: "INV4 (filterStore S Zl)" and Hg17: "INV5 S"
        and Hg14: "free_scope_store S = Environ.dom Zs" 
        and Hg15:  "Zl \<subseteq> Zs" and Hg18: "Zr \<subseteq> Zl"
        by assumption
      have "Z1\<union>Z2\<union> filterScFreeX X \<union> Z \<subseteq> Zl" by assumption
      hence Hg12: "filterScFreeX X \<subseteq> Zl" by auto
      have "\<not> hasDollars (\<lambda>^qz^\<pi>[\<tau>].e)" using Hzins Hg17 INV5 by blast
      hence Hd: "\<not> hasDollars e" by auto
      hence "\<not> (\<exists>\<rho>. \<rho> is free dollar sc in e)" using not_hasDollars_exists by auto
      hence Hdollars: "\<not> (\<exists>\<rho>. \<rho> is free dollar sc in (freshen_tm y e))" 
        using free_dollars_vsubst_simpl by auto
      hence Hg19: "INV6 (freshen_tm y e) Zr" using INV6 by auto
      have Hg20: "INV7 (e1 \<cdot> e2) (freshen_tm y e) Zs" using Hdollars INV7 by blast
      have Hg21: "INV8 (filterScFreeX X) (Z1 \<union> Z2\<union> filterScFreeX X \<union> Z) e'" 
        by (auto simp add: dom_def)
      have Hg22: "INV9 (freshen_tm y e)" using Hdollars hasDollars_exists INV9_intro by blast
      have "\<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> False" using Hd not_hasDollars_exists by blast
      hence Hd: "\<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> scopeT \<tau>1 \<noteq> ScFree \<rho>" by auto
      have "\<Gamma>v,(y:\<tau>1);\<Delta>;M;filterScFreeX X \<turnstile> vsubst_tm (TmFreeVar y) 0 e : \<tau>2" 
        apply (rule substitution[where x=k]) using Htk Hk Hy_wf Hy_comp Hd by auto
      moreover
      have "INV3 S" by assumption
      hence "(\<exists>q \<pi>' l. (\<lambda>^qz^\<pi>[\<tau>].e) = TmLoc q \<pi>' l) \<or> 
        (\<forall>l. \<not> l free loc in (\<lambda>^qz^\<pi>[\<tau>].e))"
        using INV3[of S] Hzins by blast
      hence "(\<forall>l. \<not> l free loc in (\<lambda>^qz^\<pi>[\<tau>].e))" by auto
      hence "(\<forall>l. \<not> l free loc in e)" by auto
      hence "(\<forall>l. \<not> l free loc in (freshen_tm y e))" 
        using vsubst_tm_same_loc by auto
      ultimately
      have Hg3: "\<Gamma>v,(y:\<tau>1);\<Delta>;{};filterScFreeX X \<turnstile> freshen_tm y e : \<tau>2"
        using ty_mem_strengthen_all[of "\<Gamma>v,(y:\<tau>1)" _ _ _ "freshen_tm y e"] 
        by auto
      note Hun
      also have "un (\<Gamma>s \<union> \<Gamma>m) = un \<Gamma>s \<union> un \<Gamma>m" by (auto simp add: un_def)
      also note Hunv
      finally have "\<Gamma>v \<subseteq> un \<Gamma>s \<union> un \<Gamma>m" by auto
      moreover have "un \<Gamma>s \<subseteq> \<Gamma>s" by (auto simp add: un_def)
      ultimately have "\<Gamma>v \<subseteq> \<Gamma>s \<union> un \<Gamma>m" by auto
      moreover
      note Hying
      moreover have "\<Gamma>r \<subseteq> \<Gamma>s" using t_app by auto
      ultimately have Hg4: "\<Gamma>v,(y:\<tau>1) \<union> \<Gamma>r \<subseteq> \<Gamma>s\<union>(un \<Gamma>m)" by auto
      have Hg5: "\<Gamma>v,(y:\<tau>1) \<sim> \<Gamma>r" using HqualG
      proof (auto simp add: env_compat_def qualGrelL_def)
        assume Hy: "y\<rhd>\<tau>1 \<in> \<Gamma>r"
        { assume "qualT \<tau>1 = qU"  hence "qualT \<tau>1 \<preceq> qU" by auto }
        moreover
        {
          assume "qualT \<tau>1 = qL"
          moreover have "\<Gamma>2 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK"
            using t_app wf_env_subset[of "\<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" \<Delta> "\<Gamma>2 \<union> \<Gamma>r"] by auto
          moreover have "\<Gamma>2 \<sim> \<Gamma>r"
            using t_app env_compat_subset[of "\<Gamma>1 \<union> \<Gamma>2" \<Gamma>r \<Gamma>2] by auto
          ultimately have False using Hy Hyg2 compat_ok_sanity[of \<Gamma>2 \<Gamma>r _ y \<tau>1]
            by (auto simp add: dom_def)
        }
        ultimately show "qualT \<tau>1 \<preceq> qU" by (cases "qualT \<tau>1", auto)
      qed
      have "\<Gamma>s \<union> (un \<Gamma>m);\<Delta> \<Turnstile> OK"
        using t_app wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Delta> "\<Gamma>s \<union> (un \<Gamma>m)"]
        by (auto simp add: un_def)
      hence Hg6: "\<Gamma>v,(y:\<tau>1) \<union> \<Gamma>r;\<Delta> \<Turnstile> OK"
        using wf_env_subset[of "\<Gamma>s \<union> (un \<Gamma>m)"] Hg4 by auto
      have "qualG (un \<Gamma>m) \<preceq> qU" by (auto simp add: un_def qualGrelL_def)
      hence Hg7: "(\<Gamma>s\<union>(un \<Gamma>m)) \<sim> \<Gamma>m"
        using t_app env_compat_weakening_left[of \<Gamma>s \<Gamma>m "un \<Gamma>m"] by auto
      have ?thesis
        apply (rule exI[of _ "\<Gamma>v,(y:\<tau>1)"],
          rule exI[of _ "\<Gamma>s\<union> (un \<Gamma>m)"],
          rule exI[of _ \<Gamma>m], rule exI[of _ M], 
          rule exI[of _ Zs], rule exI[of _ "filterScFreeX X"], rule exI[of _ Zl])
        using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg12 Hg14 Hg15 Hg16 Hg17 Hg18 Hg19 Hg20 
          Hg21 Hg22 e_beta Heq 
      by auto
    }
    moreover
    {
      assume Hql: "qz = qL"
      hence Heq: "S' = S\<bar>z" and Hzins: "z\<mapsto>\<lambda>^qz^\<pi>[\<tau>].e \<in> S"
        using e_beta t_app by auto
      have Hz12l: "Z1\<union>Z2\<union> filterScFreeX X \<union> Z \<subseteq> Zl" by assumption
      have "\<Gamma>1;\<Delta>;Mz;Z1 \<turnstile> e1 : Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)" by assumption
      hence "validScope (scopeT (Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2))) Z1" by (rule ty_valid_scope) 
      hence Htvalid: "validScope (scopeT (Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2))) Zl" 
        apply (rule valid_scope_subset) using Hz12l by auto
      have "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" using e_beta t_app by auto
      have Hzing: "z\<rhd>(Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)) \<in> \<Gamma>1"
        using t_app e_beta by auto
      hence Hzing: "z\<rhd>(Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)) \<in> \<Gamma>s\<union>\<Gamma>m"
        using t_app e_beta by auto
      have Hyg2: "y\<rhd>\<tau>1 \<in> \<Gamma>2" using t_app e_beta by auto
      hence Hying: "y\<rhd>\<tau>1 \<in> \<Gamma>s" using t_app e_beta by auto
      hence "y\<rhd>\<tau>1 \<in> \<Gamma>s\<union>\<Gamma>m" and "\<Gamma>s\<union>\<Gamma>m;\<Delta> \<Turnstile> OK" by auto
      hence "y \<noteq> z" using Hzing proof (cases "y=z")
        assume "y = z"
          and "y\<rhd>\<tau>1 \<in> \<Gamma>s\<union>\<Gamma>m"
          and "\<Gamma>s\<union>\<Gamma>m;\<Delta> \<Turnstile> OK"
          and "z\<rhd>(Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)) \<in> \<Gamma>s\<union>\<Gamma>m"
        hence "\<tau>1 = Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)" 
          by (auto simp add: wf_env_def[of "\<Gamma>s\<union>\<Gamma>m"])
        hence "size \<tau>1 = size (Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2))" by auto
        thus ?thesis by auto
      next
        assume "y\<noteq>z"
        thus ?thesis by auto
      qed
      hence Hyingwz: "y\<rhd>\<tau>1 \<in> \<Gamma>s\<bar>z" using Hying by (auto simp add: without_def)
      have Htr: "M;\<Delta>;Zs;Zl \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
      hence "scopeV (\<lambda>^qz^\<pi>[\<tau>].e) = scopeT (Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2))"
        apply (rule storeTyping_scope_sanity) using Hzing Hzins by auto
      hence Hvalidv: "validScope (scopeV (\<lambda>^qz^\<pi>[\<tau>].e)) Zl" using Htvalid by auto
      have "q = qL"
        using storeTyping_sanity[of M _ Zs Zl S "\<Gamma>s\<union>\<Gamma>m" z " \<lambda>^qz^\<pi>[\<tau>].e"
                                 "Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)"]
          Hzins Hzing Htr Hql Hvalidv
        by auto
      hence "\<exists>\<Gamma>v Znew. \<Gamma>v;\<Delta>;M;Znew \<turnstile> \<lambda>^qz^\<pi>[\<tau>].e : Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2) \<and>
             M;\<Delta>;Zs;Zl \<Turnstile> S\<bar>z : ((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v) \<and> un \<Gamma>v \<subseteq> un (\<Gamma>s\<union>\<Gamma>m) \<and>
             ((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<sim> \<Gamma>v) \<and> ((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v);\<Delta> \<Turnstile> OK \<and> Znew \<subseteq> Zs"
        using storeTyping_elim[of M _ Zs Zl S  "\<Gamma>s\<union>\<Gamma>m" z " \<lambda>^qz^\<pi>[\<tau>].e"
                               "Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)"] Hzins Hzing Htr Hvalidv
        by auto
      then obtain \<Gamma>v Znew
        where Ht: "\<Gamma>v;\<Delta>;M;Znew \<turnstile> \<lambda>^qz^\<pi>[\<tau>].e : Qual q ScBot (\<tau>1 \<rightarrow>^X \<tau>2)"
        and Hst: "M;\<Delta>;Zs;Zl \<Turnstile> S\<bar>z : ((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v)"
        and Hcomp: "((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<sim> \<Gamma>v)"
        and Hok: "((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v);\<Delta> \<Turnstile> OK"
        and Hun: "un \<Gamma>v \<subseteq> un (\<Gamma>s \<union> \<Gamma>m)"
        and Hzsum: "Znew \<subseteq> Zs"
        by blast
      hence Hqeq: "qz = q" and Hteq: "\<tau> = \<tau>1" and Hscope: "\<pi> = ScBot" by auto
      have Hse: "z\<rhd>(Qual qL ScBot (\<tau>1 \<rightarrow>^X \<tau>2)) \<in> \<Gamma>1"
        using t_app e_beta Hqeq Hql by auto
      hence Hsz: "z\<rhd>(Qual qL ScBot (\<tau>1 \<rightarrow>^X \<tau>2)) \<in> \<Gamma>s"
        using t_app by auto
      hence "\<not> \<Gamma>m defines z"
        using compat_ok_sanity[of \<Gamma>s \<Gamma>m _ z "Qual qL ScBot (\<tau>1 \<rightarrow>^X \<tau>2)"] 
          t_app Hql Hqeq
        by auto
      hence Hmz: "\<Gamma>m\<bar>z = \<Gamma>m" by (auto simp add: without_def dom_def)
      note Hst
      also have "(\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v = \<Gamma>s\<bar>z \<union> \<Gamma>v \<union> \<Gamma>m" using Hmz
        by (auto simp add: without_def)
      finally
      have Hg1: "M;\<Delta>;Zs;Zl \<Turnstile> S\<bar>z:((\<Gamma>s\<bar>z \<union> \<Gamma>v) \<union> \<Gamma>m)" by auto
      have Hg2: "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M"  using e_beta t_app by auto
      from Ht obtain L where Ht_b: "\<forall>x. \<Gamma>v,(x:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>v \<sim> {x\<rhd>\<tau>1} \<and> x \<notin> L \<longrightarrow>
                           \<Gamma>v,(x:\<tau>1);\<Delta>;M;(filterScFreeX X) \<turnstile> freshen_tm x e : \<tau>2"
        and Hfin: "finite L"
        by auto
      have "\<Gamma>v;\<Delta> \<Turnstile> OK" and "\<Delta> \<Turnstile>' \<tau>1" using ty_wf_env Ht by auto
      moreover
      hence "finite (dom \<Gamma>v)" using dom_finite by (auto simp add: wf_env_def)
      ultimately
      obtain k where Hk: "\<Gamma>v,(k:\<tau>1);\<Delta> \<Turnstile> OK \<and> k \<notin> L \<and> \<Gamma>v \<sim> {k\<rhd>\<tau>1} \<and> k \<notin> set (free_tm e) \<and> k \<notin> dom \<Gamma>v" 
        using new_variable_abs[where Lx="L\<union>dom \<Gamma>v\<union>(set (free_tm e))"] Hfin by blast
      hence Htk: "\<Gamma>v,(k:\<tau>1);\<Delta>;M;(filterScFreeX X) \<turnstile> freshen_tm k e : \<tau>2" using Ht_b by auto
      have "\<Gamma>s\<bar>z \<union> \<Gamma>v \<subseteq> (\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v" by (auto simp add: without_def)
      hence "\<Gamma>s\<bar>z \<union> \<Gamma>v;\<Delta> \<Turnstile> OK" using wf_env_subset Hok by auto
      hence Hy_wf: "\<Gamma>v,(y:\<tau>1);\<Delta> \<Turnstile> OK" using wf_env_union_sanity_cons Hyingwz by auto
      have "y\<rhd>\<tau>1 \<in> ((\<Gamma>s\<union>\<Gamma>m)\<bar>z)" using Hyingwz by (auto simp add: without_def)
      hence "{(y\<rhd>\<tau>1)} \<sim> \<Gamma>v" using env_compat_subset Hcomp by auto
      hence Hy_comp: "\<Gamma>v \<sim> {(y\<rhd>\<tau>1)}" using env_compat_symm by auto
      have "INV5 S" by assumption
      hence "\<not> hasDollars (\<lambda>^qz^\<pi>[\<tau>].e)" using Hzins INV5 by blast
      hence Hd: "\<not> hasDollars e" by auto
      hence "\<not> (\<exists>\<rho>. \<rho> is free dollar sc in e)" using not_hasDollars_exists by auto
      hence Hdollars: "\<not> (\<exists>\<rho>. \<rho> is free dollar sc in (freshen_tm y e))" 
        using free_dollars_vsubst_simpl by auto
      hence Hg19: "INV6 (freshen_tm y e) Zr" using INV6 by auto
      have "\<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> False" using Hd not_hasDollars_exists by blast
      hence Hd: "\<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> scopeT \<tau>1 \<noteq> ScFree \<rho>" by auto
      have "\<Gamma>v,(y:\<tau>1);\<Delta>;M;filterScFreeX X \<turnstile> vsubst_tm (TmFreeVar y) 0 e : \<tau>2" 
        apply (rule substitution[where x=k]) using Htk Hk Hy_wf Hy_comp Hd by auto
      moreover
      have "INV3 S" by assumption
      hence "(\<exists>q \<pi>' l. (\<lambda>^qz^\<pi>[\<tau>].e) = TmLoc q \<pi>' l) \<or> 
        (\<forall>l. \<not> l free loc in (\<lambda>^qz^\<pi>[\<tau>].e))"
        using INV3[of S] Hzins by blast
      hence "(\<forall>l. \<not> l free loc in (\<lambda>^qz^\<pi>[\<tau>].e))" by auto
      hence "(\<forall>l. \<not> l free loc in e)" by auto
      hence "(\<forall>l. \<not> l free loc in (freshen_tm y e))"
        using vsubst_tm_same_loc by auto
      ultimately
      have Hg3: "\<Gamma>v,(y:\<tau>1);\<Delta>;{};filterScFreeX X \<turnstile> freshen_tm y e : \<tau>2"
        using ty_mem_strengthen_all[of "\<Gamma>v,(y:\<tau>1)" _ _ _ "freshen_tm y e"] 
        by auto
      have "\<Gamma>s;\<Delta> \<Turnstile> OK" using t_app wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Delta> \<Gamma>s] by auto
      moreover
      have "\<tau>1 \<noteq> Qual qL ScBot (\<tau>1 \<rightarrow>^X \<tau>2)" proof
        assume "\<tau>1 = Qual qL ScBot (\<tau>1 \<rightarrow>^X \<tau>2)"
        hence "size \<tau>1 = size (Qual qL ScBot (\<tau>1 \<rightarrow>^X \<tau>2))" by auto
        thus False by auto
      qed
      ultimately have Hneq: "y \<noteq> z" using Hying Hsz
        by (auto simp add: wf_env_def)
      have "\<not> \<Gamma>r defines z"
        using compat_ok_sanity[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r _ z "Qual qL ScBot (\<tau>1 \<rightarrow>^X \<tau>2)"]
          Hql Hqeq Hse t_app by auto
      moreover have "\<Gamma>r \<subseteq> \<Gamma>s" using t_app by auto
      ultimately have Hrsz: "\<Gamma>r \<subseteq> \<Gamma>s\<bar>z" 
        by (auto simp add: dom_def without_def)
      moreover
      hence "y\<rhd>\<tau>1 \<in> (\<Gamma>s\<bar>z)" using Hying Hneq by (auto simp add: without_def)
      ultimately have Hg4: "\<Gamma>v,(y:\<tau>1) \<union> \<Gamma>r \<subseteq> \<Gamma>s\<bar>z \<union> \<Gamma>v" by auto
      have "\<Gamma>s\<bar>z \<sim> \<Gamma>v" 
        using Hcomp env_compat_subset[of "(\<Gamma>s\<union>\<Gamma>m)\<bar>z" \<Gamma>v "\<Gamma>s\<bar>z"]
        by (auto simp add: without_def)
      hence "\<Gamma>r \<sim> \<Gamma>v" using Hrsz env_compat_subset[of "\<Gamma>s\<bar>z" \<Gamma>v \<Gamma>r] by auto
      hence "\<Gamma>v \<sim> \<Gamma>r" by (auto simp add: env_compat_def)
      hence Hg5: "\<Gamma>v,(y:\<tau>1) \<sim> \<Gamma>r"
      proof (auto simp add: env_compat_def qualGrelL_def)
        assume Hy: "y\<rhd>\<tau>1 \<in> \<Gamma>r"
        { assume "qualT \<tau>1 = qU"  hence "qualT \<tau>1 \<preceq> qU" by auto }
        moreover
        {
          assume "qualT \<tau>1 = qL"
          moreover have "\<Gamma>2 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK"
            using t_app wf_env_subset[of "\<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" \<Delta> "\<Gamma>2 \<union> \<Gamma>r"] by auto
          moreover have "\<Gamma>2 \<sim> \<Gamma>r"
            using t_app env_compat_subset[of "\<Gamma>1 \<union> \<Gamma>2" \<Gamma>r \<Gamma>2] by auto
          ultimately have False using Hy Hyg2 compat_ok_sanity[of \<Gamma>2 \<Gamma>r _ y \<tau>1]
            by (auto simp add: dom_def)
        }
        ultimately show "qualT \<tau>1 \<preceq> qU" by (cases "qualT \<tau>1", auto)
      qed
      have "\<Gamma>s\<bar>z \<union> \<Gamma>v;\<Delta> \<Turnstile> OK"
        using Hok wf_env_subset[of "(\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v" \<Delta> "\<Gamma>s\<bar>z \<union> \<Gamma>v"]
        by (auto simp add: without_def)
      hence Hg6: "\<Gamma>v,(y:\<tau>1) \<union> \<Gamma>r;\<Delta> \<Turnstile> OK"
        using wf_env_subset[of "\<Gamma>s\<bar>z \<union> \<Gamma>v"] Hg4 by auto
      have "\<Gamma>s\<bar>z \<sim> \<Gamma>m" using t_app env_compat_subset[of \<Gamma>s \<Gamma>m "\<Gamma>s\<bar>z"]
        by (auto simp add: without_def)
      moreover
      have "\<Gamma>m\<bar>z \<sim> \<Gamma>v" 
        using Hcomp env_compat_subset[of "(\<Gamma>s\<union>\<Gamma>m)\<bar>z" \<Gamma>v "\<Gamma>m\<bar>z"]
        by (auto simp add: without_def)
      hence "\<Gamma>m \<sim> \<Gamma>v" using Hmz by auto
      hence "\<Gamma>v \<sim> \<Gamma>m" by (auto simp add: env_compat_def)
      ultimately have Hg7: "(\<Gamma>s\<bar>z \<union>\<Gamma>v) \<sim> \<Gamma>m" using env_compat_union_right 
        by auto
      have "S\<bar>z \<subseteq> S" by (auto simp add: without_def)
      hence "filterStore (S\<bar>z) Zl \<subseteq> filterStore S Zl" by (auto simp add: filterStore_def)
      moreover
      have "INV2 (filterStore S Zl)" and  "INV4 (filterStore S Zl)" by assumption
      ultimately
      have Hg10: "INV2 (filterStore (S\<bar>z) Zl)" and Hg16: "INV4 (filterStore (S\<bar>z) Zl)" 
        using inv2_subset inv4_subset by auto
      have "INV3 S" and "INV5 S" by assumption
      moreover
      have "S\<bar>z \<subseteq> S" by (auto simp add: without_def)
      ultimately
      have Hg11: "INV3 (S\<bar>z)" and Hg17: "INV5 (S\<bar>z)" using inv3_subset inv5_subset by auto
      have Hg12: "filterScFreeX X \<subseteq> Zl" using t_app.prems by (auto simp only: Un_ac)
      have Hg13: "Zs \<subseteq> Zs" by auto
      have "free_scope_store ((S\<bar>z),(z\<mapsto>(\<lambda>^qz^\<pi>[\<tau>].e)))
            = free_scope_store (S\<bar>z)" 
        using Hscope free_scope_store_cons_empty[of "\<lambda>^qz^\<pi>[\<tau>].e" z "S\<bar>z"] by auto
      also
      have Hstore: "S \<Turnstile> Store" using storeTyping_wf_store t_app by auto
      hence "(S\<bar>z),(z\<mapsto>\<lambda>^qz^\<pi>[\<tau>].e) = S" using Hzins wf_store_defines_without by auto
      also
      have "free_scope_store S = dom Zs" by assumption
      finally have Hg14: "free_scope_store (S\<bar>z) = dom Zs" by auto
      have Hg15: "Zl \<subseteq> Zs" and Hg18: "Zr \<subseteq> Zl" by assumption
      have Hg20: "INV7 (e1 \<cdot> e2) (freshen_tm y e) Zs" using Hdollars INV7 by blast
      have Hg21: "INV8 (filterScFreeX X) (Z1 \<union> Z2\<union> filterScFreeX X \<union> Z) e'"
        by (auto simp add: dom_def)
      have Hg22: "INV9 (freshen_tm y e)" using Hdollars hasDollars_exists INV9_intro by blast
      have ?thesis
        apply (rule exI[of _ "\<Gamma>v,(y:\<tau>1)"],
          rule exI[of _ "\<Gamma>s\<bar>z \<union> \<Gamma>v"],
          rule exI[of _ \<Gamma>m], rule exI[of _ M],
          rule exI[of _ Zs], rule exI[of _ "filterScFreeX X"], rule exI[of _ Zl])
        using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg12 Hg13 Hg14 Hg15 Hg16 Hg17
          Hg18 Hg19 Hg20 Hg21 Hg22 e_beta Heq by (auto simp del: INV10)
    }
    ultimately
    show ?thesis by (cases qz, auto)
  qed auto
next;

      (********************* ptyapp cases *********************)
  case (t_ptyapp \<Delta> Z Z\<phi> Mz \<Gamma> \<tau> \<phi> e q X e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;e\<langle>\<phi>\<rangle> \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_tyapp_p Sa Sa' \<mu>a \<mu>a' ea e')
    from t_ptyapp.hyps
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r Zr.
          \<lbrakk> Mz = {}; M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M;  S;\<mu>;e\<hookrightarrow>S';\<mu>';e';
           \<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma> \<sim> \<Gamma>r; \<Gamma> \<union> \<Gamma>r;\<Delta> \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; \<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK;
           Z \<subseteq> Zl; Zr \<subseteq> Zl; Zl \<subseteq> Zs; free_scope_store S = dom Zs; 
           INV2 (filterStore S Zl); INV3 S; INV4 (filterStore S Zl); INV5 S;
           INV6 e Zr; INV9 e; \<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK 
          \<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'.
            (M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e' : Qual q ScBot (\<forall>.\<tau>^X)) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (Ze' \<subseteq> Zl') \<and> (Zl' \<subseteq> Zs') \<and> (Zs \<subseteq> Zs') \<and> (Zr \<subseteq> Zl') \<and>
            (free_scope_store S' = dom Zs') \<and> (INV2 (filterStore S' Zl')) \<and>
            (INV3 S') \<and> (INV4 (filterStore S' Zl')) \<and> (INV5 S') \<and> (INV6 e' Zr) \<and>
            (INV7 e e' Zs) \<and> (INV8 Ze' Z e') \<and> (INV9 e')"
      by auto
    obtain \<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'
      where Hg1: "M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e' : Qual q ScBot (\<forall>.\<tau>^X)"
      and Hg4: "\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s'"
      and Hg5: "\<Gamma>e' \<sim> \<Gamma>r"
      and Hg6: "\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg12: "Ze' \<subseteq> Zl'"
      and Hg15: "Zl' \<subseteq> Zs'"
      and Hg13: "Zs \<subseteq> Zs'"
      and Hg16: "Zr\<union>filterScFreeX X\<union> Z\<phi> \<subseteq> Zl'"
      and Hg14: "free_scope_store S' = dom Zs'"
      and Hg10: "INV2 (filterStore S' Zl')"
      and Hg11: "INV3 S'"
      and Hg17: "INV4 (filterStore S' Zl')"
      and Hg18: "INV5 S'"
      and Hg19: "INV6 e' (Zr\<union>filterScFreeX X\<union>Z\<phi>)"
      and Hg20: "INV7 e e' Zs"
      and Hg21: "INV8 Ze' Z e'"
      and Hg22: "INV9 e'"
      using IH[of M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r "Zr\<union>filterScFreeX X\<union>Z\<phi>"] proof rule
      show "Mz = {}" and "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M" 
        and "\<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s" and "\<Gamma> \<sim> \<Gamma>r" and "\<Gamma> \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" and "\<Gamma>s \<sim> \<Gamma>m"
        and "\<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK" and "INV2 (filterStore S Zl)" and "INV3 S"
        and "INV4 (filterStore S Zl)" and "INV5 S"
        and "Zl \<subseteq> Zs" and "free_scope_store S = dom Zs" by assumption
    next
      show "Zr\<union>filterScFreeX X\<union>Z\<phi> \<subseteq> Zl" using t_ptyapp by auto
    next
      show "S;\<mu>;e\<hookrightarrow>S';\<mu>';e'" using e_tyapp_p by auto
    next
      have "INV6 (e\<langle>\<phi>\<rangle>) Zr" by assumption
      moreover
      have "\<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> \<not> (filterScFreeX X\<union>Z\<phi>) defines \<rho>" by assumption
      ultimately
      show "INV6 e (Zr\<union>filterScFreeX X\<union>Z\<phi>)" by (auto simp add: dom_def)
    next
      have "INV9 (e\<langle>\<phi>\<rangle>)" by assumption
      thus "INV9 e" by auto 
    next
      show "Z \<subseteq> Zl" using t_ptyapp by auto
    qed auto
    have "\<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> \<rho> \<notin> dom (filterScFreeX X\<union>Z\<phi>)" 
      and "Z\<union>filterScFreeX X\<union>Z\<phi> \<subseteq>Zl" and "Zl \<subseteq> Zs"
      by assumption
    hence Hdol: "\<forall>\<rho>. \<rho> is free dollar sc in e' \<longrightarrow> \<not> (filterScFreeX X\<union>Z\<phi>) defines \<rho>"
      using eval_preserves_let_dollars[of e _ Zs e'] Hg20 by blast
    have Hg3: "\<Gamma>e';\<Delta>;{};Ze'\<union>filterScFreeX X\<union>Z\<phi> \<turnstile> e'\<langle>\<phi>\<rangle> : vsubst_ty \<phi> 0 \<tau>" 
      apply rule using Hdol Ht t_ptyapp by auto
    have Hg12: "Ze'\<union>filterScFreeX X\<union>Z\<phi> \<subseteq> Zl'" using Hg12 Hg16 by auto
    have "INV6 (e\<langle>\<phi>\<rangle>) Zr" by assumption
    hence Hg19: "INV6 (e'\<langle>\<phi>\<rangle>) Zr" using Hg19 by (auto simp add: dom_def)
    have Hg21: "INV8 (Ze'\<union>filterScFreeX X\<union>Z\<phi>) (Z\<union>filterScFreeX X\<union>Z\<phi>) e'"
      using Hg21 by (auto simp add: dom_def)
    have Hg22: "INV9 (e'\<langle>\<phi>\<rangle>)" using Hg22 by auto
    show ?thesis
      apply (rule exI[of _ \<Gamma>e'], rule exI[of _ \<Gamma>s'], rule exI[of _ \<Gamma>m'], rule exI[of _ M'],
        rule exI[of _ Zs'], rule exI[of _ "Ze'\<union>filterScFreeX X\<union>Z\<phi>"], rule exI[of _ Zl'])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg12 Hg13 Hg14 Hg15 Hg16 Hg17 Hg18 Hg19 Hg20 Hg21
        Hg22 e_tyapp_p by auto
  next
    case (e_tybeta Sa Sa' \<mu>a z en qz \<pi>z \<phi>)
    have Hz12l: "Z\<union>filterScFreeX X\<union>Z\<phi> \<subseteq> Zl" by assumption
    have "\<Gamma>;\<Delta>;Mz;Z \<turnstile> e : Qual q ScBot (\<forall>.\<tau>^X)" by assumption
    hence "validScope (scopeT (Qual q ScBot (\<forall>.\<tau>^X))) Z" by (rule ty_valid_scope)
    hence Htvalid: "validScope (scopeT (Qual q ScBot (\<forall>.\<tau>^X))) Zl" apply (rule valid_scope_subset) 
      using Hz12l by auto
    have "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" using e_beta t_app by auto
    also have Hgmeq: "\<Gamma>s \<union> \<Gamma>m = \<Gamma>s \<union> (un \<Gamma>m) \<union> \<Gamma>m" 
      by (auto simp add: un_def)
    finally have Hg1: "M;\<Delta>;Zs;Zl \<Turnstile> S:((\<Gamma>s \<union> (un \<Gamma>m)) \<union> \<Gamma>m)" by auto
    have  Hg2: "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M"  using e_beta t_app by auto
    have Hzing: "z\<rhd>(Qual q ScBot (\<forall>.\<tau>^X)) \<in> \<Gamma>" 
      using t_ptyapp e_tybeta by auto
    hence Hzing: "z\<rhd>(Qual q ScBot (\<forall>.\<tau>^X)) \<in> \<Gamma>s\<union>\<Gamma>m" 
      using t_ptyapp e_tybeta by auto
    have Htr: "M;\<Delta>;Zs;Zl \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
    {
      assume Hu: "qz = qU"
      hence Heq: "S = S'" and Hzins: "z\<mapsto>\<Lambda>^qz^\<pi>z[*].en \<in> S"
        using e_tybeta by auto
      have "scopeV (\<Lambda>^qz^\<pi>z[*].en) = scopeT (Qual q ScBot (\<forall>.\<tau>^X))" using Htr
        apply (rule storeTyping_scope_sanity) using Hzing Hzins by auto
      hence Hvalidv: "validScope (scopeV (\<Lambda>^qz^\<pi>z[*].en)) Zl" using Htvalid by auto
      have "\<exists>\<Gamma>v Znew. \<Gamma>v;\<Delta>;M;Znew \<turnstile> \<Lambda>^qz^\<pi>z[*].en : Qual q ScBot (\<forall>.\<tau>^X)
                      \<and> un \<Gamma>v \<subseteq> un (\<Gamma>s \<union> \<Gamma>m) \<and> Znew \<subseteq> Zs"
        using storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" z "\<Lambda>^qz^\<pi>z[*].en"
                                 "Qual q ScBot (\<forall>.\<tau>^X)"] Hzins Hzing Htr Hvalidv
        by auto
      then obtain \<Gamma>v Znew
        where Ht: "\<Gamma>v;\<Delta>;M;Znew \<turnstile> \<Lambda>^qz^\<pi>z[*].en : Qual q ScBot (\<forall>.\<tau>^X)"
        and Hun: "un \<Gamma>v \<subseteq> un (\<Gamma>s \<union> \<Gamma>m)"
        and Hzsub: "Znew \<subseteq> Zs"
        by blast
      hence Hqeq: "qz = q" and H\<pi>_eq: "\<pi>z = ScBot" by auto
      have HqualG: "qualG \<Gamma>v \<preceq> qU" using Hu Ht by auto
      hence Hunv: "un \<Gamma>v = \<Gamma>v" using equal_unrestricted by auto
      have H\<Gamma>v: "\<Gamma>v;\<Delta> \<Turnstile> OK" using Ht ty_wf_env by auto 
      hence H\<Gamma>v_a: "\<Gamma>v;\<Delta>,(a:\<star>) \<Turnstile> OK" using poly_env_ty_cons by auto
      have HM: "M;\<Delta> \<Turnstile> OK" using Ht ty_wf_env_loc by auto 
      obtain \<tau>v Xv La
        where Ha_typ_eq: "Qual q ScBot (\<forall>.\<tau>^X) = Qual qz \<pi>z (\<forall>.\<tau>v^Xv)"
        and Ha_t: "\<forall>a. a \<notin> La \<longrightarrow> \<Gamma>v;\<Delta>,(a:\<star>);M;filterScFreeX X \<turnstile> freshen_tmty a en : freshen_ty a \<tau>v" 
        and Hznew: "minimalValidScope \<pi>z Znew" 
        and Hfin: "finite La"
        using Ht typing_poly_elim[of \<Gamma>v \<Delta> M Znew qz \<pi>z en "Qual q ScBot (\<forall>.\<tau>^X)"]
        by force
      have "\<Gamma>v;\<Delta> \<Turnstile> OK" and "M;\<Delta> \<Turnstile> OK" using Ht ty_wf_env ty_wf_env_loc by auto
      hence "finite \<Gamma>v" and "finite M" by (auto simp add: wf_env_def)
      hence "finite (free_ty_env \<Gamma>v)" and "finite (free_ty_env M)" using free_env_finite by auto
      moreover
      have "finite \<Delta>" using poly_wf_env_ty Ht by (auto simp add: wf_env_ty_def)
      hence "finite (dom \<Delta>)" using dom_finite by auto 
      ultimately
      obtain a where Ha: "\<not> a free in term en \<and> \<not> a free in type \<tau> \<and> a \<notin> free_ty_env \<Gamma>v \<and> a \<notin> dom \<Delta>
        \<and> a \<notin> free_ty_env M \<and> a \<notin> La"
        using Hfin fresh_exists[of "set (free_tmty en)\<union>set (free_ty \<tau>)\<union>free_ty_env \<Gamma>v\<union>dom \<Delta>\<union>free_ty_env M\<union>La"]
        by auto
      have H\<Gamma>subst_var: "\<Gamma>v = vsubst_ty_env (TyFreeVar a) 0 \<Gamma>v" 
        and H\<Gamma>subst_pty: "\<Gamma>v = vsubst_ty_env \<phi> 0 \<Gamma>v"
        using wf_env_subst_eq[of \<Gamma>v \<Delta> \<phi>] wf_env_subst_eq[of \<Gamma>v \<Delta> "TyFreeVar a"] H\<Gamma>v
        by auto
      have HMsubst_var: "M = vsubst_ty_env (TyFreeVar a) 0 M" 
        and HMsubst_pty: "M = vsubst_ty_env \<phi> 0 M"
        using wf_env_subst_eq[of M \<Delta> \<phi>] wf_env_subst_eq[of M \<Delta> "TyFreeVar a"] HM
        by auto
      hence Ht': "vsubst_ty_env (TyFreeVar a) 0 \<Gamma>v;\<Delta>,(a:\<star>);M;filterScFreeX X \<turnstile>
               freshen_tmty a en : freshen_ty a \<tau>"
        using H\<Gamma>subst_var Ha Ha_t Ha_typ_eq by auto
      have "vsubst_ty_env \<phi> 0 \<Gamma>v;\<Delta>;vsubst_ty_env \<phi> 0 M;filterScFreeX X \<turnstile>
                                                          vsubst_tmty \<phi> 0 en : vsubst_ty \<phi> 0 \<tau>" 
        apply (rule substitution_pty[of en _ \<phi> 0 \<Gamma>v \<Delta> M a "filterScFreeX X" \<tau>])
        using Ha Ht' H\<Gamma>v H\<Gamma>subst_var H\<Gamma>subst_pty HM HMsubst_var HMsubst_pty proof -
        show "\<Delta> \<Turnstile>' \<phi>" using t_ptyapp e_tybeta by auto
      next
        show "\<forall>i \<rho>. vsubst_sc_pty (ScFree \<rho>) i \<phi> = \<phi>" using t_ptyapp e_tybeta by auto
      qed auto
      hence "\<Gamma>v;\<Delta>;M;filterScFreeX X \<turnstile> vsubst_tmty \<phi> 0 en : vsubst_ty \<phi> 0 \<tau>" using H\<Gamma>subst_pty HMsubst_pty
        by auto
      moreover
      have "INV3 S" by assumption
      hence "(\<exists>q \<pi>' l. (\<Lambda>^qz^\<pi>z[*].en) = TmLoc q \<pi>' l) \<or> (\<forall>l. \<not> l free loc in (\<Lambda>^qz^\<pi>z[*].en))"
        using INV3[of S] Hzins by blast
      hence "(\<forall>l. \<not> l free loc in (\<Lambda>^qz^\<pi>z[*].en))" by auto
      hence "(\<forall>l. \<not> l free loc in en)" by auto
      hence "(\<forall>l. \<not> l free loc in (vsubst_tmty \<phi> 0 en))" using poly_free_loc_tmty_equal by auto
      ultimately
      have Hg3: "\<Gamma>v;\<Delta>;{};filterScFreeX X \<turnstile> vsubst_tmty \<phi> 0 en : vsubst_ty \<phi> 0 \<tau>"
        using ty_mem_strengthen_all by auto
      note Hun
      also have "un (\<Gamma>s \<union> \<Gamma>m) = un \<Gamma>s \<union> un \<Gamma>m" by (auto simp add: un_def)
      also note Hunv
      finally have "\<Gamma>v \<subseteq> un \<Gamma>s \<union> un \<Gamma>m" by auto
      moreover have "un \<Gamma>s \<subseteq> \<Gamma>s" by (auto simp add: un_def)
      ultimately have "\<Gamma>v \<subseteq> \<Gamma>s \<union> un \<Gamma>m" by auto
      hence Hg4: "\<Gamma>v\<union>\<Gamma>r \<subseteq> \<Gamma>s\<union>(un \<Gamma>m)" using t_ptyapp by auto
      have Hg5: "\<Gamma>v \<sim> \<Gamma>r" using HqualG by (auto simp add: env_compat_def qualGrelL_def)
      have "\<Gamma>s \<union> (un \<Gamma>m);\<Delta> \<Turnstile> OK" using t_ptyapp wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Delta> "\<Gamma>s \<union> (un \<Gamma>m)"]
        by (auto simp add: un_def)
      hence Hg6: "\<Gamma>v\<union>\<Gamma>r;\<Delta> \<Turnstile> OK"  using wf_env_subset[of "\<Gamma>s \<union> (un \<Gamma>m)"] Hg4 by auto
      have "qualG (un \<Gamma>m) \<preceq> qU" by (auto simp add: un_def qualGrelL_def)
      hence Hg7: "(\<Gamma>s\<union>(un \<Gamma>m)) \<sim> \<Gamma>m"
        using t_ptyapp env_compat_weakening_left[of \<Gamma>s \<Gamma>m "un \<Gamma>m"] by auto
      have "\<Gamma>;\<Delta>;Mz;Z \<turnstile> e : Qual q ScBot (\<forall>.\<tau>^X)" by assumption
      hence Hg8: "filterScFreeX X \<subseteq> Zl" using t_ptyapp by auto
      have Hg9: "Zl \<subseteq> Zs" by assumption
      have Hg11: "Zr \<subseteq> Zl" by assumption
      have Hg12: "free_scope_store S = dom Zs" 
        and Hg13: "INV2 (filterStore S Zl)" 
        and Hg14: "INV3 S" 
        and Hg15: "INV4 (filterStore S Zl)" 
        and Hg16: "INV5 S"  
        by assumption
      have "\<not> hasDollars (\<Lambda>^qz^\<pi>z[*].en)" using Hzins Hg16 INV5 by blast
      hence "\<not> hasDollars en" by auto
      hence "\<not> (\<exists>\<rho>. \<rho> is free dollar sc in en)" using not_hasDollars_exists by auto
      hence Hdollars: "\<not> (\<exists>\<rho>. \<rho> is free dollar sc in (vsubst_tmty \<phi> 0 en))" 
        using poly_free_dollars_tmty_equal by auto
      hence Hg17: "INV6 (vsubst_tmty \<phi> 0 en) Zr" by auto
      have Hg18: "INV7 (e\<langle>\<phi>\<rangle>) (vsubst_tmty \<phi> 0 en) Zs" using Hdollars by auto
      have Hg19: "INV8 (filterScFreeX X) (Z\<union>filterScFreeX X\<union>Z\<phi>) (vsubst_tmty \<phi> 0 en)" 
        using INV8 by (auto simp add: dom_def)
      have Hg20: "INV9 (vsubst_tmty \<phi> 0 en)" using Hdollars hasDollars_exists INV9_intro by blast
      have ?thesis
        apply (rule exI[of _ \<Gamma>v], rule exI[of _ "\<Gamma>s\<union>un(\<Gamma>m)"], rule exI[of _ \<Gamma>m], 
               rule exI[of _ M], rule exI[of _ Zs], 
               rule exI[of _ "filterScFreeX X"], rule exI[of _ Zl])
        using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg8 Hg9 Hg11 Hg12 Hg13 Hg14 Hg15 Hg16 Hg17 Hg18 Hg19 Hg20
          e_tybeta Hu
        by auto
    }
    moreover
    {
      assume Hql: "qz = qL"
      hence Heq: "S' = S\<bar>z" and Hzins: "z\<mapsto>\<Lambda>^qz^\<pi>z[*].en \<in> S"
        using e_tybeta by auto
      have "scopeV (\<Lambda>^qz^\<pi>z[*].en) = scopeT (Qual q ScBot (\<forall>.\<tau>^X))" using Htr
        apply (rule storeTyping_scope_sanity) using Hzing Hzins by auto
      hence Hvalidv: "validScope (scopeV (\<Lambda>^qz^\<pi>z[*].en)) Zl" using Htvalid by auto
      have "q = qL"
        using storeTyping_sanity[of M _ Zs Zl S "\<Gamma>s\<union>\<Gamma>m" z " \<Lambda>^qz^\<pi>z[*].en" "Qual q ScBot (\<forall>.\<tau>^X)"]
          Hzins Hzing Htr Hql Hvalidv
        by auto
      hence "\<exists>\<Gamma>v Znew. \<Gamma>v;\<Delta>;M;Znew \<turnstile> \<Lambda>^qz^\<pi>z[*].en : Qual q ScBot (\<forall>.\<tau>^X) \<and>
             M;\<Delta>;Zs;Zl \<Turnstile> S\<bar>z : ((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v) \<and> un \<Gamma>v \<subseteq> un (\<Gamma>s\<union>\<Gamma>m) \<and>
             ((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<sim> \<Gamma>v) \<and> ((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v);\<Delta> \<Turnstile> OK \<and> Znew \<subseteq> Zs"
        using storeTyping_elim[of M _ Zs Zl S  "\<Gamma>s\<union>\<Gamma>m" z "\<Lambda>^qz^\<pi>z[*].en"
                               "Qual q ScBot (\<forall>.\<tau>^X)"] Hzins Hzing Htr Hvalidv
        by auto
      then obtain \<Gamma>v Znew
        where Ht: "\<Gamma>v;\<Delta>;M;Znew \<turnstile> \<Lambda>^qz^\<pi>z[*].en : Qual q ScBot (\<forall>.\<tau>^X)"
        and Hst: "M;\<Delta>;Zs;Zl \<Turnstile> S\<bar>z : ((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v)"
        and Hcomp: "((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<sim> \<Gamma>v)"
        and Hok: "((\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v);\<Delta> \<Turnstile> OK"
        and Hun: "un \<Gamma>v \<subseteq> un (\<Gamma>s \<union> \<Gamma>m)"
        and Hzsum: "Znew \<subseteq> Zs"
        by blast
      hence Hqeq: "qz = q" and Hscope: "ScBot = \<pi>z" by auto
      have Hse: "z\<rhd>(Qual qL ScBot (\<forall>.\<tau>^X)) \<in> \<Gamma>"
        using t_ptyapp e_tybeta Hqeq Hql by auto
      hence Hsz: "z\<rhd>(Qual qL ScBot (\<forall>.\<tau>^X)) \<in> \<Gamma>s"
        using t_ptyapp by auto
      hence "\<not> \<Gamma>m defines z"
        using compat_ok_sanity[of \<Gamma>s \<Gamma>m _ z "Qual qL ScBot (\<forall>.\<tau>^X)"] 
          t_ptyapp Hql Hqeq by auto
      hence Hmz: "\<Gamma>m\<bar>z = \<Gamma>m" by (auto simp add: without_def dom_def)
      note Hst
      also have "(\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v = \<Gamma>s\<bar>z \<union> \<Gamma>v \<union> \<Gamma>m" using Hmz
        by (auto simp add: without_def)
      finally
      have Hg1: "M;\<Delta>;Zs;Zl \<Turnstile> S\<bar>z : \<Gamma>s\<bar>z \<union> \<Gamma>v \<union> \<Gamma>m" by auto
      have Hg2: "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M" using e_beta t_app by auto
      have HM: "M;\<Delta> \<Turnstile> OK" using Ht ty_wf_env_loc by auto 
      obtain \<tau>v Xv La
        where Ha_typ_eq: "Qual q ScBot (\<forall>.\<tau>^X) = Qual qz \<pi>z (\<forall>.\<tau>v^Xv)"
        and Ha_t: "\<forall>a. a \<notin> La \<longrightarrow> \<Gamma>v;\<Delta>,(a:\<star>);M;(filterScFreeX Xv) \<turnstile> freshen_tmty a en : freshen_ty a \<tau>v" 
        and "minimalValidScope \<pi>z Znew" 
        and Ha_finite: "finite La"
        using Ht typing_poly_elim[of \<Gamma>v \<Delta> M Znew qz \<pi>z en "Qual q ScBot (\<forall>.\<tau>^X)"]
        by force
      have "\<Gamma>v;\<Delta> \<Turnstile> OK" and "M;\<Delta> \<Turnstile> OK" using Ht ty_wf_env ty_wf_env_loc by auto
      hence "finite \<Gamma>v" and "finite M" by (auto simp add: wf_env_def)
      hence "finite (free_ty_env \<Gamma>v)" and "finite (free_ty_env M)" using free_env_finite by auto
      moreover
      have "finite \<Delta>" using poly_wf_env_ty Ht by (auto simp add: wf_env_ty_def)
      hence "finite (dom \<Delta>)" using dom_finite by auto 
      ultimately
      obtain a where Ha: "\<not> a free in term en \<and> \<not> a free in type \<tau> \<and> a \<notin> free_ty_env \<Gamma>v \<and> a \<notin> dom \<Delta>
        \<and> a \<notin> free_ty_env M \<and> a \<notin> La"
        using fresh_exists[of "set (free_tmty en)\<union>set (free_ty \<tau>)\<union>free_ty_env \<Gamma>v\<union>dom \<Delta>\<union>free_ty_env M\<union>La"]
        Ha_finite by auto
      have H\<Gamma>v: "\<Gamma>v;\<Delta> \<Turnstile> OK" using Ht ty_wf_env by auto 
      hence H\<Gamma>v_a: "\<Gamma>v;\<Delta>,(a:\<star>) \<Turnstile> OK" using poly_env_ty_cons by auto
      have H\<Gamma>subst_var: "\<Gamma>v = vsubst_ty_env (TyFreeVar a) 0 \<Gamma>v" 
        and H\<Gamma>subst_pty: "\<Gamma>v = vsubst_ty_env \<phi> 0 \<Gamma>v"
        using wf_env_subst_eq[of \<Gamma>v \<Delta> \<phi>] wf_env_subst_eq[of \<Gamma>v \<Delta> "TyFreeVar a"] H\<Gamma>v
        by auto
      have HMsubst_var: "M = vsubst_ty_env (TyFreeVar a) 0 M" 
        and HMsubst_pty: "M = vsubst_ty_env \<phi> 0 M"
        using wf_env_subst_eq[of M \<Delta> \<phi>] wf_env_subst_eq[of M \<Delta> "TyFreeVar a"] HM
        by auto
      have Ht': "vsubst_ty_env (TyFreeVar a) 0 \<Gamma>v;\<Delta>,(a:\<star>);M;filterScFreeX X \<turnstile>
               freshen_tmty a en : freshen_ty a \<tau>"
        using H\<Gamma>subst_var Ha Ha_typ_eq Ha_t by auto
      have "vsubst_ty_env \<phi> 0 \<Gamma>v;\<Delta>;vsubst_ty_env \<phi> 0 M;filterScFreeX X \<turnstile>
                                                                   vsubst_tmty \<phi> 0 en : vsubst_ty \<phi> 0 \<tau>"
        apply (rule substitution_pty[of en _ \<phi> 0 \<Gamma>v \<Delta> M a "filterScFreeX X" \<tau>])
        using Ha Ht' H\<Gamma>v H\<Gamma>subst_var H\<Gamma>subst_pty HM HMsubst_var HMsubst_pty proof -
        show "\<Delta> \<Turnstile>' \<phi>" using t_ptyapp e_tybeta by auto
      next
        show "\<forall>i \<rho>. vsubst_sc_pty (ScFree \<rho>) i \<phi> = \<phi>" using t_ptyapp e_tybeta by auto
      qed auto
      hence "\<Gamma>v;\<Delta>;M;filterScFreeX X \<turnstile> vsubst_tmty \<phi> 0 en : vsubst_ty \<phi> 0 \<tau>" using H\<Gamma>subst_pty HMsubst_pty
        by auto
      moreover
      have "INV3 S" by assumption
      hence "(\<exists>q \<pi>' l. (\<Lambda>^qz^\<pi>z[*].en) = TmLoc q \<pi>' l) \<or> (\<forall>l. \<not> l free loc in (\<Lambda>^qz^\<pi>z[*].en))"
        using INV3[of S] Hzins by blast
      hence "(\<forall>l. \<not> l free loc in (\<Lambda>^qz^\<pi>z[*].en))" by auto
      hence "(\<forall>l. \<not> l free loc in en)" by auto
      hence "(\<forall>l. \<not> l free loc in (vsubst_tmty \<phi> 0 en))" using poly_free_loc_tmty_equal by auto
      ultimately
      have Hg3: "\<Gamma>v;\<Delta>;{};filterScFreeX X \<turnstile> vsubst_tmty \<phi> 0 en : vsubst_ty \<phi> 0 \<tau>"
        using ty_mem_strengthen_all by auto
      have "\<not> \<Gamma>r defines z"
        using compat_ok_sanity[of \<Gamma> \<Gamma>r _ z "Qual qL ScBot (\<forall>.\<tau>^X)"]
          Hql Hqeq Hse t_ptyapp by auto
      moreover
      have "\<Gamma>r \<subseteq> \<Gamma>s" using t_ptyapp by auto
      ultimately
      have Hrsz: "\<Gamma>r \<subseteq> \<Gamma>s\<bar>z" by (auto simp add: dom_def without_def)
      hence Hg4: "\<Gamma>v\<union>\<Gamma>r \<subseteq> \<Gamma>s\<bar>z\<union>\<Gamma>v" by auto
      have "\<Gamma>s\<bar>z \<sim> \<Gamma>v" 
        using Hcomp env_compat_subset[of "(\<Gamma>s\<union>\<Gamma>m)\<bar>z" \<Gamma>v "\<Gamma>s\<bar>z"]
        by (auto simp add: without_def)
      hence "\<Gamma>r \<sim> \<Gamma>v" using Hrsz env_compat_subset[of "\<Gamma>s\<bar>z" \<Gamma>v \<Gamma>r] by auto
      hence Hg5: "\<Gamma>v \<sim> \<Gamma>r" using env_compat_symm by auto
      have "\<Gamma>s\<bar>z \<union> \<Gamma>v;\<Delta> \<Turnstile> OK"
        using Hok wf_env_subset[of "(\<Gamma>s\<union>\<Gamma>m)\<bar>z \<union> \<Gamma>v" \<Delta> "\<Gamma>s\<bar>z \<union> \<Gamma>v"]
        by (auto simp add: without_def)
      hence Hg6: "\<Gamma>v \<union> \<Gamma>r;\<Delta> \<Turnstile> OK"
        using wf_env_subset[of "\<Gamma>s\<bar>z \<union> \<Gamma>v"] Hg4 by auto
      have "\<Gamma>s\<bar>z \<sim> \<Gamma>m" using t_ptyapp env_compat_subset[of \<Gamma>s \<Gamma>m "\<Gamma>s\<bar>z"]
        by (auto simp add: without_def)
      moreover
      have "\<Gamma>m\<bar>z \<sim> \<Gamma>v" 
        using Hcomp env_compat_subset[of "(\<Gamma>s\<union>\<Gamma>m)\<bar>z" \<Gamma>v "\<Gamma>m\<bar>z"]
        by (auto simp add: without_def)
      hence "\<Gamma>m \<sim> \<Gamma>v" using Hmz by auto
      hence "\<Gamma>v \<sim> \<Gamma>m" by (auto simp add: env_compat_def)
      ultimately
      have Hg7: "(\<Gamma>s\<bar>z \<union>\<Gamma>v) \<sim> \<Gamma>m" using env_compat_union_right by auto
      hence Hg8: "filterScFreeX X \<subseteq> Zl" using t_ptyapp by auto
      have Hg9: "Zl \<subseteq> Zs" by assumption
      have Hg11: "Zr \<subseteq> Zl" by assumption
      have "free_scope_store ((S\<bar>z),(z\<mapsto>(\<Lambda>^qz^\<pi>z[*].en)))
            = free_scope_store (S\<bar>z)" 
        using Hscope free_scope_store_cons_empty[of "\<Lambda>^qz^\<pi>z[*].en" z "S\<bar>z"] by auto
      also
      have Hstore: "S \<Turnstile> Store" using storeTyping_wf_store t_ptyapp by auto
      hence "(S\<bar>z),(z\<mapsto>\<Lambda>^qz^\<pi>z[*].en) = S" using Hzins wf_store_defines_without by auto
      also
      have "free_scope_store S = dom Zs" by assumption
      finally 
      have Hg12: "free_scope_store (S\<bar>z) = dom Zs" by auto
      have "S\<bar>z \<subseteq> S" by (auto simp add: without_def)
      hence "filterStore (S\<bar>z) Zl \<subseteq> filterStore S Zl" by (auto simp add: filterStore_def)
      moreover
      have "INV2 (filterStore S Zl)" and  "INV4 (filterStore S Zl)" by assumption
      ultimately
      have Hg13: "INV2 (filterStore (S\<bar>z) Zl)" 
        and Hg15: "INV4 (filterStore (S\<bar>z) Zl)" 
        using inv2_subset inv4_subset by auto
      have "INV3 S" and "INV5 S" by assumption
      moreover
      have "S\<bar>z \<subseteq> S" by (auto simp add: without_def)
      ultimately
      have Hg14: "INV3 (S\<bar>z)" and Hg16: "INV5 (S\<bar>z)" using inv3_subset inv5_subset by auto
      have "INV5 S" by assumption
      hence "\<not> hasDollars (\<Lambda>^qz^\<pi>z[*].en)" using Hzins INV5 by blast
      hence "\<not> hasDollars en" by auto
      hence "\<not> (\<exists>\<rho>. \<rho> is free dollar sc in en)" using not_hasDollars_exists by auto
      hence Hdollars: "\<not> (\<exists>\<rho>. \<rho> is free dollar sc in (vsubst_tmty \<phi> 0 en))" 
        using poly_free_dollars_tmty_equal by auto
      hence Hg17: "INV6 (vsubst_tmty \<phi> 0 en) Zr" using INV6 by auto
      have Hg18: "INV7 (e\<langle>\<phi>\<rangle>) (vsubst_tmty \<phi> 0 en) Zs" using Hdollars INV7 by blast 
      have Hg19: "INV8 (filterScFreeX X) (Z\<union>filterScFreeX X\<union>Z\<phi>) (vsubst_tmty \<phi> 0 en)" 
        by (auto simp add: dom_def)
      have Hg20: "INV9 (vsubst_tmty \<phi> 0 en)" using Hdollars hasDollars_exists INV9_intro by blast
      have ?thesis
        apply (rule exI[of _ \<Gamma>v], rule exI[of _ "\<Gamma>s\<bar>z\<union>\<Gamma>v"], rule exI[of _ \<Gamma>m], 
               rule exI[of _ M], rule exI[of _ Zs], 
               rule exI[of _ "filterScFreeX X"], rule exI[of _ Zl])
        using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg8 Hg9 Hg11 Hg12 Hg13 Hg14 Hg15 Hg16 Hg17 Hg18 Hg19 Hg20
          e_tybeta Hql by auto
    }
    ultimately
    show ?thesis by (cases qz, auto)
  qed auto
next;

      (********************* value cases *********************)
  case (t_unit \<Delta> Z Mz \<Gamma>e q \<pi> e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;TmUnit q \<pi> \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_val Sa \<mu>a v z)
    note value_preservation[of "TmUnit q \<pi>" M \<Delta> Zs Zl S \<Gamma>s \<Gamma>m \<mu> \<Gamma>e Z
                               "Qual q \<pi> TyUnit" \<Gamma>r z Zr]
    also have "S,(z\<mapsto>TmUnit q \<pi>) = S'" using e_val by auto
    also have "TmFreeVar z = e'" using e_val by auto
    also have "\<mu> = \<mu>'" using e_val by auto
    finally show ?thesis apply this using t_unit e_val by auto
  qed auto
next
  case (t_loc \<Delta> Z M \<Gamma>e \<tau> l \<pi> q e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;TmLoc q \<pi> l \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_val Sa \<mu>a v z)
    note value_preservation[of "TmLoc q \<pi> l" M \<Delta> Zs Zl S \<Gamma>s \<Gamma>m \<mu>
                               \<Gamma>e Z "Qual q \<pi> (TyRef \<tau>)" \<Gamma>r z Zr]
    also have "S,(z\<mapsto>TmLoc q \<pi> l) = S'" using e_val by auto
    also have "TmFreeVar z = e'" using e_val by auto
    also have "\<mu> = \<mu>'" using e_val by auto
    finally show ?thesis apply this using e_val t_loc by auto
  qed auto
next
  case (t_abs \<Delta> X Z Mz \<Gamma>e \<tau>1 \<tau>2 e q \<pi> Lx e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;\<lambda>^q^\<pi>[\<tau>1].e \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_val Sa \<mu>a v z)
    have Htyp: "\<Gamma>e;\<Delta>;{};Z \<turnstile> \<lambda>^q^\<pi>[\<tau>1].e : Qual q \<pi> (\<tau>1 \<rightarrow>^X \<tau>2)" 
      apply (rule Typing.t_abs[where L=Lx]) using t_abs by auto
    note value_preservation[of "\<lambda>^q^\<pi>[\<tau>1].e" M \<Delta> Zs Zl S \<Gamma>s \<Gamma>m \<mu> \<Gamma>e Z
                                "Qual q \<pi> (\<tau>1\<rightarrow>^X \<tau>2)" \<Gamma>r z Zr]
    also have "S,(z\<mapsto>\<lambda>^q^\<pi>[\<tau>1].e)=S'" using e_val by auto
    also have "TmFreeVar z = e'" using e_val by auto
    also have "\<mu> = \<mu>'" using e_val by auto
    finally show ?thesis apply this using e_val t_abs Htyp by auto
  qed auto
next
  case (t_poly \<Delta> Z X M \<Gamma>e \<tau> v q \<pi> La e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;\<Lambda>^q^\<pi>[*].v \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_val Sa \<mu>a vo z)
    note value_preservation[of "\<Lambda>^q^\<pi>[*].v" M \<Delta> Zs Zl S \<Gamma>s \<Gamma>m \<mu> \<Gamma>e Z
                                "Qual q \<pi> (\<forall>.\<tau>^X)" \<Gamma>r z Zr]
    also have "S,(z\<mapsto>\<Lambda>^q^\<pi>[*].v)=S'" using e_val by auto
    also have "TmFreeVar z = e'" using e_val by auto
    also have "\<mu> = \<mu>'" using e_val by auto
    finally show ?thesis apply this using t_poly e_val by auto
  qed auto
next
  case (t_var \<Delta> Z M \<Gamma> \<tau> x e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M G Zs)
  have "S;\<mu>;TmFreeVar x \<hookrightarrow> S';\<mu>';e'" .
  thus ?case by (cases, auto)
next;

      (********************* deref cases *********************)
  case (t_deref \<pi> \<Delta> Z Z\<tau> Mz \<Gamma> \<tau> e e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;TmDeref e \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_deref_p Sa Sa' \<mu>a \<mu>a' ea e')
    from t_deref.hyps
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r Zr.
          \<lbrakk> Mz = {}; M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M;  S;\<mu>;e\<hookrightarrow>S';\<mu>';e';
           \<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma> \<sim> \<Gamma>r; \<Gamma> \<union> \<Gamma>r;\<Delta> \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; \<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK;
           Z \<subseteq> Zl; Zr \<subseteq> Zl; Zl \<subseteq> Zs; free_scope_store S = dom Zs; 
           INV2 (filterStore S Zl); INV3 S; INV4 (filterStore S Zl); INV5 S;
           INV6 e Zr; INV9 e; \<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK 
          \<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'.
            (M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e' : Qual qU \<pi> (TyRef \<tau>)) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (Ze' \<subseteq> Zl') \<and> (Zl' \<subseteq> Zs') \<and> (Zs \<subseteq> Zs') \<and> (Zr \<subseteq> Zl') \<and>
            (free_scope_store S' = dom Zs') \<and> (INV2 (filterStore S' Zl')) \<and>
            (INV3 S') \<and> (INV4 (filterStore S' Zl')) \<and> (INV5 S') \<and> (INV6 e' Zr) \<and>
            (INV7 e e' Zs) \<and> (INV8 Ze' Z e') \<and> (INV9 e')"
      by auto
    obtain \<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'
      where Hg1: "M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e' : Qual qU \<pi> (TyRef \<tau>)"
      and Hg4: "\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s'"
      and Hg5: "\<Gamma>e' \<sim> \<Gamma>r"
      and Hg6: "\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg12: "Ze' \<subseteq> Zl'"
      and Hg15: "Zl' \<subseteq> Zs'"
      and Hg13: "Zs \<subseteq> Zs'"
      and Hg16: "Zr\<union>Z\<tau> \<subseteq> Zl'"
      and Hg14: "free_scope_store S' = dom Zs'"
      and Hg10: "INV2 (filterStore S' Zl')"
      and Hg11: "INV3 S'"
      and Hg17: "INV4 (filterStore S' Zl')"
      and Hg18: "INV5 S'"
      and Hg19: "INV6 e' (Zr\<union>Z\<tau>)"
      and Hg20: "INV7 e e' Zs"
      and Hg21: "INV8 Ze' Z e'"
      and Hg22: "INV9 e'"
      using IH[of M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r "Zr\<union>Z\<tau>"] proof rule
      show "Mz = {}" and "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M" 
        and "\<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s" and "\<Gamma> \<sim> \<Gamma>r" and "\<Gamma> \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" and "\<Gamma>s \<sim> \<Gamma>m"
        and "\<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK" and "INV2 (filterStore S Zl)" and "INV3 S"
        and "INV4 (filterStore S Zl)" and "INV5 S"
        and "Zl \<subseteq> Zs" and "free_scope_store S = dom Zs" by assumption
    next
      show "S;\<mu>;e\<hookrightarrow>S';\<mu>';e'" using e_deref_p by auto
    next
      have "INV6 (TmDeref e) Zr" 
        and "\<forall>x. x is free dollar sc in e \<longrightarrow> x \<notin> dom Z\<tau>" by assumption
      thus "INV6 e (Zr\<union>Z\<tau>)"  by (auto simp add: dom_def)
    next
      have "INV9 (TmDeref e)" by assumption
      thus "INV9 e" by auto 
    next
      show "Z \<subseteq> Zl" using t_deref by auto
    next
      show "Zr\<union>Z\<tau> \<subseteq> Zl" using t_deref.prems by auto
    qed auto
    have "Z\<tau> \<subseteq> Zl" using t_deref.prems by auto
    hence Hg12: "Ze'\<union>Z\<tau> \<subseteq> Zl'" using Hg16 Hg12 by (auto simp add: Un_ac)
    have Hg21: "INV8 (Ze'\<union>Z\<tau>) (Z\<union>Z\<tau>) e'" using Hg21 by (auto simp add: dom_def)
    have "qualT \<tau> \<preceq> qU" by assumption
    moreover    
    have "\<forall>\<rho>. \<rho> is free dollar sc in e' \<longrightarrow> \<not> Z\<tau> defines \<rho>" using Hg19 by (auto simp add: dom_def)
    ultimately
    have Hg3: "\<Gamma>e';\<Delta>;{};Ze'\<union>Z\<tau> \<turnstile> TmDeref e' : \<tau>" using Ht t_deref.hyps by auto
    have Hg19: "INV6 (TmDeref e') Zr" using Hg19 by (auto simp add: dom_def)
    show ?thesis
      apply (rule exI[of _ \<Gamma>e'], rule exI[of _ \<Gamma>s'], rule exI[of _ \<Gamma>m'], rule exI[of _ M'],
        rule exI[of _ Zs'], rule exI[of _ "Ze'\<union>Z\<tau>"], rule exI[of _ Zl'])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg12 Hg13 Hg14 Hg15 Hg16 Hg17 Hg18 Hg19 Hg20 Hg21
        Hg22 e_deref_p by auto
  next
    case (e_deref_c Sa Sa' \<mu>a l q x y \<pi>o)
    have Hz12l: "Z\<union>Z\<tau> \<subseteq> Zl" by assumption
    have Ht: "\<Gamma>;\<Delta>;Mz;Z \<turnstile> e : Qual qU \<pi> (TyRef \<tau>)" by assumption
    hence "validScope (scopeT (Qual qU \<pi> (TyRef \<tau>))) Z" by (rule ty_valid_scope) 
    hence Htvalid: "validScope (scopeT (Qual qU \<pi> (TyRef \<tau>))) Zl" apply (rule valid_scope_subset)
      using Hz12l by auto
    have Hx: "x\<rhd>Qual qU \<pi> (TyRef \<tau>) \<in> \<Gamma>" using t_deref e_deref_c by auto
    hence Hsx: "x\<rhd>Qual qU \<pi> (TyRef \<tau>) \<in> \<Gamma>s" using t_deref by auto
    hence "x\<rhd>Qual qU \<pi> (TyRef \<tau>) \<in> (\<Gamma>s \<union> \<Gamma>m)" by auto
    moreover
    have HxinS: "(x\<mapsto>TmLoc q \<pi>o l) \<in> S" using t_deref e_deref_c by auto
    moreover
    have "M;\<Delta>;Zs;Zl \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
    moreover
    hence  "scopeV (TmLoc q \<pi>o l) = scopeT (Qual qU \<pi> (TyRef \<tau>))"
      apply (rule storeTyping_scope_sanity) using Hsx HxinS by auto
    hence Hvalidv: "validScope (scopeV (TmLoc q \<pi>o l)) Zl" using Htvalid by auto
    ultimately
    obtain \<Gamma>v Ze'
      where Ht: "\<Gamma>v;\<Delta>;M;Ze' \<turnstile> TmLoc q \<pi>o l : Qual qU \<pi> (TyRef \<tau>) "
      and Hun: "un \<Gamma>v \<subseteq> un (\<Gamma>s\<union>\<Gamma>m)"
      and HqU: "q = qU" and Hscope: "\<pi>o = \<pi>"
      and Hz: "Ze' \<subseteq> Zs"
      using storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x "TmLoc q \<pi>o l" "Qual qU \<pi> (TyRef \<tau>)"]
      by auto
    hence Hlinm: "l\<rhd>\<tau> \<in> M" by auto
    have Hsok: "\<Gamma>s;\<Delta> \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Delta> \<Gamma>s] by auto
    have Hs: "S' = S" using HqU e_deref_c by auto
    have "qualT \<tau> = qU" using qualT_unrestricted t_deref by auto
    hence Hyingm: "y\<rhd>\<tau> \<in> \<Gamma>m"
      using Hlinm t_deref e_deref_c memTyping_elim_qU[of \<Gamma>m _ \<mu> M l y \<tau>] by auto
    hence Hgeq: "\<Gamma>s,(y:\<tau>) \<union> \<Gamma>m = \<Gamma>s \<union> \<Gamma>m" by auto
    hence Hg1: "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s,(y:\<tau>) \<union> \<Gamma>m)" using t_deref by auto
    have Hg2: "\<Gamma>m;\<Delta> \<Turnstile>' \<mu> : M" by assumption
    have "\<Delta> \<Turnstile>' Qual qU \<pi> (TyRef \<tau>)" using t_deref poly_ty_wf_type by auto
    hence H\<tau>: "\<Delta> \<Turnstile>' \<tau>" by auto
    have H_TY: "\<Delta> \<Turnstile>' TY" using t_deref poly_env_wf_env_ty by auto
    have "minimalValidScope (scopeT \<tau>) Z\<tau>" by assumption
    hence Hg3: "{y\<rhd>\<tau>};\<Delta>;{};Z\<tau> \<turnstile> TmFreeVar y : \<tau>" 
      using H\<tau> H_TY by (auto simp add: wf_env_def without_def qualGrelL_def)
    have Hg4: "{y\<rhd>\<tau>} \<union> \<Gamma>r \<subseteq> \<Gamma>s,(y:\<tau>)" using t_deref by auto
    have Hg5: "{y\<rhd>\<tau>} \<sim> \<Gamma>r" using t_deref by (auto simp add: env_compat_def)
    have "\<Gamma>r;\<Delta> \<Turnstile> OK" using Hsok t_deref wf_env_subset[of \<Gamma>s \<Delta> \<Gamma>r] by auto
    have "\<Gamma>r,(y:\<tau>) \<subseteq> \<Gamma>s\<union>\<Gamma>m" using t_deref Hyingm by auto
    hence Hg6: "{y\<rhd>\<tau>} \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" using wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Delta> "{y\<rhd>\<tau>} \<union> \<Gamma>r"] t_deref 
      by auto
    have Hg7: "\<Gamma>s,(y:\<tau>) \<sim> \<Gamma>m" using t_deref by (auto simp add: env_compat_def)
    have Hg10: "INV2 (filterStore S Zl)"  and Hg11: "INV3 S"
      and Hg17: "INV4 (filterStore S Zl)" and Hg18: "INV5 S" 
      and Hg14: "free_scope_store S = dom Zs" and Hg15: "Zl \<subseteq> Zs" and Hg16: "Zr \<subseteq> Zl" 
      by assumption
    have Hg13: "Z\<tau> \<subseteq> Zl" using t_deref by auto
    have Hg19: "INV8 Z\<tau> (Z\<union>Z\<tau>) (TmFreeVar y)" by (auto simp add: dom_def)
    show ?thesis
      apply (rule exI[of _ "{y\<rhd>\<tau>}"], rule exI[of _ "\<Gamma>s,(y:\<tau>)"], rule exI[of _ "\<Gamma>m"], 
             rule exI[of _ M], rule exI[of _ Zs], 
             rule exI[of _ Z\<tau>], rule exI[of _ Zl])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg13 Hg14 Hg15 Hg16 Hg17 Hg18 Hg19 e_deref_c Hs
    by auto
  qed auto
next;

      (********************* free cases *********************)
  case (t_free \<Delta> Z Z\<tau> Mz \<Gamma> \<tau> e e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;TmFree e \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_free_p Sa Sa' \<mu>a \<mu>a' ea e')
    from t_free.hyps
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r Zr.
          \<lbrakk> Mz = {}; M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M;  S;\<mu>;e\<hookrightarrow>S';\<mu>';e';
           \<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma> \<sim> \<Gamma>r; \<Gamma> \<union> \<Gamma>r;\<Delta> \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; 
           Z \<subseteq> Zl; Zr \<subseteq> Zl; Zl \<subseteq> Zs; free_scope_store S = dom Zs; 
           INV2 (filterStore S Zl); INV3 S; INV4 (filterStore S Zl); INV5 S;
           INV6 e Zr; INV9 e; \<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK 
          \<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'.
            (M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e' : Qual qL ScBot (TyRef \<tau>)) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (Ze' \<subseteq> Zl') \<and> (Zl' \<subseteq> Zs') \<and> (Zs \<subseteq> Zs') \<and> (Zr \<subseteq> Zl') \<and>
            (free_scope_store S' = dom Zs') \<and> (INV2 (filterStore S' Zl')) \<and>
            (INV3 S') \<and> (INV4 (filterStore S' Zl')) \<and> (INV5 S') \<and> (INV6 e' Zr) \<and>
            (INV7 e e' Zs) \<and> (INV8 Ze' Z e') \<and> (INV9 e')"
      by auto
    obtain \<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'
      where Hg1: "M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e' : Qual qL ScBot (TyRef \<tau>)"
      and Hg4: "\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s'"
      and Hg5: "\<Gamma>e' \<sim> \<Gamma>r"
      and Hg6: "\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg12: "Ze' \<subseteq> Zl'"
      and Hg15: "Zl' \<subseteq> Zs'"
      and Hg13: "Zs \<subseteq> Zs'"
      and Hg16: "Zr\<union>Z\<tau> \<subseteq> Zl'"
      and Hg14: "free_scope_store S' = dom Zs'"
      and Hg10: "INV2 (filterStore S' Zl')"
      and Hg11: "INV3 S'"
      and Hg17: "INV4 (filterStore S' Zl')"
      and Hg18: "INV5 S'"
      and Hg19: "INV6 e' (Zr\<union>Z\<tau>)"
      and Hg20: "INV7 e e' Zs"
      and Hg21: "INV8 Ze' Z e'"
      and Hg22: "INV9 e'"
      using IH[of M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r "Zr\<union>Z\<tau>"] proof rule 
      show "Mz = {}" and "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M" 
        and "\<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s" and "\<Gamma> \<sim> \<Gamma>r" and "\<Gamma> \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" and "\<Gamma>s \<sim> \<Gamma>m"
        and "\<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK" and "INV2 (filterStore S Zl)" and "INV3 S"
        and "INV4 (filterStore S Zl)" and "INV5 S" 
        and "Zl \<subseteq> Zs" and "free_scope_store S = dom Zs" by assumption
    next
      show "S;\<mu>;e\<hookrightarrow>S';\<mu>';e'" using e_free_p by auto
    next
      have "INV6 (TmFree e) Zr" 
        and "\<forall>x. x is free dollar sc in e \<longrightarrow> x \<notin> dom Z\<tau>" by assumption
      thus "INV6 e (Zr\<union>Z\<tau>)"  by (auto simp add: dom_def)
    next
      have "INV9 (TmFree e)" by assumption
      thus "INV9 e" by auto 
    next
      show "Zr \<union> Z\<tau> \<subseteq> Zl" using t_free.prems by auto
    next
      show "Z \<subseteq> Zl" using t_free.prems by auto
    qed auto
    have Hg12: "Ze'\<union>Z\<tau> \<subseteq> Zl'" using Hg16 Hg12 by auto
    have Hg21: "INV8 (Ze'\<union>Z\<tau>) (Z\<union>Z\<tau>) (TmFree e')" using Hg21 by (auto simp add: dom_def)
    have Hg22: "INV9 (TmFree e')" using Hg22 by auto
    have "\<forall>\<rho>. \<rho> is free dollar sc in e' \<longrightarrow> \<not> Z\<tau> defines \<rho>" using Hg19 by (auto simp add: dom_def)
    hence Hg3: "\<Gamma>e';\<Delta>;{};Ze'\<union>Z\<tau> \<turnstile> TmFree e' : \<tau>" using Ht by auto
    have Hg19: "INV6 (TmFree e') Zr" using Hg19 by (auto simp add: dom_def)
    show ?thesis
      apply (rule exI[of _ \<Gamma>e'], rule exI[of _ \<Gamma>s'], rule exI[of _ \<Gamma>m'], rule exI[of _ M'],
        rule exI[of _ Zs'], rule exI[of _ "Ze'\<union>Z\<tau>"], rule exI[of _ Zl'])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg12 Hg13 Hg14 Hg15 Hg16 Hg17 Hg18 Hg19 Hg20 Hg21
        Hg22 e_free_p by auto
  next
    case (e_free_c Sa Sa' \<mu>a l q x y \<pi>)
    have Hz12l: "Z\<union>Z\<tau> \<subseteq> Zl" by assumption
    have Ht: "\<Gamma>;\<Delta>;Mz;Z \<turnstile> e : Qual qL ScBot (TyRef \<tau>)" by assumption
    hence "validScope (scopeT (Qual qL ScBot (TyRef \<tau>))) Z" by (rule ty_valid_scope) 
    hence Htvalid: "validScope (scopeT (Qual qL ScBot (TyRef \<tau>))) Zl" apply (rule valid_scope_subset) 
      using Hz12l by auto
    have Hx: "x\<rhd>Qual qL ScBot (TyRef \<tau>) \<in> \<Gamma>" using t_free e_free_c by auto
    hence Hsx: "x\<rhd>Qual qL ScBot (TyRef \<tau>) \<in> \<Gamma>s" using t_free by auto
    hence "x\<rhd>Qual qL ScBot (TyRef \<tau>) \<in> (\<Gamma>s \<union> \<Gamma>m)" by auto
    moreover
    have HxinS: "(x\<mapsto>TmLoc q \<pi> l) \<in> S" using t_free e_free_c by auto
    moreover
    have "M;\<Delta>;Zs;Zl \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
    moreover
    hence  "scopeV (TmLoc q \<pi> l) = scopeT (Qual qL ScBot (TyRef \<tau>))"
      apply (rule storeTyping_scope_sanity) using Hsx HxinS by auto
    hence Hvalidv: "validScope (scopeV (TmLoc q \<pi> l)) Zl" using Htvalid by auto
    ultimately
    obtain \<Gamma>v Ze'
      where Ht: "\<Gamma>v;\<Delta>;M;Ze' \<turnstile> TmLoc q \<pi> l : Qual qL ScBot (TyRef \<tau>)"
      and Hst: "M;\<Delta>;Zs;Zl \<Turnstile> S\<bar>x : ((\<Gamma>s \<union> \<Gamma>m)\<bar>x \<union> \<Gamma>v)"
      and Hcomp: "(\<Gamma>s \<union> \<Gamma>m)\<bar>x \<sim> \<Gamma>v"
      and Hok: "(\<Gamma>s \<union> \<Gamma>m)\<bar>x \<union> \<Gamma>v;\<Delta> \<Turnstile> OK"
      and Hzun: "Ze' \<subseteq> Zs"
      using storeTyping_elim[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x "TmLoc q \<pi> l" "Qual qL ScBot (TyRef \<tau>)"]
      by auto
    have Hsok: "\<Gamma>s;\<Delta> \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Delta> \<Gamma>s] by auto
    have Hascomp: "\<Gamma>s \<sim> \<Gamma>m" and Hasok: "\<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK" by assumption
    hence Hmx: "\<not> \<Gamma>m defines x" using Hsx compat_ok_sanity by auto
    have "\<Gamma>;\<Delta>;Mz;Z\<union>Z\<tau> \<turnstile> TmFree e : \<tau>" using t_free by auto
    have Hql: "q = qL" using Ht by auto
    hence Hs': "S\<bar>x = S'" using e_free_c by auto
    have Hrl: "\<pi> = ScBot" using Ht by auto
    have "INV2 (filterStore S Zl)" by assumption 
    moreover
    have "(x\<mapsto>TmLoc q \<pi> l) \<in> (filterStore S Zl)" 
      using HxinS Hrl by (auto simp add: filterStore_def)
    ultimately
    have Hlnotin: "\<not> l free loc in st ((filterStore S Zl)\<bar>x)" using HxinS Hql by auto
    have "(filterStore (S\<bar>x) Zl) = (filterStore S Zl)\<bar>x" 
      by (auto simp add: filterStore_def without_def)
    hence Hlnotin: "\<not> l free loc in st (filterStore (S\<bar>x) Zl)" using Hlnotin by auto
    have HlinM: "l\<rhd>\<tau> \<in> M" using Ht by fast
    hence "l\<mapsto>TmFreeVar y \<in> \<mu>" and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu> : M"
      using prems by auto
    hence Hmt1: "\<lbrakk> qualT \<tau> = qL \<rbrakk> \<Longrightarrow> \<Gamma>m\<bar>y;\<Delta> \<Turnstile>' \<mu>\<bar>l : M\<bar>l \<and> (y\<rhd>\<tau>) \<in> \<Gamma>m"
      and Hmt2: "\<lbrakk> qualT \<tau> = qU \<rbrakk> \<Longrightarrow> \<Gamma>m;\<Delta> \<Turnstile>' \<mu>\<bar>l : M\<bar>l \<and> (y\<rhd>\<tau>) \<in> \<Gamma>m"
      using HlinM memTyping_elim_qL memTyping_elim_qU by auto
    hence Hmy: "y\<rhd>\<tau> \<in> \<Gamma>m" by (cases "qualT \<tau>", auto)
    note Hst
    also have " (\<Gamma>s\<union>\<Gamma>m)\<bar>x \<union> \<Gamma>v = (\<Gamma>s\<bar>x) \<union> \<Gamma>v \<union> \<Gamma>m"
      using Hmx by (auto simp add: without_def dom_def)
    also have "(\<Gamma>s\<bar>x) \<union> \<Gamma>v \<union> \<Gamma>m = (\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v \<union> \<Gamma>m" using Hmy by auto
    finally have "M;\<Delta>;Zs;Zl \<Turnstile> S':((\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v \<union> \<Gamma>m)" using Hs' by auto 
    hence Hg1u: "M\<bar>l;\<Delta>;Zs;Zl \<Turnstile> S':((\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v \<union> \<Gamma>m)"
      using storeTyping_strengthen Hlnotin Hs' by auto
    have "\<Delta> \<Turnstile>' Qual qL ScBot (TyRef \<tau>)" using t_free poly_ty_wf_type by auto
    hence H\<tau>: "\<Delta> \<Turnstile>' \<tau>" by auto
    have "\<Delta> \<Turnstile>' TY" using t_free poly_env_wf_env_ty by auto
    hence Hg3: "{(y\<rhd>\<tau>)};\<Delta>;{};Z\<tau> \<turnstile> TmFreeVar y : \<tau>" using t_free.hyps H\<tau>
      by (auto simp add: wf_env_def qualGrelL_def without_def)
    have Hasrs: "\<Gamma>r \<subseteq> \<Gamma>s" using prems by auto
    moreover
    have "\<Gamma> \<sim> \<Gamma>r" and "\<Gamma> \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" by assumption
    hence "\<not> \<Gamma>r defines x" using Hx compat_ok_sanity by auto
    ultimately
    have "\<Gamma>r \<subseteq> (\<Gamma>s\<bar>x)" by (auto simp add: without_def dom_def)
    hence "\<Gamma>r \<subseteq> (\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v" by auto
    hence Hg4: "{(y\<rhd>\<tau>)} \<union> \<Gamma>r \<subseteq> (\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v" by auto
    have Hrnotin: "\<not> \<Gamma>r defines y \<or> (y\<rhd>\<tau> \<in> \<Gamma>r \<and> qualT \<tau> = qU)"
    proof -
      {
        assume Hry: "\<Gamma>r defines y"
        have "\<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s" by assumption
        hence "\<Gamma>s defines y" using Hry by (auto simp add: dom_def)
        hence Hsy: "y\<rhd>\<tau> \<in> \<Gamma>s"
          using prems Hmy wf_env_union_sanity[of \<Gamma>s \<Gamma>m] by auto
        hence "y\<rhd>\<tau> \<in> \<Gamma>r" using Hasrs Hsok Hry wf_env_subset_defines by auto
        moreover
        {
          assume Hl: "qualT \<tau> = qL"
          hence "\<not> \<Gamma>m defines y" using Hascomp Hasok Hsy compat_ok_sanity
            by auto
          hence False using Hmy by (auto simp add: dom_def)
        }
        hence "qualT \<tau> = qU" by (cases "qualT \<tau>", auto)
        ultimately
        have "y\<rhd>\<tau> \<in> \<Gamma>r \<and> qualT \<tau> = qU" by auto
      }
      thus ?thesis by auto
    qed
    have Hg5: "{(y\<rhd>\<tau>)} \<sim> \<Gamma>r"
      using Hrnotin by (auto simp add: dom_def env_compat_def)
    have "\<Gamma>r;\<Delta> \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>\<union>\<Gamma>r" \<Delta> \<Gamma>r] by auto
    hence Hg6: "{(y\<rhd>\<tau>)} \<union> \<Gamma>r;\<Delta> \<Turnstile> OK"
      using Hrnotin H\<tau> by (auto simp add: dom_def wf_env_def)
    have Hs: "S\<bar>x \<subseteq> S" by (auto simp add: without_def)
    hence "filterStore (S\<bar>x) Zl \<subseteq>  filterStore S Zl" by (auto simp add: filterStore_def)
    moreover
    have "INV2 (filterStore S Zl)" and "INV4 (filterStore S Zl)" by assumption
    ultimately
    have Hg10: "INV2 (filterStore (S\<bar>x) Zl)" and Hg17: "INV4 (filterStore (S\<bar>x) Zl)" 
      using inv2_subset inv4_subset by auto
    have "INV3 S" and "INV5 S" by assumption
    hence Hg11: "INV3 (S\<bar>x)" and Hg18: "INV5 (S\<bar>x)"using Hs inv3_subset inv5_subset by auto
    have Hg13: "Z\<tau> \<subseteq> Zl" using t_free.prems by auto
    have "free_scope_store ((S\<bar>x),(x\<mapsto>TmLoc q \<pi> l))
          = free_scope_store (S\<bar>x)" 
      using Hrl free_scope_store_cons_empty[of "TmLoc q \<pi> l" x "S\<bar>x"] by auto
    also
    have "S \<Turnstile> Store" using storeTyping_wf_store t_free by auto
    hence "(S\<bar>x),(x\<mapsto>TmLoc q \<pi> l) = S" using HxinS wf_store_defines_without by auto
    also
    have "free_scope_store S = dom Zs" by assumption
    finally have Hg14: "free_scope_store (S\<bar>x) = dom Zs" by auto
    have Hg15: "Zl \<subseteq> Zs" and Hg16: "Zr \<subseteq> Zl" by assumption
    have Hg19: "INV6 (TmFreeVar y) Zr" by auto
    have Hg20: "INV7 (TmFree e) (TmFreeVar y) Zs" by auto
    have Hg21: "INV8 Z\<tau> (Z\<union>Z\<tau>) (TmFreeVar y)" by (auto simp add: dom_def)
    have Hg22: "INV9 (TmFreeVar y)" by auto
    show ?thesis proof (cases "qualT \<tau>")
      case qL
      note Hst
      also
      have Heql: "(\<Gamma>s \<union> \<Gamma>m)\<bar>x \<union> \<Gamma>v = (\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v \<union> \<Gamma>m\<bar>y" proof -
        have Heq: "(\<Gamma>s \<union> \<Gamma>m)\<bar>x = \<Gamma>s\<bar>x \<union> \<Gamma>m" using Hmx
          by (auto simp add: dom_def without_def)
        also
        {
          have "(\<Gamma>s \<union> \<Gamma>m)\<bar>x;\<Delta> \<Turnstile> OK" using Hok wf_env_subset by auto
          hence "\<Gamma>s\<bar>x \<union> \<Gamma>m;\<Delta> \<Turnstile> OK" using Heq by auto
          hence "\<Gamma>s\<bar>x \<union> \<Gamma>m = (\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>m\<bar>y"
            using Hmy env_cons_without_eq_left[of "\<Gamma>s\<bar>x" \<Gamma>m] by auto
        }
        finally show ?thesis by auto
      qed
      also note Hs'
      finally
      have "M;\<Delta>;Zs;Zl \<Turnstile> S':((\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v \<union> \<Gamma>m\<bar>y)" by auto
      hence Hg1l: "M\<bar>l;\<Delta>;Zs;Zl \<Turnstile> S':((\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v \<union> \<Gamma>m\<bar>y)" 
        using storeTyping_strengthen Hlnotin Hs' by auto
      have Hg2: "\<Gamma>m\<bar>y;\<Delta> \<Turnstile>' \<mu>':M\<bar>l" using e_free_c Hmt1 qL by auto
      have Hg7: "((\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v) \<sim> \<Gamma>m\<bar>y" using Hcomp
      proof (auto simp add: env_compat_def)
        assume "(y\<rhd>\<tau>) \<in> \<Gamma>m\<bar>y"
        hence False by (auto simp add: dom_def without_def)
        thus "qualT \<tau> \<preceq> qU" by auto
      next
        fix xa \<tau>
        assume "(xa\<rhd>\<tau>) \<in> \<Gamma>m\<bar>y"
          and "(xa\<rhd>\<tau>) \<in> \<Gamma>s\<bar>x"
        moreover
        have "\<Gamma>s \<sim> \<Gamma>m" by assumption
        ultimately
        show "qualT \<tau> \<preceq> qU"
          by (auto simp add: env_compat_def without_def)
      next
        have "\<Gamma>m\<bar>x \<sim> \<Gamma>v"
          using Hcomp env_compat_subset[of "(\<Gamma>s\<union>\<Gamma>m)\<bar>x" \<Gamma>v "\<Gamma>m\<bar>x"]
          by (auto simp add: without_def)
        also have "\<Gamma>m\<bar>x = \<Gamma>m" using Hmx by (auto simp add: without_def dom_def)
        finally have "\<Gamma>m \<sim> \<Gamma>v" by auto
        thus "\<And>x \<tau>. \<lbrakk>(x\<rhd>\<tau>) \<in> \<Gamma>m\<bar>y; (x\<rhd>\<tau>) \<in> \<Gamma>v\<rbrakk> \<Longrightarrow> qualT \<tau> \<preceq> qU"
          by (auto simp add: env_compat_def without_def)
      qed
      show ?thesis
        apply (rule exI[of _ "{y\<rhd>\<tau>}"],
          rule exI[of _ "(\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v"],
          rule exI[of _ "\<Gamma>m\<bar>y"], rule exI[of _ "M\<bar>l"],
          rule exI[of _ Zs], rule exI[of _ Z\<tau>],
          rule exI[of _ Zl])
        using Hg1l Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg13 Hg14 Hg15 Hg16 Hg17 Hg18 Hg19 Hg20
          Hg21 Hg22 e_free_c Hs' by auto
    next
      case qU
      have Hg2: "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>':M\<bar>l" using e_free_c Hmt2 qU by auto
      have Hg7: "((\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v) \<sim> \<Gamma>m"
      proof (auto simp add: env_compat_def)
        assume "(y\<rhd>\<tau>) \<in> \<Gamma>m"
        thus "qualT \<tau> \<preceq> qU" using qU by auto
      next
        fix xa \<tau>
        assume "(xa\<rhd>\<tau>) \<in> \<Gamma>m"
          and "(xa\<rhd>\<tau>) \<in> \<Gamma>s\<bar>x"
        moreover
        have "\<Gamma>s \<sim> \<Gamma>m" by assumption
        ultimately
        show "qualT \<tau> \<preceq> qU"
          by (auto simp add: env_compat_def without_def)
      next
        have "\<Gamma>m\<bar>x \<sim> \<Gamma>v"
          using Hcomp env_compat_subset[of "(\<Gamma>s\<union>\<Gamma>m)\<bar>x" \<Gamma>v "\<Gamma>m\<bar>x"]
          by (auto simp add: without_def)
        also have "\<Gamma>m\<bar>x = \<Gamma>m" using Hmx by (auto simp add: without_def dom_def)
        finally have "\<Gamma>m \<sim> \<Gamma>v" by auto
        thus "\<And>x \<tau>. \<lbrakk>(x\<rhd>\<tau>) \<in> \<Gamma>m; (x\<rhd>\<tau>) \<in> \<Gamma>v\<rbrakk> \<Longrightarrow> qualT \<tau> \<preceq> qU"
          by (auto simp add: env_compat_def without_def)
      qed
      show ?thesis
        apply (rule exI[of _ "{y\<rhd>\<tau>}"],
               rule exI[of _ "(\<Gamma>s\<bar>x),(y:\<tau>) \<union> \<Gamma>v"],
               rule exI[of _ "\<Gamma>m"], rule exI[of _ "M\<bar>l"],
               rule exI[of _ Zs], rule exI[of _ Z\<tau>],
               rule exI[of _ Zl])
        using Hg1u Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg13 Hg14 Hg15 Hg16 Hg17 Hg18 Hg19 Hg20
          Hg21 Hg22 e_free_c Hs' by auto
    qed
  qed auto
next;

      (********************* let$ cases *********************)
  case (t_let' \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> Mz Z1 Z2 Z \<rho> x e1 e2 \<tau> \<tau>1 \<tau>2 Zf Lx Ly)
  have  "S;\<mu>;let$ \<rho> (TmFreeVar x) e1 in e2 \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_let_p1 So So' \<mu>o \<mu>o' xo eo1 e1' eo2 \<rho>o)
    have Hxin: "x\<rhd>convT \<rho> \<tau> \<in> \<Gamma>" using t_let' e_let_p1 by auto
    hence Hxtin: "x\<rhd>convT \<rho> \<tau> \<in> \<Gamma>s" using t_let' by auto
    have Hx: "\<not> (\<Gamma>1 \<union> \<Gamma>2) defines x" by assumption
    hence Hx1: "\<not> \<Gamma>1 defines x" by (auto simp add: dom_def)
    have Hx2: "\<not> \<Gamma>2 defines x" using Hx by (auto simp add: dom_def)
    have "\<Delta> \<Turnstile>' convT \<rho> \<tau>" using t_let' poly_ty_wf_type by auto
    hence H\<tau>: "\<Delta> \<Turnstile>' \<tau>" by (cases \<tau>, auto)
    have H\<tau>1: "\<Delta> \<Turnstile>' \<tau>1" using t_let' poly_ty_wf_type by auto
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r Zs Zr.
          \<lbrakk> Mz = {}; M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M;
            S;\<mu>; e1 \<hookrightarrow> S';\<mu>'; e1';
            \<Gamma>1,(x:convT \<rho> \<tau>) \<union> \<Gamma>r \<subseteq> \<Gamma>s;
            \<Gamma>1,(x:convT \<rho> \<tau>) \<sim> \<Gamma>r;
            \<Gamma>1,(x:convT \<rho> \<tau>) \<union> \<Gamma>r;\<Delta> \<Turnstile> OK;
            \<Gamma>s \<sim> \<Gamma>m; Zf \<subseteq> Zl; Zr \<subseteq> Zl; Zl \<subseteq> Zs;
            free_scope_store S = dom Zs;
            INV2 (filterStore S Zl); INV3 S; 
            INV4 (filterStore S Zl); INV5 S; 
            INV6 e1 Zr; INV9 e1;
            \<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK \<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'.
            (M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e1' : \<tau>1) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (Ze' \<subseteq> Zl') \<and> (Zl' \<subseteq> Zs') \<and> (Zs \<subseteq> Zs') \<and> (Zr \<subseteq> Zl') \<and>
            (free_scope_store S' = dom Zs') \<and>
            (INV2 (filterStore S' Zl')) \<and> (INV3 S') \<and>
            (INV4 (filterStore S' Zl')) \<and> (INV5 S') \<and>
            (INV6 e1' Zr) \<and>
            (INV7 e1 e1' Zs) \<and>
            (INV8 Ze' Zf e1') \<and> INV9 e1'"
      using t_let'.hyps e_let_p1 by auto
    obtain \<Gamma>e' \<Gamma>s' \<Gamma>m' M' Ze' Zs' Zl'
      where Hg1: "M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M'"
      and Htyp: "\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e1' : \<tau>1"
      and Hsub: "\<Gamma>e' \<union> (\<Gamma> \<union> \<Gamma>2 \<union> \<Gamma>r) \<subseteq> \<Gamma>s'"
      and Hcomp1: "\<Gamma>e' \<sim> (\<Gamma> \<union> \<Gamma>2 \<union> \<Gamma>r)"
      and Hwf1: "\<Gamma>e' \<union> (\<Gamma> \<union> \<Gamma>2 \<union> \<Gamma>r);\<Delta> \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hzsub: "Ze' \<subseteq> Zl'"  
      and Hg12: "Zl' \<subseteq> Zs'"
      and Hzsub2: "Zs \<subseteq> Zs'" 
      and Hzsub3: "Z2\<union>Z\<union>Zr \<subseteq> Zl'"
      and Hg13:"free_scope_store S' = Environ.dom Zs'"
      and Hg10: "INV2 (filterStore S' Zl')" 
      and Hg11: "INV3 S'"
      and Hg17: "INV4 (filterStore S' Zl')" 
      and Hg18: "INV5 S'"
      and Hg19: "INV6 e1' (Z2\<union>Z\<union>Zr)"
      and Hg20: "INV7 e1 e1' Zs"
      and Hg21: "INV8 Ze' Zf e1'"
      and Hg22: "INV9 e1'"
      using IH[of M Zs S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' "\<Gamma>\<union>\<Gamma>2\<union>\<Gamma>r" "Z2\<union>Z\<union>Zr"]
    proof rule
      show "\<Gamma>s \<sim> \<Gamma>m" and "INV2 (filterStore S Zl)" and "INV3 S" 
        and "INV4 (filterStore S Zl)" and "INV5 S" 
        and "Mz ={}" and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M" and "M;\<Delta>;Zs;Zl \<Turnstile> S:\<Gamma>s\<union>\<Gamma>m" and "\<Gamma>s\<union>\<Gamma>m;\<Delta> \<Turnstile> OK"
        and "free_scope_store S = dom Zs" and "Zl \<subseteq> Zs"
        by assumption
    next
      show "S;\<mu>;e1 \<hookrightarrow> S';\<mu>';e1'" using e_let_p1 by auto
    next
      show "\<Gamma>1,(x:convT \<rho> \<tau>) \<union> (\<Gamma> \<union> \<Gamma>2 \<union> \<Gamma>r) \<subseteq> \<Gamma>s" using t_let' Hxtin by auto
    next
      have "\<Gamma>1 \<sim> \<Gamma>" using t_let' env_compat_symm[of \<Gamma> \<Gamma>1] by auto
      hence "\<Gamma>1 \<sim> (\<Gamma>\<union>\<Gamma>2)" using t_let' env_compat_union_left[of \<Gamma>1 \<Gamma> \<Gamma>2] by auto
      moreover
      have "\<Gamma>1 \<sim> \<Gamma>r" using t_let' env_compat_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>1] by auto
      ultimately
      have "\<Gamma>1 \<sim> \<Gamma> \<union> \<Gamma>2 \<union> \<Gamma>r" 
        using env_compat_union_left[of \<Gamma>1 "\<Gamma>\<union>\<Gamma>2" \<Gamma>r] by auto
      thus "\<Gamma>1,(x:convT \<rho> \<tau>) \<sim> \<Gamma> \<union> \<Gamma>2 \<union> \<Gamma>r" 
        using env_compat_weakening_left[of _ _ "{x\<rhd>convT \<rho> \<tau>}"] 
        by (cases \<tau>, auto simp add: env_compat_def qualGrelL_def)
    next
      have "\<Gamma>s;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Delta> \<Gamma>s] by auto
      moreover
      have "\<Gamma>1 \<union> (\<Gamma> \<union> \<Gamma>2 \<union> \<Gamma>r) \<subseteq> \<Gamma>s" using t_let' by (auto simp only: Un_ac)
      hence "\<Gamma>1,(x:convT \<rho> \<tau>) \<union> (\<Gamma> \<union> \<Gamma>2 \<union> \<Gamma>r) \<subseteq> \<Gamma>s"
        using Hxtin by (auto simp only: Un_ac)
      ultimately
      show "\<Gamma>1,(x:convT \<rho> \<tau>) \<union> (\<Gamma> \<union> \<Gamma>2 \<union> \<Gamma>r);\<Delta> \<Turnstile> OK" 
        using wf_env_subset by auto
    next
      have "validScope (scopeT (convT \<rho> \<tau>)) Z" using ty_valid_scope t_let' by auto
      hence "(\<rho>, EmptyScope) \<in> Z" proof (cases \<tau>, auto simp add: dom_def)
        fix b
        assume "(\<rho>, b) \<in> Z"
        thus ?thesis by (cases b, auto)
      qed
      moreover
      have "Z \<subseteq> Zl" using t_let' by auto
      ultimately
      have "(\<rho>, EmptyScope) \<in> Zl" by auto
      moreover
      have "Z1 \<subseteq> Zl" using t_let' by auto
      ultimately
      have Hsub: "Z1,(\<rho>:\<section>) \<subseteq> Zl" by auto
      have "Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>)" by assumption
      thus "Zf \<subseteq> Zl" using Hsub by auto
    next
      show "Z2 \<union> Z \<union> Zr \<subseteq> Zl" using t_let' by (auto simp only: Un_ac)
    next
      have "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<not> (Z \<union> Z2) defines \<rho>" by assumption
      moreover
      have "INV6 (let$ \<rho> TmFreeVar x e1 in e2) Zr" by assumption
      hence "INV6 e1 Zr" by auto
      ultimately
      show "INV6 e1 (Z2 \<union> Z \<union> Zr)" by (auto simp add: dom_def)
    next
      have "INV9 let$ \<rho> TmFreeVar x e1 in e2" by assumption
      thus "INV9 e1"  using inv9_vsubst by auto
    qed auto
    have "\<Gamma> \<union> (\<Gamma>e'\<bar>x) \<union> \<Gamma>2;\<Delta>;{};Z2\<union>(Ze'\<bar>\<rho>)\<union>Z \<turnstile> let$ \<rho> (TmFreeVar x) e1' in e2 : \<tau>2"
    proof 
      have "\<Gamma>e' \<sim> \<Gamma>2" using Hcomp1 env_compat_subset_right[of \<Gamma>e' _ \<Gamma>2] by auto
      thus "(\<Gamma>e'\<bar>x) \<sim> \<Gamma>2" using env_compat_subset[of \<Gamma>e' \<Gamma>2 "\<Gamma>e'\<bar>x"] 
        by (auto simp add: without_def)
    next
      show "\<Gamma> \<sim> \<Gamma>2" by assumption
    next
      have "\<Gamma>e' \<sim> \<Gamma>" using Hcomp1 env_compat_subset_right[of \<Gamma>e' _ \<Gamma>] by auto
      hence "(\<Gamma>e'\<bar>x) \<sim> \<Gamma>" using env_compat_subset[of \<Gamma>e' \<Gamma> "\<Gamma>e'\<bar>x"] 
        by (auto simp add: without_def)
      thus "\<Gamma> \<sim> (\<Gamma>e'\<bar>x)" using env_compat_symm by auto
    next
      show "\<Gamma> \<union> (\<Gamma>e'\<bar>x) \<union> \<Gamma>2;\<Delta> \<Turnstile> OK" 
        using Hwf1 wf_env_subset[of "\<Gamma>e'\<union>(\<Gamma>\<union>\<Gamma>2\<union>\<Gamma>r)" \<Delta> "\<Gamma>\<union>\<Gamma>e'\<bar>x\<union>\<Gamma>2"]
        by (auto simp only: without_def Un_ac)
    next
      have "Ze' \<Turnstile> ScopeSet" using Htyp ty_wf_scope by auto
      thus "(Ze'\<bar>\<rho>) \<Turnstile> ScopeSet" by (auto simp add: wf_scopeset_def without_def)
    next
      let ?r = "scopeT \<tau>1"
      have "validScope (scopeT \<tau>1) Z1" by assumption
      hence "validScope ?r Z1" by (cases \<tau>1, auto)
      moreover
      have "\<not> Z1 defines \<rho>" using t_let' by (auto simp add: dom_def)
      ultimately
      have "?r \<noteq> ScFree \<rho>" by auto
      moreover
      have "validScope (scopeT \<tau>1) Ze'" using Htyp ty_valid_scope by auto
      ultimately
      show "validScope (scopeT \<tau>1) (Ze'\<bar>\<rho>)" proof (cases \<tau>1, auto )
        fix s
        assume "s \<noteq> ScFree \<rho>"
          and "validScope s Ze'"
        thus "validScope s (Ze'\<bar>\<rho>)" by (cases s, auto simp add: dom_def without_def)
      qed
    next
      show "isBangable \<tau>" using t_let' by auto 
    next
      show "\<Gamma>;\<Delta>;{};Z \<turnstile> TmFreeVar x : convT \<rho> \<tau>" using t_let' e_let_p1 by auto
    next
      have "\<not> (\<Gamma>1 \<union> \<Gamma>2) defines x" by assumption
      hence "\<not> \<Gamma>2 defines x" by (auto simp add: dom_def)
      thus "x \<notin> Environ.dom ((\<Gamma>e'\<bar>x) \<union> \<Gamma>2)" by (auto simp add: dom_def without_def)
    next
      have Hx: "x\<rhd>convT \<rho> \<tau> \<in> \<Gamma>s'" using Hxin Hsub by auto
      have Hsub: "\<Gamma>e' \<subseteq> \<Gamma>s'" using Hsub by auto
      have Hok: "\<Gamma>s';\<Delta> \<Turnstile> OK"
        using wf_env_subset[of "\<Gamma>s'\<union>\<Gamma>m'"] Hg1 Hx storeTyping_wf_env[of _ \<Delta> _ _ _ "\<Gamma>s'\<union>\<Gamma>m'"]
        by blast
      have Hok': "\<Gamma>e',(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK" using wf_env_subset[of "\<Gamma>s'"] Hok Hx Hsub by auto
      have Hqual: "qualG {x\<rhd>convT \<rho> \<tau>} \<preceq> qU" by (cases \<tau>, auto simp add: qualGrelL_def)
      have "\<Gamma>e',(x:convT \<rho> \<tau>);\<Delta>;{};Ze' \<turnstile> e1' : \<tau>1"
        using Htyp Hqual Hok' using weakening[where \<Gamma>=\<Gamma>e' and \<Gamma>'="{x\<rhd>convT \<rho> \<tau>}"] by auto
      also
      have "\<Gamma>e',(x:convT \<rho> \<tau>) = (\<Gamma>e'\<bar>x),(x:convT \<rho> \<tau>)" proof (cases "\<Gamma>e' defines x")
        case False thus ?thesis by (auto simp add: dom_def without_def)
      next
        case True
        have "\<Gamma>e' \<union> \<Gamma>s' = \<Gamma>s'" using Hsub by auto
        hence "\<Gamma>e' \<union> \<Gamma>s';\<Delta> \<Turnstile> OK" using Hok by auto
        hence "x\<rhd>convT \<rho> \<tau> \<in> \<Gamma>e'" using Hx True wf_env_union_sanity by auto
        moreover
        have "\<Gamma>e';\<Delta> \<Turnstile> OK" using Hok wf_env_subset Hsub by auto
        ultimately
        have "\<Gamma>e' = (\<Gamma>e'\<bar>x),(x:convT \<rho> \<tau>)" using wf_env_defines_without by auto
        thus ?thesis by auto
      qed
      finally
      show "(\<Gamma>e'\<bar>x),(x:convT \<rho> \<tau>);\<Delta>;{};Ze' \<turnstile> e1' : \<tau>1"
        by auto
    next
      show "\<forall>xa y. \<Gamma>2,(xa:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {xa\<rhd>\<tau>},(y:\<tau>1) \<and> {xa\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1} \<and> 
            xa \<notin> Lx \<and> y \<notin> Ly \<and> xa \<noteq> y \<longrightarrow>
                \<Gamma>2,(xa:\<tau>),(y:\<tau>1);\<Delta>;{};Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm xa e2) : \<tau>2" 
      proof auto
        fix xa y
        assume "\<Gamma>2,(xa:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK"
          and "\<Gamma>2 \<sim> {xa\<rhd>\<tau>},(y:\<tau>1)"
          and "{xa\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
          and "xa \<notin> Lx"
          and "y \<notin> Ly"
          and "xa \<noteq> y"
        thus "\<Gamma>2,(xa:\<tau>),(y:\<tau>1);\<Delta>;{};Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0)
          (vsubst_tm (TmFreeVar xa) 0 e2) : \<tau>2" using t_let' e_let_p1 by auto
      qed
    next
      have "\<not> (Z2 \<union> Z1) defines \<rho>" by assumption
      hence "\<not> Z2 defines \<rho>" by (auto simp add: dom_def)
      thus "\<rho> \<notin> dom (Z2 \<union> (Ze'\<bar>\<rho>))" by (auto simp add: dom_def without_def)
    next
      show "Ze' = (Ze'\<bar>\<rho>) \<or> Ze' = (Ze'\<bar>\<rho>),(\<rho>:\<section>)" proof (cases "Ze' defines \<rho>")
        case True
        hence "(\<rho>, EmptyScope) \<in> Ze'" proof (auto simp add: dom_def)
          fix b
          assume "(\<rho>, b) \<in> Ze'"
          thus ?thesis by (cases b, auto)
        qed
        hence "Ze' = (Ze'\<bar>\<rho>),(\<rho>:\<section>)" proof (auto simp add: without_def dom_def) 
          fix a b
          assume "(a, b) \<in> Ze'"
            and "b \<noteq> EmptyScope"
          thus False by (cases b, auto)
        qed
        thus ?thesis by auto
      next
        case False
        thus ?thesis by (auto simp add: without_def dom_def)
      qed
    next
      have "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom (Z\<union>Z2)"
        by assumption
      moreover
      have "Z2\<union>Z1\<union>Z \<subseteq> Zl" and "Zl \<subseteq> Zs" by assumption
      hence "Z\<union>Z2 \<subseteq> Zs" by auto
      ultimately
      show "\<forall>\<rho>. \<rho> is free dollar sc in e1' \<longrightarrow> \<rho> \<notin> Environ.dom (Z\<union>Z2)" 
        using eval_preserves_let_dollars[of e1 "Z\<union>Z2" Zs e1'] Hg20 free_dollars_vsubst by auto
    next
      have "\<forall>\<rho>'. \<rho>' is free dollar sc in e2 \<longrightarrow> \<rho>' \<notin> Environ.dom (Ze'\<bar>\<rho>)" proof auto
        fix \<rho>'
        assume Hassm: "\<rho>' is free dollar sc in e2"
        and Hfalse: "\<rho>' \<in> dom (Ze'\<bar>\<rho>)"
        show False proof (cases "\<rho> = \<rho>'")
          case True thus False using Hfalse by (auto simp add: without_def dom_def)
        next
          case False
          have H_lx: "finite Lx" and H_ly: "finite Ly" by assumption
          have Hinv9: "INV9 let$ \<rho> TmFreeVar x e1 in e2" by assumption
          have "\<Gamma>2;\<Delta> \<Turnstile> OK" using wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>2] t_let' by auto
          then obtain x y where "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" 
            and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}" and "x \<notin> Lx" and "y \<notin> Ly" and "x \<noteq> y"
            using new_variable_let[where Lx=Lx and Ly=Ly] H\<tau> H\<tau>1 H_lx H_ly by blast
          hence "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta>;Mz;Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2) : \<tau>2"
            using t_let' by auto
          moreover
          have "INV9 (vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2))" using Hinv9 inv9_vsubst_tm by auto
          ultimately
          have H1: "Z2 defines \<rho>'" using free_dollars_in_scopeset Hassm free_dollars_vsubst_tm
            by auto
          {
            assume "\<rho>' is free dollar sc in e1"
            hence "\<not> (Z\<union>Z2) defines \<rho>'" using t_let' by auto
            hence "\<not> Z2 defines \<rho>'" by (auto simp add: dom_def)
            hence False using H1 by auto
          }
          moreover
          {
            assume Hass: "\<not> \<rho>' is free dollar sc in e1"
            { 
              assume "\<rho>' is free dollar sc in e1'"
              hence "\<not> Zs defines \<rho>'" using Hg20 INV7 Hass free_dollars_vsubst by auto
              moreover
              have "Zl \<subseteq> Zs" and "Z2\<union>Z1\<union>Z \<subseteq> Zl" by assumption
              ultimately
              have "\<not> Z2 defines \<rho>'" by (auto simp add: dom_def)
              hence False using H1 by auto
            }
            moreover
            {
              assume Hass2: "\<not> \<rho>' is free dollar sc in e1'"
              have "\<rho>' \<notin> Environ.dom (Z\<union>Z1)" using Hassm t_let' by auto
              hence "\<rho>' \<notin> Environ.dom Z1" by (auto simp add: dom_def)
              moreover
              have "Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>)" by assumption
              ultimately
              have "\<rho>' \<notin> Environ.dom Zf" using False by auto 
              hence "\<not> Ze' defines \<rho>'" using INV8 Hg21 Hass2 free_dollars_vsubst by auto
              hence False using Hfalse by (auto simp add: without_def dom_def)
            }
            ultimately have False by auto
          }
          ultimately show False by auto
        qed
      qed
      moreover
      have "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom (Z\<union>Z1)" by assumption
      hence "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom Z" by (auto simp add: dom_def)
      ultimately
      show "\<forall>\<rho>'. \<rho>' is free dollar sc in e2 \<longrightarrow> \<rho>' \<notin> Environ.dom (Z\<union>(Ze'\<bar>\<rho>))"
        by (auto simp add: dom_def)
    qed
    hence Hg3: "\<Gamma> \<union> (\<Gamma>e'\<bar>x) \<union> \<Gamma>2;\<Delta>;{};Z2\<union>(Ze'\<bar>\<rho>)\<union>Z \<turnstile> e' : \<tau>2" using e_let_p1 by auto
    have Hg4: "\<Gamma> \<union> (\<Gamma>e'\<bar>x) \<union> \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>s'"
      using Hsub by (auto simp only: without_def Un_ac)
    have "\<Gamma> \<sim> \<Gamma>r" using t_let' env_compat_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>] by auto
    moreover
    have "\<Gamma>e' \<sim> \<Gamma>r" using Hcomp1 env_compat_subset_right[of \<Gamma>e' _ \<Gamma>r] by auto
    hence "(\<Gamma>e'\<bar>x) \<sim> \<Gamma>r" using env_compat_subset[of \<Gamma>e' _ "\<Gamma>e'\<bar>x"]
      by (auto simp add: without_def)
    ultimately
    have "\<Gamma> \<union> (\<Gamma>e'\<bar>x) \<sim> \<Gamma>r" using env_compat_union_right by auto
    moreover
    have "\<Gamma>2 \<sim> \<Gamma>r" using t_let' env_compat_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>2] by auto
    ultimately
    have Hg5: "\<Gamma> \<union> (\<Gamma>e'\<bar>x) \<union> \<Gamma>2 \<sim> \<Gamma>r" using env_compat_union_right by auto
    have "\<Gamma> \<union> (\<Gamma>e'\<bar>x) \<union> \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>e' \<union> (\<Gamma> \<union> \<Gamma>2 \<union> \<Gamma>r)" 
      by (auto simp add: dom_def without_def)
    hence Hg6: "\<Gamma> \<union> (\<Gamma>e'\<bar>x) \<union> \<Gamma>2 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" using Hwf1 wf_env_subset by auto
    have Hg14: "Z2 \<union> (Ze'\<bar>\<rho>) \<union> Z \<subseteq> Zl'"
      using Hzsub Hzsub2 Hzsub3 by (auto simp only: Un_ac without_def)
    have Hg15: "Zr \<subseteq> Zl'" using Hzsub3 by auto
    have "INV6 e1' Zr"
      using Hg19 free_dollars_vsubst by (auto simp add: dom_def)
    moreover
    have "INV6 (let$ \<rho> (TmFreeVar x) e1 in e2) Zr" using t_let' by auto
    ultimately
    have "INV6 (let$ \<rho> (TmFreeVar x) e1' in e2) Zr" by auto
    hence Hg19: "INV6 e' Zr" using e_let_p1 by auto
    have "INV7 e1 e1' Zs" using Hg20 free_dollars_vsubst by auto
    hence "INV7 (let$ \<rho> (TmFreeVar x) e1 in e2) (let$ \<rho> (TmFreeVar x) e1' in e2) Zs" 
      by auto
    hence Hg20: "INV7 (let$ \<rho> (TmFreeVar x) e1  in e2) e' Zs" using e_let_p1 by auto
    have "Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>)" by assumption
    hence "INV8 (Z2\<union>(Ze'\<bar>\<rho>)\<union>Z) (Z2\<union>Z1\<union>Z) (let$ \<rho> TmFreeVar x e1' in e2)"
    proof
      assume "Zf = Z1"
      hence "INV8 Ze' Z1 e1'" using Hg21 by auto
      thus "INV8 (Z2 \<union> Ze' \<bar> \<rho> \<union> Z) (Z2 \<union> Z1 \<union> Z) (let$ \<rho> TmFreeVar x e1' in e2)" 
        by (auto simp add: dom_def without_def)
    next
      assume "Zf = Z1,(\<rho>:\<section>)"
      hence "INV8 Ze' (Z1,(\<rho>:\<section>)) e1'" using Hg21 by auto
      thus "INV8 (Z2 \<union> Ze'\<bar>\<rho> \<union> Z) (Z2 \<union> Z1 \<union> Z) (let$ \<rho> TmFreeVar x e1' in e2)" 
        by (auto simp add: dom_def without_def)
    qed
    hence Hg21: "INV8 (Z2\<union>(Ze'\<bar>\<rho>)\<union>Z) (Z2\<union>Z1\<union>Z) e'" using e_let_p1 by auto
    have "INV9 let$ \<rho> TmFreeVar x e1 in e2" by assumption
    hence "INV9 let$ \<rho> TmFreeVar x e1' in e2" using Hg22 using inv9_vsubst by auto
    hence Hg22: "INV9 e'" using e_let_p1 by auto
    show ?thesis
      apply (rule exI[of _ "\<Gamma>\<union>\<Gamma>e'\<bar>x\<union>\<Gamma>2"],
        rule exI[of _ "\<Gamma>s'"], rule exI[of _ "\<Gamma>m'"], rule exI[of _ "M'"],
        rule exI[of _ "Zs'"], rule exI[of _ "Z2\<union>(Ze'\<bar>\<rho>)\<union>Z"],
        rule exI[of _ Zl'])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg12 Hg13 Hg14 Hg15 Hg17 Hg18 Hg19 Hg20 Hg21 Hg22
        Hzsub2 
      by auto
  next
    case (e_let_c So So' \<mu>o xo z \<rho>o v y eo2)
    have Hsin: "(x\<mapsto>v) \<in> S" using e_let_c by auto
    hence Hvalue: "value v" using wf_store_def e_let_c by auto
    hence Hvs: "scopeV (convRevV v) = ScBot" and Hvq: "qualV (convRevV v) = qL" 
      by (cases v, auto)
    moreover
    have "isBangable \<tau>" by assumption
    hence Hts: "scopeT \<tau> = ScBot" and Htq: "qualT \<tau> = qL" by (cases \<tau>, auto)
    have "x\<rhd>convT \<rho> \<tau> \<in> \<Gamma>" using t_let' by auto
    hence "x\<rhd>convT \<rho> \<tau> \<in> \<Gamma>s" using t_let' by auto
    hence Hxinsm: "x\<rhd>convT \<rho> \<tau> \<in> (\<Gamma>s\<union>\<Gamma>m)" by auto
    have "validScope (scopeT (convT \<rho> \<tau>)) Z" using ty_valid_scope t_let' by auto 
    hence "validScope (ScFree \<rho>) Z" by (cases \<tau>, auto)
    hence Hvalid: "validScope (ScFree \<rho>) Zl" apply (rule validScope_weaken) 
      using t_let'.prems by auto
    have Hst: "M;\<Delta>;Zs;Zl \<Turnstile> S:\<Gamma>s \<union> \<Gamma>m" by assumption
    have HscopeV: "scopeV v = ScFree \<rho>" 
      using storeTyping_scope_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x v "convT \<rho> \<tau>"] Hxinsm Hsin Hst
      by (cases \<tau>, auto)
    hence HvalidV: "validScope (scopeV v) Zl" using Hvalid by auto
    obtain \<Gamma>foo Zfoo where HqualV: "qualV v = qU" 
      and Hvtyp: "\<Gamma>foo;\<Delta>;M;Zfoo \<turnstile> v : convT \<rho> \<tau>"
      using storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x v "convT \<rho> \<tau>"] Hxinsm Hsin Hst HvalidV
      by (cases \<tau>, auto)
    have "M;\<Delta>;Zs;Zl \<Turnstile> S,(z\<mapsto>convRevV v) : (\<Gamma>s\<union>\<Gamma>m),(z:\<tau>)" 
      apply (rule storeTyping_without_conv_qU[of _ _ Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x v \<rho> \<tau> z])
    proof-
      show "M;\<Delta>;Zs;Zl \<Turnstile> S:\<Gamma>s \<union> \<Gamma>m" by assumption
    next
      show "x\<mapsto>v \<in> S" using Hsin by auto
    next
      show "isBangable \<tau>" by assumption
    next
      show "z \<notin> dom S" using e_let_c by auto
    next
      show "value v" using Hvalue by auto
    qed (insert HscopeV HqualV Hxinsm Hvalid, auto)
    hence "M;\<Delta>;Zs;Zl \<Turnstile> S,(z\<mapsto>convRevV v):\<Gamma>s,(z:\<tau>) \<union> \<Gamma>m" by auto
    then obtain \<Gamma>s' where "M;\<Delta>;Zs;(Zl\<bar>\<rho>) \<Turnstile> S,(z\<mapsto>convRevV v):\<Gamma>s'"
      and Hgs_sub: "\<Gamma>s,(z:\<tau>) \<union> \<Gamma>m \<subseteq> \<Gamma>s'"
      using storeTyping_strenghten_scopes by blast
    moreover hence Hnewg_eq: "\<Gamma>s' = (\<Gamma>s'-lin(\<Gamma>m)),(z:\<tau>) \<union> \<Gamma>m" by (auto simp add: lin_def)
    ultimately have Hg1: "M;\<Delta>;Zs;Zl\<bar>\<rho> \<Turnstile> S,(z\<mapsto>convRevV v): (\<Gamma>s'-lin(\<Gamma>m)),(z:\<tau>) \<union> \<Gamma>m" by auto
    let ?\<Gamma>s' = "(\<Gamma>s'-lin(\<Gamma>m)),(z:\<tau>)"
    have Hg2: "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>: M" by assumption
    have Hzsm: "\<not> (\<Gamma>s\<union>\<Gamma>m) defines z" using t_let' e_let_c store_dom_not_in by auto
    hence Hzg2: "\<not> \<Gamma>2 defines z" using t_let' by (auto simp only: dom_def)
    have Ht': "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;Mz;Zf \<turnstile> freshen_tm x (freshen_sc \<rho> (TmFreeVar y)) : \<tau>1"
      using t_let' e_let_c by auto 
    have "y \<noteq> x" proof 
      assume Heq: "y = x"
      have "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK" using Ht' ty_wf_env by auto
      moreover
      have "y\<rhd>\<tau>1 \<in> \<Gamma>1,(x:convT \<rho> \<tau>)" using Ht' by auto
      ultimately
      have "\<tau>1 = convT \<rho> \<tau>" using Heq by (auto simp add: wf_env_def)
      hence "validScope (scopeT (convT \<rho> \<tau>)) Z1" using t_let' by auto
      hence "Z1 defines \<rho>" by (cases \<tau>, auto)
      thus False using t_let' by (auto simp add: dom_def)
    qed
    hence Hying1: "y\<rhd>\<tau>1 \<in> \<Gamma>1" using Ht' by auto
    moreover
    have "\<not> \<Gamma>1 defines z" using Hzsm t_let' by (auto simp only: dom_def)
    ultimately
    have Hneq:  "z \<noteq> y" by (auto simp add: dom_def)
    have "\<Delta> \<Turnstile>' convT \<rho> \<tau>" using t_let' poly_ty_wf_type by auto
    hence H\<tau>: "\<Delta> \<Turnstile>' \<tau>" by (cases \<tau>, auto)
    have H\<tau>1: "\<Delta> \<Turnstile>' \<tau>1" using t_let' poly_ty_wf_type by auto
    have "\<Gamma>s;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Delta> \<Gamma>s] by auto
    hence "\<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK" using t_let' wf_env_subset[of \<Gamma>s \<Delta> "\<Gamma>1 \<union> \<Gamma>2"]  by auto
    hence Hwf2_new: "\<Gamma>2,(y:\<tau>1);\<Delta> \<Turnstile> OK" using wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2" _ "\<Gamma>2,(y:\<tau>1)"] Hying1 H\<tau>1 by auto
    hence "\<Gamma>2,(y:\<tau>1),(z:\<tau>);\<Delta> \<Turnstile> OK" using wf_env_cons[of "\<Gamma>2,(y:\<tau>1)"] Hneq Hzg2 H\<tau> by auto
    hence Hy2: "\<Gamma>2,(z:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" by (auto simp only: insert_commute)
    have Hy_2_comp: "{y\<rhd>\<tau>1} \<sim> \<Gamma>2" using Hying1 env_compat_subset[of \<Gamma>1 \<Gamma>2 "{y\<rhd>\<tau>1}"] t_let' by auto
    moreover
    have "{z\<rhd>\<tau>} \<sim> \<Gamma>2" using Hzg2 by (auto simp add: env_compat_def dom_def)
    ultimately
    have Hy1: "\<Gamma>2 \<sim> {z\<rhd>\<tau>},(y:\<tau>1)" 
      using env_compat_symm env_compat_union_right[of "{z\<rhd>\<tau>}" \<Gamma>2 "{y\<rhd>\<tau>1}"] by auto
    have Hy3: "{z\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}" using Hneq by (auto simp add: env_compat_def)
    have "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom (Z \<union> Z1)" by assumption
    moreover
    have "validScope (scopeT \<tau>1) Z1" by assumption
    moreover
    have "isBangable \<tau>" by assumption
    hence "validScope (scopeT \<tau>) Z" by (cases \<tau>, auto)
    ultimately
    have Hdol1: "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> scopeT \<tau>1 \<noteq> ScFree \<rho>" 
      and Hdol2: "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> scopeT \<tau> \<noteq> ScFree \<rho>" 
      by (cases \<tau>1, auto simp add: dom_def)

    have "finite \<Gamma>2" using Hwf2_new by (auto simp add: wf_env_def)
    hence "finite (dom \<Gamma>2)" using dom_finite by auto
    hence
      "finite (Lx\<union>(dom (\<Gamma>2,(y:\<tau>1)))\<union>set (free_tm (vsubst_tm (TmFreeVar y) (Suc 0) e2)))"
      and "finite (Ly\<union>set (free_tm e2)\<union>dom \<Gamma>2)" by auto
    from this new_variable_let[of "\<Gamma>2,(y:\<tau>1)" \<Delta> \<tau> \<tau>1
      "(Lx\<union>(dom (\<Gamma>2,(y:\<tau>1)))\<union>set (free_tm (vsubst_tm (TmFreeVar y) (Suc 0) e2))\<union>{y})"
      "(Ly\<union>set (free_tm e2)\<union>(dom \<Gamma>2)\<union>{y})"]
    obtain k l where "\<Gamma>2,(y:\<tau>1),(k:\<tau>),(l:\<tau>1);\<Delta> \<Turnstile> OK" and Hkl_comp: "\<Gamma>2,(y:\<tau>1) \<sim> {k\<rhd>\<tau>},(l:\<tau>1)"
      and Hk_m: "k \<notin> (Lx\<union>(dom (\<Gamma>2,(y:\<tau>1)))\<union>set (free_tm (vsubst_tm (TmFreeVar y) (Suc 0) e2))\<union>{y})"
      and Hl_m: "l \<notin> (Ly\<union>set (free_tm e2)\<union>(dom \<Gamma>2)\<union>{y})"
      and Hkl_neq: "k \<noteq> l"
      using Hwf2_new H\<tau> H\<tau>1
      by blast
    hence "\<Gamma>2,(k:\<tau>),(l:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2,(k:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2,(y:\<tau>1),(k:\<tau>);\<Delta> \<Turnstile> OK"
      using wf_env_subset[of "\<Gamma>2,(y:\<tau>1),(k:\<tau>),(l:\<tau>1)" \<Delta>  "\<Gamma>2,(k:\<tau>),(l:\<tau>1)"] 
        wf_env_subset[of "\<Gamma>2,(y:\<tau>1),(k:\<tau>),(l:\<tau>1)" \<Delta>  "\<Gamma>2,(k:\<tau>),(y:\<tau>1)"] by (auto simp add: env_permute)
    moreover
    have "\<Gamma>2 \<sim> {k\<rhd>\<tau>},(l:\<tau>1)" using Hkl_comp env_compat_subset by auto
    moreover
    have "\<not> \<Gamma>2 defines k" and "\<not> \<Gamma>2 defines l" and "k \<noteq> y" and "l \<noteq> y" and "k \<noteq> l" 
      using Hk_m Hl_m Hkl_neq by auto
    hence "\<Gamma>2,(k:\<tau>) \<sim> {l\<rhd>\<tau>1}" and "\<Gamma>2,(k:\<tau>) \<sim> {y\<rhd>\<tau>1}" and "{k\<rhd>\<tau>} \<sim> {l\<rhd>\<tau>1}" and "\<Gamma>2,(y:\<tau>1) \<sim> {k\<rhd>\<tau>}"
      using Hkl_comp Hkl_neq Hy_2_comp
      by (auto simp add: env_compat_def dom_def)
    moreover
    have "l \<notin> set (free_tm e2)" using Hl_m by auto
    hence "l \<notin> set (free_tm (freshen_tm k e2))" using not_free_subst Hkl_neq by auto
    ultimately
    have Hkl: "\<Gamma>2,(k:\<tau>),(l:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {k\<rhd>\<tau>},(l:\<tau>1) \<and> {k\<rhd>\<tau>} \<sim> {l\<rhd>\<tau>1}
      \<and> k \<notin> Lx \<and> l \<notin> Ly \<and> k \<noteq> l \<and> \<Gamma>2,(k:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2,(k:\<tau>) \<sim> {y\<rhd>\<tau>1} \<and>
      l \<notin> set (free_tm (freshen_tm k e2)) \<and> l \<notin> Environ.dom (\<Gamma>2,(k:\<tau>)) \<and> \<Gamma>2,(k:\<tau>) \<sim> {l\<rhd>\<tau>1} \<and>
      \<Gamma>2,(y:\<tau>1),(k:\<tau>);\<Delta> \<Turnstile> OK \<and> \<Gamma>2,(y:\<tau>1) \<sim> {k\<rhd>\<tau>} \<and> 
      k \<notin> set (free_tm (vsubst_tm (TmFreeVar y) (Suc 0) e2)) \<and> k \<notin> Environ.dom (\<Gamma>2,(y:\<tau>1))" 
      using Hk_m Hl_m Hkl_neq by auto
    hence Hg3': "\<Gamma>2,(k:\<tau>),(l:\<tau>1);\<Delta>;Mz;Z2 \<turnstile> vsubst_tm (TmFreeVar l) (Suc 0) (freshen_tm k e2): \<tau>2"
      using t_let' by auto
    have Hsub1: "\<forall>\<rho>. \<rho> is free dollar sc in freshen_tm k e2 \<longrightarrow> scopeT \<tau>1 \<noteq> ScFree \<rho>"
      using Hdol1 free_dollars_vsubst_simpl[of e2 k 0] by auto
    have "\<Gamma>2,(k:\<tau>),(y:\<tau>1);\<Delta>;Mz;Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm k e2): \<tau>2" 
      apply (rule substitution[where x=l]) using Hkl Hg3' Hsub1 by auto
    hence Hg3'': "\<Gamma>2,(y:\<tau>1),(k:\<tau>);\<Delta>;Mz;Z2 \<turnstile> freshen_tm k (vsubst_tm (TmFreeVar y) (Suc 0) e2): \<tau>2" 
      using subst_subst_swap[of "Suc 0" 0] by (auto simp add: env_permute)
    have Hsub2: "\<forall>\<rho>. \<rho> is free dollar sc in (vsubst_tm (TmFreeVar y) (Suc 0) e2) \<longrightarrow> scopeT \<tau> \<noteq> ScFree \<rho>"
      using Hdol2 free_dollars_vsubst_simpl[of e2] by auto
    have "\<Gamma>2,(y:\<tau>1),(z:\<tau>);\<Delta>;Mz;Z2 \<turnstile> vsubst_tm (TmFreeVar z) 0 (vsubst_tm (TmFreeVar y) (Suc 0) e2): \<tau>2" 
      apply (rule substitution[where x=k]) using Hkl Hg3'' Hsub2 Hy2 proof (auto simp add: env_permute)
      show "\<Gamma>2,(y:\<tau>1) \<sim> {z\<rhd>\<tau>}" using Hy1 Hy3 by (auto simp add: env_compat_def)
    qed
    hence Hg3: "\<Gamma>2,(z:\<tau>),(y:\<tau>1);\<Delta>;Mz;Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar z) 0 e2): \<tau>2" 
      using subst_subst_swap[of "Suc 0" 0] by (auto simp add: env_permute)
    have "y\<rhd>\<tau>1 \<in> \<Gamma>s" using Hying1 t_let' by (auto simp only: Un_ac)
    moreover
    have "\<Gamma>2 \<subseteq> \<Gamma>s" and "\<Gamma>r \<subseteq> \<Gamma>s" using t_let' by auto
    moreover
    have "\<Gamma>s \<sim> \<Gamma>m" by assumption
    hence "\<Gamma>s - lin(\<Gamma>m) = \<Gamma>s" proof (auto simp add: lin_def env_compat_def qualT_unrestricted)
      fix x \<tau>
      assume "\<forall>x \<tau>. x\<rhd>\<tau> \<in> \<Gamma>s \<and> x\<rhd>\<tau> \<in> \<Gamma>m \<longrightarrow> qualT \<tau> \<preceq> qU"
        and "x\<rhd>\<tau> \<in> \<Gamma>s"
        and "x\<rhd>\<tau> \<in> \<Gamma>m"
        and Hf: "qualT \<tau> = qL"
      hence "qualT \<tau> \<preceq> qU" by blast
      thus False using Hf by (cases "qualT \<tau>", auto)
    qed
    hence "\<Gamma>s \<subseteq> ?\<Gamma>s'" using Hgs_sub by auto
    ultimately
    have Hg4: "\<Gamma>2,(z:\<tau>),(y:\<tau>1) \<union> \<Gamma>r \<subseteq> ?\<Gamma>s'" by auto
    have "\<Gamma>r \<subseteq> \<Gamma>s \<union> \<Gamma>m" using t_let' by auto
    hence Hzgr: "\<not> \<Gamma>r defines z" using Hzsm by (auto simp add: dom_def)
    moreover
    have "\<Gamma>2 \<sim> \<Gamma>r" using t_let' env_compat_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>2] by auto
    ultimately
    have "\<Gamma>2,(z:\<tau>) \<sim> \<Gamma>r" by (auto simp add: dom_def env_compat_def)
    moreover
    have "\<Gamma>1 \<sim> \<Gamma>r" using t_let' env_compat_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>1] by auto
    hence "{y\<rhd>\<tau>1} \<sim> \<Gamma>r" using Hying1 env_compat_subset[of \<Gamma>1 \<Gamma>r "{y\<rhd>\<tau>1}"] t_let' by auto
    ultimately
    have Hg5: "\<Gamma>2,(z:\<tau>),(y:\<tau>1) \<sim> \<Gamma>r" 
      using env_compat_union_right[of _ _ "{y\<rhd>\<tau>1}"] by auto
    have "(\<Gamma>2\<union>\<Gamma>1) \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" 
      using t_let' wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>r" \<Delta> "(\<Gamma>2\<union>\<Gamma>1) \<union> \<Gamma>r"] 
      by (auto simp only: Un_ac)
    hence "\<Gamma>2,(y:\<tau>1) \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" 
      using wf_env_subset[of "(\<Gamma>2\<union>\<Gamma>1) \<union> \<Gamma>r" \<Delta> "\<Gamma>2,(y:\<tau>1) \<union> \<Gamma>r"] Hying1
      by (auto simp only: Un_ac)
    moreover
    have "\<not> (\<Gamma>2,(y:\<tau>1) \<union> \<Gamma>r) defines z" using Hneq Hzgr Hzg2 by (auto simp add: dom_def)
    ultimately
    have "(\<Gamma>2,(y:\<tau>1) \<union> \<Gamma>r),(z:\<tau>);\<Delta> \<Turnstile> OK" using wf_env_cons H\<tau> by auto
    also have "(\<Gamma>2,(y:\<tau>1) \<union> \<Gamma>r),(z:\<tau>) = \<Gamma>2,(z:\<tau>),(y:\<tau>1) \<union> \<Gamma>r" by auto
    finally have Hg6: "\<Gamma>2,(z:\<tau>),(y:\<tau>1) \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" by auto
    have "\<Gamma>s \<sim> \<Gamma>m" by assumption
    moreover
    have "\<not> \<Gamma>m defines z" using Hzsm by (auto simp add: dom_def)
    ultimately
    have Hg7: "?\<Gamma>s' \<sim> \<Gamma>m" proof (auto simp add: env_compat_def dom_def lin_def)
      fix \<tau>
      assume "qualT \<tau> \<noteq> qL"
      thus "qualT \<tau> \<preceq> qU" by (cases "qualT \<tau>", auto)
    qed
    have "Zl \<subseteq> Zs" by assumption
    hence Hg13: "(Zl\<bar>\<rho>) \<subseteq> Zs" by (auto simp add: without_def)
    have "free_scope_store S = dom Zs" by assumption
    also
    from free_scope_store_def
    have "free_scope_store (S,(z\<mapsto>(convRevV v))) = 
          free_scope_store S \<union>  set (free_scope_scope (scopeV (convRevV v)))" 
      by auto
    hence "free_scope_store S = free_scope_store (S,(z\<mapsto>(convRevV v)))"
      using Hvs by auto 
    finally have Hg14: "free_scope_store (S,(z\<mapsto>convRevV v)) = dom Zs" 
      by auto
    have "Zl\<bar>\<rho> \<subseteq> Zl" by (auto simp add: without_def)
    hence Hfsub: "filterStore S (Zl\<bar>\<rho>) \<subseteq> filterStore S Zl" 
    proof (auto simp add: filterStore_def)
      fix x v
      assume "Zl\<bar>\<rho> \<subseteq> Zl"
        and "validScope (scopeV v) (Zl\<bar>\<rho>)"
      thus "validScope (scopeV v) Zl" by (cases "scopeV v", auto simp add: dom_def)
    qed
    have "INV2 (filterStore S Zl)" by assumption
    hence Hfinv2: "INV2 (filterStore S (Zl\<bar>\<rho>))" using inv2_subset Hfsub by auto
    have "INV3 S" by assumption
    hence Hfinv3: "INV3 (filterStore S (Zl\<bar>\<rho>))" using inv3_subset 
      by (auto simp add: filterStore_def)
    have "\<not> S defines z" using e_let_c by auto
    moreover
    have "filterStore S (Zl\<bar>\<rho>) \<subseteq> S" by (auto simp add: filterStore_def)
    ultimately
    have Hfz: "\<not> (filterStore S (Zl\<bar>\<rho>)) defines z" by (auto simp add: dom_def)
    have Hval: "value (convRevV v)" using Hvalue by (cases v, auto)
    have "validScope (scopeT (convT \<rho> \<tau>)) Z" using ty_valid_scope t_let' by auto 
    hence "validScope (ScFree \<rho>) Z" by (cases \<tau>, auto)
    hence "Z defines \<rho>" by auto
    moreover 
    have "Z \<subseteq> Zl" using t_let' by auto
    ultimately
    have Hzlr: "Zl defines \<rho>" by (auto simp add: dom_def)
    hence Hvor: "(\<exists>q l \<pi>. v = TmLoc q \<pi> l) \<or> (\<forall>l. \<not> l free loc in v)"
      using Hsin t_let' by auto
    hence "INV2 ((filterStore S (Zl\<bar>\<rho>)),(z\<mapsto>convRevV v))" proof
      assume  "\<forall>l. l \<notin> set (free_loc v)"
      hence  "\<forall>l. l \<notin> set (free_loc (convRevV v))" using Hvalue by (cases v, auto)
      thus ?thesis using Hfinv2 Hfz inv2_cons by auto
    next
      assume "\<exists>q l \<pi>. v = TmLoc q \<pi> l"
      then obtain l where Hv: "convRevV v = TmLoc qL ScBot l" using Hvs Hvq Hval by (cases v, auto)
      have "\<not> l free loc in st (filterStore S (Zl\<bar>\<rho>))" proof 
        assume "l free loc in st filterStore S (Zl\<bar>\<rho>)"
        then obtain y w where Hyin: "y\<mapsto>w \<in> filterStore S (Zl\<bar>\<rho>)" 
          and Hlfree: "l free loc in w"
          by (auto simp add: free_loc_store_def)
        hence "(\<exists>q l \<pi>. w = TmLoc q \<pi> l) \<or> (\<forall>l. \<not> l free loc in w)" using Hfinv3 by auto
        hence "\<exists>q l \<pi>. w = TmLoc q \<pi> l" using Hlfree by auto
        hence "\<exists>q \<pi>. w = TmLoc q \<pi> l" using Hlfree by auto
        then obtain q \<pi> where Hwis: "w = TmLoc q \<pi> l" by auto
        hence "y\<mapsto>TmLoc q \<pi> l \<in> (filterStore S Zl)" using Hyin Hfsub by auto
        moreover
        have "x\<mapsto>TmLoc qU (ScFree \<rho>) l \<in> S" using Hvalue Hsin Hv HqualV HscopeV by (cases v, auto)
        hence "x\<mapsto>TmLoc qU (ScFree \<rho>) l \<in> (filterStore S Zl)" using Hzlr 
          by (auto simp add: filterStore_def)
        moreover
        have "INV4 (filterStore S Zl)" by assumption
        ultimately
        have "\<pi> = ScFree \<rho>" using inv4_elim by auto
        hence "y\<mapsto>TmLoc q (ScFree \<rho>) l \<in> filterStore S (Zl\<bar>\<rho>)" using Hyin Hwis by auto
        thus False by (auto simp add: filterStore_def without_def dom_def)
      qed
      thus ?thesis 
        using inv2_cons_loc[of "filterStore S (Zl\<bar>\<rho>)" z l qL ScBot] Hfz Hfinv2 Hv
        by auto
    qed
    moreover 
    have "filterStore (S,(z\<mapsto>convRevV v)) (Zl\<bar>\<rho>) \<subseteq> (filterStore S (Zl\<bar>\<rho>)),(z\<mapsto>convRevV v)"
      by (auto simp add: filterStore_def)
    ultimately
    have Hg10: "INV2 ((filterStore (S,(z\<mapsto>convRevV v)) (Zl\<bar>\<rho>)))" 
      using inv2_subset by (auto simp del: INV2)
    have "(\<exists>q l \<pi>. convRevV v = TmLoc q \<pi> l) \<or> (\<forall>l. \<not> l free loc in convRevV v)" using Hvor proof
      assume "\<exists> q l \<pi>. v = TmLoc q \<pi> l"
      then obtain q l \<pi> where "v = TmLoc q \<pi> l" by auto
      then obtain q \<pi> where "convRevV v = TmLoc q \<pi> l" using Hval by (cases v, auto)
      thus ?thesis by auto
    next
      assume "\<forall>l. l \<notin> (set (free_loc v))"
      hence "\<forall>l. l \<notin> (set (free_loc (convRevV v)))" using Hvalue by (cases v, auto)
      thus ?thesis by auto
    qed
    moreover
    have "INV3 S" by assumption
    ultimately
    have Hg11: "INV3 (S,(z\<mapsto>convRevV v))" using t_let by auto
    have "\<not> Z2 defines \<rho>" using t_let' by (auto simp only: dom_def)
    moreover
    have "Z2 \<subseteq> Zl" using t_let' by (auto simp only: Un_ac)
    ultimately
    have Hg15: "Z2 \<subseteq> (Zl\<bar>\<rho>)" by (auto simp add: dom_def without_def)
    have H: "Mz = {}" by assumption
    have "INV6 (let$ \<rho> (TmFreeVar x) e1 in e2) Zr" by assumption
    hence "\<not> Zr defines \<rho>" by auto
    moreover
    have "Zr \<subseteq> Zl" by assumption
    ultimately
    have Hg16: "Zr \<subseteq> (Zl\<bar>\<rho>)" by (auto simp add: dom_def without_def)
    hence Hinv_orig: "INV4 (filterStore S Zl)" using t_let' e_let_c by auto
    moreover
    have Hfilter_sub: "filterStore S (Zl\<bar>\<rho>) \<subseteq> filterStore S Zl"
    proof (auto simp add: filterStore_def)
      fix x v
      assume "validScope (scopeV v) (Zl\<bar>\<rho>)"
      thus "validScope (scopeV v) Zl" by (cases "scopeV v", auto simp add: without_def dom_def)
    qed
    ultimately
    have Hinv: "INV4 (filterStore S (Zl\<bar>\<rho>))" using inv4_subset by (auto simp del: INV4)
    have "INV4 ((filterStore (S) (Zl\<bar>\<rho>)),(z\<mapsto>convRevV v))" using Hvalue proof (cases v)
      case (v_unit q \<pi>) thus ?thesis using Hinv INV4 by auto
    next
      case (v_loc q \<pi> l)
      hence Hv: "convRevV v = TmLoc qL ScBot l" using Hvalue by auto
      have H: "\<forall>x l \<pi>. x\<mapsto>TmLoc qU \<pi> l \<in> (filterStore S (Zl\<bar>\<rho>)) \<longrightarrow> \<not> l free loc in (convRevV v)" 
      proof auto
        fix xa la \<pi>
        assume Hassm: "xa\<mapsto>TmLoc qU \<pi> la \<in> filterStore S (Zl\<bar>\<rho>)"
          and "la free loc in (convRevV v)"
        hence Hl: "la = l" using Hv Hvalue by auto
        have "x\<mapsto>TmLoc qU (ScFree \<rho>) l \<in> S" using Hvalue Hv e_let_c HqualV HscopeV by (cases v, auto)
        hence "x\<mapsto>TmLoc qU (ScFree \<rho>) l \<in> (filterStore S Zl)" 
          using Hzlr by (auto simp add: filterStore_def)
        hence "\<forall>x v. x\<mapsto>v \<in> (filterStore S Zl) \<and> l free loc in v \<longrightarrow> scopeV v = ScFree \<rho>"
          using Hinv_orig by (auto simp add: INV4)
        hence H: "\<And>x v. \<lbrakk> x\<mapsto>v \<in> (filterStore S Zl); l free loc in v\<rbrakk> \<Longrightarrow> scopeV v = ScFree \<rho>"
          by auto
        have Hin: "xa\<mapsto>TmLoc qU \<pi> l \<in> (filterStore S Zl)" using Hfilter_sub Hassm Hl by auto
        have "scopeV (TmLoc qU \<pi> l) = ScFree \<rho>" using H[of xa "TmLoc qU \<pi> l"] Hin by auto
        hence Hsc: "\<pi> = ScFree \<rho>" by auto
        hence "xa\<mapsto>TmLoc qU \<pi> la \<notin> filterStore S (Zl\<bar>\<rho>)"
          by (auto simp add: filterStore_def without_def dom_def)
        thus False using Hassm by auto
      qed
      show ?thesis using Hinv Hv proof (auto simp add: INV4 free_loc_store_def)
        fix x \<pi>
        assume "x\<mapsto>TmLoc qU \<pi> l \<in> filterStore S (Zl\<bar>\<rho>)"
        hence False using H Hv by auto
        thus "ScBot = \<pi>" by auto
      qed
    next
      case (v_abs q \<pi> \<tau>' e)
      moreover
      have "isBangable \<tau>" by assumption
      ultimately
      have False using Hvtyp proof (cases \<tau>)
        fix q' \<pi>' \<phi>
        assume "\<tau> = Qual q' \<pi>' \<phi>"
          and "isBangable \<tau>"
        thus False using Hvtyp v_abs by (cases \<phi>, auto)
      qed
      thus ?thesis by auto
    next 
      case (v_poly q \<pi> e)
      moreover
      have "isBangable \<tau>" by assumption
      ultimately
      have False using Hvtyp proof (cases \<tau>)
        fix q' \<pi>' \<phi>
        assume "\<tau> = Qual q' \<pi>' \<phi>"
          and "isBangable \<tau>"
        thus False using Hvtyp v_poly by (cases \<phi>, auto)
      qed
      thus ?thesis by auto
    qed
    moreover
    have "filterStore (S,(z\<mapsto>convRevV v)) (Zl\<bar>\<rho>) \<subseteq> (filterStore (S) (Zl\<bar>\<rho>)),(z\<mapsto>convRevV v)" 
      by (auto simp add: filterStore_def)
    ultimately
    have Hg17: "INV4 (filterStore (S,(z\<mapsto>convRevV v)) (Zl\<bar>\<rho>))" using inv4_subset by auto
    have "x\<mapsto>v \<in> S" using e_let_c by auto
    moreover
    have Hinv5: "INV5 S" by assumption
    ultimately
    have "\<not> hasDollars v" by (auto simp add: INV5)
    hence Hg18: "INV5 (S,(z\<mapsto>convRevV v))" using Hinv5 Hvalue by (auto simp add: INV5)
    have "INV6 (let$ \<rho> (TmFreeVar x) e1 in e2) Zr" by assumption
    hence Hg19: "INV6 (vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm z e2)) Zr"
      using free_dollars_vsubst_tm by auto
    have Hg20: "INV7 (let$ \<rho> TmFreeVar x e1 in e2) (vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm z e2)) Zs" 
      using free_dollars_vsubst_tm by auto
    have Hg21: "INV8 Z2 (Z2\<union>Z1\<union>Z) (vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm z e2))" 
      by (auto simp add: dom_def)
    have "INV9 let$ \<rho> TmFreeVar x e1 in e2" by assumption
    hence Hg22: "INV9 (vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm z e2))" using inv9_vsubst_tm by auto
    show ?thesis
      apply (rule exI[of _ "\<Gamma>2,(z:\<tau>),(y:\<tau>1)"],
        rule exI[of _ "?\<Gamma>s'"], rule exI[of _ "\<Gamma>m"], rule exI[of _ "M"], 
        rule exI[of _ "Zs"], rule exI[of _ "Z2"], rule exI[of _ "Zl\<bar>\<rho>"]) 
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg13 Hg14 Hg15 Hg16 Hg17 Hg18 Hg19 Hg20 Hg21
        Hg22 H e_let_c
      by auto
  qed auto
next;

      (********************* let! cases *********************)
  case (t_let \<Delta> \<Gamma>1 \<Gamma>2 \<Gamma> Mz Z1 Z2 Z e e1 e2 \<tau> \<tau>1 \<tau>2 Lx Ly)
  have  "S;\<mu>;let! (e) e1 in e2 \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_let_p So So' \<mu>o \<mu>o' eo e' eo1 eo2)
        from t_let e_let_p
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r Zs Zr.
      \<lbrakk> Mz = {}; M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M;
        S;\<mu>; e \<hookrightarrow> S';\<mu>'; e'; \<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma> \<sim> \<Gamma>r; \<Gamma> \<union> \<Gamma>r;\<Delta> \<Turnstile> OK;
        \<Gamma>s \<sim> \<Gamma>m; Z \<subseteq> Zl; Zr \<subseteq> Zl; free_scope_store S = dom Zs; Zl \<subseteq> Zs;
        INV2 (filterStore S Zl); INV3 S; INV4 (filterStore S Zl); INV5 S; 
        INV6 e Zr; INV9 e; \<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK \<rbrakk> \<Longrightarrow>
      \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'.
      (M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
      (\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M') \<and>
      (\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e' : \<tau>) \<and>
      (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
      (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
      (Ze' \<subseteq> Zl') \<and> (Zl' \<subseteq> Zs') \<and> (Zs \<subseteq> Zs') \<and> (Zr \<subseteq> Zl') \<and>
      (free_scope_store S' = dom Zs') \<and>
      (INV2 (filterStore S' Zl')) \<and> (INV3 S') \<and>
      INV4 (filterStore S' Zl') \<and> INV5 S' \<and> INV6 e' Zr \<and> INV7 e e' Zs \<and> INV8 Ze' Z e' \<and> 
      INV9 e'"
      by auto
    obtain \<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'
      where Hg1: "M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e' : \<tau>"
      and Hg4: "\<Gamma>e' \<union> (\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>r) \<subseteq> \<Gamma>s'"
      and H5: "\<Gamma>e' \<sim> (\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>r)"
      and H6: "\<Gamma>e' \<union> (\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>r);\<Delta> \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg12': "Ze' \<subseteq> Zl'"
      and Hg15: "Zl' \<subseteq> Zs'"
      and Hg13: "Zs \<subseteq> Zs'"
      and Hg16: "Z2\<union>Z1\<union>Zr \<subseteq> Zl'"
      and Hg14: "free_scope_store S' = dom Zs'"
      and Hg10: "INV2 (filterStore S' Zl')"
      and Hg11: "INV3 S'"
      and Hg17: "INV4 (filterStore S' Zl')"
      and Hg18: "INV5 S'"
      and Hg19: "INV6 e' (Z2\<union>Z1\<union>Zr)"
      and Hg20: "INV7 e e' Zs"
      and Hg21: "INV8 Ze' Z e'"
      and Hg22: "INV9 e'"
      using IH[of M Zs S \<Gamma>s \<Gamma>m  \<mu> S' \<mu>' "\<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" "(Z2\<union>Z1\<union>Zr)"]
    proof
      show "Mz = {}" and "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M"
        and "INV2 (filterStore S Zl)" and "INV3 S" and  "\<Gamma>s \<sim> \<Gamma>m"
        and "\<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK" and "free_scope_store S = dom Zs" and "Zl \<subseteq> Zs"
        and "INV4 (filterStore S Zl)" and "INV5 S"
        by assumption
    next
      have "Z2 \<union> Z1 \<union> Z \<subseteq> Zl" by assumption
      thus "Z \<subseteq> Zl" by (auto simp only: Un_ac)
    next
      show "S;\<mu>;e\<hookrightarrow>S';\<mu>';e'" using e_let_p by auto
    next
      have "\<Gamma> \<union> \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>s" by assumption
      thus "\<Gamma> \<union> (\<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r) \<subseteq> \<Gamma>s" by (auto simp add: Un_ac)
    next
      have "\<Gamma> \<sim> \<Gamma>2" and "\<Gamma> \<sim> \<Gamma>1" by assumption
      hence H: "\<Gamma> \<sim> (\<Gamma>1 \<union> \<Gamma>2)" using env_compat_union_left by auto
      have "\<Gamma> \<union> \<Gamma>1 \<union> \<Gamma>2 \<sim> \<Gamma>r" by assumption
      hence "\<Gamma> \<sim> \<Gamma>r" using env_compat_subset[of "\<Gamma> \<union> \<Gamma>1 \<union> \<Gamma>2" \<Gamma>r \<Gamma>] by auto
      thus "\<Gamma> \<sim> (\<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r)"
        using H env_compat_union_left[of \<Gamma> "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r] by auto
    next
      have "\<Gamma> \<union> \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" by assumption
      thus "\<Gamma> \<union> (\<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r);\<Delta> \<Turnstile> OK" by (auto simp add: Un_ac)
    next
      have "Z2 \<union> Z1 \<union> Z \<subseteq> Zl" and "Zr \<subseteq> Zl" by assumption
      thus "Z2\<union>Z1\<union>Zr \<subseteq> Zl" by auto
    next
      have "\<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> \<not> (Z1 \<union> Z2) defines \<rho>" by assumption
      moreover
      have "INV6 (let! e e1 in e2) Zr" by assumption
      hence "INV6 e Zr" by auto
      ultimately
      show "INV6 e (Z2 \<union> Z1 \<union> Zr)" by (auto simp add: dom_def)
    next
      have "INV9 let! e e1 in e2" by assumption
      thus "INV9 e" by auto
    qed blast
    have "Z1\<union>Z2 \<Turnstile> ScopeSet" using t_let by (auto simp add: wf_scopeset_def)
    moreover
    have "\<Gamma>1;\<Delta> \<Turnstile> OK" using wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> \<Gamma>1] t_let by auto 
    moreover
    have H\<tau>: "\<Delta> \<Turnstile>' \<tau>" using t_let poly_ty_wf_type by auto
    moreover
    have "finite Lx" using t_let by auto
    ultimately
    obtain \<rho>n x where "\<Gamma>1,(x:convT \<rho>n \<tau>);\<Delta> \<Turnstile> OK" and
      "\<Gamma>1 \<sim> {x\<rhd>convT \<rho>n \<tau>}" and H: "\<rho>n \<notin> set (free_let_dollar_scope e1)" 
      and "\<rho>n \<notin> Environ.dom (Z2 \<union> Z1)" and "x \<notin> Lx"
      using new_scope_let[of Z1 Z2 Lx \<Gamma>1 \<Delta> \<tau> e1] by auto
    hence Htyp1: "\<Gamma>1,(x:convT \<rho>n \<tau>);\<Delta>;Mz;Z1 \<turnstile> freshen_tm x (freshen_sc \<rho>n e1) : \<tau>1
      \<or> \<Gamma>1,(x:convT \<rho>n \<tau>);\<Delta>;Mz;Z1,(\<rho>n:\<section>) \<turnstile> freshen_tm x (freshen_sc \<rho>n e1) : \<tau>1" 
      using t_let.hyps by auto
    hence H\<tau>1: "\<Delta> \<Turnstile>' \<tau>1" using poly_ty_wf_type by auto
    have Hg3: "\<Gamma>e'\<union>\<Gamma>1\<union>\<Gamma>2;\<Delta>;{};Z2\<union>Z1\<union>Ze' \<turnstile> let! (e') e1 in e2 : \<tau>2" proof
      show "\<Gamma>e' \<sim> \<Gamma>2" using H5 env_compat_subset_right[of \<Gamma>e' "\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>r" \<Gamma>2]
        by auto
    next
      show "\<Gamma>e' \<sim> \<Gamma>1" using H5 env_compat_subset_right[of \<Gamma>e' "\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>r" \<Gamma>1]
        by auto
    next
      show "\<Gamma>e' \<union> \<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK" 
        using H6 wf_env_subset[of "\<Gamma>e' \<union> (\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>r)" \<Delta> "\<Gamma>e'\<union>\<Gamma>1\<union>\<Gamma>2"]
        by (auto simp add: Un_ac)
    next
      show "\<forall>x \<rho>. \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK \<and> \<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>} \<and> x \<notin> Lx \<and>
        \<rho> \<notin> set (free_let_dollar_scope e1) \<and> \<rho> \<notin> Environ.dom (Z2 \<union> Z1) \<longrightarrow>
        (\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;{};Z1 \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1 \<or>
         \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;{};Z1,(\<rho>:\<section>) \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1)"
      proof auto
        fix x \<rho>
        assume "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK"
          and "\<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>}"
          and "\<rho> \<notin> Environ.dom (Z2 \<union> Z1)"
          and "\<rho> \<notin> set (free_let_dollar_scope e1)"
          and "x \<notin> Lx"
          and Hfalse: "\<not> \<Gamma>1,(x:convT \<rho> \<tau>); \<Delta>; {}; Z1,(\<rho>:\<section>) \<turnstile> vsubst_tm (TmFreeVar x) 0           
            (vsubst_sc (ScFree \<rho>) 0 e1) : \<tau>1" 
        hence "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>; Mz; Z1 \<turnstile> 
            vsubst_tm (TmFreeVar x) 0 (vsubst_sc (ScFree \<rho>) 0 e1) : \<tau>1 \<or>
          \<Gamma>1,(x:convT \<rho> \<tau>); \<Delta>; Mz; Z1,(\<rho>:\<section>) \<turnstile> 
            vsubst_tm (TmFreeVar x) 0 (vsubst_sc (ScFree \<rho>) 0 e1) : \<tau>1" 
          using t_let.hyps by (auto simp add: dom_def)
        thus "\<Gamma>1,(x:convT \<rho> \<tau>); \<Delta>; {}; Z1 \<turnstile> 
            vsubst_tm (TmFreeVar x) 0 (vsubst_sc (ScFree \<rho>) 0 e1) : \<tau>1" 
          using Hfalse t_let.prems by auto
      qed
    next
      show "\<forall>x y. \<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1) \<and> {x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1} \<and> x \<notin> Lx \<and> y \<notin> Ly \<and> x \<noteq> y
        \<longrightarrow> 
              \<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta>;{};Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2) : \<tau>2" proof auto
        fix x y
        assume "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
          and "x \<notin> Lx" and "y \<notin> Ly" and "x \<noteq> y"
        thus "\<Gamma>2,(x:\<tau>),(y:\<tau>1); \<Delta>; {};Z2 \<turnstile> 
            vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar x) 0 e2) : \<tau>2"
          using t_let e_let_p by (auto simp add: dom_def)
      qed
    next
      show "\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e' : \<tau>" using Ht by auto
    next
      have "Z1\<union>Z2 \<subseteq> Zs" using t_let by (auto simp only: Un_ac)
      moreover
      have "\<forall>\<rho>. \<rho> is free dollar sc in e \<longrightarrow> \<rho> \<notin> Environ.dom (Z1 \<union> Z2)"
        using t_let by auto
      ultimately 
      show "\<forall>\<rho>. \<rho> is free dollar sc in e' \<longrightarrow> \<rho> \<notin> Environ.dom (Z1 \<union> Z2)" 
        using fresh_dollars_are_free Hg20 by blast
    next
      have "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom Ze'" proof auto
        fix \<rho> 
        assume Hassm: "\<rho> is free dollar sc in e1"
        and Hfalse: "\<rho> \<in> dom Ze'"
        have H1: "Z1 defines \<rho>" using Htyp1 proof 
          assume "\<Gamma>1,(x:convT \<rho>n \<tau>);\<Delta>;Mz;Z1 \<turnstile> freshen_tm x (freshen_sc \<rho>n e1) : \<tau>1"
          moreover
          have "INV9 let! e e1 in e2" by assumption
          hence "INV9 freshen_tm x (freshen_sc \<rho>n e1)" using inv9_vsubst by auto
          ultimately 
          show "Z1 defines \<rho>" using free_dollars_in_scopeset Hassm free_dollars_vsubst by auto
        next
          assume "\<Gamma>1,(x:convT \<rho>n \<tau>);\<Delta>;Mz;Z1,(\<rho>n:\<section>) \<turnstile> freshen_tm x (freshen_sc \<rho>n e1) : \<tau>1" 
          moreover
          have "INV9 let! e e1 in e2" by assumption
          hence "INV9 freshen_tm x (freshen_sc \<rho>n e1)" using inv9_vsubst by auto
          ultimately
          have "Z1,(\<rho>n:\<section>) defines \<rho>" 
            using free_dollars_in_scopeset[where Z="Z1,(\<rho>n:\<section>)"] Hassm free_dollars_vsubst 
            by auto
          moreover
          have "\<rho> \<noteq> \<rho>n" using Hassm H by auto
          ultimately
          show "Z1 defines \<rho>" by (auto simp add: dom_def)
        qed
        {
          assume "\<rho> is free dollar sc in e"
          hence "\<not> (Z1\<union>Z2) defines \<rho>" using t_let by auto
          hence "\<not> Z1 defines \<rho>" by (auto simp add: dom_def)
          hence False using H1 by auto
        }
        moreover
        {
          assume Hass: "\<not> \<rho> is free dollar sc in e"
          { 
            assume "\<rho> is free dollar sc in e'"
            hence "\<not> Zs defines \<rho>" using Hg20 INV7 Hass by auto
            moreover
            have "Zl \<subseteq> Zs" and "Z2\<union>Z1\<union>Z \<subseteq> Zl" by assumption
            ultimately
            have "\<not> Z1 defines \<rho>" by (auto simp add: dom_def)
            hence False using H1 by auto
          }
          moreover
          {
            assume Hass2: "\<not> \<rho> is free dollar sc in e'"
            have "\<rho> \<notin> Environ.dom (Z2 \<union> Z)" using Hassm t_let by auto
            hence "\<rho> \<notin> Environ.dom Z" by (auto simp add: dom_def)
            hence "\<not> Ze' defines \<rho>" using INV8 Hg21 Hass2 by auto
            hence False using Hfalse by auto
          }
          ultimately have False by auto
        }
        ultimately show False by auto
      qed
      moreover
      have "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom (Z2 \<union> Z)" by assumption
      hence "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom Z2" by (auto simp add: dom_def)
      ultimately
      show "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom (Z2 \<union> Ze')" 
        by (auto simp add: dom_def)
    next
      have "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom Ze'" proof auto
        fix \<rho>
        assume Hassm: "\<rho> is free dollar sc in e2"
        and Hfalse: "\<rho> \<in> dom Ze'" 
        have "\<Gamma>2;\<Delta> \<Turnstile> OK" using wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" _ \<Gamma>2] t_let by auto 
        moreover
        have "finite Lx" and "finite Ly" by assumption
        ultimately
        obtain x y where "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {x\<rhd>\<tau>},(y:\<tau>1)" 
          and "{x\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}" and "x \<notin> Lx" and "y \<notin> Ly" and "x \<noteq> y"
          using new_variable_let[where Lx=Lx and Ly=Ly] H\<tau> H\<tau>1 by blast
        hence "\<Gamma>2,(x:\<tau>),(y:\<tau>1);\<Delta>;Mz;Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2) : \<tau>2"
          using t_let by auto
        moreover
        have "INV9 let! e e1 in e2" by assumption
        hence "INV9 vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm x e2)" using inv9_vsubst_tm by auto
        ultimately
        have H1: "Z2 defines \<rho>"
          using free_dollars_in_scopeset Hassm free_dollars_vsubst_tm by auto
        {
          assume "\<rho> is free dollar sc in e"
          hence "\<not> (Z1\<union>Z2) defines \<rho>" using t_let by auto
          hence "\<not> Z2 defines \<rho>" by (auto simp add: dom_def)
          hence False using H1 by auto
        }
        moreover
        {
          assume Hass: "\<not> \<rho> is free dollar sc in e"
          { 
            assume "\<rho> is free dollar sc in e'"
            hence "\<not> Zs defines \<rho>" using Hg20 INV7 Hass free_dollars_vsubst by auto
            moreover
            have "Zl \<subseteq> Zs" and "Z2\<union>Z1\<union>Z \<subseteq> Zl" by assumption
            ultimately
            have "\<not> Z2 defines \<rho>" by (auto simp add: dom_def)
            hence False using H1 by auto
          }
          moreover
          {
            assume Hass2: "\<not> \<rho> is free dollar sc in e'"
            have "\<rho> \<notin> Environ.dom (Z1\<union>Z)" using Hassm t_let by auto
            hence "\<rho> \<notin> Environ.dom Z" by (auto simp add: dom_def)
            hence "\<not> Ze' defines \<rho>" using INV8 Hg21 Hass2 free_dollars_vsubst by auto
            hence False using Hfalse by (auto simp add: without_def dom_def)
          }
          ultimately have False by auto
        }
        ultimately show False by auto
      qed
      moreover
      have "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom (Z1\<union>Z)" by assumption 
      hence "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom Z1" by (auto simp add: dom_def)
      ultimately
      show "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom (Z1\<union>Ze')"
        by (auto simp add: dom_def)
    qed
    have "\<Gamma>e' \<sim> \<Gamma>r" using H5 env_compat_subset_right[of \<Gamma>e' "\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>r" \<Gamma>r] by auto
    moreover
    have "\<Gamma>1 \<sim> \<Gamma>r" using t_let env_compat_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>1] by auto
    ultimately
    have "\<Gamma>e'\<union>\<Gamma>1 \<sim> \<Gamma>r" using env_compat_union_right by auto
    moreover
    have "\<Gamma>2 \<sim> \<Gamma>r" using t_let env_compat_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>2] by auto
    ultimately
    have Hg5: "\<Gamma>e'\<union>\<Gamma>1\<union>\<Gamma>2 \<sim> \<Gamma>r" using env_compat_union_right by auto
    have Hg6: "\<Gamma>e' \<union> \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" using H6 by (auto simp add: Un_ac)
    have "Z2\<union>Z1 \<subseteq> Zl'" using Hg16 by auto
    hence Hg12: "Z2\<union>Z1\<union>Ze' \<subseteq> Zl'" using Hg12' Hg15 by auto
    have Hg16: "Zr \<subseteq> Zl'" using Hg16 by auto
    have "INV6 (let! (e) e1 in e2) Zr" by assumption
    hence Hg19: "INV6 (let! (e') e1 in e2) Zr" using Hg19 by (auto simp add: dom_def)
    have Hg20: "INV7 (let! (e) e1 in e2) (let! (e') e1 in e2) Zs" using Hg20 by auto
    have Hg21: "INV8 (Z2\<union>Z1\<union>Ze') (Z2\<union>Z1\<union>Z) (let! (e') e1 in e2)" 
      using Hg21 by (auto simp add: dom_def)
    have "INV9 let! e e1 in e2" by assumption
    hence Hg22: "INV9 let! e' e1 in e2" using Hg22 by auto
    show ?thesis
      apply (rule exI[of _ "\<Gamma>e'\<union>\<Gamma>1\<union>\<Gamma>2"],
        rule exI[of _ "\<Gamma>s'"], rule exI[of _ "\<Gamma>m'"], 
        rule exI[of _ "M'"], rule exI[of _ "Zs'"],
        rule exI[of _ "Z2\<union>Z1\<union>Ze'"], rule exI[of _ "Zl'"])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg12 Hg13 Hg14 Hg15 Hg16 Hg17 Hg18
        Hg19 Hg20 Hg21 Hg22 e_let_p 
      by auto
  next
    case (e_let_p' S' S \<mu> x e1o e2o \<rho> v)
    have Hbang: "isBangable \<tau>" by assumption
    hence Hlin: "qualT \<tau> = qL" and Hbot: "scopeT \<tau> = ScBot" by auto
    have Hst: "M;\<Delta>;Zs;Zl \<Turnstile> S:\<Gamma>s \<union> \<Gamma>m" using t_let e_let_p' by auto
    hence Hstr: "M;\<Delta>;Zs;Zl,(\<rho>:\<section>) \<Turnstile> S:\<Gamma>s \<union> \<Gamma>m" apply (rule storeTyping_weaken_scopes) proof -
      show "\<rho> \<notin> free_scope_store S" by assumption    
    qed
    have HxinS: "x\<mapsto>v \<in> S" using e_let_p' t_let by auto
    have Hxing: "x\<rhd>\<tau> \<in> \<Gamma>" using t_let e_let_p' by auto
    hence Hxin: "x\<rhd>\<tau> \<in> \<Gamma>s \<union> \<Gamma>m" using t_let by auto
    have "M;\<Delta>;Zs;Zl \<Turnstile> S : \<Gamma>s\<union>\<Gamma>m" using t_let.prems e_let_p' by auto
    hence "scopeV v = scopeT \<tau>"
      apply (rule storeTyping_scope_sanity) using HxinS Hxin by auto
    hence Hvbot: "scopeV v = ScBot" using Hbot by auto
    have "M;\<Delta>;Zs,(\<rho>:\<section>);Zl,(\<rho>:\<section>) \<Turnstile> (S\<bar>x),(x\<mapsto>convV \<rho> v):((\<Gamma>s \<union> \<Gamma>m)\<bar>x),(x:convT \<rho> \<tau>)"
      apply (rule storeTyping_without_conv_qL[of M \<Delta> Zs \<rho> Zl S "\<Gamma>s\<union>\<Gamma>m" x v \<tau>])
    using Hvbot HxinS Hxin Hstr Hbang by auto
    also
    obtain \<Gamma>new Znew where Hvtyp: "\<Gamma>new;\<Delta>;M;Znew \<turnstile> v : \<tau>"
      and Hvlin: "qualV v = qL" and Hvbot: "scopeV v = ScBot" 
      using storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x v \<tau>] 
        t_let e_let_p' Hxin Hlin HxinS Hvbot Hst
      by auto
    hence Hs: "S\<bar>x = S'" using e_let_p' by auto
    also
    have "((\<Gamma>s\<union>\<Gamma>m)\<bar>x),(x:convT \<rho> \<tau>) = (\<Gamma>s\<bar>x),(x:convT \<rho> \<tau>) \<union> (\<Gamma>m\<bar>x)"
      by (auto simp add: without_def)
    also
    hence "x\<rhd>\<tau> \<in> \<Gamma>s" using Hxing t_let by auto
    hence Hxnontinm: "\<not> \<Gamma>m defines x" using t_let compat_ok_sanity Hlin by auto
    hence "\<Gamma>m\<bar>x = \<Gamma>m" by (auto simp add: without_def dom_def)
    finally
    have Hg1: "M;\<Delta>;Zs,(\<rho>:\<section>);Zl,(\<rho>:\<section>) \<Turnstile> S',(x\<mapsto>convV \<rho> v):(\<Gamma>s\<bar>x),(x:convT \<rho> \<tau>) \<union> \<Gamma>m"
      by auto
    have Hg_compat: "\<Gamma> \<sim> \<Gamma>1 \<union> \<Gamma>2"
      using t_let by (auto simp add: env_compat_union_left)
    moreover
    have "\<Gamma> \<union> (\<Gamma>1 \<union> \<Gamma>2);\<Delta> \<Turnstile> OK" using t_let by (auto simp add: Un_ac)
    ultimately
    have Hxnotin: "\<not> (\<Gamma>1\<union>\<Gamma>2) defines x"
      using Hxing Hlin compat_ok_sanity[of \<Gamma> "\<Gamma>1\<union>\<Gamma>2" _ x \<tau>] by auto
    have Hrs: "\<not> Zs defines \<rho>" using t_let e_let_p' by auto
    moreover
    have "Z2 \<union> Z1 \<subseteq> Zs" using t_let by (auto simp only: Un_ac)
    ultimately
    have Hr: "\<not> (Z2\<union>Z1) defines \<rho>" by (auto simp add: dom_def)
    have "\<rho> \<notin> set (free_let_dollar_scope (let! e e1 in e2))" proof 
      assume "\<rho> is free dollar sc in (let! e e1 in e2)" 
      moreover
      have "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2;\<Delta>;Mz;Z2\<union>Z1\<union>Z \<turnstile> let! e e1 in e2 : \<tau>2" 
        apply (rule  Typing.t_let[of \<Gamma>1 \<Gamma>2 \<Gamma> \<Delta> Z1 Z2 \<tau>1 \<tau> Mz Z e Lx Ly], assumption+)
        using t_let.hyps by auto
      moreover 
      have "INV9 let! e e1 in e2" by assumption
      ultimately
      have "(Z2\<union>Z1\<union>Z) defines \<rho>" 
        using free_dollars_in_scopeset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> Mz "Z2\<union>Z1\<union>Z" "let! e e1 in e2" \<tau>2]
        by auto
      moreover
      have "Z2\<union>Z1\<union>Z \<subseteq> Zl" and "Zl \<subseteq> Zs" by assumption
      hence "Z2\<union>Z1\<union>Z \<subseteq> Zs" by auto
      ultimately
      have "Zs defines \<rho>" by (auto simp only: dom_def Un_ac)
      thus False using Hrs by auto
    qed
    hence Hr': "\<rho> \<notin> set (free_let_dollar_scope e1)" 
      and Hr'': "\<rho> \<notin> set (free_let_dollar_scope e2)" 
      by auto 
    have Hxnotin1: "\<not> \<Gamma>1 defines x" using Hxnotin by (auto simp add: dom_def)
    have H\<tau>: "\<Delta> \<Turnstile>' \<tau>" using t_let poly_ty_wf_type by auto
    hence HconvT\<tau>: "\<Delta> \<Turnstile>' convT \<rho> \<tau>" by (cases \<tau>, auto)
    have Hwf_1_new: "\<Gamma>1;\<Delta> \<Turnstile> OK" by (rule wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2"], assumption, auto)
    hence Hx: "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK \<and> \<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>}" using Hxnotin1 HconvT\<tau> 
      using wf_env_cons by (auto simp add: dom_def env_compat_def)
    have "\<And>j. free_let_dollar_scope e1 = free_let_dollar_scope (vsubst_sc (ScFree \<rho>) j e1)"
      by (induct e1 arbitrary: j, auto)
    hence Hd: "\<forall>\<rho>'. \<rho>' is free dollar sc in freshen_sc \<rho> e1 \<longrightarrow> scopeT (convT \<rho> \<tau>) \<noteq> ScFree \<rho>'" 
      using Hr' by (cases \<tau>, auto)
        (*
        lemma free_dollars_vsubst_scope:
        shows "free_let_dollar_scope e = 
        free_let_dollar_scope (vsubst_sc (ScFree \<rho>) j e))"
        by (induct e arbitrary: j, auto)
        *)
    have "\<forall>x \<rho>. \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK \<and> \<Gamma>1 \<sim> {x\<rhd>convT \<rho> \<tau>} \<and> x \<notin> Lx \<and>
           (\<rho> \<notin> set (free_let_dollar_scope e1)) \<and> \<rho> \<notin> Environ.dom (Z2 \<union> Z1) \<longrightarrow>
            \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;Mz;Z1 \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1 \<or>
            \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;Mz;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1"
      using t_let.hyps by auto
    moreover
    have Hfin_new: "finite (dom \<Gamma>1)" using dom_finite Hwf_1_new by (auto simp add: wf_env_def)
    have Hfin2_new: "finite Lx" by assumption
    obtain k 
      where Hk: "\<Gamma>1,(k:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK \<and> \<Gamma>1 \<sim> {k\<rhd>convT \<rho> \<tau>} \<and> 
      k \<notin> (Lx\<union>set (free_tm (freshen_sc \<rho> e1))\<union>dom \<Gamma>1)" 
      using Hwf_1_new HconvT\<tau> Hfin_new Hfin2_new
        new_variable_abs[of \<Gamma>1 \<Delta> "Lx\<union>dom \<Gamma>1\<union>(set (free_tm (freshen_sc \<rho> e1)))" "convT \<rho> \<tau>"]
      by blast
    ultimately
    have "\<Gamma>1,(k:convT \<rho> \<tau>);\<Delta>;Mz;Z1 \<turnstile> freshen_tm k (freshen_sc \<rho> e1) : \<tau>1 \<or>
      \<Gamma>1,(k:convT \<rho> \<tau>);\<Delta>;Mz;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm k (freshen_sc \<rho> e1) : \<tau>1"
      using Hr Hr' by auto
    hence "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;Mz;Z1 \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1 \<or>
      \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;Mz;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1" proof rule
      assume Ht: "\<Gamma>1,(k:convT \<rho> \<tau>);\<Delta>;Mz;Z1 \<turnstile> freshen_tm k (freshen_sc \<rho> e1) : \<tau>1"
      have "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;Mz;Z1 \<turnstile> vsubst_tm (TmFreeVar x) 0 (freshen_sc \<rho> e1) : \<tau>1"
        apply (rule substitution[where x=k]) using Hk Hx Ht Hd by auto
      thus ?thesis by auto
    next
      assume Ht: "\<Gamma>1,(k:convT \<rho> \<tau>);\<Delta>;Mz;Z1,(\<rho>:\<section>) \<turnstile> freshen_tm k (freshen_sc \<rho> e1) : \<tau>1"
      have "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;Mz;Z1,(\<rho>:\<section>) \<turnstile> vsubst_tm (TmFreeVar x) 0 (freshen_sc \<rho> e1) : \<tau>1"
        apply (rule substitution[where x=k]) using Hk Hx Ht Hd by auto
      thus ?thesis by auto
    qed
    hence "\<exists>Zf. \<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;Mz;Zf \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1 \<and>
             (Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>))" 
      by blast
    then obtain Zf where Hf: "Zf = Z1 \<or> Zf = Z1,(\<rho>:\<section>)" 
      and Ht: "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;Mz;Zf \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1" by auto
    have Hg3: "(\<Gamma>\<bar>x),(x:convT \<rho> \<tau>)\<union>\<Gamma>1\<union>\<Gamma>2;\<Delta>;{};Z2\<union>Z1\<union>({},(\<rho>:\<section>)) \<turnstile> 
      let$ \<rho> TmFreeVar x (freshen_tm x (freshen_sc \<rho> e1)) in e2 : \<tau>2"
    proof
      show "\<Gamma>1,(x:convT \<rho> \<tau>);\<Delta>;{};Zf \<turnstile> freshen_tm x (freshen_sc \<rho> e1) : \<tau>1"
        using Ht t_let.prems by auto
    next
      show "\<forall>xa y. \<Gamma>2,(xa:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK \<and> \<Gamma>2 \<sim> {xa\<rhd>\<tau>},(y:\<tau>1) \<and> {xa\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1} \<and> xa \<notin> Lx \<and> 
        y \<notin> Ly \<and> xa \<noteq> y \<longrightarrow>
              \<Gamma>2,(xa:\<tau>),(y:\<tau>1);\<Delta>;{};Z2 \<turnstile> vsubst_tm (TmFreeVar y) (Suc 0) (freshen_tm xa e2) : \<tau>2"
      proof auto
        fix xa y
        assume "\<Gamma>2,(xa:\<tau>),(y:\<tau>1);\<Delta> \<Turnstile> OK" and "\<Gamma>2 \<sim> {xa\<rhd>\<tau>},(y:\<tau>1)" and "{xa\<rhd>\<tau>} \<sim> {y\<rhd>\<tau>1}"
          and "xa \<notin> Lx" and "y \<notin> Ly" and "xa \<noteq> y"
        thus "\<Gamma>2,(xa:\<tau>),(y:\<tau>1);\<Delta>;{};Z2 \<turnstile> 
              vsubst_tm (TmFreeVar y) (Suc 0) (vsubst_tm (TmFreeVar xa) 0 e2) : \<tau>2" 
          using t_let e_let_p' by auto
      qed
    next
      have "qualG (\<Gamma>\<bar>x) \<preceq> qU" using t_let e_let_p' by auto
      moreover
      have "((\<Gamma>\<bar>x),(x:convT \<rho> \<tau>))\<bar>x = (\<Gamma>\<bar>x)" by (auto simp add: without_def)
      ultimately
      have "qualG ((\<Gamma> \<bar> x),(x:convT \<rho> \<tau>) \<bar> x) \<preceq> qU" by auto
      moreover
      have "\<Gamma>;\<Delta> \<Turnstile> OK" using t_let ty_wf_env by auto
      hence "(\<Gamma>\<bar>x),(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK" using HconvT\<tau>
        by (auto simp add: wf_env_def without_def)
      moreover
      have "minimalValidScope (scopeT (convT \<rho> \<tau>)) ({},(\<rho>:\<section>))"
        by (cases \<tau>, auto simp add: minimalValidScope_def)
      ultimately
      show "(\<Gamma>\<bar>x),(x:convT \<rho> \<tau>);\<Delta>;{};{},(\<rho>:\<section>) \<turnstile> TmFreeVar x : convT \<rho> \<tau>"  
        by (auto simp add: wf_env_def)
    next
      have "(\<Gamma>\<bar>x) \<sim> \<Gamma>2" using t_let env_compat_subset[of \<Gamma> \<Gamma>2 "\<Gamma>\<bar>x"] 
        by (auto simp add: without_def)
      moreover
      have "\<not> \<Gamma>2 defines x" using Hxnotin by (auto simp add: dom_def)
      ultimately
      show "(\<Gamma>\<bar>x),(x:convT \<rho> \<tau>) \<sim> \<Gamma>2" by (auto simp add: dom_def env_compat_def)
    next
      have "(\<Gamma>\<bar>x) \<sim> \<Gamma>1" using t_let env_compat_subset[of \<Gamma> \<Gamma>1 "\<Gamma>\<bar>x"] 
        by (auto simp add: without_def)
      moreover
      have "\<not> \<Gamma>1 defines x" using Hxnotin by (auto simp add: dom_def)
      ultimately
      show "(\<Gamma>\<bar>x),(x:convT \<rho> \<tau>) \<sim> \<Gamma>1" by (auto simp add: dom_def env_compat_def)
    next
      have "(\<Gamma>\<bar>x) \<union> \<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK" 
        using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Delta> "(\<Gamma>\<bar>x)\<union>\<Gamma>1\<union>\<Gamma>2"]
        by (auto simp add: without_def)
      hence "((\<Gamma>\<bar>x) \<union> \<Gamma>1 \<union> \<Gamma>2),(x:convT \<rho> \<tau>);\<Delta> \<Turnstile> OK" using Hxnotin HconvT\<tau> 
        by (auto simp add: dom_def wf_env_def without_def)
      thus "(\<Gamma>\<bar>x),(x:convT \<rho> \<tau>) \<union> \<Gamma>1 \<union> \<Gamma>2;\<Delta> \<Turnstile> OK" by auto
    next
      show "\<Gamma>1 \<sim> \<Gamma>2" using t_let by auto
    next
      show "Z1 \<Turnstile> ScopeSet" using t_let by auto
    next
      show "validScope (scopeT \<tau>1) Z1" using t_let by auto
    next
      show "isBangable \<tau>" using t_let by auto
    next
      show "\<not> (\<Gamma>1 \<union> \<Gamma>2) defines x" using Hxnotin by auto
    next
      show "\<rho> \<notin> Environ.dom (Z2 \<union> Z1)" using Hr by auto 
    next
      have "\<forall>\<rho>'. \<rho>' is free dollar sc in e1 \<longrightarrow> \<rho>' \<notin> Environ.dom (Z2 \<union> Z)" by assumption
      hence "\<forall>\<rho>'. \<rho>' is free dollar sc in e1 \<longrightarrow> \<rho>' \<notin> Environ.dom (Z,(\<rho>:\<section>) \<union> Z2)" 
        using Hr' by (auto simp add: dom_def)
      thus "\<forall>\<rho>'. \<rho>' is free dollar sc in (freshen_tm x (freshen_sc \<rho> e1)) \<longrightarrow> 
                  \<rho>' \<notin> Environ.dom ({},(\<rho>:\<section>) \<union> Z2)" 
        using free_dollars_vsubst by (auto simp add: dom_def)
    next
      have "\<forall>\<rho>'. \<rho>' is free dollar sc in e2 \<longrightarrow> \<rho>' \<notin> Environ.dom (Z1 \<union> Z)" by assumption
      hence "\<forall>\<rho>'. \<rho>' is free dollar sc in e2 \<longrightarrow> \<rho>' \<notin> Environ.dom (Z1 \<union> Z)" 
        by (auto simp add: dom_def)
      thus "\<forall>\<rho>'. \<rho>' is free dollar sc in e2 \<longrightarrow> \<rho>' \<notin> Environ.dom ({},(\<rho>:\<section>) \<union> Z1)"
        using Hr'' by (auto simp add: dom_def)
    qed
    have Hg2: "\<Gamma>m;\<Delta> \<Turnstile>' \<mu> : M" using t_let e_let_p' by auto
    have H: "\<not> \<Gamma>r defines x" 
      using t_let Hxing Hlin compat_ok_sanity[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Delta> x \<tau>] 
      by (auto simp add: dom_def)
    have "\<Gamma> \<union> \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>s" by assumption 
    hence "(\<Gamma> \<union> \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r)\<bar>x \<subseteq> (\<Gamma>s\<bar>x)" by (auto simp add: without_def)
    also have "(\<Gamma> \<union> \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r)\<bar>x = (\<Gamma>\<bar>x) \<union> \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r"
      using Hxnotin H by (auto simp add: without_def dom_def)
    finally 
    have Hg4: "(\<Gamma>\<bar>x),(x:convT \<rho> \<tau>) \<union> \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r \<subseteq> (\<Gamma>s\<bar>x),(x:convT \<rho> \<tau>)" 
      by auto
    have Hg5: "(\<Gamma>s\<bar>x) \<sim> \<Gamma>m" using t_let env_compat_subset[of \<Gamma>s \<Gamma>m "\<Gamma>s\<bar>x"]
      by (auto simp add: without_def)
    hence Hg5: "(\<Gamma>s\<bar>x),(x:convT \<rho> \<tau>) \<sim> \<Gamma>m" 
      using Hxnontinm by (auto simp add: without_def env_compat_def dom_def)
    have "Z2 \<union> Z1 \<union> Z \<subseteq> Zl" by assumption
    hence Hg6: "Z2\<union>Z1\<union>({},(\<rho>:\<section>)) \<subseteq> Zl,(\<rho>:\<section>)" by auto
    have "Zl \<subseteq> Zs" by assumption
    hence Hg7: "Zl,(\<rho>:\<section>) \<subseteq> Zs,(\<rho>:\<section>)" by auto
    have Hvalue: "value v" using e_let_p' wf_store_def by auto
    have Hstore: "S \<Turnstile> Store" and Hvin: "x\<mapsto>v \<in> S" using e_let_p' by auto
    have "free_scope_store S = dom Zs" using t_let e_let_p' by auto
    also have "free_scope_store S = free_scope_store (S\<bar>x)"
      using Hstore Hvin Hvbot free_scope_store_without_empty by auto
    finally have "free_scope_store ((S\<bar>x),(x\<mapsto>convV \<rho> v)) = dom (Zs,(\<rho>:\<section>))" 
      using free_scope_store_cons_eq[of v x \<rho> "S\<bar>x"] Hvalue
      by (auto simp add: dom_def)
    hence Hg8: "free_scope_store (S',(x\<mapsto>convV \<rho> v)) = dom (Zs,(\<rho>:\<section>))" 
      using Hs by auto
    have Hstore_sub: "filterStore S Zl \<subseteq> S" by (auto simp add: filterStore_def)
    have Hinv2: "INV2 (filterStore S Zl)" using t_let e_let_p' by auto
    moreover
    have "INV3 S" using t_let e_let_p' by auto
    hence "INV3 (filterStore S Zl)" using inv3_subset Hstore_sub by auto
    moreover
    have "(filterStore S Zl) \<Turnstile> Store" using Hstore Hstore_sub wf_store_subset
      by auto
    moreover
    have HxinS: "x\<mapsto>v \<in> S" using e_let_p' by auto
    hence HxinSfilter: "x\<mapsto>v \<in> (filterStore S Zl)" 
      using Hvbot by (auto simp add: filterStore_def)
    ultimately
    have "INV2 (((filterStore S Zl)\<bar>x),(x\<mapsto>convV \<rho> v))"
      using inv2_without_conv[of "filterStore S Zl"] by auto
    also 
    have "(filterStore S Zl)\<bar>x = filterStore (S\<bar>x) Zl"
      by (auto simp add: filterStore_def without_def)
    also
    have "\<rho> \<notin> free_scope_store S" by assumption
    hence Hfilter_equal: "filterStore (S\<bar>x) Zl = filterStore (S\<bar>x) (Zl,(\<rho>:\<section>))"
    proof (auto simp add: filterStore_def free_scope_store_def valid_scope_weaken
      free_scope_store_def)
      fix x v
      assume "validScope (scopeV v) Zl"
      thus "validScope (scopeV v) (Zl,(\<rho>:\<section>))"
        by (cases "scopeV v", auto)
    next
      fix xa v
      assume Hvalid: "validScope (scopeV v) (Zl,(\<rho>:\<section>))"
        and H: "\<forall>x. (\<forall>z v. x = set (free_scope_scope (scopeV v)) \<longrightarrow> z\<mapsto>v \<notin> S) \<or> \<rho> \<notin> x"
        and "xa\<mapsto>v \<in> (S\<bar>x)"
      hence H1: "xa\<mapsto>v \<in> S" by (auto simp add: without_def)
      let ?x = "set (free_scope_scope (scopeV v))"
      have "(\<forall>z v. ?x = set (free_scope_scope (scopeV v)) \<longrightarrow> z\<mapsto>v \<notin> S) \<or> \<rho> \<notin> ?x"
        using H by auto
      hence "\<rho> \<notin> set (free_scope_scope (scopeV v))" using H1 by auto
      thus "validScope (scopeV v) Zl" using Hvalid by (cases "scopeV v", auto)
    qed
    also
    have "validScope (scopeV (convV \<rho> v)) (Zl,(\<rho>:\<section>))"
      using Hvalue by (cases v, auto simp add: dom_def)
    hence "filterStore (S\<bar>x) (Zl,(\<rho>:\<section>)),(x\<mapsto>convV \<rho> v)
           = filterStore ((S\<bar>x),(x\<mapsto>convV \<rho> v)) (Zl,(\<rho>:\<section>))" 
      by (auto simp add: filterStore_def)
    finally
    have Hg9: "INV2 (filterStore (S',(x\<mapsto>convV \<rho> v)) (Zl,(\<rho>:\<section>)))"
      using Hs by auto
    have "INV3 S" using e_let_p' t_let by auto
    hence "(\<exists>q l \<pi>. v = TmLoc q \<pi> l) \<or> (\<forall>l. l \<notin> set (free_loc v))" 
      using HxinS by auto
    hence "(\<exists>q l \<pi>. (convV \<rho> v) = TmLoc q \<pi> l) \<or> (\<forall>l. l \<notin> set (free_loc (convV \<rho> v)))"
    proof
      assume "\<exists>q l \<pi>. v = TmLoc q \<pi> l"
      then obtain q l \<pi> where "v = TmLoc q \<pi> l" by auto
      hence "convV \<rho> v = TmLoc qU (ScFree \<rho>) l" by auto
      thus ?thesis by blast
    next
      assume "\<forall>l. l \<notin> set (free_loc v)"
      hence "\<forall>l. l \<notin> set (free_loc (convV \<rho> v))" using Hvalue by (cases v, auto)
      thus ?thesis by auto
    qed
    moreover
    have "INV3 S" using t_let e_let_p' by auto
    hence "INV3 (S\<bar>x)" using inv3_subset[of S "S\<bar>x"] by (auto simp add: without_def)
    ultimately
    have "INV3 ((S\<bar>x),(x\<mapsto>convV \<rho> v))" by auto
    hence Hg10: "INV3 (S',(x\<mapsto>convV \<rho> v))" using Hs by auto
    have "(\<Gamma>\<bar>x) \<union> \<Gamma>1 \<union> \<Gamma>2 \<sim> \<Gamma>r" 
      using t_let env_compat_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2" \<Gamma>r "(\<Gamma>\<bar>x)\<union>\<Gamma>1\<union>\<Gamma>2"]
      by (auto simp add: without_def)
    hence Hg11: "(\<Gamma>\<bar>x),(x:convT \<rho> \<tau>) \<union> \<Gamma>1 \<union> \<Gamma>2 \<sim> \<Gamma>r"
      using H by (auto simp add: dom_def env_compat_def)
    have "(\<Gamma>\<bar>x) \<union> \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" 
      using t_let wf_env_subset[of "\<Gamma>\<union>\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>r" \<Delta> "(\<Gamma>\<bar>x)\<union>\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>r"] 
      by (auto simp add: without_def)
    hence Hg12: "(\<Gamma>\<bar>x),(x:convT \<rho> \<tau>) \<union> \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK"
      using wf_env_cons[of "\<Gamma>\<bar>x\<union>\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>r" \<Delta> x "convT \<rho> \<tau>"] Hxnotin H HconvT\<tau>
      by (auto simp add: dom_def without_def)
    have Hg16: "Zr \<subseteq> Zl,(\<rho>:\<section>)" using t_let by (auto simp only: Un_ac)
    hence "INV4 (filterStore S Zl)" using t_let e_let_p' by auto
    moreover
    have "filterStore (S\<bar>x) Zl \<subseteq> filterStore S Zl"
      by (auto simp add: without_def filterStore_def)
    ultimately
    have Hinv: "INV4 (filterStore (S\<bar>x) Zl)" using inv4_subset by (auto simp del: INV4)
    have Hbang: "isBangable \<tau>" using t_let e_let_p' by auto
    have "INV4 ((filterStore (S\<bar>x) Zl),(x\<mapsto>convV \<rho> v))" using Hvtyp Hvalue proof (cases v)
      case (TmUnit q \<pi>) thus ?thesis using Hinv INV4 by auto
    next
      case (TmLoc q \<pi> l)
      hence Hv: "v = TmLoc qL \<pi> l" using Hvlin by auto
      hence "\<not> l free loc in st ((filterStore S Zl)\<bar>x)" using Hinv2 HxinSfilter by auto
      also have "(filterStore S Zl)\<bar>x = filterStore (S\<bar>x) Zl"
        by (auto simp add: filterStore_def without_def dom_def)
      finally have "\<not> l free loc in st (filterStore (S\<bar>x) Zl)" by auto 
      thus ?thesis
        using inv4_cons_loc[where S="(filterStore (S\<bar>x) Zl)" ] Hinv Hv by auto
    next
      case (TmAbs q \<pi> \<tau>' e)
      hence False using Hvtyp Hbang proof (cases \<tau>)
        fix q' \<pi>' \<phi>
        assume "\<tau> = Qual q' \<pi>' \<phi>"
          and "isBangable \<tau>"
        thus False using Hvtyp TmAbs by (cases \<phi>, auto)
      qed
      thus ?thesis by auto
    next
      case (TmProd q \<pi> e)
      hence False using Hvtyp Hbang proof (cases \<tau>)
        fix q' \<pi>' \<phi>
        assume "\<tau> = Qual q' \<pi>' \<phi>"
          and "isBangable \<tau>"
        thus False using Hvtyp TmProd by (cases \<phi>, auto)
      qed
      thus ?thesis by auto
    qed auto
    also note Hfilter_equal
    finally have "INV4 ((filterStore (S\<bar>x) (Zl,(\<rho>:\<section>))),(x\<mapsto>convV \<rho> v))" by auto
    moreover 
    have "filterStore ((S\<bar>x),(x\<mapsto>convV \<rho> v)) (Zl,(\<rho>:\<section>)) \<subseteq> 
      (filterStore (S\<bar>x) (Zl,(\<rho>:\<section>))),(x\<mapsto>convV \<rho> v)" by (auto simp add: filterStore_def)
    ultimately
    have Hg17: "INV4 (filterStore ((S\<bar>x),(x\<mapsto>convV \<rho> v)) (Zl,(\<rho>:\<section>)))" 
      using inv4_subset by blast
    have "INV5 S" using t_let e_let_p' by auto
    hence "INV5 (S\<bar>x)" by (rule inv5_subset, auto simp add: without_def)
    hence Hg18: "INV5 ((S\<bar>x),(x\<mapsto>convV \<rho> v))" using INV5 Hvalue proof (cases v)
      case (TmAbs q \<pi> \<tau>' e)
      hence False using Hvtyp Hbang proof (cases \<tau>)
        fix q' \<pi>' \<phi>
        assume "\<tau> = Qual q' \<pi>' \<phi>"
          and "isBangable \<tau>"
        thus False using Hvtyp TmAbs by (cases \<phi>, auto)
      qed
      thus ?thesis by auto
    next
      case (TmProd q \<pi> e)
      hence False using Hvtyp Hbang proof (cases \<tau>)
        fix q' \<pi>' \<phi>
        assume "\<tau> = Qual q' \<pi>' \<phi>"
          and "isBangable \<tau>"
        thus False using Hvtyp TmProd by (cases \<phi>, auto)
      qed
      thus ?thesis by auto
    qed auto
    have HrnotinS: "\<not> Zs defines \<rho>" using prems by auto
    moreover
    have "Zr \<subseteq> Zs" using t_let by auto
    ultimately
    have "\<not> Zr defines \<rho>" by (auto simp add: dom_def)
    moreover
    have "INV6 (let! (TmFreeVar x) e1 in e2) Zr" using t_let e_let_p' by auto
    ultimately
    have Hg19: "INV6 (let$ \<rho> (TmFreeVar x) (freshen_tm x (freshen_sc \<rho> e1)) in e2) Zr" 
      using free_dollars_vsubst by auto
    have Hg20: "INV7 (let! (TmFreeVar x) e1 in e2) 
      (let$ \<rho> (TmFreeVar x) (freshen_tm x(freshen_sc \<rho> e1)) in e2) Zs"
      using HrnotinS free_dollars_vsubst by auto 
    have "INV9 let! e e1 in e2" by assumption
    hence Hg21: "INV9 let$ \<rho> e (freshen_tm x (freshen_sc \<rho> e1)) in e2"
      using inv9_vsubst by auto
    have Hg22: "INV8 (Z2\<union>Z1\<union>({},(\<rho>:\<section>))) (Z2\<union>Z1\<union>Z) 
      (let$ \<rho> TmFreeVar x freshen_tm x (freshen_sc \<rho> e1) in e2)" by (auto simp add: dom_def)
    show ?thesis
      apply (rule exI[of _ "((\<Gamma>\<bar>x),(x:convT \<rho> \<tau>))\<union>\<Gamma>1\<union>\<Gamma>2"],
        rule exI[of _ "(\<Gamma>s\<bar>x),(x:convT \<rho> \<tau>)"], rule exI[of _ "\<Gamma>m"], 
        rule exI[of _ "M"], rule exI[of _ "Zs,(\<rho>:\<section>)"],
        rule exI[of _ "Z2\<union>Z1\<union>({},(\<rho>:\<section>))"], rule exI[of _ "Zl,(\<rho>:\<section>)"])
      using e_let_p' Hs Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg8 Hg9 Hg10 Hg11 Hg12 Hg16 Hg17 Hg18 Hg19 
        Hg20 Hg21 Hg22
      by auto
  qed auto
next;

      (********************* swap cases *********************)
  case (t_swap \<Delta> Z1 Z2 Mz \<Gamma>1 \<Gamma>2 \<tau> e1 e2 \<pi> e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;e1 :=: e2 \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
  case (e_swap_p1 Sa Sa' \<mu>a \<mu>a' e1a e1' e2a)
    from t_swap.hyps
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r Zs Zr.
          \<lbrakk> Mz = {}; M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M;  S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1';
           \<Gamma>1 \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma>1 \<sim> \<Gamma>r; \<Gamma>1 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; 
           Z1 \<subseteq> Zl; Zr \<subseteq> Zl; Zl \<subseteq> Zs;
           free_scope_store S = dom Zs; INV2 (filterStore S Zl); INV3 S;
           INV4 (filterStore S Zl); INV5 S; INV6 e1 Zr; INV9 e1;
           \<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK\<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'.
            (M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e1' : Qual qU \<pi> (TyRef \<tau>)) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (Ze' \<subseteq> Zl') \<and> (Zl' \<subseteq> Zs') \<and> (Zs \<subseteq> Zs') \<and> (Zr \<subseteq> Zl') \<and>
            (free_scope_store S' = Environ.dom Zs') \<and>
            (INV2 (filterStore S' Zl')) \<and> (INV3 S') \<and>
            (INV4 (filterStore S' Zl')) \<and> (INV5 S') \<and>
            (INV6 e1' Zr) \<and> (INV7 e1 e1' Zs) \<and> (INV8 Ze' Z1 e1') \<and> (INV9 e1')"
      by auto
    have Hsubeq: "\<Gamma>1 \<union> (\<Gamma>r\<union>\<Gamma>2) = \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    obtain \<Gamma>1' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'
      where Hg1: "M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>1';\<Delta>;{};Ze' \<turnstile> e1' : Qual qU \<pi> (TyRef \<tau>)"
      and H4_1: "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2) \<subseteq> \<Gamma>s'"
      and H5_1_1: "\<Gamma>1' \<sim> (\<Gamma>r\<union>\<Gamma>2)"
      and H6_1: "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2);\<Delta> \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg10: "INV2 (filterStore S' Zl')"
      and Hg11: "INV3 S'"
      and Hg12: "Zs \<subseteq> Zs'"
      and Hg13: "Ze' \<subseteq> Zl'"
      and Hg14: "free_scope_store S' = dom Zs'"
      and Hg15g: "Zl' \<subseteq> Zs'"
      and Hg15: "Zr\<union>Z2 \<subseteq> Zl'"
      and Hg16: "INV4 (filterStore S' Zl')"
      and Hg17: "INV5 S'"
      and Hg18: "INV6 e1' (Zr\<union>Z2)"
      and Hg19: "INV7 e1 e1' Zs"
      and Hg20: "INV8 Ze' Z1 e1'"
      and Hg21: "INV9 e1'"
      using IH[of M Zs S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' "\<Gamma>r\<union>\<Gamma>2" "Zr\<union>Z2"] proof 
      show  "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M" and "\<Gamma>s \<sim> \<Gamma>m" 
        and "\<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK" and "INV2 (filterStore S Zl)" and "Mz = {}" and "INV3 S"
        and "INV4 (filterStore S Zl)" and "INV5 S"
        and " free_scope_store S = dom Zs" and "Zl \<subseteq> Zs" by assumption
    next
      show "S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1'" using e_swap_p1 by auto
    next
      have "\<Gamma>1 \<sim> \<Gamma>r" using t_swap env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>1] by auto
      thus "\<Gamma>1 \<sim> (\<Gamma>r\<union>\<Gamma>2)" using env_compat_union_left[of \<Gamma>1 \<Gamma>r \<Gamma>2] t_swap by auto
    next
      show "\<Gamma>1 \<union> (\<Gamma>r\<union>\<Gamma>2);\<Delta> \<Turnstile> OK" using Hsubeq t_swap by auto
    next
      show "\<Gamma>1 \<union> (\<Gamma>r \<union> \<Gamma>2) \<subseteq> \<Gamma>s" using Hsubeq t_swap by auto
    next
      show "Z1 \<subseteq> Zl" using t_swap by (auto simp add: Un_ac)
    next
      show "Zr \<union> Z2 \<subseteq> Zl" using t_swap by (auto simp add: Un_ac)
    next
      have "\<forall>x. x is free dollar sc in e1 \<longrightarrow> x \<notin> Environ.dom Z2" 
        and "INV6 (e1 :=: e2) Zr" by assumption
      thus "INV6 e1 (Zr \<union> Z2)" by (auto simp add: dom_def)
    next
      have "INV9 e1:=:e2" by assumption
      thus "INV9 e1" by auto
    qed auto
    have H5_1: "\<Gamma>1' \<sim> \<Gamma>r" using H5_1_1 env_compat_def by auto
    have "\<Gamma>1' \<sim> \<Gamma>2" using H5_1_1 env_compat_def by auto moreover
    have "\<Gamma>1' \<union> \<Gamma>2;\<Delta> \<Turnstile> OK" 
      using H6_1 wf_env_subset[of "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2)" \<Delta> "\<Gamma>1' \<union> \<Gamma>2"]
      by auto 
    moreover
    have "\<Gamma>1';\<Delta>;{};Ze' \<turnstile> e1' : Qual qU \<pi> (TyRef \<tau>)" using Ht . 
    moreover
    have "\<Gamma>2;\<Delta>;{};Z2 \<turnstile> e2 : \<tau>" using t_swap by auto 
    moreover
    have "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> dom Z2" and "Z2\<union>Z1 \<subseteq> Zl" and "Zl \<subseteq> Zs" 
      by assumption
    hence "\<forall>\<rho>. \<rho> is free dollar sc in e1' \<longrightarrow> \<rho> \<notin> Environ.dom Z2"
      using eval_preserves_let_dollars[of e1 Z2 Zs e1'] Hg19 by auto
    moreover
    {
      have "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom Z1" 
        and "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom Z2"
        by assumption
      moreover
      note Hg19 Hg20
      moreover
      have "Z2\<union>Z1 \<subseteq> Zl" and "Zl \<subseteq> Zs" by assumption
      hence "Z1\<union>Z2 \<subseteq> Zs" by auto
      moreover
      have "\<Gamma>2; \<Delta>; Mz; Z2 \<turnstile> e2 : \<tau>" by assumption
      moreover
      have "INV9 e2" using t_swap.prems by auto      
      ultimately
      have "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom Ze'"  
        using eval_preserves_let_dollars_scope[of e2 Z1 e1 Z2 e1' Zs Ze' \<Gamma>2 \<Delta> Mz \<tau>] by auto
    }
    moreover
    have "qualT \<tau> = qL" using t_swap by auto
    ultimately
    have Hg3: "\<Gamma>1' \<union> \<Gamma>2;\<Delta>;{};(Z2\<union>Ze') \<turnstile> e1' :=: e2 : \<tau>" by auto
    have Hg4: "\<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>s'" using H4_1 by auto
    have "\<Gamma>2 \<sim> \<Gamma>r" using t_swap env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>2] by auto
    hence Hg5: "(\<Gamma>1' \<union> \<Gamma>2) \<sim> \<Gamma>r"
      using H5_1 env_compat_union_right[of \<Gamma>1' \<Gamma>r \<Gamma>2] by auto
    have "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2) = \<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    hence Hg6: "\<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" using H6_1 by auto
    have Hg13: "Ze' \<union> Z2 \<subseteq> Zl'" using Hg15 Hg13 by auto
    have "INV6 (e1:=:e2) Zr" by assumption
    hence Hg18: "INV6 (e1':=:e2) Zr" using Hg18 by (auto simp add: dom_def)
    have Hg19: "INV7 (e1:=:e2) (e1':=:e2) Zs" using Hg19 by (auto simp add: dom_def)
    have Hg20: "INV8 (Z2\<union>Ze') (Z2 \<union> Z1) (e1':=:e2)" using Hg20 by (auto simp add: dom_def)
    have "INV9 e1:=:e2" by assumption
    hence Hg21: "INV9 e1':=:e2" using Hg21 by auto
    show ?thesis
      apply (rule exI[of _ "\<Gamma>1'\<union>\<Gamma>2"], rule exI[of _ \<Gamma>s'])
      apply (rule exI[of _ \<Gamma>m'], rule exI[of _ M'])
      apply (rule exI[of _ Zs'], rule exI[of _ "Z2\<union>Ze'"], rule exI[of _ Zl'])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg12 Hg13 Hg14 Hg15 Hg15g Hg16 Hg17 Hg18 Hg19 
        Hg20 Hg21 e_swap_p1 
      by auto
  next
    case (e_swap_p2 Sa Sa' \<mu>a \<mu>a' e1a e2' e2a)
    from t_swap.hyps
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r Zs Zr.
         \<lbrakk> Mz = {}; M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M;  S;\<mu>;e2\<hookrightarrow>S';\<mu>';e2';
           \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma>2 \<sim> \<Gamma>r; \<Gamma>2 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; 
           Z2 \<subseteq> Zl; Zr \<subseteq> Zl; Zl \<subseteq> Zs;
           free_scope_store S = dom Zs; INV2 (filterStore S Zl); INV3 S;
           INV4 (filterStore S Zl); INV5 S; INV6 e2 Zr; INV9 e2;
           \<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK\<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'.
            (M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e2' : \<tau>) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (Ze' \<subseteq> Zl') \<and> (Zl' \<subseteq> Zs') \<and> (Zs \<subseteq> Zs') \<and> (Zr \<subseteq> Zl') \<and>
            (free_scope_store S' = Environ.dom Zs') \<and>
            (INV2 (filterStore S' Zl')) \<and> (INV3 S') \<and>
            (INV4 (filterStore S' Zl')) \<and> (INV5 S') \<and>
            (INV6 e2' Zr) \<and> (INV7 e2 e2' Zs) \<and> (INV8 Ze' Z2 e2') \<and> (INV9 e2')"
      by auto
    have Hsubeq: "\<Gamma>2 \<union> (\<Gamma>r\<union>\<Gamma>1) = \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    obtain \<Gamma>2' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'
      where Hg1: "M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>2';\<Delta>;{};Ze' \<turnstile> e2' : \<tau>"
      and H4_1: "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1) \<subseteq> \<Gamma>s'"
      and H5_1_1: "\<Gamma>2' \<sim> (\<Gamma>r\<union>\<Gamma>1)"
      and H6_1: "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1);\<Delta> \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg10: "INV2 (filterStore S' Zl')"
      and Hg11: "INV3 S'"
      and Hg12: "Zs \<subseteq> Zs'"
      and Hg13: "Ze' \<subseteq> Zl'"
      and Hg14: "free_scope_store S' = dom Zs'"
      and Hg15g: "Zl' \<subseteq> Zs'"
      and Hg15: "Zr\<union>Z1 \<subseteq> Zl'"
      and Hg16: "INV4 (filterStore S' Zl')"
      and Hg17: "INV5 S'"
      and Hg18: "INV6 e2' (Zr\<union>Z1)"
      and Hg19: "INV7 e2 e2' Zs"
      and Hg20: "INV8 Ze' Z2 e2'"
      and Hg21: "INV9 e2'"
      using IH[of M Zs S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' "\<Gamma>r\<union>\<Gamma>1" "Zr\<union>Z1"] proof 
      show  "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M" and "\<Gamma>s \<sim> \<Gamma>m" 
        and "\<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK" and "INV2 (filterStore S Zl)" and "Mz = {}" and "INV3 S"
        and "INV4 (filterStore S Zl)" and "INV5 S"
        and " free_scope_store S = dom Zs" and "Zl \<subseteq> Zs" by assumption
    next
      show "S;\<mu>;e2\<hookrightarrow>S';\<mu>';e2'" using e_swap_p2 by auto
    next
      have "\<Gamma>2 \<sim> \<Gamma>r" using t_swap env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>2] by auto
      thus "\<Gamma>2 \<sim> (\<Gamma>r\<union>\<Gamma>1)"
      using t_swap env_compat_union_left[of \<Gamma>2 \<Gamma>r \<Gamma>1] 
        env_compat_symm[of \<Gamma>1 \<Gamma>2] by auto
    next
      show "\<Gamma>2 \<union> (\<Gamma>r\<union>\<Gamma>1);\<Delta> \<Turnstile> OK" using Hsubeq t_swap by auto
    next
      show "\<Gamma>2 \<union> (\<Gamma>r \<union> \<Gamma>1) \<subseteq> \<Gamma>s" using Hsubeq t_swap by auto
    next
      show "Z2 \<subseteq> Zl" using t_swap by (auto simp add: Un_ac)
    next
      show "Zr \<union> Z1 \<subseteq> Zl" using t_swap by (auto simp add: Un_ac)
    next
      have "\<forall>x. x is free dollar sc in e2 \<longrightarrow> x \<notin> Environ.dom Z1" 
        and "INV6 (e1:=:e2) Zr" by assumption
      thus "INV6 e2 (Zr \<union> Z1)" by (auto simp add: dom_def)
    next
      have "INV9 e1:=:e2" by assumption
      thus "INV9 e2" by auto
    next 
    qed auto
    have H5_1: "\<Gamma>2' \<sim> \<Gamma>r" using H5_1_1 env_compat_def by auto
    have "\<Gamma>1 \<sim> \<Gamma>2'" using H5_1_1 env_compat_def by auto moreover
    have "\<Gamma>1 \<union> \<Gamma>2';\<Delta> \<Turnstile> OK" 
      using H6_1 wf_env_subset[of "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1)" \<Delta> "\<Gamma>1 \<union> \<Gamma>2'"]
      by auto moreover
    have  "\<Gamma>1;\<Delta>;{};Z1 \<turnstile> e1 : Qual qU \<pi> (TyRef \<tau>)" using t_swap by auto
    moreover
    have  "\<Gamma>2';\<Delta>;{};Ze' \<turnstile> e2' : \<tau>" using Ht by auto
    moreover
    have "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> dom Z1" and "Z2\<union>Z1 \<subseteq> Zl" and "Zl \<subseteq> Zs" 
      by assumption
    hence "\<forall>\<rho>. \<rho> is free dollar sc in e2' \<longrightarrow> \<rho> \<notin> Environ.dom Z1"
      using eval_preserves_let_dollars[of e2 Z1 Zs e2'] Hg19 by auto
    moreover
    {
      have "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom Z2"
        and "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom Z1" 
        by assumption
      moreover
      note Hg19 Hg20
      moreover
      have "Z2\<union>Z1 \<subseteq> Zl" and "Zl \<subseteq> Zs" by assumption
      hence "Z1\<union>Z2 \<subseteq> Zs" by auto
      moreover
      have "\<Gamma>1; \<Delta>; Mz; Z1 \<turnstile> e1 : Qual qU \<pi> (TyRef \<tau>)" by assumption
      moreover
      have "INV9 e1" using t_swap.prems by auto
      ultimately
      have "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom Ze'"  
        using eval_preserves_let_dollars_scope[of e1 Z2 e2 Z1 e2' Zs Ze' \<Gamma>1 \<Delta> Mz] by auto
    }
    moreover
    have "qualT \<tau> = qL" using t_swap by auto
    ultimately
    have Hg3: "\<Gamma>1 \<union> \<Gamma>2';\<Delta>;{};Ze'\<union>Z1 \<turnstile> e1 :=: e2' : \<tau>" by auto
    have Hg4: "\<Gamma>2' \<union> \<Gamma>1 \<union> \<Gamma>r \<subseteq> \<Gamma>s'" using H4_1 by auto
    have "\<Gamma>1 \<sim> \<Gamma>r" using t_swap env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>1] by auto
    hence Hg5: "(\<Gamma>1 \<union> \<Gamma>2') \<sim> \<Gamma>r"
      using H5_1 env_compat_union_right[of \<Gamma>1 \<Gamma>r \<Gamma>2'] by auto
    have "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1) = \<Gamma>1 \<union> \<Gamma>2' \<union> \<Gamma>r" by auto
    hence Hg6: "\<Gamma>1 \<union> \<Gamma>2' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" using H6_1 by auto
    have Hg13: "Ze' \<union> Z1 \<subseteq> Zl'" using Hg15 Hg13 by auto
    have "INV6 (e1:=:e2) Zr" by assumption
    hence Hg18: "INV6 (e1:=:e2') Zr" using Hg18 by (auto simp add: dom_def)
    have Hg19: "INV7 (e1:=:e2) (e1:=:e2') Zs" using Hg19 by (auto simp add: dom_def)
    have Hg20: "INV8 (Ze'\<union>Z1) (Z2\<union>Z1) (e1:=:e2')" using Hg20 by (auto simp add: dom_def)
    have "INV9 e1:=:e2" by assumption
    hence Hg21: "INV9 e1:=:e2'" using Hg21 by auto
    show ?thesis
      apply (rule exI[of _ "\<Gamma>1\<union>\<Gamma>2'"], rule exI[of _ \<Gamma>s'])
      apply (rule exI[of _ \<Gamma>m'], rule exI[of _ M'])
      apply (rule exI[of _ Zs'], rule exI[of _ "Ze'\<union>Z1"], rule exI[of _ Zl'])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg12 Hg13 Hg14 Hg15 Hg15g Hg16 Hg17 Hg18 Hg19 
        Hg20 Hg21 e_swap_p2 
    by auto
  next
    case (e_swap_c Sa Sa' \<mu>a l q x w z \<pi>o)
    have Hz12l: "Z2\<union>Z1 \<subseteq> Zl" by assumption
    have "\<Gamma>1;\<Delta>;Mz;Z1 \<turnstile> e1 : Qual qU \<pi> (TyRef \<tau>)" by assumption
    hence "validScope (scopeT (Qual qU \<pi> (TyRef \<tau>))) Z1" by (rule ty_valid_scope) 
    hence Htvalid: "validScope (scopeT (Qual qU \<pi> (TyRef \<tau>))) Zl" apply (rule valid_scope_subset) 
      using Hz12l by auto
    hence Hvalid: "validScope \<pi> Zl" by auto
    have Hsin: "x\<mapsto>TmLoc q \<pi>o l \<in> S" using e_swap_c by auto
    moreover
    have "x\<rhd>Qual qU \<pi> (TyRef \<tau>) \<in> \<Gamma>1" using t_swap e_swap_c by auto
    hence Hxin: "x\<rhd>Qual qU \<pi> (TyRef \<tau>) \<in> (\<Gamma>s\<union>\<Gamma>m)" using t_swap by auto
    moreover
    have "M;\<Delta>;Zs;Zl \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
    moreover
    hence "scopeV (TmLoc q \<pi>o l) = scopeT (Qual qU \<pi> (TyRef \<tau>))"
      apply (rule storeTyping_scope_sanity) using Hsin Hxin by auto
    hence Hvalidv: "validScope (scopeV (TmLoc q \<pi>o l)) Zl" using Htvalid by auto
    ultimately
    have Hq: "q = qU" and Hscope: "\<pi>o = \<pi>" 
      using storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x "TmLoc q \<pi>o l" "Qual qU \<pi> (TyRef \<tau>)"]
      by auto
    hence Hs: "S' = S" using e_swap_c by auto
    have Hzing2: "z\<rhd>\<tau> \<in> \<Gamma>2" using t_swap e_swap_c by auto
    hence Hin: "z\<rhd>\<tau> \<in> \<Gamma>s" using t_swap by auto
    have Hsttyp: "M;\<Delta>;Zs;Zl \<Turnstile> S':(\<Gamma>s \<union> \<Gamma>m)" using t_swap Hs by auto
    have "M;\<Delta>;Zs;Zl \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
    then obtain \<Gamma>v Ze' where "\<Gamma>v;\<Delta>;M;Ze' \<turnstile> TmLoc qU \<pi>o l : Qual qU \<pi> (TyRef \<tau>)"
      using Hq Hxin Hsin Hvalidv
        storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x "TmLoc qU \<pi>o l" "Qual qU \<pi> (TyRef \<tau>)"]
      by auto
    hence HlinM: "l\<rhd>\<tau> \<in> M" by auto 
    moreover
    hence Hlinm: "\<mu> defines l" using prems by (auto simp add: dom_def)
    ultimately
    have Hwingm: "w\<rhd>\<tau> \<in> \<Gamma>m" using t_swap e_swap_c memTyping_elim_qL[of \<Gamma>m _ \<mu> M l w \<tau>] by auto
    hence Hwnotings: "\<not> \<Gamma>s defines w" using t_swap compat_ok_sanity_right by auto
    hence Hrnotw: "\<not> \<Gamma>r defines w" using t_swap by (auto simp add: dom_def)
    have "\<Gamma>s \<union> \<Gamma>m = \<Gamma>s\<bar>z \<union> \<Gamma>m,(z:\<tau>)" using t_swap Hin env_cons_without_eq_right by auto
    also
    {
      hence "\<Gamma>s\<bar>z \<union> \<Gamma>m,(z:\<tau>);\<Delta> \<Turnstile> OK" using t_swap by auto
      moreover have "w\<rhd>\<tau> \<in> \<Gamma>m,(z:\<tau>)" using Hwingm by auto
      ultimately
      have "\<Gamma>s\<bar>z \<union> \<Gamma>m,(z:\<tau>) = (\<Gamma>s\<bar>z),(w:\<tau>) \<union> \<Gamma>m,(z:\<tau>)\<bar>w"
        using env_cons_without_eq_left[of "\<Gamma>s\<bar>z" "\<Gamma>m,(z:\<tau>)" _ w \<tau>] by auto
    }
    also
    have Hneqzw: "z \<noteq> w" using Hin Hwnotings by (auto simp add: dom_def)
    hence Hgszw: "(\<Gamma>s\<bar>z),(w:\<tau>) = \<Gamma>s,(w:\<tau>)\<bar>z" by (auto simp add: without_def)
    finally
    have Heqenvs: "\<Gamma>s \<union> \<Gamma>m = \<Gamma>s,(w:\<tau>)\<bar>z \<union> \<Gamma>m,(z:\<tau>)\<bar>w" by auto
    hence Hg1: "M;\<Delta>;Zs;Zl \<Turnstile> S':(\<Gamma>s,(w:\<tau>)\<bar>z \<union> \<Gamma>m,(z:\<tau>)\<bar>w)" using Hsttyp by auto
    have "\<Gamma>m;\<Delta> \<Turnstile>' \<mu> : M" by assumption
    moreover
    have "\<not> \<Gamma>m defines z" using t_swap Hin compat_ok_sanity by auto
    ultimately
    have Hg2: "\<Gamma>m,(z:\<tau>)\<bar>w;\<Delta> \<Turnstile>' update \<mu> l (TmFreeVar z) : M" using t_swap e_swap_c Hneqzw
      HlinM Hwingm memTyping_update_qL_without[of \<Gamma>m _ \<mu> M l w \<tau> z]
      by auto
    have "z\<rhd>\<tau> \<in> (\<Gamma>1\<union>\<Gamma>2)" using Hzing2 by (auto simp add: dom_def)
    hence Hznotingr: "\<not> \<Gamma>r defines z" using t_swap compat_ok_sanity[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r _ z] 
      by auto
    have H\<tau>: "\<Delta> \<Turnstile>' \<tau>" and H\<Delta>: "\<Delta> \<Turnstile>' TY" using poly_ty_wf_type poly_wf_env_ty t_swap by auto
    have Hval: "validScope (scopeT \<tau>) Z2" using t_swap ty_valid_scope by auto
    then obtain Ze' where Hmin: "minimalValidScope (scopeT \<tau>) Ze'" 
      by (cases "scopeT \<tau>", auto simp add: minimalValidScope_def dom_def)
    hence Hg3: "{(w\<rhd>\<tau>)};\<Delta>;{};Ze' \<turnstile> TmFreeVar w : \<tau>"
      using H\<tau> H\<Delta> by (auto simp add: without_def wf_env_def qualGrelL_def)
    have Hze_sub: "Ze' \<subseteq> Z2" using Hmin Hval proof (auto simp add: minimalValidScope_def dom_def)
      fix \<rho> b 
      assume "(\<rho>, b) \<in> Z2"
      thus "(\<rho>, EmptyScope) \<in> Z2" by (cases b, auto)
    qed
    have "{w\<rhd>\<tau>} \<union> \<Gamma>r \<subseteq> (\<Gamma>s\<bar>z),(w:\<tau>)" using t_swap Hznotingr
      by (auto simp add: without_def dom_def)
    hence Hg4: "{w\<rhd>\<tau>} \<union> \<Gamma>r \<subseteq> \<Gamma>s,(w:\<tau>)\<bar>z" using Hgszw by auto
    have Hg5: "{w\<rhd>\<tau>} \<sim> \<Gamma>r" using Hrnotw by (auto simp add: dom_def env_compat_def)
    have "\<Gamma>r;\<Delta> \<Turnstile> OK" using t_swap wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Delta> \<Gamma>r] by force
    hence Hg6: "{w\<rhd>\<tau>} \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" using t_swap Hrnotw wf_env_cons H\<tau> by auto
    have "\<Gamma>s \<sim> \<Gamma>m" by assumption
    hence Hg7: "\<Gamma>s,(w:\<tau>)\<bar>z \<sim> \<Gamma>m,(z:\<tau>)\<bar>w" by (auto simp add: env_compat_def without_def)
    have Hg10: "INV2 (filterStore S' Zl)" and Hg16: "INV4 (filterStore S' Zl)" 
      using Hs t_swap by auto
    have Hg11: "INV3 S'" and  Hg17: "INV5 S'" using Hs t_swap by auto
    have Hg13: "Zl \<subseteq> Zs" and Hg14: "Zr \<subseteq> Zl"  by assumption
    have Hg23: "Ze' \<subseteq> Zl" using t_swap.prems Hze_sub by auto
    have Hg15: "free_scope_store S' = dom Zs" using Hs t_swap by auto
    have Hg18: "INV6 (TmUnit qU ScBot) Zr" by auto
    have Hg19: "INV7 (TmFreeVar x:=TmFreeVar y) (TmUnit qU ScBot) Zs" by auto
    have Hg20: "INV8 Ze' (Z2\<union>Z1) (TmUnit qU ScBot)" using Hze_sub by (auto simp add: dom_def)
    have Hg21: "INV9 (TmUnit qU ScBot)" by auto
    show ?thesis
      apply (rule exI[of _ "{w\<rhd>\<tau>}"], rule exI[of _ "\<Gamma>s,(w:\<tau>)\<bar>z"])
      apply (rule exI[of _ "\<Gamma>m,(z:\<tau>)\<bar>w"], rule exI[of _ M],
        rule exI[of _ Zs], rule exI[of _ Ze'], rule exI[of _ Zl])
      using e_swap_c Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg13 Hg14 Hg15 
        Hg16 Hg17 Hg18 Hg19 Hg20 Hg21 Hg23 Hs by auto
  qed auto
next

      (********************* assign cases *********************)
  case (t_assign \<Delta> Z1 Z2 Mz \<Gamma>1 \<Gamma>2 \<tau> e1 e2 \<pi> e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;e1 := e2 \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_assign_p1 Sa Sa' \<mu>a \<mu>a' e1a e1' e2a)
    from t_assign.hyps
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r Zs Zr.
          \<lbrakk> Mz = {}; M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M;  S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1';
           \<Gamma>1 \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma>1 \<sim> \<Gamma>r; \<Gamma>1 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; 
           Z1 \<subseteq> Zl; Zr \<subseteq> Zl; Zl \<subseteq> Zs;
           free_scope_store S = dom Zs; INV2 (filterStore S Zl); INV3 S;
           INV4 (filterStore S Zl); INV5 S; INV6 e1 Zr; INV9 e1;
           \<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK\<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'.
            (M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e1' : Qual qU \<pi> (TyRef \<tau>)) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (Ze' \<subseteq> Zl') \<and> (Zl' \<subseteq> Zs') \<and> (Zs \<subseteq> Zs') \<and> (Zr \<subseteq> Zl') \<and>
            (free_scope_store S' = Environ.dom Zs') \<and>
            (INV2 (filterStore S' Zl')) \<and> (INV3 S') \<and>
            (INV4 (filterStore S' Zl')) \<and> (INV5 S') \<and>
            (INV6 e1' Zr) \<and> (INV7 e1 e1' Zs) \<and> (INV8 Ze' Z1 e1') \<and> (INV9 e1')"
      by auto
    have Hsubeq: "\<Gamma>1 \<union> (\<Gamma>r\<union>\<Gamma>2) = \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    obtain \<Gamma>1' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'
      where Hg1: "M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>1';\<Delta>;{};Ze' \<turnstile> e1' : Qual qU \<pi> (TyRef \<tau>)"
      and H4_1: "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2) \<subseteq> \<Gamma>s'"
      and H5_1_1: "\<Gamma>1' \<sim> (\<Gamma>r\<union>\<Gamma>2)"
      and H6_1: "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2);\<Delta> \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg10: "INV2 (filterStore S' Zl')"
      and Hg11: "INV3 S'"
      and Hg12: "Zs \<subseteq> Zs'"
      and Hg13: "Ze' \<subseteq> Zl'"
      and Hg14: "free_scope_store S' = dom Zs'"
      and Hg15g: "Zl' \<subseteq> Zs'"
      and Hg15: "Zr\<union>Z2 \<subseteq> Zl'"
      and Hg16: "INV4 (filterStore S' Zl')"
      and Hg17: "INV5 S'"
      and Hg18: "INV6 e1' (Zr\<union>Z2)"
      and Hg19: "INV7 e1 e1' Zs"
      and Hg20: "INV8 Ze' Z1 e1'"
      and Hg21: "INV9 e1'"
      using IH[of M Zs S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' "\<Gamma>r\<union>\<Gamma>2" "Zr\<union>Z2"] proof 
      show  "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M" and "\<Gamma>s \<sim> \<Gamma>m" 
        and "\<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK" and "INV2 (filterStore S Zl)" and "Mz = {}" and "INV3 S"
        and "INV4 (filterStore S Zl)" and "INV5 S"
        and " free_scope_store S = dom Zs" and "Zl \<subseteq> Zs" by assumption
    next
      show "S;\<mu>;e1\<hookrightarrow>S';\<mu>';e1'" using e_assign_p1 by auto
    next
      have "\<Gamma>1 \<sim> \<Gamma>r" using t_assign env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>1] by auto
      thus "\<Gamma>1 \<sim> (\<Gamma>r\<union>\<Gamma>2)" using env_compat_union_left[of \<Gamma>1 \<Gamma>r \<Gamma>2] t_assign by auto
    next
      show "\<Gamma>1 \<union> (\<Gamma>r\<union>\<Gamma>2);\<Delta> \<Turnstile> OK" using Hsubeq t_assign by auto
    next
      show "\<Gamma>1 \<union> (\<Gamma>r \<union> \<Gamma>2) \<subseteq> \<Gamma>s" using Hsubeq t_assign by auto
    next
      show "Z1 \<subseteq> Zl" using t_assign by (auto simp add: Un_ac)
    next
      show "Zr \<union> Z2 \<subseteq> Zl" using t_assign by (auto simp add: Un_ac)
    next
      have "\<forall>x. x is free dollar sc in e1 \<longrightarrow> x \<notin> Environ.dom Z2" 
        and "INV6 (e1 := e2) Zr" by assumption
      thus "INV6 e1 (Zr \<union> Z2)" by (auto simp add: dom_def)
    next
      have "INV9 e1:=e2" by assumption
      thus "INV9 e1" by auto
    qed auto
    have H5_1: "\<Gamma>1' \<sim> \<Gamma>r" using H5_1_1 env_compat_def by auto
    have "\<Gamma>1' \<sim> \<Gamma>2" using H5_1_1 env_compat_def by auto moreover
    have "\<Gamma>1' \<union> \<Gamma>2;\<Delta> \<Turnstile> OK" 
      using H6_1 wf_env_subset[of "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2)" \<Delta> "\<Gamma>1' \<union> \<Gamma>2"]
      by auto 
    moreover
    have "\<Gamma>1';\<Delta>;{};Ze' \<turnstile> e1' : Qual qU \<pi> (TyRef \<tau>)" using Ht . 
    moreover
    have "\<Gamma>2;\<Delta>;{};Z2 \<turnstile> e2 : \<tau>" using t_assign by auto 
    moreover
    have "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> dom Z2" and "Z1\<union>Z2 \<subseteq> Zl" and "Zl \<subseteq> Zs" 
      by assumption
    hence "\<forall>\<rho>. \<rho> is free dollar sc in e1' \<longrightarrow> \<rho> \<notin> Environ.dom Z2"
      using eval_preserves_let_dollars[of e1 Z2 Zs e1'] Hg19 by auto
    moreover
    {
      have "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom Z1" 
        and "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom Z2"
        by assumption
      moreover
      note Hg19 Hg20
      moreover
      have "Z1\<union>Z2 \<subseteq> Zl" and "Zl \<subseteq> Zs" by assumption
      hence "Z1\<union>Z2 \<subseteq> Zs" by auto
      moreover
      have "\<Gamma>2; \<Delta>; Mz; Z2 \<turnstile> e2 : \<tau>" by assumption
      moreover
      have "INV9 e2" using t_assign.prems by auto
      ultimately
      have "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom Ze'"  
        using eval_preserves_let_dollars_scope[of e2 Z1 e1 Z2 e1' Zs Ze' \<Gamma>2 \<Delta> Mz \<tau>] by auto
    }
    ultimately
    have Hg3: "\<Gamma>1' \<union> \<Gamma>2;\<Delta>;{};(Ze' \<union> Z2) \<turnstile> e1' := e2 : Qual qU ScBot TyUnit" by auto
    have Hg4: "\<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>s'" using H4_1 by auto
    have "\<Gamma>2 \<sim> \<Gamma>r" using t_assign env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>2] by auto
    hence Hg5: "(\<Gamma>1' \<union> \<Gamma>2) \<sim> \<Gamma>r"
      using H5_1 env_compat_union_right[of \<Gamma>1' \<Gamma>r \<Gamma>2] by auto
    have "\<Gamma>1' \<union> (\<Gamma>r\<union>\<Gamma>2) = \<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    hence Hg6: "\<Gamma>1' \<union> \<Gamma>2 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" using H6_1 by auto
    have Hg13: "Ze' \<union> Z2 \<subseteq> Zl'" using Hg15 Hg13 by auto
    have "INV6 (e1:=e2) Zr" by assumption
    hence Hg18: "INV6 (e1':=:e2) Zr" using Hg18 by (auto simp add: dom_def)
    have Hg19: "INV7 (e1:=e2) (e1':=e2) Zs" using Hg19 by (auto simp add: dom_def)
    have Hg20: "INV8 (Ze' \<union> Z2) (Z1 \<union> Z2) (e1':=e2)" using Hg20 by (auto simp add: dom_def)
    have "INV9 e1:=e2" by assumption
    hence Hg21: "INV9 e1':=e2" using Hg21 by auto
    show ?thesis
      apply (rule exI[of _ "\<Gamma>1'\<union>\<Gamma>2"], rule exI[of _ \<Gamma>s'])
      apply (rule exI[of _ \<Gamma>m'], rule exI[of _ M'])
      apply (rule exI[of _ Zs'], rule exI[of _ "Ze' \<union> Z2"], rule exI[of _ Zl'])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg12 Hg13 Hg14 Hg15 Hg15g Hg16 Hg17 Hg18 Hg19 
        Hg20 Hg21 e_assign_p1 
      by auto
  next
    case (e_assign_p2 Sa Sa' \<mu>a \<mu>a' e1a e2' e2a)
    from t_assign.hyps
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r Zs Zr.
         \<lbrakk> Mz = {}; M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M;  S;\<mu>;e2\<hookrightarrow>S';\<mu>';e2';
           \<Gamma>2 \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma>2 \<sim> \<Gamma>r; \<Gamma>2 \<union> \<Gamma>r;\<Delta> \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; 
           Z2 \<subseteq> Zl; Zr \<subseteq> Zl; Zl \<subseteq> Zs;
           free_scope_store S = dom Zs; INV2 (filterStore S Zl); INV3 S;
           INV4 (filterStore S Zl); INV5 S; INV6 e2 Zr; INV9 e2;
           \<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK\<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'.
            (M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e2' : \<tau>) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (Ze' \<subseteq> Zl') \<and> (Zl' \<subseteq> Zs') \<and> (Zs \<subseteq> Zs') \<and> (Zr \<subseteq> Zl') \<and>
            (free_scope_store S' = Environ.dom Zs') \<and>
            (INV2 (filterStore S' Zl')) \<and> (INV3 S') \<and>
            (INV4 (filterStore S' Zl')) \<and> (INV5 S') \<and>
            (INV6 e2' Zr) \<and> (INV7 e2 e2' Zs) \<and> (INV8 Ze' Z2 e2') \<and> (INV9 e2')"
      by auto
    have Hsubeq: "\<Gamma>2 \<union> (\<Gamma>r\<union>\<Gamma>1) = \<Gamma>1 \<union> \<Gamma>2 \<union> \<Gamma>r" by auto
    obtain \<Gamma>2' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'
      where Hg1: "M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>2';\<Delta>;{};Ze' \<turnstile> e2' : \<tau>"
      and H4_1: "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1) \<subseteq> \<Gamma>s'"
      and H5_1_1: "\<Gamma>2' \<sim> (\<Gamma>r\<union>\<Gamma>1)"
      and H6_1: "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1);\<Delta> \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg10: "INV2 (filterStore S' Zl')"
      and Hg11: "INV3 S'"
      and Hg12: "Zs \<subseteq> Zs'"
      and Hg13: "Ze' \<subseteq> Zl'"
      and Hg14: "free_scope_store S' = dom Zs'"
      and Hg15g: "Zl' \<subseteq> Zs'"
      and Hg15: "Zr\<union>Z1 \<subseteq> Zl'"
      and Hg16: "INV4 (filterStore S' Zl')"
      and Hg17: "INV5 S'"
      and Hg18: "INV6 e2' (Zr\<union>Z1)"
      and Hg19: "INV7 e2 e2' Zs"
      and Hg20: "INV8 Ze' Z2 e2'"
      and Hg21: "INV9 e2'"
      using IH[of M Zs S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' "\<Gamma>r\<union>\<Gamma>1" "Zr\<union>Z1"] proof 
      show  "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M" and "\<Gamma>s \<sim> \<Gamma>m" 
        and "\<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK" and "INV2 (filterStore S Zl)" and "Mz = {}" and "INV3 S"
        and "INV4 (filterStore S Zl)" and "INV5 S" 
        and " free_scope_store S = dom Zs" and "Zl \<subseteq> Zs" by assumption
    next
      show "S;\<mu>;e2\<hookrightarrow>S';\<mu>';e2'" using e_assign_p2 by auto
    next
      have "\<Gamma>2 \<sim> \<Gamma>r" using t_assign env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>2] by auto
      thus "\<Gamma>2 \<sim> (\<Gamma>r\<union>\<Gamma>1)"
      using t_assign env_compat_union_left[of \<Gamma>2 \<Gamma>r \<Gamma>1] 
        env_compat_symm[of \<Gamma>1 \<Gamma>2] by auto
    next
      show "\<Gamma>2 \<union> (\<Gamma>r\<union>\<Gamma>1);\<Delta> \<Turnstile> OK" using Hsubeq t_assign by auto
    next
      show "\<Gamma>2 \<union> (\<Gamma>r \<union> \<Gamma>1) \<subseteq> \<Gamma>s" using Hsubeq t_assign by auto
    next
      show "Z2 \<subseteq> Zl" using t_assign by (auto simp add: Un_ac)
    next
      show "Zr \<union> Z1 \<subseteq> Zl" using t_assign by (auto simp add: Un_ac)
    next
      have "\<forall>x. x is free dollar sc in e2 \<longrightarrow> x \<notin> Environ.dom Z1" 
        and "INV6 (e1:=e2) Zr" by assumption
      thus "INV6 e2 (Zr \<union> Z1)" by (auto simp add: dom_def)
    next
      have "INV9 e1:=e2" by assumption
      thus "INV9 e2" by auto
    next 
    qed auto
    have H5_1: "\<Gamma>2' \<sim> \<Gamma>r" using H5_1_1 env_compat_def by auto
    have "\<Gamma>1 \<sim> \<Gamma>2'" using H5_1_1 env_compat_def by auto moreover
    have "\<Gamma>1 \<union> \<Gamma>2';\<Delta> \<Turnstile> OK" 
      using H6_1 wf_env_subset[of "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1)" \<Delta> "\<Gamma>1 \<union> \<Gamma>2'"]
      by auto 
    moreover
    have  "\<Gamma>1;\<Delta>;{};Z1 \<turnstile> e1 : Qual qU \<pi> (TyRef \<tau>)" using t_assign by auto
    moreover
    have  "\<Gamma>2';\<Delta>;{};Ze' \<turnstile> e2' : \<tau>" using Ht by auto
    moreover
    have "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> dom Z1" and "Z1\<union>Z2 \<subseteq> Zl" and "Zl \<subseteq> Zs" 
      by assumption
    hence "\<forall>\<rho>. \<rho> is free dollar sc in e2' \<longrightarrow> \<rho> \<notin> Environ.dom Z1"
      using eval_preserves_let_dollars[of e2 Z1 Zs e2'] Hg19 by auto
    moreover
    {
      have "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom Z2"
        and "\<forall>\<rho>. \<rho> is free dollar sc in e2 \<longrightarrow> \<rho> \<notin> Environ.dom Z1" 
        by assumption
      moreover
      note Hg19 Hg20
      moreover
      have "Z1\<union>Z2 \<subseteq> Zl" and "Zl \<subseteq> Zs" by assumption
      hence "Z1\<union>Z2 \<subseteq> Zs" by auto
      moreover
      have "\<Gamma>1; \<Delta>; Mz; Z1 \<turnstile> e1 : Qual qU \<pi> (TyRef \<tau>)" by assumption
      moreover
      have "INV9 e1" using t_assign.prems by auto
      ultimately
      have "\<forall>\<rho>. \<rho> is free dollar sc in e1 \<longrightarrow> \<rho> \<notin> Environ.dom Ze'"  
        using eval_preserves_let_dollars_scope[of e1 Z2 e2 Z1 e2' Zs Ze' \<Gamma>1 \<Delta> Mz] by auto
    }
    ultimately
    have Hg3: "\<Gamma>1 \<union> \<Gamma>2';\<Delta>;{};Z1\<union>Ze' \<turnstile> e1 := e2' : Qual qU ScBot TyUnit"
      using Typing.t_assign t_assign by auto
    have Hg4: "\<Gamma>2' \<union> \<Gamma>1 \<union> \<Gamma>r \<subseteq> \<Gamma>s'" using H4_1 by auto
    have "\<Gamma>1 \<sim> \<Gamma>r" using t_assign env_compat_subset[of "\<Gamma>1\<union>\<Gamma>2" \<Gamma>r \<Gamma>1] by auto
    hence Hg5: "(\<Gamma>1 \<union> \<Gamma>2') \<sim> \<Gamma>r"
      using H5_1 env_compat_union_right[of \<Gamma>1 \<Gamma>r \<Gamma>2'] by auto
    have "\<Gamma>2' \<union> (\<Gamma>r\<union>\<Gamma>1) = \<Gamma>1 \<union> \<Gamma>2' \<union> \<Gamma>r" by auto
    hence Hg6: "\<Gamma>1 \<union> \<Gamma>2' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" using H6_1 by auto
    have Hg13: "Z1 \<union> Ze' \<subseteq> Zl'" using Hg15 Hg13 by auto
    have "INV6 (e1:=e2) Zr" by assumption
    hence Hg18: "INV6 (e1:=e2') Zr" using Hg18 by (auto simp add: dom_def)
    have Hg19: "INV7 (e1:=e2) (e1:=e2') Zs" using Hg19 by (auto simp add: dom_def)
    have Hg20: "INV8 (Z1 \<union> Ze') (Z1 \<union> Z2) (e1:=e2')" using Hg20 by (auto simp add: dom_def)
    have "INV9 e1:=e2" by assumption
    hence Hg21: "INV9 e1:=e2'" using Hg21 by auto
    show ?thesis
      apply (rule exI[of _ "\<Gamma>1\<union>\<Gamma>2'"], rule exI[of _ \<Gamma>s'])
      apply (rule exI[of _ \<Gamma>m'], rule exI[of _ M'])
      apply (rule exI[of _ Zs'], rule exI[of _ "Z1 \<union> Ze'"], rule exI[of _ Zl'])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg12 Hg13 Hg14 Hg15 Hg15g Hg16 Hg17 Hg18 Hg19 
        Hg20 Hg21 e_assign_p2 
    by auto
  next
    case (e_assign_c Sa Sa' \<mu>a l q x z \<pi>o)
    have Hz12l: "Z1\<union>Z2 \<subseteq> Zl" by assumption
    have "\<Gamma>1;\<Delta>;Mz;Z1 \<turnstile> e1 : Qual qU \<pi> (TyRef \<tau>)" by assumption
    hence "validScope (scopeT (Qual qU \<pi> (TyRef \<tau>))) Z1" by (rule ty_valid_scope) 
    hence Htvalid: "validScope (scopeT (Qual qU \<pi> (TyRef \<tau>))) Zl" apply (rule valid_scope_subset) 
      using Hz12l by auto
    hence Hvalid: "validScope \<pi> Zl" by auto
    have Hsin: "x\<mapsto>TmLoc q \<pi>o l \<in> S" using e_assign_c by auto
    moreover
    have "x\<rhd>Qual qU \<pi> (TyRef \<tau>) \<in> \<Gamma>1" using t_assign e_assign_c by auto
    hence Hxin: "x\<rhd>Qual qU \<pi> (TyRef \<tau>) \<in> (\<Gamma>s\<union>\<Gamma>m)" using t_assign by auto
    moreover
    have "M;\<Delta>;Zs;Zl \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
    moreover
    hence "scopeV (TmLoc q \<pi>o l) = scopeT (Qual qU \<pi> (TyRef \<tau>))"
      apply (rule storeTyping_scope_sanity) using Hsin Hxin by auto
    hence Hvalidv: "validScope (scopeV (TmLoc q \<pi>o l)) Zl" using Htvalid by auto
    ultimately
    have Hq: "q = qU" and Hp: "\<pi>o = \<pi>" 
      using storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x "TmLoc q \<pi>o l" "Qual qU \<pi> (TyRef \<tau>)"] 
      by auto
    hence Hs: "S' = S" using e_assign_c by auto
    have Hvalid: "validScope \<pi> Zl" using Htvalid by auto
    have "z\<rhd>\<tau> \<in> \<Gamma>2" using t_assign e_assign_c by auto
    hence Hin: "z\<rhd>\<tau> \<in> \<Gamma>s" using t_assign by auto
    have "M;\<Delta>;Zs;Zl \<Turnstile> S':(\<Gamma>s \<union> \<Gamma>m)" using t_assign Hs by auto
    also have "\<Gamma>s \<union> \<Gamma>m = \<Gamma>s \<union> \<Gamma>m,(z:\<tau>)" using Hin by auto
    finally
    have Hg1: "M;\<Delta>;Zs;Zl \<Turnstile> S':(\<Gamma>s \<union> \<Gamma>m,(z:\<tau>))" using t_assign by auto
    moreover
    have "M;\<Delta>;Zs;Zl \<Turnstile> S : (\<Gamma>s\<union>\<Gamma>m)" by assumption
    ultimately
    obtain \<Gamma>v Znew where "\<Gamma>v;\<Delta>;M;Znew \<turnstile> TmLoc qU \<pi>o l : Qual qU \<pi> (TyRef \<tau>)"
      using Hq Hxin Hsin Hvalidv
        storeTyping_sanity[of M \<Delta> Zs Zl S "\<Gamma>s\<union>\<Gamma>m" x "TmLoc qU \<pi>o l" "Qual qU \<pi> (TyRef \<tau>)"]
      by auto
    hence "l\<rhd>\<tau> \<in> M" by auto
    moreover
    hence "\<mu> defines l" using prems mem_dom_in[of \<Gamma>m \<Delta> \<mu> M l] by (auto simp add: dom_def)
    moreover
    have "\<Gamma>m,(z:\<tau>);\<Delta> \<Turnstile> OK" using e_assign_c t_assign wf_env_union_sanity_cons by auto
    moreover
    have "\<Gamma>m;\<Delta> \<Turnstile>' \<mu> : M" by assumption
    moreover
    hence "\<Gamma>m \<sim> {z\<rhd>\<tau>}" by (auto simp add: env_compat_def)
    ultimately
    have Hg2: "\<Gamma>m,(z:\<tau>);\<Delta> \<Turnstile>' update \<mu> l (TmFreeVar z) : M"
      using memTyping_update by auto
    have  H\<Delta>: "\<Delta> \<Turnstile>' TY" using t_assign poly_wf_env_ty by auto
    have "{}\<union>{} \<Turnstile> Qual qU ScBot TyUnit" apply (rule ty_qual) 
      by (auto simp add: minimalValidScope_def)
    hence Hg3: "{};\<Delta>;{};{} \<turnstile> TmUnit qU ScBot : Qual qU ScBot TyUnit" using H\<Delta>
      by (auto simp add: wf_env_def qualGrelL_def minimalValidScope_def wf_env_ty_def)
    have Hg4: "{} \<union> \<Gamma>r \<subseteq> \<Gamma>s" using prems by auto
    have Hg5: "{} \<sim> \<Gamma>r" by (auto simp add: env_compat_def)
    hence Hg6: "{} \<union> \<Gamma>r;\<Delta> \<Turnstile> OK"  using t_assign wf_env_subset[of "\<Gamma>1\<union>\<Gamma>2\<union>\<Gamma>r" \<Delta> \<Gamma>r] by auto
    have "\<Gamma>s \<sim> \<Gamma>m" and "qualT \<tau> \<preceq> qU" by assumption
    hence Hg7: "\<Gamma>s \<sim> \<Gamma>m,(z:\<tau>)" by (auto simp add: env_compat_def)
    have Hg10: "INV2 (filterStore S' Zl)" and Hg16: "INV4 (filterStore S' Zl)" 
      using Hs t_assign by auto
    have Hg11: "INV3 S'" and  Hg17: "INV5 S'" using Hs t_assign by auto
    have Hg13: "Zl \<subseteq> Zs" and Hg14: "Zr \<subseteq> Zl"  by assumption
    have Hg15: "free_scope_store S' = dom Zs" using Hs t_assign by auto
    have Hg18: "INV6 (TmUnit qU ScBot) Zr" by auto
    have Hg19: "INV7 (TmFreeVar x:=TmFreeVar y) (TmUnit qU ScBot) Zs" by auto
    have Hg20: "INV8 {} (Z1\<union>Z2) (TmUnit qU ScBot)" by (auto simp add: dom_def)
    have Hg21: "INV9 (TmUnit qU ScBot)" by auto
    show ?thesis
      apply (rule exI[of _ "{}"], rule exI[of _ \<Gamma>s])
      apply (rule exI[of _ "\<Gamma>m,(z:\<tau>)"], rule exI[of _ M],
        rule exI[of _ Zs], rule exI[of _ "{}"], rule exI[of _ Zl])
      using e_assign_c Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg13 Hg14 Hg15 Hg16 Hg17 Hg18 Hg19
        Hg20 Hg21 Hs by auto
  qed auto
next;

      (********************* new cases *********************)
  case (t_new \<Delta> Z Mz \<Gamma> \<tau> e e' S S' \<mu> \<mu>' \<Gamma>s \<Gamma>r \<Gamma>m M)
  have "S;\<mu>;TmNew e \<hookrightarrow> S';\<mu>';e'" .
  thus ?case proof cases
    case (e_new_p Sa Sa' \<mu>a \<mu>a' ea e')
    from t_new.hyps
    have IH: "\<And>M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r.
          \<lbrakk> Mz = {}; M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m); \<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M;  S;\<mu>;e\<hookrightarrow>S';\<mu>';e';
           \<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s; \<Gamma> \<sim> \<Gamma>r; \<Gamma> \<union> \<Gamma>r;\<Delta> \<Turnstile> OK; \<Gamma>s \<sim> \<Gamma>m; 
           Z \<subseteq> Zl; Zr \<subseteq> Zl; Zl \<subseteq> Zs; free_scope_store S = dom Zs; 
           INV2 (filterStore S Zl); INV3 S; INV4 (filterStore S Zl); INV5 S;
           INV6 e Zr; INV9 e; \<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK
          \<rbrakk> \<Longrightarrow>
          \<exists>\<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'.
            (M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')) \<and>
            (\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M') \<and>
            (\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e' : \<tau>) \<and>
            (\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s') \<and>
            (\<Gamma>e' \<sim> \<Gamma>r) \<and> (\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK) \<and> (\<Gamma>s' \<sim> \<Gamma>m') \<and>
            (Ze' \<subseteq> Zl') \<and> (Zl' \<subseteq> Zs') \<and> (Zs \<subseteq> Zs') \<and> (Zr \<subseteq> Zl') \<and>
            (free_scope_store S' = dom Zs') \<and> (INV2 (filterStore S' Zl')) \<and>
            (INV3 S') \<and> (INV4 (filterStore S' Zl')) \<and> (INV5 S') \<and> (INV6 e' Zr) \<and>
            (INV7 e e' Zs) \<and> (INV8 Ze' Z e') \<and> (INV9 e')"
      by auto
    obtain \<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'
      where Hg1: "M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
      and Hg2: "\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M'"
      and Ht: "\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e' : \<tau>"
      and Hg4: "\<Gamma>e' \<union> \<Gamma>r \<subseteq> \<Gamma>s'"
      and Hg5: "\<Gamma>e' \<sim> \<Gamma>r"
      and Hg6: "\<Gamma>e' \<union> \<Gamma>r;\<Delta> \<Turnstile> OK"
      and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
      and Hg12: "Ze' \<subseteq> Zl'"
      and Hg15: "Zl' \<subseteq> Zs'"
      and Hg13: "Zs \<subseteq> Zs'"
      and Hg16: "Zr \<subseteq> Zl'"
      and Hg14: "free_scope_store S' = dom Zs'"
      and Hg10: "INV2 (filterStore S' Zl')"
      and Hg11: "INV3 S'"
      and Hg17: "INV4 (filterStore S' Zl')"
      and Hg18: "INV5 S'"
      and Hg19: "INV6 e' Zr"
      and Hg20: "INV7 e e' Zs"
      and Hg21: "INV8 Ze' Z e'"
      and Hg22: "INV9 e'"
      using IH[of M S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' \<Gamma>r] proof rule 
      show "Mz = {}" and "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M" 
        and "\<Gamma> \<union> \<Gamma>r \<subseteq> \<Gamma>s" and "\<Gamma> \<sim> \<Gamma>r" and "\<Gamma> \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" and "\<Gamma>s \<sim> \<Gamma>m"
        and "\<Gamma>s \<union> \<Gamma>m;\<Delta> \<Turnstile> OK" and "INV2 (filterStore S Zl)" and "INV3 S"
        and "INV4 (filterStore S Zl)" and "INV5 S" and "Zr \<subseteq> Zl"
        and "Zl \<subseteq> Zs" and "free_scope_store S = dom Zs" and "Z \<subseteq> Zl" by assumption
    next
      show "S;\<mu>;e\<hookrightarrow>S';\<mu>';e'" using e_new_p by auto
    next
      have "INV6 (TmNew e) Zr" by assumption
      thus "INV6 e Zr" by auto
    next
      have "INV9 (TmNew e)" by assumption
      thus "INV9 e" by auto
    qed auto
    moreover
    have Hg3: "\<Gamma>e';\<Delta>;{};Ze' \<turnstile> TmNew e' : Qual qL ScBot (TyRef \<tau>)" using Ht by auto
    moreover
    have Hg19: "INV6 (TmNew e') Zr" using Hg19 by auto
    moreover
    have Hg22: "INV9 (TmNew e')" using Hg22 by auto
    show ?thesis
      apply (rule exI[of _ \<Gamma>e'], rule exI[of _ \<Gamma>s'], rule exI[of _ \<Gamma>m'], rule exI[of _ M'],
        rule exI[of _ Zs'], rule exI[of _ Ze'], rule exI[of _ Zl'])
      using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg12 Hg13 Hg14 Hg15 Hg16 Hg17 Hg18 Hg19 Hg20 Hg21
        Hg22 e_new_p by auto
  next
    case (e_new_c Sa \<mu>a l x z)
    have Hm: "M;\<Delta> \<Turnstile> OK" using prems memTyping_wf_loc by auto
    moreover
    have Hlnotinm: "\<not> M defines l" using prems mem_dom_not_in by auto
    moreover
    have H\<tau>: "\<Delta> \<Turnstile>' \<tau>" using t_new poly_ty_wf_type by auto
    ultimately
    have Hmok: "M,(l:\<tau>);\<Delta> \<Turnstile> OK" using e_new_c wf_env_cons by auto
    have Hlfresh: "\<not> l free loc in st (filterStore S Zl)" apply (rule storeTyping_no_loc)
      using Hlnotinm t_new.prems by auto
    have Hg1: "M,(l:\<tau>);\<Delta>;Zs;Zl \<Turnstile> S,(z\<mapsto>TmLoc qL ScBot l) : 
                (\<Gamma>s \<union> \<Gamma>m),(z:Qual qL ScBot (TyRef \<tau>))" 
      apply (rule st_cons[of "{}" \<Delta> "M,(l:\<tau>)" "{}" "TmLoc qL ScBot l" 
        "Qual qL ScBot (TyRef \<tau>)" Zs Zl S "\<Gamma>s\<union>\<Gamma>m" z])
    proof -
      show "{};\<Delta>;M,(l:\<tau>);{} \<turnstile> TmLoc qL ScBot l : Qual qL ScBot (TyRef \<tau>)"
        using Hmok by (auto simp add: wf_env_def qualGrelL_def minimalValidScope_def)
    next
      have "M;\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m)" by assumption 
      thus "M,(l:\<tau>);\<Delta>;Zs;Zl \<Turnstile> S:(\<Gamma>s \<union> \<Gamma>m \<union> {})" using storeTyping_weakening Hmok by auto
    next
      show "\<Gamma>s \<union> \<Gamma>m \<union> {};\<Delta> \<Turnstile> OK" using storeTyping_wf_env by auto
    next
      show "\<not> S defines z" using prems by auto
    next 
      show "S \<Turnstile> Store" using t_new by (auto simp add: storeTyping_wf_store)
    qed (auto simp add: un_def env_compat_def storeTyping_wf_store st_cons)
    have "\<Delta> \<Turnstile>' TY" using t_new poly_env_wf_env_ty by auto
    hence Hg3: "{z\<rhd>Qual qL ScBot (TyRef \<tau>)};\<Delta>;{};{} \<turnstile> TmFreeVar z : Qual qL ScBot (TyRef \<tau>)"
      using H\<tau> by (auto simp add: without_def wf_env_def qualGrelL_def minimalValidScope_def)
    have Hsmnotin: "\<not> (\<Gamma>s\<union>\<Gamma>m) defines z" using prems store_dom_not_in by auto
    hence Hnotin: "\<not> \<Gamma>r defines z" using prems by (auto simp add: dom_def)
    hence Hg5: "{z\<rhd>Qual qL ScBot (TyRef \<tau>)} \<sim> \<Gamma>r"
      by (auto simp add: env_compat_def dom_def)
    have H\<tau>_ref:  "\<Delta> \<Turnstile>' Qual qL ScBot (TyRef \<tau>)" using H\<tau> by auto
    have "\<Gamma>r \<subseteq> \<Gamma>s\<union>\<Gamma>m" using prems by auto
    hence "\<Gamma>r;\<Delta> \<Turnstile> OK" using wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Delta> \<Gamma>r] prems by auto
    hence Hg6: "{z\<rhd>Qual qL ScBot (TyRef \<tau>)} \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" 
      using Hnotin H\<tau>_ref wf_env_cons[of \<Gamma>r _ z] by auto
    have Hs_sub: "filterStore S Zl \<subseteq> S" by (auto simp add: filterStore_def)
    have "\<not> S defines z" using prems by auto
    hence "\<not> (filterStore S Zl) defines z" using Hs_sub by (auto simp add: dom_def)
    hence "INV2 ((filterStore S Zl),(z\<mapsto>TmLoc qL ScBot l))" 
      using Hlfresh inv2_cons_loc[of "filterStore S Zl" z l] t_new.prems by blast
    also
    have Hinv2inv4_2:
      "((filterStore S Zl),(z\<mapsto>TmLoc qL ScBot l)) = (filterStore (S,(z\<mapsto>TmLoc qL ScBot l)) Zl)" 
      by (auto simp add: filterStore_def)
    finally
    have Hg10: "INV2 (filterStore (S,(z\<mapsto>TmLoc qL ScBot l)) Zl)" by auto
    have "INV3 S" by assumption
    hence Hg11: "INV3 (S,(z\<mapsto>TmLoc qL ScBot l))" by auto
    have "\<Gamma>;\<Delta>;{};Z \<turnstile> TmFreeVar x : \<tau>" using prems by auto
    hence Hxin: "(x\<rhd>\<tau>) \<in> \<Gamma>" by auto
    hence "{x\<rhd>\<tau>} \<subseteq> \<Gamma>" by auto
    hence Hxsub: "{x\<rhd>\<tau>} \<subseteq> \<Gamma>s" using t_new by auto
    have Hsmcomp: "\<Gamma>s \<sim> \<Gamma>m" and Hsmok: "\<Gamma>s \<union>  \<Gamma>m;\<Delta> \<Turnstile> OK" by assumption
    have Hg13: "{} \<subseteq> Zs" by auto
    have "free_scope_store (S,(z\<mapsto>TmLoc qL ScBot l)) = free_scope_store S"
      using free_scope_store_cons_empty by auto
    hence Hg14: "free_scope_store (S,(z\<mapsto>TmLoc qL ScBot l)) = Environ.dom Zs" 
      using t_new by auto
    have Hg15: "Zl \<subseteq> Zs"
      and Hg16: "Zr \<subseteq> Zl" by assumption
    have Hg17: "{} \<subseteq> Zl" by auto
    have "INV4 (filterStore S Zl)" by assumption
    hence "INV4 ((filterStore S Zl),(z\<mapsto>TmLoc qL ScBot l))"
      using Hlfresh inv4_cons_loc by auto
    hence Hg18: "INV4 (filterStore (S,(z\<mapsto>TmLoc qL ScBot l)) Zl)" using Hinv2inv4_2 by auto
    have "INV5 S" by assumption
    hence Hg19: "INV5 (S,(z\<mapsto>TmLoc qL ScBot l))" by (auto simp add: INV5)
    have Hg20: "INV6 (TmFreeVar z) Zr" by auto
    have Hg21: "INV7 (TmNew e) (TmFreeVar z) Zs" by auto
    have Hg22: "INV8 {} Z (TmFreeVar z)" by (auto simp add: dom_def)
    have Hg23: "INV9 (TmFreeVar z)" by auto
    show ?thesis proof (cases "qualT \<tau>")
      case qL
      note Hg1
      also have "\<Gamma>s \<union> \<Gamma>m = (\<Gamma>s - {x\<rhd>\<tau>}) \<union> (\<Gamma>m,(x:\<tau>))" using Hxsub by auto
      also have "(((\<Gamma>s - {x\<rhd>\<tau>})) \<union> (\<Gamma>m,(x:\<tau>))),(z:Qual qL ScBot (TyRef \<tau>)) =
                (((\<Gamma>s - {x\<rhd>\<tau>}),(z:Qual qL ScBot (TyRef \<tau>))) \<union> (\<Gamma>m,(x:\<tau>)))" by auto
      finally have Hg1: "M,(l:\<tau>);\<Delta>;Zs;Zl \<Turnstile> S,(z\<mapsto>TmLoc qL ScBot l) :
        (((\<Gamma>s - {x\<rhd>\<tau>}),(z:Qual qL ScBot (TyRef \<tau>))) \<union> (\<Gamma>m,(x:\<tau>)))"
        by auto
      have Hmx: "\<not> \<Gamma>m defines x"
        using qL Hxsub Hsmcomp Hsmok compat_ok_sanity[of \<Gamma>s \<Gamma>m _ x \<tau>] by auto
      have Hg2: "\<Gamma>m,(x:\<tau>);\<Delta> \<Turnstile>' \<mu>,(l\<mapsto>TmFreeVar x):M,(l:\<tau>)" proof
        show "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M" using t_new by auto
      next
        show "\<Gamma>m \<sim> {(x, Term \<tau>)}"
          using Hmx by (auto simp add: env_compat_def dom_def)
      next
        have "\<Gamma>m;\<Delta> \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Delta> \<Gamma>m] by auto
        thus "\<Gamma>m,(x:\<tau>);\<Delta> \<Turnstile> OK" using H\<tau> Hmx wf_env_cons by auto
      next
        show "l \<notin> Environ.dom \<mu>" using prems by auto
      next
        show "M;\<Delta> \<Turnstile> OK" using Hm by auto
      next
        show "\<mu> \<Turnstile> Mem" using prems memTyping_wf_mem by auto
      qed
      have "\<Gamma> \<sim> \<Gamma>r" and "\<Gamma> \<union> \<Gamma>r;\<Delta> \<Turnstile> OK" by assumption
      hence "\<not> \<Gamma>r defines x"
        using qL Hxin compat_ok_sanity[of \<Gamma> \<Gamma>r _ x \<tau>] by auto
      moreover
      have "\<Gamma>r \<subseteq> \<Gamma>s" using prems by auto
      ultimately
      have "\<Gamma>r \<subseteq> \<Gamma>s - {x\<rhd>\<tau>}" by (auto simp add: dom_def)
      hence Hg4: "{z\<rhd>Qual qL ScBot (TyRef \<tau>)} \<union> \<Gamma>r \<subseteq>
                 (\<Gamma>s - {x\<rhd>\<tau>}),(z:Qual qL ScBot (TyRef \<tau>))" by auto
      have "\<Gamma>s \<sim> \<Gamma>m" by assumption
      hence Hg7: "((\<Gamma>s - {x\<rhd>\<tau>}),(z:Qual qL ScBot (TyRef \<tau>))) \<sim> (\<Gamma>m,(x:\<tau>))"
      using Hsmnotin proof (auto simp add: env_compat_def dom_def)
        assume "Qual qL ScBot (TyRef \<tau>) = \<tau>"
        hence "size (Qual qL ScBot (TyRef \<tau>)) = size \<tau>" by auto
        thus "qualT \<tau> \<preceq> qU" by auto
      qed
      show ?thesis
        apply (rule exI[of _ "{z\<rhd>Qual qL ScBot (TyRef \<tau>)}"],
               rule exI[of _ "(\<Gamma>s - {x\<rhd>\<tau>}),(z:Qual qL ScBot (TyRef \<tau>))"],
               rule exI[of _ "\<Gamma>m,(x:\<tau>)"], rule exI[of _ "M,(l:\<tau>)"],
               rule exI[of _ Zs], rule exI[of _ "{}"],
               rule exI[of _ Zl])
        using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg13 Hg14 Hg15 Hg16 Hg17 Hg18 Hg19 Hg20 
          Hg21 Hg22 Hg23 e_new_c by auto
    next
      case qU
      note Hg1
      also have Hueq: "\<Gamma>s \<union> \<Gamma>m = \<Gamma>s \<union> (\<Gamma>m,(x:\<tau>))" using Hxsub by auto
      also have "(\<Gamma>s \<union> (\<Gamma>m,(x:\<tau>))),(z:Qual qL ScBot (TyRef \<tau>)) =
                \<Gamma>s,(z:Qual qL ScBot (TyRef \<tau>)) \<union> (\<Gamma>m,(x:\<tau>))" by auto
      finally have Hg1: "M,(l:\<tau>);\<Delta>;Zs;Zl \<Turnstile> S,(z\<mapsto>TmLoc qL ScBot l) :
        (\<Gamma>s,(z:Qual qL ScBot (TyRef \<tau>)) \<union> (\<Gamma>m,(x:\<tau>)))"
        by auto
      have Hg2: "\<Gamma>m,(x:\<tau>);\<Delta> \<Turnstile>' \<mu>,(l\<mapsto>TmFreeVar x):M,(l:\<tau>)" proof
        show "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M" using t_new by auto
      next
        show "\<Gamma>m \<sim> {(x, Term \<tau>)}"
          using qU by (auto simp add: env_compat_def dom_def)
      next
        have Hgmok: "\<Gamma>m;\<Delta> \<Turnstile> OK" using prems wf_env_subset[of "\<Gamma>s\<union>\<Gamma>m" \<Delta> \<Gamma>m] by auto
        show "\<Gamma>m,(x:\<tau>);\<Delta> \<Turnstile> OK" proof (cases "\<Gamma>m defines x")
          case True
          have "x\<rhd>\<tau> \<in> \<Gamma>s" using Hxsub by auto
          moreover
          have "\<Gamma>m \<union> \<Gamma>s;\<Delta> \<Turnstile> OK" using Hsmok by (auto simp add: wf_env_def)
          ultimately
          have "x\<rhd>\<tau> \<in> \<Gamma>m" using wf_env_union_sanity[of \<Gamma>m \<Gamma>s _ x \<tau>]
            using True by auto
          hence "\<Gamma>m,(x:\<tau>) = \<Gamma>m" by auto
          thus ?thesis using Hgmok by auto
        next
          case False
          thus ?thesis using Hgmok wf_env_cons H\<tau> by auto
        qed
      next
        show "l \<notin> Environ.dom \<mu>" using prems by auto
      next
        show "M;\<Delta> \<Turnstile> OK" using Hm by auto
      next
        show "\<mu> \<Turnstile> Mem" using prems memTyping_wf_mem by auto
      qed
      have "\<Gamma>r \<subseteq> \<Gamma>s" using prems by auto
      hence Hg4: "{(z, Term (Qual qL ScBot (TyRef \<tau>)))} \<union> \<Gamma>r \<subseteq>
                 \<Gamma>s,(z:Qual qL ScBot (TyRef \<tau>))" by auto
      have Hg7: "(\<Gamma>s,(z:Qual qL ScBot (TyRef \<tau>))) \<sim> (\<Gamma>m,(x:\<tau>))"
      proof (auto simp add: env_compat_def)
        show  "qualT \<tau> \<preceq> qU" using prems by auto
      next
        have "\<not> \<Gamma>m defines z" using Hsmnotin by (auto simp add: dom_def)
        thus "(z\<rhd>(Qual qL ScBot (TyRef \<tau>))) \<in> \<Gamma>m \<Longrightarrow> False"
          using Hnotin by (auto simp add: dom_def)
      next
        show "qualT \<tau> \<preceq> qU" using prems by auto
      next
        show "\<And>x \<tau>. \<lbrakk>(x\<rhd>\<tau>) \<in> \<Gamma>s; (x\<rhd>\<tau>) \<in> \<Gamma>m\<rbrakk> \<Longrightarrow> qualT \<tau> \<preceq> qU"
          using Hsmcomp by (auto simp add: env_compat_def)
      qed
      show ?thesis
        apply (rule exI[of _ "{z\<rhd>Qual qL ScBot (TyRef \<tau>)}"],
               rule exI[of _ "\<Gamma>s,(z:Qual qL ScBot (TyRef \<tau>))"],
               rule exI[of _ "\<Gamma>m,(x:\<tau>)"], rule exI[of _ "M,(l:\<tau>)"],
               rule exI[of _ Zs], rule exI[of _ "{}"], rule exI[of _ Zl])
        using Hg1 Hg2 Hg3 Hg4 Hg5 Hg6 Hg7 Hg10 Hg11 Hg13 Hg14 Hg15 Hg16 Hg17 Hg18 Hg19 Hg20 
          Hg21 Hg22 Hg23 e_new_c by auto
    qed
  qed auto
next;

qed
qed

theorem safety:
  assumes Hsem: "S;\<mu>;e \<hookrightarrow>* S';\<mu>';e'"
  and Ht: "\<Gamma>e;\<Delta>;{};Ze \<turnstile> e : \<tau>"
  and Hst: "M;\<Delta>;Zs;Zl \<Turnstile> S : \<Gamma>s \<union> \<Gamma>m"
  and H2: "\<Gamma>m;\<Delta> \<Turnstile>' \<mu> : M"
  and H3: "\<Gamma>e \<subseteq> \<Gamma>s"
  and H4: "\<Gamma>s \<sim> \<Gamma>m"
  and H7: "Ze \<subseteq> Zl" 
  and H9: "Zl \<subseteq> Zs"
  and Hscopes: "free_scope_store S = dom Zs"
  and Hinv2: "INV2 (filterStore S Zl)"
  and Hinv3: "INV3 S"
  and Hinv4: "INV4 (filterStore S Zl)" 
  and Hinv5: "INV5 S" 
  and Hinv7: "INV9 e"
  shows "not_stuck e' S' \<mu>'"
using prems proof (induct arbitrary: \<Gamma>e \<Gamma>s \<Gamma>m M Zs Ze Zl)
  case (e_refl S \<mu> e \<Gamma>e \<Gamma>s \<Gamma>m M Zs Ze Zl) thus ?case using progress by auto
next
  case (e_trans S \<mu> e S' \<mu>' e' S'' \<mu>'' e'' \<Gamma>e \<Gamma>s \<Gamma>m M Zs Ze Zl)
  obtain \<Gamma>e' \<Gamma>s' \<Gamma>m' M' Zs' Ze' Zl'
    where Hg1: "M';\<Delta>;Zs';Zl' \<Turnstile> S':(\<Gamma>s' \<union> \<Gamma>m')"
    and Hg2: "\<Gamma>m';\<Delta> \<Turnstile>' \<mu>':M'"
    and Ht: "\<Gamma>e';\<Delta>;{};Ze' \<turnstile> e' : \<tau>"
    and Hg4: "\<Gamma>e' \<union> {} \<subseteq> \<Gamma>s'"
    and Hg5: "\<Gamma>e' \<sim> {}"
    and Hg6: "\<Gamma>e' \<union> {};\<Delta> \<Turnstile> OK"
    and Hg7: "\<Gamma>s' \<sim> \<Gamma>m'"
    and Hg12: "Ze' \<subseteq> Zl'"
    and Hg15: "Zl' \<subseteq> Zs'"
    and Hg13: "Zs \<subseteq> Zs'"
    and Hg16: "{} \<subseteq> Zl'"
    and Hg14: "free_scope_store S' = dom Zs'"
    and Hg10: "INV2 (filterStore S' Zl')"
    and Hg11: "INV3 S'"
    and Hg17: "INV4 (filterStore S' Zl')"
    and Hg18: "INV5 S'"
    and Hg19: "INV6 e' {}"
    and Hg20: "INV7 e e' Zs"
    and Hg21: "INV8 Ze' Ze e'"
    and Hg22: "INV9 e'"
    using preservation[of \<Gamma>e \<Delta> "{}" Ze e \<tau> M Zs Zl S \<Gamma>s \<Gamma>m \<mu> S' \<mu>' e' "{}" "{}"]  proof rule
    show "\<Gamma>e;\<Delta>;{};Ze \<turnstile> e : \<tau>" by assumption
  next
    show "{} = {}" by auto
  next
    show "M;\<Delta>;Zs;Zl \<Turnstile> S:\<Gamma>s\<union>\<Gamma>m" and "\<Gamma>m;\<Delta> \<Turnstile>' \<mu>:M" and "S;\<mu>;e\<hookrightarrow>S';\<mu>';e'" by assumption
  next
    show "\<Gamma>e \<union> {} \<subseteq> \<Gamma>s" and "\<Gamma>e \<sim> {}" and "\<Gamma>e \<union> {};\<Delta> \<Turnstile> OK" 
      using e_trans by (auto simp add: env_compat_def ty_wf_env)
  next
    show "\<Gamma>s \<sim> \<Gamma>m" and "Ze \<subseteq> Zl" by assumption 
  next
    show "{} \<subseteq> Zl" by auto 
  next
    show "Zl \<subseteq> Zs" and "free_scope_store S = Environ.dom Zs" and "INV2 (filterStore S Zl)"
    and "INV3 S" and "INV4 (filterStore S Zl)" and "INV5 S" 
      by assumption
  next
    show "INV6 e {}" by (auto simp add: dom_def)
  next
    show "INV9 e" by assumption
  qed auto
  thus ?case using e_trans.hyps by (auto simp del: INV6 INV5 INV4 INV3 INV2)
qed

theorem program_safety:
  assumes Hsem: "{};{};e \<hookrightarrow>* S';\<mu>';e'"
  and Htyp: "{};{};{};{} \<turnstile> e : \<tau>"
  and Hsurface: "\<not> hasDollars e"
  shows "not_stuck e' S' \<mu>'"
using Hsem Htyp apply (rule safety[of "{}" "{}" _ _ _ _ _ "{}" "{}" _ "{}" "{}" "{}" "{}" "{}"])
proof (auto simp add: wf_env_def env_compat_def dom_def free_scope_store_def wf_env_ty_def
    filterStore_def INV4 INV5 INV10)
  show "INV9 e" using Hsurface INV9_intro by auto
qed

end
